/*
 * PDFlow -- conv.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <assert.h>
#include <string.h>

#include "grid/grid.h"
#include "grid/layout.h"
#include "grid/lc.h"
#include "app/conv.h"

typedef enum s_status (*page_transl) (struct conv *conv, 
				      struct gspg_page *spdfp,
				      struct gr_grid *grid,
				      struct gspg_page **dpdfp,
				      struct ref_owner *p);

struct float_closure {
	struct gspg_page *pdfp;
	struct conv *conv;
};

struct resize_closure {
	struct conv *conv;
	struct pg_pages *dpages;
	struct bt_cursor *dpc;
};

static void destroy_src (struct conv *conv)
{
	struct conv_sstate *sstate;

	sstate = &conv -> sstate;

	gf_cache_destroy (sstate -> doc -> rdb);

	pd_unref (sstate -> doc, &conv -> owner);
	if (sstate -> pi != NULL)
		pdfi_destroy (sstate -> pi);
}

static void destroy_dest (struct conv *conv, int retain)
{
	struct conv_dstate *dstate;

	dstate = &conv -> dstate;

	agc_unref (dstate -> agc, &conv -> owner);	

	if (retain)
		pd_unref (dstate -> doc, &conv -> owner);
	else if (dstate -> doc != NULL)
		pd_unlink (dstate -> doc, &conv -> owner);

	v_destroy_selection (&dstate -> sel);
}

static void conv_destroy (struct conv *conv, int retain)
{
	destroy_src (conv);
	destroy_dest (conv, retain);

	if (conv -> pdc != NULL)
		pdfc_destroy (conv -> pdc);
	ob_unref (conv -> ob, &conv -> owner);

	free (conv);
}

static enum s_status init_src (struct conv *conv, const char *fname,
			       struct args *args)
{
	struct conv_sstate *sstate;
	enum s_status res;

	sstate = &conv -> sstate;

	sstate -> heur = args -> heur;
	sstate -> amys = args -> amys;
	sstate -> ranges = args -> ranges;

	res = pd_init (&sstate -> doc, &conv -> owner);
	if (res != S_CSC_OK)
		return res;

	res = pd_read (sstate -> doc, fname);
	if (res != S_CSC_OK)
		return res;

	res = pdfi_create (sstate -> doc -> rdb, &sstate -> pi);
	if (res != S_CSC_OK)
		return res;

	gf_cache_init (sstate -> doc -> rdb);

	return S_CSC_OK;
}

static enum s_status init_dest (struct conv *conv, const char *fname,
				struct args *args)
{
	struct conv_dstate *dstate;
	enum s_status res;

	dstate = &conv -> dstate;

	dstate -> aout = args -> aout;

	dstate -> bounds.bottom = 0;
	dstate -> bounds.left = 0;
	dstate -> bounds.top = args -> aout.height;
	dstate -> bounds.right = args -> aout.width;

	res = pd_init (&dstate -> doc, &conv -> owner);
	if (res != S_CSC_OK)
		return res;

	memset (&dstate -> sel, 0, sizeof (struct v_selection));
	res = pd_add_selection (&dstate -> sel, &args -> prefs);
	if (res != S_CSC_OK)
		return res;

	res = pd_creat (dstate -> doc, fname, &dstate -> sel);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status init (struct conv *conv, const char *sfname,
			   const char *dfname, struct args *args)
{
	enum s_status res;

	res = init_src (conv, sfname, args);
	if (res != S_CSC_OK)
		return res;

	res = init_dest (conv, dfname, args);
	if (res != S_CSC_OK)
		return res;

	res = ob_new (&conv -> ob, conv -> sstate.doc -> rdb, 
		      conv -> dstate.doc -> rdb, &conv -> owner);
	if (res != S_CSC_OK)
		return res;

	res = pdfc_create (conv -> dstate.doc -> trailer -> root, conv -> ob,
			   &conv -> dstate.sel, &conv -> pdc);
	if (res != S_CSC_OK)
		return res;

	conv -> rfp = args -> debug.rfp;

	return S_CSC_OK;
}

static enum s_status conv_init (struct conv **convp, const char *sfname,
				const char *dfname, struct args *args)
{
	struct conv *conv;
	enum s_status res;

	conv = malloc (sizeof (struct conv));
	if (conv == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (conv, 0, sizeof (struct conv));

	ref_init_owner (&conv -> owner, "conv");

	res = init (conv, sfname, dfname, args);
	if (res != S_CSC_OK) {
		conv_destroy (conv, 0);
		return res;
	}

	*convp = conv;

	return S_CSC_OK;
}

static enum s_status pull_page (struct conv *conv, struct args_range *range,
				const char *fmt, struct gspg_page **spdfpp, 
				struct ref_owner *p)
{
	struct gspg_page *pdfp;
	struct conv_sstate *st;
	struct cat_catalog *cat;
	struct pg_page *spage;
	pdf_bool_t eof;
	enum s_status res;

	st = &conv -> sstate;

	if (range == NULL)
		range = &conv -> sstate.ranges.doc;

	spage = NULL;
	eof = conv -> sstate.eof;
	while (!conv -> sstate.eof) {
		if (range -> hi >= 0 && st -> page > range -> hi) {
			eof = conv -> sstate.eof = 1;
			break;
		}
			
		spage = pg_get (st -> spc);
		conv -> sstate.eof = !bt_cursor_next (st -> spc);
		if (st -> page++ >= range -> lo)
			break;
		eof = conv -> sstate.eof;
	} 

	if (eof) {
		*spdfpp = NULL;
		return S_CSC_OK;
	}
	cat = st -> doc -> trailer -> root;

	res = pdfi_eval (st -> pi, cat, spage, &pdfp, p);
	if (res != S_CSC_OK)
		return res;

	pdfp -> number = st -> page;
	printf (fmt == NULL ? "[%d]" : fmt, st -> page);
	fflush (stdout);

	*spdfpp = pdfp;

	return S_CSC_OK;
}

static enum s_status noop_transl (struct conv *conv, struct gspg_page *spdfp,
				  struct gr_grid *grid,
				  struct gspg_page **dpdfp, 
				  struct ref_owner *p)
{
	enum s_status res;

	agc_unref (conv -> dstate.agc, &conv -> owner);	

	res = agc_new (&conv -> dstate.agc, conv -> dstate.doc -> rdb, 
		       spdfp, &conv -> owner);
	if (res != S_CSC_OK)
		return res;
	*dpdfp = spdfp;
	gspg_page_ref (*dpdfp, p);


	return S_CSC_OK;
}

static enum s_status float_grid_f (struct gr_grid *grid, void *closure)
{
	struct float_closure *fc;
	enum s_status res;

	if (!gr_is_word (grid))
		return gr_traverse (grid, float_grid_f, closure);

	fc = closure;

	res = a_add_statement (grid -> d.area, fc -> conv -> dstate.agc);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status float_transl (struct conv *conv, struct gspg_page *spdfp,
				   struct gr_grid *grid,
				   struct gspg_page **dpdfpp, 
				   struct ref_owner *p)
{
	struct float_closure fc;
	struct gspg_page *dpdfp;
	enum s_status res;
	
	res = gspg_page_new (&dpdfp, &spdfp -> bounds, 
			     spdfp -> rotate, p);
	if (res != S_CSC_OK)
		return res;

	agc_unref (conv -> dstate.agc, &conv -> owner);	

	res = agc_new (&conv -> dstate.agc, conv -> dstate.doc -> rdb, 
		       dpdfp, &conv -> owner);
	if (res != S_CSC_OK) {
		gspg_page_unref (dpdfp, p);
		return res;
	}

	fc.pdfp = dpdfp;
	fc.conv = conv;
	res = gr_traverse (grid, float_grid_f, &fc);
	if (res != S_CSC_OK) {
		gspg_page_unref (dpdfp, p);
		return res;
	}

	*dpdfpp = dpdfp;

	return S_CSC_OK;
}

static enum s_status draw_margins (struct agc_context *agc, 
				   struct args_margins *amys,
				   struct cobj_rect *pgrect)
{
	struct args_margin *amy;
	struct cobj_rect rect;
	enum s_status res;
	int i;

	res = agc_rgb (agc, 1, 0, 0);
	if (res != S_CSC_OK)
		return res;

	res = agc_ctm (agc, NULL);
	if (res != S_CSC_OK)
		return res;

	for (i = 0; i < amys -> margins; i++) {
		amy = &amys -> margin [i];
		rect = *pgrect;
		if (amy -> type == ARGS_LD_NORTH ||
		    amy -> type == ARGS_LD_SOUTH)
			rect.top = rect.bottom = amy -> value;
		else
			rect.left = rect.right = amy -> value;
		res = agc_rect (agc, &rect, &gs_paint_stroke);
		if (res != S_CSC_OK)
			return res;
	}

	return S_CSC_OK;
}

static enum s_status do_highlight (struct conv *conv, struct pg_pages *dpages,
				   struct bt_cursor *dpc, 
				   struct gspg_page *spdfp, 
				   struct gr_al_machine *dc,
				   page_transl transl)
{
	struct gspg_page *dpdfp;
	struct pg_page *dpage;
	struct gr_grid *grid;
	enum s_status res;

	res = pg_page_insert (&dpage, dpages, dpc);
	if (res != S_CSC_OK)
		return res;

	dpage -> inh.media_box = spdfp -> bounds;

	res = gr_get_grid (&grid, spdfp, dc, &conv -> sstate.amys,
			   &conv -> sstate.heur, &conv -> owner);
	if (res != S_CSC_OK)
		return res;

	res = transl (conv, spdfp, grid, &dpdfp, &conv -> owner);
	if (res != S_CSC_OK) {
		gr_unref (grid, &conv -> owner);
		return res;
	}

	res = gr_highlight (grid, conv -> dstate.agc, 
			    conv -> dstate.aout.hlspec, dpdfp);
	gr_unref (grid, &conv -> owner);
	if (res == S_CSC_OK)
		res = draw_margins (conv -> dstate.agc, 
				    &conv -> sstate.amys,
				    &spdfp -> bounds);

	if (res == S_CSC_OK)
		res = agc_done (conv -> dstate.agc);

	if (res != S_CSC_OK) {
		gspg_page_unref (dpdfp, &conv -> owner);
		return res;
	}

	res = pdfc_compile (conv -> pdc, dpdfp, dpage);
	gspg_page_unref (dpdfp, &conv -> owner);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status high_app (struct conv *conv, struct pg_pages *dpages, 
			       struct bt_cursor *dpc, page_transl transl)
{
	struct gr_al_machine *dc;
	struct gspg_page *spdfp;
	enum s_status res;

	res = gr_al_create_machine (&dc, &conv -> sstate.heur, conv -> rfp);
	if (res != S_CSC_OK)
		return res;

	res = S_CSC_OK;
	while (1) {
		res = pull_page (conv, NULL, NULL, &spdfp, &conv -> owner);
		if (res != S_CSC_OK)
			break;

		if (spdfp == NULL)
			break;

		res = do_highlight (conv, dpages, dpc, spdfp, dc, transl);
		gspg_page_unref (spdfp, &conv -> owner);
		if (res != S_CSC_OK)
			break;
	}

	gr_al_destroy_machine (dc);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status eject (struct agc_context *pagc, 
			    struct cobj_rect *bounds, void *closure)
{
	struct resize_closure *rc;
	struct pg_page *dpage;
	struct gspg_page *dpdfp;
	struct agc_context *agc;
	struct gs_point point;
	struct conv *conv;
	enum s_status res;

	rc = closure;
	conv = rc -> conv;

	res = pg_page_insert (&dpage, rc -> dpages, rc -> dpc);
	if (res != S_CSC_OK)
		return res;

	res = gspg_page_new (&dpdfp, &conv -> dstate.bounds, 
			     0, &conv -> owner);
	if (res != S_CSC_OK)
		return res;

	res = agc_new (&agc, conv -> dstate.doc -> rdb, 
		       dpdfp, &conv -> owner);
	if (res != S_CSC_OK) {
		gspg_page_unref (dpdfp, &conv -> owner);
		return res;
	}
				     
	dpage -> inh.media_box = conv -> dstate.bounds;
	point.x = point.y = 0;
	res = agc_apply (pagc, agc, &point);
	if (res == S_CSC_OK)
		res = agc_done (agc);

	agc_unref (agc, &conv -> owner);
	if (res != S_CSC_OK) {
		gspg_page_unref (dpdfp, &conv -> owner);
		return res;
	}

	res = pdfc_compile (conv -> pdc, dpdfp, dpage);
	gspg_page_unref (dpdfp, &conv -> owner);
	if (res != S_CSC_OK)
		return res;

	*bounds = conv -> dstate.bounds;

	return S_CSC_OK;
}

static enum s_status detect (struct conv *conv, const char *fmt)
{
	struct gspg_page *spdfp;
	struct gr_layout gly;
	enum s_status res;

	gr_layout_init (&gly, &conv -> sstate.heur);
	res = S_CSC_OK;
	while (1) {
		res = pull_page (conv, &conv -> sstate.ranges.detect,
				 fmt, &spdfp, &conv -> owner);
		if (res != S_CSC_OK)
			break;

		if (spdfp == NULL)
			break;

		res = gr_layout_learn (&gly, spdfp);
		gspg_page_unref (spdfp, &conv -> owner);
		if (res != S_CSC_OK)
			break;
	}

	gr_layout_apply (&gly, &conv -> sstate.amys);
	gr_layout_destroy (&gly);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status detect_app (struct conv *conv)
{
	struct args_margin *amy;
	enum s_status res;
	char c;
	int i;

	res = detect (conv, NULL);
	if (res != S_CSC_OK)
		return res;

	printf ("\n");

	if (!conv -> sstate.amys.margins) {
		printf ("No margins detected\n");
		return S_CSC_OK;
	}

	printf ("Autodetected layout code: -D ");

	for (i = 0; i < conv -> sstate.amys.margins; i++) {
		amy = &conv -> sstate.amys.margin [i];
		c = -1;
		switch (amy -> type) {
		case ARGS_LD_NORTH:
			c = 'n';
			break;

		case ARGS_LD_EAST:
			c = 'e';
			break;

		case ARGS_LD_WEST:
			c = 'w';
			break;

		case ARGS_LD_SOUTH:
			c = 's';
			break;
		}

		printf ("%c%g", c, amy -> value);
	}

	printf ("/%d", (int) (conv -> sstate.amys.weight * 100));
				
	return S_CSC_OK;
}

static enum s_status resize_app (struct conv *conv, struct pg_pages *dpages, 
				 struct bt_cursor *dpc)
{
	struct gr_line_composer *lc;
	struct gr_lc_param lcp;
	struct resize_closure rc;
	struct gr_al_machine *dc;
	struct gspg_page *spdfp;
	struct gr_grid *grid;
	enum s_status res;

	lcp.aout = &conv -> dstate.aout;
	lcp.heur = &conv -> sstate.heur;

	rc.conv = conv;	
	rc.dpages = dpages;
	rc.dpc = dpc;

	res = gr_lc_new (&lc, eject, &rc, &lcp, &conv -> dstate.bounds, 
			 &conv -> owner);
	if (res != S_CSC_OK)
		return res;
	
	res = gr_al_create_machine (&dc, &conv -> sstate.heur, conv -> rfp);
	if (res != S_CSC_OK) {
		gr_lc_unref (lc, &conv -> owner);
		return res;
	}

	res = S_CSC_OK;
	while (1) {
		res = pull_page (conv, NULL, NULL, &spdfp, &conv -> owner);
		if (res != S_CSC_OK)
			break;

		if (spdfp == NULL)
			break;

		res = gr_get_grid (&grid, spdfp, dc, &conv -> sstate.amys,
				   &conv -> sstate.heur, &conv -> owner);
		gspg_page_unref (spdfp, &conv -> owner);
		if (res != S_CSC_OK)
			break;

		res = gr_lc_feed (lc, grid);
		gr_unref (grid, &conv -> owner);
		if (res != S_CSC_OK)
			break;
	}

	gr_al_destroy_machine (dc);

	if (res != S_CSC_OK) {
		gr_lc_unref (lc, &conv -> owner);
		return res;
	}

	res = gr_lc_finalize (lc);
	gr_lc_unref (lc, &conv -> owner);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status rewind_doc (struct conv *conv, struct pg_pages *spages)
{
	enum s_status res;

	res = bt_cursor_init (spages -> tree, &conv -> sstate.spc, 
			      &conv -> owner, &conv -> sstate.eof);
	if (res != S_CSC_OK)
		return res;

	conv -> sstate.page = 0;

	return S_CSC_OK;
}

static enum s_status process_pages (struct conv *conv, struct args *args)
{
	struct bt_cursor *dpc;
	struct tri_trailer *tri;
	struct cat_catalog *cat;
	struct pg_pages *spages, *dpages;
	enum s_status res;
	int i;

	tri = conv -> sstate.doc -> trailer;
	if (tri == NULL)
		return S_CSC_OK;

	cat = tri -> root;
	if (cat == NULL)
		return S_CSC_OK;

	spages = cat -> pages;
	if (spages == NULL || spages -> tree == NULL)
		return S_CSC_OK;

	if (args -> app != ARGS_A_DETECT && 
	    !conv -> sstate.amys.skip_m &&
	    !conv -> sstate.amys.skip_w) {
		res = rewind_doc (conv, spages);
		if (res != S_CSC_OK)
			return res;

		res = detect (conv, "<%d>");
		bt_cursor_unref (conv -> sstate.spc, &conv -> owner);
		if (res != S_CSC_OK)
			return res;
	}

	res = rewind_doc (conv, spages);
	if (res != S_CSC_OK)
		return res;

	dpages = conv -> dstate.doc -> trailer -> root -> pages;
	res = bt_cursor_init (dpages -> tree, &dpc, &conv -> owner, NULL);
	if (res != S_CSC_OK) {
		bt_cursor_unref (conv -> sstate.spc, &conv -> owner);
		return res;
	}

	switch (args -> app) {
	case ARGS_A_HIGHLIGHT:
		res = high_app (conv, dpages, dpc, noop_transl);
		break;
			
	case ARGS_A_FLOAT:
		res = high_app (conv, dpages, dpc, float_transl);
		break;

	case ARGS_A_RESIZE:
		res = resize_app (conv, dpages, dpc);
		break;

	case ARGS_A_DETECT:
		res = detect_app (conv);
		break;

	default:
		assert (0);
	}

	res = S_CSC_OK;
	i = 0;

	printf ("\n");

	bt_cursor_unref (dpc, &conv -> owner);
	bt_cursor_unref (conv -> sstate.spc, &conv -> owner);
	if (res != S_CSC_OK)
		return res;

	res = pdfc_finalize (conv -> pdc);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status do_conv (struct conv *conv, struct args *args)
{
	enum s_status res;

	res = pd_derive (conv -> dstate.doc, conv -> sstate.doc, conv -> ob);
	if (res != S_CSC_OK)
		return res;

	res = pd_sync_start (conv -> dstate.doc, &conv -> dstate.sel);
	if (res != S_CSC_OK)
		return res;

	res = process_pages (conv, args);
	if (res != S_CSC_OK)
		return res;

	res = pd_sync_end (conv -> dstate.doc, &conv -> dstate.sel);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

enum s_status conv_run (const char *sfname, const char *dfname, 
			struct args *args)
{
	struct conv *conv;
	enum s_status res;

	printf ("%s:\n", sfname);
	res = conv_init (&conv, sfname, dfname, args);
	if (res != S_CSC_OK)
		return res;

	res = do_conv (conv, args);
	conv_destroy (conv, res == S_CSC_OK);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

