/* 
 * PDFlow -- args.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <ctype.h>
#include <stdlib.h>
#include <getopt.h>
#include <string.h>
#include "args.h"

#define CHARSTRING "a:O:o:d:x:hS:W:H:V:t:D:r:L:C:R:s:w:"

#define O_COLUMN_GLUE 		1001
#define O_WORD_GLUE 		1002
#define O_INDENT_TOLERANCE 	1003
#define O_BASELINE_TOLERANCE 	1004
#define O_HF_TOLERANCE	 	1005
#define O_HF_THRESH	 	1007
#define O_HEADER_THRESH	 	1008
#define O_FOOTER_THRESH	 	1009
#define O_HF_PERCENTAGE	        1010
#define O_HF_GAP	        1011
#define O_DETECT_QUORUM	        1012
#define O_PAR_DECADENCE	        1013
#define O_GRAPH_SPACE	        1014
#define O_SPACE_RATIO	        1015
#define O_WSPACE_WIDTH	        1016
#define O_GRAPH_SPACE_EL        1017
#define O_PAR_BREAK	        1018
#define O_FILL_THRESH	        1019
#define O_W_GRAPH_SPACE	        1020
#define O_W_TOLERANCE	        1021
#define O_PAGE_WEIGHT		1022
#define O_WEIGHT_TOLERANCE	1023
#define O_GRAPH_RATIO		1024
#define O_GRAPH_BAD_SLOPE	1025
#define O_MAX_INDENT		1026
#define O_GLYPH_GLUE 		1027
#define O_MERGE_LINE_AFFINITY	1028

#if defined (_WIN32) ||	defined (WIN32)
#define SEPARATOR '\\'
#elif defined (__APPLE__)
#define SEPARATOR ':'
#else
#define SEPARATOR '/'
#endif

static struct option opts [] = {
	{ "application", 1, NULL, 'a' },
	{ "optimization", 1, NULL, 'O' },
	{ "output", 1, NULL, 'o' },
	{ "dest-dir", 1, NULL, 'd' },
	{ "extension", 1, NULL, 'x' },
	{ "help", 0, NULL, 'h' },
	{ "column-glue", 1, NULL, O_COLUMN_GLUE },
	{ "word-glue", 1, NULL, O_WORD_GLUE },
	{ "glyph-glue", 1, NULL, O_GLYPH_GLUE },
	{ "indent-tolerance", 1, NULL, O_INDENT_TOLERANCE },
	{ "baseline-tolerance", 1, NULL, O_BASELINE_TOLERANCE },
	{ "hf-tolerance", 1, NULL, O_HF_TOLERANCE },
	{ "hf-thresh", 1, NULL, O_HF_THRESH },
	{ "header-thresh", 1, NULL, O_HEADER_THRESH },	
	{ "footer-thresh", 1, NULL, O_FOOTER_THRESH },
	{ "hf-percentage", 1, NULL, O_HF_PERCENTAGE },
	{ "hf-gap", 1, NULL, O_HF_GAP },
	{ "size", 1, NULL, 'S' },
	{ "width", 1, NULL, 'W' },
	{ "height", 1, NULL, 'H' },
	{ "range", 1, NULL, 'r' },
	{ "pdf-version", 1, NULL, 'V' },
	{ "text-ratio", 1, NULL, 't' },
	{ "detect-data", 1, NULL, 'D' },
	{ "margin-clip", 1, NULL, 'C' },
	{ "detect-range", 1, NULL, 'R' },
	{ "margins", 1, NULL, 'L' },
	{ "detect-quorum", 1, NULL, O_DETECT_QUORUM },
	{ "par-decadence", 1, NULL, O_PAR_DECADENCE },
	{ "par-break", 1, NULL, O_PAR_BREAK },
	{ "graph-space", 1, NULL, O_GRAPH_SPACE },
	{ "w-graph-space", 1, NULL, O_W_GRAPH_SPACE },
	{ "w-tolerance", 1, NULL, O_W_TOLERANCE },
	{ "space-ratio", 1, NULL, O_SPACE_RATIO },
	{ "hspace-width", 1, NULL, O_WSPACE_WIDTH },
	{ "graph-space-el", 1, NULL, O_GRAPH_SPACE_EL },
	{ "fill-thresh", 1, NULL, O_FILL_THRESH },
	{ "hlspec", 1, NULL, 's' },
	{ "page-weight", 1, NULL, O_PAGE_WEIGHT },
	{ "weight-tolerance", 1, NULL, O_WEIGHT_TOLERANCE },
	{ "graph-ratio", 1, NULL, O_GRAPH_RATIO },
	{ "graph-bad-slop", 1, NULL, O_GRAPH_BAD_SLOPE },
	{ "max-indent", 1, NULL, O_MAX_INDENT },
	{ "report-file", 1, NULL, 'w' },
	{ "merge-line-affinity", 1, NULL, O_MERGE_LINE_AFFINITY },
	{ NULL }
};

static void set_defaults (struct args *args)
{
	memset (args, 0, sizeof (struct args));
	
	args -> app = ARGS_A_RESIZE;

	args -> heur.column_glue = 4;
	args -> heur.word_glue = 0.1;
	args -> heur.glyph_glue = 0.01;
	args -> heur.indent_tolerance = 1.1;
	args -> heur.baseline_tolerance = 1.1;
	args -> heur.hf_tolerance = 1.5;
	args -> heur.header_thresh = 0.25;
	args -> heur.footer_thresh = 0.25;
	args -> heur.hf_percentage = .60;
	args -> heur.hf_gap = 1.2;
	args -> heur.par_decadence = 0.3;
	args -> heur.graph_space = 0.07;
	args -> heur.w_graph_space = .07;
	args -> heur.w_tolerance = 2;
	args -> heur.weight_tolerance = .05;
	args -> heur.par_break = 2;
	args -> heur.fill_thresh = 5;
	args -> heur.max_indent = 7;
	args -> heur.merge_line_affinity = .8;

	args -> aout.width = 255.1181;
	args -> aout.height = 340.1575;
	args -> aout.text_ratio = .707;
	args -> aout.space_ratio = -1;
	args -> aout.graph_ratio = -1;
	args -> aout.graph_bad_slope = 50;
	args -> aout.wspace_width = .2;
	args -> aout.graph_space_el = 0.3;
	args -> aout.hlspec = "";

	args -> ranges.doc.lo = 0;
	args -> ranges.doc.hi = -1;

	args -> ranges.detect.lo = -1;
	args -> ranges.detect.hi = -1;

	args -> debug.report = NULL;

	args -> amys.skip_m = 0;
	args -> amys.skip_w = 0;
	args -> amys.margins = 0;
	args -> amys.max_margins = 10;
	args -> amys.quorum = 2;
	args -> amys.weight = .5;
}

static enum s_status parse_real (pdf_real_t *r, const char *s)
{
	pdf_real_t ans;
	char *end;

	ans = strtod (s, &end);
	if (*end)
		return s_make_indirect (S_CSC_PARSE_ERROR, 
					"Bad number: %s", s);
	
	*r = ans;

	return S_CSC_OK;
}

static enum s_status parse_integer (pdf_integer_t *i, const char *s)
{
	pdf_integer_t ans;
	char *end;

	ans = strtol (s, &end, 0);
	if (*end)
		return s_make_indirect (S_CSC_PARSE_ERROR, 
					"Bad int: %s", s);
	
	*i = ans;

	return S_CSC_OK;
}

static int do_parse_rect (struct args_rect *arect, const char *s)
{
	char *end;

	arect -> left = strtod (s, &end);
	if (*end != ':')
		return 0;
	if (end == s)
		arect -> left = 0;
	s = end + 1;

	arect -> bottom = strtod (s, &end);
	if (*end != ':')
		return 0;
	if (end == s)
		arect -> bottom = 0;
	s = end + 1;

	arect -> right = strtod (s, &end);
	if (*end != ':')
		return 0;
	if (end == s)
		arect -> right = -1;
	s = end + 1;

	arect -> top = strtod (s, &end);
	if (*end)
		return 0;
	if (end == s)
		arect -> top = -1;

	return 1;
}

static enum s_status parse_rect (struct args_rect *arect, const char *s)
{
	if (!do_parse_rect (arect, s))
		return s_make_indirect (S_CSC_PARSE_ERROR,
					"Bad rect string, format is "
					"left:bottom:right:top");
	return S_CSC_OK;
}

static int do_parse_detect (struct args_margins *amys, const char *s)
{
	struct args_margin *amy;
	char *end;

	for (amy = &amys -> margin [amys -> margins]; 
	     amys -> margins < ARGS_MAX_MARGINS;
	     amy = &amys -> margin [++amys -> margins]) {
		
		switch (*s) {
		case 'N':
			amy -> type = ARGS_LD_NORTH;
			break;

		case 'E':
			amy -> type = ARGS_LD_EAST;
			break;

		case 'W':
			amy -> type = ARGS_LD_WEST;
			break;

		case 'S':
			amy -> type = ARGS_LD_SOUTH;
			break;

		case '/':
			break;

		default:
			return 0;
		}

		if (*s == '/')
			break;
		
		s++;

		amy -> value = strtod (s, &end);
		if (end == s)
			return 0;

		s = end;
	}

	if (*s == '/') {
		s++;
		amys -> weight = strtod (s, &end);
		amys -> weight /= 100;
	}

	if (*end) {
#define s(s) #s
		fprintf (stderr, "Warning: exceeded number of margins "
			 " (max is " s (ARGS_MAX_MARGINS) ")\n");
#undef s
		return 1;
	}


	return 1;
}

static enum s_status parse_detect (struct args_margins *amys, const char *s)
{
	if (!do_parse_detect (amys, s))
		return s_make_indirect (S_CSC_PARSE_ERROR,
					"Bad detect string, format is "
					"N<..>E<...>W<...>S<...>.../...");

	return S_CSC_OK;
}

static int do_parse_range (struct args_range *range, const char *s)
{
	char *end;

	range -> lo = strtoul (s, &end, 10);
	if (*end != ':')
		return 0;
	if (end == s)
		range -> lo = 0;
	else
		range -> lo--;	/* First page is page zero */
	s = end + 1;

	range -> hi = strtoul (s, &end, 10);
	if (*end)
		return 0;
	if (end == s)
		range -> hi = -1;
	else
		range -> hi--;

	if (range -> lo > range -> hi && range -> hi != -1)
		return 0;

	return 1;
}

static enum s_status parse_range (struct args_range *range, const char *s)
{
	if (!do_parse_range (range, s))
		return s_make_indirect (S_CSC_PARSE_ERROR,
					"Bad range string, format is "
					"[lo]:[hi], e.g. 1:5");

	return S_CSC_OK;
}

static enum s_status parse_dots (pdf_real_t *r, const char *s)
{
	pdf_real_t ans;
	char *end;

	ans = strtod (s, &end);
	if (end == s)
		return s_make_indirect (S_CSC_PARSE_ERROR, 
					"Bad number: %s", s);
	if (!strcmp (end, "cm"))
		ans *= 28.346;
	else if (!strcmp (end, "inch") || !strcmp (end, "i"))
		ans *= 72;
	else if (!strcmp (end, "pt"))
		;
	else
		return s_make_indirect (S_CSC_PARSE_ERROR, 
					"Bad unit: %s", end);

	*r = ans;

	return S_CSC_OK;

}

static enum s_status parse_version (struct v_pref *prefs, const char *s)
{
	if (sscanf (s, "%u.%u", &prefs -> major, &prefs -> minor) != 2)
		return s_make_indirect (S_CSC_PARSE_ERROR, 
					"Bad version: %s", s);

	return S_CSC_OK;
}

static enum s_status parse_app (struct args *args, const char *appname)
{
	if (!strncasecmp ("resize", appname, strlen (appname)))
		args -> app = ARGS_A_RESIZE;
	else if (!strncasecmp ("float", appname, strlen (appname)))
		args -> app = ARGS_A_FLOAT;
	else if (!strncasecmp ("highlight", appname, strlen (appname)))
		args -> app = ARGS_A_HIGHLIGHT;
	else if (!strncasecmp ("detect", appname, strlen (appname)))
		args -> app = ARGS_A_DETECT;
	else
		return s_make_indirect (S_CSC_PARSE_ERROR, 
					"Bad appname (%s)", appname);

	return S_CSC_OK;
}

static enum s_status parse_opt (struct args *args, const char *flag)
{
	int good;

	good = 1;
	if (*flag) {
		if (isdigit (*flag)) {
			args -> prefs.speed = *flag - '0';
			if (!args -> prefs.speed)
				args -> prefs.bspeed = 5;
		} else if (*flag == 's')
			args -> prefs.size = 5;
		else if (*flag == 't')
			args -> prefs.text = 1;
		else if (*flag == 'x')
			args -> prefs.skip_optimizer = 1;
		else
			good = 0;
	} else
		good = 0;

	if (!good)
		return s_make_indirect (S_CSC_PARSE_ERROR, 
					"Bad optimization (%s)", flag);

	return S_CSC_OK;
}

static void helpmsg (void)
{
	fprintf (stderr, 
                 "Usage: pdflow [options] filename[s]           \n"
                 "    Most useful options are:                     \n"
                 "      -o fname: set output filename (may be dir) \n"
                 "      -x ext: set output extension (e.g. file.ext.pdf)\n"
                 "      -S size: set display width&height (def. inches) \n"
                 "      -W width: set display width (def. inches)       \n"
                 "      -H height: set display height (def. inches)     \n"
                 "      -r [start]:[end]: set range (def. :)     	\n"
                 "Examples:                                             \n"
                 "   Create file.qdf.pdf for 6 inches display:          \n"
                 "       $  pdflow file.pdf                             \n"
                 "   Create out.pdf for 15cm display:                   \n"
                 "       $  pdflow -o out.pdf -S 15cm file.pdf          \n"
                 "   Convert all the files from in/ to out/, skip page 1:\n"
                 "       $  pdflow -o out in/*.pdf -r 2:  (or)          \n"
                 "       $  pdflow -d out in/*.pdf -r 2:                \n"
                 "   Convert all the files and place extension:         \n"
                 "       $  pdflow -x qdf *.pdf                         \n"
                 "See manpage for further options & examples\n");
}

static enum s_status parse_file (struct args_file *af, const char *arg,
				 const char *output, const char *ext,
				 int todir)
{
	size_t size;
	char *p;

	if (!*arg)
		return s_make_indirect (S_CSC_PARSE_ERROR, 
					"Filename too funny");
	af -> src = strdup (arg);
	if (af -> src == NULL)
		return S_CSC_OUT_OF_MEMORY;

	if (todir)
		size = strlen (output) + 1 + strlen (arg);
	else if (output != NULL)
		size = strlen (output);
	else
		size = strlen (arg);
	if (ext != NULL)
		size += 1 + strlen (ext);

	size++;
	af -> dest = malloc (size);
	if (af -> dest == NULL)
		return S_CSC_OUT_OF_MEMORY;
	
	if (todir)
		snprintf (af -> dest, size, "%s%c%s", 
			  output, SEPARATOR, arg);
	else if (output != NULL)
		strcpy (af -> dest, output);
	else
		strcpy (af -> dest, arg);
	if (ext != NULL) {
		p = strrchr (af -> dest, '.');
		if (p != NULL)
			memmove (p + strlen (ext) + 1, p, strlen (p) + 1);
		else {
			p = af -> dest + strlen (af -> dest);
			af -> dest [size - 1] = 0;
		}
		*(p++) = '.';
		memcpy (p, ext, strlen (ext));
	}

	return S_CSC_OK;
}

static enum s_status parse_files (struct args *args, int argc, char *argv [],
				  const char *output, const char *ext, int tod)
{
	enum s_status res;
	struct stat st;
	int todir, i;

	if (output == NULL) {
		if (ext == NULL)
			ext = "qdf";
		todir = 0;
	} else if (output != NULL) {
		todir = !stat (output, &st) && S_ISDIR (st.st_mode);
		if (!todir)
			ext = NULL;
	}

	if (tod && !todir) 
		return s_make_indirect (S_CSC_PARSE_ERROR, 
					"-d not applied to directory");

	if (!argc)
		return s_make_indirect (S_CSC_PARSE_ERROR, "No input files");

	args -> file = calloc (argc, sizeof (struct args_file));
	if (args -> file == NULL)
		return S_CSC_OUT_OF_MEMORY;

	args -> files = argc;

	for (i = 0; i < argc; i++) {
		res = parse_file (&args -> file [i], argv [i], 
				  output, ext, todir);
		if (res != S_CSC_OK)
			return res;
	}

	if (args -> debug.report != NULL) {
		args -> debug.rfp = fopen (args -> debug.report, "w");
		if (args -> debug.rfp == NULL)
			return s_make_indirect (S_ERRNO, 
						"Can't create report file: "
						"%s: %s",
						args -> debug.report,
						strerror (errno));
	}

	return S_CSC_OK;
}

static enum s_status parse (struct args *args, int argc, char *argv [])
{
	const char *output, *ext;
	pdf_real_t size;
	enum s_status res;
	int c, need_help, todir;

	output = ext = NULL;
	need_help = todir = 0;
	while (1) {
		c = getopt_long (argc, argv, CHARSTRING, opts, NULL);
		if (c == -1)
			break;

		res = S_CSC_OK;

		switch (c) {
		case 'a':
			res = parse_app (args, optarg);
			break;

		case 'O':
			res = parse_opt (args, optarg);
			break;

		case 'o':
			output = optarg;
			todir = 0;
			break;

		case 'd':
			output = optarg;
			todir = 1;
			break;

		case 'x':
			ext = optarg;
			break;

		case 'h':
			need_help = 1;
			break;

		case O_COLUMN_GLUE:
			res = parse_real (&args -> heur.column_glue, optarg);
			break;

		case O_WORD_GLUE:
			res = parse_real (&args -> heur.word_glue, optarg);
			break;

		case O_GLYPH_GLUE:
			res = parse_real (&args -> heur.glyph_glue, optarg);
			break;

		case O_INDENT_TOLERANCE:
			res = parse_real (&args -> heur.indent_tolerance, 
					  optarg);
			break;

		case O_MAX_INDENT:
			res = parse_real (&args -> heur.max_indent, 
					  optarg);
			break;

		case O_BASELINE_TOLERANCE:
			res = parse_real (&args -> heur.baseline_tolerance, 
					  optarg);
			break;

		case O_HF_TOLERANCE:
			res = parse_real (&args -> heur.hf_tolerance, optarg);
			break;

		case O_HF_THRESH:
			res = parse_real (&args -> heur.header_thresh, optarg);
			args -> heur.footer_thresh = 
				args -> heur.header_thresh;
			break;

		case O_HEADER_THRESH:
			res = parse_real (&args -> heur.header_thresh, optarg);
			break;

		case O_FOOTER_THRESH:
			res = parse_real (&args -> heur.footer_thresh, optarg);
			break;

		case O_HF_PERCENTAGE:
			res = parse_real (&args -> heur.hf_percentage, optarg);
			break;

		case O_HF_GAP:
			res = parse_real (&args -> heur.hf_gap, optarg);
			break;

		case O_PAR_DECADENCE:
			res = parse_real (&args -> heur.par_decadence, optarg);
			break;

		case O_GRAPH_SPACE:
			res = parse_real (&args -> heur.graph_space, optarg);
			break;

		case O_PAR_BREAK:
			res = parse_real (&args -> heur.par_break, optarg);
			break;

		case O_FILL_THRESH:
			res = parse_integer (&args -> heur.fill_thresh, 
					     optarg);
			break;

		case 'r':
			res = parse_range (&args -> ranges.doc, optarg);
			break;

		case 't':
			res = parse_real (&args -> aout.text_ratio, optarg);
			break;

		case O_GRAPH_RATIO:
			res = parse_real (&args -> aout.graph_ratio, optarg);
			break;

		case O_GRAPH_BAD_SLOPE:
			res = parse_real (&args -> aout.graph_bad_slope, 
					  optarg);
			break;

		case O_SPACE_RATIO:
			res = parse_real (&args -> aout.space_ratio, optarg);
			break;

		case O_WSPACE_WIDTH:
			res = parse_real (&args -> aout.wspace_width, optarg);
			break;

		case O_GRAPH_SPACE_EL:
			res = parse_real (&args -> aout.graph_space_el, 
					  optarg);
			break;

		case O_W_GRAPH_SPACE:
			res = parse_real (&args -> heur.w_graph_space, 
					  optarg);
			break;

		case O_W_TOLERANCE:
			res = parse_real (&args -> heur.w_tolerance, 
					  optarg);
			break;

		case O_MERGE_LINE_AFFINITY:
			res = parse_real (&args -> heur.merge_line_affinity, 
					  optarg);
			args -> heur.merge_line_affinity /= 100;
			break;

		case 'D':
			res = parse_detect (&args -> amys, optarg);
			args -> amys.skip_m = 1;
			break;

		case 'C':
			res = parse_rect (&args -> amys.clip, optarg);
			args -> amys.skip_m = 1;
			break;

		case O_PAGE_WEIGHT:
			res = parse_real (&args -> amys.weight, optarg);
			args -> amys.weight /= 100;
			args -> amys.skip_w = 1;
			break;

		case O_WEIGHT_TOLERANCE:
			res = parse_real (&args -> heur.weight_tolerance, 
					  optarg);
			break;

		case 'R':
			res = parse_range (&args -> ranges.detect, optarg);
			break;

		case O_DETECT_QUORUM:
			res = parse_integer (&args -> amys.quorum, optarg);
			break;

		case 'L':
			res = parse_integer (&args -> amys.max_margins, 
					     optarg);
			if (args -> amys.max_margins > ARGS_MAX_MARGINS)
				args -> amys.max_margins = ARGS_MAX_MARGINS;
			break;

		case 'S':
			res = parse_dots (&size, optarg);
			if (res == S_CSC_OK) {
				args -> aout.width = (3 * size) / 5;
				args -> aout.height = (4 * size) / 5;
			}
			break;

		case 'W':
			res = parse_dots (&args -> aout.width, optarg);
			break;

		case 'H':
			res = parse_dots (&args -> aout.height, optarg);
			break;
			
		case 's':
			args -> aout.hlspec = optarg;
			break;

		case 'V':
			res = parse_version (&args -> prefs, optarg);
			break;

		case 'w':
			args -> debug.report = optarg;
			break;

		default:
			need_help = 1;
		}

		if (res != S_CSC_OK) {
			helpmsg ();
			return res;
		}
	}

	if (need_help) {
		helpmsg ();
		return S_CSC_USER_CANCEL;
	}

	res = parse_files (args, argc - optind, argv + optind, 
			   output, ext, todir);
	if (res != S_CSC_OK)
		return res;

	if (args -> ranges.detect.lo == -1)
		args -> ranges.detect = args -> ranges.doc;

	if (args -> aout.space_ratio == -1)
		args -> aout.space_ratio = args -> aout.text_ratio;

	if (args -> aout.graph_ratio == -1)
		args -> aout.graph_ratio = args -> aout.text_ratio;

	return S_CSC_OK;
}

enum s_status args_init (struct args *args, int argc, char *argv [])
{
	enum s_status res;

	set_defaults (args);
	res = parse (args, argc, argv);
	if (res != S_CSC_OK) {
		args_destroy (args);
		return res;
	}

	return S_CSC_OK;
}

void args_destroy (struct args *args)
{
	int i;

	if (args -> debug.rfp)
		fclose (args -> debug.rfp);

	if (args -> file != NULL) {
		for (i = 0; i < args -> files; i++) {
			if (args -> file [i].src != NULL)
				free (args -> file [i].src);
			if (args -> file [i].dest != NULL)
				free (args -> file [i].dest);
		}
		free (args -> file);
	}
}

