/*
 * PDFlow -- a85.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "filters_priv.h"

#define WS "\t\r\n\f "

static enum s_status encode (const struct obj_region *src,
			     struct obj_region *dest,
			     struct xref_db *rdb,
			     struct obj *parms);
static enum s_status decode (const struct obj_region *src,
			     struct obj_region *dest,
			     struct xref_db *rdb,
			     pdf_integer_t dl,
			     struct obj *parms);

struct flt_interface flt_a85_ifc = {
	"ASCII85Decode", encode, decode
};

static void encode_pkt(unsigned char *pdest, unsigned long curr_dword) {
	int ii;

	for (ii = 4; ii >= 0; ii--) {
		pdest[ii] = '!' + (curr_dword % 85);
		curr_dword /= 85;
	}
}

static enum s_status encode (const struct obj_region *src,
			     struct obj_region *dest,
			     struct xref_db *rdb,
			     struct obj *parms)
{
	int src_len = src -> end - src -> begin;
	/*
	 * at most, dest will have 5 bytes for each 4 src bytes,
	 * plus 2 bytes for the ~> EOD marker
	 */
	int max_dest_len = (src_len * 5) / 4 + 2;
	unsigned char *buf, *pdest, *newbuf;
	const uint8_t *psrc;
	unsigned const char *end_dword_boundary = src -> begin + (src_len & 0xfffffffc);
	int remaining_bytes = src_len % 4;
	unsigned long curr_dword;
	int idx, true_size;

	buf = malloc(max_dest_len);
	if (buf == NULL)
		return S_CSC_OUT_OF_MEMORY;

	psrc = src -> begin;
	pdest = buf;
	curr_dword = 0;
	idx =0;
	while (psrc < end_dword_boundary) {
		curr_dword <<= 8;
		curr_dword |= *psrc;
		psrc++;
		idx++;
		if (idx >= 4) {
			/* a dword is ready to be encoded */
			idx = 0;
			if (curr_dword == 0) {
				/* special encoding:
				 * 0x00, 0x00, 0x00, 0x00 -> 'z' */
				*pdest = 'z';
				pdest++;
			} else {
				/* normal encoding:
				 * 4 bytes base 256 -> 5 bytes base 85 */
				encode_pkt(pdest, curr_dword);
				pdest += 5;
				curr_dword = 0;
			}
		}
	}

	if (remaining_bytes) {
		unsigned char last_bytes[5];

		for ( ; idx < 4; idx++) {
			curr_dword <<= 8;
			if (idx < remaining_bytes) {
				curr_dword |= *psrc;
				psrc++;
			}
		}
		encode_pkt(last_bytes, curr_dword);
		memcpy(pdest, last_bytes, remaining_bytes + 1);
		pdest += remaining_bytes + 1;
	}

	memcpy(pdest, "~>", 2);
	pdest += 2;

	true_size = pdest - buf;

	/*
	 * Try to (likely) shrink buf.
	 * Even if realloc() fails, we still have the previous pointer.
	 */
	newbuf = realloc(buf, true_size);
	if (newbuf != NULL) {
		buf = newbuf;
		pdest = buf+true_size;
	}

	obj_init_malloc_region (dest, buf, pdest);

	return S_CSC_OK;
}

static void decode_pkt (unsigned char in [5], unsigned char *out)
{
	unsigned long b;

	b = in [4] +
		85 * (in [3] +
		      85 * (in [2] +
			    85 * (in [1] +
				  85 * in [0])));
	out [0] = b >> 24;
	out [1] = b >> 16;
	out [2] = b >> 8;
	out [3] = b;
}

static enum s_status decode (const struct obj_region *src,
			     struct obj_region *dest,
			     struct xref_db *rdb,
			     pdf_integer_t dl,
			     struct obj *parms)
{
	const unsigned char *p, *eos;
	unsigned char *q, *buf, *newbuf;
	unsigned char in [5];
	int c, dest_len;


	eos = src -> end-1;
	p = src -> begin;
	while (eos >= p && *eos != '>') {
		eos--;
	}
	if (*eos != '>')
		return s_make_indirect (S_CSC_PARSE_ERROR,
					"Unterminated a85 stream: missing '>'");

	do {
		eos--;
	} while((strchr (WS, *eos) != NULL) && (eos >= p));

	if (*eos != '~')
		return s_make_indirect (S_CSC_PARSE_ERROR,
					"Unterminated a85 stream: missing '~' before '>'");

	c = eos - src -> begin;

	for (p = src -> begin; p < eos; p++)
		if (*p == 'z')
			c += 4;

	dest_len = (c + 4) / 5 * 4;

	buf = calloc (1, dest_len ? dest_len : 1);
	if (buf == NULL)
		return S_CSC_OUT_OF_MEMORY;

	q = buf;
	c = 0;
	for (p = src -> begin; p < eos; p++) {
		if (!*p || strchr (WS, *p) != NULL)
			continue;

		if (*p == 'z') {
			if (c) {
				free (buf);
				return s_make_indirect (S_CSC_PARSE_ERROR,
							"'z' in wrong place");
			}
			q += 4;
			continue;
		} else if (*p >= 0x21 && *p <= 0x75)
			in [c++] = *p - 0x21;
		else {
			free (buf);
			return s_make_indirect (S_CSC_PARSE_ERROR,
						"bad char '%c'", *p);
		}

		if (c == 5) {
			decode_pkt (in, q);
			q += 4;
			c = 0;
		}
	}

	if (c) {
		int ii;

		if (c < 2)
			return s_make_indirect (S_CSC_PARSE_ERROR,
						"bad a85 length");

		for (ii = c; ii < 5; ii++) {
			in[ii] = 84;
		}
		decode_pkt (in, q);
		q += c - 1;
	}

	assert (q - buf <= dest_len);

	dest_len = q-buf;
	newbuf = realloc(buf, dest_len);
	if (newbuf != NULL) {
		buf = newbuf;
		q = buf+dest_len;
	}

	obj_init_malloc_region (dest, buf, q);

	return S_CSC_OK;
}
