/*
 * PDFlow -- flate.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <zlib.h>
#include <string.h>

#include "filters_priv.h"
#include "predictor.h"

static enum s_status encode (const struct obj_region *src,
			     struct obj_region *dest,
			     struct xref_db *rdb,
			     struct obj *parm);
static enum s_status decode (const struct obj_region *src,
			     struct obj_region *dest,
			     struct xref_db *rdb,
			     pdf_integer_t dl,
			     struct obj *parm);

struct flt_interface flt_flate_ifc = {
	"FlateDecode", encode, decode
};

static void *f_alloc (void *opaque, uInt items, uInt size)
{
	return calloc (items, size);
}

static void f_free (void *opaque, void *data)
{
	free (data);
}

static enum s_status encode (const struct obj_region *src,
			     struct obj_region *dest,
			     struct xref_db *rdb,
			     struct obj *parm)
{
	/* FIXME: how to configure level? */
	int level = Z_DEFAULT_COMPRESSION;
	z_stream sin;
	void *out, *p;
	size_t outSize;
	int res;

	sin.next_in = (void *) src -> begin;
	sin.avail_in = src -> end - src -> begin;

	sin.zalloc = f_alloc;
	sin.zfree = f_free;
	sin.opaque = NULL;

	res = deflateInit(&sin, level);
	if (res != Z_OK) {
		return s_make_indirect(S_CSC_INTERNAL_ERROR,
				"zlib initialization error %d: %s", res, sin.msg);
	}

	outSize = deflateBound(&sin, sin.avail_in);
	out = malloc(outSize);
	if (out == NULL)
		return S_CSC_OUT_OF_MEMORY;

	sin.next_out = out;
	sin.avail_out = outSize;

	do {
		res = deflate (&sin, Z_FINISH);
		if (res == Z_STREAM_END || res != Z_OK)
			break;
		sin.avail_out = 1024;
		outSize += sin.avail_out;
		p = realloc (out, outSize);
		if (p == NULL) {
			free (out);
			out = NULL;
			break;
		} else {
			out = p;
		}
	} while (res == Z_OK);

	deflateEnd (&sin);

	if (res == Z_OK && out == NULL)
		return S_CSC_OUT_OF_MEMORY;

	if (res != Z_STREAM_END) {
		if (out != NULL)
			free (out);

		return s_make_indirect (S_CSC_INTERNAL_ERROR,
				"zlib error %d: %s", res, sin.msg);
	}

	p = realloc (out, sin.total_out);
	if (p == NULL)
		return S_CSC_OUT_OF_MEMORY;

	out = p;

	obj_init_malloc_region (dest, out, out + sin.total_out);
	if (parm != NULL) {
		res = pred_decode (dest, rdb, parm);
		if (res != S_CSC_OK) {
			obj_region_destroy (dest);
			return res;
		}
	}

	return S_CSC_OK;
}

static enum s_status decode (const struct obj_region *src,
			     struct obj_region *dest,
			     struct xref_db *rdb,
			     pdf_integer_t dl,
			     struct obj *parm)
{
	Bytef buf [1024];
	z_stream sin;
	void *out, *p;
	size_t valid;
	int res;

	sin.next_in = (void *) src -> begin;
	sin.avail_in = src -> end - src -> begin;

	sin.zalloc = f_alloc;
	sin.zfree = f_free;
	sin.opaque = NULL;

	if (inflateInit (&sin) != Z_OK)
		return s_make_indirect (S_CSC_PARSE_ERROR,
					sin.msg);

	out = NULL;
	valid = 0;
	do {
		sin.next_out = buf;
		sin.avail_out = sizeof (buf);

		res = inflate (&sin, Z_SYNC_FLUSH);
		if (out == NULL) {
			out = malloc (sin.total_out);
			if (out == NULL)
				break;
		} else {
			p = realloc (out, sin.total_out);
			if (p == NULL) {
				free (out);
				out = NULL;
				break;
			} else
				out = p;
		}
		memcpy (out + valid, buf, sin.total_out - valid);
		valid = sin.total_out;
	} while (res == Z_OK);

	inflateEnd (&sin);

	if (res == Z_OK && out == NULL)
		return S_CSC_OUT_OF_MEMORY;

	if (res != Z_STREAM_END) {
		if (out != NULL)
			free (out);

		return s_make_indirect (S_CSC_PARSE_ERROR,
					sin.msg);
	}

	obj_init_malloc_region (dest, out, out + valid);
	if (parm != NULL) {
		res = pred_decode (dest, rdb, parm);
		if (res != S_CSC_OK) {
			obj_region_destroy (dest);
			return res;
		}
	}

	return S_CSC_OK;
}
