/*
 * PDFlow -- lzw.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>

#include <string.h>
#include <stdlib.h>

#include "fmt/fields.h"
#include "filters_priv.h"
#include "predictor.h"

struct lzw_data {
#include "lzw.tab.h"
};

#define XREF_IMPLEMENTATION struct lzw_data
#include "lzw.tab.h"
#undef XREF_IMPLEMENTATION

static enum s_status encode (const struct obj_region *src, 
			     struct obj_region *dest,
			     struct xref_db *rdb,
			     struct obj *parm);
static enum s_status decode (const struct obj_region *src, 
			     struct obj_region *dest,
			     struct xref_db *rdb,
			     pdf_integer_t dl,
			     struct obj *parm);

struct flt_interface flt_lzw_ifc = {
	"LZWDecode", encode, decode
};

struct entry {
	unsigned int idx;
	unsigned len; 
};

struct input_data {
	const unsigned char *p, *end;
	int shift, clen, entries;
};

struct output_data {
	unsigned char *buf;
	unsigned idx, size;
};

const int threshold [] = {
	512, 1024, 2048
};

static enum s_status encode (const struct obj_region *src, 
			     struct obj_region *dest,
			     struct xref_db *rdb,
			     struct obj *parm)
{
	return S_CSC_NOT_SUPPORTED;
}

static void reset_input (struct input_data *in)
{
	in -> entries = 258;
	in -> clen = 9;
}

static void init_input (struct input_data *in,
			const struct obj_region *src)
{
	in -> p = src -> begin;
	in -> end = src -> end;
	in -> shift = 7;
	reset_input (in);
}

static int init_output (struct output_data *out,
			const struct obj_region *src,
			pdf_integer_t dl)
{
	out -> idx = 0;
	out -> size = dl > 0 ? dl : (src -> end - src -> begin) * 3;
	out -> buf = malloc (out -> size);
	if (out -> buf == NULL)
		return 0;

	return 1;
}

static int get (struct input_data *in, unsigned *inc)
{
	int i, bit;

	*inc = 0;
	for (i = in -> clen - 1; i >= 0; i--) {
		if (in -> p >= in -> end)
			return 0;

		bit = *in -> p & (1 << in -> shift) ? 1 : 0;
		*inc |= bit << i;
		if (!in -> shift--) {
			in -> shift = 7;
			in -> p++;
		}
	}
	
	return 1;
}

static int ensure (struct output_data *out, int outs)
{
	unsigned char *bbuf; 
	
	while (out -> idx + outs >= out -> size) {
		out -> size += 1024;
		bbuf = realloc (out -> buf, out -> size);
		if (bbuf == NULL)
			return 0;
		out -> buf = bbuf;
	}

	return 1;
}

static void put (struct output_data *out, unsigned char *outc, int outs)
{
	int i;

	for (i = 0; i < outs; i++)
		out -> buf [out -> idx++] = outc [i];
}

static enum s_status decode (const struct obj_region *src, 
			     struct obj_region *dest,
			     struct xref_db *rdb,
			     pdf_integer_t dl,
			     struct obj *parm)
{
	struct entry table [4096];
	struct output_data out;
	struct input_data in;
	struct lzw_data lzwd;
	unsigned inc;
	unsigned char *outc, c;
	enum s_status res;
	int outs;

	if (src -> begin + 1  >= src -> end)
		return s_make_indirect (S_CSC_PARSE_ERROR, "Bad LZW stream");

	if (parm != NULL && parm -> type == OT_DICTIONARY) {
		res = xref_fields_parse (rdb, NULL, NULL, parm,
					 &lzw_field, &lzwd);
		if (res != S_CSC_OK)
			return res;
	}  else
		lzwd.early = 1;

	if (lzwd.early)
		lzwd.early = 1;
	
	init_input (&in, src);
	if (!init_output (&out, src, dl))
		return S_CSC_OUT_OF_MEMORY;

	if (!get (&in, &inc) || inc != 256) {
		free (out.buf);
		return s_make_indirect (S_CSC_PARSE_ERROR, "Bad LZW stream");
	}

	while (1) {
		if (!get (&in, &inc)) {
			free (out.buf);
			return s_make_indirect (S_CSC_PARSE_ERROR, 
						"Truncated LZW stream");
		}
		if (inc < 256) {
			c = inc;
			outs = 1;
			if (!ensure (&out, outs)) {
				free (out.buf);
				return S_CSC_OUT_OF_MEMORY;
			}
			if (c != inc)
				abort ();
			outc = &c;
		} else if (inc >= 258 && inc < in.entries) {
			outs = table [inc].len;
			if (!ensure (&out, outs)) {
				free (out.buf);
				return S_CSC_OUT_OF_MEMORY;
			}
			outc = out.buf + table [inc].idx;
		} else if (inc == 257)
			break;
		else /*if (*p == 256) */ {
			reset_input (&in);
			continue;
		}

		if (in.entries < 4096) {
			table [in.entries].idx = out.idx;
			table [in.entries].len = outs + 1;
			if ((in.entries++ + lzwd.early) == 
			    threshold [in.clen - 9])
				in.clen++;
		}

		put (&out, outc, outs);
	}

	obj_init_malloc_region (dest, out.buf, out.buf + out.idx);
	if (parm != NULL) {
		res = pred_decode (dest, rdb, parm);
		if (res != S_CSC_OK) {
			obj_region_destroy (dest);
			return res;
		}
	}

	return S_CSC_OK;
}
