/*
 * PDFlow -- predictor.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdint.h>
#include <string.h>
#include <stdlib.h>

#include "fmt/fields.h"
#include "predictor.h"

struct pred_data {
#include "predictor.tab.h"
};

#define XREF_IMPLEMENTATION struct pred_data
#include "predictor.tab.h"
#undef XREF_IMPLEMENTATION

typedef void (*p_png_fun_t) (uint8_t *q, const uint8_t *p,
			     int ridx, int rowsz, int ofs);

static void p_png_none (uint8_t *q, const uint8_t *p, 
			int ridx, int rowsz, int ofs);
static void p_png_sub (uint8_t *q, const uint8_t *p, 
		       int ridx, int rowsz, int ofs);
static void p_png_up (uint8_t *q, const uint8_t *p, 
		      int ridx, int rowsz, int ofs);
static void p_png_avg (uint8_t *q, const uint8_t *p,
		       int ridx, int rowsz, int ofs);
static void p_png_paeth (uint8_t *q, const uint8_t *p,
			 int ridx, int rowsz, int ofs);

static const p_png_fun_t png_enc [] = { 
	p_png_none, p_png_sub, p_png_up, p_png_avg, p_png_paeth 
};


#define min3(a,b,c) ((a) < (b) ? 		\
		    (a) < (c) ? (a) : (c) :	\
		    (b) < (c) ? (b) : (c))

struct bmp_data {
	uint8_t *p, *end;
	int crow, bpc;
	int mask;
};

static int unpack (struct bmp_data *in, uint32_t *buf)
{
	uint32_t u;
	int i, j;

	for (i = 0; i < in -> crow; i++) {
		u = 0;
		for (j = 0; j < in -> bpc; j++) {
			u <<= 1;
			u |= *in -> p & in -> mask;
			if (!(in -> mask >>= 1)) {
				in -> mask = 0x80;
				in -> p++;
				if (in -> p >= in -> end)
					return 0;
			}
				
		}
		*(buf++) = u;
	}

	return 1;
}

static int pack (uint32_t *buf, struct bmp_data *out)
{
	uint32_t u;
	int i, j, umask;

	for (i = 0; i < out -> crow; i++) {
		u = *buf;
		umask = 1 << out -> bpc;
		for (j = 0; j < out -> bpc; j++) {
			if (u & umask)
				*out -> p |= out -> mask;
			else
				*out -> p &= ~out -> mask;
			umask >>= 1;
			if (!(out -> mask >>= 1)) {
				out -> mask = 0x80;
				out -> p++;
				if (out -> p >= out -> end)
					return 0;
			}
				
		}
		buf++;
	}

	return 1;
}

static void pack_start (struct bmp_data *bmp, struct obj_region *reg, 
			int colors, int bpc, int cols)
{
	bmp -> p = (uint8_t *) reg -> begin;
	bmp -> end = (uint8_t *) reg -> end;
	bmp -> crow = colors * cols;
	bmp -> bpc = bpc;
	bmp -> mask = 0x80;
}

static enum s_status p_tiff2 (struct obj_region *reg, int colors, 
			      int bpc, int cols)
{
	struct bmp_data rin, win;
	uint32_t *pbuf;
	int i, crow;

	pack_start (&rin, reg, colors, bpc, cols);
	pack_start (&win, reg, colors, bpc, cols);

	pbuf = calloc ((colors * bpc * cols + 7) >> 3, sizeof (uint32_t));
	if (pbuf == NULL)
		return S_CSC_OUT_OF_MEMORY;

	crow = colors * cols;
	while (unpack (&rin, pbuf)) {
		for (i = colors; i < crow; i++)
			pbuf [i] += pbuf [i - colors];
		pack (pbuf, &win);
	}

	free (pbuf);

	return S_CSC_OK;
}

static enum s_status p_tiff (int pred, struct obj_region *reg, int colors, 
			     int bpc, int cols)
{
	switch (pred) {
	case 0:
		return S_CSC_OK;
	case 1:
		return p_tiff2 (reg, colors, bpc, cols);
	}

	return S_CSC_NOT_SUPPORTED;
}

static p_png_fun_t get_png_predictor (int pred)
{
	if (pred < 0 || pred >= sizeof (png_enc) / sizeof (p_png_fun_t))
		return NULL;

	return png_enc [pred];
}

static void p_png_none (uint8_t *q, const uint8_t *p, 
			int ridx, int rowsz, int ofs)
{
	memmove (q, p, rowsz);
}

static void p_png_sub (uint8_t *q, const uint8_t *p, 
		       int ridx, int rowsz, int ofs)
{
	int i;
	
	for (i = ofs; i < rowsz; i++)
		q [i] = p [i] + q [i - ofs];
}

static void p_png_up (uint8_t *q, const uint8_t *p, 
		      int ridx, int rowsz, int ofs)
{
	int i;

	if (ridx)
		for (i = 0; i < rowsz; i++)
			q [i] = p [i] + q [i - rowsz];
	else
		memmove (q, p, rowsz);
}

static void p_png_avg (uint8_t *q, const uint8_t *p, 
		       int ridx, int rowsz, int ofs)
{
	int i;

	if (ridx)
		for (i = 0; i < rowsz; i++)
			q [i] = p [i] + (q [i - ofs] + q [i - rowsz]) / 2;
	else
		for (i = ofs; i < rowsz; i++)
			q [i] = p [i] + q [i - ofs] / 2;
}

static void p_png_paeth (uint8_t *q, const uint8_t *p, 
			 int ridx, int rowsz, int ofs)
{
	int i, e, da, db, dc;
	
	if (ridx) {
		for (i = rowsz; i >= ofs; i++) {
			e = q [i - ofs] + q [i - rowsz] - q [i - ofs - rowsz];

			da = abs (e - q [i - ofs]);
			db = abs (e - q [i - rowsz]);
			dc = abs (e - q [i - ofs - rowsz]);

			e = min3 (da, db, dc);
			if (e == da) 
				q [i] = p [i] + q [i - ofs];
			else if (e == db)
				q [i] = p [i] + q [i - rowsz];
			else
				q [i] = p [i] + q [i - ofs - rowsz];
		}
	} else
		for (i = ofs; i < rowsz; i++)
			q [i] = p [i] + q [i - ofs];

}

static enum s_status p_png (int pred, struct obj_region *reg, int colors, 
			    int bpc, int cols)
{
	p_png_fun_t p_fun;
	const uint8_t *p;
	uint8_t *q;
	int rowsz, rows, ridx, ofs;

	rowsz = (cols * bpc * colors + 7) / 8;
	ofs = (bpc * colors + 7) / 8;

	p_fun = get_png_predictor (pred);	
	if (p_fun == NULL && pred != 5)
		return S_CSC_NOT_SUPPORTED;
	
	rows = (reg -> end - reg -> begin) / (rowsz + 1);
	p = reg -> begin + 1;
	q = (uint8_t *) reg -> begin;
	for (ridx = 0; ridx < rows; ridx++) {
		if (pred == 5) {
			p_fun = get_png_predictor (p [-1]);
			if (p_fun == NULL)
				return S_CSC_NOT_SUPPORTED;
		}
		p_fun (q, p, ridx, rowsz, ofs);

		p += rowsz + 1;
		q += rowsz;
	}

	reg -> end = q;

	return S_CSC_OK;
}


enum s_status pred_decode (struct obj_region *reg,
			   struct xref_db *rdb, struct obj *parm)
{
	struct pred_data predd;
	enum s_status res;

	res = xref_fields_parse (rdb, NULL, NULL, parm, &pred_field, &predd);
	if (res != S_CSC_OK)
		return res;

	if (predd.predictor < 10)
		res = p_tiff (predd.predictor, reg, predd.colors, 
			      predd.bpc, predd.columns);
	else if (predd.predictor < 20)
		res = p_png (predd.predictor - 10, reg, predd.colors, 
			     predd.bpc, predd.columns);
	else
		res = S_CSC_NOT_SUPPORTED;
		
	if (res == S_CSC_NOT_SUPPORTED)
		return obj_error (parm, "Unsupported predictor: %d", 
				  predd.predictor);

	return S_CSC_OK;
}
