/*
 * PDFlow -- fields.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <version.h>
#include <assert.h>

#include "fmt/fields.h"

#define RELOC(fd, base) (((uint8_t *) base) + (int) (fd) -> field)
#define MAX_FIELDS 30

#define DRF_DRY_RUN 1
#define DRF_REPLACE 2

static enum s_status tag_parser (struct xref_db *rdb, 
				 struct xref_bind *robj,
				 struct ref_owner *p,
				 struct obj *src,
				 struct xref_field *fd,
				 const char *fname, int line);
static enum s_status tag_serializer (struct xref_db *rdb, 
				     struct xref_bind *robj,
				     struct v_selection *sel,
				     struct obj *dest,
				     enum xref_changed *changed,
				     struct xref_field *fd);

static enum s_status array_parser (struct xref_db *rdb, 
				   struct xref_bind *robj,
				   struct ref_owner *p,
				   struct obj *src,
				   struct xref_field *fd,
				   int dflags,
				   const char *fname, int line);
static enum s_status array_serializer (struct xref_db *rdb, 
				       struct xref_bind *robj,
				       struct v_selection *sel,
				       struct obj *dest,
				       enum xref_changed *changed,
				       struct xref_field *fd);
static void array_funref (struct ref_owner *p, struct xref_field *fd);

static enum s_status obj_recursive_serializer (struct xref_db *rdb, 
					       struct xref_bind *robj,
					       struct v_selection *sel,
					       struct obj *dest,
					       enum xref_changed *changed,
					       struct xref_field *fd);

static enum s_status pstable_parser (struct xref_db *rdb, 
				     struct xref_bind *robj,
				     struct ref_owner *p,
				     struct obj *src,
				     struct xref_field *fd,
				     int dflags,
				     const char *fname, int line);
static enum s_status pstable_serializer (struct xref_db *rdb, 
					 struct xref_bind *robj,
					 struct v_selection *sel,
					 struct obj *dest,
					 enum xref_changed *changed,
					 struct xref_field *fd);
static void pstable_funref (struct ref_owner *p, struct xref_field *fd);

static enum s_status bool_parser (struct xref_db *rdb, 
				  struct xref_bind *robj,
				  struct ref_owner *p,
				  struct obj *src,
				  struct xref_field *fd,
				  const char *fname, int line);
static enum s_status bool_serializer (struct xref_db *rdb, 
				      struct xref_bind *robj,
				      struct v_selection *sel,
				      struct obj *dest,
				      enum xref_changed *changed,
				      struct xref_field *fd);
static enum s_status int_parser (struct xref_db *rdb, 
				 struct xref_bind *robj,
				 struct ref_owner *p,
				 struct obj *src,
				 struct xref_field *fd,
				 const char *fname, int line);
static enum s_status int_serializer (struct xref_db *rdb, 
				     struct xref_bind *robj,
				     struct v_selection *sel,
				     struct obj *dest,
				     enum xref_changed *changed,
				     struct xref_field *fd);
static enum s_status real_parser  (struct xref_db *rdb, 
				   struct xref_bind *robj,
				   struct ref_owner *p,
				   struct obj *src,
				   struct xref_field *fd,
				   const char *fname, int line);
static enum s_status real_serializer (struct xref_db *rdb, 
				      struct xref_bind *robj,
				      struct v_selection *sel,
				      struct obj *dest,
				      enum xref_changed *changed,
				      struct xref_field *fd);

static enum s_status name_parser (struct xref_db *rdb, 
				  struct xref_bind *robj,
				  struct ref_owner *p,
				  struct obj *src,
				  struct xref_field *fd,
				  const char *fname, int line);
static enum s_status dictionary_parser (struct xref_db *rdb, 
					struct xref_bind *robj,
					struct ref_owner *p,
					struct obj *src,
					struct xref_field *fd,
					const char *fname, int line);
static enum s_status stream_parser (struct xref_db *rdb, 
				    struct xref_bind *robj,
				    struct ref_owner *p,
				    struct obj *src,
				    struct xref_field *fd,
				    const char *fname, int line);
static enum s_status obj_parser (struct xref_db *rdb, 
				 struct xref_bind *robj,
				 struct ref_owner *p,
				 struct obj *src,
				 struct xref_field *fd,
				 const char *fname, int line);
static enum s_status obj_serializer  (struct xref_db *rdb, 
				      struct xref_bind *robj,
				      struct v_selection *sel,
				      struct obj *dest,
				      enum xref_changed *changed,
				      struct xref_field *fd);
static void obj_funref (struct ref_owner *p, struct xref_field *fd);

static void destroy_dict (struct ref_owner *p, struct xref_field *fd);
static void destroy_array (struct ref_owner *p, struct xref_field *fd);

static enum s_status parse_dict (struct xref_db *rdb, 
				 struct xref_bind *robj,
				 struct ref_owner *p,
				 struct obj *obj,
				 struct xref_field *fd,
				 int dflags,
				 const char *fname, int line);

static enum s_status parse_stream (struct xref_db *rdb, 
				   struct xref_bind *robj,
				   struct ref_owner *p,
				   struct obj *obj,
				   struct xref_field *fd,
				   int dflags,
				   const char *fname, int line);

static enum s_status parse_array (struct xref_db *rdb, 
				  struct xref_bind *robj,
				  struct ref_owner *p,
				  struct obj *obj, 
				  struct xref_field *fd,
				  int dflags,
				  const char *fname, int line);

static enum s_status ser_dict (struct xref_db *rdb, struct xref_bind *robj,
			       struct v_selection *sel, struct obj *tobj, 
			       enum xref_changed *changed, 
			       struct xref_field *fd);

static enum s_status ser_stream (struct xref_db *rdb, struct xref_bind *robj,
				 struct v_selection *sel, struct obj *tobj, 
				 enum xref_changed *changed, 
				 struct xref_field *fd);

static enum s_status ser_array (struct xref_db *rdb, struct xref_bind *robj,
				struct v_selection *sel, struct obj *tobj, 
				enum xref_changed *changed, 
				struct xref_field *fd);

struct xref_pstable xref_tag_pstable = {
	0, tag_parser, tag_serializer, NULL
};

struct xref_pstable xref_bool_pstable = {
	sizeof (pdf_bool_t),
	bool_parser, bool_serializer, NULL,
};

struct xref_pstable xref_int_pstable = {
	sizeof (pdf_integer_t),
	int_parser, int_serializer, NULL,
};

struct xref_pstable xref_real_pstable = {
	sizeof (pdf_real_t),
	real_parser, real_serializer, NULL,
};

struct xref_pstable xref_name_pstable = {
	sizeof (struct obj),
	name_parser, obj_serializer, obj_funref
};

struct xref_pstable xref_dictionary_pstable = {
	sizeof (struct obj),
	dictionary_parser, obj_serializer, obj_funref
};

struct xref_pstable xref_stream_pstable = {
	sizeof (struct obj),
	stream_parser, obj_serializer, obj_funref
};

struct xref_pstable xref_obj_pstable = {
	sizeof (struct obj),
	obj_parser, obj_serializer, obj_funref
};

const struct xref_state {
	const char *state;
}	xref_mandatory_tag = { "mandatory" }, 
	xref_optional_tag = {"optional" };

#ifdef DEBUG
int *xref_p_cast (xref_parser_t f)
{
	return (int *) f;
}

xref_serializer_t xref_s_cast (xref_serializer_t f)
{
	return f;
}
#endif

/* ============
 * Bind support
 * ============ */
void xref_bind_init (struct xref_bind *robj)
{
	robj -> size = 0;
}

void xref_bind_data (struct xref_bind *robj, const char *name, void *p)
{
	assert (robj -> size < XREF_BIND_SIZE);
	robj -> e [robj -> size].name = name;
	robj -> e [robj -> size].p = p;
	robj -> size++;
}

void *xref_bind_get (struct xref_bind *robj, const char *name)
{
	int i;

	for (i = 0; i < robj -> size; i++)
		if (robj -> e [i].name == name)
			return robj -> e [i].p;

	assert (0);
}

void xref_init_field (struct xref_field *fd, void *field, void *data)
{
	fd -> name = "<onthefly>";
	fd -> field = field;
	fd -> data = data;
	fd -> ser = XREF_S_INLINE;
	fd -> des = XREF_D_UNWRAP;
	fd -> major = 1;
	fd -> minor = 0;
	fd -> state = &xref_mandatory_tag;
}

/* ============
 * Array map/unmap support
 * ============ */
enum s_status xref__array_map (struct xref_db *rdb, struct xref_bind *robj, 
			       struct ref_owner *p, struct obj *src, 
			       void *mapped_array, enum obj_type ot,
			       size_t elsize, const char *label,
			       xref_parser_t parse, 
			       xref_funref_t funref, void *pdata,
			       const char *fname, int line)
{
	struct xref_mapped_array ma;
	struct obj_array_node *an, *anp;
	struct xref_field fd;
	struct obj eobj;
	uint8_t *field;
	enum s_status res;
	int i, pres;

	switch (src -> type) {

	case OT_NULL:
		XREF_A_SET_NULL (&ma);
		break;

	case OT_ARRAY:
		i = 0;
		for (an = src -> val.array.head; an != NULL; an = an -> next)
			i++;

		field = ma.data = calloc (i, elsize);
		if (ma.data == NULL)
			return S_CSC_OUT_OF_MEMORY;

		ma.n = i;
		
		res = S_CSC_OK;
		for (an = src -> val.array.head; an != NULL; an = an -> next) {
			if (rdb != NULL)
				res = xref_resolve (rdb, &an -> obj, ot, 
						    &eobj, &pres, label);
			else
				res = obj_dup (&an -> obj, &eobj);

			if (res != S_CSC_OK) 
				break;

			xref_init_field (&fd, field, pdata);
			res = parse (rdb, robj, p, &eobj, &fd, fname, line);
			obj_destroy (&eobj);

			if (res != S_CSC_OK)
				break;

			field += elsize;
		}
		
		if (res != S_CSC_OK) {
			field = ma.data;
			for (anp = src -> val.array.head; 
			     anp != an; anp = anp -> next) {

				xref_init_field (&fd, field, pdata);
				if (funref != NULL)
					funref (p, &fd);

				field += elsize;
			}

			free (ma.data);

			XREF_A_SET_NULL (&ma);

			return res;
		}
		break;

	default:
		res = xref_resolve (rdb, src, ot, &eobj, &pres, label);
		if (res != S_CSC_OK)
			return res;

		ma.data = malloc (elsize);
		if (ma.data == NULL)
			return S_CSC_OUT_OF_MEMORY;
		memset (ma.data, 0, elsize);
		ma.n = 1;

		xref_init_field (&fd, ma.data, pdata);
		res = parse (rdb, robj, p, &eobj, &fd, fname, line);
		obj_destroy (&eobj);

		if (res != S_CSC_OK) {
			free (ma.data);
			return res;
		}
	}

	*(struct xref_mapped_array *) mapped_array = ma;

	return S_CSC_OK;
}

enum s_status xref__array_map_ps (struct xref_db *rdb, struct xref_bind *robj,
				  struct ref_owner *p, struct obj *src, 
				  void *mapped_array, const char *label, 
				  struct xref_pstable *pst, void *data,
				  const char *fname, int line)
{
	return xref__array_map (rdb, robj, p, src, mapped_array, OT_ANY,
				pst -> size, label, pst -> parser, 
				pst -> funref, data, fname, line);
}

enum s_status xref_array_unmap (struct xref_db *rdb, struct xref_bind *robj,
				struct v_selection *sel, void *mapped_array,
				struct obj *dest, size_t elsize, int simplify,
				xref_serializer_t ser, void *pdata)
{
	struct obj_array array;
	struct xref_field fd;
	struct xref_mapped_array *ma;
	enum xref_changed changed;
	enum s_status res;
	struct obj obj;
	uint8_t *p;
	int i;

	ma = mapped_array;
	p = ma -> data;

	if (XREF_A_IS_NULL (ma)) {
		obj_set_null (dest, &f_null);
		return S_CSC_OK;
	}

	if (simplify) {
		switch (XREF_A_SIZE (ma)) {
		case 0:
			obj_set_null (dest, &f_null);
			return S_CSC_OK;
		case 1:
			xref_init_field (&fd, p, pdata);
			changed = XREF_C_UNKNOWN;
			return ser (rdb, robj, sel, dest, &changed, &fd);
		}
	}

	res = S_CSC_OK;
	obj_init_array (&array);
	for (i = 0; i < XREF_A_SIZE (ma); i++) {
		xref_init_field (&fd, p, pdata);
		changed = XREF_C_UNKNOWN;
		res = ser (rdb, robj, sel, &obj, &changed, &fd);
		if (res != S_CSC_OK)
			break;

		res = obj_xfer_t_array (&array, &obj, NULL);
		if (res != S_CSC_OK)
			break;

		p += elsize;
	}

	if (res != S_CSC_OK) {
		obj_array_destroy (&array);
		return res;
	}

	obj_set_array (dest, &f_null, &array);

	return S_CSC_OK;
}	

enum s_status xref_array_unmap_ps (struct xref_db *rdb, struct xref_bind *robj,
				   struct v_selection *sel, void *mapped_array,
				   struct obj *dest, int simplify,
				   struct xref_pstable *pst, void *data)
{
	return xref_array_unmap (rdb, robj, sel, mapped_array, dest, 
				 pst -> size, simplify, 
				 pst -> serializer, data);
}

void xref_array_map_rel (void *mapped_array, struct ref_owner *p,
			 size_t elsize, xref_funref_t funref, void *pdata)
{
	struct xref_mapped_array *ma;
	struct xref_field fd;
	uint8_t *field;
	int i;

	ma = mapped_array;

	if (ma -> data == NULL)
		return;

	field = ma -> data;
	for (i = 0; i < XREF_A_SIZE (ma); i++) {
		xref_init_field (&fd, field, pdata);
		if (funref != NULL)
			funref (p, &fd);
		field += elsize;
	}

	free (ma -> data);

	ma -> n = 0;
	ma -> data = NULL;
}

void xref_array_map_rel_ps (void *mapped_array, struct ref_owner *p,
			    struct xref_pstable *pst, void *data)
{
	xref_array_map_rel (mapped_array, p, pst -> size, 
			    pst -> funref, data);
}

static enum s_status resolve_dyn (enum xref_op op, struct xref_field *fd, 
				  void *refdata, int *stop)
{
	enum s_status res;
	void *cdata;

	*stop = 0;
	while (fd -> des == XREF_D_DYNAMIC) {
		/* refdata points to the parent field, cdata to the child */
		cdata = fd -> field;
		res = fd -> u.dyn (op, fd, refdata, stop);
		if (res != S_CSC_OK)
			return res;
		if (*stop)
			return S_CSC_OK;
		/* The dynamic interface is relative to the child */
		fd -> field = RELOC (fd, cdata);
		refdata = fd -> data;
	}

	return S_CSC_OK;
}

static void update_changed (enum xref_changed *changed, enum xref_changed lc)
{
	if (lc != XREF_C_UNCHANGED && *changed != XREF_C_UNKNOWN)
		*changed = lc;
}

/* ============
 * Destroy methods
 * ============ */
static void destroy_nest (struct ref_owner *p, struct xref_field *fd)
{
	switch (fd -> ot) {
	case OT_NULL:
	case OT_STREAM:
	case OT_DICTIONARY:
		destroy_dict (p, fd);
		break;

	case OT_ARRAY:
		destroy_array (p, fd);
		break;

	default:
		assert (0);
	}
}

static void destroy_field (struct ref_owner *p, struct xref_field *fd,
			   void *base)
{
	struct xref_field sfd;
	enum s_status res;
	struct obj obj;
	int stop;

	sfd = *fd;
	sfd.field = RELOC (fd, base);

	fd = &sfd;

	res = resolve_dyn (XREF_OP_DESTROY, fd, base, &stop);
	/* res != S_CSC_OK violates contract */
	assert (res == S_CSC_OK);

	if (stop)
		return;

	switch (fd -> des) {
	case XREF_D_UNWRAP:
		obj_wrap (fd -> field, fd -> ot, &f_null, &obj);
		obj_destroy (&obj);
		break;

	case XREF_D_OBJ:
		obj_destroy (fd -> field);
		*(struct obj *) fd -> field = obj_null;
		break;

	case XREF_D_PSTABLE:
		pstable_funref (p, fd);
		break;

	case XREF_D_ARRAY:
		array_funref (p, fd);
		break;

	case XREF_D_NEST:
		destroy_nest (p, fd);
		break;

	case XREF_D_REGION:
		xref_stream_destroy (fd -> field, p);
		break;

	case XREF_D_DYNAMIC:
		assert (0);
	}
}

static void destroy_fields (struct ref_owner *p, void *refdata, 
			    struct xref_field *fd, int limit)
{
	/* Going backwards makes destructor's environment
	 * more consistent */
	if (limit < 0) {
		limit = 0;
		while (fd [limit].name != NULL)
			limit++;
	}
	while (limit--)
		destroy_field (p, fd + limit, refdata);
}

static void destroy_dict (struct ref_owner *p, struct xref_field *fd)
{
	destroy_fields (p, fd -> field, fd -> u.fdesc -> fd, -1);
}

static void destroy_array (struct ref_owner *p, struct xref_field *fd)
{
	destroy_fields (p, fd -> field, fd -> u.fdesc -> fd, -1);
}

void xref_fields_destroy (struct xref_field *fd, void *base,
			  struct ref_owner *p)
{
	destroy_field (p, fd, base);
}

/* ============
 * Parse methods
 * ============ */
static enum s_status parse_nest (struct xref_db *rdb, struct xref_bind *robj,
				 struct ref_owner *p, struct obj *obj, 
				 struct xref_field *fd, int dflags,
				 const char *fname, int line)
{
	enum s_status res;

	switch (fd -> ot) {
	case OT_NULL:
	case OT_DICTIONARY:
		res = parse_dict (rdb, robj, p, obj, fd, dflags,
				  fname, line);
		if (res != S_CSC_OK)
			return res;
		break;

	case OT_STREAM:
		res = parse_stream (rdb, robj, p, obj, fd, dflags,
				    fname, line);
		if (res != S_CSC_OK)
			return res;
		break;

	case OT_ARRAY:
		res = parse_array (rdb, robj, p, obj, fd, dflags,
				   fname, line);
		if (res != S_CSC_OK)
			return res;
		break;

	default:
		assert (0);
	}

	if (fd -> u.fdesc -> val != NULL) {
		res = fd -> u.fdesc -> val (fd -> field);
		if (res != S_CSC_OK) {
			destroy_nest (p, fd);
			return res;
		}
	}

	return S_CSC_OK;
}

static enum s_status parse_field (struct xref_db *rdb, 
				  struct xref_bind *robj,
				  struct ref_owner *p,
				  struct obj *refobj,
				  struct obj *sobj,
				  struct xref_field *fd,
				  void *base,
				  int dflags,
				  const char *fname, int line)
{
	struct xref_field sfd;
	struct obj obj;
	enum s_status res;
	int present, stop;

	sfd = *fd;
	sfd.field = RELOC (fd, base);

	fd = &sfd;

	res = resolve_dyn (XREF_OP_PARSE, fd, base, &stop);
	if (res != S_CSC_OK)
		return res;

	if (stop)
		return S_CSC_OK;

	if (sobj != NULL) {
		res = xref_resolve_no_cast
			(rdb, sobj, fd -> simplify ? OT_ANY : fd -> ot, 
			 &obj, &present, fd -> name);
		if (res != S_CSC_OK)
			return res;
	} else
		obj_set_null (&obj, &f_null);

	if (obj.type == OT_NULL && !(dflags & DRF_DRY_RUN) &&
	    fd -> state == &xref_mandatory_tag) {
		obj_destroy (&obj);
		return obj_error (refobj, "Missing mandatory field (%s)",
				  fd -> name);
	}

	switch (fd -> des) {
	case XREF_D_UNWRAP:
		obj_cast (&obj, fd -> ot);
		if (obj.type != OT_NULL)
			obj_unwrap (&obj, fd -> field);
		else
			obj_unwrap (&fd -> defval, fd -> field);
		break;

	case XREF_D_OBJ:
		obj_cast (&obj, fd -> ot);
		if (dflags & DRF_REPLACE)
			obj_destroy (fd -> field);

		*(struct obj *) fd -> field = obj;
		break;

	case XREF_D_PSTABLE:
		obj_cast (&obj, fd -> ot);
		res = pstable_parser (rdb, robj, p, &obj, fd, 
				      dflags, fname, line);
		obj_destroy (&obj);
		if (res != S_CSC_OK)
			return res;
		break;

	case XREF_D_ARRAY:
		res = array_parser (rdb, robj, p, &obj, fd, 
				    dflags, fname, line);
		obj_destroy (&obj);
		if (res != S_CSC_OK)
			return res;
		break;

	case XREF_D_NEST:
		res = parse_nest (rdb, robj, p, &obj, fd, 
				  dflags, fname, line);
		obj_destroy (&obj);
		if (res != S_CSC_OK)
			return res;
		break;
		
	case XREF_D_REGION:
		obj_destroy (&obj);
		if (dflags & DRF_REPLACE)
			obj_stream_destroy (fd -> field);
		if (dflags & DRF_DRY_RUN) {
			memset (fd -> field, 0, sizeof (struct obj_stream));
			return S_CSC_OK;
		}
		assert (sobj -> type == OT_STREAM);
		res = xref_resolve_stream_xs (rdb, sobj, fd -> field, p);
		if (res != S_CSC_OK)
			return res;
		break;

	case XREF_D_DYNAMIC:
		assert (0);
	}

	return S_CSC_OK;
}

static struct obj *get_dict_node (struct obj *refobj,
				  struct obj_dictionary *dict,
				  const char *name,
				  struct obj_dict_node ***np)
{
	struct obj_dict_node **n;
	int nlen;

	nlen = strlen (name);
	if (!nlen)
		return refobj;

	for (n = &dict -> head; *n != NULL; n = &(*n) -> next) {
		if (nlen == (*n) -> key.size &&
		    !memcmp ((*n) -> key.s, name, nlen)) {
			break;
		}
	}

	if (np != NULL)
		*np = n;

	return *n != NULL ? &(*n) -> obj : NULL;
}

static enum s_status parse_dict (struct xref_db *rdb, 
				 struct xref_bind *robj,
				 struct ref_owner *p,
				 struct obj *refobj, 
				 struct xref_field *fd,
				 int dflags,
				 const char *fname, int line)
{
	struct obj_dictionary *dict;
	struct xref_field_desc *fdesc;
	struct xref_field sfd;
	struct obj *obj;
	enum s_status res;
	int i;

	dict = obj_get_rw_dictionary (refobj);
	if (dict == NULL)
		return obj_error (refobj, "Expecting dictionary-like type");

	fdesc = fd -> u.fdesc;

	for (i = 0; fdesc -> fd [i].name != NULL; i++) {

		sfd = fdesc -> fd [i];
		obj = get_dict_node (refobj, dict, sfd.name, NULL);
		if (obj != NULL || !(dflags & DRF_REPLACE)) {
			res = parse_field (rdb, robj, p, refobj, obj, &sfd, 
					   fd -> field, dflags, fname, line);
			if (res != S_CSC_OK) {
				destroy_fields (p, fd -> field, 
						fdesc -> fd, i);
				return res;
			}
		}
	}

	return S_CSC_OK;
}

static enum s_status parse_stream (struct xref_db *rdb, 
				   struct xref_bind *robj,
				   struct ref_owner *p,
				   struct obj *refobj, 
				   struct xref_field *fd,
				   int dflags,
				   const char *fname, int line)
{
	enum s_status res;

	if (refobj -> type != OT_NULL &&
	    refobj -> type != OT_STREAM)
		obj_error (refobj, "Stream expected");

	res = parse_dict (rdb, robj, p, refobj, fd, dflags, fname, line);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status parse_array (struct xref_db *rdb, 
				  struct xref_bind *robj,
				  struct ref_owner *p,
				  struct obj *refobj, 
				  struct xref_field *fd,
				  int dflags,
				  const char *fname, int line)
{
	struct xref_obj_array ma;
	struct xref_field_desc *fdesc;
	struct xref_field sfd;
	enum s_status res;
	struct obj *obj;
	int i;

	if (!fd -> simplify && 
	    refobj -> type != OT_ARRAY && refobj -> type != OT_NULL)
		return obj_error (refobj, "Expecting array");

	res = xref_array_map_ps (rdb, robj, NULL, refobj, &ma, 
				 "Element", &xref_obj_pstable, NULL);
	if (res != S_CSC_OK)
		return res;

	fdesc = fd -> u.fdesc;
	
	for (i = 0; fdesc -> fd [i].name != NULL; i++) {
		sfd = fdesc -> fd [i];
		obj = i < XREF_A_SIZE (&ma) ? &ma.data [i] : NULL;
		if (obj != NULL || !(dflags & DRF_REPLACE)) {
			res = parse_field (rdb, robj, p, refobj, obj, &sfd, 
					   fd -> field, dflags, fname, line);
			if (res != S_CSC_OK) {
				destroy_fields (p, fd -> field, 
						fdesc -> fd, i);
				xref_array_map_rel_ps (&ma, NULL, 
						       &xref_obj_pstable, 
						       NULL);
				return res;
			}
		}
	}

	xref_array_map_rel_ps (&ma, NULL, &xref_obj_pstable, NULL);

	return S_CSC_OK;
}

enum s_status xref__fields_parse (struct xref_db *rdb, struct xref_bind *robj,
				  struct ref_owner *p, struct obj *obj, 
				  struct xref_field *fd, void *base,
				  const char *fname, int line)
{
	enum s_status res;

	res = parse_field (rdb, robj, p, obj, obj, fd, base, 
			   obj -> type == OT_NULL ? DRF_DRY_RUN : 0,
			   fname, line);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

enum s_status xref__fields_replace (struct xref_db *rdb, 
				    struct xref_bind *robj,
				    struct ref_owner *p, struct obj *obj, 
				    struct xref_field *fd, void *base,
				    const char *fname, int line)
{
	enum s_status res;

	res = parse_field (rdb, robj, p, obj, obj, fd, base, 
			   DRF_REPLACE |
			   (obj -> type == OT_NULL ? DRF_DRY_RUN : 0),
			   fname, line);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

enum s_status xref_dict_is (struct xref_db *rdb, struct obj *obj,
			    const char *type, pdf_bool_t *ans)
{
	struct obj tobj;
	enum s_status res;

	res = xref_dict_resolve (rdb, obj, "Type", OT_NAME, &tobj);
	if (res != S_CSC_OK)
		return res;

	*ans = tobj.type != OT_NULL && 
		obj_string_equal (&tobj.val.string, type);
	obj_destroy (&tobj);

	return S_CSC_OK;
}

/* ============
 * Serialize methods
 * ============ */
static enum s_status ser_nest (struct xref_db *rdb, struct xref_bind *robj,
			       struct v_selection *sel, struct obj *tobj, 
			       enum xref_changed *changed, 
			       struct xref_field *fd)
{
	enum s_status res;

	if (!fd -> simplify && !obj_cast (tobj, fd -> ot))
		return obj_error (tobj, "Can't cast %s to %s",
				  obj_tag (tobj -> type), obj_tag (fd -> ot));

	switch (fd -> ot) {
	case OT_NULL:
	case OT_DICTIONARY:
		res = ser_dict (rdb, robj, sel, tobj, changed, fd);
		break;

	case OT_STREAM:
		res = ser_stream (rdb, robj, sel, tobj, changed, fd);
		break;

	case OT_ARRAY:
		res = ser_array (rdb, robj, sel, tobj, changed, fd);
		break;

	default:
		assert (0);
	}

	return res;
}

static enum s_status serialize_field (struct xref_db *rdb,
				      struct xref_bind *robj,
				      struct v_selection *sel,
				      struct obj *sobj, 
				      enum xref_changed *changed,
				      struct xref_field *fd, void *base)
{
	struct xref_field sfd;
	struct obj_stream *stream;
	enum s_status res;
	struct obj uobj, tobj;
	enum xref_changed fchanged;
	int stop;

	sfd = *fd;
	sfd.field = RELOC (fd, base);

	fd = &sfd;

	/* First: version check */
	if (sel != NULL) {
		if (!v_supported (sel -> pref, fd -> major, fd -> minor)) {
			if (fd -> state == &xref_mandatory_tag)
				return s_make_indirect 
					(S_CSC_NOT_SUPPORTED,
					 "Need at least PDF %d.%d "
					 "to store this object",
					 fd -> major, fd -> minor);
			obj_destroy (sobj);
			obj_set_null (sobj, &f_null);
			return S_CSC_OK;
		}
		v_upgrade (sel, fd -> major, fd -> minor);
	}

	res = resolve_dyn (XREF_OP_SERIALIZE, fd, base, &stop);
	if (res != S_CSC_OK)
		return res;

	if (stop)
		return S_CSC_OK;

	/* Next: obtain an updated version of the object into uobj.
	 * if the fields' contents are changed, set fchanged accordingly */
	fchanged = XREF_C_UNKNOWN;
	switch (fd -> des) {
	case XREF_D_UNWRAP:
	case XREF_D_OBJ:
		res = obj_recursive_serializer (rdb, robj, sel, &uobj, 
						&fchanged, fd);
		if (res != S_CSC_OK)
			return res;
		break;

	case XREF_D_PSTABLE:
		res = pstable_serializer (rdb, robj, sel, &uobj, 
					  &fchanged, fd);
		if (res != S_CSC_OK)
			return res;
		break;

	case XREF_D_ARRAY:
		obj_set_null (&uobj, &f_null);
		res = array_serializer (rdb, robj, sel, &uobj, &fchanged, fd);
		if (res != S_CSC_OK)
			return res;
		break;

	case XREF_D_NEST:
		res = obj_dup (sobj, &uobj);
		if (res != S_CSC_OK)
			return res;
		res = ser_nest (rdb, robj, sel, &uobj, &fchanged, fd);
		if (res != S_CSC_OK) {
			obj_destroy (&uobj);
			return res;
		}
		break;

	case XREF_D_REGION:
		assert (obj_cast (sobj, OT_STREAM));
		stream = fd -> field;
		if (!obj_region_equals (&sobj -> val.stream.decoded, 
					&stream -> decoded) ||
		    !obj_region_equals (&sobj -> val.stream.region,
					&stream -> region)) {
			fchanged = XREF_C_CHANGED;
			res = xref_stream_ser (NULL, robj, sel, sobj, 
					       fd -> field);
			if (res != S_CSC_OK)
				return res;
		} else
			fchanged = XREF_C_UNCHANGED;

		res = obj_dup (sobj, &uobj);
		if (res != S_CSC_OK)
			return res;
		break;

	case XREF_D_DYNAMIC:
		assert (0);
	}
	
	/* Now serialize the object if needed, and determine whether
	 * sobj has been changed too. The results are copied into tobj 
	 * and uobj is destroyed */
	if (rdb != NULL && fd -> ser == XREF_S_REFERENCE) {
		if (uobj.type != OT_NULL) {
			if (fchanged == XREF_C_UNKNOWN)
				fchanged = obj_equals (sobj, &uobj) ?
					XREF_C_UNCHANGED : XREF_C_CHANGED;

			if (fchanged == XREF_C_CHANGED ||
			    !obj_exists (&uobj)) {
				assert (sel != NULL);
				res = xref_serialize_obj (rdb, &uobj);
				if (res != S_CSC_OK) {
					obj_destroy (&uobj);
					return res;
				}
			}
			obj_copy_reference (&tobj, &uobj);
			obj_destroy (&uobj);
		} else
			obj_set_null (&tobj, &f_null);

		fchanged = XREF_C_UNKNOWN;
	} else
		tobj = uobj;

	if (fchanged == XREF_C_UNKNOWN)
		fchanged = obj_equals (sobj, &tobj) ?
			XREF_C_UNCHANGED : XREF_C_CHANGED;

	obj_destroy (sobj);
	*sobj = tobj;
	*changed = fchanged;
		
	return S_CSC_OK;
}

static enum s_status ser_dict (struct xref_db *rdb, struct xref_bind *robj,
			       struct v_selection *sel, struct obj *refobj,
			       enum xref_changed *changed, 
			       struct xref_field *fd)
{
	struct obj_dict_node **n;
	struct xref_field sfd;
	struct xref_field_desc *fdesc;
	struct obj_dictionary *dict;
	struct obj null, *obj;
	enum s_status res;
	enum xref_changed lc;
	int i;

	dict = obj_get_rw_dictionary (refobj);
	if (dict == NULL)
		return obj_error (refobj, "Expecting dictionary-like type");

	*changed = XREF_C_UNCHANGED;

	fdesc = fd -> u.fdesc;

	/* We serialize backwards so the dictionary is filled in
	 * the right order */
	i = 0;
	while (fdesc -> fd [i].name != NULL)
		i++;

	while (--i >= 0) {
		sfd = fdesc -> fd [i];
		
		obj_set_null (&null, &f_null);
		obj = get_dict_node (refobj, dict, sfd.name, &n);
		if (obj == NULL)
			obj = &null;

		res = serialize_field (rdb, robj, sel, obj, &lc, 
				       &sfd, fd -> field);
		if (res != S_CSC_OK)
			return res;

		update_changed (changed, lc);

		if (obj == refobj)
			;
		else if (*n == NULL && obj -> type != OT_NULL) {
			res = obj_add_dict (dict, sfd.name, obj);
			obj_destroy (obj);
			if (res != S_CSC_OK)
				return res;
		} else if (*n != NULL && obj -> type == OT_NULL)
			obj_dict_node_destroy (n);
	}

	return S_CSC_OK;
}

static enum s_status ser_stream (struct xref_db *rdb, struct xref_bind *robj,
				 struct v_selection *sel, struct obj *refobj,
				 enum xref_changed *changed, 
				 struct xref_field *fd)
{
	return ser_dict (rdb, robj, sel, refobj, changed, fd);
}

static enum s_status ser_array (struct xref_db *rdb, struct xref_bind *robj,
				struct v_selection *sel, struct obj *refobj, 
				enum xref_changed *changed, 
				struct xref_field *fd)
{
	struct xref_field_desc *fdesc;
	struct xref_obj_array ma;
	struct xref_field sfd;
	struct obj *rvec, aobj;
	enum xref_changed lc;
	enum s_status res;
	int i, tlen;

	if (!fd -> simplify && 
	    refobj -> type != OT_ARRAY && refobj -> type != OT_NULL)
		return obj_error (refobj, "Expecting array");

	fdesc = fd -> u.fdesc;

	res = xref_array_map_ps (rdb, robj, NULL, refobj, &ma, 
				 "Element", &xref_obj_pstable, NULL);
	if (res != S_CSC_OK)
		return res;

	for (i = 0; fdesc -> fd [i].name != NULL; i++)
		;
	tlen = i;

	if (tlen > XREF_A_SIZE (&ma)) {
		rvec = realloc (ma.data, tlen * sizeof (struct obj));
		if (rvec == NULL) {
			xref_array_map_rel_ps (&ma, NULL, 
					       &xref_obj_pstable, NULL);
			return S_CSC_OUT_OF_MEMORY;
		}
		ma.data = rvec;
		for (i = XREF_A_SIZE (&ma); i < tlen; i++)
			obj_set_null (&ma.data [i], &f_null);
	}

	*changed = XREF_C_UNCHANGED;
	for (i = 0; fdesc -> fd [i].name != NULL; i++) {
		sfd = fdesc -> fd [i];

		res = serialize_field (rdb, robj, sel, &ma.data [i], 
				       &lc, &sfd, fd -> field);
		if (res != S_CSC_OK) {
			xref_array_map_rel_ps (&ma, NULL, 
					       &xref_obj_pstable, NULL);
			return res;
		}

		update_changed (changed, lc);
	}

	while (i < XREF_A_SIZE (&ma))
		obj_destroy (&ma.data [i++]);

	for (i = tlen - 1; i >= 0; i--)
		if (ma.data [i].type != OT_NULL)
			break;

	i++;
	if (XREF_A_SIZE (&ma) != i) {
		update_changed (changed, XREF_C_CHANGED);
		ma.n = i;
	}
	res = xref_array_unmap_ps (rdb, robj, sel, &ma, &aobj, 
				   fd -> simplify, &xref_obj_pstable, NULL);
	xref_array_map_rel_ps (&ma, NULL, &xref_obj_pstable, NULL);
	if (res != S_CSC_OK)
		return res;

	obj_destroy (refobj);
	*refobj = aobj;

	return S_CSC_OK;
}

enum s_status xref_fields_ser (struct xref_db *rdb, struct xref_bind *robj,
			       struct v_selection *sel, 
			       struct obj *sobj, enum xref_changed *changed,
			       struct xref_field *fd, void *base)
{
	return serialize_field (rdb, robj, sel, sobj, changed, fd, base);
}

enum s_status xref_dict_resolve (struct xref_db *rdb, struct obj *dobj,
				 const char *key, enum obj_type ot, 
				 struct obj *obj)
{
	struct obj_dictionary *dict;
	struct obj *fobj;
	int present;

	assert (obj_can_cast (dobj, OT_DICTIONARY));

	dict = obj_get_rw_dictionary (dobj);

	fobj = get_dict_node (dobj, dict, key, NULL);
	if (fobj == NULL) {
		obj_set_null (obj, &dobj -> ptr);
		return S_CSC_OK;
	}

	return xref_resolve (rdb, fobj, ot, obj, &present, key);
}

static enum s_status tag_parser (struct xref_db *rdb, 
				 struct xref_bind *robj,
				 struct ref_owner *p,
				 struct obj *src,
				 struct xref_field *fd,
				 const char *fname, int line)
{
	if (src -> type == OT_NULL)
		return S_CSC_OK;

	if (src -> type != OT_NAME)
		return obj_error (src, "Bad tag type");

	if (!obj_string_equal (&src -> val.string, fd -> defval.val.string.s))
		return obj_error (src, "Expected `%s', found `%.*s'",
				  fd -> defval.val.string.s,
				  src -> val.string.size,
				  src -> val.string.s);

	return S_CSC_OK;
}

static enum s_status tag_serializer (struct xref_db *rdb, 
				     struct xref_bind *robj,
				     struct v_selection *sel,
				     struct obj *dest,
				     enum xref_changed *changed,
				     struct xref_field *fd)
{
	struct obj_bstr str;
	enum s_status res;

	res = obj_string_from_c (fd -> defval.val.string.s, &str);
	if (res != S_CSC_OK)
		return res;

	obj_set_name (dest, &f_null, &str);

	return S_CSC_OK;
}

static enum s_status array_parser (struct xref_db *rdb, 
				   struct xref_bind *robj,
				   struct ref_owner *p,
				   struct obj *src,
				   struct xref_field *fd,
				   int dflags,
				   const char *fname, int line)
{
	struct xref_mapped_array *ma, sma;
	enum s_status res;

	ma = fd -> field;
	sma = *ma;

	memset (ma, 0, sizeof (struct xref_mapped_array));

	res =  xref__array_map_ps (rdb, robj, p, src, ma,
				   fd -> name, fd -> u.pstable, fd -> data,
				   fname, line);
	if (res != S_CSC_OK) {
		*ma = sma;
		return res;
	}

	if (dflags & DRF_REPLACE)
		xref_array_map_rel_ps (&sma, p, fd -> u.pstable, fd -> data);

	return S_CSC_OK;
}

static enum s_status array_serializer (struct xref_db *rdb, 
				       struct xref_bind *robj,
				       struct v_selection *sel,
				       struct obj *dest,
				       enum xref_changed *changed,
				       struct xref_field *fd)
{
	enum s_status res;
	
	res = xref_array_unmap_ps (rdb, robj, sel, fd -> field, 
				   dest, fd -> simplify, 
				   fd -> u.pstable, fd -> data); 
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static void array_funref (struct ref_owner *p, struct xref_field *fd)
{
	xref_array_map_rel_ps (fd -> field, p, fd -> u.pstable, fd -> data);
}

static enum s_status pstable_parser (struct xref_db *rdb, 
				     struct xref_bind *robj,
				     struct ref_owner *p,
				     struct obj *src,
				     struct xref_field *fd,
				     int dflags,
				     const char *fname, int line)
{
	struct xref_field sfd;
	void *buffer;
	enum s_status res;

	if (dflags & DRF_REPLACE) {
		buffer = malloc (fd -> u.pstable -> size);
		if (buffer == NULL)
			return S_CSC_OUT_OF_MEMORY;
		memcpy (buffer, fd -> field,
			fd -> u.pstable -> size);
		memset (fd -> field, 0, fd -> u.pstable -> size);
	} else
		buffer = NULL;

	res = fd -> u.pstable -> parser (rdb, robj, p, src, fd, fname, line);
	if (buffer != NULL) {
		if (res != S_CSC_OK) 
			memcpy (fd -> field, buffer, fd -> u.pstable -> size);
		else if (fd -> u.pstable != NULL) {
			xref_init_field (&sfd, buffer, fd -> data);
			fd -> u.pstable -> funref (p, &sfd);
		}
	}

	if (buffer != NULL)
		free (buffer);

	return res;
}

static enum s_status serialize_children (struct xref_db *rdb,
					 struct obj *obj)
{
	struct obj_dict_node *dn;
	struct obj_array_node *an;
	enum s_status res;

	switch (obj -> type) {
	case OT_DICTIONARY:
	case OT_STREAM:
		if (obj -> type == OT_DICTIONARY)
			dn = obj -> val.stream.dictionary.head;
		else
			dn = obj -> val.dictionary.head;
		while (dn != NULL) {
			if (rdb != NULL && obj_exists (&dn -> obj)) {
				res = xref_serialize_obj (rdb, &dn -> obj);
				if (res != S_CSC_OK)
					return res;
			}
			res = serialize_children (rdb, &dn -> obj);
			if (res != S_CSC_OK)
				return res;

			dn = dn -> next;
		}
		break;

	case OT_ARRAY:
		an = obj -> val.array.head;
		while (an != NULL) {
			if (rdb != NULL && obj_exists (&an -> obj)) {
				res = xref_serialize_obj (rdb, &an -> obj);
				if (res != S_CSC_OK)
					return res;
			}
			res = serialize_children (rdb, &an -> obj);
			if (res != S_CSC_OK)
				return res;

			an = an -> next;
		}
		break;
	default:
		break;
	}

	return S_CSC_OK;
}

static enum s_status obj_recursive_serializer (struct xref_db *rdb, 
					       struct xref_bind *robj,
					       struct v_selection *sel,
					       struct obj *dest,
					       enum xref_changed *changed,
					       struct xref_field *fd)
{
	enum s_status res;

	if (fd -> des == XREF_D_UNWRAP) {
		obj_wrap (fd -> field, fd -> ot, &f_null, dest);
		if (fd -> state == &xref_optional_tag &&
		    obj_equals (dest, &fd -> defval))
			obj_set_null (dest, &f_null);
	} else {
		res = obj_dup (fd -> field, dest);
		if (res != S_CSC_OK)
			return res;
	}

	res = serialize_children (rdb, dest);
	if (res != S_CSC_OK) {
		obj_destroy (dest);
		return res;
	}

	*changed = XREF_C_UNCHANGED;

	return S_CSC_OK;
}

static enum s_status pstable_serializer (struct xref_db *rdb, 
					 struct xref_bind *robj,
					 struct v_selection *sel,
					 struct obj *dest,
					 enum xref_changed *changed,
					 struct xref_field *fd)
{
	obj_set_null (dest, &f_null);

	return fd -> u.pstable -> serializer (rdb, robj, sel,  
					      dest, changed, fd);
}

static void pstable_funref (struct ref_owner *p, struct xref_field *fd)
{
	if (fd -> u.pstable -> funref != NULL)
		fd -> u.pstable -> funref (p, fd);
	memset (fd -> field, 0, fd -> u.pstable -> size);
}

static enum s_status bool_parser (struct xref_db *rdb, 
				  struct xref_bind *robj,
				  struct ref_owner *p,
				  struct obj *src,
				  struct xref_field *fd,
				  const char *fname, int line)
{
	if (!obj_cast (src, OT_BOOLEAN))
		return obj_error (src, "Bad object type");

	*(pdf_bool_t *) fd -> field = src -> val.bool;

	return S_CSC_OK;
}

static enum s_status bool_serializer (struct xref_db *rdb, 
				      struct xref_bind *robj,
				      struct v_selection *sel,
				      struct obj *dest,
				      enum xref_changed *changed,
				      struct xref_field *fd)
{
	obj_set_boolean (dest, &f_null, *(pdf_bool_t *) fd -> field);

	return S_CSC_OK;
}

static enum s_status int_parser (struct xref_db *rdb, 
				 struct xref_bind *robj,
				 struct ref_owner *p,
				 struct obj *src,
				 struct xref_field *fd,
				 const char *fname, int line)
{
	if (!obj_cast (src, OT_INTEGER))
		return obj_error (src, "Bad object type");

	*(pdf_integer_t *) fd -> field = src -> val.integer;

	return S_CSC_OK;
}

static enum s_status int_serializer (struct xref_db *rdb, 
				     struct xref_bind *robj,
				     struct v_selection *sel,
				     struct obj *dest,
				     enum xref_changed *changed,
				     struct xref_field *fd)
{
	obj_set_integer (dest, &f_null, *(pdf_integer_t *) fd -> field);

	return S_CSC_OK;
}

static enum s_status real_parser (struct xref_db *rdb, 
				  struct xref_bind *robj,
				  struct ref_owner *p,
				  struct obj *src,
				  struct xref_field *fd,
				  const char *fname, int line)
{
	if (!obj_cast (src, OT_REAL))
		return obj_error (src, "Bad object type");

	*(pdf_real_t *) fd -> field = src -> val.real;

	return S_CSC_OK;
}

static enum s_status real_serializer (struct xref_db *rdb, 
				      struct xref_bind *robj,
				      struct v_selection *sel,
				      struct obj *dest,
				      enum xref_changed *changed,
				      struct xref_field *fd)
{
	obj_set_real (dest, &f_null, *(pdf_real_t *) fd -> field);

	return S_CSC_OK;
}

static enum s_status common_obj_parser (struct xref_db *rdb, 
					struct xref_bind *robj,
					struct ref_owner *p,
					struct obj *src,
					struct xref_field *fd,
					enum obj_type ot)
{
	if (!obj_can_cast (src, ot))
		return obj_error (src, "Can't cast as %s", obj_tag (ot));

	return obj_dup (src, fd -> field);
}

static enum s_status name_parser (struct xref_db *rdb, 
				  struct xref_bind *robj,
				  struct ref_owner *p,
				  struct obj *src,
				  struct xref_field *fd,
				  const char *fname, int line)
{
	return common_obj_parser (rdb, robj, p, src, fd, OT_NAME);
}

static enum s_status dictionary_parser (struct xref_db *rdb, 
					struct xref_bind *robj,
					struct ref_owner *p,
					struct obj *src,
					struct xref_field *fd,
					const char *fname, int line)
{
	return common_obj_parser (rdb, robj, p, src, fd, OT_DICTIONARY);
}

static enum s_status stream_parser (struct xref_db *rdb, 
				    struct xref_bind *robj,
				    struct ref_owner *p,
				    struct obj *src,
				    struct xref_field *fd,
				    const char *fname, int line)
{
	return common_obj_parser (rdb, robj, p, src, fd, OT_STREAM);
}

static enum s_status obj_parser (struct xref_db *rdb, 
				 struct xref_bind *robj,
				 struct ref_owner *p,
				 struct obj *src,
				 struct xref_field *fd,
				 const char *fname, int line)
{
	return common_obj_parser (rdb, robj, p, src, fd, OT_ANY);
}

static enum s_status obj_serializer (struct xref_db *rdb, 
				     struct xref_bind *robj,
				     struct v_selection *sel,
				     struct obj *dest,
				     enum xref_changed *changed,
				     struct xref_field *fd)
{
	struct obj *src;
	enum s_status res;

	src = fd -> field;
	if (rdb != NULL && src -> type == OT_STREAM) {
		if (obj_exists (src))
			return S_CSC_OK;

		res = xref_serialize_obj (rdb, src);
		if (res != S_CSC_OK)
			return res;

		obj_copy_reference (dest, src);
	} else {
		res = obj_dup (src, dest);
		if (res != S_CSC_OK)
			return res;
	}

	return S_CSC_OK;
}

static void obj_funref (struct ref_owner *p, struct xref_field *fd)
{
	obj_destroy (fd -> field);
}


