/*
 * PDFlow -- file.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if defined(_WIN32) || defined(WIN32)
#include <windows.h>
#else
#include <sys/mman.h>
#endif

#include <sys/stat.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <assert.h>

#include "fmt/file.h"
#include "fmt/pdf-lex.h"

struct f_ptr f_null = { NULL, NULL };

static enum s_status init (struct f_file **file, enum f_file_type ftype)
{
	struct f_file *f;

	f = malloc (sizeof (struct f_file));
	if (f == NULL)
		return S_CSC_OUT_OF_MEMORY;

	/* Minimal initialization to make f_unref work */
	memset (f, 0, sizeof (struct f_file));
	f -> type = ftype;
	f -> refcount = 1;
	f -> c.fd = -1;

	*file = f;

	return S_CSC_OK;
}

#ifdef F_EMULATE_MMAP
static enum s_status init_file (struct f_file *f, const char *fname)
{
	enum s_status res;
	HANDLE hfile, hmap;

	f -> fname = strdup (fname);
	if (f -> fname == NULL)
		return S_CSC_OUT_OF_MEMORY;

	hfile = CreateFile (fname, GENERIC_READ, FILE_SHARE_READ, NULL, 
			    OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hfile == NULL) {
		free (f -> fname);
		return S_WIN_ERROR;
	}

	hmap = CreateFileMapping (hfile, NULL, PAGE_READONLY, 0, 0, NULL);
	if (hmap == NULL) {
		free (f -> fname);
		res = S_WIN_ERROR;
		CloseHandle (hmap);
		CloseHandle (hfile);
		return res;
	}

	f -> pdf = MapViewOfFile (hmap, FILE_MAP_READ, 0, 0, 0);
	if (f -> pdf == NULL) {
		free (f -> fname);
		res = S_WIN_ERROR;
		CloseHandle (hfile);
		return res;
	}

	f -> c.wm.hfile = hfile;
	f -> c.wm.hmap = hmap;
	f -> end = f -> pdf + GetFileSize (hfile, NULL);

	return S_CSC_OK;
}
#else
static enum s_status init_file (struct f_file *f, const char *fname)
{
	struct stat stat;

	f -> fname = strdup (fname);
	if (f -> fname == NULL)
		return S_CSC_OUT_OF_MEMORY;

	f -> c.fd = open (fname, O_RDONLY);
	if (f -> c.fd < 0) {
		free (f -> fname);
		return S_ERRNO;
	}

	if (fstat (f -> c.fd, &stat)) {
		free (f -> fname);
		return S_ERRNO;
	}

	if (!S_ISREG (stat.st_mode)) {
		free (f -> fname);
		return s_make_indirect (S_CSC_NOT_SUPPORTED, 
					"Not a regular file");
	}

	f -> pdf = mmap (NULL, stat.st_size, 
			 PROT_READ, MAP_SHARED, f -> c.fd, 0);
	if (f -> pdf == NULL) {
		free (f -> fname);
		return S_ERRNO;
	}

	f -> end = f -> pdf + stat.st_size;

	return S_CSC_OK;
}
#endif

enum s_status f_open (const char *fname, struct f_file **pfile)
{
	struct f_file *f;
	enum s_status res;

	res = init (&f, F_T_FILE);
	if (res != S_CSC_OK)
		return res;

	res = init_file (f, fname);
	if (res != S_CSC_OK) {
		f_unref (f);
		return res;
	}

	*pfile = f;

	return S_CSC_OK;
}

static enum s_status creat_file (struct f_file *f, const char *fname)
{
	f -> c.fp = fopen (fname, "wb");
	if (f -> c.fd < 0)
		return S_ERRNO;

	f -> fname = strdup (fname);
	if (f -> fname == NULL)
		return S_CSC_OUT_OF_MEMORY;

	return S_CSC_OK;
}

enum s_status f_creat (const char *fname, struct f_file **pfile)
{
	struct f_file *f;
	enum s_status res;

	res = init (&f, F_T_WRITER);
	if (res != S_CSC_OK)
		return res;

	res = creat_file (f, fname);
	if (res != S_CSC_OK) {
		f_unref (f);
		return res;
	}

	*pfile = f;

	return S_CSC_OK;
}

static enum s_status append_file (struct f_file *f, const char *fname)
{
	f -> c.fp = fopen (fname, "ab");
	if (f -> c.fd < 0)
		return S_ERRNO;

	f -> fname = strdup (fname);
	if (f -> fname == NULL)
		return S_CSC_OUT_OF_MEMORY;

	return S_CSC_OK;
}

enum s_status f_append (struct f_file *file, struct f_file **pfile)
{
	struct f_file *f;
	enum s_status res;

	res = init (&f, F_T_WRITER);
	if (res != S_CSC_OK)
		return res;

	res = append_file (f, file -> fname);
	if (res != S_CSC_OK) {
		f_unref (f);
		return res;
	}

	*pfile = f;

	return S_CSC_OK;
}

enum s_status f_reg_open (struct f_file *parent, const uint8_t *begin, 
			  const uint8_t *end, void *data,
			  void (*deallocator) (void *p),
			  struct f_file **pfile)
{
	struct f_file *f;

	f = malloc (sizeof (struct f_file));
	if (f == NULL)
		return S_CSC_OUT_OF_MEMORY;

	/* Minimal initialization to make f_unref work */
	memset (f, 0, sizeof (struct f_file));

	f -> type = F_T_REGION;
	f -> fname = parent != NULL ? parent -> fname : NULL;
	f -> pdf = begin;
	f -> end = end;
	f -> refcount = 1;
	f -> c.reg.parent = parent;
	f -> c.reg.data = data;
	f -> c.reg.deallocator = deallocator;
	if (parent != NULL)
		f_ref (parent);

	*pfile = f;

	return S_CSC_OK;
}

enum s_status f_buffer_open (const char *name, struct f_file **pfile)
{
	struct f_file *f;

	f = malloc (sizeof (struct f_file));
	if (f == NULL)
		return S_CSC_OUT_OF_MEMORY;

	/* Minimal initialization to make f_unref work */
	memset (f, 0, sizeof (struct f_file));

	f -> type = F_T_BUFFER;
	f -> fname = strdup (name);
	if (f -> fname == NULL) {
		free (f);
		return S_CSC_OUT_OF_MEMORY;
	}
	f -> pdf = NULL;
	f -> end = NULL;
	f -> refcount = 1;
	f -> c.buffer.eob = NULL;
	f -> c.buffer.increment = 1024;

	*pfile = f;

	return S_CSC_OK;
}

void f_ref (struct f_file *f)
{
	f -> refcount++;
}

void f_unref (struct f_file *f)
{
	if (--f -> refcount)
		return;

	switch (f -> type) {
	case F_T_FILE:
#if defined(_WIN32) || defined(WIN32)
		if (f -> pdf != NULL)
			UnmapViewOfFile ((void *) f -> pdf);
		
		if (f -> c.wm.hmap != NULL)
			CloseHandle (f -> c.wm.hmap);

		if (f -> c.wm.hfile != NULL)
			CloseHandle (f -> c.wm.hfile);
#else
		if (f -> pdf != NULL)
			munmap ((void *) f -> pdf, f -> end - f -> pdf);

		if (f -> c.fd >= 0)
			close (f -> c.fd);
#endif

		if (f -> fname != NULL)
			free (f -> fname);
		break;

	case F_T_WRITER:
		if (f -> c.fp != NULL)
			fclose (f -> c.fp);

		if (f -> fname != NULL) {
			if (f -> unlink)
				unlink (f -> fname);
			free (f -> fname);
		}
		break;

	case F_T_REGION:
		if (f -> c.reg.deallocator != NULL)	
			f -> c.reg.deallocator (f -> c.reg.data);
		if (f -> c.reg.parent != NULL)
			f_unref (f -> c.reg.parent);
		break;

	case F_T_BUFFER:
		if (f -> pdf != NULL)
			free ((void *) f -> pdf);
		if (f -> fname != NULL)
			free (f -> fname);
		break;
	}

	free (f);
}

enum s_status f_verror (const struct f_ptr *ptr, const char *fmt, va_list ap)
{
	char buf [1024];
	uint8_t *p, *q;
	int lenp, lenq, len;

	vsnprintf (buf, sizeof (buf), fmt, ap);
	buf [sizeof (buf) - 1] = 0;

	if (ptr -> file == NULL)
		return s_make_indirect (S_CSC_PARSE_ERROR,
					"(input not avail): %s", buf);
	
	len = ptr -> file -> end - ptr -> cptr;
	if (len > 25)
		len = 25;
	p = memchr (ptr -> cptr, '\r', ptr -> file -> end - ptr -> cptr);
	q = memchr (ptr -> cptr, '\n', ptr -> file -> end - ptr -> cptr);
	lenp = p == NULL ? len : p - ptr -> cptr;
	lenq = q == NULL ? len : q - ptr -> cptr;

	return s_make_indirect (S_CSC_PARSE_ERROR,
				"%s:%u(0x%x): %.*s: %s", 
				ptr -> file -> fname,
				ptr -> cptr - ptr -> file -> pdf, 
				ptr -> cptr - ptr -> file -> pdf, 
				lenp < lenq ? lenp : lenq,
				ptr -> cptr, buf);
}

enum s_status f_error (const struct f_ptr *ptr, const char *fmt, ...) 
{
	enum s_status res;
	va_list ap;

	va_start (ap, fmt);
	res = f_verror (ptr, fmt, ap);
	va_end (ap);

	return res;
}


int f_ptr_init (struct f_ptr *fptr, struct f_file *file, long ofs)
{
	fptr -> file = file;
	fptr -> cptr = file -> pdf + ofs;

	return f_ptr_valid (fptr);
}

void f_ptr_eof (struct f_ptr *fptr, struct f_file *file)
{
	if (file != NULL)
		fptr -> file = file;
	fptr -> cptr = fptr -> file -> end;
}

int f_ptr_valid (struct f_ptr *fptr)
{
	return f_ptr_test (fptr, 0);
}

int f_ptr_test (struct f_ptr *fptr, long s)
{
	return 	fptr -> cptr + s >= fptr -> file -> pdf &&
		fptr -> cptr + s < fptr -> file -> end;
}

int f_ptr_seek (struct f_ptr *fptr, long s)
{
	fptr -> cptr += s;

	return f_ptr_valid (fptr);
}

size_t f_ptr_avail (struct f_ptr *fptr)
{
	return fptr -> file -> end - fptr -> cptr;
}

static enum s_status fvwrite (struct f_file *f, const char *fmt, va_list ap)
{
	int wr;

	wr = vfprintf (f -> c.fp, fmt, ap);

	return wr >= 0 ? S_CSC_OK : S_ERRNO;
}

static enum s_status assure (struct f_file *f, size_t size)
{
	uint8_t *p;

	if (f -> pdf != NULL && f -> c.buffer.eob - f -> end >= size)
		return S_CSC_OK;

	size += f -> c.buffer.increment;
	p = realloc ((void *) f -> pdf, size);
	if (p == NULL)
		return S_CSC_OUT_OF_MEMORY;

	f -> end = p + (f -> end - f -> pdf);
	f -> pdf = p;
	f -> c.buffer.eob = f -> pdf + size;

	return S_CSC_OK;
}

static enum s_status bvwrite (struct f_file *f, const char *fmt, va_list sap)
{
	const uint8_t *needed;
	enum s_status res;
	va_list ap;
	int wr;

	res = assure (f, 0);
	if (res != S_CSC_OK)
		return res;

	ap = sap;
	wr = vsnprintf ((char *) f -> end, 
			f -> c.buffer.eob - f -> end, fmt, ap);
	needed = f -> end + wr + 1;

	if (needed >= f -> c.buffer.eob) {
		res = assure (f, needed - f -> pdf);
		wr = vsnprintf ((char *) f -> end, 
				f -> c.buffer.eob - f -> end, fmt, ap);
	}

	f -> end += wr;

	return S_CSC_OK;
}

static enum s_status vwrite (struct f_file *f, const char *fmt, va_list ap)
{
	switch (f -> type) {
	case F_T_FILE:
		return s_make_indirect (S_CSC_INTERNAL_ERROR,
					"Attempting to write on RO file");

	case F_T_WRITER:
		return fvwrite (f, fmt, ap);

	case F_T_REGION:
		return s_make_indirect (S_CSC_INTERNAL_ERROR,
					"Attempting to write on RO region");
		
	case F_T_BUFFER:
		return bvwrite (f, fmt, ap);
	}

	assert (0);
}

enum s_status f_write (struct f_file *f, const char *fmt, ...)
{
	enum s_status res;
	va_list ap;

	va_start (ap, fmt);
	res = vwrite (f, fmt, ap);
	va_end (ap);

	return res;
}

enum s_status f_writeln (struct f_file *f, const char *fmt, ...)
{
	enum s_status res;
	va_list ap;
	
	va_start (ap, fmt);
	res = vwrite (f, fmt, ap);
	va_end (ap);
	if (res != S_CSC_OK)
		return res;

	return f_newline (f);
}

enum s_status f_newline (struct f_file *file)
{
	return f_write (file, "%c", 10);
}

enum s_status f_dump (struct f_file *dest, struct f_file *src)
{
	return f_raw (dest, src -> pdf, src -> end - src -> pdf);
}

enum s_status f_raw (struct f_file *dest, const void *data, size_t size)
{
	enum s_status res;
	size_t ans;

	switch (dest -> type) {
	case F_T_FILE:
		return s_make_indirect (S_CSC_INTERNAL_ERROR,
					"Attempting to write on RO file");

	case F_T_WRITER:
		ans = fwrite (data, 1, size, dest -> c.fp);
		return ans == size ? S_CSC_OK : S_ERRNO;

	case F_T_REGION:
		return s_make_indirect (S_CSC_INTERNAL_ERROR,
					"Attempting to write on RO region");
		
	case F_T_BUFFER:
		res = assure (dest, dest -> end - dest -> pdf + size);
		if (res != S_CSC_OK)
			return res;

		memcpy ((void *) dest -> end, data, size);
		dest -> end += size;

		return S_CSC_OK;
	}

	assert (0);
}

enum s_status f_tell (struct f_file *file, pdf_ofs_t *pos)
{
	long ofs;

	ofs = ftell (file -> c.fp);
	if (ofs < 0)
		return S_ERRNO;

	*pos = ofs;

	return S_CSC_OK;
}

void f_unlink_on_close (struct f_file *file)
{
	file -> unlink = 1;
}
