/*
 * PDFlow -- machine.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "fmt/obj.h"
#include "fmt/machine.h"

#include <string.h>
#include <stdlib.h>

#define CHUNK_SIZE 30

struct m_slot {

	struct m_slot *next;

	struct obj obj;
	
	int token;
};

struct m_chunk {

	struct m_chunk *next;

	struct m_slot slot [CHUNK_SIZE];

};

struct stack_ctxt {

	struct m_slot *sslot, *qslot;

	int scope;
};

struct m_stack {

	struct m_chunk *chunks;

	struct m_slot *fslot;

	struct stack_ctxt ctxt;

};

struct stack_ctxt_node {

	struct stack_ctxt_node *next;

	struct stack_ctxt ctxt;

};

struct m_state {

	struct stack_ctxt_node *ctxts;

	struct m_stack stack;

	struct m_machine *mach;

	struct m_block_desc *desc;

	int bscope;
};

struct m_machine {

	m_trigger *ttable;

	int blocks;
	struct m_block_desc *block;

	int eoftok;
};

struct m_block {

	int refs;

	struct m_stack *stack;

	struct m_slot *qslot;
};

static enum s_status block_dup (const struct obj_alive *sal, 
				struct obj_alive *dal);
static void block_deallocate (struct obj_alive *obj_alive);

static struct obj_alive_allocator block_allocator = { block_dup, 
						      block_deallocate };


static struct m_slot *alloc_slot (struct m_stack *stack)
{
	struct m_chunk *p;
	struct m_slot *s;
	int i;

	if (stack -> fslot == NULL) {
		p = malloc (sizeof (struct m_chunk));
		if (p == NULL)
			return NULL;

		p -> next = stack -> chunks;
		stack -> chunks = p;
		p -> slot [0].next = NULL;
		for (i = 1; i < CHUNK_SIZE; i++)
			p -> slot [i].next = &p -> slot [i - 1];
		stack -> fslot = &p -> slot [CHUNK_SIZE - 1];
	}

	s = stack -> fslot;
	stack -> fslot = s -> next;
	
	return s;
}

static void release_slot (struct m_stack *stack, struct m_slot *s)
{
	s -> next = stack-> fslot;
	stack -> fslot = s;
}

static int q_empty (struct m_stack *stack)
{
	return stack -> ctxt.qslot == NULL;
}

static enum s_status q_put (struct m_stack *stack, struct obj *obj, int tt)
{
	struct m_slot *s;

	s = alloc_slot (stack);
	if (s == NULL)
		return S_CSC_OUT_OF_MEMORY;

	if (stack -> ctxt.qslot != NULL) {
		s -> next = stack -> ctxt.qslot -> next;
		stack -> ctxt.qslot -> next = s;
	} else
		s -> next = s;

	stack -> ctxt.qslot = s;

	s -> obj = *obj;
	s -> token = tt;
	
	return S_CSC_OK;
}

static int q_get (struct m_stack *stack, struct obj *obj, int *tt)
{
	struct m_slot *s;

	if (stack -> ctxt.qslot == NULL)
		return 0;

	s = stack -> ctxt.qslot -> next;
	
	if (s -> next != s)
		stack -> ctxt.qslot -> next = s -> next;
	else
		stack -> ctxt.qslot = NULL;

	*obj = s -> obj;
	*tt = s -> token;

	release_slot (stack, s);

	return 1;
}

static int q_next (struct m_slot *qslot, void **pos,
		   struct obj *obj, int *tt)
{
	struct m_slot *slot;

	slot = *pos;
	if (slot != NULL) {
		if (slot == qslot)
			return 0;
		slot = slot -> next;
	} else {
		if (qslot == NULL)
			return 0;
		slot = qslot -> next;
	}

	*obj = slot -> obj;
	*tt = slot -> token;

	*pos = slot;

	return 1;
}

static void q_export (struct m_stack *stack, struct m_slot **slot)
{
	*slot = stack -> ctxt.qslot;
	stack -> ctxt.qslot = NULL;
}

static void q_clear (struct m_stack *stack, struct m_slot *slot)
{
	struct m_slot *p, *q;

	p = slot -> next;
	slot -> next = NULL;

	while (p != NULL) {
		q = p -> next;
		obj_destroy (&p -> obj);
		release_slot (stack, p);
		p = q;
	}
}

int m_empty (struct m_stack *stack)
{
	return stack -> ctxt.sslot == NULL;
}

unsigned m_size (struct m_stack *stack)
{
	struct m_slot *s;
	unsigned n;

	n = 0;
	for (s = stack -> ctxt.sslot; s != NULL; s = s -> next)
		n++;
	
	return n;
}

static enum s_status m_push_at (struct m_stack *stack, struct m_slot **head,
				struct obj *obj, int tt)
{
	struct m_slot *s;

	s = alloc_slot (stack);
	if (s == NULL)
		return S_CSC_OUT_OF_MEMORY;

	s -> next = *head;
	*head = s;

	s -> obj = *obj;
	s -> token = tt;
	
	return S_CSC_OK;
}

enum s_status m_push (struct m_stack *stack, struct obj *obj, int tt)
{
	return m_push_at (stack, &stack -> ctxt.sslot, obj, tt);
}

static int m_pop_at (struct m_stack *stack, struct m_slot **head,
		     struct obj *obj, int *tt)
{
	struct m_slot *s;

	if (*head == NULL)
		return 0;

	s = *head;
	*head = s -> next;

	*obj = s -> obj;
	*tt = s -> token;

	release_slot (stack, s);

	return 1;
}

int m_pop (struct m_stack *stack, struct obj *obj, int *tt)
{
	return m_pop_at (stack, &stack -> ctxt.sslot, obj, tt);
}


enum s_status m_copy (struct m_stack *stack, struct obj *refobj, 
		      pdf_integer_t n)
{
	struct m_slot *s, *head;
	enum s_status res;
	struct obj obj;
	int i, tt;

	head = NULL;
	s = stack -> ctxt.sslot;
	res = S_CSC_OK;
	for (i = 0; i < n; i++) {
		if (s == NULL) {
			res = obj_error (refobj, "Stack underflow");
			break;
		}

		res = obj_dup (&s -> obj, &obj);
		if (res != S_CSC_OK)
			break;

		res = m_push_at (stack, &head, &obj, s -> token);
		if (res != S_CSC_OK) {
			obj_destroy (&obj);
			break;
		}
		s = s -> next;
	}

	if (res == S_CSC_OK) {
		while (m_pop_at (stack, &head, &obj, &tt)) {
			res = m_push (stack, &obj, tt);
			if (res != S_CSC_OK)
				break;
		}
	}

	if (res != S_CSC_OK) {
		while (m_pop_at (stack, &head, &obj, &tt))
			obj_destroy (&obj);

		return res;
	}

	return S_CSC_OK;
}

enum s_status m_roll (struct m_stack *stack, struct obj *refobj, 
		      pdf_integer_t j, pdf_integer_t n)
{
	struct m_slot *s1, *sn, **sjp;
	pdf_integer_t i;

	if (!n)
		return S_CSC_OK;

	if (j < 0)
		j = n - ((-j) % n);

	j %= n;

	s1 = stack -> ctxt.sslot;
	sn = stack -> ctxt.sslot;
	sjp = &stack -> ctxt.sslot;
	for (i = 0; i < n - 1; i++) {
		if (sn == NULL)
			break;
		if (i == j - 1)
			sjp = &sn -> next;
		sn = sn -> next;
	}
	if (sn == NULL)
		return obj_error (refobj, "Stack underflow");

	if (j == 0)	
		return S_CSC_OK;

	stack -> ctxt.sslot = *sjp;
	*sjp = sn -> next;
	sn -> next = s1;
	
	return S_CSC_OK;
}

enum s_status m_index (struct m_stack *stack, struct obj *refobj, 
		       pdf_integer_t n)
{
	struct m_slot *s;
	enum s_status res;
	struct obj obj;
	int i;

	s = stack -> ctxt.sslot;
	for (i = 0; i < n; i++) {
		if (s == NULL)
			break;
		s = s -> next;
	}
	if (s == NULL)
		return obj_error (refobj, "Stack underflow");

	res = obj_dup (&s -> obj, &obj);
	if (res != S_CSC_OK)
		return res;

	res = m_push (stack, &obj, s -> token);
	if (res != S_CSC_OK) {
		obj_destroy (&obj);
		return res;
	}

	return S_CSC_OK;
}

enum s_status m_exch (struct m_stack *stack, struct obj *refobj)
{
	struct m_slot *s1, *s2;

	if (m_size (stack) < 2)
		return obj_error (refobj, "Stack underflow");
		
	s1 = stack -> ctxt.sslot;
	s2 = s1 -> next;
	
	s1 -> next = s2 -> next;
	s2 -> next = s1;

	stack -> ctxt.sslot = s2;

	return S_CSC_OK;
}

struct obj *m_top (struct m_stack *stack, int *tt)
{
	if (stack -> ctxt.sslot != NULL) {
		*tt = stack -> ctxt.sslot -> token;
		return &stack -> ctxt.sslot -> obj;
	} else
		return NULL;
}

void m_open (struct m_stack *stack)
{
	stack -> ctxt.scope++;
}

void m_close (struct m_stack *stack)
{
	--stack -> ctxt.scope;
}

static void m_clear (struct m_stack *stack)
{
	struct obj obj;
	int tt;

	while (m_pop (stack, &obj, &tt))
		obj_destroy (&obj);

	while (q_get (stack, &obj, &tt))
		obj_destroy (&obj);

}

static void m_save (struct m_stack *stack, struct stack_ctxt_node *n)
{
	n -> ctxt = stack -> ctxt;
	memset (&stack -> ctxt, 0, sizeof (struct stack_ctxt));
}

static void m_restore (struct m_stack *stack, struct stack_ctxt_node *n)
{
	m_clear (stack);
	stack -> ctxt = n -> ctxt;
}

static void s_destroy (struct m_stack *stack)
{
	struct m_chunk *p;

	while (stack -> chunks != NULL) {
		p = stack -> chunks;
		stack -> chunks = p -> next;
		free (p);
	}
}

enum s_status err_trigger (struct m_stack *stack,
			   struct obj *obj, int token,	
			   void *closure)
{
	return obj_error (obj, "Unexpected token");
}

enum s_status open_block_trigger (struct m_stack *stack,
				  struct obj *obj, int token,	
				  void *closure)
{
	struct m_block_desc *desc;
	struct m_state *state;
	int i;

	state = closure;
	for (i = 0; i < state -> mach -> blocks; i++)
		if (state -> mach -> block [i].open == token)
			break;
	
	desc = &state -> mach -> block [i];
	if (i == state -> mach -> blocks)
		return s_make_indirect (S_CSC_INTERNAL_ERROR,
					"Unexpected call to open_block");

	if (state -> bscope ||
	    state -> desc != NULL ||
	    !q_empty (&state -> stack))
		return s_make_indirect (S_CSC_INTERNAL_ERROR,
					"Nested block encountered");
	state -> desc = desc;
	state -> bscope++;

	return S_CSC_OK;
}

enum s_status close_block_trigger (struct m_stack *stack,
				   struct obj *obj, int token,	
				   void *closure)
{
	return s_make_indirect (S_CSC_INTERNAL_ERROR,
				"Unexpected close block token");
}

static enum s_status wrap_block (struct m_stack *stack, 
				 struct obj *tobj)
{
	enum s_status res;
	struct m_block *block;
	struct obj obj;

	block = malloc (sizeof (struct m_block));
	if (block == NULL)
		return S_CSC_OUT_OF_MEMORY;

	q_export (stack, &block -> qslot);
	block -> refs = 1;
	block -> stack = stack;

	obj_set_alive (&obj, &tobj -> ptr, &block_allocator, block);
	res = m_push (stack, &obj, M_TOK_BLOCK);
	if (res != S_CSC_OK) {
		obj_destroy (&obj);
		return res;
	}

	return S_CSC_OK;
}

static enum s_status block_dup (const struct obj_alive *sal, 
				struct obj_alive *dal)
{
	struct m_block *block;

	*dal = *sal;
	block = dal -> ptr;

	block -> refs++;

	return S_CSC_OK;
}

static void block_deallocate (struct obj_alive *obj_alive)
{
	struct m_block *block;

	block = obj_alive -> ptr;

	if (--block -> refs)
		return;

	q_clear (block -> stack, block -> qslot);
	free (block);
}

enum s_status m_create_machine (struct m_rule *rules, int toks,
				struct m_block_desc *block, int blocks,
				int eoftok, struct m_machine **machp)
{
	struct m_machine *mach;
	m_trigger *tt;
	int i;

	mach = malloc (sizeof (struct m_machine));
	if (mach == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (mach, 0, sizeof (struct m_machine));

	tt = calloc (toks, sizeof (m_trigger));
	if (tt == NULL) {
		m_destroy_machine (mach);
		return S_CSC_OUT_OF_MEMORY;
	}

	if (blocks) {
		mach -> blocks = blocks;
		mach -> block = calloc (blocks, sizeof (struct m_block_desc));
		if (mach -> block == NULL) {
			m_destroy_machine (mach);
			return S_CSC_OUT_OF_MEMORY;
		}
		memcpy (mach -> block, block,
			blocks * sizeof (struct m_block_desc));
	}

	mach -> ttable = tt;

	for (i = 0; i < toks; i++)
		tt [i] = err_trigger;

	for (i = 0; rules [i].token >= 0; i++)
		tt [rules [i].token] = rules [i].trigger;

	for (i = 0; i < blocks; i++) {
		tt [block [i].open] = open_block_trigger;
		tt [block [i].close] = close_block_trigger;
	}

	mach -> eoftok = eoftok;

	*machp = mach;

	return S_CSC_OK;
}

void m_destroy_machine (struct m_machine *mach)
{
	if (mach -> block != NULL)
		free (mach -> block);
	if (mach -> ttable != NULL)
		free (mach -> ttable);
	free (mach);
}

enum s_status m_start_machine (struct m_machine *mach,
			       struct m_state **state)
{
	struct m_state *s;

	s = malloc (sizeof (struct m_state));
	if (s == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (s, 0, sizeof (struct m_state));

	s -> mach = mach;

	*state = s;

	return S_CSC_OK;
}

int m_run_machine_cyc (struct m_state *state, struct obj *obj, int tt,
		       enum s_status *res, void *closure)
{
	struct m_stack *stack;
	m_trigger t;

	stack = &state -> stack;
	if (state -> desc != NULL) {
		*res = S_CSC_OK;
		if (tt == state -> desc -> open)
			state -> bscope++;
		else if (tt == state -> desc -> close) {
			if (!--state -> bscope) {
				*res = wrap_block (stack, obj);
				state -> desc = NULL;
			}
		} else if (tt == state -> mach -> eoftok)
			*res = obj_error (obj, "Unexpected EOF");
		else
			*res = q_put (stack, obj, tt);
	} else {
		t = state -> mach -> ttable [tt];
		if (t == NULL)
			*res = m_push (stack, obj, tt);
		else if (t == open_block_trigger || 
			 t == close_block_trigger)
			*res = t (stack, obj, tt, state);
		else
			*res = t (stack, obj, tt, closure);
	}	

	return *res == S_CSC_OK && (stack -> ctxt.scope || state -> bscope);
}

enum s_status m_eval_block (struct m_state *state, struct obj *obj,
			    void *closure)
{
	struct m_stack *stack;
	struct m_block *block;
	struct obj cobj;
	enum s_status res;
	void *pos;
	int tt;
	
	block = obj -> val.alive.ptr;
	pos = NULL;	
	stack = &state -> stack;
	while (q_next (block -> qslot, &pos, &cobj, &tt)) {
		m_run_machine_cyc (state, &cobj, tt, &res, closure);
		if (res != S_CSC_OK)
			return res;
	}	

	return S_CSC_OK;
}

int m_compare_blocks (struct obj *obj1, struct obj *obj2)
{
	struct m_block *block1, *block2;

	block1 = obj1 -> val.alive.ptr;
	block2 = obj2 -> val.alive.ptr;

	return block1 == block2;
}

enum s_status m_push_stack (struct m_state *state, struct obj *obj, 
			    int tt, int count)
{
	enum s_status res;
	int i;

	for (i = count - 1; i >= 0; i--) {
		res = m_push (&state -> stack, &obj [i], tt);
		if (res != S_CSC_OK)
			return res;
	}

	return S_CSC_OK;
}

enum s_status m_dump_stack (struct m_state *state, struct obj *obj, 
			    int *count)
{
	int i, tt, n;

	n = m_size (&state -> stack);
	for (i = 0; i < *count; i++) {
		if (!m_pop (&state -> stack, obj + i, &tt))
			break;
	}
	n = *count;

	return S_CSC_OK;
}

enum s_status m_open_machine (struct m_state *state)
{
	struct stack_ctxt_node *ctxt;

	ctxt = malloc (sizeof (struct stack_ctxt_node));
	if (ctxt == NULL)
		return S_CSC_OUT_OF_MEMORY;

	m_save (&state -> stack, ctxt);

	ctxt -> next = state -> ctxts;
	state -> ctxts = ctxt;

	return S_CSC_OK;
}

void m_close_machine (struct m_state *state)
{
	struct stack_ctxt_node *ctxt;

	ctxt = state -> ctxts;
	state -> ctxts = ctxt -> next;

	m_restore (&state -> stack, ctxt);
	
	free (ctxt);
}

void m_stop_machine (struct m_state *state)
{
	s_destroy (&state -> stack);

	free (state);
}


