/*
 * PDFlow -- obj.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include "status.h"
#include "fmt/obj.h"

const struct obj obj_null = { OT_NULL };
const struct obj obj_null_dictionary = { OT_DICTIONARY };
const struct obj obj_null_array = { OT_ARRAY };
const struct obj obj_null_name = { OT_NAME };

static enum s_status file_dup (const struct obj_region *sregion,
			       struct obj_region *dregion);
static void file_deallocate (struct obj_region *region);
static enum s_status malloc_dup (const struct obj_region *sreg,
				 struct obj_region *dreg);
static void malloc_deallocate (struct obj_region *region);

static struct obj_region_allocator file_allocator = {
	file_dup, file_deallocate
};

static struct obj_region_allocator malloc_allocator = {
	malloc_dup, malloc_deallocate
};

void obj_set_null (struct obj *obj, struct f_ptr *ptr)
{
	obj -> type = OT_NULL;
	obj -> ptr = *ptr;
	obj -> ref.oid = 0;
	obj -> ref.gnum = 0;
}

void obj_set_boolean (struct obj *obj, struct f_ptr *ptr, pdf_bool_t val)
{
	obj -> type = OT_BOOLEAN;
	obj -> ptr = *ptr;
	obj -> val.bool = val ? 1 : 0;
	obj -> ref.oid = 0;
	obj -> ref.gnum = 0;
}

void obj_set_real (struct obj *obj, struct f_ptr *ptr, pdf_real_t val)
{
	obj -> type = OT_REAL;
	obj -> ptr = *ptr;
	obj -> val.real = val;
	obj -> ref.oid = 0;
	obj -> ref.gnum = 0;
}

void obj_set_integer (struct obj *obj, struct f_ptr *ptr, pdf_integer_t val)
{
	obj -> type = OT_INTEGER;
	obj -> ptr = *ptr;
	obj -> val.integer = val;
	obj -> ref.oid = 0;
	obj -> ref.gnum = 0;
}

void obj_set_string (struct obj *obj, struct f_ptr *ptr, 
		     struct obj_bstr *string)
{
	obj -> type = OT_STRING;
	obj -> ptr = *ptr;
	obj -> val.string = *string;
	obj -> ref.oid = 0;
	obj -> ref.gnum = 0;
}

void obj_set_name (struct obj *obj, struct f_ptr *ptr, struct obj_bstr *string)
{
	obj -> type = OT_NAME;
	obj -> ptr = *ptr;
	obj -> val.string = *string;
	obj -> ref.oid = 0;
	obj -> ref.gnum = 0;
}

void obj_set_region (struct obj *obj, struct f_ptr *ptr,
		     struct obj_region *reg)
{
	obj -> type = OT_REGION;
	obj -> ptr = *ptr;
	obj -> val.region = *reg;
	obj -> ref.oid = 0;
	obj -> ref.gnum = 0;
}

void obj_set_dictionary (struct obj *obj, struct f_ptr *ptr, 
			 const struct obj_dictionary *dict)
{
	obj -> type = OT_DICTIONARY;
	obj -> ptr = *ptr;
	obj -> val.dictionary = *dict;
	obj -> ref.oid = 0;
	obj -> ref.gnum = 0;
}

void obj_set_array (struct obj *obj, struct f_ptr *ptr, 
		    const struct obj_array *array)
{
	obj -> type = OT_ARRAY;
	obj -> ptr = *ptr;
	obj -> val.array = *array;
	obj -> ref.oid = 0;
	obj -> ref.gnum = 0;
}

void obj_set_reference (struct obj *obj, const struct f_ptr *ptr, 
			pdf_oid_t oid, pdf_gen_t gnum)
{
	obj -> type = OT_REFERENCE;
	obj -> ptr = *ptr;
	obj -> ref.oid = oid;
	obj -> ref.gnum = gnum;
}

void obj_copy_reference (struct obj *obj, const struct obj *robj)
{
	obj_set_reference (obj, &robj -> ptr, 
			   robj -> ref.oid, robj -> ref.gnum);
}

void obj_set_xref (struct obj *obj, struct f_ptr *ptr, 
		   pdf_ofs_t f1, pdf_gen_t gnum, int valid)
{
	obj -> type = OT_XREF;
	obj -> ptr = *ptr;

	obj_init_xref (&obj -> val.xref, f1, gnum, valid);

	obj -> ref.oid = 0;
	obj -> ref.gnum = 0;
}

enum s_status obj_set_definition (struct obj *obj, struct f_ptr *ptr, 
				  struct obj *robj)
{
	obj -> type = OT_DEFINITION;
	obj -> ptr = *ptr;
	obj -> val.definition.obj = malloc (sizeof (struct obj));
	if (obj -> val.definition.obj == NULL)
		return S_CSC_OUT_OF_MEMORY;

	*obj -> val.definition.obj = *robj;

	obj -> ref.oid = 0;
	obj -> ref.gnum = 0;

	return S_CSC_OK;
}

void obj_set_stream (struct obj *obj, struct f_ptr *ptr,
		     struct obj_dictionary *dict, struct obj_region *region)
{
	obj -> type = OT_STREAM;
	obj -> ptr = *ptr;
	obj -> val.stream.dictionary = *dict;

	obj -> val.stream.region = *region;
	obj -> val.stream.decoded.allocator = NULL;
	obj -> val.stream.decoded.begin = NULL;
	obj -> val.stream.decoded.end = NULL;
	obj -> ref.oid = 0;
	obj -> ref.gnum = 0;
}

void obj_set_alive (struct obj *obj, struct f_ptr *ptr,
		    struct obj_alive_allocator *allocator, void *p)
{
	obj -> type = OT_ALIVE;
	obj -> ptr = *ptr;
	obj -> val.alive.ptr = p;
	obj -> val.alive.allocator = *allocator;
	obj -> ref.oid = 0;
	obj -> ref.gnum = 0;
}

void obj_set_keyword (struct obj *obj, struct f_ptr *ptr,
		      struct obj_keyword *keyword)
{
	obj -> type = OT_KEYWORD;
	obj -> ptr = *ptr;
	obj -> val.keyword = keyword;
	obj -> ref.oid = 0;
	obj -> ref.gnum = 0;
}

void obj_init_dictionary (struct obj_dictionary *dict)
{
	dict -> head = NULL;
}

void obj_init_dict_node (struct obj_dict_node *dnode,
			 struct obj *key, struct obj *obj)
{
	dnode -> next = NULL;
	dnode -> key = key -> val.string;
	dnode -> obj = *obj;
}

static enum s_status xfer_dict (struct obj_dict_node **ptr, 
				struct obj_dict_node *dnode)
{
	struct obj_dict_node *n;

	n = malloc (sizeof (struct obj_dict_node));
	if (n == NULL) 
		return S_CSC_OUT_OF_MEMORY;

	*n = *dnode;
	n -> next = *ptr;
	*ptr = n;

	return S_CSC_OK;
}

enum s_status obj_xfer_dict (struct obj_dictionary *dict, 
			     struct obj_dict_node *dnode)
{
	return xfer_dict (&dict -> head, dnode);
}

enum s_status obj_xfer_t_dict (struct obj_dictionary *dict, 
			       struct obj_dict_node *dnode)
{
	struct obj_dict_node **dnp;

	for (dnp = &dict -> head; *dnp != NULL; dnp = &(*dnp) -> next)
		;

	return xfer_dict (dnp, dnode);
}

enum s_status obj_add_dict_s (struct obj_dictionary *dict,
			      struct obj *kobj, struct obj *val)
{
	struct obj_dict_node *dn, dnode;
	enum s_status res;
	struct obj obj, sobj;
	
	for (dn = dict -> head; dn != NULL; dn = dn -> next) {
		if (obj_string_equal_s (&dn -> key, &kobj -> val.string)) {
			res = obj_dup (val, &obj);
			if (res != S_CSC_OK)
				return res;

			obj_destroy (&dn -> obj);
			dn -> obj = obj;

			return S_CSC_OK;
		}
	}

	res = obj_dup (val, &obj);
	if (res != S_CSC_OK)
		return res;

	res = obj_dup (kobj, &sobj);
	if (res != S_CSC_OK) {
		obj_destroy (&obj);
		return res;
	}

	obj_init_dict_node (&dnode, &sobj, &obj);

	res = obj_xfer_dict (dict, &dnode);
	if (res != S_CSC_OK) {
		obj_destroy (&obj);
		obj_destroy (&sobj);
		return res;
	}

	return S_CSC_OK;
}

enum s_status obj_add_dict (struct obj_dictionary *dict,
			    const char *key, struct obj *val)
{
	struct obj_dict_node *dn, dnode;
	enum s_status res;
	struct obj_bstr s;
	struct obj obj, sobj;
	
	for (dn = dict -> head; dn != NULL; dn = dn -> next) {
		if (obj_string_equal (&dn -> key, key)) {
			res = obj_dup (val, &obj);
			if (res != S_CSC_OK)
				return res;

			obj_destroy (&dn -> obj);
			dn -> obj = obj;

			return S_CSC_OK;
		}
	}

	res = obj_dup (val, &obj);
	if (res != S_CSC_OK)
		return res;

	res = obj_string_from_c (key, &s);
	if (res != S_CSC_OK) {
		obj_destroy (&obj);
		return res;
	}

	obj_set_string (&sobj, &f_null, &s);
	obj_init_dict_node (&dnode, &sobj, &obj);

	res = obj_xfer_dict (dict, &dnode);
	if (res != S_CSC_OK) {
		obj_destroy (&obj);
		obj_destroy (&sobj);
		return res;
	}

	return S_CSC_OK;
}

void obj_init_xref (struct obj_xref *ref, pdf_ofs_t f1, 
		    pdf_gen_t gnum, int valid)
{
	if (valid)
		ref -> u.ofs = f1;
	else
		ref -> u.next = f1;
	ref -> valid = valid;
	ref -> gnum = gnum;
}

void obj_dict_node_destroy (struct obj_dict_node **dn)
{
	struct obj_dict_node *p;

	p = *dn;
	*dn = (*dn) -> next;

	free (p -> key.s);
	obj_destroy (&p -> obj);
	free (p);
}

void obj_dict_destroy (struct obj_dictionary *dict)
{
	while (dict -> head != NULL)
		obj_dict_node_destroy (&dict -> head);
}

struct obj_dict_node *obj_dict_find_node (struct obj_dictionary *dict,
					  struct obj_bstr *name)
{
	struct obj_dict_node *dn;
	
	for (dn = dict -> head; dn != NULL; dn = dn -> next)
		if (obj_string_equal_s (&dn -> key, name))
			return dn;

	return NULL;
}

enum s_status obj_dict_dup (const struct obj_dictionary *sdict,
			    struct obj_dictionary *ddict)
{
	struct obj_dict_node *n, **p;
	struct obj_bstr key;
	struct obj obj;
	enum s_status res;

	res = S_CSC_OK;

	p = &ddict -> head;
	for (n = sdict -> head; n != NULL; n = n -> next) {
		res = obj_dup (&n -> obj, &obj);
		if (res != S_CSC_OK)
			break;
		
		key.s = malloc (n -> key.size);
		if (key.s == NULL) {
			obj_destroy (&obj);
			res = S_CSC_OUT_OF_MEMORY;
			break;
		}

		memcpy (key.s, n -> key.s, n -> key.size);
		key.size = n -> key.size;

		*p = malloc (sizeof (struct obj_dict_node));
		if (*p == NULL) {
			obj_destroy (&obj);
			res = S_CSC_OUT_OF_MEMORY;
			break;
		}

		(*p) -> next = NULL;
		(*p) -> key = key;
		(*p) -> obj = obj;

		p = &(*p) -> next;
	}

	if (res != S_CSC_OK) {
		obj_dict_destroy (ddict);
		return res;
	}

	return S_CSC_OK;
}

void obj_dict_merge (struct obj_dictionary *dest, struct obj_dictionary *src)
{
	struct obj_dict_node **dn;

	dn = &dest -> head; 
	while (*dn != NULL)
		dn = &(*dn) -> next;

	*dn = src -> head;
	src -> head = NULL;
}

void obj_init_array (struct obj_array *array)
{
	array -> head = NULL;
}

static enum s_status xfer_array (struct obj_array_node **ptr, struct obj *obj)
{
	struct obj_array_node *n;

	n = malloc (sizeof (struct obj_array_node));
	if (n == NULL) 
		return S_CSC_OUT_OF_MEMORY;

	n -> obj = *obj;
	n -> next = *ptr;
	*ptr = n;

	return S_CSC_OK;
}

enum s_status obj_xfer_array (struct obj_array *array, struct obj *obj)
{
	return xfer_array (&array -> head, obj);
}

enum s_status obj_xfer_t_array (struct obj_array *array, struct obj *obj,
				struct obj_array_node **anpp)
{
	struct obj_array_node **anp;
	enum s_status res;

	for (anp = &array -> head; *anp != NULL; anp = &(*anp) -> next)
		;

	res = xfer_array (anp, obj);
	if (res != S_CSC_OK)
		return res;

	if (anpp != NULL)
		*anpp = *anp;
	
	return S_CSC_OK;
}

void obj_array_node_destroy (struct obj_array_node **an)
{
	struct obj_array_node *p;

	p = *an;
	*an = (*an) -> next;

	obj_destroy (&p -> obj);
	free (p);
}

void obj_array_destroy (struct obj_array *array)
{
	while (array -> head != NULL) 
		obj_array_node_destroy (&array -> head);
}

enum s_status obj_array_dup (const struct obj_array *sarray,
			     struct obj_array *darray)
{
	struct obj_array_node *n, **p;
	struct obj obj;
	enum s_status res;

	res = S_CSC_OK;

	p = &darray -> head;
	for (n = sarray -> head; n != NULL; n = n -> next) {
		res = obj_dup (&n -> obj, &obj);
		if (res != S_CSC_OK)
			break;
		
		*p = malloc (sizeof (struct obj_array_node));
		if (*p == NULL) {
			obj_destroy (&obj);
			res = S_CSC_OUT_OF_MEMORY;
			break;
		}

		(*p) -> next = NULL;
		(*p) -> obj = obj;

		p = &(*p) -> next;
	}

	if (res != S_CSC_OK) {
		obj_array_destroy (darray);
		return res;
	}

	return S_CSC_OK;
}

static enum s_status file_dup (const struct obj_region *sregion,
			       struct obj_region *dregion)
{
	*dregion = *sregion;
	f_ref (dregion -> ddata);

	return S_CSC_OK;
}

static void file_deallocate (struct obj_region *region)
{
	f_unref (region -> ddata);
}

void obj_init_file_region (struct obj_region *region,
			   const void *begin, const void *end, 
			   struct f_file *file)
{
	region -> begin = begin;
	region -> end = end;
	region -> allocator = &file_allocator;
	region -> ddata = file;
	f_ref (file);
}

static enum s_status malloc_dup (const struct obj_region *sreg,
				 struct obj_region *dreg)
{
	size_t size;

	size = sreg -> end - sreg -> begin;
	dreg -> begin = malloc (size);
	if (dreg -> begin == NULL)
		return S_CSC_OUT_OF_MEMORY;
	dreg -> end = dreg -> begin + size;

	memcpy ((void *) dreg -> begin, sreg -> begin, size);

	return S_CSC_OK;
}

static void malloc_deallocate (struct obj_region *region)
{
	free ((void *) region -> begin);
}

void obj_init_malloc_region (struct obj_region *region,
			     const void *begin, const void *end)
{
	region -> begin = begin;
	region -> end = end;
	region -> allocator = &malloc_allocator;
}

static void obj_f_deallocator (void *data)
{
	struct obj *obj;

	obj = data;
	obj_destroy (obj);
	free (obj);
}

static enum s_status obj_file (struct obj *sobj, 
			       struct obj_region *reg,
			       struct f_file **f)
{
	struct obj *obj;
	enum s_status res;
	int ofs;

	obj = malloc (sizeof (struct obj));
	if (obj == NULL)
		return S_CSC_OUT_OF_MEMORY;

	res = obj_dup (sobj, obj);
	if (res != S_CSC_OK) {
		free (obj);
		return res;
	}

	ofs = ((uint8_t *) reg) - ((uint8_t *) sobj);
	reg = (struct obj_region *) (((uint8_t *) obj) + ofs);

	res = f_reg_open (obj -> ptr.file, reg -> begin, reg -> end,
			  obj, obj_f_deallocator, f);
	if (res != S_CSC_OK) {
		obj_destroy (obj);
		free (obj);
		return res;
	}

	return S_CSC_OK;
}
			       

enum s_status obj_region_file (struct obj *reg, struct f_file **f)
{
	return obj_file (reg, &reg -> val.region, f);
}

enum s_status obj_stream_file (struct obj *str, struct f_file **f)
{
	return obj_file (str, &str -> val.stream.decoded, f);
}

void obj_stream_destroy (struct obj_stream *stream)
{
	obj_dict_destroy (&stream -> dictionary);
	obj_region_destroy (&stream -> region);
	obj_region_destroy (&stream -> decoded);
}

void obj_region_destroy (struct obj_region *region)
{
	if (region -> allocator != NULL)
		region -> allocator -> deallocate (region);
	memset (region, 0, sizeof (struct obj_region));
}

enum s_status obj_region_dup (const struct obj_region *sregion,
			      struct obj_region *dregion)
{
	*dregion = *sregion;
	if (dregion -> allocator == NULL)
		return S_CSC_OK;

	return dregion -> allocator -> dup (sregion, dregion);
}

enum s_status obj_string_dup (const struct obj_bstr *sstr,
			      struct obj_bstr *dstr)
{
	dstr -> s = malloc (sstr -> size);
	if (dstr -> s == NULL)
		return S_CSC_OUT_OF_MEMORY;

	dstr -> size = sstr -> size;
	memcpy (dstr -> s, sstr -> s, sstr -> size);

	return S_CSC_OK;
}

void obj_alive_destroy (struct obj_alive *alive)
{
	if (alive -> allocator.deallocate != NULL)
		alive -> allocator.deallocate (alive);
}

enum s_status obj_alive_dup (const struct obj_alive *sal,
			     struct obj_alive *dal)
{
	*dal = *sal;
	if (dal -> allocator.dup == NULL)
		return S_CSC_OK;

	return dal -> allocator.dup (sal, dal);
}

int obj_string_equal (const struct obj_bstr *str, const char *s)
{
	size_t len;

	len = strlen (s);
	return 	str -> size == len &&
		!memcmp (str -> s, s, len);
}

int obj_string_equal_s (const struct obj_bstr *str1, 
			const struct obj_bstr *str2)
{
	return 	str1 -> size == str2 -> size &&
		!memcmp (str1 -> s, str2 -> s, str1 -> size);
}

int obj_string_compare (const struct obj_bstr *str1, 
			const char *str2)
{
	struct obj_bstr s;

	s.s = (char *) str2;
	s.size = strlen (str2);

	return obj_string_compare_s (str1, &s);
}

int obj_string_compare_s (const struct obj_bstr *str1, 
			  const struct obj_bstr *str2)
{
	int ss, ans;

	ss = str1 -> size < str2 -> size ? str1 -> size : str2 -> size;
	ans = memcmp (str1 -> s, str2 -> s, ss);
	if (ans)
		return ans;
	
	return str1 -> size - str2 -> size;
}

enum s_status obj_string_to_c (const struct obj_bstr *str, char **s)
{
	char *p;

	p = malloc (str -> size + 1);
	if (p == NULL)
		return S_CSC_OUT_OF_MEMORY;

	memcpy (p, str -> s, str -> size);
	p [str -> size] = 0;

	*s = p;
	
	return S_CSC_OK;
}

enum s_status obj_string_from_c (const char *s, struct obj_bstr *str)
{
	char *p;

	p = strdup (s);
	if (p == NULL)
		return S_CSC_OUT_OF_MEMORY;

	str -> size = strlen (s);
	str -> s = p;

	return S_CSC_OK;
}

void obj_string_destroy (struct obj_bstr *str)
{
	if (str -> s != NULL)
		free (str -> s);
	memset (str, 0, sizeof (struct obj_bstr));
}

enum s_status obj_ptr_dup (struct obj *sobj, struct obj *dobj)
{
	struct obj_ptr *op;
	enum s_status res;

	op = malloc (sizeof (struct obj_ptr));
	if (op == NULL)
		return S_CSC_OUT_OF_MEMORY;

	op -> refs = 1;
	res = obj_dup (sobj, &op -> obj);
	if (res != S_CSC_OK) {
		free (op);
		return res;
	}

	memset (dobj, 0, sizeof (struct obj));
	dobj -> type = OT_OBJ_PTR;
	dobj -> val.obj_ptr = op;
	
	return S_CSC_OK;
}

static enum s_status deref_dict (struct obj_dictionary *dict)
{
	struct obj_dict_node *dn;
	struct obj grip;
	enum s_status res;

	for (dn = dict -> head; dn != NULL; dn = dn -> next) {
		res = obj_ptr_deref (&dn -> obj, &grip);
		if (res != S_CSC_OK)
			return res;

		obj_destroy (&dn -> obj);
		dn -> obj = grip;
	}

	return S_CSC_OK;
}

static enum s_status deref_array (struct obj_array *array)
{
	struct obj_array_node *an;
	struct obj grip;
	enum s_status res;

	for (an = array -> head; an != NULL; an = an -> next) {
		res = obj_ptr_deref (&an -> obj, &grip);
		if (res != S_CSC_OK)
			return res;

		obj_destroy (&an -> obj);
		an -> obj = grip;
	}

	return S_CSC_OK;
}

enum s_status obj_ptr_deref (struct obj *sobj, struct obj *dobj)
{
	enum s_status res;

	res = obj_dup (obj_ptr_get (sobj), dobj);
	if (res != S_CSC_OK)
		return res;

	switch (dobj -> type) {
	case OT_DICTIONARY:
		res = deref_dict (&dobj -> val.dictionary);
		break;

	case OT_STREAM:
		res = deref_dict (&dobj -> val.stream.dictionary);
		break;

	case OT_ARRAY:
		res = deref_array (&dobj -> val.array);
		break;

	default:	
		res = S_CSC_OK;
		break;
	}

	return res;
}

struct obj *obj_ptr_get (struct obj *ptr)
{
	while (ptr -> type == OT_OBJ_PTR)
		ptr = &ptr -> val.obj_ptr -> obj;
	
	return ptr;
}


void obj_definition_unwrap (struct obj_definition *def, struct obj *obj)
{
	*obj = *def -> obj;
	free (def -> obj);
}

enum s_status obj_navigate (struct obj *obj, 
			    enum s_status (*f) (struct obj *o, void *data),
			    void *data)
{
	struct obj_array_node *an;
	struct obj_dict_node *dn;
	enum s_status res;

	if (obj -> type == OT_DICTIONARY) {
		for (dn = obj -> val.dictionary.head; 
		     dn != NULL; dn = dn -> next) {
			res = obj_navigate (&dn -> obj, f, data);
			if (res != S_CSC_OK)
				return res;
		}
	} else if (obj -> type == OT_ARRAY) {
		for (an = obj -> val.array.head; an != NULL; an = an -> next) {
			res = obj_navigate (&an -> obj, f, data);
			if (res != S_CSC_OK)
				return res;
		}
	} 
		
	return f (obj, data);
}

void obj_unwrap (struct obj *obj, void *val)
{
	switch (obj -> type) {
	case OT_ANY:
	case OT_NULL:
		break;

	case OT_BOOLEAN:
		*(pdf_bool_t *) val = obj -> val.bool;
		break;

	case OT_REAL:
		*(pdf_real_t *) val = obj -> val.real;
		break;

	case OT_INTEGER:
		*(pdf_integer_t *) val = obj -> val.integer;
		break;

	case OT_NAME:
	case OT_STRING:
		*(struct obj_bstr *) val = obj -> val.string;
		break;

	case OT_REGION:
		*(struct obj_region *) val = obj -> val.region;
		break;

	case OT_ARRAY:
		*(struct obj_array *) val = obj -> val.array;
		break;

	case OT_DICTIONARY:
		*(struct obj_dictionary *) val = obj -> val.dictionary;
		break;

	case OT_REFERENCE:
		break;

	case OT_XREF:
		*(struct obj_xref *) val = obj -> val.xref;
		break;

	case OT_DEFINITION:
		*(struct obj_definition *) val = obj -> val.definition;
		break;

	case OT_STREAM:
		*(struct obj_stream *) val = obj -> val.stream;
		break;

	case OT_ALIVE:
		*(struct obj_alive *) val = obj -> val.alive;
		break;

	case OT_KEYWORD:
		*(struct obj_keyword **) val = obj -> val.keyword;
		break;

	case OT_OBJ_PTR:
		obj_unwrap (&obj -> val.obj_ptr -> obj, val);
	}
}

void obj_wrap (void *val, enum obj_type ot, 
	       struct f_ptr *ptr, struct obj *obj)
{
	memset (obj, 0, sizeof (struct obj));

	switch (ot) {
	case OT_ANY:
	case OT_NULL:
		obj_set_null (obj, ptr);
		break;

	case OT_BOOLEAN:
		obj_set_boolean (obj, ptr, *(pdf_bool_t *) val);
		break;

	case OT_REAL:
		obj_set_real (obj, ptr, *(pdf_real_t *) val);
		break;

	case OT_INTEGER:
		obj_set_integer (obj, ptr, *(pdf_integer_t *) val);
		break;

	case OT_NAME:
	case OT_STRING:
		obj_set_string (obj, ptr, (struct obj_bstr *) val);
		break;

	case OT_REGION:
		obj_set_region (obj, ptr, (struct obj_region *) val);
		break;

	case OT_ARRAY:
		obj_set_array (obj, ptr, (struct obj_array *) val);
		break;

	case OT_DICTIONARY:
		obj_set_dictionary (obj, ptr, (struct obj_dictionary *) val);
		break;

	case OT_REFERENCE:
		break;

		/* These are not really needed */
	case OT_XREF:
	case OT_DEFINITION:
	case OT_STREAM:
	case OT_ALIVE:
	case OT_KEYWORD:
	case OT_OBJ_PTR:
		assert (0);
		break;
	}
}

int obj_region_equals (struct obj_region *reg1, struct obj_region *reg2)
{
	size_t size;

	if (reg1 -> begin == reg2 -> begin)
		return 1;

	if (reg1 -> begin == NULL || reg2 -> begin == NULL)
		return 0;

	size = reg1 -> end - reg1 -> begin;
	if (size != reg2 -> end - reg2 -> begin)
		return 0;

	return !memcmp (reg1 -> begin, reg2 -> begin, size);
}

int obj_array_equals (struct obj_array *array1, struct obj_array *array2)
{
	struct obj_array_node *an1, *an2;

	an1 = array1 -> head;
	an2 = array2 -> head;

	while (an1 != NULL) {
		if (an2 == NULL)
			return 0;

		if (!obj_equals (&an1 -> obj, &an2 -> obj))
			return 0;

		an1 = an1 -> next;
		an2 = an2 -> next;
	}

	if (an2 != NULL)
		return 0;

	return 1;
}

int obj_xref_equals (struct obj_xref *xref1, struct obj_xref *xref2)
{
	if (xref1 -> valid != xref2 -> valid ||
	    xref1 -> gnum != xref2 -> gnum)
		return 0;

	return xref1 -> valid ? 
		xref1 -> u.ofs == xref2 -> u.ofs :
		xref1 -> u.next == xref2 -> u.next;
}

int obj_dict_equals (struct obj_dictionary *dict1, 
		     struct obj_dictionary *dict2)
{
	struct obj_dict_node *dn1, *dn2;
	int size;

	size = 0;
	for (dn1 = dict1 -> head; dn1 != NULL; dn1 = dn1 -> next)
		size++;
	for (dn2 = dict2 -> head; dn2 != NULL; dn2 = dn2 -> next)
		size--;
	if (size)
		return 0;

	for (dn1 = dict1 -> head; dn1 != NULL; dn1 = dn1 -> next) {
		for (dn2 = dict2 -> head; dn2 != NULL; dn2 = dn2 -> next) {
			if (obj_string_equal_s (&dn1 -> key, &dn2 -> key)) {
				if (!obj_equals (&dn1 -> obj, &dn2 -> obj))
					return 0;
				break;
			}

			if (dn2 == NULL)
				return 0;
		}
	}

	return 1;
}

int obj_equals (struct obj *obj1, struct obj *obj2)
{
	struct obj mobj1, mobj2;

	if (obj1 -> type == OT_REAL || obj2 -> type == OT_REAL) {
		mobj1 = *obj1;
		mobj2 = *obj2;
		if (!obj_cast (&mobj1, OT_REAL))
			return 0;

		if (!obj_cast (&mobj1, OT_REAL))
			return 0;

		return mobj1.val.real == mobj2.val.real;
	}

	if (obj1 -> type != obj2 -> type)
		return 0;

	obj1 = obj_ptr_get (obj1);
	obj2 = obj_ptr_get (obj2);

	switch (obj1 -> type) {
	case OT_ANY:
	case OT_NULL:
		return 1;

	case OT_BOOLEAN:
		return obj1 -> val.bool == obj2 -> val.bool;

	case OT_REAL:
		return obj1 -> val.real == obj2 -> val.real;

	case OT_INTEGER:
		return obj1 -> val.integer == obj2 -> val.integer;

	case OT_NAME:
	case OT_STRING:
		return obj_string_equal_s (&obj1 -> val.string,
					   &obj2 -> val.string);

	case OT_REGION:
		return obj_region_equals (&obj1 -> val.region,
					  &obj2 -> val.region);

	case OT_ARRAY:
		return obj_array_equals (&obj1 -> val.array,
					 &obj2 -> val.array);

	case OT_DICTIONARY:
		return obj_dict_equals (&obj1 -> val.dictionary,
					&obj2 -> val.dictionary);

	case OT_REFERENCE:
		return obj_same (obj1, obj2);

		/* These should never be called */
	case OT_DEFINITION:
	case OT_XREF:
	case OT_STREAM:
	case OT_ALIVE:
	case OT_KEYWORD:
	case OT_OBJ_PTR:
		assert (0);
	}

	return 0;
}

int obj_same (const struct obj *obj1, const struct obj *obj2)
{
	return 	obj1 -> ref.oid &&
		obj1 -> ref.oid == obj2 -> ref.oid &&
		obj1 -> ref.gnum == obj2 -> ref.gnum;
}

int obj_exists (const struct obj *obj)
{
	return obj -> ref.oid;
}

struct obj_dictionary *obj_get_rw_dictionary (struct obj *obj)
{
	struct obj_dictionary *dict;

	obj = obj_ptr_get (obj);

	switch (obj -> type) {
	case OT_NULL:
		/* Can't use set_dictionary because it will reset
		 * its oid and gnum */
		obj -> type = OT_DICTIONARY;
		obj -> val.dictionary = obj_null_dictionary.val.dictionary;

		/* Fall trough */
	case OT_DICTIONARY:
		dict = &obj -> val.dictionary;
		break;

	case OT_STREAM:
		dict = &obj -> val.stream.dictionary;
		break;
			
	default:
		return NULL;
	}

	return dict;
}

struct obj_array *obj_get_rw_array (struct obj *obj)
{
	struct obj_array *array;

	obj = obj_ptr_get (obj);

	switch (obj -> type) {
	case OT_NULL:
		/* Can't use set_array because it will reset
		 * its oid and gnum */
		obj -> type = OT_ARRAY;
		obj -> val.array = obj_null_array.val.array;

		/* Fall trough */
	case OT_ARRAY:
		array = &obj -> val.array;
		break;
			
	default:
		assert (0);
	}

	return array;
}

void obj_unload (struct obj *obj)
{
	struct obj robj;

	obj_copy_reference (&robj, obj);
	obj_destroy (obj);
	*obj = robj;
}

void obj_destroy (struct obj *obj)
{
	switch (obj -> type) {
	case OT_ANY:
	case OT_NULL:
	case OT_BOOLEAN:
	case OT_REAL:
	case OT_INTEGER:
		break;

	case OT_NAME:
	case OT_STRING:
		obj_string_destroy (&obj -> val.string);
		break;

	case OT_REGION:
		obj_region_destroy (&obj -> val.region);
		break;

	case OT_ARRAY:
		obj_array_destroy (&obj -> val.array);
		break;

	case OT_DICTIONARY:
		obj_dict_destroy (&obj -> val.dictionary);
		break;

	case OT_REFERENCE:
	case OT_XREF:
		break;

	case OT_DEFINITION:
		obj_destroy (obj -> val.definition.obj);
		free (obj -> val.definition.obj);
		break;

	case OT_STREAM:
		obj_stream_destroy (&obj -> val.stream);
		break;

	case OT_ALIVE:
		obj_alive_destroy (&obj -> val.alive);
		break;

	case OT_KEYWORD:
		break;

	case OT_OBJ_PTR:
		if (!--obj -> val.obj_ptr -> refs) {
			obj_destroy (&obj -> val.obj_ptr -> obj);
			free (obj -> val.obj_ptr);
		}
	}

	*obj = obj_null;
}

enum s_status obj_dup1 (struct obj *obj)
{
	struct obj sobj;

	sobj = *obj;
	
	return obj_dup (&sobj, obj);
}

enum s_status obj_dup (const struct obj *sobj, struct obj *dobj)
{
	enum s_status res;

	*dobj = *sobj;

	switch (sobj -> type) {
	case OT_ANY:
	case OT_NULL:
	case OT_BOOLEAN:
	case OT_REAL:
	case OT_INTEGER:
		break;

	case OT_NAME:
	case OT_STRING:
		res = obj_string_dup (&sobj -> val.string,
				      &dobj -> val.string);
		if (res != S_CSC_OK)
			return res;
		break;

	case OT_REGION:
		res = obj_region_dup (&sobj -> val.region,
				      &dobj -> val.region);
		if (res != S_CSC_OK)
			return res;
		break;

	case OT_ARRAY:
		res = obj_array_dup (&sobj -> val.array,	
				     &dobj -> val.array);
		if (res != S_CSC_OK)
			return res;
		break;

	case OT_DICTIONARY:
		res = obj_dict_dup (&sobj -> val.dictionary,
				    &dobj -> val.dictionary);
		if (res != S_CSC_OK)
			return res;
		break;

	case OT_REFERENCE:
	case OT_XREF:
		break;

	case OT_DEFINITION:
		dobj -> val.definition.obj = malloc (sizeof (struct obj));
		if (dobj -> val.definition.obj == NULL)
			return S_CSC_OUT_OF_MEMORY;

		res = obj_dup (sobj -> val.definition.obj,
			       dobj -> val.definition.obj);
		if (res != S_CSC_OK) {
			free (dobj -> val.definition.obj);
			return res;
		}
		break;

	case OT_STREAM:
		res = obj_dict_dup (&sobj -> val.stream.dictionary,
				    &dobj -> val.stream.dictionary);
		if (res != S_CSC_OK)
			return res;

		res = obj_region_dup (&sobj -> val.stream.region,
				      &dobj -> val.stream.region);
		if (res != S_CSC_OK) {
			obj_dict_destroy (&dobj -> val.stream.dictionary);
			return res;
		}

		res = obj_region_dup (&sobj -> val.stream.decoded,
				      &dobj -> val.stream.decoded);
		if (res != S_CSC_OK) {
			obj_region_destroy (&dobj -> val.stream.region);
			obj_dict_destroy (&dobj -> val.stream.dictionary);
			return res;
		}
		break;

	case OT_ALIVE:
		res = obj_alive_dup (&sobj -> val.alive,
				     &dobj -> val.alive);
		if (res != S_CSC_OK)
			return res;
		break;

	case OT_KEYWORD:
		break;

	case OT_OBJ_PTR:
		sobj -> val.obj_ptr -> refs++;

	}

	return S_CSC_OK;
}

const char *obj_tag (enum obj_type ot)
{
	switch (ot) {
	case OT_ANY:
		return "(any)";

	case OT_NULL:
		return "null";

	case OT_BOOLEAN:
		return "boolean";

	case OT_REAL:
		return "real";

	case OT_INTEGER:
		return "integer";

	case OT_NAME:
		return "name";

	case OT_STRING:
		return "string";

	case OT_REGION:
		return "region";

	case OT_ARRAY:
		return "array";

	case OT_DICTIONARY:
		return "dictionary";

	case OT_REFERENCE:
		return "reference";

	case OT_XREF:
		return "xref";

	case OT_DEFINITION:
		return "definition";

	case OT_STREAM:
		return "stream";

	case OT_ALIVE:
		return "some_brisky_object";

	case OT_KEYWORD:
		return "keyword";

	case OT_OBJ_PTR:
		return "obj_ptr";
	}

	return "???";
}

void obj_store_init (struct obj_store *lobj)
{
	lobj -> next = NULL;
	lobj -> level = 0;
}

enum s_status obj_store_add (struct obj_store *lobj, struct obj *obj)
{
	while (lobj -> next != NULL)
		lobj = lobj -> next;
	if (lobj -> level >= OBJ_STORE_SIZE) {
		lobj -> next = malloc (sizeof (struct obj_store));
		lobj = lobj -> next;
		if (lobj == NULL)
			return S_CSC_OUT_OF_MEMORY;
		obj_store_init (lobj);
	}

	lobj -> obj [lobj -> level++] = *obj;

	return S_CSC_OK;
}

int obj_mark (struct obj_store *lobj)
{
	int ans;

	for (ans = 0; lobj -> next != NULL; lobj = lobj -> next)
		ans += OBJ_STORE_SIZE;

	return ans + lobj -> level;
}

enum s_status obj_rollback (struct obj_store *lobj, int p, enum s_status res)
{
	while (p > OBJ_STORE_SIZE) {
		lobj = lobj -> next;
		p -= OBJ_STORE_SIZE;
	}
		
	while (p < lobj -> level)
		obj_destroy (&lobj -> obj [p++]);

	obj_commit (lobj);

	return res;
}

enum s_status obj_cleanup (struct obj_store *lobj, enum s_status res)
{
	return obj_rollback (lobj, 0, res);
}

int obj_can_cast (const struct obj *obj, enum obj_type ot)
{
	if (obj -> type == OT_NULL)
		return 1;

	switch (ot) {
	case OT_ANY:
		return 1;

	case OT_STREAM:
	case OT_DICTIONARY:
		return 	obj -> type == OT_DICTIONARY ||
			obj -> type == OT_STREAM;

	case OT_ARRAY:
		return obj -> type == OT_ARRAY;

	case OT_INTEGER:
	case OT_REAL:
		return 	obj -> type == OT_REAL ||
			obj -> type == OT_INTEGER;

	default:
		return obj -> type == ot;

	}
}

int obj_cast (struct obj *obj, enum obj_type ot)
{
	if (!obj_can_cast (obj, ot))
		return 0;

	if (obj -> type == OT_NULL)
		return 1;

	switch (ot) {
	case OT_ANY:
		return 1;

	case OT_DICTIONARY:
		if (obj -> type == OT_STREAM) {
			obj_region_destroy (&obj -> val.stream.region);
			obj_region_destroy (&obj -> val.stream.decoded);
		}

		obj -> val.dictionary = *obj_get_rw_dictionary (obj);
		break;

	case OT_STREAM:
		if (obj -> type == OT_DICTIONARY) {
			obj -> val.stream.dictionary = 
				*obj_get_rw_dictionary (obj);
			memset (&obj -> val.stream.region, 0,
				sizeof (struct obj_region));
			memset (&obj -> val.stream.decoded, 0,
				sizeof (struct obj_region));
		}
		break;

	case OT_ARRAY:
		obj -> val.array = *obj_get_rw_array (obj);
		break;

	case OT_INTEGER:
		if (obj -> type == OT_REAL) 
			obj -> val.integer = obj -> val.real;
		else if (obj -> type == OT_NULL)
			obj -> val.integer = 0;
		break;

	case OT_REAL:
		if (obj -> type == OT_INTEGER) 
			obj -> val.real = obj -> val.integer;
		else if (obj -> type == OT_NULL)
			obj -> val.real = 0;
		break;

	default:
		return 1;
	}

	obj -> type = ot;

	return 1;
}

void obj_commit (struct obj_store *lobj)
{
	struct obj_store *lnobj;
	int i;

	while (lobj -> next != NULL) {
		lnobj = lobj -> next;
		lobj -> next = lnobj -> next;

		for (i = 0; i < lnobj -> level; i++)
			obj_destroy (&lnobj -> obj [i]);

		free (lnobj);
	}
}

enum s_status obj_verror (const struct obj *obj, const char *fmt, va_list ap)
{
	return f_verror (&obj -> ptr, fmt, ap);
}

enum s_status obj_error (const struct obj *obj, const char *fmt, ...)
{
	enum s_status res;
	va_list ap;

	va_start (ap, fmt);
	res = obj_verror (obj, fmt, ap);
	va_end (ap);

	return res;
}
