/*
 * PDFlow -- objbr.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <assert.h>
#include <string.h>

#include "fmt/objbr.h"
#include "fmt/fields.h"

#define BTREE_MIN_SIZE 70

static int compare (bt_key_t *key1, bt_key_t *key2);
static void destroy_data (bt_tree_data_t *tdata, bt_data_t *data);

static enum s_status cross_traverse (struct ob_bridge *ob, struct obj *obj);
static void cancel_cross_traverse (struct ob_bridge *ob, struct obj *obj);

static struct bt_param tree_param = {
	
	sizeof (struct obj_reference),
	BTREE_MIN_SIZE,
	NULL,
	compare, NULL, NULL, NULL, destroy_data

};

static struct ob_bridge *cast_tree_data (bt_tree_data_t *tdata)
     __attribute__ ((used));

static struct ob_bridge *cast_tree_data (bt_tree_data_t *tdata)
{
	return (struct ob_bridge *) tdata;
}

static bt_tree_data_t *uncast_tree_data (struct ob_bridge *ob)
{
	return (struct bt_imp_data_t *) ob;
}

static struct obj_reference *cast_key (bt_key_t *key)
{
	return (struct obj_reference *) key;
}

static bt_key_t *uncast_key (struct obj_reference *ref)
{
	return (struct bt_key_t *) ref;
}

static struct obj_reference *cast_data (bt_key_t *data)
{
	return (struct obj_reference *) data;
}

static bt_data_t *uncast_data (struct obj_reference *ref)
{
	return (struct bt_data_t *) ref;
}

static int compare (bt_key_t *key1, bt_key_t *key2)
{
	struct obj_reference *ref1, *ref2;
	int ans;

	ref1 = cast_key (key1);
	ref2 = cast_key (key2);

	ans = ref1 -> oid - ref2 -> oid;
	if (ans)
		return ans;

	return ref1 -> gnum - ref2 -> gnum;
}

static void destroy_data (bt_tree_data_t *tdata, bt_data_t *data)
{
	free (cast_data (data));
}

static enum s_status init (struct ob_bridge *ob)
{
	struct bt_param param;
	enum s_status res;

	param = tree_param;
	param.tdata = uncast_tree_data (ob);

	res = bt_init (&ob -> tree, &param, ref_rown (&ob -> refs));
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

enum s_status ob_new (struct ob_bridge **obp, struct xref_db *srdb, 
		      struct xref_db *drdb, struct ref_owner *p)
{
	struct ob_bridge *ob;
	enum s_status res;

	ob = malloc (sizeof (struct ob_bridge));
	if (ob == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (ob, 0, sizeof (struct ob_bridge));

	ref_init (&ob -> refs, "obridge", p);
	xref_ref_db (ob -> srdb = srdb, ref_rown (&ob -> refs));
	xref_ref_db (ob -> drdb = drdb, ref_rown (&ob -> refs));
	
	res = init (ob);
	if (res != S_CSC_OK) {
		ob_unref (ob, p);
		return res;
	}

	*obp = ob;

	return S_CSC_OK;
}

static void cancel_cross (struct ob_bridge *ob, struct obj *obj)
{
	xref_remove_object (ob -> drdb, obj);
	memset (&obj -> ref, 0, sizeof (struct obj_reference));
}

static void array_cancel_cross (struct ob_bridge *ob, struct obj *obj)
{
	struct obj_array_node *an;
	struct obj_array *array;

	if (obj -> type == OT_NULL)
		return;

	assert (obj -> type == OT_ARRAY);
	array = &obj -> val.array;
	for (an = array -> head; an != NULL; an = an -> next)
		cancel_cross_traverse (ob, &an -> obj);
}

static void dict_cancel_cross (struct ob_bridge *ob, struct obj *obj)
{
	struct obj_dictionary *dict;
	struct obj_dict_node *dn;

	if (obj -> type == OT_NULL)
		return;

	assert (obj -> type == OT_DICTIONARY || obj -> type == OT_STREAM);

	if (obj -> type == OT_DICTIONARY)
		dict = &obj -> val.dictionary;
	else 
		dict = &obj -> val.stream.dictionary;

	for (dn = dict -> head; dn != NULL; dn = dn -> next)
		cancel_cross_traverse (ob, &dn -> obj);
}

static void cancel_cross_traverse (struct ob_bridge *ob, struct obj *obj)
{
	cancel_cross (ob, obj);

	switch (obj -> type) {
	case OT_ANY:
	case OT_NULL:
	case OT_BOOLEAN:
	case OT_REAL:
	case OT_INTEGER:
	case OT_NAME:
	case OT_STRING:
	case OT_REGION:
	case OT_REFERENCE:
	case OT_XREF:
	case OT_DEFINITION:
	case OT_ALIVE:
	case OT_KEYWORD:
	case OT_OBJ_PTR:
		break;

	case OT_ARRAY:
		array_cancel_cross (ob, obj);
			
		break;

	case OT_DICTIONARY:
	case OT_STREAM:
		dict_cancel_cross (ob, obj);
		break;
	}
}

static int must_be_serialized (struct obj *obj)
{
	return obj_exists (obj) || obj -> type == OT_STREAM;
}

static enum s_status cross (struct ob_bridge *ob, struct obj *obj)
{
	struct obj_reference *ref, sref;
	bt_data_t *data;
	struct obj dobj;
	int pres, keyed;
	enum s_status res;

	if (!must_be_serialized (obj))
		return S_CSC_OK;

	keyed = obj_exists (obj);

	res = xref_resolve (ob -> srdb, obj, OT_ANY, &dobj, 
			    &pres, "object to be bridged");
	if (res != S_CSC_OK)
		return res;

	obj_destroy (obj);
	*obj = dobj;
	
	res = bt_get (ob -> tree, uncast_key (&obj -> ref), &data);
	if (res != S_CSC_OK)
		return res;

	if (data != NULL) {
		obj -> ref = *cast_data (data);
		return S_CSC_OK;
	}

	sref = obj -> ref;
	memset (&obj -> ref, 0, sizeof (struct obj_reference));

	res = xref_register_object (ob -> drdb, obj);
	if (res != S_CSC_OK) {
		obj -> ref = sref;
		return res;
	}	

	if (keyed) {
		ref = malloc (sizeof (struct obj_reference));
		if (ref == NULL)
			return S_CSC_OUT_OF_MEMORY;

		*ref = obj -> ref;

		res = bt_insert (ob -> tree, uncast_key (&sref), 
				 uncast_data (ref), NULL);
		if (res != S_CSC_OK) {
			cancel_cross (ob, obj);
			free (ref);
			return res;
		}
	}

	return S_CSC_OK;
}

static enum s_status array_cross (struct ob_bridge *ob, struct obj *obj)
{
	struct obj_array_node *an, *bn;
	struct obj_array *array;
	enum s_status res;

	if (obj -> type == OT_NULL)
		return S_CSC_OK;

	assert (obj -> type == OT_ARRAY);
	array = &obj -> val.array;
	res = S_CSC_OK;
	for (an = array -> head; an != NULL; an = an -> next) {
		res = cross_traverse (ob, &an -> obj);
		if (res != S_CSC_OK)
			break;
	}
	if (res != S_CSC_OK) {
		for (bn = array -> head; bn != an; bn = bn -> next)
			cancel_cross_traverse (ob, &bn -> obj);
		return res;
	}

	return S_CSC_OK;
}

static enum s_status dict_cross (struct ob_bridge *ob, struct obj *obj)
{
	struct obj_dictionary *dict;
	struct obj_dict_node *dn, *bn;
	enum s_status res;

	if (obj -> type == OT_NULL)
		return S_CSC_OK;

	assert (obj -> type == OT_DICTIONARY || obj -> type == OT_STREAM);

	if (obj -> type == OT_DICTIONARY)
		dict = &obj -> val.dictionary;
	else 
		dict = &obj -> val.stream.dictionary;

	res = S_CSC_OK;
	for (dn = dict -> head; dn != NULL; dn = dn -> next) {
		res = cross_traverse (ob, &dn -> obj);
		if (res != S_CSC_OK)
			break;
	}

	if (res != S_CSC_OK) {
		for (bn = dict -> head; bn != dn; bn = bn -> next)
			cancel_cross_traverse (ob, &bn -> obj);
		return res;
	}

	return S_CSC_OK;
}

enum s_status cross_traverse (struct ob_bridge *ob, struct obj *obj)
{
	enum s_status res;

	res = cross (ob, obj);
	if (res != S_CSC_OK)
		return res;

	switch (obj -> type) {
	case OT_ANY:
	case OT_NULL:
	case OT_BOOLEAN:
	case OT_REAL:
	case OT_INTEGER:
	case OT_NAME:
	case OT_STRING:
	case OT_REGION:
	case OT_REFERENCE:
	case OT_XREF:
	case OT_DEFINITION:
	case OT_ALIVE:
	case OT_KEYWORD:
	case OT_OBJ_PTR:
		res = S_CSC_OK;
		break;

	case OT_ARRAY:
		res = array_cross (ob, obj);
			
		break;

	case OT_DICTIONARY:
	case OT_STREAM:
		res = dict_cross (ob, obj);
		break;
	}

	if (res != S_CSC_OK) {
		cancel_cross_traverse (ob, obj);
		return res;
	}

	return S_CSC_OK;
}

enum s_status ob_lookup (struct ob_bridge *ob, 
			 struct obj *sobj, struct obj *dobj)
{
	enum s_status res;
	struct obj obj;

	res = obj_dup (sobj, &obj);
	if (res != S_CSC_OK)
		return res;

	res = cross_traverse (ob, &obj);
	if (res != S_CSC_OK) {
		obj_destroy (&obj);
		return res;
	}

	*dobj = obj;

	return S_CSC_OK;
}

void ob_ref (struct ob_bridge *ob, struct ref_owner *p)
{
	if (ob != NULL)
		ref_up (&ob -> refs, p);
}

void ob_unref (struct ob_bridge *ob, struct ref_owner *p)
{
	if (ob == NULL || ref_down (&ob -> refs, p))
		return;

	if (ob -> srdb != NULL)
		xref_unref_db (ob -> srdb, ref_rown (&ob -> refs));
	if (ob -> drdb != NULL)
		xref_unref_db (ob -> drdb, ref_rown (&ob -> refs));

	if (ob -> tree != NULL)
		bt_unref (ob -> tree, ref_rown (&ob -> refs));

	ref_zero (&ob -> refs);

	free (ob);
}

