/*
 * PDFlow -- pdf-lex.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <stdio.h>
#include <ctype.h>
#include <assert.h>

#include "fmt/obj.h"
#include "fmt/pdf-lex.h"

#define WS "\t\r\n\f "
#define DELIMITER "()<>[]{}/%"
#define NONREG WS DELIMITER "#"

static enum s_status stream_handler (struct pl_state *ps, 
				     enum pl_token_type *tt);

struct keyword {
	const char *name;
	int token;
	enum s_status (*handler) (struct pl_state *ps, enum pl_token_type *tt);
} keys [] = {
	{ "null", TOK_NULL, NULL },
	{ "true", TOK_TRUE, NULL },
	{ "false", TOK_FALSE, NULL },
	{ "obj", TOK_OBJ, NULL },
	{ "endobj", TOK_ENDOBJ, NULL },
	{ "stream", TOK_STREAM, stream_handler },
	{ "endstream", TOK_ENDSTREAM, NULL },
	{ "R", TOK_REF, NULL },
	{ "xref", TOK_XREF, NULL },
	{ "n", TOK_N, NULL },
	{ "f", TOK_F, NULL },
	{ "trailer", TOK_TRAILER, NULL },
	{ "startxref", TOK_STARTXREF, NULL },
	{ NULL }
};

void pl_init (struct pl_state *ps, struct f_ptr *ptr)
{
	pl_init_kw (ps, ptr, NULL);
}

void pl_init_kw (struct pl_state *ps, struct f_ptr *ptr, 
		 struct obj_keyword *keys)
{
	ps -> ptr = *ptr;
	ps -> eof = 0;
	ps -> keys = keys;
}

static int adv (struct pl_state *ps, int *eol)
{
	if (ps -> eof)
		return 0;

	ps -> ptr.cptr++;
	if (!f_ptr_valid (&ps -> ptr)) {
		ps -> eof = 1;
		return 0;
	}
	if (*ps -> ptr.cptr == '\r' &&
	    f_ptr_test (&ps -> ptr, 1) && ps -> ptr.cptr [1] == '\n')
		ps -> ptr.cptr++;

	if (eol != NULL)
		*eol = *ps -> ptr.cptr == '\r' || *ps -> ptr.cptr == '\n';

	return 1;
}

static int is_ws (char c)
{
	return !c || strchr (WS, c) != NULL;
}

int pl_is_blank (char c)
{
	return is_ws (c);
}

static int is_del (char c)
{
	return is_ws (c) || strchr (DELIMITER, c) != NULL;
}

static int skip_ws (struct pl_state *ps)
{
	char c;

	if (!f_ptr_valid (&ps -> ptr))
		return 0;

	do {
		c = *ps -> ptr.cptr;
		if (!is_ws (*ps -> ptr.cptr))
			return 1;
	} while (adv (ps, NULL));


	return 0;
}

int pl_skip_ws (struct pl_state *ps)
{
	return skip_ws (ps);
}

static int skip_eol (struct pl_state *ps)
{
	int eol;

	while (adv (ps, &eol))
		if (eol)
			return adv (ps, NULL);

	return 0;
}

static enum s_status parse_number (struct pl_state *ps, 
				   enum pl_token_type *tt)
{
	pdf_real_t dans, dig;
	pdf_integer_t ians;
	int dec, real;
	char c;

	dans = ians = real = 0;
	dec = dig = 0;
	do {
		c = *ps -> ptr.cptr;
		if (c == '.' && !dec) {
			dig = .1;
			dec = 1;
		} else if (!isdigit (c))
			break;
		else if (dec) {
			dans += dig * (c - '0');
			dig /= 10;
		} else if (real) {
			dans *= 10;
			dans += c - '0';
		} else {
			dans *= 10;
			dans += c - '0';

			ians *= 10;
			ians += c - '0';
			if (ians < dans && (!ians || ians != -ians))
				real = 1;
		}
	} while (adv (ps, NULL));

	if (dec) {
		*tt = TOK_REAL;
		obj_set_real (&ps -> token, &ps -> sptr, dans);
	} else if (!real) {
		*tt = TOK_INTEGER;
		obj_set_integer (&ps -> token, &ps -> sptr, ians);
	} else
		return pl_error (ps, "Integer too large");

	return S_CSC_OK;
}

static enum s_status parse_escape (struct pl_state *ps, int *val)
{
	struct pl_state pps;
	char c, ans;
	int i;

	pps = *ps;
	ans = i = 0;
	do {
		c = *ps -> ptr.cptr;
		if (c < '0' || c >= '8')
			break;
		ans <<= 3;
		ans |= c - '0';
		if (++i >= 3)
			break;

	} while (adv (ps, NULL));

	if (i > 0) {
		*val = ans;
		return S_CSC_OK;
	}

	*ps = pps;
	switch (*ps -> ptr.cptr) {
	case 'n':
		*val = '\n';
		break;

	case 'r':
		*val = '\r';
		break;

	case 't':
		*val = '\t';
		break;

	case 'b':
		*val = '\b';
		break;
			
	case 'f':
		*val = '\f';
		break;

	case '\r':
	case '\n':
		*val = -1;
		break;
			
	case '(':
	case ')':
	case '\\':
		*val = *ps -> ptr.cptr;
		break;

	default:
		return pl_error (ps, "Unknown escape character (%c)", 
				 *ps -> ptr.cptr);	
	}

	return S_CSC_OK;
}

static enum s_status parse_literal (struct pl_state *ps, 
				    enum pl_token_type *tt)
{
	enum s_status res;
	struct pl_state pps;
	struct obj_bstr s;
	char *q;
	int nest, esc, sz;

	pps = *ps;

	nest = 1;
	while (adv (&pps, NULL)) {
		switch (*pps.ptr.cptr) {
		case '\\':
			pps.ptr.cptr++;
			break;

		case '(':
			nest++;
			break;

		case ')':
			nest--;
			break;
		}
		if (!nest)
			break;
	}

	if (pps.eof)
		return pl_error (&pps, "Unterminated string");

	adv (ps, NULL);

	sz = pps.ptr.cptr - ps -> ptr.cptr;
	if (!sz)
		sz++;
	q = s.s = malloc (sz);
	if (s.s == NULL)
		return pl_error (ps, "Out of memory");

	while (ps -> ptr.cptr < pps.ptr.cptr) {
		switch (*ps -> ptr.cptr) {
		case '\\':
			if (!adv (ps, NULL)) {
				free (s.s);
				return pl_error (ps, "End of file while " 
						 "parsing escape char");
			}
			esc = 0x100;
			res = parse_escape (ps, &esc);
			if (res != S_CSC_OK) {
				free (s.s);
				return res;
			}
			assert (esc != 0x100);

			if (esc == -1)
				q--;
			else
				*q = esc;
			break;

		case '\r':
			if (ps -> ptr.cptr + 1 < pps.ptr.cptr && 
			    ps -> ptr.cptr [1] == '\n') {
				adv (ps, NULL);
				q--;
			} else
				*q = '\n';
			break;
			
		default:
			*q = *ps -> ptr.cptr;
		}

		q++;
		adv (ps, NULL);
	}

	s.size = q - s.s;

	obj_set_string (&ps -> token, &ps -> sptr, &s);

	*tt = TOK_LITERAL;

	adv (ps, NULL);	/* Drop ')' */

	return S_CSC_OK;
}

static enum s_status parse_hexstring (struct pl_state *ps,
				      enum pl_token_type *tt)
{
	struct pl_state pps;
	struct obj_bstr s;
	char *q, c;
	int size, shift, n;

	pps = *ps;
	size = 0;
	while (adv (&pps, NULL)) {
		if (isxdigit (*pps.ptr.cptr))
			size++;
		if (is_ws (*pps.ptr.cptr))
			continue;
		if (*pps.ptr.cptr == '>')
			break;
	}

	if (pps.eof)
		return pl_error (ps, "Unterminated hex string");
	if (size & 1)
		size++;

	s.size = size >> 1;
	q = s.s = malloc (s.size);
	memset (q, 0, s.size);
	shift = 4;
	while (ps -> ptr.cptr < pps.ptr.cptr) {
		c = tolower (*ps -> ptr.cptr);
		if (isxdigit (c)) {
			if (isdigit (c))
				n = c - '0';
			else
				n = c - 'a' + 10;
			*q |= n << shift;

			if (!shift)
				q++;

			shift ^= 4;
		}
		adv (ps, NULL);
	}

	adv (ps, NULL);	/* Drop '>' */

	obj_set_string (&ps -> token, &ps -> sptr, &s);

	*tt = TOK_HEX;

	return S_CSC_OK;
}

static enum s_status parse_name (struct pl_state *ps,
				 enum pl_token_type *tt)
{
	struct pl_state pps;
	struct obj_bstr s;
	char *q, c;
	int esc;

	pps = *ps;
	while (adv (&pps, NULL)) {
		if (*pps.ptr.cptr == '#') {
			if (!adv (&pps, NULL) || !adv (&pps, NULL))
				return pl_error (ps, "Unterminated name");
		} else if (is_del (*pps.ptr.cptr))
			break;
	}

	if (!adv (ps, NULL) || pps.ptr.cptr == ps -> ptr.cptr)
		return pl_error (ps, "Empty name");
	
	q = s.s = malloc (pps.ptr.cptr - ps -> ptr.cptr);
	if (s.s == NULL)
		return S_CSC_OUT_OF_MEMORY;

	while (ps -> ptr.cptr < pps.ptr.cptr) {
		switch (*ps -> ptr.cptr) {
		case '#':
			adv (ps, NULL);
			c = tolower (*ps -> ptr.cptr);
			if (!isxdigit (c)) {
				free (s.s);
				return pl_error (ps, "Bad escape name");
			}
			if (isdigit (c))
				esc = c - '0';
			else
				esc = c - 'a' + 10;
			*q = esc << 4;

			adv (ps, NULL);
			c = tolower (*ps -> ptr.cptr);
			if (!isxdigit (c)) {
				free (s.s);
				return pl_error (ps, "Bad escape name");
			}
			if (isdigit (c))
				esc = c - '0';
			else
				esc = c - 'a' + 10;
			*q |= esc;
			break;
			
		default:
			*q = *ps -> ptr.cptr;
		}

		q++;
		adv (ps, NULL);
	}

	s.size = q - s.s;
	obj_set_name (&ps -> token, &ps -> sptr, &s);

	*tt = TOK_NAME;

	return S_CSC_OK;
}

/* Need a special handler to avoid swallowing blanks after the
 * stream newline */
static enum s_status stream_handler (struct pl_state *ps, 
				     enum pl_token_type *tt)
{
	ps -> ptr.cptr--;	/* eol starts by adv */
	skip_eol (ps);
	*tt = TOK_STREAM;

	return S_CSC_OK;
}

static enum s_status parse_keyword (struct pl_state *ps,
				    enum pl_token_type *tt)
{
	struct obj_keyword *kwp;
	struct keyword *p;
	struct obj_bstr key;
	size_t spare;
	int i, kwlen;

	spare = f_ptr_avail (&ps -> ptr);

	if (ps -> keys != NULL) {
		for (kwp = ps -> keys; kwp -> name != NULL; kwp++) {

			kwlen = strlen (kwp -> name);
			if (spare < kwlen)
				continue;

			if (spare > kwlen && !is_del (ps -> ptr.cptr [kwlen]))
				continue;

			if (!memcmp (ps -> ptr.cptr, kwp -> name, kwlen)) {

				obj_set_keyword (&ps -> token, 
						 &ps -> ptr, kwp);
				ps -> ptr.cptr += kwlen;
				skip_ws (ps);
				*tt = TOK_KEYWORD;
				return S_CSC_OK;
			}
		}
	}

	for (p = keys; p -> name != NULL; p++) {
		kwlen = strlen (p -> name);
		if (spare < kwlen)
			continue;

		if (spare > kwlen && !is_del (ps -> ptr.cptr [kwlen]))
			continue;

		if (!memcmp (ps -> ptr.cptr, p -> name, kwlen)) {
			obj_set_null (&ps -> token, &ps -> ptr);
			ps -> ptr.cptr += kwlen;
			if (p -> handler == NULL) {
				skip_ws (ps);
				*tt = p -> token;
				return S_CSC_OK;
			} else
				return p -> handler (ps, tt);
		}
	}

	for (i = 0; i < spare; i++)
		if (is_del (ps -> ptr.cptr [i]))
			break;

	key.size = i;
	key.s = malloc (key.size);
	if (key.s == NULL)
		return S_CSC_OUT_OF_MEMORY;

	memcpy (key.s, ps -> ptr.cptr, key.size);

	obj_set_string (&ps -> token, &ps -> ptr, &key);
	*tt = TOK_OTH_KEYWORD;

	ps -> ptr.cptr += i;

	return S_CSC_OK;
}

enum s_status pl_seek_bol (struct pl_state *ps, const char *key)
{
	size_t len;

	len = strlen (key);

	do {
		if (!f_ptr_test (&ps -> ptr, len))
			break;

		if (!strncmp ((char *) ps -> ptr.cptr, key, len) &&
		    (!f_ptr_test (&ps -> ptr, len + 1) ||
		     is_ws (ps -> ptr.cptr [len]))) {
			ps -> ptr.cptr += len;
			return S_CSC_OK;
		}
	} while (skip_eol (ps));

	return pl_error (ps, "EOF reached while looking for %s", key);
}

enum s_status pl_lex (struct pl_state *ps, enum pl_token_type *tt)
{
	char c;

	ps -> token.type = OT_ANY;

	do {
		if (!skip_ws (ps)) {
			*tt = TOK_EOF;
			return S_CSC_OK;
		}

		c = *ps -> ptr.cptr;
		if (c == '%')
			skip_eol (ps);
	} while (c == '%');

	ps -> sptr = ps -> ptr;
	ps -> token.ptr = ps -> sptr;

	switch (c) {
	case '(':
		return parse_literal (ps, tt);

	case '<':
		if (f_ptr_test (&ps -> ptr, 1) && ps -> ptr.cptr [1] == '<') {
			adv (ps, NULL);
			adv (ps, NULL);
			*tt = TOK_DICT_BEGIN;
			return S_CSC_OK;
		} else
			return parse_hexstring (ps, tt);

	case '>':
		if (f_ptr_test (&ps -> ptr, 1) && ps -> ptr.cptr [1] == '>') {
			adv (ps, NULL);
			adv (ps, NULL);
			*tt = TOK_DICT_END;
			return S_CSC_OK;
		} else
			return pl_error (ps, "Unexpected closed bracked");

	case '/':
		return parse_name (ps, tt);

	case '+':
		*tt = TOK_PLUS;
		break;

	case '-':
		*tt = TOK_MINUS;
		break;

	case '[':
		*tt = TOK_ARRAY_BEGIN;
		break;

	case ']':
		*tt = TOK_ARRAY_END;
		break;

	case '{':
		*tt = TOK_LCURLY;
		break;

	case '}':
		*tt = TOK_RCURLY;
		break;

	default:
		if (isdigit (c) || c == '.')
			return parse_number (ps, tt);

		return parse_keyword (ps, tt);
	}
	
	adv (ps, NULL);

	return S_CSC_OK;
}

enum s_status pl_lex_predict (struct pl_state *ps, enum pl_token_type tt,
			      const char *tag)
{
	enum pl_token_type gtt;
	enum s_status res;

	res = pl_lex (ps, &gtt);
	if (res != S_CSC_OK)
		return res;

	if (tt != gtt)
		return pl_error (ps, "Expecting %s", tag);

	return S_CSC_OK;
}

enum s_status pl_backline (struct pl_state *ps, unsigned n)
{
	ps -> ptr.cptr--;
	if (!f_ptr_valid (&ps -> ptr))
		return S_CSC_PARSE_ERROR;

	if (*ps -> ptr.cptr == '\r' || *ps -> ptr.cptr == '\n')
		n++;
	
	while (f_ptr_valid (&ps -> ptr)) {
		if (*ps -> ptr.cptr == '\n' || *ps -> ptr.cptr == '\r')
			n--;
		if (!n)
			break;

		if (*ps -> ptr.cptr == '\n' && 
		    f_ptr_test (&ps -> ptr, -1) && ps -> ptr.cptr [-1] == '\r')
			ps -> ptr.cptr--;

		ps -> ptr.cptr--;
	}

	if (n)
		return S_CSC_PARSE_ERROR;

	adv (ps, NULL);

	return S_CSC_OK;
}

static enum s_status serialize_name (struct f_file *file, 
				     struct obj_bstr *str)
{
	enum s_status res;
	const char *p;
	int i, ls;
	
	res = f_write (file, "/");
	if (res != S_CSC_OK)
		return res;
	p = (const char *) str -> s;
	ls = 0;
	for (i = 0; i < str -> size; i++)
		if (strchr (NONREG, p [i]) != NULL || !p [i]) {
			res = f_write (file, "%.*s#%02x", 
				       i - ls, p + ls, p [i]);
			if (res != S_CSC_OK)
				return res;
			ls = i + 1;
		}

	if (i > ls) {
		res = f_write (file, "%.*s", i - ls, p + ls);
		if (res != S_CSC_OK)
			return res;
	}

	return S_CSC_OK;
}

static enum s_status serialize_string (struct f_file *file, 
				       struct obj_bstr *str)
{
	enum s_status res;
	const char *q, *next;
	int i, j;
	
	res = f_write (file, "(");
	if (res != S_CSC_OK)
		return res;
	i = 0;
	while (i < str -> size) {
		next = NULL;
		q = memchr (str -> s + i, '\\', str -> size - i);
		if (q != NULL && (next == NULL || q < next))
			next = q;
		q = memchr (str -> s + i, '(', str -> size - i);
		if (q != NULL && (next == NULL || q < next))
			next = q;
		q = memchr (str -> s + i, ')', str -> size - i);
		if (q != NULL && (next == NULL || q < next))
			next = q;

		j = next != NULL ? next - str -> s : str -> size;
		res = f_raw (file, str -> s + i, j - i);
		if (res != S_CSC_OK)
			return res;

		if (next == NULL)
			break;
		f_write (file, "\\%c", str -> s [j]);

		i = j + 1;
	}

	res = f_write (file, ")");
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status serialize_reference (struct f_file *file,
					  struct obj *obj)
{
	return f_write (file, "%ld %ld R", obj -> ref.oid, obj -> ref.gnum);
}

static enum s_status serialize_array (struct f_file *file, 
				      struct obj_array *array)
{
	struct obj_array_node *an;
	enum s_status res;
	
	res = f_write (file, "[");
	if (res != S_CSC_OK)
		return res;

	for (an = array -> head; an != NULL; an = an -> next) {
		if (an != array -> head) {
			res = f_write (file, " ");
			if (res != S_CSC_OK)
				return res;
		}

		if (obj_exists (&an -> obj))
			res = serialize_reference (file, &an -> obj);
		else
			res = pl_serialize (file, &an -> obj);
		if (res != S_CSC_OK)
			return res;
	}

	res = f_write (file, "]");
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status serialize_dictionary (struct f_file *file, 
					   struct obj_dictionary *dict)
{
	struct obj_dict_node *dn;
	enum s_status res;
	
	res = f_write (file, "<<");
	if (res != S_CSC_OK)
		return res;

	for (dn = dict -> head; dn != NULL; dn = dn -> next) {
		if (dn != dict -> head) {
			res = f_write (file, " ");
			if (res != S_CSC_OK)
				return res;
		}

		res = serialize_name (file, &dn -> key);
		if (res != S_CSC_OK)
			return res;

		res = f_write (file, " ");
		if (res != S_CSC_OK)
			return res;

		if (obj_exists (&dn -> obj))
			res = serialize_reference (file, &dn -> obj);
		else
			res = pl_serialize (file, &dn -> obj);
		if (res != S_CSC_OK)
			return res;
	}

	res = f_write (file, ">>");
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status serialize_stream (struct f_file *file, 
				       struct obj_stream *stream)
{
	struct obj_region *reg;
	enum s_status res;
	struct obj obj;
	size_t len;
	
	reg = stream -> region.begin != NULL ? 
		&stream -> region : &stream -> decoded;
	len = reg -> end - reg -> begin;

	obj_set_integer (&obj, &f_null, len);
	res = obj_add_dict (&stream -> dictionary, "Length", &obj);
	if (res != S_CSC_OK)
		return res;
	
	res = serialize_dictionary (file, &stream -> dictionary);
	if (res != S_CSC_OK)
		return res;

	F_NEWLINE (file);
	F_WRITELN (file, "stream");
	
	res = f_raw (file, reg -> begin, len);
	if (res != S_CSC_OK)
		return res;

	F_NEWLINE (file);
	F_WRITE (file, "endstream");

	return S_CSC_OK;
}

enum s_status pl_serialize (struct f_file *file, struct obj *obj)
{
	enum s_status res;

	switch (obj -> type) {
	case OT_NULL:
		res = f_write (file, "null");
		break;

	case OT_BOOLEAN:
		res = f_write (file, obj -> val.bool ? "true" : "false");
		break;

	case OT_REAL:
		res = f_write (file, "%f", obj -> val.real);
		break;

	case OT_INTEGER:
		res = f_write (file, "%ld", obj -> val.integer);
		break;

	case OT_NAME:
		res = serialize_name (file, &obj -> val.string);
		break;

	case OT_STRING:
		res = serialize_string (file, &obj -> val.string);
		break;

	case OT_ARRAY:
		res = serialize_array (file, &obj -> val.array);
		break;

	case OT_DICTIONARY:
		res = serialize_dictionary (file, &obj -> val.dictionary);
		break;

	case OT_REFERENCE:
		res = serialize_reference (file, obj);
		break;

	case OT_STREAM:
		res = serialize_stream (file, &obj -> val.stream);
		break;

	default:
		return s_make_indirect (S_CSC_INTERNAL_ERROR,
					"Object type '%s' can't be serialized",
					obj_tag (obj -> type));
	}

	return res;
}

enum s_status pl_get_header (struct f_file *file, const char *name,
			     char **result)
{
	struct pl_state ps, qs;
	struct f_ptr fptr;
	size_t len, lsize;

	f_ptr_init (&fptr, file, 0);
	pl_init (&ps, &fptr);
	len = strlen (name);
	while (f_ptr_avail (&ps.ptr) >= len) {
		qs = ps;
		skip_eol (&qs);
		if (!memcmp (ps.ptr.cptr, name, len)) {
			ps.ptr.cptr += len;
			if (!skip_ws (&ps) || ps.ptr.cptr >= qs.ptr.cptr)
				break;
			if (*ps.ptr.cptr != ':')
				break;
			ps.ptr.cptr++;
			if (!skip_ws (&ps) || ps.ptr.cptr >= qs.ptr.cptr)
				break;
			do {
				qs.ptr.cptr--;
			} while (is_ws (*qs.ptr.cptr) &&
				 qs.ptr.cptr >= ps.ptr.cptr);

			lsize = qs.ptr.cptr - ps.ptr.cptr + 1;
			*result = malloc (lsize + 1);
			if (*result == NULL)
				return S_CSC_OUT_OF_MEMORY;
			memcpy (*result, ps.ptr.cptr, lsize);
			(*result) [lsize] = 0;

			return S_CSC_OK;
		}
		ps = qs;
	}

	*result = NULL;

	return S_CSC_OK;
}

enum s_status pl_verror (struct pl_state *ps, const char *fmt, va_list ap) 
{
	return f_verror (&ps -> ptr, fmt, ap);
}

enum s_status pl_error (struct pl_state *ps, const char *fmt, ...) 
{
	enum s_status res;
	va_list ap;

	va_start (ap, fmt);
	res = pl_verror (ps, fmt, ap);
	va_end (ap);

	return res;
}


