/*
 * PDFlow -- pdf-syn.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdarg.h>
#include <stdlib.h>

#include "fmt/pdf-lex.h"
#include "fmt/pdf-syn.h"
#include "fmt/machine.h"

#include "fmt/fields.h"
#include "fmt/fields.tab.h"

struct stream_info {
#include "pdf-syn.tab.h"	
};

#define XREF_IMPLEMENTATION struct stream_info
#include "pdf-syn.tab.h"
#undef XREF_IMPLEMENTATION

static enum s_status tr_array_begin (struct m_stack *stack,
				     struct obj *obj, int token,
				     void *closure);
static enum s_status tr_array_end (struct m_stack *stack,
				   struct obj *obj, int token,
				   void *closure);
static enum s_status tr_null (struct m_stack *stack,
			      struct obj *obj, int token,
			      void *closure);
static enum s_status tr_true (struct m_stack *stack,
			      struct obj *obj, int token,
			      void *closure);
static enum s_status tr_false (struct m_stack *stack,
			       struct obj *obj, int token,
			       void *closure);
static enum s_status tr_dict_begin (struct m_stack *stack,
				    struct obj *obj, int token,
				    void *closure);
static enum s_status tr_dict_end (struct m_stack *stack,
				  struct obj *obj, int token,
				  void *closure);
static enum s_status tr_real (struct m_stack *stack,
			      struct obj *obj, int token,
			      void *closure);
static enum s_status tr_obj (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure);
static enum s_status tr_endobj (struct m_stack *stack,
				struct obj *obj, int token,
				void *closure);
static enum s_status tr_integer (struct m_stack *stack,
				 struct obj *obj, int token,
				 void *closure);
static enum s_status tr_ref (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure);
static enum s_status tr_stream (struct m_stack *stack,
				struct obj *obj, int token,
				void *closure);
static enum s_status tr_endstream (struct m_stack *stack,
				   struct obj *obj, int token,
				   void *closure);
static enum s_status tr_xref_n (struct m_stack *stack,
				struct obj *obj, int token,
				void *closure);

static enum s_status tr_xref_f (struct m_stack *stack,
				struct obj *obj, int token,
				void *closure);
static enum s_status tr_keyword (struct m_stack *stack,
				 struct obj *obj, int token,
				 void *closure);

static enum s_status tr_oth_keyword (struct m_stack *stack,
				     struct obj *obj, int token,
				     void *closure);
static enum s_status tr_eof (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure);

static struct m_rule rules [] = {
	{ TOK_PLUS, NULL },
	{ TOK_MINUS, NULL },
	{ TOK_ARRAY_BEGIN, tr_array_begin },
	{ TOK_ARRAY_END, tr_array_end },
	{ TOK_NULL, tr_null },
	{ TOK_TRUE, tr_true },
	{ TOK_FALSE, tr_false },
	{ TOK_BOOLEAN, NULL },
	{ TOK_INTEGER, tr_integer },
	{ TOK_REAL, tr_real },
	{ TOK_NAME, NULL },
	{ TOK_LITERAL, NULL },
	{ TOK_HEX, NULL },
	{ TOK_STREAM, tr_stream },
	{ TOK_ENDSTREAM, tr_endstream },
	{ TOK_OBJ, tr_obj },
	{ TOK_ENDOBJ, tr_endobj },
	{ TOK_DICT_BEGIN, tr_dict_begin },
	{ TOK_DICT_END, tr_dict_end },
	{ TOK_REF, tr_ref }, 
	{ TOK_N, tr_xref_n }, 
	{ TOK_F, tr_xref_f }, 
	{ TOK_TRAILER, NULL },
	{ TOK_STARTXREF, NULL },
	{ TOK_KEYWORD, tr_keyword },
	{ TOK_EOF, tr_eof },
	{ TOK_OTH_KEYWORD, tr_oth_keyword },
	{ -1 }
};

struct ps_parser {
	struct m_machine *mach;
	struct m_state *state;
	struct xref_db *rdb;
};

struct ps_closure {

	struct ps_parser *parser;
	struct pl_state *ps;
	void *pst_closure;
	m_trigger def_handler;

};

static struct ps_parser template;

static enum s_status pc_error (struct ps_closure *pc, const char *fmt, ...) 
{
	enum s_status res;
	va_list ap;

	va_start (ap, fmt);
	res = pl_verror (pc -> ps, fmt, ap);
	va_end (ap);

	return res;
}

static enum s_status tr_array_begin (struct m_stack *stack,
				     struct obj *obj, int token,
				     void *closure)
{
	m_open (stack);

	return m_push (stack, obj, token);
}

static enum s_status tr_array_end (struct m_stack *stack,
				   struct obj *nobj, int token,
				   void *closure)
{
	struct obj_array array;
	struct obj obj, aobj;
	enum s_status res;
	int tt;

	obj_init_array (&array);
	while (m_pop (stack, &obj, &tt)) {
		if (tt == TOK_ARRAY_BEGIN)
			break;

		res = obj_xfer_array (&array, &obj);
		if (res != S_CSC_OK) {
			obj_array_destroy (&array);
			return res;
		}
	}
	
	if (tt != TOK_ARRAY_BEGIN) {
		obj_array_destroy (&array);
		return pc_error (closure, "Unmatched ']'");
	}

	obj_set_array (&aobj, &obj.ptr, &array);
	res = m_push (stack, &aobj, TOK_ARRAY);
	if (res != S_CSC_OK) {
		obj_destroy (&aobj);
		return res;
	}

	m_close (stack);

	return S_CSC_OK;
}

static enum s_status tr_dict_begin (struct m_stack *stack,
				    struct obj *obj, int token,
				    void *closure)
{
	m_open (stack);

	return m_push (stack, obj, token);
}

static enum s_status tr_null (struct m_stack *stack,
			      struct obj *obj, int token,
			      void *closure)
{
	obj -> type = OT_NULL;

	return m_push (stack, obj, token);
}

static enum s_status tr_true (struct m_stack *stack,
			      struct obj *obj, int token,
			      void *closure)
{
	obj -> type = OT_BOOLEAN;
	obj -> val.bool = 1;

	return m_push (stack, obj, token);
}

static enum s_status tr_false (struct m_stack *stack,
			       struct obj *obj, int token,
			       void *closure)
{
	obj -> type = OT_BOOLEAN;
	obj -> val.bool = 0;

	return m_push (stack, obj, token);
}


static enum s_status tr_dict_end (struct m_stack *stack,
				  struct obj *nobj, int token,
				  void *closure)
{
	struct obj_dictionary dictionary;
	struct obj_dict_node dnode;
	struct obj kobj, vobj, dobj;
	enum s_status res;
	int tt;

	obj_init_dictionary (&dictionary);
	while (m_pop (stack, &vobj, &tt)) {
		if (tt == TOK_DICT_BEGIN)
			break;

		if (!m_pop (stack, &kobj, &tt)) {
			obj_destroy (&vobj);
			obj_dict_destroy (&dictionary);
			return pc_error (closure, "Bad dictionary");
		}

		if (tt != TOK_NAME) {
			obj_destroy (&kobj);
			obj_destroy (&vobj);
			obj_dict_destroy (&dictionary);
			return pc_error (closure, "Bad dictionary");
		}

		obj_init_dict_node (&dnode, &kobj, &vobj);

		res = obj_xfer_dict (&dictionary, &dnode);
		if (res != S_CSC_OK) {
			obj_dict_destroy (&dictionary);
			return res;
		}
	}
	
	if (tt != TOK_DICT_BEGIN) {
		obj_dict_destroy (&dictionary);
		return pc_error (closure, "Unmatched '>>'");
	}

	obj_set_dictionary (&dobj, &vobj.ptr, &dictionary);
	res = m_push (stack, &dobj, TOK_DICTIONARY);
	if (res != S_CSC_OK) {
		obj_destroy (&dobj);
		return res;
	}

	m_close (stack);

	return S_CSC_OK;
}

static enum s_status tr_obj (struct m_stack *stack,
			     struct obj *nobj, int token,
			     void *closure)
{
	struct obj obj;
	pdf_oid_t oid;
	pdf_gen_t gen;
	int tt;

	if (!m_pop (stack, &obj, &tt) || tt != TOK_INTEGER)
		return pc_error (closure, "Bad reference");

	gen = obj.val.integer;

	if (!m_pop (stack, &obj, &tt) || tt != TOK_INTEGER)
		return pc_error (closure, "Bad reference");

	oid = obj.val.integer;
		
	obj_set_reference (&obj, &obj.ptr, oid, gen);

	m_open (stack);
	
	return m_push (stack, &obj, TOK_DEF_REFERENCE);
}

static enum s_status tr_endobj (struct m_stack *stack,
				struct obj *nobj, int token,
				void *closure)
{
	struct obj dobj, obj, ref;
	enum s_status res;
	int tt;

	if (!m_pop (stack, &obj, &tt))
		return pc_error (closure, "Misplaced endobj");

	if (!m_pop (stack, &ref, &tt)) {
		obj_destroy (&obj);
		return pc_error (closure, "Misplaced endobj");
	}

	if (tt != TOK_DEF_REFERENCE) {
		obj_destroy (&obj);
		obj_destroy (&ref);
		return pc_error (closure, "Misplaced endobj");
	}

	obj.ref = ref.ref;

	res = obj_set_definition (&dobj, &ref.ptr, &obj);
	if (res != S_CSC_OK) {
		obj_destroy (&obj);
		obj_destroy (&ref);
		return S_CSC_OK;
	}

	res = m_push (stack, &dobj, TOK_DEFINITION);
	if (res != S_CSC_OK) {
		obj_destroy (&dobj);
		return res;
	}

	m_close (stack);

	return S_CSC_OK;
}

static enum s_status tr_real (struct m_stack *stack,
			      struct obj *nobj, int token,
			      void *closure)
{
	struct obj *obj, sobj;
	int tt, neg;

	neg = 0;
	obj = m_top (stack, &tt);
	if (obj != NULL) {
		if (tt == TOK_PLUS || tt == TOK_MINUS) {
			m_pop (stack, &sobj, &tt);
			if (tt == TOK_MINUS)
				neg = 1;
		}	
	}

	if (neg)
		nobj -> val.real = -nobj -> val.real;

	return m_push (stack, nobj, token);
}

static enum s_status tr_integer (struct m_stack *stack,
				 struct obj *nobj, int token,
				 void *closure)
{
	struct obj *obj, sobj;
	int tt, neg;

	neg = 0;
	obj = m_top (stack, &tt);
	if (obj != NULL) {
		if (tt == TOK_PLUS || tt == TOK_MINUS) {
			m_pop (stack, &sobj, &tt);
			if (tt == TOK_MINUS)
				neg = 1;
		}
	}

	if (neg)
		nobj -> val.integer = -nobj -> val.integer;
	else if (nobj -> val.integer < 0)
		return obj_error (nobj, "Integer too large");

	return m_push (stack, nobj, token);
}

static enum s_status tr_ref (struct m_stack *stack,
			     struct obj *nobj, int token,
			     void *closure)
{
	struct obj obj;
	pdf_oid_t oid;
	pdf_gen_t gen;
	int tt;

	if (!m_pop (stack, &obj, &tt))
		return pc_error (closure, "Bad reference");

	if (tt != TOK_INTEGER) {
		obj_destroy (&obj);
		return pc_error (closure, "Bad reference");
	}

	gen = obj.val.integer;

	if (!m_pop (stack, &obj, &tt))
		return pc_error (closure, "Bad reference");

	if (tt != TOK_INTEGER) {
		obj_destroy (&obj);
		return pc_error (closure, "Bad reference");
	}

	oid = obj.val.integer;
		
	obj_set_reference (&obj, &obj.ptr, oid, gen);
	
	return m_push (stack, &obj, TOK_REFERENCE);
}

static enum s_status tr_stream (struct m_stack *stack,
				struct obj *obj, int token,
				void *closure)
{
	struct obj_region region;
	struct obj dobj, sobj;
	struct ps_closure *pc;
	const void *begin, *end;
	struct stream_info info;
	enum s_status res;
	int tt;

	if (!m_pop (stack, &dobj, &tt))
		return pc_error (closure, "Bad stream");

	if (tt != TOK_DICTIONARY) {
		obj_destroy (&dobj);
		return pc_error (closure, "No dict stream");
	}

	pc = closure;
	res = xref_fields_parse (pc -> parser -> rdb, NULL, NULL,
				 &dobj, &stream_field, &info);
	if (res != S_CSC_OK) {
		obj_destroy (&dobj);
		return res;
	}

	begin = pc -> ps -> ptr.cptr;
	if (!f_ptr_seek (&pc -> ps -> ptr, info.length))
		return pc_error (closure, "Bad length");
	end = pc -> ps -> ptr.cptr;
	obj_init_file_region (&region, begin, end, pc -> ps -> ptr.file);

	obj_set_stream (&sobj, &dobj.ptr, 
			&dobj.val.dictionary, &region);
	
	res = m_push (stack, &sobj, TOK_STREAM);
	if (res != S_CSC_OK) {
		obj_destroy (&sobj);
		return res;
	}

	res = pl_lex_predict (pc -> ps, TOK_ENDSTREAM, "endstream");
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status tr_endstream (struct m_stack *stack,
				   struct obj *obj, int token,
				   void *closure)
{
	return pc_error (closure, "Unexpected endstream keyword");
}

static enum s_status tr_xref_f (struct m_stack *stack,
				struct obj *nobj, int token,
				void *closure)
{
	struct obj obj;
	pdf_ofs_t next;
	pdf_gen_t gen;
	int tt;

	if (!m_pop (stack, &obj, &tt))
		return pc_error (closure, "Bad xref");

	if (tt != TOK_INTEGER) {
		obj_destroy (&obj);
		return pc_error (closure, "Bad xref");
	}

	gen = obj.val.integer;

	if (!m_pop (stack, &obj, &tt))
		return pc_error (closure, "Bad xref");

	if (tt != TOK_INTEGER) {
		obj_destroy (&obj);
		return pc_error (closure, "Bad xref");
	}

	next = obj.val.integer;
		
	obj_set_xref (&obj, &obj.ptr, next, gen, 0);
	
	return m_push (stack, &obj, TOK_XREF);
}

static enum s_status tr_keyword (struct m_stack *stack,
				 struct obj *obj, int token,
				 void *closure)
{
	struct ps_closure *pc;

	pc = closure;

	return ((m_trigger) obj -> val.keyword -> data) (stack, obj, token, 
							 pc -> pst_closure);
}

static enum s_status tr_xref_n (struct m_stack *stack,
				struct obj *nobj, int token,
				void *closure)
{
	struct obj obj;
	pdf_gen_t ofs;
	pdf_gen_t gen;
	int tt;

	if (!m_pop (stack, &obj, &tt))
		return pc_error (closure, "Bad xref");

	if (tt != TOK_INTEGER) {
		obj_destroy (&obj);
		return pc_error (closure, "Bad xref");
	}

	gen = obj.val.integer;

	if (!m_pop (stack, &obj, &tt))
		return pc_error (closure, "Bad xref");

	if (tt != TOK_INTEGER) {
		obj_destroy (&obj);
		return pc_error (closure, "Bad xref");
	}

	ofs = obj.val.integer;
		
	obj_set_xref (&obj, &obj.ptr, ofs, gen, 1);
	
	return m_push (stack, &obj, TOK_XREF);
}

static enum s_status tr_oth_keyword (struct m_stack *stack,
				     struct obj *obj, int token,
				     void *closure)
{
	struct ps_closure *pc;

	pc = closure;

	if (pc -> def_handler != NULL)
		return pc -> def_handler (stack, obj, token, 
					  pc -> pst_closure);
	else
		return pc_error (closure, "Unexpected keyword (%.*s)",
				 obj -> val.string.size, obj -> val.string.s);
}

static enum s_status tr_eof (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure)
{
	return pc_error (closure, "Unexpected end of file");
}

enum s_status ps_build_obj (struct ps_parser *parser, 
			    struct pl_state *ps, pdf_ofs_t limit,
			    struct obj *obj, enum obj_type ot)
{
	const uint8_t *sentinel;
	struct ps_closure pc;
	enum pl_token_type tt;
	enum s_status res;
	int ctxt, n;

	if (ot == OT_REFERENCE || 
	    ot == OT_XREF ||
	    ot == OT_DEFINITION)
		ctxt = 2;
	else
		ctxt = 0;

	pc.parser = parser;
	pc.ps = ps;
	pc.pst_closure = NULL;
	pc.def_handler = NULL;

	res = m_open_machine (parser -> state);
	if (res != S_CSC_OK)
		return res;

	while (ctxt--) {
		res = pl_lex (ps, &tt);
		if (res != S_CSC_OK) {
			m_close_machine (parser -> state);
			return res;
		}

		m_run_machine_cyc (parser -> state, &ps -> token, 
				   tt, &res, &pc);
		if (res != S_CSC_OK) {
			obj_destroy (&ps -> token);
			m_close_machine (parser -> state);
			return res;
		}
	}

	sentinel = limit > 0 ? ps -> ptr.cptr + limit : ps -> ptr.file -> end;

	do {
		res = pl_lex (ps, &tt);
		if (res != S_CSC_OK) {
			m_close_machine (parser -> state);
			return res;
		}

		if (ps -> sptr.cptr >= sentinel)
			break;

	} while (m_run_machine_cyc (parser -> state, &ps -> token, 
				    tt, &res, &pc));
	if (res != S_CSC_OK) {
		obj_destroy (&ps -> token);
		m_close_machine (parser -> state);
		return res;
	}

	n = 1;
	res = m_dump_stack (parser -> state, obj, &n);
	m_close_machine (parser -> state);
	if (res != S_CSC_OK)
		return res;

	if (n != 1)
		return pc_error (&pc, "Stack nonempty after creation");
	
	if (ot == OT_REAL && obj -> type == OT_INTEGER)
		obj_set_real (obj, &obj -> ptr, obj -> val.integer);
	else if (ot == OT_INTEGER && obj -> type == OT_REAL)
		obj_set_integer (obj, &obj -> ptr, obj -> val.real);

	if (!obj_cast (obj, ot))
		return pc_error (&pc, "Bad object type");

	return S_CSC_OK;
}

enum s_status ps_serialize_obj (struct f_file *file, struct obj *obj)
{
	enum s_status res;

	res = f_writeln (file, "%lu %lu obj", obj -> ref.oid, obj -> ref.gnum);
	if (res != S_CSC_OK)
		return res;

	res = pl_serialize (file, obj);
	if (res != S_CSC_OK)
		return res;

	res = f_newline (file);
	if (res != S_CSC_OK)
		return res;

	res = f_writeln (file, "endobj");
	if (res != S_CSC_OK)
		return res;

	f_newline (file);

	return S_CSC_OK;
}

enum s_status ps_eval (struct ps_parser *parser, m_trigger def_handler,
		       struct pl_state *ps, pdf_ofs_t limit,
		       void *pst_closure, ps_eval_feed feed, void *closure)
{
	const uint8_t *sentinel;
	struct ps_closure pc;
	enum pl_token_type tt;
	enum s_status res;
	int n;

	pc.parser = parser;
	pc.ps = ps;
	pc.pst_closure = pst_closure;
	pc.def_handler = def_handler;

	res = m_open_machine (parser -> state);
	if (res != S_CSC_OK)
		return res;

	sentinel = limit > 0 ? ps -> ptr.cptr + limit : ps -> ptr.file -> end;

	do {
		res = pl_lex (ps, &tt);
		if (res != S_CSC_OK) {
			m_close_machine (parser -> state);
			return res;
		}

		if (tt == TOK_EOF || ps -> sptr.cptr >= sentinel) {
			if (feed == NULL || !feed (ps, &limit, closure))
				break;
			sentinel = limit > 0 ?
				ps -> ptr.cptr + limit :
				ps -> ptr.file -> end;

		} else
			m_run_machine_cyc (parser -> state, &ps -> token, 
					   tt, &res, &pc);
	} while (res == S_CSC_OK);
	if (res != S_CSC_OK) {
		obj_destroy (&ps -> token);
		m_close_machine (parser -> state);
		return res;
	}

	n = 0;
	res = m_dump_stack (parser -> state, NULL, &n);
	m_close_machine (parser -> state);
	if (res != S_CSC_OK)
		return res;

	if (n)
		return pc_error (&pc, "Stack nonempty after execution");
	
	return S_CSC_OK;
}

enum s_status ps_create (struct xref_db *rdb, struct ps_parser **parser)
{
	struct ps_parser *p;

	p = malloc (sizeof (struct ps_parser));
	if (p == NULL)
		return S_CSC_OUT_OF_MEMORY;

	*p = template;
	p -> rdb = rdb;

	*parser = p;

	return S_CSC_OK;
}

void ps_destroy (struct ps_parser *parser)
{
	free (parser);
}

enum s_status ps_init (void)
{
	enum s_status res;

	res = m_create_machine (rules, TOK__LAST__, NULL, 0,
				TOK_EOF, &template.mach);
	if (res != S_CSC_OK)
		return res;

	res = m_start_machine (template.mach, &template.state);
	if (res != S_CSC_OK) {
		m_destroy_machine (template.mach);
		return res;
	}

	return S_CSC_OK;
}

void ps_cleanup (void)
{
	m_stop_machine (template.state);
	m_destroy_machine (template.mach);
}



