/*
 * PDFlow -- xref-t10.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>
#include <stdlib.h>

#include "fmt/pdf-lex.h"
#include "fmt/pdf-syn.h"
#include "xref_private.h"

struct xref10_subsection {

	struct xref_subsection s;

	struct f_ptr ptr;
};

static enum s_status load (struct xref_db *rdb,
			   struct xref_section **spp,
			   struct pl_state *ps);

static enum s_status scratch (struct xref_db *rdb,
			      struct xref_section **spp);

static enum s_status load_object (struct xref_db *rdb,
				  struct xref_section *sp,
				  struct xref_subsection *ssp,
				  struct obj_reference *ref,
				  struct obj *obj);

static void cleanup_subsection (struct xref_subsection *ssp);

static enum s_status read_entry (struct xref_db *rdb,
				 struct xref_section *sp,
				 struct xref_subsection *ssp,
				 pdf_integer_t oid,
				 struct obj_xref *xobj);

static enum s_status write_entry (struct xref_db *rdb,
				  struct xref_section *sp,
				  struct xref_subsection *ssp,
				  pdf_integer_t oid,
				  struct obj_xref *xobj);
	
static enum s_status insert_subsection (struct xref_db *rdb,
					struct xref_section *sp,
					struct xref_subsection **prev,
					pdf_integer_t lo,
					pdf_integer_t hi);

static enum s_status sync_section (struct xref_db *rdb,
				   struct xref_section *sp,
				   struct v_selection *sel);

struct xref_ifc xref10_ifc = { 
	load, scratch, load_object, NULL, cleanup_subsection,
	read_entry, write_entry, insert_subsection, sync_section
};


static enum s_status add_xref (struct xref_subsection **sspp, 
			       pdf_oid_t oid, pdf_integer_t entries, 
			       struct f_ptr *ptr)
{
	struct xref10_subsection *ssp;

	ssp = malloc (sizeof (struct xref10_subsection));
	if (ssp == NULL)
		return S_CSC_OUT_OF_MEMORY;

	ssp -> ptr = *ptr;

	f_ref (ssp -> ptr.file);

	xref_add (sspp, &ssp -> s, oid, entries);

	return S_CSC_OK;
}

static enum s_status load (struct xref_db *rdb,
			   struct xref_section **spp,
			   struct pl_state *ps)
{
	enum pl_token_type tt;
	struct xref_section *sp;
	struct obj trailer;
	pdf_oid_t oid;
	pdf_integer_t entries;
	enum s_status res;

	sp = *spp;

	res = pl_lex_predict (ps, TOK_XREF, "'xref'");
	if (res != S_CSC_OK)
		return res;

	while (1) {
		res = pl_lex (ps, &tt);
		if (res != S_CSC_OK)
			return res;
		if (tt != TOK_INTEGER)
			break;

		oid = ps -> token.val.integer;

		res = pl_lex_predict (ps, TOK_INTEGER, "integral type");
		if (res != S_CSC_OK)
			return res;
		entries = ps -> token.val.integer;

		pl_skip_ws (ps);

		res = add_xref (&sp -> head, oid, entries, &ps -> ptr);
		if (res != S_CSC_OK) {
			xref10_ifc.cleanup (sp);
			return res;
		}

		if (!f_ptr_seek (&ps -> ptr, 20 * entries))
			return pl_error (ps, "bad offset");
	}

	if (tt != TOK_TRAILER) {
		obj_destroy (&ps -> token);
		return pl_error (ps, "expecting 'trailer'");
	}

	res = ps_build_obj (rdb -> parser, ps, -1,
			    &trailer, OT_DICTIONARY);
	if (res != S_CSC_OK)
		return res;

	sp -> dict = trailer;

	return S_CSC_OK;
}

static enum s_status scratch (struct xref_db *rdb,
			      struct xref_section **spp)
{
	return S_CSC_OK;
}

static enum s_status load_xref (struct xref_db *rdb,
				struct xref_section *sp,
				struct xref_subsection *ssp,
				struct obj_reference *ref,
				struct obj_xref *xref)
{
	struct xref10_subsection *ssp10;
	enum s_status res;
	struct obj xobj;
	struct pl_state ps;
	unsigned lofs;

	ssp10 = (struct xref10_subsection *) ssp;

	lofs = (ref -> oid - ssp -> oid) * 20;

	pl_init (&ps, &ssp10 -> ptr);
	if (!f_ptr_seek (&ps.ptr, lofs))
		return f_error (&ssp10 -> ptr, "Bad offset");

	res = ps_build_obj (rdb -> parser, &ps, -1, &xobj, OT_XREF);
	if (res != S_CSC_OK)
		return res;

	*xref = xobj.val.xref;

	return S_CSC_OK;
}

static void store_xref (struct xref_db *rdb,
			struct xref_section *sp,
			struct xref_subsection *ssp,
			struct obj_reference *ref,
			pdf_integer_t first, 
			char flag)
{
	struct xref10_subsection *ssp10;
	unsigned lofs;

	ssp10 = (struct xref10_subsection *) ssp;

	lofs = (ref -> oid - ssp -> oid) * 20;

	snprintf ((char *) ssp10 -> ptr.file -> pdf + lofs, 20, 
		  "%010lu %05lu %c ", first, ref -> gnum, flag);
	((uint8_t *) ssp10 -> ptr.file -> pdf) [lofs + 19] = 10;
}

static enum s_status load_object (struct xref_db *rdb,
				  struct xref_section *sp,
				  struct xref_subsection *ssp,
				  struct obj_reference *ref,
				  struct obj *obj)
{
	struct obj dobj;
	struct obj_xref xref; 
	struct pl_state ps;
	struct f_ptr ptr;
	enum s_status res;

	res = load_xref (rdb, sp, ssp, ref, &xref);
	if (!xref.valid || xref.gnum != ref -> gnum) {
		*obj = obj_null;
		return S_CSC_OK;
	}
		
	if (!f_ptr_init (&ptr, rdb -> file, xref.u.ofs))
		return pl_error (&ps, "Bad offset");

	pl_init (&ps, &ptr);

	res = ps_build_obj (rdb -> parser, &ps, -1,
			    &dobj, OT_DEFINITION);
	if (res != S_CSC_OK)
		return res;

	obj_definition_unwrap (&dobj.val.definition, obj);

	return S_CSC_OK;
}

static void cleanup_subsection (struct xref_subsection *ssp)
{
	struct xref10_subsection *ssp10;

	ssp10 = (struct xref10_subsection *) ssp;

	if (ssp10 -> ptr.file != NULL)
		f_unref (ssp10 -> ptr.file);
}

static enum s_status read_entry (struct xref_db *rdb,
				 struct xref_section *sp,
				 struct xref_subsection *ssp,
				 pdf_integer_t oid,
				 struct obj_xref *xobj)
{
	struct obj_reference ref;

	ref.oid = oid;

	return load_xref (rdb, sp, ssp, &ref, xobj);
}

static enum s_status write_entry (struct xref_db *rdb,
				  struct xref_section *sp,
				  struct xref_subsection *ssp,
				  pdf_integer_t oid,
				  struct obj_xref *xobj)
{
	struct obj_reference ref;

	ref.oid = oid;
	ref.gnum = xobj -> gnum;

	if (xobj -> valid) 
		store_xref (rdb, sp, ssp, &ref, xobj -> u.ofs, 'n');
	else
		store_xref (rdb, sp, ssp, &ref, xobj -> u.next, 'f');

	return S_CSC_OK;
}
	
static enum s_status insert_subsection (struct xref_db *rdb,
					struct xref_section *sp,
					struct xref_subsection **prev,
					pdf_integer_t lo,
					pdf_integer_t hi)
{
	struct xref10_subsection *ssp10;
	struct f_file *file;
	enum s_status res;
	size_t size;
	void *data;

	size = 20 * (hi - lo);
	data = malloc (size);
	if (data == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (data, 0, size);

	res = f_reg_open (NULL, data, data + size, data, free, &file);
	if (res != S_CSC_OK) {
		free (data);
		return res;
	}

	ssp10 = malloc (sizeof (struct xref10_subsection));
	if (ssp10 == NULL) {
		f_unref (file);
		return S_CSC_OUT_OF_MEMORY;
	}

	f_ptr_init (&ssp10 -> ptr, file, 0);

	xref_add (prev, &ssp10 -> s, lo, hi - lo);

	return S_CSC_OK;
}

static enum s_status sync_section (struct xref_db *rdb,
				   struct xref_section *sp,
				   struct v_selection *sel)
{
	struct xref10_subsection *ssp10;
	enum s_status res;

	res = f_writeln (rdb -> scfile, "xref");
	if (res != S_CSC_OK)
		return res;

	for (ssp10 = (struct xref10_subsection *) sp -> head;
	     ssp10 != NULL;
	     ssp10 = (struct xref10_subsection *) ssp10 -> s.next) {
		res = f_writeln (rdb -> scfile, "%ld %ld",
				 ssp10 -> s.oid, ssp10 -> s.entries);
		if (res != S_CSC_OK)
			return res;

		res = f_dump (rdb -> scfile, ssp10 -> ptr.file);
		if (res != S_CSC_OK)
			return res;
	}

	res = f_writeln (rdb -> scfile, "trailer");
	if (res != S_CSC_OK)
		return res;

	return xref_serialize_trailer (rdb, sel);
}
