/*
 * PDFlow -- xref-t15.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <string.h>

#include "fmt/pdf-lex.h"
#include "fmt/pdf-syn.h"
#include "xref_private.h"

#define F1W 1
#define F2W 4
#define F3W 4

#define FxW (F1W + F2W + F3W)

struct stream_object {

	struct stream_object *next;

#define XREF_T15_STREAM_OBJECT
#include "xref-t15.tab.h"

	struct obj sobj;

	struct f_file *file;

	pdf_integer_t oid;

	struct stream_object_entry {
		pdf_oid_t oid;
		pdf_ofs_t ofs, end;
	} *entry;
};

struct xref15_subsection {

	struct xref_subsection s;

	void *scdata;

	const uint8_t *start;
};

struct xref15_section {

	struct xref_section s;

#define XREF_T15_SECTION
#include "xref-t15.tab.h"	

	size_t entrysz;

	struct stream_object *sobjs;
};

#define XREF_IMPLEMENTATION struct xref15_section
#define XREF_T15_SECTION
#include "xref-t15.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct stream_object
#define XREF_T15_STREAM_OBJECT
#include "xref-t15.tab.h"
#undef XREF_IMPLEMENTATION

static enum s_status load (struct xref_db *rdb,
			   struct xref_section **spp,
			   struct pl_state *ps);

static enum s_status scratch (struct xref_db *rdb,
			      struct xref_section **spp);

static enum s_status load_object (struct xref_db *rdb,
				  struct xref_section *sp,
				  struct xref_subsection *ssp,
				  struct obj_reference *ref,
				  struct obj *obj);

static void cleanup (struct xref_section *sp);

static void cleanup_subsection (struct xref_subsection *ssp);

static enum s_status read_entry (struct xref_db *rdb,
				 struct xref_section *sp,
				 struct xref_subsection *ssp,
				 pdf_integer_t oid,
				 struct obj_xref *xobj);

static enum s_status write_entry (struct xref_db *rdb,
				  struct xref_section *sp,
				  struct xref_subsection *ssp,
				  pdf_integer_t oid,
				  struct obj_xref *xobj);
	
static enum s_status insert_subsection (struct xref_db *rdb,
					struct xref_section *sp,
					struct xref_subsection **prev,
					pdf_integer_t lo,
					pdf_integer_t hi);

static enum s_status sync_section (struct xref_db *rdb,
				   struct xref_section *sp,
				   struct v_selection *sel);

struct xref_ifc xref15_ifc = { 
	load, scratch, load_object, cleanup, cleanup_subsection,
	read_entry, write_entry, insert_subsection, sync_section
};


static enum s_status add_xref (struct xref_subsection **sspp, 
			       pdf_oid_t oid, pdf_integer_t entries,
			       const uint8_t *start)
{
	struct xref15_subsection *ssp;

	ssp = malloc (sizeof (struct xref15_subsection));
	if (ssp == NULL)
		return S_CSC_OUT_OF_MEMORY;

	xref_add (sspp, &ssp -> s, oid, entries);

	ssp -> scdata = NULL;
	ssp -> start = start;

	return S_CSC_OK;
}

static enum s_status do_read_dict (struct xref_db *rdb,
				   struct xref15_section *sp,
				   struct obj *xref)
{
	pdf_integer_t oid, entries;
	enum s_status res;
	const uint8_t *sofs, *eofs;
	int i;

	res = xref_fields_parse (rdb, NULL, NULL, xref, 
				 &xref15_section_field, sp);
	if (res != S_CSC_OK)
		return res;

	sp -> entrysz = 
		sp -> w.data [0] + sp -> w.data [1] + sp -> w.data [2];
	if (!XREF_A_IS_NULL (&sp -> idx)) {
		sofs = sp -> data.decoded.begin;
		for (i = 0; i < XREF_A_SIZE (&sp -> idx); i += 2) {
			oid = sp -> idx.data [i];
			entries = sp -> idx.data [i + 1];
			eofs = sofs + entries * sp -> entrysz;
			if (eofs > sp -> data.decoded.end) {
				cleanup (&sp -> s);
				return obj_error (xref, "Entries overflow");
			}

			res = add_xref (&sp -> s.head, oid, entries, sofs);
			if (res != S_CSC_OK) {
				cleanup (&sp -> s);
				return res;
			}

			sofs = eofs;
		}
	} else {
		res = add_xref (&sp -> s.head, 0, sp -> size,
				sp -> data.decoded.begin);
		if (res != S_CSC_OK) {
			cleanup (&sp -> s);
			return res;
		}
	}

	obj_set_dictionary (&sp -> s.dict, &xref -> ptr, 
			    &xref -> val.stream.dictionary);
	obj_region_destroy (&xref -> val.stream.region);

	return S_CSC_OK;
}

static enum s_status load (struct xref_db *rdb,
			   struct xref_section **spp,
			   struct pl_state *ps)
{
	struct xref15_section *sp;
	struct obj xrefd, xref;
	enum s_status res;

	res = ps_build_obj (rdb -> parser, ps, -1,
			    &xrefd, OT_DEFINITION);
	if (res != S_CSC_OK)
		return res;

	sp = realloc (*spp, sizeof (struct xref15_section));
	if (sp == NULL) {
		obj_destroy (&xrefd);
		return S_CSC_OUT_OF_MEMORY;
	}

	*spp = &sp -> s;

	obj_definition_unwrap (&xrefd.val.definition, &xref);
	if (xref.type != OT_STREAM) {
		obj_destroy (&xref);
		return obj_error (&xrefd, "Not a stream definition");
	}

	sp -> sobjs = NULL;
	sp -> data.decoded.begin = NULL;

	res = do_read_dict (rdb, sp, &xref);
	if (res != S_CSC_OK) {
		obj_destroy (&xref);
		return res;
	}

	return S_CSC_OK;
}

static enum s_status scratch (struct xref_db *rdb,
			      struct xref_section **spp)
{
	struct xref15_section *sp;

	sp = malloc (sizeof (struct xref15_section));
	if (sp == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (sp, 0, sizeof (struct xref15_section));

	sp -> w.n = 3;
	sp -> w.data = calloc (3, sizeof (pdf_integer_t));
	if (sp -> w.data == NULL) {
		free (sp);
		return S_CSC_OUT_OF_MEMORY;
	}

	sp -> w.data [0] = F1W;
	sp -> w.data [1] = F2W;
	sp -> w.data [2] = F3W;
	sp -> entrysz = FxW;

	sp -> s = **spp;
	sp -> data.decoded.begin = NULL;
	sp -> sobjs = NULL;

	free (*spp);

	*spp = &sp -> s;

	return S_CSC_OK;
}

static enum s_status load_plain_obj (struct xref_db *rdb,
				     struct xref15_section *sp15,
				     struct xref15_subsection *ssp15,
				     struct obj_reference *ref,
				     struct obj *obj,	
				     pdf_integer_t ofs, 
				     pdf_integer_t gen)
{
	struct pl_state ps;
	struct f_ptr ptr;
	struct obj dobj;
	enum s_status res;

	if (gen != ref -> gnum) {
		*obj = obj_null;
		return S_CSC_OK;
	}
		
	if (!f_ptr_init (&ptr, rdb -> file, ofs))
		return pl_error (&ps, "Bad offset");

	pl_init (&ps, &ptr);
	res = ps_build_obj (rdb -> parser, &ps, -1,
			    &dobj, OT_DEFINITION);
	if (res != S_CSC_OK)
		return res;

	obj_definition_unwrap (&dobj.val.definition, obj);

	return S_CSC_OK;
}

static void destroy_stream_obj (struct stream_object *so)
{
	xref_fields_destroy (&xref15_stream_object_field, so, NULL);

	if (so -> file != NULL)
		f_unref (so -> file);

	if (so -> entry != NULL)
		free (so -> entry);

	obj_destroy (&so -> sobj);

	free (so);
}

static enum s_status init_stream_object (struct xref_db *rdb,
					 struct obj *sobj,
					 pdf_oid_t oid,
					 struct stream_object *so)
{
	struct pl_state ps;
	struct f_ptr fptr;
	enum s_status res;
	struct obj iobj;
	int i;

	res = xref_fields_parse (rdb, NULL, NULL, sobj,
				 &xref15_stream_object_field, so);
	if (res != S_CSC_OK)
		return res;

	so -> entry = calloc (so -> n, sizeof (struct stream_object_entry));
	if (so -> entry == NULL)
		return res;

	res = f_reg_open (sobj -> ptr.file,
			  so -> data.decoded.begin,
			  so -> data.decoded.end, 
			  NULL, NULL,
			  &so -> file);
	if (res != S_CSC_OK)
		return res;

	if (!f_ptr_init (&fptr, so -> file, 0))
		return obj_error (sobj, "Bad stream offset");

	pl_init (&ps, &fptr);

	for (i = 0; i < so -> n; i++) {
		res = ps_build_obj (rdb -> parser, &ps, -1,
				    &iobj, OT_INTEGER);
		if (res != S_CSC_OK)
			return res;

		so -> entry [i].oid = iobj.val.integer;

		res = ps_build_obj (rdb -> parser, &ps, -1,
				    &iobj, OT_INTEGER);
		if (res != S_CSC_OK)
			return res;

		so -> entry [i].ofs = so -> first + iobj.val.integer;
		if (i > 0)
			so -> entry [i - 1].end = so -> entry [i].ofs;
	}

	if (i > 0)
		so -> entry [i - 1].end = 
			so -> data.decoded.end - so -> data.decoded.begin;

	so -> sobj = *sobj;
	so -> oid = oid;
	so -> next = NULL;	

	return S_CSC_OK;
}
	
static enum s_status create_stream_obj (struct xref_db *rdb,
					struct obj *sobj,
					pdf_oid_t oid,
					struct stream_object **stobj)

{
	struct stream_object *so;
	enum s_status res;

	if (sobj -> type != OT_STREAM) {
		obj_destroy (sobj);
		return obj_error (sobj, "Not a stream object definition");
	}

	so = malloc (sizeof (struct stream_object));
	if (so == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (so, 0, sizeof (struct stream_object));

	res = init_stream_object (rdb, sobj, oid, so);
	if (res != S_CSC_OK) {
 		destroy_stream_obj (so);
		return res;
	}

	*stobj = so;

	return S_CSC_OK;
}

static enum s_status load_stream_obj (struct xref_db *rdb,
				      struct xref15_section *sp15,
				      struct xref15_subsection *ssp15,
				      struct obj_reference *ref,
				      struct obj *obj,	
				      pdf_oid_t oid, 
				      pdf_integer_t snum)
{
	struct obj_reference sref;
	struct stream_object **s, *stobj;
	struct stream_object_entry *e;
	struct pl_state ps;
	struct f_ptr fptr;
	struct obj sobj;
	enum s_status res;
	pdf_integer_t i;

	sref.oid = oid;
	sref.gnum = 0;

	for (s = &sp15 -> sobjs; *s != NULL; s = &(*s) -> next)
		if ((*s) -> oid >= oid)
			break;

	if ((*s) == NULL || (*s) -> oid != oid) {
		res = xref_get_object (rdb, &sref, &sobj);
		if (res != S_CSC_OK)
			return res;
		       
		stobj = NULL;	/* To make the compiler happy */
		res = create_stream_obj (rdb, &sobj, oid, &stobj);
		if (res != S_CSC_OK) {
			obj_destroy (&sobj);
			return res;
		}

		stobj -> next = *s;
		*s = stobj;
	}

	for (i = 0; i < (*s) -> n; i++)
		if ((*s) -> entry [i].oid == oid)
			break;

	if (snum >= (*s) -> n)
		return obj_error (&sp15 -> s.dict, 
				  "Stream does not contain entry %d", snum);

	e = &(*s) -> entry [snum];

	f_ptr_init (&fptr, (*s) -> file, e -> ofs);
	pl_init (&ps, &fptr);
	res = ps_build_obj (rdb -> parser, &ps, e -> end, obj, OT_ANY);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static const uint8_t *read_mb (const uint8_t *p, pdf_integer_t *i, 
			       int size, int *pres)
{
	if (!(*pres = size)) {
		*i = 0;			/* To make compiler happy */
		return p;
	}
	
	*i = 0;
	while (size--) {
		*i <<= 8;
		*i |= *(p++);
	}

	return p;
}

static uint8_t *write_mb (uint8_t *p, pdf_integer_t i, int size)
{
	int shift;

	shift = 8 * (size - 1);

	while (size--) {
		*(p++) = i >> shift;
		shift -= 8;
	}

	return p;
}

static enum s_status load_object (struct xref_db *rdb,
				  struct xref_section *sp,
				  struct xref_subsection *ssp,
				  struct obj_reference *ref,
				  struct obj *obj)
{
	struct xref15_subsection *ssp15;
	struct xref15_section *sp15;
	pdf_integer_t f1, f2, f3;
	int pf1, pf2, pf3;
	enum s_status res;
	const uint8_t *p;

	sp15 = (struct xref15_section *) sp;
	ssp15 = (struct xref15_subsection *) ssp;

	p = ssp15 -> start + ((ref -> oid - ssp -> oid) * sp15 -> entrysz);
	p = read_mb (p, &f1, sp15 -> w.data [0], &pf1);
	p = read_mb (p, &f2, sp15 -> w.data [1], &pf2);
	p = read_mb (p, &f3, sp15 -> w.data [2], &pf3);

	if (!pf1)
		f1 = 1;
	
	switch (f1) {
	case 0:
		*obj = obj_null;
		break;
		
	case 1:
		if (!pf2)
			return obj_error (&sp -> dict,
					  "Field 2 of entry type 1 "
					  "can't be zero");

		if (!pf3)
			f3 = 0;

		res = load_plain_obj (rdb, sp15, ssp15, ref, obj, f2, f3);
		if (res != S_CSC_OK)
			return res;
		
		break;
	case 2:
		if (!pf2)
			return obj_error (&sp -> dict,
					  "Field 2 entry type 2 "
					  "can't be zero");
		if (!pf3)
			f3 = 0;

		res = load_stream_obj (rdb, sp15, ssp15, ref, obj, f2, f3);
		if (res != S_CSC_OK)
			return res;

		break;

	default:
		return obj_error (&sp -> dict,
				  "Unknown entry type (%d)", f1);
	}

	return S_CSC_OK;
}

static void cleanup (struct xref_section *sp)
{
	struct xref15_section *sp15;
	struct stream_object *sobj;

	sp15 = (struct xref15_section *) sp;
	if (sp15 -> data.decoded.begin != NULL)
		obj_region_destroy (&sp15 -> data.decoded);

	while (sp15 -> sobjs != NULL) {
		sobj = sp15 -> sobjs;
		sp15 -> sobjs = sobj -> next;
		destroy_stream_obj (sobj);
	}

	xref_fields_destroy (&xref15_section_field, sp, NULL);

	xref_default_cleanup (sp);
}

static void cleanup_subsection (struct xref_subsection *ssp)
{
	struct xref15_subsection *ssp15;

	ssp15 = (struct xref15_subsection *) ssp;

	if (ssp15 -> scdata != NULL)
		free (ssp15 -> scdata);
}

static enum s_status read_entry (struct xref_db *rdb,
				 struct xref_section *sp,
				 struct xref_subsection *ssp,
				 pdf_integer_t oid,
				 struct obj_xref *xobj)
{
	struct xref15_subsection *ssp15;
	struct xref15_section *sp15;
	pdf_integer_t f1, f2, f3;
	const uint8_t *p;
	unsigned lofs;
	int pres;

	sp15 = (struct xref15_section *) sp;
	ssp15 = (struct xref15_subsection *) ssp;

	lofs = (oid - ssp -> oid) * sp15 -> entrysz;

	p = ssp15 -> start + lofs;

	p = read_mb (p, &f1, sp15 -> w.data [0], &pres);
	p = read_mb (p, &f2, sp15 -> w.data [1], &pres);
	p = read_mb (p, &f3, sp15 -> w.data [2], &pres);


	switch (f1) {
	case 0:
		xobj -> valid = 0;
		xobj -> u.next = f2;
		xobj -> gnum = f3;
		break;

	case 1:
		xobj -> valid = 1;
		xobj -> u.ofs = f2;
		xobj -> gnum = f3;
		break;

	case 2:
		/* Compressed objects are not supported yet */
		return S_CSC_INTERNAL_ERROR;
	}

	return S_CSC_OK;
}

static enum s_status write_entry (struct xref_db *rdb,
				  struct xref_section *sp,
				  struct xref_subsection *ssp,
				  pdf_integer_t oid,
				  struct obj_xref *xobj)
{
	struct xref15_subsection *ssp15;
	struct xref15_section *sp15;
	uint8_t *p;
	unsigned lofs;

	sp15 = (struct xref15_section *) sp;
	ssp15 = (struct xref15_subsection *) ssp;

	lofs = (oid - ssp -> oid) * sp15 -> entrysz;

	p = (uint8_t *) ssp15 -> start + lofs;

	if (!xobj -> valid) {
		p = write_mb (p, 0, sp15 -> w.data [0]);
		p = write_mb (p, xobj -> u.next, sp15 -> w.data [1]);
		p = write_mb (p, xobj -> gnum, sp15 -> w.data [2]);
	} else {
		p = write_mb (p, 1, sp15 -> w.data [0]);
		p = write_mb (p, xobj -> u.ofs, sp15 -> w.data [1]);
		p = write_mb (p, xobj -> gnum, sp15 -> w.data [2]);
	}

	return S_CSC_OK;
}
	
static enum s_status insert_subsection (struct xref_db *rdb,
					struct xref_section *sp,
					struct xref_subsection **prev,
					pdf_integer_t lo,
					pdf_integer_t hi)
{
	struct xref15_subsection *ssp15;
	size_t size;
	void *data;

	size = FxW * (hi - lo);
	data = malloc (size);
	if (data == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (data, 0, size);

	ssp15 = malloc (sizeof (struct xref15_subsection));
	if (ssp15 == NULL) {
		free (data);
		return S_CSC_OUT_OF_MEMORY;
	}

	ssp15 -> scdata = data;
	ssp15 -> start = data;

	xref_add (prev, &ssp15 -> s, lo, hi - lo);

	return S_CSC_OK;
}

static void update_line (const uint8_t *p, pdf_integer_t line [3], 
			 struct xref15_section *sp15)
{
	pdf_integer_t f;
	int i, pres;

	for (i = 0; i < 3; i++) {
		p = read_mb (p, &f, sp15 -> w.data [i], &pres);
		line [i] |= f;
	}
}

static void copy_line (const uint8_t *src, uint8_t *dest, 
		       pdf_integer_t w [3], struct xref15_section *sp15)
{
	int i, j;

	for (i = 0; i < 3; i++) {
		src += sp15 -> w.data [i] - w [i];
		for (j = 0; j < w [i]; j++)
			*(dest++) = *(src++);
	}
}

static void get_widths (pdf_integer_t line [3], pdf_integer_t w [3],
			size_t *size)
{
	int i, mask, width;

	*size = 0;
	for (i = 0; i < 3; i++) {
		w [i] = 1;
		mask = 0xff;
		for (width = 1; width <= sizeof (pdf_integer_t); width++) {
			if (line [i] & mask)
				w [i] = width;
			mask <<= 8;
		}
		*size += w [i];
	}
}

static enum s_status do_make_region (struct xref_section *sp, 
				     struct xref_int_array *w, 
				     struct xref_int_array *idx,
				     struct obj_region *reg)
{
	struct xref15_subsection *ssp15;
	struct xref15_section *sp15;
	struct xref_subsection *ssp;
	pdf_integer_t line [3];
	enum s_status res;
	uint8_t *data;
	size_t size, entries, destsz;
	int i, j, k;

	sp15 = (struct xref15_section *) sp;

	/* Pass I: get index sizes */
	line [0] = line [1] = line [2] = 0;

	res = S_CSC_OK;
	entries = k = 0;
	for (ssp = sp -> head; ssp != NULL; ssp = ssp -> next) {
		ssp15 = (struct xref15_subsection *) ssp;
		idx -> data [k++] = ssp -> oid;
		idx -> data [k++] = ssp -> entries;

		j = 0;
		for (i = 0; i < ssp -> entries; i++) {
			update_line (ssp15 -> scdata + j, line, sp15);
			j += sp15 -> entrysz;
		}

		entries += ssp -> entries;
	}

	/* Pass II fill region */
	get_widths (line, w -> data, &destsz);
	size = entries * destsz;

	data = malloc (size);
	if (data == NULL)
		return S_CSC_OUT_OF_MEMORY;

	obj_init_malloc_region (reg, data, data + size);

	size = 0;
	for (ssp = sp -> head; ssp != NULL; ssp = ssp -> next) {
		ssp15 = (struct xref15_subsection *) ssp;

		j = 0;
		for (i = 0; i < ssp -> entries; i++) {
			copy_line (ssp15 -> scdata + j, 
				   data + size, w -> data, sp15);
			j += sp15 -> entrysz;
			size += destsz;
		}
	}

	return S_CSC_OK;
}

static enum s_status make_region (struct xref_section *sp, 
				  struct xref_int_array *wp, 
				  struct xref_int_array *idxp,
				  struct obj_region *reg)
{
	struct xref_subsection *ssp;
	struct xref_int_array w, idx;
	enum s_status res;
	int secs;

	secs = 0;
	for (ssp = sp -> head; ssp != NULL; ssp = ssp -> next)
		secs++;

	w.n = 3;
	w.data = calloc (w.n, sizeof (pdf_integer_t));
	
	idx.n = secs * 2;
	idx.data = calloc (idx.n, sizeof (pdf_integer_t));

	if (w.data == NULL || idx.data == NULL) {
		if (w.data != NULL)
			free (w.data);
		if (idx.data != NULL)
			free (idx.data);
		return S_CSC_OUT_OF_MEMORY;
	}

	res = do_make_region (sp, &w, &idx, reg);
	if (res != S_CSC_OK) {
		free (w.data);
		free (idx.data);
		return res;
	}

	if (wp -> data != NULL)
		free (wp -> data);
	if (idxp -> data != NULL)
		free (idxp -> data);

	*wp = w;
	*idxp = idx;
	

	return S_CSC_OK;
}

static enum s_status sync_section (struct xref_db *rdb,
				   struct xref_section *sp,
				   struct v_selection *sel)
{
	struct xref15_section *sp15;
	struct obj_dictionary dict;
	struct obj_region reg;
	struct obj scobj, pobj;
	enum s_status res;
	enum xref_changed changed;
	pdf_ofs_t ofs;

	sp15 = (struct xref15_section *) sp;

	obj_set_null (&pobj, &f_null);

	F_TELL (rdb -> scfile, &ofs);
	res = xref_place_object (rdb, &pobj, ofs);
	if (res != S_CSC_OK)
		return res;

	if (sp -> dict.type != OT_NULL) {
		res = obj_dict_dup (&sp -> dict.val.dictionary, &dict);
		if (res != S_CSC_OK) {
 			xref_remove_object (rdb, &pobj);
			return res;
		}	
	} else
		dict = obj_null_dictionary.val.dictionary;

	sp15 -> prev = sp -> next != NULL ? sp -> next -> trofs : 0;

	res = make_region (sp, &sp15 -> w, &sp15 -> idx, &reg);
	if (res != S_CSC_OK) {
		obj_dict_destroy (&dict);
		xref_remove_object (rdb, &pobj);
		return res;
	}

	obj_region_destroy (&sp15 -> data.decoded);
	sp15 -> data.encoded = reg;

	scobj.ref = pobj.ref;
	sp15 -> size = rdb -> entries;

	changed = XREF_C_UNKNOWN;
	res = xref_fields_ser (rdb, NULL, sel, &scobj, 
			       &changed, &xref15_section_field, sp);
	if (res == S_CSC_OK) {
		obj_dict_merge (&scobj.val.stream.dictionary,
				&rdb -> xrefs -> dict.val.dictionary);
		res = ps_serialize_obj (rdb -> scfile, &scobj);
	}

	obj_destroy (&scobj);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}
