/*
 * PDFlow -- xref.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <assert.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <version.h>

#include "fmt/pdf-lex.h"
#include "fmt/pdf-syn.h"
#include "xref_private.h"
#include "fmt/file.h"
#include "version.h"
#include "filters.h"

/* This is somehow circular, but PDF is like that ... */
#include "types/filespec.h"

#define SUBSECTION_PART 40

static pdf_integer_t id;

extern struct xref_pstable fspec_pstable;

struct xref_trailer_info {
#define XREF_TRAILER_INFO
#include "fmt/xref.tab.h"	
};

struct xref_back_ptr {
#define XREF_BACK_PTR
#include "fmt/xref.tab.h"	
};

#define XREF_IMPLEMENTATION struct xref_stream
#define XREF_STREAM
#include "fmt/xref.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct xref_trailer_info
#define XREF_TRAILER_INFO
#include "fmt/xref.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct xref_back_ptr
#define XREF_BACK_PTR
#include "fmt/xref.tab.h"
#undef XREF_IMPLEMENTATION

static struct xref_ifc *ifc [] = {
	&xref10_ifc, 
	&xref15_ifc,
	NULL
};

static struct v_option options [] = {
      	/*             		            Vers  sp  bs  sz  */
	V_OPTION_BSPEED ("Default", 	    1, 0,  5,  5,  5),
	V_OPTION_BSPEED ("XrefStream",      1, 5, 10,  4, 10),
	V_END
};

static struct v_optable optable = {
	"xref", options, ifc, sizeof (struct xref_interface *)
};

static enum s_status get_scratch_ssp (struct xref_db *rdb, 
				      pdf_integer_t oid,
				      struct xref_subsection **sspp);

enum s_status xref_cache_add (struct xref_db *rdb, const char *id,
			      xref_cache_unref unref, 
			      void *closure)
{
	struct xref_cache *cache;

	cache = malloc (sizeof (struct xref_cache));
	if (cache == NULL)
		return S_CSC_OUT_OF_MEMORY;

	ref_init_owner (&cache -> owner, id);
	cache -> list = NULL;
	cache -> id = id;
	cache -> unref = unref;
	cache -> closure = closure;

	cache -> next = rdb -> cache;
	rdb -> cache = cache;

	return S_CSC_OK;
}

static void cache_destroy (struct xref_cache *cache)
{
	struct xref_cache_node *n;

	while (cache -> list != NULL) {
		n = cache -> list;
		cache -> unref (n -> ptr, &cache -> owner, cache -> closure);
		cache -> list = n -> next;
		free (n);
	}

	free (cache);
}

static void cache_destroy_all (struct xref_db *rdb)
{
	struct xref_cache *cache;

	while (rdb -> cache != NULL) {
		cache = rdb -> cache;
		rdb -> cache = cache -> next;
		cache_destroy (cache);
	}
}

struct xref_cache *xref_cache_get (struct xref_db *rdb, const char *id)
{
	struct xref_cache *cache;

	for (cache = rdb -> cache; cache != NULL; cache = cache -> next)
		if (cache -> id == id)
			return cache;

	return NULL;
}

void xref_cache_remove (struct xref_db *rdb, const char *id)
{
	struct xref_cache *cache, **cp;

	cp = &rdb -> cache; 
	while (*cp != NULL)
		if ((*cp) -> id == id) {
			cache = *cp;	
			*cp = cache -> next;
			cache -> next = NULL;
			cache_destroy (cache);
		} else
			cp = &(*cp) -> next;
}

void *xref_cache_find (struct xref_cache *cache, const struct obj *obj)
{
	struct xref_cache_node *n;

	if (cache == NULL || !obj_exists (obj))
		return NULL;

	for (n = cache -> list; n != NULL; n = n -> next)
		if (obj_same (obj, &n -> obj))
			return n -> ptr;

	return NULL;
}

struct ref_owner *xref_cache_put (struct xref_cache *cache, 
				  const struct obj *obj, void *ptr)
{
	struct xref_cache_node *n;

	if (cache == NULL)
		return NULL;

	n = malloc (sizeof (struct xref_cache_node));
	if (n == NULL)
		return NULL;

	n -> next = cache -> list;
	obj_copy_reference (&n -> obj, obj);
	n -> ptr = ptr;

	cache -> list = n;

	return &cache -> owner;
}

enum s_status xref_create_db (struct xref_db **prdb, struct ref_owner *p)
{
	struct xref_db *rdb;
	enum s_status res;

	rdb = malloc (sizeof (struct xref_db));
	if (rdb == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (rdb, 0, sizeof (struct xref_db));

	rdb -> xrefs = NULL;
	rdb -> id = id++;
	ref_init (&rdb -> refs, "xref", p);

	res = ps_create (rdb, &rdb -> parser);
	if (res != S_CSC_OK) {
		free (rdb);
		return res;
	}

	*prdb  = rdb;

	return S_CSC_OK;
}

static void cleanup_subsection (struct xref_section *sp,
				struct xref_subsection **sspp)
{
	struct xref_subsection *ssp;

	ssp = *sspp;
	*sspp = ssp -> next;

	if (sp -> ifc -> cleanup_subsection != NULL)
		sp -> ifc -> cleanup_subsection (ssp);

	free (ssp);
}

static void cleanup_section (struct xref_section *sp)
{
	while (sp -> head != NULL)
		cleanup_subsection (sp, &sp -> head);

	if (sp -> secin && sp -> ifc -> cleanup != NULL)
		sp -> ifc -> cleanup (sp);

	obj_destroy (&sp -> dict);

	free (sp);
}

static void cleanup_xrefs (struct xref_section **xrefs)
{
	struct xref_section *sp;

	while (*xrefs != NULL) {
		sp = *xrefs;
		*xrefs = sp -> next;

		cleanup_section (sp);
	}
}

void xref_add (struct xref_subsection **head, 
	       struct xref_subsection *ssp,
	       pdf_oid_t oid, pdf_integer_t entries)
{
	ssp -> oid = oid;
	ssp -> entries = entries;

	while (*head != NULL && (*head) -> oid < oid)
		head = &(*head) -> next;
	
	ssp -> next = *head;
	*head = ssp;
}

void xref_default_cleanup (struct xref_section *sp)
{
	struct xref_subsection *ssp;

	while (sp -> head != NULL) {
		ssp = sp -> head;
		sp -> head = ssp -> next;
		free (ssp);
	}

	obj_destroy (&sp -> dict);
}

static struct xref_subsection *fetch_subsection (struct xref_db *rdb,
						 pdf_integer_t oid,
						 struct xref_section **spp)
{
	struct xref_subsection *ssp;
	struct xref_section *sp;

	if (oid >= rdb -> entries)
		return NULL;

	for (sp = rdb -> xrefs; sp != NULL; sp = sp -> next) {
		for (ssp = sp -> head; ssp != NULL; ssp = ssp -> next) {
			if (oid >= ssp -> oid &&
			    oid < ssp -> oid + ssp -> entries) {
				if (spp != NULL)
					*spp = sp;
				return ssp;
			}
		}
	}

	return NULL;
}

static enum s_status read_entry (struct xref_db *rdb,
				 pdf_integer_t oid,
				 struct obj_xref *ref)
{
	struct xref_subsection *ssp;
	struct xref_section *sp;

	ssp = fetch_subsection (rdb, oid, &sp);
	if (ssp == NULL) {
		obj_init_xref (ref, 0, 0xffff, 0);
		return S_CSC_OK;
	}

	return sp -> ifc -> read_entry (rdb, sp, ssp, oid, ref);
}

static enum s_status write_entry (struct xref_db *rdb,
				  pdf_integer_t oid,
				  struct obj_xref *ref)
{
	struct xref_subsection *ssp;
	enum s_status res;

	res = get_scratch_ssp (rdb, oid, &ssp);
	if (res != S_CSC_OK)
		return res;

	return rdb -> xrefs -> ifc -> write_entry (rdb, rdb -> xrefs, 
						   ssp, oid, ref);
}

static enum s_status get_first_free (struct xref_db *rdb, 
				     pdf_integer_t *oid)
{
	struct obj_xref xref;
	enum s_status res;

	res = rdb -> xrefs -> ifc -> 
		read_entry (rdb, rdb -> xrefs, 
			    rdb -> xrefs -> head, 0, &xref);	

	if (res != S_CSC_OK)
		return res;

	*oid = xref.u.next;

	return S_CSC_OK;
}

static enum s_status set_first_free (struct xref_db *rdb, 
				     pdf_integer_t oid)
{
	struct obj_xref ref;

	obj_init_xref (&ref, oid, 0xffff, 0);
	return rdb -> xrefs -> ifc -> write_entry (rdb, rdb -> xrefs, 
						   rdb -> xrefs -> head, 0,
						   &ref);	
}

static enum s_status init_subsection (struct xref_db *rdb,
				      struct xref_section *sp,
				      struct xref_subsection *ssp,
				      pdf_integer_t lo,
				      pdf_integer_t hi)
{
	pdf_integer_t oid, foid;
	struct obj_xref ref;
	struct xref_ifc *ifc;
	enum s_status res;

	ifc = sp -> ifc;

	for (oid = lo; oid < hi; oid++) {
		if (oid >= rdb -> entries)
			break;

		res = read_entry (rdb, oid, &ref);
		if (res != S_CSC_OK)
			return res;

		res = ifc -> write_entry (rdb, sp, ssp, oid, &ref);
		if (res != S_CSC_OK)
			return res;
	}

	if (oid == hi)
		return S_CSC_OK;

	res = get_first_free (rdb, &foid);
	if (res != S_CSC_OK)
		return res;

	while (oid < hi) {
		obj_init_xref (&ref, foid, 0, 0);
		res = ifc -> write_entry (rdb, sp, ssp, oid, &ref);
		if (res != S_CSC_OK)
			return res;

		foid = oid;
		oid++;
	}
	
	res = set_first_free (rdb, foid);
	if (res != S_CSC_OK)
		return res;

	rdb -> entries = hi;

	return S_CSC_OK;
}
				      

static enum s_status get_scratch_ssp (struct xref_db *rdb, 
				      pdf_integer_t oid,
				      struct xref_subsection **sspp)
{
	struct xref_subsection **ssp;
	struct xref_section *sp;
	pdf_integer_t poid, lo, hi;
	enum s_status res;

	poid = 0;
	sp = rdb -> xrefs;
	for (ssp = &sp -> head; *ssp != NULL; ssp = &(*ssp) -> next) {
		if ((*ssp) -> oid > oid)
			break;

		poid = (*ssp) -> oid + (*ssp) -> entries;
		if (oid >= (*ssp) -> oid && oid < poid) {
			if (sspp != NULL)
				*sspp = *ssp;

			return S_CSC_OK;
		}
	}

	lo = (oid / SUBSECTION_PART) * SUBSECTION_PART;
	hi = lo + SUBSECTION_PART;

	if (lo < poid)
		lo = poid;
	if (*ssp != NULL && hi >= (*ssp) -> oid)
		hi = (*ssp) -> oid - 1;

	res = sp -> ifc -> insert_subsection (rdb, sp, ssp, lo, hi);
	if (res != S_CSC_OK)
		return res;

	res = init_subsection (rdb, sp, *ssp, lo, hi);
	if (res != S_CSC_OK) {
		cleanup_subsection (sp, ssp);
		return res;
	}

	if (hi > rdb -> entries)
		rdb -> entries = hi;

	if (sspp != NULL)
		*sspp = *ssp;

	return S_CSC_OK;
}

static enum s_status update_entry (struct xref_db *rdb, struct obj *obj,
				   pdf_ofs_t ofs)
{
	struct obj_xref ref;

	obj_init_xref (&ref, ofs, obj -> ref.gnum, 1);

	return write_entry (rdb, obj -> ref.oid, &ref);
}

static enum s_status allocate_entry (struct xref_db *rdb, struct obj *obj,
				     pdf_ofs_t ofs)
{
	struct obj_xref ref, wref;
	enum s_status res;
	pdf_integer_t foid;

	res = get_first_free (rdb, &foid);
	if (res != S_CSC_OK)	
		return res;

	if (!foid) {
		res = get_scratch_ssp (rdb, rdb -> entries, NULL);
		if (res != S_CSC_OK)
			return res;

		res = get_first_free (rdb, &foid);
		if (res != S_CSC_OK)
			return res;

		assert (foid);
	}

	res = read_entry (rdb, foid, &ref);
	if (res != S_CSC_OK)
		return res;

	obj -> ref.oid = foid;
	obj -> ref.gnum = ref.gnum;

	obj_init_xref (&wref, ofs, ref.gnum, 1);

	res = write_entry (rdb, foid, &wref);
	if (res != S_CSC_OK)
		return res;

	return set_first_free (rdb, ref.u.next);
}

enum s_status xref_place_object (struct xref_db *rdb, struct obj *obj,
				 pdf_ofs_t ofs)
{
	enum s_status res;

	if (obj_exists (obj))
		res = update_entry (rdb, obj, ofs);
	else
		res = allocate_entry (rdb, obj, ofs);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

enum s_status xref_register_object (struct xref_db *rdb, struct obj *obj)
{
	return xref_place_object (rdb, obj, 0);
}

static enum s_status remove_entry (struct xref_db *rdb, struct obj *obj)
{
	struct obj_xref wref;
	enum s_status res;
	pdf_integer_t foid, gnum;

	gnum = obj -> ref.gnum + 1;
	if (gnum >= 0xffff) {
		obj_init_xref (&wref, 0, 0xffff, 0);
		return write_entry (rdb, foid, &wref);
	}

	res = get_first_free (rdb, &foid);
	if (res != S_CSC_OK)	
		return res;

	obj_init_xref (&wref, foid, gnum, 0);
	res = write_entry (rdb, obj -> ref.oid, &wref);
	if (res != S_CSC_OK)
		return res;

	return set_first_free (rdb, obj -> ref.oid);
}

enum s_status xref_remove_object (struct xref_db *rdb, struct obj *obj)
{
	enum s_status res;

	if (obj_exists (obj))
		res = remove_entry (rdb, obj);
	else
		res = S_CSC_OK;

	return res;
}

enum s_status xref_serialize_trailer (struct xref_db *rdb, 
				      struct v_selection *sel)
{
	struct obj obj;
	enum s_status res;

	obj_set_integer (&obj, &f_null, rdb -> entries);
	res = obj_add_dict (&rdb -> xrefs -> dict.val.dictionary, 
			    "Size", &obj);
	if (res != S_CSC_OK)
		return res;

	res = pl_serialize (rdb -> scfile, &rdb -> xrefs -> dict);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status init_section (struct xref_section **spp,
				   struct xref_ifc *ifc)
{
	struct xref_section *sp;

	sp = malloc (sizeof (struct xref_section));
	if (sp == NULL)
		return S_CSC_OUT_OF_MEMORY;

        sp -> next = NULL;
	sp -> head = NULL;
	sp -> ifc = ifc;
	sp -> dict = obj_null;
	sp -> secin = 0;

	*spp = sp;

	return S_CSC_OK;
}

static enum s_status load_xrefs (struct xref_db *rdb,
				 struct pl_state *ops,
				 struct xref_section **xrefs,
				 pdf_ofs_t ofs)
{
	struct xref_ifc *ifc;
	struct pl_state ps;
	struct f_ptr ptr;
	enum pl_token_type tt;
	enum s_status res;
	struct xref_section *sp;

	if (!f_ptr_init (&ptr, rdb -> file, ofs))
		return pl_error (ops, "Bad xref pointer");

	pl_init (&ps, &ptr);

	res = pl_lex (&ps, &tt);
	if (res != S_CSC_OK)
		return res;

	if (tt == TOK_XREF)
		ifc = &xref10_ifc;
	else if (tt == TOK_INTEGER)
		ifc = &xref15_ifc;
	else {
		obj_destroy (&ps.token);
		return pl_error (&ps, "Bad xref table");
	}
	
	pl_init (&ps, &ptr);

	res = init_section (&sp, ifc);
	if (res != S_CSC_OK)
		return res;
	
	res = ifc -> load (rdb, &sp, &ps);
	if (res != S_CSC_OK) {
		cleanup_xrefs (&sp);
		return res;
	}

	sp -> secin = 1;

	*xrefs = sp;

	return S_CSC_OK;
}

static enum s_status parse_header (struct f_file *f)
{
	struct obj_bstr header;
	const uint8_t *p;
	int version [2];

	for (p = f -> pdf; p < f -> end; p++)
		if (pl_is_blank (*p))
			break;
	if (p >= f -> end ||
	    p - f -> pdf < 1 || f -> pdf [0] != '%')
		return s_make_indirect (S_CSC_PARSE_ERROR, 
					"Can't find header");

	header.size = (uint8_t *) p - f -> pdf - 1;
	header.s = (char *) f -> pdf + 1;

	return v_validate (&header, version);
}

static enum s_status get_trailer_info (struct xref_db *rdb,
				       struct obj *tdict)
{
	struct xref_trailer_info tinfo;
	enum s_status res;

	res = xref_fields_parse (rdb, NULL, NULL, tdict, 
				 &xref_trailer_info_field, &tinfo);
	if (res != S_CSC_OK)
		return res;

	rdb -> entries = tinfo.size;

	return S_CSC_OK;
}

enum s_status xref_init_db (struct xref_db *rdb, struct f_file *f)
{
	struct xref_section **xp;
	struct xref_back_ptr bptr;
	struct xref_db backup;
	struct pl_state ps;
	struct f_ptr ptr;
	enum s_status res;
	pdf_ofs_t ofs;
	int i;

	res = parse_header (f);
	if (res != S_CSC_OK)
		return res;

	f_ptr_eof (&ptr, f);
	pl_init (&ps, &ptr);
	for (i = 0; i < 5; i++) {
		res = pl_backline (&ps, 1);
		if (res != S_CSC_OK)
			return pl_error (&ps, "Can't find trailer");
		if (f_ptr_avail (&ps.ptr) >= 5 &&
		    !memcmp (ps.ptr.cptr, "%%EOF", 5))
			break;
	}

	res = pl_backline (&ps, 2);
	if (res != S_CSC_OK)
		return pl_error (&ps, "Can't find trailer");
		
	res = pl_lex_predict (&ps, TOK_STARTXREF, "'startxref'");
	if (res != S_CSC_OK)
		return res;

	res = pl_lex_predict (&ps, TOK_INTEGER, "integral type");
	if (res != S_CSC_OK)
		return res;

	backup = *rdb;

	rdb -> file = f;
	rdb -> xrefs = NULL;

	xp = &rdb -> xrefs;
	ofs = ps.token.val.integer;
	do {
		res = load_xrefs (rdb, &ps, xp, ofs);
		if (res != S_CSC_OK) {
			cleanup_xrefs (&rdb -> xrefs);
			backup.refs = rdb -> refs;
			*rdb = backup;
			return res;
		}

		(*xp) -> trofs = ofs;

		res = xref_fields_parse (rdb, NULL, NULL, &(*xp) -> dict, 
					 &xref_back_ptr_field, &bptr);
		if (res == S_CSC_OK) {
			ofs = bptr.prev;
			if (ofs != -1 && ofs >= (f -> end - f -> pdf))
				res = obj_error (&(*xp) -> dict, 
						 "Bad Prev pointer");
		}
		if (res != S_CSC_OK) {
			cleanup_xrefs (&rdb -> xrefs);
			*rdb = backup;
			return res;
		}

		xp = &(*xp) -> next;

		if (bptr.xref_stm < 0)
			continue;

		/* Hybrid file, load xref stream too */
		res = load_xrefs (rdb, &ps, xp, bptr.xref_stm);
		if (res == S_CSC_OK)
			xp = &(*xp) -> next;

	} while (bptr.prev != -1);

	res = get_trailer_info (rdb, &rdb -> xrefs -> dict);
	if (res != S_CSC_OK) {
		cleanup_xrefs (&rdb -> xrefs);
		*rdb = backup;
		return res;
	}

	if (backup.file != NULL)
		f_unref (rdb -> file);
	cleanup_xrefs (&backup.xrefs);

	f_ref (f);

	return S_CSC_OK;
}

enum s_status init_tdict (struct xref_db *rdb, struct v_selection *sel,
			  struct xref_section *sp)
{
	struct obj_dictionary dict;
	enum s_status res;
	struct obj obj;

	obj_init_dictionary (&dict);

	obj_set_integer (&obj, &f_null, 0);
	res = obj_add_dict (&dict, "Size", &obj);
	if (res != S_CSC_OK) {
		obj_dict_destroy (&dict);
		return res;
	}

	if (sp -> next != NULL) {
		obj_set_integer (&obj, &f_null, sp -> next -> trofs);
		res = obj_add_dict (&dict, "Prev", &obj);
		if (res != S_CSC_OK) {
			obj_dict_destroy (&dict);
			return res;
		}
	}

	obj_set_dictionary (&sp -> dict, &f_null, &dict);

	return S_CSC_OK;
}

static enum s_status init_page0 (struct xref_db *rdb, struct xref_section *sp)
{
	struct xref_subsection *ssp;
	struct xref_ifc *ifc;
	struct obj_xref ref;
	pdf_integer_t oid;
	enum s_status res;

	ssp = fetch_subsection (rdb, 0, NULL);
	if (ssp != NULL)
		return get_scratch_ssp (rdb, 0, NULL);

	ifc = sp -> ifc;
	rdb -> entries = SUBSECTION_PART;
	res = ifc -> insert_subsection (rdb, sp, &ssp,  0, rdb -> entries);
	if (res != S_CSC_OK)
		return res;

	obj_init_xref (&ref, rdb -> entries - 1, 0xffff, 0);
	res = ifc -> write_entry (rdb, sp, ssp, 0, &ref);
	if (res != S_CSC_OK) {
		cleanup_subsection (sp, &ssp);
		return res;
	}
		
	for (oid = 1; oid < rdb -> entries; oid++) {
		obj_init_xref (&ref, oid - 1, 0, 0);
		res = ifc -> write_entry (rdb, sp, ssp, oid, &ref);
		if (res != S_CSC_OK) {
			cleanup_subsection (sp, &ssp);
			return res;
		}
	}

	sp -> head = ssp;

	return S_CSC_OK;
}

static enum s_status init_scratch_section (struct xref_db *rdb,
					   struct xref_section **sp,
					   struct v_selection *sel)
{
	enum s_status res;

	res = init_tdict (rdb, sel, *sp);
	if (res != S_CSC_OK)
		return res;

	res = get_trailer_info (rdb, &(*sp) -> dict);
	if (res != S_CSC_OK)
		return res;

	res = (*sp) -> ifc -> scratch (rdb, sp);
	if (res != S_CSC_OK)
		return res;

	(*sp) -> secin = 1;

	res = init_page0 (rdb, *sp);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

enum s_status xref_scratch_db (struct xref_db *rdb, struct v_selection *sel,
			       struct f_file *file)
{
	struct xref_section *sp;
	struct xref_ifc *ifc;
	struct f_file *f;
	enum s_status res;

	if (file == NULL) {
		res = f_append (rdb -> file, &f);
		if (res != S_CSC_OK)
			return res;

		file = f;
	} else
		f_ref (file);

	if (rdb -> scfile != NULL)
		f_unref (rdb -> scfile);

	rdb -> scfile = file;

	ifc = *(struct xref_ifc **) v_get_fval (sel, &optable);

	res = init_section (&sp, ifc);
	if (res != S_CSC_OK)
		return res;
	
	res = init_scratch_section (rdb, &sp, sel);
	if (res != S_CSC_OK) {
		cleanup_section (sp);
		return res;
	}

	sp -> next = rdb -> xrefs;
	rdb -> xrefs = sp;

	return S_CSC_OK;
}

enum s_status xref_sync_start (struct xref_db *rdb, struct v_selection *sel)
{
	enum s_status res;

	if (rdb -> xrefs -> next == NULL) {
		F_WRITELN (rdb -> scfile, "%%%s", v_get_hdr (sel));
		if (!sel -> pref -> text) 
			F_WRITELN (rdb -> scfile, "%%%c%c%c%c", 
				   0xff, 0xfe, 0xfd, 0xfc);
	}

	return S_CSC_OK;
}

enum s_status xref_sync_db (struct xref_db *rdb, struct v_selection *sel,
			    struct obj *trailer)
{
	enum s_status res;

	obj_destroy (&rdb -> xrefs -> dict);
	res = obj_dup (trailer, &rdb -> xrefs -> dict);
	if (res != S_CSC_OK)
		return res;

	F_TELL (rdb -> scfile, &rdb -> xrefs -> trofs);

	res = rdb -> xrefs -> ifc -> sync_section (rdb, rdb -> xrefs, sel);
	if (res != S_CSC_OK)
		return res;

	F_NEWLINE (rdb -> scfile);
	F_WRITELN (rdb -> scfile, "startxref");
	F_WRITELN (rdb -> scfile, "%lu", rdb -> xrefs -> trofs);
	F_WRITELN (rdb -> scfile, "%%%%EOF");

	f_unref (rdb -> scfile);
	rdb -> scfile = NULL;

	return S_CSC_OK;
}

enum s_status xref_get_file_info (struct xref_db *rdb, 
				  const char **fname, pdf_ofs_t *size)
{
	*fname = rdb -> scfile -> fname;

	return f_tell (rdb -> scfile, size);
}

void xref_ref_db (struct xref_db *rdb, struct ref_owner *p)
{
	if (rdb != NULL)
		ref_up (&rdb -> refs, p);
}

void xref_unref_db (struct xref_db *rdb, struct ref_owner *p)
{	
	if (rdb == NULL || ref_down (&rdb -> refs, p))
		return;

	cache_destroy_all (rdb);

	if (rdb -> file != NULL)
		f_unref (rdb -> file);

	if (rdb -> scfile != NULL)
		f_unref (rdb -> scfile);

	ps_destroy (rdb -> parser);

	cleanup_xrefs (&rdb -> xrefs);

	ref_zero (&rdb -> refs);

	free (rdb);
}

void xref_set_aborted (struct xref_db *rdb)
{
	if (rdb -> scfile != NULL)
		f_unlink_on_close (rdb -> scfile);
}

enum s_status xref_serialize_obj (struct xref_db *rdb, struct obj *obj)
{
	enum s_status res;
	pdf_ofs_t ofs;

	res = f_tell (rdb -> scfile, &ofs);
	if (res != S_CSC_OK)
		return res;

	res = xref_place_object (rdb, obj, ofs);
	if (res != S_CSC_OK)
		return res;

	res = ps_serialize_obj (rdb -> scfile, obj);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status load_object (struct xref_db *rdb, 
				  struct obj_reference *ref,
				  struct obj *obj)
{
	struct xref_subsection *ssp;
	struct xref_section *sp;
	enum s_status res;

	ssp = fetch_subsection (rdb, ref -> oid, &sp);
	if (ssp == NULL) {
		*obj = obj_null;
		return S_CSC_OK;
	}

	res = sp -> ifc -> load_object (rdb, sp, ssp, ref, obj);
	if (res != S_CSC_OK)
		return res;

	obj -> ref = *ref;

	return S_CSC_OK;
}

enum s_status xref_get_object (struct xref_db *rdb, 
			       struct obj_reference *ref,
			       struct obj *obj)
{
	return load_object (rdb, ref, obj);
}

enum s_status xref_resolve_no_cast (struct xref_db *rdb, struct obj *sobj,
				    enum obj_type ot, struct obj *dobj, 
				    int *present, const char *label)
{
	enum s_status res;
	struct obj tobj;

	if (sobj -> type == OT_REFERENCE && rdb -> file != NULL)
		res = xref_get_object (rdb, &sobj -> ref, &tobj);
	else
		res = obj_dup (sobj, &tobj);
	if (res != S_CSC_OK)
		return res;

	if (tobj.type == OT_NULL) {
		if (present == NULL)
			return obj_error (sobj,
					  "Required object (%s) is null",
					  label);
		*present = 0;
	} else {
		if (present != NULL)
			*present = 1;
		
		if (!obj_can_cast (&tobj, ot)) {
			res = obj_error (&tobj, 
					 "Bad object type: "
					 "%s is a %s (not a %s)",
					 label, obj_tag (tobj.type),
					 obj_tag (ot));
			obj_destroy (&tobj);
			return res;
		}
	}

	*dobj = tobj;

	return S_CSC_OK;
}

enum s_status xref_resolve (struct xref_db *rdb, struct obj *sobj,
			    enum obj_type ot, struct obj *dobj, 
			    int *present, const char *label)
{
	enum s_status res;

	res = xref_resolve_no_cast (rdb, sobj, ot, dobj, present, label);
	if (res != S_CSC_OK)
		return res;

	obj_cast (dobj, ot);

	return S_CSC_OK;
}

static enum s_status stream_resolve (struct xref_db *rdb,
				     struct xref_stream *xs,
				     struct obj *sobj)
{
	struct xref_obj_array *filter, *param;
	struct obj_region sreg, dreg;
	struct obj_stream *stream;
	struct f_file *file;
	struct obj *cparm;
	enum s_status res;
	int i;

	stream = &sobj -> val.stream;

	res = S_CSC_OK;
	if (xs -> fspec != NULL) {
		obj_region_destroy (&stream -> region);
		res = fspec_get (xs -> fspec, &file, NULL, NULL, NULL);
		if (res != S_CSC_OK)
			return res;

		obj_init_file_region (&stream -> region,
				      file -> pdf, file -> end, file);
		f_unref (file);

		filter = &xs -> ffilter;
		param = &xs -> fdecode_parms;
	} else {
		filter = &xs -> filter;
		param = &xs -> decode_parms;
	}

	res = obj_region_dup (&stream -> region, &sreg);
	if (res != S_CSC_OK)
		return res;

	res = S_CSC_OK;
	for (i = 0; i < XREF_A_SIZE (filter); i++) {
		if (i < XREF_A_SIZE (param) &&
		    param -> data [i].type != OT_NULL)
			cparm = &param -> data [i];
		else
			cparm = NULL;
		res = flt_decode (&filter -> data [i].val.string,
				  rdb, xs -> dl, cparm, &sreg, &dreg);
		obj_region_destroy (&sreg);
		if (res != S_CSC_OK) {
			if (s_get_code (res) == S_CSC_NOT_SUPPORTED)
				res = S_CSC_OK;

			return res;
		}

		sreg = dreg;
	}

	obj_region_destroy (&stream -> decoded);
	stream -> decoded = sreg;

	return S_CSC_OK;
}

enum s_status xref_resolve_stream_xs (struct xref_db *rdb,
				      struct obj *sobj,
				      struct xref_stream *xsp,
				      struct ref_owner *p)
{
	struct xref_stream xs;
	enum s_status res;

	memset (&xs, 0, sizeof (struct xref_stream));

	res = xref_fields_parse (rdb, NULL, p, sobj, 
				 &xref_stream_field, &xs);
	if (res != S_CSC_OK)
		return res;

	res = stream_resolve (rdb, &xs, sobj);
	if (res == S_CSC_OK)
		res = obj_region_dup (&sobj -> val.stream.region,
				      &xs.encoded);
	if (res == S_CSC_OK)
		res = obj_region_dup (&sobj -> val.stream.decoded,
				      &xs.decoded);
	if (res != S_CSC_OK) {
		xref_stream_destroy (&xs, p);
		return res;
	}

	*xsp = xs;

	return S_CSC_OK;
}

enum s_status xref_resolve_stream (struct xref_db *rdb, struct obj *sobj)
{
	struct xref_stream xs;
	struct ref_owner own;
	enum s_status res;

	if (sobj -> val.stream.decoded.begin != NULL)
		return S_CSC_OK;

	ref_init_owner (&own, "resolve_stream");
	res = xref_resolve_stream_xs (rdb, sobj, &xs, &own);
	if (res != S_CSC_OK)
		return res;

	xref_stream_destroy (&xs, &own);

	if (sobj -> val.stream.decoded.begin == NULL)
		return obj_error (sobj, "Unsupported filter");

	return S_CSC_OK;
}

enum s_status xref_stream_ser (struct xref_db *rdb, struct xref_bind *robj,
			       struct v_selection *sel, struct obj *obj,
			       struct xref_stream *xs)
{
	enum xref_changed changed;
	enum s_status res;

	obj_destroy (obj);

	memset (obj, 0, sizeof (struct obj));

	res = obj_region_dup (&xs -> decoded, &obj -> val.stream.decoded);
	if (res != S_CSC_OK)
		return res;

	res = obj_region_dup (&xs -> encoded, &obj -> val.stream.region);
	if (res != S_CSC_OK) {
		obj_region_destroy (&obj -> val.stream.decoded);
		return res;
	}

	xs -> dl = 
		obj -> val.stream.decoded.end - 
		obj -> val.stream.decoded.begin;

	if (!xs -> dl)
		xs -> dl--;

	obj -> type = OT_STREAM;
	res = xref_fields_ser (rdb, robj, sel, obj, &changed, 
			       &xref_stream_field, xs);
	if (res != S_CSC_OK) {
		obj_region_destroy (&obj -> val.stream.decoded);
		obj_region_destroy (&obj -> val.stream.region);
		return res;
	}

	return S_CSC_OK;
}

void xref_stream_destroy (struct xref_stream *xs, struct ref_owner *p)
{
	xref_fields_destroy (&xref_stream_field, xs, p);
	obj_region_destroy (&xs -> encoded);
	obj_region_destroy (&xs -> decoded);
	memset (xs, 0, sizeof (struct xref_stream));
}

enum s_status xref_add_selection (struct v_selection *sel)
{
	return v_add_selection (sel, &optable);
}

struct obj *xref_get_trailer (struct xref_db *rdb)
{
	return &rdb -> xrefs -> dict;
}

pdf_integer_t xref_get_id (struct xref_db *rdb)
{
	return rdb != NULL ? rdb -> id : -1;
}
