/*
 * PDFlow -- agc.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <assert.h>
#include <string.h>

#include "graph/agc.h"

static enum s_status pg_done (struct agc_context *agc);
static void pg_destroy (struct agc_context *agc);
static enum s_status pg_rgb (struct agc_context *agc, pdf_real_t r, 
			     pdf_real_t g, pdf_real_t b, pdf_bool_t stroking);
static enum s_status pg_ctm (struct agc_context *agc, 
			     struct cobj_matrix *ctm);
static enum s_status pg_rect (struct agc_context *agc, 
			      struct cobj_rect *rect, struct gs_paint *paint);
static enum s_status pg_move_to (struct agc_context *agc, 
				 struct gs_point *p);
static enum s_status pg_line_to (struct agc_context *agc, 
				 struct gs_point *p);
static enum s_status pg_bezier_to (struct agc_context *agc, 
				   struct gs_point *b,
				   struct gs_point *ca,
				   struct gs_point *cb);
static enum s_status pg_stroke (struct agc_context *agc, 
				struct gs_paint *paint);
static enum s_status pg_xo (struct agc_context *agc,
			    struct xo_object *xo);
static enum s_status pg_shading (struct agc_context *agc,
				 struct shd_shading *shd);
static enum s_status pg_set_state (struct agc_context *agc,
				   struct gs_text_state *tstate,
				   struct gs_graph_state *gstate,
				   struct cobj_rect *cpath);
static enum s_status pg_write_char (struct agc_context *agc,
				    struct gs_glyph_info *gi,
				    struct gs_paint *paint);
static enum s_status pg_play (struct agc_context *agc,
			      struct agc_context *pagc);
static void pg_set_enable (struct agc_context *agc,
			   agc_code_t code, pdf_bool_t enabled);

static enum s_status ft_done (struct agc_context *agc);
static void ft_destroy (struct agc_context *agc);
static enum s_status ft_rgb (struct agc_context *agc, pdf_real_t r, 
			     pdf_real_t g, pdf_real_t b, pdf_bool_t stroking);
static enum s_status ft_ctm (struct agc_context *agc, 
			     struct cobj_matrix *ctm);
static enum s_status ft_rect (struct agc_context *agc, 
			      struct cobj_rect *rect, struct gs_paint *paint);
static enum s_status ft_move_to (struct agc_context *agc, 
				 struct gs_point *p);
static enum s_status ft_line_to (struct agc_context *agc, 
				 struct gs_point *p);
static enum s_status ft_bezier_to (struct agc_context *agc, 
				   struct gs_point *b,
				   struct gs_point *ca,
				   struct gs_point *cb);
static enum s_status ft_stroke (struct agc_context *agc, 
				struct gs_paint *paint);
static enum s_status ft_xo (struct agc_context *agc,
			    struct xo_object *xo);
static enum s_status ft_shading (struct agc_context *agc,
				 struct shd_shading *shd);
static enum s_status ft_set_state (struct agc_context *agc,
				   struct gs_text_state *tstate,
				   struct gs_graph_state *gstate,
				   struct cobj_rect *cpath);
static enum s_status ft_write_char (struct agc_context *agc,
				    struct gs_glyph_info *gi,
				    struct gs_paint *paint);
static enum s_status ft_play (struct agc_context *agc,
			      struct agc_context *pagc);
static void ft_set_enable (struct agc_context *agc,
			   agc_code_t code, pdf_bool_t enabled);

static enum s_status mt_done (struct agc_context *agc);
static void mt_destroy (struct agc_context *agc);
static enum s_status mt_rgb (struct agc_context *agc, pdf_real_t r, 
			     pdf_real_t g, pdf_real_t b, pdf_bool_t stroking);
static enum s_status mt_ctm (struct agc_context *agc, 
			     struct cobj_matrix *ctm);
static enum s_status mt_rect (struct agc_context *agc, 
			      struct cobj_rect *rect, struct gs_paint *paint);
static enum s_status mt_move_to (struct agc_context *agc, 
				 struct gs_point *p);
static enum s_status mt_line_to (struct agc_context *agc, 
				 struct gs_point *p);
static enum s_status mt_bezier_to (struct agc_context *agc, 
				   struct gs_point *b,
				   struct gs_point *ca,
				   struct gs_point *cb);
static enum s_status mt_stroke (struct agc_context *agc, 
				struct gs_paint *paint);
static enum s_status mt_xo (struct agc_context *agc,
			    struct xo_object *xo);
static enum s_status mt_shading (struct agc_context *agc,
				 struct shd_shading *shd);
static enum s_status mt_set_state (struct agc_context *agc,
				   struct gs_text_state *tstate,
				   struct gs_graph_state *gstate,
				   struct cobj_rect *cpath);
static enum s_status mt_write_char (struct agc_context *agc,
				    struct gs_glyph_info *gi,
				    struct gs_paint *paint);
static enum s_status mt_play (struct agc_context *agc,
			      struct agc_context *pagc);
static void mt_set_enable (struct agc_context *agc,
			   agc_code_t code, pdf_bool_t enabled);

static struct agc_ifc pg_ifc = {
	pg_done, pg_destroy, pg_rgb, pg_ctm, 
	pg_rect, pg_move_to, pg_line_to, pg_bezier_to, pg_stroke,
	pg_xo, pg_shading, pg_set_state, pg_write_char, pg_play, pg_set_enable
};

static struct agc_ifc ft_ifc = {
	ft_done, ft_destroy, ft_rgb, ft_ctm,
	ft_rect, ft_move_to, ft_line_to, ft_bezier_to, ft_stroke,
	ft_xo, ft_shading, ft_set_state, ft_write_char, ft_play, ft_set_enable
};

static struct agc_ifc mt_ifc = {
	mt_done, mt_destroy, mt_rgb, mt_ctm,
	mt_rect, mt_move_to, mt_line_to, mt_bezier_to, mt_stroke,
	mt_xo, mt_shading, mt_set_state, mt_write_char, mt_play, mt_set_enable
};


static enum s_status flush_text (struct agc_context *agc);

static enum s_status pg_init (struct agc_context *agc, struct xref_db *rdb,
			      struct gspg_page *gspg)
{
	enum s_status res;

	res = T_CALL (gs_context_new, &agc -> d.pg.gpc, rdb, 
		      ref_rown (&agc -> refs));
	if (res != S_CSC_OK)
		return res;

	agc -> d.pg.gspg = gspg;
	gspg_page_ref (gspg, ref_rown (&agc -> refs));

	return S_CSC_OK;
}

enum s_status agc__new (struct agc_context **agcp, struct xref_db *rdb,
			struct gspg_page *gspg, struct ref_owner *p, 
			const char *fname, int line)
{
	struct agc_context *agc;
	enum s_status res;

	agc = malloc (sizeof (struct agc_context));
	if (agc == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (agc, 0, sizeof (struct agc_context));

	ref__init (&agc -> refs, "agc-pg", p, fname, line);
	
	agc -> ifc = &pg_ifc;
	agc -> bctm = cobj_eye;

	res = pg_init (agc, rdb, gspg);
	if (res != S_CSC_OK) {
		agc_unref (agc, p);
		return res;
	}

	*agcp = agc;

	return S_CSC_OK;
}

enum s_status agc__float_new (struct agc_context **agcp, struct ref_owner *p,
			      const char *fname, int line)
{
	struct agc_context *agc;

	agc = malloc (sizeof (struct agc_context));
	if (agc == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (agc, 0, sizeof (struct agc_context));

	ref__init (&agc -> refs, "agc-ft", p, fname, line);

	agc -> ifc = &ft_ifc;
	agc -> bctm = cobj_eye;

	*agcp = agc;

	return S_CSC_OK;
}

enum s_status agc__measure_new (struct agc_context **agcp, struct ref_owner *p,
				const char *fname, int line)
{
	struct agc_context *agc;

	agc = malloc (sizeof (struct agc_context));
	if (agc == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (agc, 0, sizeof (struct agc_context));

	ref__init (&agc -> refs, "agc-ms", p, fname, line);

	agc -> ifc = &mt_ifc;
	agc -> bctm = cobj_eye;
	agc -> d.mt.ctm = cobj_eye;
	cobj_rect_set_null (&agc -> d.mt.rect);

	*agcp = agc;

	return S_CSC_OK;
}

void agc__ref (struct agc_context *agc, struct ref_owner *p, 
	       const char *fname, int line)
{
	if (agc != NULL)
		ref__up (&agc -> refs, p, fname, line);
}

void agc_unref (struct agc_context *agc, struct ref_owner *p)
{
	if (agc == NULL || ref_down (&agc -> refs, p))
		return;

	agc -> ifc -> destroy (agc);

	ref_zero (&agc -> refs);

	free (agc);
}

static enum s_status pg_done (struct agc_context *agc)
{
	enum s_status res;

	res = flush_text (agc);
	if (res != S_CSC_OK)
		return res;

	res = gs_done (agc -> d.pg.gpc);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static void pg_destroy (struct agc_context *agc)
{
	gs_context_unref (agc -> d.pg.gpc, ref_rown (&agc -> refs));
	gspg_page_unref (agc -> d.pg.gspg, ref_rown (&agc -> refs));
}

static enum s_status pg_ctm (struct agc_context *agc, 
			     struct cobj_matrix *ctm)
{
	return gs_set_ctm (agc -> d.pg.gpc, ctm);
}

static enum s_status pg_rgb (struct agc_context *agc, pdf_real_t r, 
			     pdf_real_t g, pdf_real_t b, pdf_bool_t stroking)
{
	enum s_status res;
	pdf_real_t rgb [3];

	rgb [0] = r;
	rgb [1] = g;
	rgb [2] = b;

	if (stroking)
		res = gs_set_stroking_colour (agc -> d.pg.gpc, 
					      GC_CSF_DEVICE_RGB, rgb);
	else
		res = gs_set_nonstroking_colour (agc -> d.pg.gpc,
						 GC_CSF_DEVICE_RGB, rgb);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status pg_rect (struct agc_context *agc, 
			      struct cobj_rect *rect, struct gs_paint *paint)
{
	enum s_status res;

	res = flush_text (agc);
	if (res != S_CSC_OK)
		return res;

	res = gs_rect (agc -> d.pg.gpc, NULL, rect);
	if (res != S_CSC_OK)
		return res;

	res = gspg_stmt_add_path (agc -> d.pg.gspg, agc -> d.pg.gpc, paint, 1);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status pg_move_to (struct agc_context *agc, 
				 struct gs_point *p)
{
	return gs_move (agc -> d.pg.gpc, NULL, p -> x, p -> y);
}

static enum s_status pg_line_to (struct agc_context *agc, 
				 struct gs_point *p)
{
	return gs_lineto (agc -> d.pg.gpc, NULL, p -> x, p -> y);
}

static enum s_status pg_bezier_to (struct agc_context *agc, 
				   struct gs_point *b,
				   struct gs_point *ca,
				   struct gs_point *cb)
{
	return gs_bezierto (agc -> d.pg.gpc, NULL, 
			    ca -> x, ca -> y,
			    cb -> x, cb -> y,
			    b -> x, b -> y);
}

static enum s_status pg_stroke (struct agc_context *agc, 
				struct gs_paint *paint)
{
	return gspg_stmt_add_path (agc -> d.pg.gspg, 
				   agc -> d.pg.gpc, paint, 1);
}

static enum s_status pg_xo (struct agc_context *agc,
			    struct xo_object *xo)
{
	enum s_status res;

	res = gspg_stmt_add_xobj (agc -> d.pg.gspg, agc -> d.pg.gpc, xo, 1);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status pg_shading (struct agc_context *agc,
				 struct shd_shading *sh)
{
	enum s_status res;

	res = gspg_stmt_add_shading (agc -> d.pg.gspg, agc -> d.pg.gpc, sh, 1);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status flush_text (struct agc_context *agc)
{
	struct agc_text *at;
	struct gs_text text;
	enum s_status res;

	at = &agc -> d.pg.at;
	if (at -> text.head == NULL)
		return S_CSC_OK;

	if (at -> text.head -> next == NULL)
		obj_set_string (&text.text, &f_null,
				&at -> text.head -> obj.val.string);
	else
		obj_set_array (&text.text, &f_null, &at -> text);

	text.tstate = NULL;
	text.gstate = NULL;

	res = gs_begin_text (agc -> d.pg.gpc, &obj_null);
	if (res != S_CSC_OK)
		return res;

	res = gs_set_mat (agc -> d.pg.gpc, &at -> tm0);
	if (res != S_CSC_OK)
		return res;

	res = gs_write_text (agc -> d.pg.gpc, &text);
	obj_array_destroy (&at -> text);
	at -> tail = NULL;
	if (res != S_CSC_OK)
		return res;

	res = gspg_stmt_add_path (agc -> d.pg.gspg, agc -> d.pg.gpc, 
				  &at -> paint, 1);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status pg_set_state (struct agc_context *agc,
				   struct gs_text_state *tstate,
				   struct gs_graph_state *gstate,
				   struct cobj_rect *cpath)
{
	enum s_status res;
	int tchan, gchan;

	if (tstate != NULL) {
		res = gs_import_tstate (&tstate, ref_rown (&agc -> refs));
		if (res != S_CSC_OK)
			return res;

		tchan = !gs_text_equal (gs_ro_text (agc -> d.pg.gpc),
					tstate);
	} else
		tchan = 0;
	
	if (gstate != NULL) {
		res = gs_import_gstate (&gstate, ref_rown (&agc -> refs));
		if (res != S_CSC_OK) {
			gs_text_unref (tstate, ref_rown (&agc -> refs));
			return res;
		}

		gchan = !gs_graph_equal (gs_ro_graph (agc -> d.pg.gpc),
					 gstate);
 
	} else
		gchan = 0;

	if (tchan || gchan) {
		res = flush_text (agc);
		if (res != S_CSC_OK) {
			gs_text_unref (tstate, ref_rown (&agc -> refs));
			gs_graph_unref (gstate, ref_rown (&agc -> refs));
			return res;
		}
	}

	if (tchan)
		gs_set_tstate (agc -> d.pg.gpc, tstate);

	if (gchan)
		gs_set_gstate (agc -> d.pg.gpc, gstate);
		       
	gs_text_unref (tstate, ref_rown (&agc -> refs));
	gs_graph_unref (gstate, ref_rown (&agc -> refs));

	return S_CSC_OK;
}

static enum s_status mt_done (struct agc_context *agc)
{
	return S_CSC_OK;
}

static void mt_destroy (struct agc_context *agc)
{
	/* empty */
}

static enum s_status mt_rgb (struct agc_context *agc, pdf_real_t r, 
			     pdf_real_t g, pdf_real_t b, pdf_bool_t stroking)
{
	return S_CSC_OK;
}

static void add_rect (struct agc_context *agc, struct cobj_rect *rect)
{
	struct cobj_rect r;

	r = *rect;
	gs_translate_rect (&r, &agc -> d.mt.ctm);
	if (!cobj_rect_is_null (&agc -> d.mt.cpath))
		cobj_rect_intersect (&r, &agc -> d.mt.cpath);
	cobj_rect_union (&agc -> d.mt.rect, &r);
}

static enum s_status mt_ctm (struct agc_context *agc, 
			     struct cobj_matrix *ctm)
{
	struct cobj_matrix m;

	cobj_matrix_mul (ctm, &agc -> d.mt.ctm, &m, 0);
	agc -> d.mt.ctm = m;

	return S_CSC_OK;
}

static enum s_status mt_rect (struct agc_context *agc, 
			      struct cobj_rect *rect, struct gs_paint *paint)
{
	add_rect (agc, rect);

	return S_CSC_OK;
}

static enum s_status mt_move_to (struct agc_context *agc, 
				 struct gs_point *p)
{
	agc -> d.mt.point = *p;

	cobj_rect_add_point (&agc -> d.mt.path, p -> x, p -> y);

	return S_CSC_OK;
}

static enum s_status mt_line_to (struct agc_context *agc, 
				 struct gs_point *p)
{
	agc -> d.mt.point = *p;

	cobj_rect_add_point (&agc -> d.mt.path, p -> x, p -> y);

	return S_CSC_OK;
}

static enum s_status mt_bezier_to (struct agc_context *agc, 
				   struct gs_point *b,
				   struct gs_point *ca,
				   struct gs_point *cb)
{
	struct gs_bezier bezier;
	struct cobj_rect r;

	bezier.a = agc -> d.mt.point;
	bezier.b = *b;
	bezier.ca = *ca;
	bezier.cb = *cb;

	agc -> d.mt.point = *b;

	gs_get_bezier_rect (&r, &bezier);

	add_rect (agc, &r);

	return S_CSC_OK;
}

static enum s_status mt_stroke (struct agc_context *agc, 
				struct gs_paint *paint)
{
	add_rect (agc, &agc -> d.mt.path);
	cobj_rect_set_null (&agc -> d.mt.path);

	return S_CSC_OK;
}

static enum s_status mt_xo (struct agc_context *agc,
			    struct xo_object *xo)
{
	struct cobj_rect rect;

	switch (xo_get_subtype (xo)) {
	case XO_T_NONE:
		assert (0);
		break;

	case XO_T_PS:
		/* Dunno how to measure it */
		return S_CSC_OK;

	case XO_T_FORM:
		rect = xo -> u.form.bbox;
		break;

	case XO_T_IMAGE:
		rect.left = rect.bottom = 0;
		rect.right = rect.top = 1;
		break;

	default:
		assert (0);
	}


	add_rect (agc, &rect);

	return S_CSC_OK;
}

static enum s_status mt_shading (struct agc_context *agc,
				 struct shd_shading *sh)
{
	add_rect (agc, &agc -> d.mt.cpath);

	return S_CSC_OK;
}

static enum s_status mt_set_state (struct agc_context *agc,
				   struct gs_text_state *tstate,
				   struct gs_graph_state *gstate,
				   struct cobj_rect *cpath)
{
	if (gstate != NULL) {
		agc -> d.mt.ctm = gstate -> ctm;
		agc -> d.mt.cpath = *cpath;
	}

	return S_CSC_OK;
}

static enum s_status mt_write_char (struct agc_context *agc,
				    struct gs_glyph_info *gi,
				    struct gs_paint *paint)
{
	struct cobj_rect r;

	r = gi -> f.rect;
	gs_translate_rect (&r, &agc -> d.mt.ctm);
	cobj_rect_union (&agc -> d.mt.rect, &r);

	return S_CSC_OK;
}

static enum s_status mt_play (struct agc_context *agc,
			      struct agc_context *pagc)
{
	assert (0);
}

static void mt_set_enable (struct agc_context *agc,
			   agc_code_t code, pdf_bool_t enabled)
{
	assert (0);
}

static enum s_status init_text (struct agc_context *agc,
				struct gs_glyph_info *gi,
				struct gs_paint *paint)
{
	struct agc_text *at;
	enum s_status res;

	at = &agc -> d.pg.at;

	if (at -> tail != NULL &&
	    !memcmp (&at -> paint, paint, sizeof (struct gs_paint)))
		return S_CSC_OK;

	res = flush_text (agc);
	if (res != S_CSC_OK)
		return res;

	at -> paint = *paint;
	at -> tm = at -> tm0 = gi -> t_m;

	return S_CSC_OK;
}

static enum s_status add_string (struct agc_text *at,
				 pdf_real_t kern,
				 struct gs_glyph_info *gi)
{
	enum s_status res;
	struct obj_bstr s;
	struct obj obj;

	if (at -> tail != NULL) {
		obj_set_real (&obj, &f_null, kern);
		res = obj_xfer_t_array (&at -> text, &obj, &at -> tail);
		if (res != S_CSC_OK)
			return res;
	}

	s.size = gi -> c.size;
	s.s = malloc (s.size);
	if (s.s == NULL)
		return S_CSC_OUT_OF_MEMORY;

	memcpy (s.s, gi -> c.bytes, s.size);

	obj_set_string (&obj, &f_null, &s);
	res = obj_xfer_t_array (&at -> text, &obj, &at -> tail);
	if (res != S_CSC_OK) {
		obj_destroy (&obj);
		return res;
	}

	return S_CSC_OK;
}


static enum s_status append_string (struct agc_text *at,
				    struct gs_glyph_info *gi)
{
	struct obj_bstr *s;
	size_t size;
	void *np;

	if (at -> tail == NULL)
		return add_string (at, 0, gi);

	s = &at -> tail -> obj.val.string;
	size = s -> size + gi -> c.size;
	np = realloc (s -> s, size);
	if (np == NULL)
		return S_CSC_OUT_OF_MEMORY;
	
	memcpy (np + s -> size, gi -> c.bytes, gi -> c.size);

	s -> size = size;
	s -> s = np;

	return S_CSC_OK;
}

static enum s_status pg_write_char (struct agc_context *agc,
				    struct gs_glyph_info *gi,
				    struct gs_paint *paint)
{
	struct agc_text *at;
	struct gs_point vec;
	enum s_status res;
	pdf_bool_t can_kern;
	pdf_real_t kern;

	res = init_text (agc, gi, paint);
	if (res != S_CSC_OK)
		return res;

	at = &agc -> d.pg.at;

	can_kern = 1;
	kern = 0;
	if (at -> tail != NULL)
		can_kern = gs_get_kerning (agc -> d.pg.gpc, &at -> tm, 
					   &gi -> t_m, &kern);

	if (!can_kern) {
		res = flush_text (agc);
		if (res != S_CSC_OK)
			return res;

		at -> tm = at -> tm0 = gi -> t_m;
	}

	if (kern) {
		res = add_string (at, kern * 1000, gi);
		if (res != S_CSC_OK)
			return res;
	} else {
		res = append_string (at, gi);
		if (res != S_CSC_OK)
			return res;
	}

	gs_get_vector_gc (agc -> d.pg.gpc, gi -> f.wx - kern, 
			  gi -> f.wy - kern, &vec);

	at -> tm.a [2][0] += vec.x;
	at -> tm.a [2][1] += vec.y;

	return S_CSC_OK;
}

static enum s_status pg_play (struct agc_context *agc,
			      struct agc_context *pagc)
{
	assert (0);
}

static void pg_set_enable (struct agc_context *agc,
			   agc_code_t code, pdf_bool_t enabled)
{
	assert (0);
}

static enum s_status ft_done (struct agc_context *agc)
{
	return S_CSC_OK;
}

static enum s_status new_cmd (struct agc_context *agc,
			      enum agc_float_command_type type,	
			      struct agc_float_command **cmdp)
{
	struct agc_float_command *cmd;

	cmd = malloc (sizeof (struct agc_float_command));
	if (cmd == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (cmd, 0, sizeof (struct agc_float_command));
	
	cmd -> type = type;
	cmd -> code = agc -> current_code;
	cmd -> enabled = 1;
	if (agc -> d.ft.cmd != NULL) {
		cmd -> next = agc -> d.ft.cmd -> next;
		agc -> d.ft.cmd -> next = cmd;
	} else
		cmd -> next = cmd;

	agc -> d.ft.cmd = cmd;

	*cmdp = cmd;

	return S_CSC_OK;
}

static void destroy_cmd (struct agc_context *agc, 
			 struct agc_float_command *cmd)
{
	switch (cmd -> type) {
	case AGC_FTC_RGB:
	case AGC_FTC_RECT:
		break;

	case AGC_FTC_SET_STATE:
		gs_text_unref (cmd -> d.set_state.tstate,
			       ref_rown (&agc -> refs));
		gs_graph_unref (cmd -> d.set_state.gstate,
				ref_rown (&agc -> refs));
		break;

	case AGC_FTC_WRITE_CHAR:
	case AGC_FTC_CTM:
	case AGC_FTC_MOVE_TO:
	case AGC_FTC_LINE_TO:
	case AGC_FTC_BEZIER_TO:
	case AGC_FTC_STROKE:
		break;

	case AGC_FTC_XOBJ:
		xo_unref (cmd -> d.xobj.xo, ref_rown (&agc -> refs));
		break;

	case AGC_FTC_SHADING:
		shd_unref (cmd -> d.shading.sh, ref_rown (&agc -> refs));
		break;
	}

	free (cmd);
}

static void ft_destroy (struct agc_context *agc)
{
	struct agc_float_command *cmd;

	cmd = agc -> d.ft.cmd;
	if (cmd == NULL)
		return;

	agc -> d.ft.cmd = cmd -> next;;
	cmd -> next = NULL;

	while (agc -> d.ft.cmd != NULL) {
		cmd = agc -> d.ft.cmd;
		agc -> d.ft.cmd = cmd -> next;
		destroy_cmd (agc, cmd);
	}
}

static enum s_status ft_rgb (struct agc_context *agc, pdf_real_t r, 
			     pdf_real_t g, pdf_real_t b, pdf_bool_t stroking)
{
	struct agc_float_command *cmd;
	enum s_status res;

	res = new_cmd (agc, AGC_FTC_RGB, &cmd);
	if (res != S_CSC_OK)
		return res;

	cmd -> d.rgb.r = r;
	cmd -> d.rgb.g = g;
	cmd -> d.rgb.b = b;
	cmd -> d.rgb.stroking = stroking;

	return S_CSC_OK;
}

static enum s_status ft_ctm (struct agc_context *agc, 
			     struct cobj_matrix *ctm)
{
	struct agc_float_command *cmd;
	enum s_status res;

	res = new_cmd (agc, AGC_FTC_CTM, &cmd);
	if (res != S_CSC_OK)
		return res;

	cmd -> d.ctm.mat = *ctm;

	return S_CSC_OK;
}

static enum s_status ft_rect (struct agc_context *agc, 
			      struct cobj_rect *rect, struct gs_paint *paint)
{
	struct agc_float_command *cmd;
	enum s_status res;

	res = new_cmd (agc, AGC_FTC_RECT, &cmd);
	if (res != S_CSC_OK)
		return res;

	cmd -> d.rect.rect = *rect;
	cmd -> d.rect.paint = *paint;

	return S_CSC_OK;
}

static enum s_status ft_move_to (struct agc_context *agc, 
				 struct gs_point *p)
{
	struct agc_float_command *cmd;
	enum s_status res;

	res = new_cmd (agc, AGC_FTC_MOVE_TO, &cmd);
	if (res != S_CSC_OK)
		return res;

	cmd -> d.move_to.p = *p;

	return S_CSC_OK;
}

static enum s_status ft_line_to (struct agc_context *agc, 
				 struct gs_point *p)
{
	struct agc_float_command *cmd;
	enum s_status res;

	res = new_cmd (agc, AGC_FTC_LINE_TO, &cmd);
	if (res != S_CSC_OK)
		return res;

	cmd -> d.line_to.p = *p;

	return S_CSC_OK;
}

static enum s_status ft_bezier_to (struct agc_context *agc, 
				   struct gs_point *b,
				   struct gs_point *ca,
				   struct gs_point *cb)
{
	struct agc_float_command *cmd;
	enum s_status res;

	res = new_cmd (agc, AGC_FTC_BEZIER_TO, &cmd);
	if (res != S_CSC_OK)
		return res;

	cmd -> d.bezier_to.b = *b;
	cmd -> d.bezier_to.ca = *ca;
	cmd -> d.bezier_to.cb = *cb;

	return S_CSC_OK;
}

static enum s_status ft_stroke (struct agc_context *agc, 
				struct gs_paint *paint)
{
	struct agc_float_command *cmd;
	enum s_status res;

	res = new_cmd (agc, AGC_FTC_STROKE, &cmd);
	if (res != S_CSC_OK)
		return res;

	cmd -> d.stroke.paint = *paint;

	return S_CSC_OK;
}

static enum s_status ft_xo (struct agc_context *agc,
			    struct xo_object *xo)
{
	struct agc_float_command *cmd;
	enum s_status res;

	res = new_cmd (agc, AGC_FTC_XOBJ, &cmd);
	if (res != S_CSC_OK)
		return res;

	cmd -> d.xobj.xo = xo;
	T_CALL (xo_ref, xo, ref_rown (&agc -> refs));

	return S_CSC_OK;
}

static enum s_status ft_shading (struct agc_context *agc,
				 struct shd_shading *sh)
{
	struct agc_float_command *cmd;
	enum s_status res;

	res = new_cmd (agc, AGC_FTC_SHADING, &cmd);
	if (res != S_CSC_OK)
		return res;

	cmd -> d.shading.sh = sh;
	T_CALL (shd_ref, sh, ref_rown (&agc -> refs));

	return S_CSC_OK;
}

static enum s_status ft_set_state (struct agc_context *agc,
				   struct gs_text_state *tstate,
				   struct gs_graph_state *gstate,
				   struct cobj_rect *cpath)
{
	struct agc_float_command *cmd;
	enum s_status res;

	res = new_cmd (agc, AGC_FTC_SET_STATE, &cmd);
	if (res != S_CSC_OK)
		return res;

	/* Note that if something bad happens, this command becomes
	 * a noop */
	if (tstate != NULL) {
		res = gs_import_tstate (&tstate, ref_rown (&agc -> refs));
		if (res != S_CSC_OK)
			return res;
	}

	if (gstate != NULL) {
		res = gs_import_gstate (&gstate, ref_rown (&agc -> refs));
		if (res != S_CSC_OK) {
			gs_text_unref (tstate, ref_rown (&agc -> refs));
			return res;
		}
	}

	cmd -> d.set_state.tstate = tstate;
	cmd -> d.set_state.gstate = gstate;
	cmd -> d.set_state.cpath = *cpath;

	return S_CSC_OK;
}

static enum s_status ft_write_char (struct agc_context *agc,
				    struct gs_glyph_info *gi,
				    struct gs_paint *paint)
{
	struct agc_float_command *cmd;
	enum s_status res;

	res = new_cmd (agc, AGC_FTC_WRITE_CHAR, &cmd);
	if (res != S_CSC_OK)
		return res;

	cmd -> d.write_char.gi = *gi;
	cmd -> d.write_char.paint = *paint;

	return S_CSC_OK;
}

static enum s_status ft_play (struct agc_context *agc,
			      struct agc_context *pagc)
{
	struct agc_float_command *cmd;
	enum s_status res;

	cmd = agc -> d.ft.cmd;
	if (cmd == NULL)
		return S_CSC_OK;

	do {
		cmd = cmd -> next;

		if (!cmd -> enabled)
			continue;

		res = S_CSC_OK;
		switch (cmd -> type) {
		case AGC_FTC_RGB:
			res = agc_rgb_s (pagc, cmd -> d.rgb.r,
					 cmd -> d.rgb.g, cmd -> d.rgb.b,
					 cmd -> d.rgb.stroking);
			break;

		case AGC_FTC_RECT:
			res = agc_rect (pagc, &cmd -> d.rect.rect,
					&cmd -> d.rect.paint);
			break;

		case AGC_FTC_SET_STATE:
			res = agc_set_state (pagc, cmd -> d.set_state.tstate,
					     cmd -> d.set_state.gstate,
					     &cmd -> d.set_state.cpath);
			break;

		case AGC_FTC_WRITE_CHAR:
			res = agc_write_char (pagc, &cmd -> d.write_char.gi,
					      &cmd -> d.write_char.paint);
			break;

		case AGC_FTC_CTM:
			res = agc_ctm (pagc, &cmd -> d.ctm.mat);
			break;

		case AGC_FTC_MOVE_TO:
			res = agc_move_to (pagc, &cmd -> d.move_to.p);
			break;

		case AGC_FTC_LINE_TO:
			res = agc_line_to (pagc, &cmd -> d.line_to.p);
			break;

		case AGC_FTC_BEZIER_TO:
			res = agc_bezier_to (pagc, &cmd -> d.bezier_to.b,
					     &cmd -> d.bezier_to.ca,
					     &cmd -> d.bezier_to.cb);
			break;

		case AGC_FTC_STROKE:
			res = agc_stroke (pagc, &cmd -> d.stroke.paint);
			break;

		case AGC_FTC_XOBJ:
			res = agc_draw_xobject (pagc, cmd -> d.xobj.xo);
			break;

		case AGC_FTC_SHADING:
			res = agc_draw_shading (pagc, cmd -> d.shading.sh);
		}

		if (res != S_CSC_OK)
			return res;

	} while (cmd != agc -> d.ft.cmd);

	return S_CSC_OK;
}

static void ft_set_enable (struct agc_context *agc,
			   agc_code_t code, pdf_bool_t enabled)
{
	struct agc_float_command *cmd;

	if (agc -> d.ft.cmd == NULL)
		return;

	cmd = agc -> d.ft.cmd;
	do {
		cmd = cmd -> next;
		if (cmd -> code == code)
			cmd -> enabled = enabled;
	} while (cmd != agc -> d.ft.cmd);
}

enum s_status agc_done (struct agc_context *agc)
{
	return agc -> ifc -> done (agc);
}

enum s_status agc_rgb (struct agc_context *agc, pdf_real_t r, 
		       pdf_real_t g, pdf_real_t b)
{
	return agc -> ifc -> rgb (agc, r, g, b, 1);
}

enum s_status agc_rgb_s (struct agc_context *agc, pdf_real_t r, 
			 pdf_real_t g, pdf_real_t b, pdf_bool_t stroking)
{
	return agc -> ifc -> rgb (agc, r, g, b, stroking);
}

enum s_status agc_ctm (struct agc_context *agc, struct cobj_matrix *mat)
{
	struct cobj_matrix xm;

	cobj_matrix_mul (mat == NULL ? &cobj_eye : mat, &agc -> bctm, &xm, 1);

	return agc -> ifc -> ctm (agc, &xm);
}

enum s_status agc_rect (struct agc_context *agc, 
			struct cobj_rect *rect, struct gs_paint *paint)
{
	return agc -> ifc -> rect (agc, rect, paint);
}

enum s_status agc_triangle (struct agc_context *agc,
			    struct gs_point *a,
			    struct gs_point *b,
			    struct gs_point *c,
			    struct gs_paint *paint)
{
	enum s_status res;

	res = agc_move_to (agc, a);
	if (res != S_CSC_OK)
		return res;

	res = agc_line_to (agc, b);
	if (res != S_CSC_OK)
		return res;

	res = agc_line_to (agc, c);
	if (res != S_CSC_OK)
		return res;

	res = agc_line_to (agc, a);
	if (res != S_CSC_OK)
		return res;

	res = agc_stroke (agc, paint);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

enum s_status agc_move_to (struct agc_context *agc,
			   struct gs_point *p)
{
	return agc -> ifc -> move_to (agc, p);
}

enum s_status agc_line_to (struct agc_context *agc,
			   struct gs_point *p)
{
	return agc -> ifc -> line_to (agc, p);
}

enum s_status agc_bezier_to (struct agc_context *agc,
			     struct gs_point *b,
			     struct gs_point *ca,
			     struct gs_point *cb)
{
	return agc -> ifc -> bezier_to (agc, b, ca, cb);
}

enum s_status agc_stroke (struct agc_context *agc,
			  struct gs_paint *paint)
{
	return agc -> ifc -> stroke (agc, paint);
}

enum s_status agc_draw_xobject (struct agc_context *agc,
				struct xo_object *xo)
{
	return agc -> ifc -> xo (agc, xo);
}

enum s_status agc_draw_shading (struct agc_context *agc,
				struct shd_shading *sh)
{
	return agc -> ifc -> shading (agc, sh);
}

enum s_status agc_draw_subpath (struct agc_context *agc,
				struct gs_subpath *sp,
				struct gs_paint *paint)
{
	struct gs_element *e;
	enum s_status res;
	int i;

	if (!sp -> elements)
		return S_CSC_OK;

	for (i = 0; i < sp -> elements; i++) {
		e = &sp -> e [i];
		switch (e -> type) {
		case GS_ET_LINE:
			if (!i) {
				res = agc_move_to (agc, &e -> u.line.a);
				if (res != S_CSC_OK)
					return res;
			}
			res = agc_line_to (agc, &e -> u.line.b);
			if (res != S_CSC_OK)
				return res;
			break;

		case GS_ET_BEZIER:
			if (!i) {
				res = agc_move_to (agc, &e -> u.bezier.a);
				if (res != S_CSC_OK)
					return res;
			}
			res = agc_bezier_to (agc, 
					     &e -> u.bezier.b, 
					     &e -> u.bezier.ca, 
					     &e -> u.bezier.cb);
			if (res != S_CSC_OK)
				return res;
			break;

		case GS_ET_TEXT:
			assert (0);
		}
	}

	res = agc_stroke (agc, paint);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

enum s_status agc_set_state (struct agc_context *agc,
			     struct gs_text_state *tstate,
			     struct gs_graph_state *gstate,
			     struct cobj_rect *cpath)
{
	struct cobj_matrix ctm;
	enum s_status res;

	if (gstate != NULL) {
		ctm = gstate -> ctm;
		cobj_matrix_mul (&ctm, &agc -> bctm, &gstate -> ctm, 1);
	}
	res = agc -> ifc -> set_state (agc, tstate, gstate, cpath);
	if (gstate != NULL)
		gstate -> ctm = ctm;
	
	return res;
}

enum s_status agc_write_char (struct agc_context *agc,
			      struct gs_glyph_info *gi,
			      struct gs_paint *paint)
{
	return agc -> ifc -> write_char (agc, gi, paint);
}

enum s_status agc_translate_like (struct agc_context *agc,
				  struct cobj_rect *frect,
				  struct cobj_rect *trect)
{
	struct cobj_matrix m;

	m = cobj_eye;
	m.a [0][0] = cobj_rect_width (trect) / cobj_rect_width (frect);
	m.a [1][1] = cobj_rect_height (trect) / cobj_rect_height (frect);
	m.a [2][0] = (trect -> left - frect -> left) * m.a [0][0];
	m.a [2][1] = (trect -> bottom - frect -> bottom) * m.a [1][1];

	agc -> bctm = m;

	return S_CSC_OK;
}

enum s_status agc_apply (struct agc_context *agc,
			 struct agc_context *pagc,
			 struct gs_point *point)
{
	static enum s_status res;
	struct cobj_matrix m;
	
	m = pagc -> bctm;
	
	pagc -> bctm.a [2][0] += point -> x;
	pagc -> bctm.a [2][1] += point -> y;

	res = agc -> ifc -> play (agc, pagc);
	pagc -> bctm = m;
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

enum s_status agc_measure (struct agc_context *agc, struct cobj_rect *rect)
{
	struct agc_context *mt;
	enum s_status res;
	struct ref_owner own;
	struct gs_point o;

	ref_init_owner (&own, "agc_measure");
	res = agc_measure_new (&mt, &own);
	if (res != S_CSC_OK)
		return res;

	o.x = o.y = 0;
	res = agc_apply (agc, mt, &o);
	if (res != S_CSC_OK) {
		agc_unref (mt, &own);
		return res;
	}

	rect -> left = 0;
	rect -> right = cobj_rect_width (&mt -> d.mt.rect);

	rect -> bottom = 0;
	rect -> top = cobj_rect_height (&mt -> d.mt.rect);

	agc_unref (mt, &own);

	return S_CSC_OK;
}

void agc_set_code (struct agc_context *agc,
		   agc_code_t code)
{
	agc -> current_code = code;
}

void agc_unset_code (struct agc_context *agc)
{
	agc -> current_code = 0;
}

void agc_enable_code (struct agc_context *agc,
		      agc_code_t code)
{
	assert (agc -> ifc -> set_enable != NULL);
	agc -> ifc -> set_enable (agc, code, 1);
}

void agc_disable_code (struct agc_context *agc,
		       agc_code_t code)
{
	assert (agc -> ifc -> set_enable != NULL);
	agc -> ifc -> set_enable (agc, code, 0);

}
