/*
 * PDFlow -- grdtree.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <assert.h>
#include <string.h>

#include "graph/grdtree.h"

#define BTREE_MIN_SIZE 15

static int compare (bt_key_t *key1, bt_key_t *key2);
static void destroy_data (bt_tree_data_t *tdata, bt_data_t *data);

static struct bt_param tree_param = {
	
	sizeof (struct t_tag),
	BTREE_MIN_SIZE,
	NULL,
	compare, NULL, NULL, NULL, destroy_data

};

static struct grd_tree *cast_tree_data (bt_tree_data_t *tdata)
{
	return (struct grd_tree *) tdata;
}

static bt_tree_data_t *uncast_tree_data (struct grd_tree *grd)
{
	return (struct bt_tree_data_t *) grd;
}

static struct t_tag *cast_key (bt_key_t *key)
{
	return (struct t_tag *) key;
}

static bt_key_t *uncast_key (struct t_tag *tag)
{
	return (struct bt_key_t *) tag;
}

static struct grd_data *cast_data (bt_key_t *data)
{
	return (struct grd_data *) data;
}

static bt_data_t *uncast_data (struct grd_data *data)
{
	return (struct bt_data_t *) data;
}

static int compare (bt_key_t *key1, bt_key_t *key2)
{
	struct t_tag *tag1, *tag2;
	int ans;

	tag1 = cast_key (key1);
	tag2 = cast_key (key2);

	ans = tag1 -> rid - tag2 -> rid;
	if (ans)
		return ans;

	ans = tag1 -> ref.oid - tag2 -> ref.oid;
	if (ans)
		return ans;

	ans = tag1 -> ref.gnum - tag2 -> ref.gnum;
	if (ans)
		return ans;

	return obj_string_compare_s (&tag1 -> name, &tag2 -> name);
}

static void destroy_data (bt_tree_data_t *tdata, bt_data_t *data)
{
	struct grd_tree *tree;
	struct grd_data *fdata;

	tree = cast_tree_data (tdata);
	fdata = cast_data (data);

	if (tree -> cc != NULL && fdata -> tcopy != NULL)
		t_cache_unref (fdata -> tcopy, ref_rown (&tree -> refs),
			       tree -> cc);

	obj_string_destroy (&fdata -> name);

	obj_destroy (&fdata -> obj);
	t_tag_destroy (&fdata -> tag);

	free (fdata);
}

static enum s_status init (struct grd_tree *grd, struct ob_bridge *ob)
{
	struct bt_param param;
	enum s_status res;

	grd -> ob = ob;
	ob_ref (ob, ref_rown (&grd -> refs));

	param = tree_param;
	param.tdata = uncast_tree_data (grd);

	res = bt_init (&grd -> tree, &param, ref_rown (&grd -> refs));
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

enum s_status grd_new (struct grd_tree **grdp, const char *prefix,
		       struct xref_db *rdb, struct ob_bridge *ob,
		       struct t_cache_closure *cc,
		       struct v_selection *sel, struct ref_owner *p)
{
	struct grd_tree *grd;
	enum s_status res;

	grd = malloc (sizeof (struct grd_tree));
	if (grd == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (grd, 0, sizeof (struct grd_tree));

	ref_init (&grd -> refs, "grd_tree", p);

	grd -> rdb = rdb;
	xref_ref_db (rdb, ref_rown (&grd -> refs));

	grd -> sel = sel;
	grd -> prefix = prefix;
	grd -> cc = cc;

	res = init (grd, ob);
	if (res != S_CSC_OK) {
		grd_unref (grd, p);
		return res;
	}

	*grdp = grd;

	return S_CSC_OK;
}

static enum s_status init_data (struct grd_tree *grd, struct grd_data *data,
				struct t_tag *tag, struct obj *obj,
				struct xref_db *rdb, struct obj *name,
				struct obj *xobj)
{
	char buf [2 + 1 + sizeof (pdf_integer_t) * 2 + 1];
	struct obj_bstr str;
	enum s_status res;

	res = S_CSC_OK;
	if (rdb == grd -> ob -> drdb) {
		if (t_tag_is_valid (tag))
			res = xref_serialize_obj (rdb, obj);
		if (res == S_CSC_OK)
			res = obj_dup (obj, &data -> obj);
	} else if (rdb == grd -> ob -> srdb) {
		assert (t_tag_is_valid (tag));
		res = ob_lookup (grd -> ob, obj, &data -> obj);
	} else
		assert (0);
	if (res != S_CSC_OK)
		return res;

	snprintf (buf, sizeof (buf), "%s%lx", grd -> prefix, grd -> seq);
	res = obj_string_from_c (buf, &data -> name);
	if (res != S_CSC_OK)
		return res;

	res = t_tag_dup (tag, &data -> tag);
	if (res != S_CSC_OK)
		return res;

	res = obj_string_dup (&data -> name, &str);
	if (res != S_CSC_OK)
		return res;

	obj_set_name (name, &f_null, &str);

	if (xobj != NULL) {
		res = obj_dup (&data -> obj, xobj);
		if (res != S_CSC_OK) {
			obj_destroy (name);
			return res;
		}
	}

	grd -> seq++;

	return S_CSC_OK;
}

enum s_status grd_lookup (struct grd_tree *grd, struct t_tag *tag, 
			  struct obj *obj, struct xref_db *rdb,
			  struct obj *name, struct obj *xobj,
			  struct t_type ***tcopy)
{
	struct grd_data *nd;
	struct obj_bstr str;
	enum s_status res;
	bt_data_t *found;

	res = bt_get (grd -> tree, uncast_key (tag), &found);
	if (res != S_CSC_OK)
		return res;

	if (found != NULL) {
		res = obj_string_dup (&cast_data (found) -> name, &str);
		if (res != S_CSC_OK)
			return res;

		obj_set_name (name, &f_null, &str);

		if (xobj != NULL) {
			res = obj_dup (&cast_data (found) -> obj, xobj);
			if (res != S_CSC_OK) {
				obj_destroy (name);
				return res;
			}
		}

		if (tcopy != NULL)
			*tcopy = &cast_data (found) -> tcopy;

		return S_CSC_OK;
	}

	nd = malloc (sizeof (struct grd_data));
	if (nd == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (nd, 0, sizeof (struct grd_data));
	
	res = init_data (grd, nd, tag, obj, rdb, name, xobj);
	if (res != S_CSC_OK) {
		destroy_data (uncast_tree_data (grd), 
			      uncast_data (nd));
		return res;
	}

	res = bt_insert (grd -> tree, uncast_key (&nd -> tag), 
			 uncast_data (nd), NULL);
	if (res != S_CSC_OK) {
		if (xobj != NULL)
			obj_destroy (xobj);
		obj_destroy (name);
		destroy_data (uncast_tree_data (grd), 
			      uncast_data (nd));
		return res;
	}

	if (tcopy != NULL)
		*tcopy = &nd -> tcopy;
	
	return S_CSC_OK;
}

static enum s_status rdict_f (bt_node_data_t *ndata, bt_data_t *d,
			      va_list ap)
{
	struct obj_dictionary *dict;
	struct grd_data *data;
	struct obj_bstr kstr;
	struct obj_dict_node dn;
	struct obj key, value;
	enum s_status res;

	if (d == NULL)
		return S_CSC_OK;
	
	data = cast_data (d);

	dict = va_arg (ap, struct obj_dictionary *);
	
	res = obj_string_dup (&data -> name, &kstr);
	if (res != S_CSC_OK)
		return res;

	obj_set_name (&key, &f_null, &kstr);

	res = obj_dup (&data -> obj, &value);
	if (res != S_CSC_OK) {
		obj_destroy (&key);
		return res;
	}

	obj_init_dict_node (&dn, &key, &value);
	res = obj_xfer_dict (dict, &dn);
	if (res != S_CSC_OK) {
		obj_destroy (&key);
		obj_destroy (&value);
		return res;
	}

	return S_CSC_OK;
}

enum s_status grd_rdict (struct grd_tree *tree, struct obj *dobj)
{
	struct obj_dictionary dict;
	enum s_status res;

	obj_init_dictionary (&dict);

	res = bt_traverse (tree -> tree, rdict_f, &dict);
	if (res != S_CSC_OK) {
		obj_dict_destroy (&dict);
		return res;
	}

	if (dict.head != NULL)
		obj_set_dictionary (dobj, &f_null, &dict);
	else
		obj_set_null (dobj, &f_null);

	return S_CSC_OK;
}

void grd_ref (struct grd_tree *grd, struct ref_owner *p)
{
	ref_up (&grd -> refs, p);
}

void grd_unref (struct grd_tree *grd, struct ref_owner *p)
{
	if (ref_down (&grd -> refs, p))
		return;

	if (grd -> tree != NULL)
		bt_unref (grd -> tree, ref_rown (&grd -> refs));

	ob_unref (grd -> ob, ref_rown (&grd -> refs));

	xref_unref_db (grd -> rdb, ref_rown (&grd -> refs));

	ref_zero (&grd -> refs);

	free (grd);
}

