/*
 * PDFlow -- gspg.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>
#include <stdlib.h>

#include "graph/gspg.h"

static const char *ii_sh_key [][2] = {
	{ "BitsPerComponent", "BPC" },
	{ "ColorSpace", "CS" },
	{ "Decode", "D" },
	{ "DecodeParms", "DP" },
	{ "Filter", "F" },
	{ "Height", "H" },
	{ "ImageMask", "IM" },
	{ "Interpolate", "I" },
	{ "Width", "W" },
	{ NULL }
};

static const char *ii_sh_val [][2] = {
	{ "DeviceGray", "G" },
	{ "DeviceRGB", "RGB" },
	{ "DeviceCMYK", "CMYK" },
	{ "Indexed", "I" },
	{ "ASCIIHexDecode", "AHx" },
	{ "ASCII85Decode", "A85" },
	{ "LZWDecode", "LZW" },
	{ "FlateDecode", "Fl" },
	{ "RunLengthDecode", "RL" },
	{ "CCITTFaxDecode", "CCF" },
	{ "DCTDecode", "DCT" },
	{ NULL }
};

enum s_status gspg_stmt_create (struct gspg_stmt **stmtp)
{
	struct gspg_stmt *stmt;

	stmt = malloc (sizeof (struct gspg_stmt));
	if (stmt == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (stmt, 0, sizeof (struct gspg_stmt));

	*stmtp = stmt;

	return S_CSC_OK;
}

void gspg_stmt_destroy (struct gspg_stmt *stmt, struct ref_owner *p)
{
	switch (stmt -> type) {
	case GSPG_ST_PATH:
		gs_path_unref (stmt -> u.path, p);
		break;

	case GSPG_ST_SHADING:
		shd_unref (stmt -> u.shading, p);
		break;

	case GSPG_ST_XOBJECT:
	case GSPG_ST_INLINE_IMAGE:
		xo_unref (stmt -> u.xobject, p);
	}

	gs_graph_unref (stmt -> gstate, p);
	gs_text_unref (stmt -> tstate, p);

	free (stmt);
}

enum s_status gspg_stmt_add (struct gspg_page *page,
			     struct gs_context *gc,
			     struct gspg_stmt **stmtp)
{
	enum s_status res;
	struct gspg_stmt *stmt;

	res = gspg_stmt_create (&stmt);
	if (res != S_CSC_OK)
		return res;

	stmt -> gstate = gs_snap_graph (gc, ref_rown (&page -> refs));
	stmt -> tstate = gs_snap_text (gc, ref_rown (&page -> refs));
	if (page -> tail != NULL)
		page -> tail -> next = stmt;
	else
		page -> head = stmt;

	page -> tail = stmt;

	if (stmtp != NULL)
		*stmtp = stmt;

	return S_CSC_OK;
}

static void destroy_stmt (struct gspg_page *pg, struct gspg_stmt *stmt)
{
	gspg_stmt_destroy (stmt, ref_rown (&pg -> refs));
}

enum s_status gspg_stmt_add_path (struct gspg_page *page, 
				  struct gs_context *gc,
				  struct gs_paint *paint,
				  pdf_bool_t no_cpath)
{
	struct gspg_stmt *stmt;
	enum s_status res;

	res = gspg_stmt_add (page, gc, &stmt);
	if (res != S_CSC_OK)
		return res;

	stmt -> no_cpath = no_cpath;
	page -> tail -> type = GSPG_ST_PATH;
	page -> tail -> u.path = gc -> path;
	page -> tail -> paint = *paint;
	gs_path_ref (gc -> path, ref_rown (&page -> refs));

	res = gs_path_reset (gc);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

enum s_status gspg_stmt_add_xobj (struct gspg_page *page, 
				  struct gs_context *gc,
				  struct xo_object *xo,
				  pdf_bool_t no_cpath)
{
	struct gspg_stmt *stmt;
	enum s_status res;

	res = gspg_stmt_add (page, gc, &stmt);
	if (res != S_CSC_OK)
		return res;

	stmt -> no_cpath = no_cpath;
	page -> tail -> type = GSPG_ST_XOBJECT;
	page -> tail -> u.xobject = xo;
	T_CALL (xo_ref, xo, ref_rown (&page -> refs));

	return S_CSC_OK;
}

enum s_status gspg_stmt_add_shading (struct gspg_page *page, 
				     struct gs_context *gc,
				     struct shd_shading *sh,
				     pdf_bool_t no_cpath)
{
	struct gspg_stmt *stmt;
	enum s_status res;

	res = gspg_stmt_add (page, gc, &stmt);
	if (res != S_CSC_OK)
		return res;

	stmt -> no_cpath = no_cpath;
	page -> tail -> type = GSPG_ST_SHADING;
	page -> tail -> u.shading = sh;
	T_CALL (shd_ref, sh, ref_rown (&page -> refs));

	return S_CSC_OK;
}

enum s_status gspg_page_new (struct gspg_page **pagep, 
			     struct cobj_rect *bounds,
			     pdf_integer_t rotate,
			     struct ref_owner *p)
{
	struct gspg_page *page;

	page = malloc (sizeof (struct gspg_page));
	if (page == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (page, 0, sizeof (struct gspg_page));

	if (bounds != NULL)
		page -> bounds = *bounds;
	else
		cobj_rect_set_null (&page -> bounds);

	page -> rotate = rotate;

	ref_init (&page -> refs, "pdf_page", p);

	*pagep = page;

	return S_CSC_OK;
}

void gspg_page_ref (struct gspg_page *page, struct ref_owner *p)
{
	ref_up (&page -> refs, p);
}

void gspg_page_unref (struct gspg_page *page, struct ref_owner *p)
{
	struct gspg_stmt *stmt;

	if (ref_down (&page -> refs, p))
		return;
	
	while (page -> head != NULL) {
		stmt = page -> head;
		page -> head = stmt -> next;

		destroy_stmt (page, stmt);
	}

	ref_zero (&page -> refs);

	free (page);
}

static enum s_status replace (const char *table [][2], 
			      struct obj_bstr *str, int apply)
{
	struct obj_bstr s;
	enum s_status res;
	int i, from, to;

	if (apply) {
		from = 0;
		to = 1;
	} else {
		from = 1;
		to = 0;
	}

	for (i = 0; table [i][from] != NULL; i++)
		if (strlen (table [i][from]) == str -> size &&
		    !strncmp (table [i][from], str -> s, str -> size))
			break;

	if (table [i][from] == NULL)
		return S_CSC_OK;

	res = obj_string_from_c (table [i][to], &s);
	if (res != S_CSC_OK)
		return res;

	obj_string_destroy (str);

	*str = s;

	return S_CSC_OK;
}

static enum s_status replace_fun (struct obj *obj, void *data)
{
	return obj -> type == OT_NAME ?
		replace (ii_sh_val, &obj -> val.string, *(int *) data) 
		: S_CSC_OK;
}

enum s_status gspg_expand_shorthands (struct obj_dictionary *dict)
{
	struct obj_dict_node *dn;
	struct obj_bstr value;
	struct obj vobj;
	enum s_status res;
	int apply;

	/* Add a "Subtype" entry */
	res = obj_string_from_c ("Image", &value);
	if (res != S_CSC_OK)
		return res;

	obj_set_name (&vobj, &f_null, &value);
	res = obj_add_dict (dict, "Subtype", &vobj);
	obj_destroy (&vobj);

	if (res != S_CSC_OK) 
		return res;

	apply = 0;
	for (dn = dict -> head; dn != NULL; dn = dn -> next) {
		res = replace (ii_sh_key, &dn -> key, apply);
		if (res != S_CSC_OK)
			return res;

		res = obj_navigate (&dn -> obj, replace_fun, &apply);
		if (res != S_CSC_OK)
			return res;
	}

	return S_CSC_OK;
}

enum s_status gspg_apply_shorthands (struct obj_dictionary *dict)
{
	struct obj_dict_node **dn, *node;
	enum s_status res;
	int apply;

	apply = 1;
	dn = &dict -> head;
	while (*dn != NULL) {
		if (obj_string_equal (&(*dn) -> key, "Type") ||
		    obj_string_equal (&(*dn) -> key, "Subtype")) {
			node = *dn;
			*dn = node -> next;
			node -> next = NULL;
			obj_dict_node_destroy (&node);
		} else {
			res = replace (ii_sh_key, &(*dn) -> key, apply);
			if (res != S_CSC_OK)
				return res;

			res = obj_navigate (&(*dn) -> obj, 
					    replace_fun, &apply);
			if (res != S_CSC_OK)
				return res;

			dn = &(*dn) -> next;
		}
	}

	return S_CSC_OK;
}

