/*
 * PDFlow -- gstate.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <assert.h>
#include <string.h>
#include <stdlib.h>

#include "fmt/fields.h"
#include "graph/gstate.h"

static enum s_status g_ini (struct gs_graph_state *gs,
			    struct xref_db *rdb,
			    struct xref_bind *robj,
			    struct ref_owner *p);
static enum s_status g_ini00 (struct gs_graph_state *gs,
			      struct xref_db *rdb,
			      struct xref_bind *robj,
			      struct ref_owner *p);
static void g_fini (struct gs_graph_state *gs, struct ref_owner *p);

static enum s_status tws_ini (struct gs_text_state *tws,
			      struct xref_db *rdb,
			      struct xref_bind *robj,
			      struct ref_owner *p);
static void tws_fini (struct gs_text_state *tws, struct ref_owner *p);

static enum s_status gc_ini (struct gs_context *gc,
			     struct xref_db *rdb,
			     struct xref_bind *robj,
			     struct ref_owner *p);
static void gc_fini (struct gs_context *gc, struct ref_owner *p);

static enum s_status text_set_key (struct gs_text_state *tws);
static enum s_status graph_set_key (struct gs_graph_state *gss);

void gsg_cleanup (void);

struct gs_paint gs_paint_stroke = { 1, 0, GS_FR_NONE };
struct gs_paint gs_paint_fill = { 0, 1, GS_FR_WINDING };
struct gs_paint gs_paint_stroke_fill = { 1, 1, GS_FR_WINDING };

enum gs_dict_key {

	GS_DK_OVERPRINT_STROKE,
	GS_DK_OVERPRINT_NONSTROKE,
	GS_DK_OVERPRINT_MODE,
	GS_DK_BLACK_GENERATION,
	GS_DK_UNDERCOLOR_REMOVAL,
	GS_DK_TRANSFER_FUNCTION,
	GS_DK_HALFTONE_DICTIONARY,
	GS_DK_SMOOTHNESS_TOLERANCE,
	GS_DK_STROKE_ADJUSTMENT,
	GS_DK_BLEND_MODE,
	GS_DK_SOFT_MASK,
	GS_DK_ALPHA_CONSTANT_STROKE,
	GS_DK_ALPHA_CONSTANT_NONSTROKE,
	GS_DK_ALPHA_IS_SHAPE,

	GS_DK_KNOCKOUT
};

#define KSDBIT(i) (((gs_keyset_t) 1) << key_desc [i].key)

static struct gs_key_desc {
	enum gs_dict_key key;
	const char *name;
	const char *dstring;
	struct obj kobj, dobj;
} key_desc [] = {
	{ GS_DK_OVERPRINT_STROKE, "OP" },
	{ GS_DK_OVERPRINT_NONSTROKE, "op" },
	{ GS_DK_OVERPRINT_MODE, "OPM" },
	{ GS_DK_BLACK_GENERATION, "BG2", "Default" },
	{ GS_DK_BLACK_GENERATION, "BG" },
	{ GS_DK_UNDERCOLOR_REMOVAL, "UCR2", "Default" },
	{ GS_DK_UNDERCOLOR_REMOVAL, "UCR" },
	{ GS_DK_TRANSFER_FUNCTION, "TR2", "Default" },
	{ GS_DK_TRANSFER_FUNCTION, "TR" },
	{ GS_DK_HALFTONE_DICTIONARY, "HT", "Default" },
	{ GS_DK_SMOOTHNESS_TOLERANCE, "SM" },
	{ GS_DK_STROKE_ADJUSTMENT, "SA" },
	{ GS_DK_BLEND_MODE, "BM", "Normal" },
	{ GS_DK_SOFT_MASK, "SMask", "None" },
	{ GS_DK_ALPHA_CONSTANT_STROKE, "CA" },
	{ GS_DK_ALPHA_CONSTANT_NONSTROKE, "ca" },
	{ GS_DK_ALPHA_IS_SHAPE, "AIS" },
	{ GS_DK_KNOCKOUT, "TK" },
	{ -1, NULL }
};

struct cobj_enum_desc gs_rendering_intent_desc [] = {
	{ GS_RI_ABSOLUTE_COLOMETRIC, "AbsoluteColorimetric" },
	{ GS_RI_RELATIVE_COLOMETRIC, "RelativeColorimetric" },
	{ GS_RI_SATURATION, "Saturation" },
	{ GS_RI_PERCEPTUAL, "Perceptual" },

	{ GS_RI_RELATIVE_COLOMETRIC, "/Default" },

	{ -1, NULL }
};

static struct cobj_enum_desc black_generation_desc [] = {
	{ GS_BG_EXPLICIT_DEFAULT, "Default", 1, 3 },

	{ GS_BG_FUNCTION, "/PSTable", 1, 0, &fun_pstable },

	{ GS_BG_DEFAULT, "/Default" },
	
	{ -1, NULL }
};

static struct cobj_enum_desc undercolor_removal_desc [] = {
	{ GS_UCR_EXPLICIT_DEFAULT, "Default", 1, 3 },

	{ GS_UCR_FUNCTION, "/PSTable", 1, 0, &fun_pstable },

	{ GS_UCR_DEFAULT, "/Default" },
	
	{ -1, NULL }
};

static struct cobj_enum_desc transfer_function_desc [] = {
	{ GS_TR_IDENTITY, "Identity", 1, 0 },
	{ GS_TR_EXPLICIT_DEFAULT, "Default", 1, 3 },

	{ GS_TR_FUNCTION, "/PSTable", 1, 0, &fun_pstable },

	{ GS_TR_DEFAULT, "/Default" },
	
	{ -1, NULL }
};

static struct cobj_enum_desc halftone_dictionary_desc [] = {
	{ GS_HT_EXPLICIT_DEFAULT, "Default", 1, 0 },

	{ GS_HT_FUNCTION, "/PSTable", 1, 0, &to_pstable },

	{ GS_HT_DEFAULT, "/Default" },

	{ -1, NULL }
};

static struct cobj_enum_desc blend_mode_desc [] = {

	{ GS_BM_NORMAL, "Normal" },
	{ GS_BM_COMPATIBLE, "Compatible" },
	{ GS_BM_MULTIPLY, "Multiply" },

	{ GS_BM_SCREEN, "Screen" },
	{ GS_BM_OVERLAY, "Overlay" },
	{ GS_BM_DARKEN, "Darken" },
	{ GS_BM_LIGHTEN, "Lighten" },
	{ GS_BM_COLOR_DODGE, "ColorDodge" },
	{ GS_BM_COLOR_BURN, "ColorBurn" },
	{ GS_BM_HARD_LIGHT, "HardLight" },
	{ GS_BM_SOFT_LIGHT, "SoftLight" },
	{ GS_BM_DIFFERENCE, "Difference" },
	{ GS_BM_EXCLUSION, "Exclusion" },
	
	{ GS_BM_HUE, "Hue" },
	{ GS_BM_SATURATION, "Saturation" },
	{ GS_BM_COLOR, "Color" },
	{ GS_BM_LUMINOSITY, "Luminosity" },

	{ GS_BM_NORMAL, "/Default" },

	{ -1, NULL }
};

static struct cobj_enum_desc soft_mask_desc [] = {
	{ GS_SM_NONE, "None", 1, 0 },
	
	{ GS_SM_DICTIONARY, "/PSTable", 1, 0, &to_pstable },

	{ GS_SM_DEFAULT, "/Default" },

	{ -1, NULL }
};

#define XREF_IMPLEMENTATION struct gs_dash_pattern
#define GS_DASH_PATTERN
#include "graph/gstate.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct gs_font
#define GS_FONT
#include "graph/gstate.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct gs_text_state
#define GS_TEXT_STATE
#include "graph/gstate.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct gs_graph_state
#define GS_GRAPH_STATE
#include "graph/gstate.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct gs_state
#define GS_GSTATE
#include "graph/gstate.tab.h"
#undef XREF_IMPLEMENTATION

T_IMPLEMENTATIONS_LOC0 (gs_graph, gs_graph_state, &graph_state_field,
			g_ini, NULL, g_ini00, g_fini);
T_IMPLEMENTATIONS_LOC (gs_text, gs_text_state, &text_state_field,
		       tws_ini, NULL, tws_fini);
T_IMPLEMENTATIONS (gs, gs_state, &gstate_field);
T_IMPLEMENTATIONS (gs_dash_pattern, gs_dash_pattern, &dash_pattern_field);
T_IMPLEMENTATIONS (gs_font, gs_font, &font_field);
T_IMPLEMENTATIONS_LOC (gs_context, gs_context, NULL, gc_ini, NULL, gc_fini);

int gs_point_same (struct gs_point *a, struct gs_point *b)
{
	return 	a -> x == b -> x &&
		a -> y == b -> y;
}

static void replace_key (const char *key, const char *key2, struct obj *dobj)
{
	struct obj_dict_node **dnp, *dn;

	for (dnp = &dobj -> val.dictionary.head; *dnp != NULL;
	     dnp = &(*dnp) -> next) 
		if (obj_string_equal (&(*dnp) -> key, key))
			break;

	if (*dnp == NULL)
		return;

	for (dn = dobj -> val.dictionary.head; dn != NULL; dn = dn -> next) 
		if (obj_string_equal (&dn -> key, key2))
			break;
	if (dn == NULL)
		return;

	dn = *dnp;
	*dnp = dn -> next;

	dn -> next = NULL;
	obj_dict_node_destroy (&dn);
}

static enum s_status g_ini00 (struct gs_graph_state *gss, struct xref_db *rdb,
			      struct xref_bind *robj, struct ref_owner *p)
{
	if (gss -> t.obj.type != OT_NULL) {
		replace_key ("BG", "BG2", &gss -> t.obj);
		replace_key ("UCR", "UCR2", &gss -> t.obj);
		replace_key ("TR", "TR2", &gss -> t.obj);
	}

	return S_CSC_OK;
}

static enum s_status g_ini (struct gs_graph_state *gss, struct xref_db *rdb,
			    struct xref_bind *robj, struct ref_owner *p)
{
	enum s_status res;

	gss -> ctm = cobj_eye;

	res = gc_cstate_init (&gss -> stroking_colour_state, rdb,
			      ref_rown (&gss -> t.refs));
	if (res != S_CSC_OK)
		return res;

	res = gc_cstate_init (&gss -> nonstroking_colour_state, rdb,
			      ref_rown (&gss -> t.refs));
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static void g_fini (struct gs_graph_state *gs, struct ref_owner *p)
{
	gs_unref_clipping_path (gs);

	gc_cstate_unref (gs -> stroking_colour_state,
			 ref_rown (&gs -> t.refs));
	gc_cstate_unref (gs -> nonstroking_colour_state,
			 ref_rown (&gs -> t.refs));

	gs_state_key_destroy (&gs -> key);
}

static enum s_status do_dup_graph (struct gs_graph_state *src,
				   struct gs_graph_state *dest)
{
	struct ref_owner *p;
	struct t_type tsave;
	int i;

	tsave = dest -> t;
	*dest = *src;
	dest -> t = tsave;

	dest -> frozen = 0;

	memset (&dest -> key, 0, sizeof (struct gs_state_key));

	p = ref_rown (&dest -> t.refs);

	if (dest -> clipping_path != NULL)
		ref_up (&dest -> clipping_path -> refs, p);

	gc_cstate_ref (dest -> stroking_colour_state, p);
	gc_cstate_ref (dest -> nonstroking_colour_state, p);
	T_CALL (gs_dash_pattern_ref, dest -> dash_pattern, p);
	T_CALL (fun_ref, dest -> black_generation.t, p);
	T_CALL (fun_ref, dest -> undercolor_removal.t, p);
	T_CALL (to_ref, dest -> soft_mask.t, p);
	T_CALL (to_ref, dest -> halftone_dictionary.t, p);
	
	/* Init objects that may be inadvertently destroyed when failing */
 	memset (&dest -> blend_mode, 0, sizeof (dest -> blend_mode));
	memset (&dest -> transfer_function, 0, 
		sizeof (dest -> transfer_function));

	if (XREF_A_SIZE (&src -> blend_mode)) {
		dest -> blend_mode.data = 
			calloc (XREF_A_SIZE (&src -> blend_mode),
				sizeof (struct cobj_enum));
		if (dest -> blend_mode.data == NULL)
			return S_CSC_OUT_OF_MEMORY;
		dest -> blend_mode.n = XREF_A_SIZE (&src -> blend_mode);
		memcpy (dest -> blend_mode.data,
			src -> blend_mode.data,
			XREF_A_SIZE (&dest -> blend_mode) *
			sizeof (struct cobj_enum));
	}		

	if (XREF_A_SIZE (&src -> transfer_function)) {
		dest -> transfer_function.data = 
			calloc (XREF_A_SIZE (&src -> transfer_function),
				sizeof (struct cobj_enum));
		if (dest -> transfer_function.data == NULL)
			return S_CSC_OUT_OF_MEMORY;
		dest -> transfer_function.n = 
			XREF_A_SIZE (&src -> transfer_function);
		for (i = 0; i < XREF_A_SIZE (&src -> transfer_function); i++) {
			dest -> transfer_function.data [i] =
				src -> transfer_function.data [i];
			T_CALL (fun_ref, 
				dest -> transfer_function.data [i].t, p);
		}
	}		

	return S_CSC_OK;
}

static enum s_status dup_graph (struct gs_graph_state *src,
				struct gs_graph_state **dest,
				struct ref_owner *p)
{
	struct gs_graph_state *gss;
	enum s_status res;
	
	res = T_CALL (gs_graph_new, &gss, src -> t.rdb, p);
	if (res != S_CSC_OK)
		return res;

	gc_cstate_unref (gss -> stroking_colour_state, 
			 ref_rown (&gss -> t.refs));
	gss -> stroking_colour_state = NULL;
	gc_cstate_unref (gss -> nonstroking_colour_state, 
			 ref_rown (&gss -> t.refs));
	gss -> nonstroking_colour_state = NULL;

	res = do_dup_graph (src, gss);
	if (res != S_CSC_OK) {
		gs_graph_unref (gss, p);
		return res;
	}

	*dest = gss;

	return S_CSC_OK;
}

enum s_status gs_set_ctm (struct gs_context *gpc, struct cobj_matrix *coef)
{
	struct gs_graph_state *gss;
	struct cobj_matrix m;
	enum s_status res;

	res = gs_graph (gpc, GS_GO_OP, &gss);
	if (res != S_CSC_OK)
		return res;
	
	cobj_matrix_mul (coef, &gss -> ctm, &m, 1);

	gss -> ctm = m;

	return S_CSC_OK;
}

enum s_status gs_add_clipping_path (struct gs_context *gpc,
				    struct gs_path *path,
				    enum gs_fill_rule rule)
{
	struct gs_graph_state *gss, *gstate;
	struct gs_text_state *tstate;
	struct gs_clipping_path *cpath;
	struct ref_owner owner;
	enum s_status res;

	/* This must be handled with care to avoid having a path
	 * that is one of its clipping paths: since we have a simple
	 * refcounter, objects won't be deallocated.  */
	ref_init_owner (&owner, "add_clipping_path");

	gstate = gs_snap_graph (gpc, &owner);
	tstate = gs_snap_text (gpc, &owner);

	res = gs_graph (gpc, GS_GO_CLIPPING_PATH, &gss);
	if (res != S_CSC_OK)
		return res;

	cpath = malloc (sizeof (struct gs_clipping_path));
	if (cpath == NULL)
		return S_CSC_OUT_OF_MEMORY;

	cpath -> next = gss -> clipping_path;
	ref_init (&cpath -> refs, "clipping_path", ref_rown (&gss -> t.refs));
	cpath -> path = path;
	cpath -> rule = rule;
	cpath -> gstate = gstate;
	cpath -> tstate = tstate;

	T_CALL (gs_graph_ref, gstate, ref_rown (&cpath -> refs));
	T_CALL (gs_text_ref, tstate, ref_rown (&cpath -> refs));

	gs_graph_unref (gstate, &owner);
	gs_text_unref (tstate, &owner);
	
	gs_path_ref (path, ref_rown (&cpath -> refs));
	if (cpath -> next != NULL) {
		ref_up (&cpath -> next -> refs, ref_rown (&cpath -> refs));
		ref_down (&cpath -> next -> refs, ref_rown (&gss -> t.refs));
	}

	gss -> clipping_path = cpath;

	return S_CSC_OK;
}

void gs_unref_clipping_path (struct gs_graph_state *gss)
{
	struct gs_clipping_path *cpath, *next;
	int again;
       
	cpath = gss -> clipping_path;
	if (cpath == NULL ||
	    ref_down (&cpath -> refs, ref_rown (&gss -> t.refs)))
		return;

	do {
		gs_path_unref (cpath -> path, ref_rown (&cpath -> refs));

		next = cpath -> next;
		again = next != NULL && 
			!ref_down (&next -> refs, ref_rown (&cpath -> refs));


		gs_graph_unref (cpath -> gstate, ref_rown (&cpath -> refs));
		gs_text_unref (cpath -> tstate, ref_rown (&cpath -> refs));

		ref_zero (&cpath -> refs);
		free (cpath);

		cpath = next;
	} while (again);
}

enum s_status gs_dash_pattern_init_alt (struct gs_dash_pattern **dp,
					struct xref_db *rdb,
					struct xref_bind *robj,
					pdf_integer_t phase,
					const struct obj *font,
					struct ref_owner *p)
{
	struct xref_obj_array ma;
	struct obj array;
	enum s_status res;

	ma.n = 2;
	ma.data = calloc (2, sizeof (struct obj));
	if (ma.data == NULL)
		return S_CSC_OUT_OF_MEMORY;

	ma.data [0] = *font;
	obj_set_integer (&ma.data [1], &f_null, phase);

	res = xref_array_unmap_ps (rdb, robj, NULL, &ma, &array, 0,
				   &xref_obj_pstable, NULL);
	free (ma.data);
	if (res != S_CSC_OK)
		return res;

	res = T_CALL (gs_dash_pattern_init, dp, rdb, robj, &array, p);
	obj_destroy (&array);

	return res;
}

int gs_dash_pattern_eq (struct gs_dash_pattern *dp1, 
			struct gs_dash_pattern *dp2)
{
	struct xref_real_array *da1, *da2;

	if (dp1 != NULL &&
	    !dp1 -> dash_phase && !XREF_A_SIZE (&dp1 -> dash_array))
		dp1 = NULL;
	if (dp2 != NULL &&
	    !dp2 -> dash_phase && !XREF_A_SIZE (&dp2 -> dash_array))
		dp2 = NULL;

	if (dp1 == dp2)
		return 1;

	if (dp1 == NULL || dp2 == NULL)
		return 0;

	da1 = &dp1 -> dash_array;
	da2 = &dp2 -> dash_array;

	if (dp1 -> dash_phase != dp2 -> dash_phase ||
	    XREF_A_SIZE (da1) != XREF_A_SIZE (da2))
		return 0;

	if (!XREF_A_SIZE (da1))
		return 1;

	return !memcmp (da1 -> data, da2 -> data,
			XREF_A_SIZE (da1) * sizeof (pdf_integer_t));
}

int gs_rendering_intent_eq (struct fun_enum *ri1, struct fun_enum *ri2)
{
	if (ri1 == ri2)
		return 1;

	if (ri1 == NULL || ri2 == NULL)
		return 0;

	return ri1 -> type == ri2 -> type;
}

static enum s_status tws_ini (struct gs_text_state *tws,
			      struct xref_db *rdb,
			      struct xref_bind *robj,
			      struct ref_owner *p)
{
	tws -> char_space = 0;
	tws -> word_space = 0;
	tws -> scale = 100;
	tws -> leading = 0;
	tws -> render = 0;
	tws -> rise = 0;

	tws -> mat.t_m = cobj_eye;
	tws -> mat.t_lm = cobj_eye;

	return S_CSC_OK;
}

static void tws_fini (struct gs_text_state *tws, struct ref_owner *p)
{
	gs_state_key_destroy (&tws -> key);
}

static enum s_status do_dup_text (struct gs_text_state *src,
				  struct gs_text_state *dest)
{
	struct t_type tsave;
	struct ref_owner *p;

	tsave = dest -> t;
	*dest = *src;
	dest -> t = tsave;

	dest -> frozen = 0;

	memset (&dest -> key, 0, sizeof (struct gs_state_key));

	p = ref_rown (&dest -> t.refs);

	T_CALL (gs_font_ref, dest -> font, p);

	return S_CSC_OK;
}

static enum s_status dup_text (struct gs_text_state *src,
			       struct gs_text_state **dest,
			       struct ref_owner *p)
{
	struct gs_text_state *tws;
	enum s_status res;

	res = T_CALL (gs_text_new, &tws, src -> t.rdb, p);
	if (res != S_CSC_OK)
		return res;

	res = do_dup_text (src, tws);
	if (res != S_CSC_OK) {
		gs_text_unref (tws, p);
		return res;
	}

	*dest = tws;

	return S_CSC_OK;
}

static enum s_status gs_dup (struct gs_context *gc, struct gs_state **destp,
			     struct ref_owner *p)
{
	struct gs_state *dest;
	enum s_status res;

	res = T_CALL (gs_new, &dest, gc -> t.rdb, p);
	if (res != S_CSC_OK)
		return res;
		
	gs_graph_unref (dest -> gstate, ref_rown (&dest -> t.refs));
	gs_text_unref (dest -> tstate, ref_rown (&dest -> t.refs));

	dest -> gstate = gs_snap_graph (gc, ref_rown (&dest -> t.refs));
	dest -> tstate = gs_snap_text (gc, ref_rown (&dest -> t.refs));

	*destp = dest;

	return S_CSC_OK;
}

static enum s_status gc_ini (struct gs_context *gpc,
			     struct xref_db *rdb,
			     struct xref_bind *robj,
			     struct ref_owner *p)
{
	enum s_status res;

	res = T_CALL (gs_new, &gpc -> state, rdb, ref_rown (&gpc -> t.refs));
	if (res != S_CSC_OK)
		return res;

	gpc -> path = NULL;
	gpc -> gss = NULL;
	gpc -> sb_path_closed = 1;

	res = gs_path_reset (gpc);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static void gc_fini (struct gs_context *gpc, struct ref_owner *p)
{
	int eos;

	/* Should not happen, so we can ignore error */
	do {
		gs_pop (gpc, &eos);
	} while (!eos);

	if (gpc -> path != NULL)
		gs_path_unref (gpc -> path, ref_rown (&gpc -> t.refs));
}

enum s_status gs_context_set_dash_pattern_alt (struct gs_context *gc,
					       pdf_integer_t phase,
					       const struct obj *font)
{
	struct gs_dash_pattern *dp;
	struct gs_graph_state *gss;
	enum s_status res;

	res = gs_graph (gc, GS_GO_OP, &gss);
	if (res != S_CSC_OK)
		return res;

	res = gs_dash_pattern_init_alt (&dp, gc -> t.rdb, NULL, phase,
					font, ref_rown (&gss -> t.refs));
	if (res != S_CSC_OK)
		return res;
	
	gs_dash_pattern_unref (gss -> dash_pattern, ref_rown (&gss -> t.refs));
	gss -> dash_pattern = dp;

	return S_CSC_OK;
}

enum s_status gs_context_set_rendering_intent (struct gs_context *gc,
					       struct obj *obj)
{
	struct gs_graph_state *gss;
	struct xref_field sfd;
	struct fun_enum ce;
	enum s_status res;

	res = gs_graph (gc, GS_GO_OP, &gss);
	if (res != S_CSC_OK)
		return res;

	xref_init_field (&sfd, &ce, gs_rendering_intent_desc);
	memset (&ce, 0, sizeof (struct fun_enum));
	res = cobj_enum_pstable.parser (gc -> t.rdb, NULL, 
					ref_rown (&gss -> t.refs), 
					obj, &sfd, __FILE__, __LINE__);
	if (res != S_CSC_OK)
		return res;

	xref_init_field (&sfd, &gss -> rendering_intent, 
			 gs_rendering_intent_desc);
	cobj_enum_pstable.funref (ref_rown (&gss -> t.refs), &sfd);
	gss  -> rendering_intent = ce;

	return S_CSC_OK;
}

static void replace_tags (struct gs_graph_state *gs,
			  struct gs_tags *tags)
{
	if (gs -> halftone_dictionary.type == GS_HT_FUNCTION) {
		t_tag_destroy (&gs -> halftone_dictionary.t -> t.tag);
		gs -> halftone_dictionary.t -> t.tag = tags -> halftone;
	} else
		t_tag_destroy (&tags -> halftone);

	if (gs -> halftone_dictionary.type == GS_SM_DICTIONARY) {
		t_tag_destroy (&gs -> soft_mask.t -> t.tag);
		gs -> soft_mask.t -> t.tag = tags -> soft_mask;
	} else
		t_tag_destroy (&tags -> soft_mask);
}

enum s_status gs_set_device_colour (struct gs_context *gc,
				    struct rd_resdict *rd,
				    struct gc_colour_state **cst,
				    enum gc_colour_space_family family)
{
	const char *name;
	enum s_status res;
	struct obj_bstr str;
	struct obj obj;

	switch (family) {
	case GC_CSF_DEVICE_GRAY:
		name = "DeviceGray";
		break;

	case GC_CSF_DEVICE_RGB:
		name = "DeviceRGB";
		break;

	case GC_CSF_DEVICE_CMYK:
		name = "DeviceRGB";
		break;
		
	default:
		assert (0);
	}

	res = obj_string_from_c (name, &str);
	if (res != S_CSC_OK)
		return res;

	obj_set_name (&obj, &f_null, &str);

	res = gc_set_colour_state (&obj, gc -> t.rdb, rd,
				   cst, gs_graph_owner (gc));

	obj_destroy (&obj);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

enum s_status gs_set_stroking_colour (struct gs_context *gc,
				      enum gc_colour_space_family family,
				      pdf_real_t *data)
{
	struct gs_graph_state *gss;
	enum s_status res;

	res = gs_graph (gc, GS_GO_COLOUR, &gss);
	if (res != S_CSC_OK)
		return res;

	res = gs_set_device_colour (gc, NULL, 
				    &gss -> stroking_colour_state, family);
	if (res != S_CSC_OK)
		return res;

	res = gc_select_colour (&gss -> stroking_colour_state, NULL, 
				data, gs_graph_owner (gc));
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

enum s_status gs_set_nonstroking_colour (struct gs_context *gc,
					 enum gc_colour_space_family family,
					 pdf_real_t *data)
{
	struct gs_graph_state *gss;
	enum s_status res;

	res = gs_graph (gc, GS_GO_COLOUR, &gss);
	if (res != S_CSC_OK)
		return res;

	res = gs_set_device_colour (gc, NULL, 
				    &gss -> nonstroking_colour_state, family);
	if (res != S_CSC_OK)
		return res;

	res = gc_select_colour (&gss -> nonstroking_colour_state, NULL, 
				data, gs_graph_owner (gc));
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

enum s_status gs_tags_get (struct gs_graph_state *gs, struct gs_tags *tags)
{
	enum s_status res;

	memset (tags, 0, sizeof (struct gs_tags));

	if (gs -> halftone_dictionary.type == GS_HT_FUNCTION) {
		res = t_tag_dup (&gs -> halftone_dictionary.t -> t.tag,
				 &tags -> halftone);
		if (res != S_CSC_OK)
			return res;
	} 

	if (gs -> halftone_dictionary.type == GS_SM_DICTIONARY) {
		res = t_tag_dup (&gs -> soft_mask.t -> t.tag,
				 &tags -> soft_mask);
		if (res != S_CSC_OK) {
			gs_tags_destroy (tags);
			return res;
		}
	}	

	return S_CSC_OK;
}

enum s_status gs_tags_dup (struct gs_tags *stags, struct gs_tags *dtags)
{
	enum s_status res;

	memset (dtags, 0, sizeof (struct gs_tags));
	res = t_tag_dup (&stags -> halftone, &dtags -> halftone);
	if (res != S_CSC_OK)	
		return res;

	res = t_tag_dup (&stags -> soft_mask, &dtags -> soft_mask);
	if (res != S_CSC_OK) {
		gs_tags_destroy (dtags);
		return res;
	}

	return S_CSC_OK;
}

void gs_tags_destroy (struct gs_tags *tags)
{
	t_tag_destroy (&tags -> halftone);
	t_tag_destroy (&tags -> soft_mask);
}

void gs_translate_vector (struct gs_point *point, struct cobj_matrix *mat)
{
	struct gs_point dpoint;

	dpoint.x =
		mat -> a [0][0] * point -> x +
		mat -> a [1][0] * point -> y;

	dpoint.y = 
		mat -> a [0][1] * point -> x +
		mat -> a [1][1] * point -> y;

	*point = dpoint;
}

void gs_translate_coords (struct gs_point *point, struct cobj_matrix *mat)
{
	struct gs_point dpoint;

	dpoint.x =
		mat -> a [0][0] * point -> x +
		mat -> a [1][0] * point -> y +
		mat -> a [2][0];

	dpoint.y = 
		mat -> a [0][1] * point -> x +
		mat -> a [1][1] * point -> y +
		mat -> a [2][1];

	*point = dpoint;
}

void gs_translate_rect (struct cobj_rect *rect, struct cobj_matrix *mat)
{
	cobj_translate (rect, mat);
}

void gs_revert_coords (struct gs_point *point, struct cobj_matrix *mat)
{
	struct cobj_matrix inv;

	cobj_matrix_inv (mat, &inv, 0);
	gs_translate_coords (point, &inv);
}

void gs_revert_rect (struct cobj_rect *rect, struct cobj_matrix *mat)
{
	struct cobj_matrix inv;

	cobj_matrix_inv (mat, &inv, 0);
	gs_translate_rect (rect, &inv);
}

static void transform_text_rect (struct cobj_rect *rect,
				 struct gs_text_state *tws,
				 struct cobj_matrix *tm)
{
	pdf_real_t hscale, vscale, grip;

	hscale = tws -> font -> size * tws -> scale / 100;
	vscale = tws -> font -> size;

	rect -> left *= hscale;
	rect -> top *= vscale;
	rect -> right *= hscale;
	rect -> bottom *= vscale;

	rect -> top += tws -> rise;
	rect -> bottom += tws -> rise;

	gs_translate_rect (rect, tm);
	if (rect -> left > rect -> right) {
		grip = rect -> left;
		rect -> left = rect -> right;
		rect -> right = grip;
	}
	if (rect -> bottom > rect -> top) {
		grip = rect -> bottom;
		rect -> bottom = rect -> top;
		rect -> top = grip;
	}
}

static void adv_chars (struct cobj_matrix *tm, const struct gs_text_state *tws,
		       pdf_real_t hval, pdf_real_t vval, pdf_real_t sp)
{
	struct gs_point point;

	if (tws -> font -> font -> wmode == CMAP_WM_HORIZONTAL) {
		point.x = (hval * tws -> font -> size + sp) * 
			tws -> scale / 100;
		point.y = 0;
	} else {
		point.x = 0;
		point.y = vval * tws -> font -> size + sp;
	}

	gs_translate_coords (&point, tm);
	tm -> a [2][0] = point.x;
	tm -> a [2][1] = point.y;
}

static pdf_bool_t is_parallel (pdf_real_t *vec,
			       const struct cobj_matrix *mat, int row,
			       pdf_real_t *ratiop)
{
	pdf_real_t ratio;
	int i;

	for (i = 0; i < 3; i++) {
		if (mat -> a [row][i])
			break;
		if (vec [i])
			return 0;
	} 

	if (i == 3) {
		*ratiop = 1;
		return 1;
	}

	ratio = vec [i] / mat -> a [row][i];
	while (++i < 3)
		if (vec [i] != mat -> a [row][i] * ratio)
			return 0;

	*ratiop = ratio;
	
	return 1;
}
			       

pdf_bool_t gs_get_kerning (struct gs_context *gpc, 
			   struct cobj_matrix *otm, struct cobj_matrix *ntm,
			   pdf_real_t *kerning)
{
	const struct gs_text_state *tws;
	pdf_real_t delta, dvec [3];
	int i, j;
	
	tws = gs_ro_text (gpc);

	for (i = 0; i < 2; i++)
		for (j = 0; j < 3; j++)
			if (otm -> a [i][j] != ntm -> a [i][j])
				return 0;

	for (j = 0; j < 3; j++)
		dvec [j] = otm -> a [2][j] - ntm -> a [2][j];

	if (tws -> font -> font -> wmode == CMAP_WM_HORIZONTAL) {
		if (!is_parallel (dvec, &tws -> mat.t_m, 0, &delta))
			return 0;

		*kerning = delta * tws -> scale / (100 * tws -> font -> size);
	} else {
		if (!is_parallel (dvec, &tws -> mat.t_m, 1, &delta))
			return 0;

	        delta = otm -> a [2][1] - ntm -> a [2][1];
		*kerning = delta / tws -> font -> size;
	}

	return 1;
}

static void map_chars (struct obj *text, struct cobj_matrix *tm,
		       struct gs_text_state *tws,
		       struct gs_glyph_info *glyph, int *idx)
{
	struct obj_bstr *str;
	struct gs_glyph_info gi, *gip;
	pdf_real_t sp;
	int i;
	
	str = &text -> val.string;

	i = 0;
	while (i < str -> size) {

		gip = glyph != NULL ? &glyph [*idx] : &gi;

		gip -> f = *gf_get_glyph_info (tws -> font -> font, 
					       (uint8_t *) str -> s + i, 
					       str -> size);
		gip -> t_m = *tm;
		if (glyph != NULL)
			transform_text_rect (&gip -> f.rect, tws, tm);

		sp = tws -> char_space;
		if (str -> s [i] == ' ')			
			sp += tws -> word_space;
		adv_chars (tm, tws, gip -> f.wx, gip -> f.wy, sp);

		gip -> c.size = gip -> f.size;
		/* This will lead to garbage .. */
		if (gip -> c.size > sizeof (gip -> c.bytes))
			gip -> c.size = sizeof (gip -> c.bytes);
		memcpy (gip -> c.bytes, str -> s + i, gip -> f.size);

		i += gip -> f.size;
		(*idx)++;
	}
}

void gs_get_bezier_rect (struct cobj_rect *rect, struct gs_bezier *bezier)
{
	if (bezier -> a.x < bezier -> b.x) {
		rect -> left = bezier -> a.x;
		rect -> right = bezier -> b.x;
	} else {
		rect -> left = bezier -> b.x;
		rect -> right = bezier -> a.x;
	}

	if (bezier -> a.y < bezier -> a.y) {
		rect -> bottom = bezier -> a.y;
		rect -> top = bezier -> b.y;
	} else {
		rect -> bottom = bezier -> b.y;
		rect -> top = bezier -> a.y;
	}
}

enum s_status gs_get_string (struct gs_text *text, 
			     struct gs_text_string *string,
			     struct cobj_matrix *matp)
{
	struct obj_array_node *an;
	struct gs_glyph_info *gi;
	struct cobj_matrix mat;
	pdf_real_t disp;
	int i, size;

	if (string != NULL) {
		switch (text -> text.type) {
		case OT_NULL:
			size = 0;
			break;

		case OT_STRING:
			size = text -> text.val.string.size;
			break;

		case OT_ARRAY:
			size = 0;
			for (an = text -> text.val.array.head; 
			     an != NULL; an = an -> next)
				if (an -> obj.type == OT_STRING)
					size += an -> obj.val.string.size;
			break;

		default:
			assert (0);
		}

		if (!size) {
			if (matp != NULL)
				*matp = mat;
			memset (string, 0, sizeof (struct gs_text_string));
			return S_CSC_OK;
		}

		gi = calloc (size, sizeof (struct gs_glyph_info));
		if (gi == NULL)
			return S_CSC_OUT_OF_MEMORY;

		string -> glyphs = gi;
		string -> elements = size;
	} else {
		size = 0;
		gi = NULL;
	}

	mat = text -> tstate -> mat.t_m;

	i = 0;
	switch (text -> text.type) {
	case OT_STRING:
		map_chars (&text -> text, &mat, text -> tstate, 
			   gi, &i);
		break;

	case OT_ARRAY:
		for (an = text -> text.val.array.head; 
		     an != NULL; an = an -> next) {
			disp = 0;
			if (an -> obj.type == OT_STRING)
				map_chars (&an -> obj, &mat, text -> tstate, 
					   gi, &i);
			else if (an -> obj.type == OT_REAL)
				disp = an -> obj.val.real;
			else if (an -> obj.type == OT_INTEGER)
				disp = an -> obj.val.integer;
			if (disp) {
				disp /= 1000;
				adv_chars (&mat, text -> tstate, 
					   -disp, -disp, 0);
			}
		}
		break;

	default:
		assert (0);
	}

	if (matp != NULL)
		*matp = mat;

	if (gi != NULL) {
		for (i = 0; i < size; i++)
			gs_translate_rect (&gi [i].f.rect, 
					   &text -> gstate -> ctm);
	}

	return S_CSC_OK;
}

void gs_get_vector (struct gs_text *text, pdf_real_t wx, pdf_real_t wy,
		    struct gs_point *vec)
{
	struct cobj_matrix tm;

	tm = text -> tstate -> mat.t_m;
	adv_chars (&tm, text -> tstate, wx, wy, text -> tstate -> char_space);
	vec -> x = tm.a [2][0] - text -> tstate -> mat.t_m.a [2][0];
	vec -> y = tm.a [2][1] - text -> tstate -> mat.t_m.a [2][1];
}

void gs_get_vector_gc (struct gs_context *gpc, pdf_real_t wx, pdf_real_t wy,
		       struct gs_point *vec)
{
	const struct gs_text_state *tws;
	struct cobj_matrix tm;

	tws = gs_ro_text (gpc);

	tm = tws -> mat.t_m;
	adv_chars (&tm, tws, wx, wy, tws -> char_space);
	vec -> x = tm.a [2][0] - tws -> mat.t_m.a [2][0];
	vec -> y = tm.a [2][1] - tws -> mat.t_m.a [2][1];
}

void gs_string_destroy (struct gs_text_string *string)
{
	if (string -> glyphs != NULL)
		free (string -> glyphs);

	memset (string, 0, sizeof (struct gs_text_string));
}


enum s_status gs_context_set_dictionary (struct gs_context *gc,
					 struct obj *dict,
					 struct gs_tags *tags)
{
	struct gs_graph_state *gss;
	struct gs_text_state *tws;
	enum s_status res;
	
	res = gs_graph (gc, GS_GO_DICT, &gss);
	if (res != S_CSC_OK)
		return res;

	res = gs_text (gc, GS_GO_DICT, &tws);
	if (res != S_CSC_OK)
		return res;

	res = xref_fields_replace (gc -> t.rdb, NULL, 
				   ref_rown (&gss -> t.refs),
				   dict, &graph_state_field, gss);
	if (res != S_CSC_OK)
		return res;

	res = xref_fields_replace (gc -> t.rdb, NULL, 
				   ref_rown (&tws -> t.refs),
				   dict, &text_state_field, tws);
	if (res != S_CSC_OK)
		return res;

	if (tags != NULL)
		replace_tags (gc -> state -> gstate, tags);

	return S_CSC_OK;
}

enum s_status gs_context_set_font (struct gs_context *gpc, 
				   struct gf_font *gf, pdf_real_t size)
{
	struct gs_text_state *tws;
	struct gs_font *font;
	enum s_status res;

	res = gs_text (gpc, GS_GO_OP, &tws);
	if (res != S_CSC_OK)
		return res;

	res = T_CALL (gs_font_new, &font, gpc -> t.rdb, 
		      ref_rown (&tws -> t.refs));
	if (res != S_CSC_OK)
		return res;

	font -> size = size;
	font -> font = gf;
	T_CALL (gf_ref, gf, ref_rown (&font -> t.refs));

	gs_font_unref (tws -> font, ref_rown (&tws -> t.refs));
	tws -> font = font;

	return S_CSC_OK;
}

struct ref_owner *gs_text_owner (struct gs_context *gc)
{
	return ref_rown (&gc -> state -> tstate -> t.refs);
}

struct ref_owner *gs_graph_owner (struct gs_context *gc)
{
	return ref_rown (&gc -> state -> gstate -> t.refs);
}

enum s_status gs_state_key_dup (struct gs_state_key *skey, 
				struct gs_state_key *dkey)
{
	enum s_status res;

	memset (dkey, 0, sizeof (struct gs_state_key));
	res = obj_string_dup (&skey -> opkey, &dkey -> opkey);
	if (res != S_CSC_OK) {
		gs_state_key_destroy (dkey);
		return res;
	}

	res = obj_string_dup (&skey -> dkey, &dkey -> dkey);
	if (res != S_CSC_OK) {
		gs_state_key_destroy (dkey);
		return res;
	}

	return S_CSC_OK;
}

static void reset_key (struct gs_state_key *key, enum gs_state_op op)
{
	switch (op) {
	case GS_GO_OP:
		obj_string_destroy (&key -> opkey);
		break;

	case GS_GO_DICT:
		obj_string_destroy (&key -> dkey);
		break;

	case GS_GO_COLOUR:
	case GS_GO_CLIPPING_PATH:
		break;
	}
}


void gs_state_key_destroy (struct gs_state_key *key)
{
	obj_string_destroy (&key -> opkey);
	obj_string_destroy (&key -> dkey);
}

int gs_state_key_compare (const struct gs_state_key *key1,
			  const struct gs_state_key *key2)
{
	int ans;

	ans = obj_string_compare_s (&key1 -> opkey, &key2 -> opkey);
	if (ans)
		return ans;

	ans = obj_string_compare_s (&key1 -> dkey, &key2 -> dkey);
	if (ans)
		return ans;

	return 0;
}

enum s_status gs_state_dkey_dup (struct gs_state_key *skey, 
				 struct gs_state_key *dkey)
{
	enum s_status res;

	res = obj_string_dup (&skey -> dkey, &dkey -> dkey);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

int gs_state_dkey_compare (struct gs_state_key *key1,
			   struct gs_state_key *key2)
{
	int ans;

	ans = obj_string_compare_s (&key1 -> dkey, &key2 -> dkey);
	if (ans)
		return ans;

	return 0;
}

#define SUBPATH_INC 5
static enum s_status grow_subpath (struct gs_subpath *gss)
{
	struct gs_element *e;
	int ncap;

	ncap = gss -> cap + SUBPATH_INC;
	e = realloc (gss -> e, ncap * sizeof (struct gs_element));
	if (e == NULL)
		return S_CSC_OUT_OF_MEMORY;

	gss -> e = e;
	gss -> cap = ncap;

	return S_CSC_OK;
}

enum s_status gs_push (struct gs_context *gpc)
{
	struct gs_state *p;
	enum s_status res;

	res = gs_dup (gpc, &p, ref_rown (&gpc -> t.refs));
	if (res != S_CSC_OK)
		return res;

	p -> next = gpc -> state;
	gpc -> state = p;

	return S_CSC_OK;
}

enum s_status gs_pop (struct gs_context *gpc, int *eos)
{
	struct gs_state *gs;
	enum s_status res;

	if (gpc -> state == NULL) {
		if (eos != NULL)
			*eos = 1;
		return S_CSC_OK;
	}

	gs = gpc -> state -> next;

	if (gpc -> state -> gstate != NULL) {
		res = graph_set_key (gpc -> state -> gstate);
		if (res != S_CSC_OK)
			return res;
	}

	if (gpc -> state -> tstate != NULL) {
		res = text_set_key (gpc -> state -> tstate);
		if (res != S_CSC_OK)
			return res;
	}

	gs_unref (gpc -> state, ref_rown (&gpc -> t.refs));

	gpc -> state = gs;

	if (eos != NULL)
		*eos = 0;

	return S_CSC_OK;
}

enum s_status gs_done (struct gs_context *gc)
{
	enum s_status res;

	if (gc -> state -> gstate -> frozen) {
		res = graph_set_key (gc -> state -> gstate); 
		if (res != S_CSC_OK)
			return res;
	}

	if (gc -> state -> tstate -> frozen) {
		res = text_set_key (gc -> state -> tstate); 
		if (res != S_CSC_OK)
			return res;
	}

	return S_CSC_OK;
}

static enum s_status add_element (struct gs_subpath *gss,
				  struct gs_element **e)
{
	enum s_status res;

	if (gss -> elements == gss -> cap) {
		res = grow_subpath (gss);
		if (res != S_CSC_OK)
			return res;
	}

	*e = &gss -> e [gss -> elements++];

	return S_CSC_OK;
}

static enum s_status init_subpath (struct gs_subpath **gssp)
{
	struct gs_subpath *gss;
	enum s_status res;

 	gss = malloc (sizeof (struct gs_subpath));
	if (gss == NULL)
		return S_CSC_OUT_OF_MEMORY;
		
	memset (gss, 0, sizeof (struct gs_subpath));
	res = grow_subpath (gss);
	if (res != S_CSC_OK) {
		free (gss);
		return res;
	}
	
	*gssp = gss;
	
	return S_CSC_OK;
}

void gs_subpath_destroy (struct gs_subpath *gss, struct ref_owner *p)
{
	int i;

	if (gss -> e != NULL) {
		for (i = 0; i < gss -> elements; i++) {
			switch (gss -> e [i].type) {
			case GS_ET_LINE:
			case GS_ET_BEZIER:
				break;

			case GS_ET_TEXT:
				gs_text_unref (gss -> e [i].u.text.tstate, p);
				gs_graph_unref (gss -> e [i].u.text.gstate, p);
				obj_destroy (&gss -> e [i].u.text.text);
			}
 
		}
		free (gss -> e);
	}
}

enum s_status gs_subpath_copy (struct gs_subpath *dsp, struct gs_subpath *ssp, 
			       struct ref_owner *p)
{
	struct gs_element *se, *de;
	enum s_status res;

	dsp -> next = NULL;
	dsp -> cap = ssp -> cap;
	dsp -> e = calloc (dsp -> cap, sizeof (struct gs_element));
	if (dsp -> e == NULL)
		return S_CSC_OUT_OF_MEMORY;

	for (dsp -> elements = 0; dsp -> elements < ssp -> elements;
	     dsp -> elements++) {
		se = &ssp -> e [dsp -> elements];
		de = &dsp -> e [dsp -> elements];
		*de = *se;

		switch (de -> type) {
		case GS_ET_LINE:
		case GS_ET_BEZIER:
			break;

		case GS_ET_TEXT:
			gs_text_unref (de -> u.text.tstate, p);
			gs_graph_unref (de -> u.text.gstate, p);
			res = obj_dup (&de -> u.text.text, &se -> u.text.text);
			if (res != S_CSC_OK) {
				gs_subpath_destroy (dsp, p);
				return res;
			}
		}
	}

	return S_CSC_OK;
}

static void destroy_subpath (struct gs_path *gp, struct gs_subpath *gss)
{
	gs_subpath_destroy (gss, ref_rown (&gp -> refs));
}

enum s_status gs_path_reset (struct gs_context *gpc)
{
	struct gs_path *path;

	path = malloc (sizeof (struct gs_path));
	if (path == NULL)
		return S_CSC_OUT_OF_MEMORY;

	memset (path, 0, sizeof (struct gs_path));
	ref_init (&path -> refs, "path", ref_rown (&gpc -> t.refs));

	if (gpc -> path != NULL)
		gs_path_unref (gpc -> path, ref_rown (&gpc -> t.refs));

	gpc -> path = path;
	gpc -> gss = NULL;

	return S_CSC_OK;
}

void gs_path_ref (struct gs_path *gp, struct ref_owner *p)
{
	ref_up (&gp -> refs, p);
}

void gs_path_unref (struct gs_path *gp, struct ref_owner *p)
{
	struct gs_subpath *gss;

	if (ref_down (&gp -> refs, p))
		return;

	while (gp -> head != NULL) {
		gss = gp -> head;
		gp -> head = gss -> next;
		destroy_subpath (gp, gss);
		free (gss);
	}

	ref_zero (&gp -> refs);

	free (gp);
}

enum s_status gs_move (struct gs_context *gpc, struct obj *ref,
			pdf_real_t x, pdf_real_t y)
{

	struct gs_subpath *gss;
	enum s_status res;

	if (gpc -> gss == NULL || gpc -> gss -> elements) {

		res = init_subpath (&gss);
		if (res != S_CSC_OK)
			return res;

		if (gpc -> gss != NULL)
			gpc -> gss -> next = gss;
		else
			gpc -> path -> head = gss;
		gpc -> gss = gss;
	}

	gpc -> curr_point.x = x;
	gpc -> curr_point.y = y;

	gpc -> start_point = gpc -> curr_point;

	return S_CSC_OK;
}

enum s_status gs_lineto (struct gs_context *gpc, struct obj *ref,
			  pdf_real_t x, pdf_real_t y)
{
	struct gs_element *e;
	enum s_status res;

	if (gpc -> gss == NULL)
		return obj_error (ref, "Path is not open");

	if (gpc -> sb_path_closed) {
		res = gs_move (gpc, ref, 
			       gpc -> curr_point.x, gpc -> curr_point.y);
		if (res != S_CSC_OK)
			return res;
		gpc -> sb_path_closed = 0;
	}

	res = add_element (gpc -> gss, &e);
	if (res != S_CSC_OK)
		return res;

	e -> type = GS_ET_LINE;
	e -> u.line.a = gpc -> curr_point;

	gpc -> curr_point.x = x;
	gpc -> curr_point.y = y;
	e -> u.line.b = gpc -> curr_point;

	return S_CSC_OK;
}

enum s_status gs_bezierto (struct gs_context *gpc, struct obj *ref,
			    pdf_real_t cx1, pdf_real_t cy1,
			    pdf_real_t cx2, pdf_real_t cy2,
			    pdf_real_t x, pdf_real_t y)
{
	struct gs_element *e;
	enum s_status res;

	if (gpc -> gss == NULL)
		return obj_error (ref, "Path is not open");

	if (gpc -> sb_path_closed) {
		res = gs_move (gpc, ref, 
				gpc -> curr_point.x, gpc -> curr_point.y);
		if (res != S_CSC_OK)
			return res;
		gpc -> sb_path_closed = 0;
	}

	res = add_element (gpc -> gss, &e);
	if (res != S_CSC_OK)
		return res;

	e -> type = GS_ET_BEZIER;
	e -> u.bezier.a = gpc -> curr_point;

	gpc -> curr_point.x = x;
	gpc -> curr_point.y = y;
	e -> u.bezier.b = gpc -> curr_point;

	e -> u.bezier.ca.x = cx1;
	e -> u.bezier.ca.y = cy1;

	e -> u.bezier.cb.x = cx2;
	e -> u.bezier.cb.y = cy2;

	return S_CSC_OK;
}

enum s_status gs_close (struct gs_context *gpc, struct obj *ref)
{
	enum s_status res;

	if (gpc -> gss == NULL)
		return obj_error (ref, "Path is not open");

	if (gpc -> sb_path_closed)
		return S_CSC_OK;

	if (gpc -> gss -> elements) {
		res = gs_lineto (gpc, ref, 
				 gpc -> start_point.x, gpc -> start_point.y);
		if (res != S_CSC_OK)
			return res;
	}
	
	gpc -> sb_path_closed = 1;

	return S_CSC_OK;
}

enum s_status gs_rect (struct gs_context *gpc, struct obj *ref,
		       struct cobj_rect *rect)
{
	enum s_status res;

	res = gs_move (gpc, ref, rect -> left, rect -> top);
	if (res != S_CSC_OK)
		return res;

	res = gs_lineto (gpc, ref, rect -> right, rect -> top);
	if (res != S_CSC_OK)
		return res;

	res = gs_lineto (gpc, ref, rect -> right, rect -> bottom);
	if (res != S_CSC_OK)
		return res;

	res = gs_lineto (gpc, ref, rect -> left, rect -> bottom);
	if (res != S_CSC_OK)
		return res;

	res = gs_close (gpc, ref);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status build_topkey (struct gs_text_state *tws,
				   struct obj_bstr *s)
{
	enum s_status res;
	struct hash hash;
	pdf_bool_t bool;
	size_t i, size;
	char *p;

	size = 	sizeof (tws -> char_space) +
		sizeof (tws -> word_space) +
		sizeof (tws -> scale) +
		sizeof (tws -> leading) +
		sizeof (tws -> rise);

	if (tws -> font != NULL) {
		size += sizeof (pdf_bool_t) +
			sizeof (struct hash) +
			sizeof (tws -> font -> size);
	} else
		size += sizeof (pdf_bool_t);

#ifdef DEBUG
	size += 6 * 4;
#endif
	
	p = malloc (size);
	if (p == NULL)
		return S_CSC_OUT_OF_MEMORY;

	i = 0;
#define ENCODE(field)\
	do { memcpy (p + i, &field, sizeof (field)); \
	i += sizeof (field); } while (0)
#ifdef DEBUG
#define TAG(name)\
	do { memcpy (p + i, name "=", 4); i += 4; } while (0)
#else
#define TAG(name)
#endif

	TAG ("chs");
	ENCODE (tws -> char_space);

	TAG ("wds");
	ENCODE (tws -> word_space);

	TAG ("scl");
	ENCODE (tws -> scale);

	TAG ("ldi");
	ENCODE (tws -> leading);

	TAG ("ris");
	ENCODE (tws -> rise);

	TAG ("fnt");
	if (tws -> font != NULL) {
		bool = 1;
		ENCODE (bool);
		res = t_tag_get_checksum (&tws -> font -> font -> t.tag,
					  &hash);
		if (res != S_CSC_OK) {
			free (p);
			return res;
		}
		ENCODE (hash);
		ENCODE (tws -> font -> size);
	} else {
		bool = 0;
		ENCODE (bool);
	}
		
#undef ENCODE
#undef TAG

	assert (i == size);

	s -> size = size;
	s -> s = p;

	return S_CSC_OK;	
}

static enum s_status build_tdkey (struct gs_text_state *tws,
				  struct obj_bstr *s)
{
	size_t i, size;
	char *p;

	size = 	sizeof (tws -> knockout);

#ifdef DEBUG
	size += 1 * 4;
#endif

	p = malloc (size);
	if (p == NULL)
		return S_CSC_OUT_OF_MEMORY;

	i = 0;
#define ENCODE(field)\
	do { memcpy (p + i, &field, sizeof (field)); \
	i += sizeof (field); } while (0)
#ifdef DEBUG
#define TAG(name)\
	do { memcpy (p + i, name "=", 4); i += 4; } while (0)
#else
#define TAG(name)
#endif
	TAG ("kno");
	ENCODE (tws -> knockout);
		
#undef ENCODE
#undef TAG

	assert (i == size);

	s -> size = size;
	s -> s = p;

	return S_CSC_OK;	
}

static enum s_status text_set_key (struct gs_text_state *tws)
{
	struct gs_state_key *key;
	enum s_status res;

	key = &tws -> key;
	if (key -> opkey.s == NULL) {
		res = build_topkey (tws, &key -> opkey);
		if (res != S_CSC_OK)
			return res;
	}

	if (key -> dkey.s == NULL) {
		res = build_tdkey (tws, &key -> dkey);
		if (res != S_CSC_OK)
			return res;
	}

	return S_CSC_OK;
}

pdf_bool_t gs_text_equal (const struct gs_text_state *tws1,
			  const struct gs_text_state *tws2)
{
	return !gs_state_key_compare (&tws1 -> key, &tws2 -> key);
}

enum s_status gs_text_update_key (struct gs_text_state *tws, 
				  enum gs_state_op op)
{
	reset_key (&tws -> key, op);

	return text_set_key (tws);
}

#ifdef DEBUG
void gs_text_check_key (struct gs_text_state *tws)
{
	struct gs_state_key key;
	enum s_status res;

	key = tws -> key;

	memset (&tws -> key, 0, sizeof (struct gs_state_key)); 
	res = text_set_key (tws);
	if (res != S_CSC_OK)
		return;

	assert (!gs_state_key_compare (&tws -> key, &key));

	reset_key (&key, GS_GO_OP);
	reset_key (&key, GS_GO_DICT);
}
#endif

enum s_status gs_text (struct gs_context *gpc, enum gs_state_op op,
		       struct gs_text_state **twsp)
{
	struct gs_text_state *tws;
	enum s_status res;

	if (!gpc -> state -> tstate -> frozen) {
		*twsp = gpc -> state -> tstate;
		return S_CSC_OK;
	}
       
	res = dup_text (gpc -> state -> tstate, &tws,
			ref_rown (&gpc -> state -> t.refs));
	if (res != S_CSC_OK)
		return res;

	res = text_set_key (gpc -> state -> tstate);
	if (res != S_CSC_OK) {
		gs_text_unref (tws, ref_rown (&gpc -> state -> t.refs));
		return res;
	}

	gs_text_unref (gpc -> state -> tstate, 
		       ref_rown (&gpc -> state -> t.refs));

	gpc -> state -> tstate = *twsp = tws;

	return S_CSC_OK;
}

const struct gs_text_state *gs_ro_text (struct gs_context *gpc)
{
	return gpc -> state -> tstate;
}

struct gs_text_state *gs_snap_text (struct gs_context *gpc, 
				    struct ref_owner *p)
{
	struct gs_text_state *tws;

	tws = gpc -> state -> tstate;

	tws -> frozen = 1;

	T_CALL (gs_text_ref, tws, p);

	return tws;
}

enum s_status gs_set_tstate (struct gs_context *gpc,
			     struct gs_text_state *tws)
{
	enum s_status res;

	res = text_set_key (gpc -> state -> tstate);
	if (res != S_CSC_OK)
		return res;

	gs_text_unref (gpc -> state -> tstate, 
		       ref_rown (&gpc -> state -> t.refs));
	gpc -> state -> tstate = tws;
	T_CALL (gs_text_ref, gpc -> state -> tstate,
		ref_rown (&gpc -> state -> t.refs));

	tws -> frozen = 0;

	return S_CSC_OK;
}

enum s_status gs_import_tstate (struct gs_text_state **twsp,
				struct ref_owner *p)
{
	struct gs_text_state *tws;
	enum s_status res;

	res = dup_text (*twsp, &tws, p);
	if (res != S_CSC_OK)
		return res;

	res = text_set_key (tws);
	if (res != S_CSC_OK) {
		gs_text_unref (tws, p);
		return res;
	}

	tws -> frozen = 0;
	*twsp = tws;

	return S_CSC_OK;
}

static enum s_status build_gopkey (struct gs_graph_state *gs,
				   struct obj_bstr *s)
{
	pdf_integer_t zero;
	size_t i, size, n;
	char *p;
	int j;

	size = 	sizeof (gs -> ctm) +
		sizeof (gs -> line_width) +
		sizeof (gs -> line_cap) +
		sizeof (gs -> miter_limit) +
		sizeof (gs -> rendering_intent.type) +
		sizeof (gs -> flatness_tolerance);
	if (gs -> dash_pattern != NULL)
		size += sizeof (gs -> dash_pattern -> dash_phase) +
			sizeof (gs -> dash_pattern -> dash_array.n) +
			XREF_A_SIZE (&gs -> dash_pattern -> dash_array) *
			sizeof (gs -> dash_pattern -> dash_array.data [0]);
	else
		size += sizeof (gs -> dash_pattern -> dash_phase) +
			sizeof (gs -> dash_pattern -> dash_array.n);

#ifdef DEBUG
	size += 7 * 4;
#endif

	p = malloc (size);
	if (p == NULL)
		return S_CSC_OUT_OF_MEMORY;

	i = 0;
#define ENCODE(field)\
	do { memcpy (p + i, &field, sizeof (field)); \
	i += sizeof (field); } while (0)
#ifdef DEBUG
#define TAG(name)\
	do { memcpy (p + i, name "=", 4); i += 4; } while (0)
#else
#define TAG(name)
#endif

	TAG ("ctm");
	ENCODE (gs -> ctm);
	TAG ("lwd");
	ENCODE (gs -> line_width);
	TAG ("lcp");
	ENCODE (gs -> line_cap);
	TAG ("mli");
	ENCODE (gs -> miter_limit);
	TAG ("dpt");
	if (gs -> dash_pattern) {
		ENCODE (gs -> dash_pattern -> dash_phase);
		ENCODE (gs -> dash_pattern -> dash_array.n);
		n = XREF_A_SIZE (&gs -> dash_pattern -> dash_array);
		for (j = 0; j < n; j++)
			ENCODE (gs -> dash_pattern -> dash_array.data [j]);
	} else {
		zero = 0;
		ENCODE (zero);
		ENCODE (zero);
	}

	TAG ("rin");
	ENCODE (gs -> rendering_intent.type);
	TAG ("flt");
	ENCODE (gs -> flatness_tolerance);
#undef ENCODE
#undef TAG

	s -> size = size;
	s -> s = p;

	assert (i == size);

	return S_CSC_OK;
}

static enum s_status build_gdkey (struct gs_graph_state *gs,
				  struct obj_bstr *s)
{
	enum s_status res;
	pdf_integer_t n;
	size_t i, size;
	struct hash hash;
	int def;
	char *p;
	int j;

	size = 	sizeof (gs -> overprint_stroke) +
		sizeof (gs -> overprint_nonstroke) +
		sizeof (gs -> overprint_mode) +
		sizeof (gs -> smoothness_tolerance) +
		sizeof (gs -> stroke_adjustment) +
		sizeof (gs -> alpha_constant_stroke) +
		sizeof (gs -> alpha_constant_nonstroke) +
		sizeof (gs -> alpha_is_shape);

	size += sizeof (gs -> black_generation.type);
	if (gs -> black_generation.type == GS_BG_FUNCTION)
		size += sizeof (struct hash);

	size += sizeof (gs -> undercolor_removal.type);
	if (gs -> undercolor_removal.type == GS_UCR_FUNCTION)
		size += sizeof (struct hash);

	size += sizeof (gs -> transfer_function.n);
	n = XREF_A_SIZE (&gs -> transfer_function);
	if (n == 1 && 
	    gs -> transfer_function.data [0].type == GS_TR_EXPLICIT_DEFAULT)
		n = 0;
	for (j = 0; j < n; j++) {
		size += sizeof (gs -> transfer_function.data [j].type);
		if (gs -> transfer_function.data [j].type == GS_TR_FUNCTION)
			size += sizeof (struct hash);
	}

	size += sizeof (gs -> halftone_dictionary.type);
	if (gs -> halftone_dictionary.type == GS_HT_FUNCTION)
		size += sizeof (struct hash);

	size += sizeof (gs -> blend_mode.n);
	n = XREF_A_SIZE (&gs -> blend_mode);
	if (n == 1 && gs -> blend_mode.data [0].type == GS_BM_NORMAL)
		n = 0;
	size += sizeof (gs -> blend_mode.data [j].type) * n;

	size += sizeof (gs -> soft_mask.type);
	if (gs -> soft_mask.type == GS_SM_DICTIONARY)
		size += sizeof (struct hash);
	
#ifdef DEBUG
	size += 14 * 4;
#endif
	p = malloc (size);
	if (p == NULL)
		return S_CSC_OUT_OF_MEMORY;

	i = 0;
#define ENCODE(field)\
	do { memcpy (p + i, &field, sizeof (field)); \
	i += sizeof (field); } while (0)
#ifdef DEBUG
#define TAG(name)\
	do { memcpy (p + i, name "=", 4); i += 4; } while (0)
#else
#define TAG(name)
#endif

	TAG ("ops");
	ENCODE (gs -> overprint_stroke);
	TAG ("opn");
	ENCODE (gs -> overprint_nonstroke);
	TAG ("opm");
	ENCODE (gs -> overprint_mode);
	TAG ("smt");
	ENCODE (gs -> smoothness_tolerance);
	TAG ("sta");
	ENCODE (gs -> stroke_adjustment);
	TAG ("acs");
	ENCODE (gs -> alpha_constant_stroke);
	TAG ("acn");
	ENCODE (gs -> alpha_constant_nonstroke);
	TAG ("ais");
	ENCODE (gs -> alpha_is_shape);

	TAG ("blg");
	def = gs -> black_generation.type;
	if (def == GS_BG_EXPLICIT_DEFAULT)
		def = GS_BG_DEFAULT;
	ENCODE (def);
	if (gs -> black_generation.type == GS_BG_FUNCTION) {
		res = fun_get_checksum (gs -> black_generation.t, &hash);
		if (res != S_CSC_OK) {
			free (p);
			return res;
		}
		ENCODE (hash);
	}

	TAG ("ucr");
	def = gs -> undercolor_removal.type;
	if (def == GS_UCR_EXPLICIT_DEFAULT)
		def = GS_UCR_DEFAULT;
	ENCODE (def);
	if (gs -> undercolor_removal.type == GS_UCR_FUNCTION) {
		res = fun_get_checksum (gs -> undercolor_removal.t, &hash);
		if (res != S_CSC_OK) {
			free (p);
			return res;
		}
		ENCODE (hash);
	}

	TAG ("xfn");
	n = XREF_A_SIZE (&gs -> transfer_function);
	if (n == 1 && 
	    gs -> transfer_function.data [0].type == GS_TR_EXPLICIT_DEFAULT)
		n = 0;
	ENCODE (n);
	for (j = 0; j < n; j++) {
		def = gs -> transfer_function.data [j].type;
		if (def == GS_TR_EXPLICIT_DEFAULT)
			def = GS_TR_DEFAULT;
		ENCODE (def);
		if (gs -> transfer_function.data [j].type == GS_TR_FUNCTION) {
			res = fun_get_checksum 
				(gs -> transfer_function.data [j].t, &hash);
			if (res != S_CSC_OK) {
				free (p);
				return res;
			}
			ENCODE (hash);
		}
	}

	TAG ("htd");
	def = gs -> halftone_dictionary.type;
	if (def == GS_HT_EXPLICIT_DEFAULT)
		def = GS_HT_DEFAULT;
	ENCODE (def);

	if (gs -> halftone_dictionary.type == GS_HT_FUNCTION) {
		res = t_tag_get_checksum 
			(&gs -> halftone_dictionary.t -> t.tag, &hash);
		if (res != S_CSC_OK)
			return res;
		ENCODE (hash);
	}

	TAG ("blm");
	n = XREF_A_SIZE (&gs -> blend_mode);
	if (n == 1 && gs -> blend_mode.data [0].type == GS_BM_NORMAL)
		n = 0;
	ENCODE (n);
	for (j = 0; j < n; j++)
		ENCODE (gs -> blend_mode.data [j].type);

	TAG ("smk");
	def = gs -> soft_mask.type;
	if (def == GS_SM_DEFAULT)
		def = GS_SM_NONE;
	ENCODE (def);
	if (gs -> soft_mask.type == GS_SM_DICTIONARY) {
		res = t_tag_get_checksum (&gs -> soft_mask.t -> t.tag, &hash);
		if (res != S_CSC_OK)
			return res;
		ENCODE (hash);
	}
#undef ENCODE
#undef TAG

	assert (i == size);

	s -> size = size;
	s -> s = p;

	return S_CSC_OK;
}

enum s_status gs_filter_gs (struct obj *obj, gs_keyset_t *keyset)
{
	struct obj_dictionary *dict;
	struct obj_dict_node **dn, *p, ndn;
	struct obj_bstr *key;
	struct obj kobj, dobj;
	enum s_status res;
	gs_keyset_t ks, mks;
	int i;

	dict = obj_get_rw_dictionary (obj);
	assert (dict != NULL);

	dn = &dict -> head;
	ks = 0;
	while (*dn != NULL) {
		key = &(*dn) -> key;
		for (i = 0; key_desc [i].name != NULL; i++)
			if (obj_string_equal (key, key_desc [i].name))
				break;

		if (key_desc [i].name == NULL) {
			p = *dn;
			*dn = p -> next;
			p -> next = NULL;
			obj_dict_node_destroy (&p);
		} else {
			ks |= KSDBIT (i);
			dn = &(*dn) -> next;
		}
	}

	/* Now we add all the missing keys */
	mks = *keyset & ~ks;
	for (i = 0; mks && key_desc [i].name != NULL; i++)
		if (mks & KSDBIT (i)) {
			assert (key_desc [i].dobj.type != OT_NULL);
			res = obj_dup (&key_desc [i].kobj, &kobj);
			if (res != S_CSC_OK)
				return res;

			res = obj_dup (&key_desc [i].dobj, &dobj);
			if (res != S_CSC_OK) {
				obj_destroy (&kobj);
				return res;
			}

			obj_init_dict_node (&ndn, &kobj, &dobj);
			res = obj_xfer_dict (dict, &ndn);
			if (res != S_CSC_OK)
				return res;

			mks &= ~KSDBIT (i);
		}

	assert (!mks);
	
	*keyset |= ks;

	return S_CSC_OK;
}

static enum s_status graph_set_key (struct gs_graph_state *gss)
{
	struct gs_state_key *key;
	enum s_status res;

	key = &gss -> key;
	if (key -> opkey.s == NULL) {
		res = build_gopkey (gss, &key -> opkey);
		if (res != S_CSC_OK)
			return res;
	}

	if (key -> dkey.s == NULL) {
		res = build_gdkey (gss, &key -> dkey);
		if (res != S_CSC_OK)
			return res;
	}

	return S_CSC_OK;
}

static pdf_bool_t cst_equal (const struct gc_colour_state *cst1,
			     const struct gc_colour_state *cst2)
{
	return 	gc_cspace_equal (cst1 -> sp, cst2 -> sp) &&
		gc_cstate_equal (cst1, cst2);
}

pdf_bool_t gs_graph_equal (const struct gs_graph_state *gss1,
			   const struct gs_graph_state *gss2)
{
	return 	!gs_state_key_compare (&gss1 -> key, &gss2 -> key) &&
		cst_equal (gss1 -> stroking_colour_state,
			   gss2 -> stroking_colour_state) &&
		cst_equal (gss1 -> nonstroking_colour_state,
			   gss2 -> nonstroking_colour_state);
}

enum s_status gs_graph_update_key (struct gs_graph_state *gss, 
				   enum gs_state_op op)
{
	reset_key (&gss -> key, op);

	return graph_set_key (gss);
}

#ifdef DEBUG
void gs_graph_check_key (struct gs_graph_state *gss)
{
	struct gs_state_key key;
	enum s_status res;

	key = gss -> key;

	memset (&gss -> key, 0, sizeof (struct gs_state_key)); 
	res = graph_set_key (gss);
	if (res != S_CSC_OK)
		return;

	assert (!gs_state_key_compare (&gss -> key, &key));

	reset_key (&key, GS_GO_OP);
	reset_key (&key, GS_GO_DICT);
}
#endif

enum s_status gs_graph (struct gs_context *gpc, enum gs_state_op op,
			struct gs_graph_state **gssp)
{
	struct gs_graph_state *gs;
	enum s_status res;

	if (!gpc -> state -> gstate -> frozen) {
		*gssp = gpc -> state -> gstate;
		return S_CSC_OK;
	}
       
	res = dup_graph (gpc -> state -> gstate, &gs,
			 ref_rown (&gpc -> state -> t.refs));
	if (res != S_CSC_OK)
		return res;

	res = graph_set_key (gpc -> state -> gstate);
	if (res != S_CSC_OK) {
		gs_graph_unref (gs, ref_rown (&gpc -> state -> t.refs));
		return res;
	}

	gs_graph_unref (gpc -> state -> gstate, 
			ref_rown (&gpc -> state -> t.refs));

	gpc -> state -> gstate = *gssp = gs;

	return S_CSC_OK;
}

const struct gs_graph_state *gs_ro_graph (struct gs_context *gpc)
{
	return gpc -> state -> gstate;
}

struct gs_graph_state *gs_snap_graph (struct gs_context *gpc,
				      struct ref_owner *p)
{
	struct gs_graph_state *gss;

	gss = gpc -> state -> gstate;

	gss -> frozen = 1;

	T_CALL (gs_graph_ref, gss, p);

	return gss;
}

enum s_status gs_set_gstate (struct gs_context *gpc,
			     struct gs_graph_state *gss)
{
	enum s_status res;

	res = graph_set_key (gpc -> state -> gstate);
	if (res != S_CSC_OK)
		return res;

	gs_graph_unref (gpc -> state -> gstate, 
		       ref_rown (&gpc -> state -> t.refs));
	gpc -> state -> gstate = gss;
	T_CALL (gs_graph_ref, gpc -> state -> gstate,
		ref_rown (&gpc -> state -> t.refs));

	gss -> frozen = 0;

	return S_CSC_OK;
}

enum s_status gs_import_gstate (struct gs_graph_state **gssp,
				struct ref_owner *p)
{
	struct gs_graph_state *gss;
	enum s_status res;

	res = dup_graph (*gssp, &gss, p);
	if (res != S_CSC_OK)
		return res;

	res = graph_set_key (gss);
	if (res != S_CSC_OK) {
		gs_graph_unref (gss, p);
		return res;
	}

	gss -> frozen = 0;
	*gssp = gss;

	return S_CSC_OK;
}

enum s_status gs_begin_text (struct gs_context *gpc, const struct obj *ref)
{
	struct gs_subpath *gss;	
	enum s_status res;

	if (gpc -> gss != NULL)
		return obj_error (ref, "Path is open");
	
	res = init_subpath (&gss);
	if (res != S_CSC_OK)
		return res;

	gpc -> path -> head = gss;
	gpc -> gss = gss;

	res = gs_set_mat (gpc, &cobj_eye);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

enum s_status gs_newline (struct gs_context *gpc, 
			   pdf_real_t tx, pdf_real_t ty)
{
	struct gs_text_state *tws;
	enum s_status res;

	res = gs_text (gpc, GS_GO_OP, &tws);
	if (res != S_CSC_OK)
		return res;

	tws -> mat.t_lm.a [2][0]= 
		tws -> mat.t_lm.a [0][0] * tx +
		tws -> mat.t_lm.a [1][0] * ty +
		tws -> mat.t_lm.a [2][0];
	tws -> mat.t_lm.a [2][1] = 
		tws -> mat.t_lm.a [0][1] * tx +
		tws -> mat.t_lm.a [1][1] * ty +
		tws -> mat.t_lm.a [2][1];
	tws -> mat.t_lm.a [2][2]= 
		tws -> mat.t_lm.a [0][2] * tx +
		tws -> mat.t_lm.a [1][2] * ty +
		tws -> mat.t_lm.a [2][2];
		
	tws -> mat.t_m = tws -> mat.t_lm;

	return S_CSC_OK;
}

enum s_status gs_set_mat (struct gs_context *gpc, struct cobj_matrix *mat)
{
	struct gs_text_state *tws;
	enum s_status res;

	res = gs_text (gpc, GS_GO_OP, &tws);
	if (res != S_CSC_OK)
		return res;

	tws -> mat.t_m = tws -> mat.t_lm = *mat;

	return S_CSC_OK;
}

int gs_t_clip (struct gs_context *gpc)
{
	const struct gs_text_state *tws;

	tws = gs_ro_text (gpc);

	return tws -> render >= GS_RT_FILL_CLIP;
}

void gs_paint_to_render (struct gs_paint *paint, 
			 enum gs_render_type *rt, enum gs_fill_rule cfrule)
{
	if (paint != NULL) {
		if (paint -> stroke)
			if (paint -> fill)
				*rt = GS_RT_FILL_STROKE;
			else
				*rt = GS_RT_STROKE;
		else
			if (paint -> fill)
				*rt = GS_RT_FILL;
			else
				*rt = GS_RT_NONE;
	} else
		*rt = GS_RT_NONE;

	if (cfrule != GS_FR_NONE)
		*rt += GS_RT_FILL_CLIP;
}

void gs_render_to_paint (enum gs_render_type rt, struct gs_paint *paint)
{
	paint -> stroke = 
		rt == GS_RT_STROKE ||
		rt == GS_RT_STROKE_CLIP ||
		rt == GS_RT_FILL_STROKE_CLIP;

	paint -> fill = 
		rt == GS_RT_FILL ||
		rt == GS_RT_FILL_CLIP ||
		rt == GS_RT_FILL_STROKE_CLIP;

	paint -> rule = GS_FR_WINDING;
}

enum s_status gs_write_text (struct gs_context *gpc, struct gs_text *text)
{
	struct gs_text_state *tws;
	struct gs_element *e;
	struct obj tobj;
	enum s_status res;

	res = obj_dup (&text -> text, &tobj);
	if (res != S_CSC_OK)
		return res;

	res = add_element (gpc -> gss, &e);
	if (res != S_CSC_OK) {
		obj_destroy (&tobj);
		return res;
	}

	e -> type = GS_ET_TEXT;
	e -> u.text.text = tobj;
	if (text -> tstate != NULL) {
		e -> u.text.tstate = text -> tstate;
		T_CALL (gs_text_ref, text -> tstate, 
			ref_rown (&gpc -> path -> refs));
	} else
		e -> u.text.tstate = 
			gs_snap_text (gpc, ref_rown (&gpc -> path -> refs));

	if (text -> gstate != NULL) {
		e -> u.text.gstate = text -> gstate;
		T_CALL (gs_graph_ref, text -> gstate, 
			ref_rown (&gpc -> path -> refs));
	} else
		e -> u.text.gstate = 
			gs_snap_graph (gpc, ref_rown (&gpc -> path -> refs));

	res = gs_text (gpc, GS_GO_OP, &tws);
	if (res != S_CSC_OK)
		return res;

	res = gs_get_string (&e -> u.text, NULL, &tws -> mat.t_m);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

enum s_status gs_write (struct gs_context *gpc, struct obj *text)
{
	struct gs_text txt;
	enum s_status res;

	txt.text = *text;
	txt.tstate = NULL;
	txt.gstate = NULL;
	res = gs_write_text (gpc, &txt);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

void gs_origin_half_init (struct gs_point *point, struct cobj_rect *rect,
			  enum gs_direction dir)
{
	switch (dir) {
	case GS_D_LEFT_TO_RIGHT:
		point -> x = rect -> left;
		break;

	case GS_D_RIGHT_TO_LEFT:
		point -> x = rect -> right;
		break;

	case GS_D_BOTTOM_TO_TOP:
		point -> y = rect -> bottom;
		break;

	case GS_D_TOP_TO_BOTTOM:
		point -> y = rect -> top;
		break;

	default:
		assert (0);
	}
}

void gs_origin_init (struct gs_point *point, struct cobj_rect *rect,
		     enum gs_direction pdir, enum gs_direction sdir)
{
	gs_origin_half_init (point, rect, pdir);
	gs_origin_half_init (point, rect, sdir);
}

struct ref_owner *gs_cstate_owner (struct gs_context *gpc)
{
	return ref_rown (&gpc -> state -> gstate -> t.refs);
}

static pdf_real_t left_rect_edge_f (struct cobj_rect *rect)
{
	return rect -> left;
}

static pdf_real_t right_rect_edge_f (struct cobj_rect *rect)
{
	return rect -> right;
}

static pdf_real_t bottom_rect_edge_f (struct cobj_rect *rect)
{
	return rect -> bottom;
}

static pdf_real_t top_rect_edge_f (struct cobj_rect *rect)
{
	return rect -> top;
}

gs_edge gs_get_lo_edge (enum gs_direction dir)
{
	switch (dir) {
	case GS_D_LEFT_TO_RIGHT:
		return left_rect_edge_f;

	case GS_D_RIGHT_TO_LEFT:
		return right_rect_edge_f;

	case GS_D_BOTTOM_TO_TOP:
		return bottom_rect_edge_f;

	case GS_D_TOP_TO_BOTTOM:
		return top_rect_edge_f;

	case GS_D_UNKNOWN:
	case GS_D__LAST__:
		assert (0);
	}

	assert (0);
}

gs_edge gs_get_hi_edge (enum gs_direction dir)
{
	switch (dir) {
	case GS_D_LEFT_TO_RIGHT:
		return right_rect_edge_f;

	case GS_D_RIGHT_TO_LEFT:
		return left_rect_edge_f;

	case GS_D_BOTTOM_TO_TOP:
		return top_rect_edge_f;

	case GS_D_TOP_TO_BOTTOM:
		return bottom_rect_edge_f;

	case GS_D_UNKNOWN:
	case GS_D__LAST__:
		assert (0);
	}

	assert (0);
}

static pdf_real_t cmp_up (pdf_real_t a, pdf_real_t b)
{
	return a - b;
}

static pdf_real_t cmp_down (pdf_real_t a, pdf_real_t b)
{
	return b - a;
}

gs_cmp gs_get_cmp (enum gs_direction dir)
{
	switch (dir) {
	case GS_D_LEFT_TO_RIGHT:
		return cmp_up;

	case GS_D_RIGHT_TO_LEFT:
		return cmp_down;

	case GS_D_BOTTOM_TO_TOP:
		return cmp_up;

	case GS_D_TOP_TO_BOTTOM:
		return cmp_down;

	case GS_D_UNKNOWN:
	case GS_D__LAST__:
		assert (0);
	}

	assert (0);
}

enum gs_direction gs_dir_neg (enum gs_direction dir)
{
	switch (dir) {
	case GS_D_LEFT_TO_RIGHT:
		return GS_D_RIGHT_TO_LEFT;

	case GS_D_RIGHT_TO_LEFT:
		return GS_D_LEFT_TO_RIGHT;

	case GS_D_BOTTOM_TO_TOP:
		return GS_D_TOP_TO_BOTTOM;

	case GS_D_TOP_TO_BOTTOM:
		return GS_D_BOTTOM_TO_TOP;

	case GS_D_UNKNOWN:
	case GS_D__LAST__:
		assert (0);
	}

	assert (0);
}

static struct xref_field *find_field (const char *name)
{
	int i;
	
	for (i = 0; graph_state_field_desc.fd [i].name != NULL; i++)
		if (!strcmp (graph_state_field_desc.fd [i].name, name))
			return &graph_state_field_desc.fd [i];

	for (i = 0; text_state_field_desc.fd [i].name != NULL; i++)
		if (!strcmp (text_state_field_desc.fd [i].name, name))
			return &text_state_field_desc.fd [i];

	return NULL;
}

enum s_status gsg_init (void)
{
	struct xref_field *field;
	struct obj_bstr bstr;
	gs_keyset_t tks;
	enum s_status res;
	int i;

	res = S_CSC_OK;

	for (i = 0; key_desc [i].name != NULL; i++) {
		field = find_field (key_desc [i].name);
		assert (field != NULL);

		switch (field -> des) {
		case XREF_D_UNWRAP:
			assert (key_desc [i].dstring == NULL);
			res = obj_dup (&field -> defval, &key_desc [i].dobj);
			if (res != S_CSC_OK)
				break;
			break;

		case XREF_D_PSTABLE:
		case XREF_D_ARRAY:
			assert (field -> u.pstable == &cobj_enum_pstable);
			if (key_desc [i].dstring != NULL) {
				res = obj_string_from_c (key_desc [i].dstring, 
							 &bstr);
				if (res != S_CSC_OK)
					break;
					
				obj_set_name (&key_desc [i].dobj, 
					      &f_null, &bstr);
			} else	
				obj_set_null (&key_desc [i].dobj, &f_null);
			break;
			
		default:
			assert (0);
		}

		if (res != S_CSC_OK)
			break;

		res = obj_string_from_c (key_desc [i].name, &bstr);
		if (res != S_CSC_OK)
			break;

		obj_set_string (&key_desc [i].kobj, &f_null, &bstr);
	}

	if (res != S_CSC_OK) {
		gsg_cleanup ();
		return res;
	}

	/* Just to make sure the keyset type is large enough */
	tks = 1 << i;
	assert (tks);

	return S_CSC_OK;
}

void gsg_cleanup (void)
{
	int i;

	for (i = 0; key_desc [i].name != NULL; i++) {
		obj_destroy (&key_desc [i].kobj);
		obj_destroy (&key_desc [i].dobj);
	}
}
