/*
 * PDFlow -- gstree.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>

#include "graph/gstree.h"

#define BTREE_MIN_SIZE 15

static int compare (bt_key_t *key1, bt_key_t *key2);
static void destroy_data (bt_tree_data_t *tdata, bt_data_t *data);

static struct bt_param tree_param = {
	
	sizeof (struct gst_data *),
	BTREE_MIN_SIZE,
	NULL,
	compare, NULL, NULL, NULL, destroy_data

};

static struct gst_tree *cast_tree_data (bt_tree_data_t *tdata)
     __attribute__ ((used));

static struct gst_tree *cast_tree_data (bt_tree_data_t *tdata)
{
	return (struct gst_tree *) tdata;
}

static bt_tree_data_t *uncast_tree_data (struct gst_tree *gst)
{
	return (struct bt_tree_data_t *) gst;
}

static struct gst_data **cast_key (bt_key_t *key)
{
	return (struct gst_data **) key;
}

static bt_key_t *uncast_key (struct gst_data **gdata)
{
	return (struct bt_key_t *) gdata;
}

static struct gst_data *cast_data (bt_key_t *data)
{
	return (struct gst_data *) data;
}

static bt_data_t *uncast_data (struct gst_data *data)
{
	return (struct bt_data_t *) data;
}

static int compare (bt_key_t *key1, bt_key_t *key2)
{
	struct gst_data *d1, *d2;
	int ans;

	d1 = *cast_key (key1);
	d2 = *cast_key (key2);

	ans = gs_state_dkey_compare (&d1 -> gkey, &d2 -> gkey);
	if (ans)
		return ans;

	ans = gs_state_dkey_compare (&d1 -> tkey, &d2 -> tkey);
	if (ans)
		return ans;

	return 0;
}

static void destroy_data (bt_tree_data_t *tdata, bt_data_t *data)
{
	struct gst_data *gdata;

	gdata = cast_data (data);

	gs_tags_destroy (&gdata -> tags);
	gs_state_key_destroy (&gdata -> gkey);
	gs_state_key_destroy (&gdata -> tkey);
	obj_string_destroy (&gdata -> name);
	obj_destroy (&gdata -> gs);

	free (gdata);
}

static enum s_status init (struct gst_tree *gst)
{
	struct bt_param param;
	enum s_status res;

	param = tree_param;
	param.tdata = uncast_tree_data (gst);

	res = bt_init (&gst -> tree, &param, ref_rown (&gst -> refs));
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

enum s_status gst_new (struct gst_tree **gstp, struct xref_db *rdb,
		       struct v_selection *sel, struct ref_owner *p)
{
	struct gst_tree *gst;
	enum s_status res;

	gst = malloc (sizeof (struct gst_tree));
	if (gst == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (gst, 0, sizeof (struct gst_tree));

	ref_init (&gst -> refs, "gst_tree", p);

	gst -> rdb = rdb;
	xref_ref_db (rdb, ref_rown (&gst -> refs));

	gst -> sel = sel;

	res = init (gst);
	if (res != S_CSC_OK) {
		gst_unref (gst, p);
		return res;
	}

	*gstp = gst;

	return S_CSC_OK;
}

static enum s_status update_gs (struct gst_tree *gst,
				struct gst_data *data)
{
	enum s_status res;

	res = gs_filter_gs (&data -> gs, &gst -> keyset);
	if (res != S_CSC_OK)
		return res;

	data -> keyset = gst -> keyset;

	return S_CSC_OK;
}
				

static enum s_status init_data (struct gst_tree *gst, struct gst_data *data, 
				struct gs_graph_state *gs, 
				struct gs_text_state *tws,
				struct obj *name)
{
	char buf [2 + 1 + sizeof (pdf_integer_t) * 2 + 1];
	struct obj gstate, tstate;
	struct obj_bstr str;
	enum s_status res;

	res = gs_state_dkey_dup (&gs -> key, &data -> gkey);
	if (res != S_CSC_OK)
		return res;

	res = gs_state_dkey_dup (&tws -> key, &data -> tkey);
	if (res != S_CSC_OK)
		return res;

	res = gs_tags_get (gs, &data -> tags);
	if (res != S_CSC_OK)
		return res;

	res = gs_graph_sync (gst -> rdb, NULL, gst -> sel, &gstate, NULL, gs);
	if (res != S_CSC_OK)
		return res;

	res = gs_text_sync (gst -> rdb, NULL, gst -> sel, &tstate, NULL, tws);
	if (res != S_CSC_OK) {
		obj_destroy (&gstate);
		return res;
	}

	obj_dict_merge (&gstate.val.dictionary, &tstate.val.dictionary);
	data -> gs = gstate;

	res = update_gs (gst, data);
	if (res != S_CSC_OK)
		return res;

	snprintf (buf, sizeof (buf), "gs%lx", gst -> seq);
	res = obj_string_from_c (buf, &data -> name);
	if (res != S_CSC_OK)
		return res;

	res = obj_string_dup (&data -> name, &str);
	if (res != S_CSC_OK)
		return res;

	obj_set_name (name, &f_null, &str);


	gst -> seq++;

	return S_CSC_OK;
}

enum s_status gst_lookup (struct gst_tree *gst, 
			  struct gs_graph_state *gs, 
			  struct gs_text_state *tws,
			  struct obj *name,
			  struct obj *gsobj,
			  struct gs_tags *tags)
{
	struct gst_data data, *nd, *k;
	struct obj_bstr str;
	enum s_status res;
	bt_data_t *found;

	data.gkey = gs -> key;
	data.tkey = tws -> key;

	k = &data;

	res = bt_get (gst -> tree, uncast_key (&k), &found);
	if (res != S_CSC_OK)
		return res;

	if (found != NULL) {
		res = update_gs (gst, cast_data (found));
		if (res != S_CSC_OK)
			return res;

		res = obj_string_dup (&cast_data (found) -> name, &str);
		if (res != S_CSC_OK)
			return res;

		res = gs_tags_dup (&cast_data (found) -> tags, tags);
		if (res != S_CSC_OK) {
			obj_string_destroy (&str);
			return res;
		}
				   

		res = obj_dup (&cast_data (found) -> gs, gsobj);
		if (res != S_CSC_OK) {
			gs_tags_destroy (tags);
			obj_string_destroy (&str);
			return res;
		}

		obj_set_name (name, &f_null, &str);

		return S_CSC_OK;
	}

	nd = malloc (sizeof (struct gst_data));
	if (nd == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (nd, 0, sizeof (struct gst_data));
	
	res = init_data (gst, nd, gs, tws, name);
	if (res != S_CSC_OK) {
		destroy_data (uncast_tree_data (gst), 
			      uncast_data (nd));
		return res;
	}

	k = nd;

	res = bt_insert (gst -> tree, uncast_key (&k), uncast_data (nd), NULL);
	if (res != S_CSC_OK) {
		obj_destroy (name);
		destroy_data (uncast_tree_data (gst), 
			      uncast_data (nd));
		return res;
	}

	res = gs_tags_dup (&nd -> tags, tags);
	if (res != S_CSC_OK) {
		obj_destroy (name);
		return res;
	}

	res = obj_dup (&nd -> gs, gsobj);
	if (res != S_CSC_OK) {
		gs_tags_destroy (tags);
		obj_destroy (name);
		return res;
	}

	return S_CSC_OK;
}

static enum s_status rdict_f (bt_node_data_t *ndata, bt_data_t *d,
			      va_list ap)
{
	struct obj_dictionary *dict;
	struct gst_tree *gst;
	struct gst_data *data;
	struct obj_bstr kstr;
	struct obj_dict_node dn;
	struct obj key, value;
	enum s_status res;

	if (d == NULL)
		return S_CSC_OK;

	data = cast_data (d);

	gst = va_arg (ap, struct gst_tree *);
	dict = va_arg (ap, struct obj_dictionary *);

	res = update_gs (gst, data);
	if (res != S_CSC_OK)
		return res;
	
	res = obj_string_dup (&data -> name, &kstr);
	if (res != S_CSC_OK)
		return res;

	obj_set_name (&key, &f_null, &kstr);

	res = obj_dup (&data -> gs, &value);
	if (res != S_CSC_OK) {
		obj_destroy (&key);
		return res;
	}

	obj_init_dict_node (&dn, &key, &value);
	res = obj_xfer_dict (dict, &dn);
	if (res != S_CSC_OK) {
		obj_destroy (&key);
		obj_destroy (&value);
		return res;
	}

	return S_CSC_OK;
}

enum s_status gst_rdict (struct gst_tree *tree, struct obj *dobj)
{
	struct obj_dictionary dict;
	enum s_status res;

	obj_init_dictionary (&dict);

	res = bt_traverse (tree -> tree, rdict_f, tree, &dict);
	if (res != S_CSC_OK) {
		obj_dict_destroy (&dict);
		return res;
	}

	if (dict.head != NULL)
		obj_set_dictionary (dobj, &f_null, &dict);
	else
		obj_set_null (dobj, &f_null);

	return S_CSC_OK;
}

void gst_ref (struct gst_tree *gst, struct ref_owner *p)
{
	ref_up (&gst -> refs, p);
}

void gst_unref (struct gst_tree *gst, struct ref_owner *p)
{
	if (ref_down (&gst -> refs, p))
		return;

	if (gst -> tree != NULL)
		bt_unref (gst -> tree, ref_rown (&gst -> refs));

	xref_unref_db (gst -> rdb, ref_rown (&gst -> refs));

	ref_zero (&gst -> refs);

	free (gst);
}

