/*
 * PDFlow -- align.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <assert.h>
#include <string.h>

#include "grid/align.h"
#include "grid/thread.h"
#include "grid/sig.h"

#ifndef min
#define min(a, b) ((a) < (b) ? (a) : (b))
#endif

#ifndef max
#define max(a, b) ((a) > (b) ? (a) : (b))
#endif

#define OFS0_INIT_SCORE 10

#define SC_INDENT 	       1
#define SC_WIDTH	       .25
#define SC_HEIGHT	       .5
#define SN_AT_BEGIN  	       0
#define SN_AT_JUSTIFIED	       .7
#define SN_AT_END	       0
#define SN_AT_CENTER	       1

struct gr_al_pole {

	struct gr_al_pole *next;

	pdf_real_t offset;

	pdf_real_t occurrences;

};

struct gr_al_poles {

	struct gr_al_pole *begin, *center, *end;
};

struct gr_al_margin {

	pdf_real_t lo, hi;

};

struct gr_al_margins {

	struct gr_al_margin begin, end;
};

struct gr_al_model {

	struct gr_al_model *next;

	pdf_real_t occurrences;

	struct gr_al_alignment al;

};

struct paragraph {

	struct gr_al_alignment al;

	struct line_entry *le;

	const char *reason, *merge_reason;

	struct gr_al_poles poles;

	struct gr_al_margins margins;

	pdf_real_t solidity;

	pdf_bool_t hard_split, margins_set, merge_with_next, settled;
};

struct line_entry {

	struct line_entry *next;

	struct gr_grid *grid;

	enum gs_direction wdir, ldir;

	pdf_bool_t has_graph_wspace;

	struct gr_al_poles poles;

	pdf_real_t first_word, lsize, lwidth;

	struct td_space lspace;

	struct gr_sig_signature sig;

	pdf_bool_t par_break;
};

struct gr_al_machine {

	FILE *rfp;

	struct gr_al_poles poles;

	struct gr_al_model *models;

	struct args_heur *heur;

	struct gr_grid *last_line;

	struct gr_al_margins margins;

	struct line_entry *les;

	enum gr_al_align_type last_al [GR_AT__LAST__];

	int paras, avail;
	struct paragraph *para;
};

static int para_height (struct paragraph *para);

static int dump_al_sent (FILE *fp, enum gr_al_align_type at,
			 pdf_real_t head, pdf_real_t tail)
{
	const char *s;

	s = NULL;
	
	switch (at) {
	case GR_AT_NONE:
		return 0;

	case GR_AT_PROPORTIONAL:
	case GR_AT_UNKNOWN:
	case GR_AT__LAST__:
		assert (0);
		break;

	case GR_AT_BEGIN:
		s = ">>";
		break;

	case GR_AT_END:
		s = "<<";
		break;

	case GR_AT_CENTER:
		s = "<>";
		break;

		/* flush_at set to justified simply
		 * does not make sense */
	case GR_AT_JUSTIFIED:
		s = "><";
	}

	assert (s != NULL);

	fprintf (fp, "%s %g/%g ", s, head, tail);

	return 1;
}

static void dump_align (FILE *fp, struct gr_al_alignment *al)
{
	if (dump_al_sent (fp, al -> at, al -> fhead, al -> ftail))
		dump_al_sent (fp, al -> flush_at, al -> lhead, al -> ltail);
	fprintf (fp, "%d \n", al -> size);
}

static void dump_pole (struct gr_al_pole *pole)
{
	if (pole != NULL)
		printf ("%g ", pole -> offset);
	else
		printf ("| ");
}

static void dump_poles (struct gr_al_poles *poles)
{
	dump_pole (poles -> begin);
	dump_pole (poles -> center);
	dump_pole (poles -> end);
}

static void dump_margin (struct gr_al_margin *margin)
{
	printf ("%g/%g ", margin -> lo, margin -> hi);
}

static void dump_margins (struct gr_al_margins *margins)
{
	dump_margin (&margins -> begin);
	dump_margin (&margins -> end);
}

static void dump_para (struct gr_al_machine *dc) __attribute__ ((unused));
static void dump_para (struct gr_al_machine *dc)
{
	struct paragraph *para;
	int i, h;

	h = 1;
	for (i = 0; i < dc -> paras; i++) {
		para = &dc -> para [i];
		printf ("%2d|%2d) %2d[%3d]", i, h,
			para_height (para),
			(int) (100 * para -> solidity));

		h += para_height (para);
		
		dump_align (stderr, &para -> al);
		dump_poles (&para -> poles);
		dump_margins (&para -> margins);
		printf ("\n");
	}
}

static enum s_status add_pole (struct gr_al_pole **list,
			       pdf_real_t offset,
			       struct args_heur *heur,
			       struct gr_al_pole **polep)
{
	struct gr_al_pole **pp, *np;
	pdf_real_t diff, indent_tolerance;
	
	indent_tolerance = heur != NULL ? heur -> indent_tolerance : 0;

	for (pp = list; *pp != NULL; pp = &(*pp) -> next) {
		diff = (*pp) -> offset - offset;
		if (labs (diff) < indent_tolerance)
			diff = 0;
		if (diff >= 0)
			break;
	}

	if (*pp == NULL || diff > 0) {
		np = malloc (sizeof (struct gr_al_pole));
		if (np == NULL)
			return S_CSC_OUT_OF_MEMORY;

		np -> occurrences = offset == 0 ? OFS0_INIT_SCORE : 0;
		np -> offset = offset;

		np -> next = *pp;
		*pp = np;
	}

	(*pp) -> occurrences++;

	if (polep != NULL)
		*polep = *pp;

	return S_CSC_OK;
}

static void destroy_poles (struct gr_al_pole **list)
{
	struct gr_al_pole *pole;

	while (*list != NULL) {
		pole = *list;
		*list = pole -> next;
		free (pole);
	}
}

static void decad_poles (struct gr_al_pole **list, struct args_heur *heur)
{
	struct gr_al_pole **pp, *pole;

	for (pp = list; (*pp) != NULL; pp = &(*pp) -> next) {
		(*pp) -> occurrences *= heur -> par_decadence;
		if ((*pp) -> offset && !(*pp) -> occurrences) {
			pole = *pp;
			*pp = pole -> next;
			free (pole);

			if (*pp == NULL)
				break;
		}	
	}
}

static void destroy_models (struct gr_al_model **glmp)
{
	struct gr_al_model *glm;

	while (*glmp != NULL) {
		glm = *glmp;
		*glmp = glm -> next;

		free (glm);
	}
}

static int diff (pdf_real_t x, pdf_real_t y, struct args_heur *heur)
{
	if (x < 0 || y < 0)
		return 0;

	if (labs (x - y) < heur -> indent_tolerance)
		return 0;

	return x > y ? 1 : -1;
}

static int compare_models (struct gr_al_model *glm, 
			   struct gr_al_alignment *al,
			   struct args_heur *heur)
{
	int ans;

	ans = diff (glm -> al.fhead, al -> fhead, heur);
	if (ans)
		return ans;

	ans = diff (glm -> al.ftail, al -> ftail, heur);
	if (ans)
		return ans;

	ans = diff (glm -> al.lhead, al -> lhead, heur);
	if (ans)
		return ans;

	ans = diff (glm -> al.ltail, al -> ltail, heur);
	if (ans)
		return ans;

	return 0;
}

static enum s_status add_model (struct gr_al_machine *dc,
				struct gr_al_alignment *al)
{
	struct gr_al_model **glm, *nglm;
	int cmp;

	for (glm = &dc -> models; *glm != NULL; glm = &(*glm) -> next) {
		cmp = compare_models (*glm, al, dc -> heur);

		if (!cmp && (*glm) -> al.at == al -> at &&
			 (*glm) -> al.flush_at == al -> flush_at) {
			(*glm) -> occurrences++;
			return S_CSC_OK;
		} else if (cmp < 0)
			break;
	}

	nglm = malloc (sizeof (struct gr_al_model));
	if (nglm == NULL)
		return S_CSC_OUT_OF_MEMORY;

	nglm -> al = *al;
	nglm -> occurrences = 1;
	nglm -> next = *glm;
	*glm = nglm;

	return S_CSC_OK;
}

struct gr_al_model *find_model (struct gr_al_machine *dc,
				struct gr_al_alignment *al)
{
	struct gr_al_model *glm, *best;
	int cmp;

	best = NULL;
	for (glm = dc -> models; glm != NULL; glm = glm -> next) {
		cmp = compare_models (glm, al, dc -> heur);
		if (!cmp &&
		    (best == NULL ||
		     glm -> occurrences > best -> occurrences))
			best = glm;
		if (cmp < 0)
			break;
	}

	return best;
}


static int match_margins (struct gr_al_margin *marg, pdf_real_t ofs,
			  struct args_heur *heur)
{
	return 	ofs >= marg -> lo - heur -> indent_tolerance &&
		ofs <= marg -> hi + heur -> indent_tolerance;
}

static struct gr_al_model *find_breaking_model (struct gr_al_machine *dc,
						struct paragraph *para)
{
	struct gr_al_model *glm, *best;
	int begin_ok, end_ok, ok;
	pdf_real_t occurrences;

	best = NULL;
	occurrences = -1;
	for (glm = dc -> models; glm != NULL; glm = glm -> next) {
		
		begin_ok = match_margins (&para -> margins.begin, 
					  glm -> al.fhead,
					  dc -> heur);

		end_ok = match_margins (&para -> margins.begin, 
					glm -> al.ftail,
					dc -> heur);

		ok = -1;
		switch (glm -> al.at) {
			/* These should never enter the library */
		case GR_AT_NONE:
		case GR_AT_PROPORTIONAL:
		case GR_AT_UNKNOWN:
		case GR_AT__LAST__:
		case GR_AT_CENTER:
			assert (0);
			break;

		case GR_AT_BEGIN:
			ok = begin_ok;
			break;

		case GR_AT_END:
			ok = end_ok;
			break;

		case GR_AT_JUSTIFIED:
			ok = begin_ok && end_ok;
		}

		if (ok) {
			best = glm;
			occurrences = glm -> occurrences;
		}
	}

	return best;
}

static enum s_status add_line_entry (struct gr_al_machine *dc,
				     struct gr_grid *grid,
				     struct gr_grid *lgrid,
				     pdf_real_t lsize, pdf_real_t lwidth,
				     pdf_real_t begin, pdf_real_t center,
				     pdf_real_t end, pdf_real_t first_word,
				     struct td_space *lspace)
{
	struct gr_al_pole *beginp, *centerp, *endp;
	struct line_entry *le;
	enum s_status res;

	res = add_pole (&dc -> poles.begin, begin, 
			dc -> heur, &beginp);
	if (res != S_CSC_OK)
		return res;

	res = add_pole (&dc -> poles.center, center, 
			dc -> heur, &centerp);
	if (res != S_CSC_OK)
		return res;

	res = add_pole (&dc -> poles.end, end, 
			dc -> heur, &endp);
	if (res != S_CSC_OK)
		return res;

	le = malloc (sizeof (struct line_entry));
	if (le == NULL)
		return S_CSC_OUT_OF_MEMORY;

	le -> grid = grid;
	le -> wdir = lgrid -> wdir;
	le -> ldir = lgrid -> ldir;
	le -> lsize = lsize;
	le -> lwidth = lwidth;
	le -> poles.begin = beginp;
	le -> poles.center = centerp;
	le -> poles.end = endp;
	le -> first_word = first_word;
	le -> lspace = *lspace;
	le -> has_graph_wspace = lgrid -> fill.first != NULL;
	le -> par_break = 0;

	gr_sig_mksig (&le -> sig, lgrid);

	if (dc -> les != NULL) {
		le -> next = dc -> les -> next;
		dc -> les -> next = le;
	} else
		le -> next = le;

	dc -> les = le;

	return S_CSC_OK;
}

static void destroy_line_table (struct line_entry *le)
{
	struct line_entry *dle;

	while (le != NULL) {
		dle = le;
		le = dle -> next;
		free (dle);
	}
}

static void destroy_para_table (struct gr_al_machine *dc)
{
	int i;

	if (dc -> para == NULL)
		return;

	for (i = 0; i < dc -> paras; i++)
		destroy_line_table (dc -> para [i].le);

	free (dc -> para);
	
	dc -> para = NULL;
}

enum s_status gr_al_create_machine (struct gr_al_machine **dcp,
				    struct args_heur *heur,
				    FILE *rfp)
{
	struct gr_al_machine *dc;
	enum gr_al_align_type at;
	enum s_status res;

	dc = malloc (sizeof (struct gr_al_machine));
	if (dc == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (dc, 0, sizeof (struct gr_al_machine));

	dc -> rfp = rfp;
	dc -> heur = heur;

	for (at = 0; at < GR_AT__LAST__; at++)
		dc -> last_al [at] = at;

	/* A matter of taste. I like justified paragraphs best,
	 * however it is quite unlikely this default setting will be used */
	dc -> last_al [GR_AT_BEGIN] = GR_AT_JUSTIFIED;

	res = add_pole (&dc -> poles.begin, 0, NULL, NULL);
	if (res == S_CSC_OK)
		res = add_pole (&dc -> poles.center, 0, NULL, NULL);
	if (res == S_CSC_OK)
		res = add_pole (&dc -> poles.end, 0, NULL, NULL);
	if (res != S_CSC_OK) {
		gr_al_destroy_machine (dc);
		return res;
	}

	*dcp = dc;

	return S_CSC_OK;
}

void gr_al_destroy_machine (struct gr_al_machine *dc)
{
	struct line_entry *le;

	destroy_poles (&dc -> poles.begin);
	destroy_poles (&dc -> poles.center);
	destroy_poles (&dc -> poles.end);
	destroy_models (&dc -> models);
	destroy_para_table (dc);

	if (dc -> les != NULL) {
		le = dc -> les -> next;
		dc -> les -> next = NULL;

		destroy_line_table (dc -> les);
	}

	free (dc);
}

static void do_decad (struct gr_al_machine *dc)
{
	decad_poles (&dc -> poles.begin, dc -> heur);
	decad_poles (&dc -> poles.center, dc -> heur);
	decad_poles (&dc -> poles.end, dc -> heur);
}

static struct gr_grid *get_first_word (struct gr_grid *grid)
{
	struct gr_cell *cell;
	struct gr_line *line;

	if (gr_is_word (grid))
		return grid;

	line = grid -> d.lines;
	assert (line != NULL);
	
	cell = line -> cells;
	assert (cell != NULL);

	return get_first_word (cell -> grid);
}

static enum s_status next_line (struct gr_grid *lgrid,
				struct gr_grid *cell,
				struct gr_th_info *info, 
				struct gr_al_machine *dc)
{
	pdf_real_t begin, end, center, first_word, lsize, lwidth;
	struct gr_grid *cgrid, *fwgrid;
	enum s_status res;

	fwgrid = get_first_word (lgrid);
	cgrid = info -> column;

	lwidth = lsize = first_word = begin = end = 0;

	switch (lgrid -> wdir) {
	case GS_D_LEFT_TO_RIGHT:
		begin = lgrid -> g.rect.left - cgrid -> g.rect.left;
		end = cgrid -> g.rect.right - lgrid -> g.rect.right;
		first_word = cobj_rect_width (&fwgrid -> g.rect);
		lsize = cobj_rect_height (&lgrid -> g.rect);
		lwidth = cobj_rect_width (&lgrid -> g.rect);
		break;

	case GS_D_TOP_TO_BOTTOM:
		begin = cgrid -> g.rect.top - lgrid -> g.rect.top;
		end = lgrid -> g.rect.bottom - cgrid -> g.rect.bottom;
		first_word = cobj_rect_height (&fwgrid -> g.rect);
		lsize = cobj_rect_width (&lgrid -> g.rect);
		lwidth = cobj_rect_height (&lgrid -> g.rect);
		break;

	case GS_D_RIGHT_TO_LEFT:
		begin = cgrid -> g.rect.right - lgrid -> g.rect.right;
		end = lgrid -> g.rect.left - cgrid -> g.rect.left;
		first_word = cobj_rect_width (&fwgrid -> g.rect);
		lsize = cobj_rect_height (&lgrid -> g.rect);
		lwidth = cobj_rect_width (&lgrid -> g.rect);
		break;

	case GS_D_BOTTOM_TO_TOP:
		begin = lgrid -> g.rect.bottom - cgrid -> g.rect.bottom;
		end = cgrid -> g.rect.top - lgrid -> g.rect.top;
		first_word = cobj_rect_height (&fwgrid -> g.rect);
		lsize = cobj_rect_width (&lgrid -> g.rect);
		lwidth = cobj_rect_height (&lgrid -> g.rect);
		break;

	case GS_D_UNKNOWN:
	case GS_D__LAST__:
		assert (0);
	}

	if (fwgrid -> wdir != cgrid -> wdir)
		first_word = 0;

	center = (begin - end) / 2;

	res = add_line_entry (dc, cell, lgrid, lsize, lwidth, 
			      begin, center, end, 
			      first_word, &info -> lspace);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status create_table (struct gr_grid *grid, 
				   struct gr_th_info *info,
				   void *closure)
{
	struct gr_al_machine *dc;
	struct gr_grid *line;
	enum s_status res;

	dc = closure;

	/* The enclosing "line". In cases of boxes, it is the
	 * box itself because it is placed directly into the column */
	line = info -> parent == info -> column ? grid : info -> parent;

	if (line != dc -> last_line) {
		dc -> last_line = line;
		res = next_line (line, grid, info, dc);
		if (res != S_CSC_OK)
			return res;
	}

	dc -> les -> has_graph_wspace 
		|= grid -> wspace.type == TD_ST_GRAPHICAL;

	return S_CSC_OK;
}

enum s_status create_para (struct gr_al_machine *dc)
{
	struct line_entry *le;
	int i;

	if (dc -> les == NULL) {
		dc -> paras = 0;
		dc -> avail = 0;
		return S_CSC_OK;
	}

	i = 0;
	le = dc -> les;
	do {
		le = le -> next;
		i++;
	} while (le != dc -> les);

	dc -> para = calloc (i, sizeof (struct paragraph));
	if (dc -> para == NULL)
		return S_CSC_OUT_OF_MEMORY;

	dc -> paras = 1;
	dc -> avail = i;
	dc -> para [0].le = dc -> les -> next;
	dc -> para [0].reason = "<init>";
	dc -> para [0].al.wdir = dc -> para [0].le -> wdir;
	dc -> para [0].al.ldir = dc -> para [0].le -> ldir;
	dc -> les -> next = NULL;
	dc -> les = NULL;

	return S_CSC_OK;
}

static void split (struct gr_al_machine *dc, int idx, 
		   struct paragraph *new_para,
		   struct line_entry **lep)
{
	idx++;
	if (idx < dc -> paras)
		memmove (&dc -> para [idx + 1],
			 &dc -> para [idx], 
			 (dc -> paras - idx) * sizeof (struct paragraph));

	dc -> paras++;

	dc -> para [idx] = *new_para;
	dc -> para [idx].le = *lep;
	dc -> para [idx].al.wdir = dc -> para [idx].le -> wdir;
	dc -> para [idx].al.ldir = dc -> para [idx].le -> ldir;
	*lep = NULL;

	assert (dc -> paras <= dc -> avail);
}

static pdf_bool_t split_by_sig (struct gr_al_machine *dc, 
				struct paragraph *para,
				struct line_entry *le,
				struct paragraph *new_para,
				pdf_bool_t cross,
				pdf_bool_t *is_hard)
{
	struct line_entry *nle;

	*is_hard = 1;

	nle = le -> next;
	if (nle == NULL)
		return 0;

	return !gr_sig_can_merge (&le -> sig, &nle -> sig, dc -> heur);
}

static pdf_bool_t split_by_lspace (struct gr_al_machine *dc, 
				   struct paragraph *para,
				   struct line_entry *le,
				   struct paragraph *new_para,
				   pdf_bool_t cross,
				   pdf_bool_t *is_hard)
{
	struct line_entry *nle;

	*is_hard = 1;

	nle = le -> next;
	if (le -> wdir != nle -> wdir ||
	    le -> ldir != nle -> ldir)
		return 1;

	switch (nle -> lspace.type) {

	case TD_ST_GRAPHICAL:
		return 1;

	case TD_ST_ORDINARY:
		return nle -> lspace.amount >= 
			nle -> lsize * dc -> heur -> par_break;
	}

	assert (0);
}

static pdf_bool_t split_by_wspace (struct gr_al_machine *dc, 
				   struct paragraph *para,
				   struct line_entry *le,
				   struct paragraph *new_para,
				   pdf_bool_t cross,
				   pdf_bool_t *is_hard)
{
	*is_hard = 1;

	return le -> has_graph_wspace;
}

static enum gr_al_align_type match_poles (struct gr_al_poles *poles,
					  struct line_entry *le1,
					  struct line_entry *le2)
{
	enum gr_al_align_type at;

	at = GR_AT_NONE;
	if (le1 -> poles.begin == le2 -> poles.begin) {
		at = GR_AT_BEGIN;
		poles -> begin = le2 -> poles.begin;
	} else
		poles -> begin = NULL;

	if (le1 -> poles.center == le2 -> poles.center) {
		if (at == GR_AT_NONE)
			at = GR_AT_CENTER;
		poles -> center = le2 -> poles.center;
	} else
		poles -> center = NULL;

	if (le1 -> poles.end == le2 -> poles.end) {
		if (at == GR_AT_BEGIN)
			at = GR_AT_JUSTIFIED;
		else
			at = GR_AT_END;
		poles -> end = le2 -> poles.end;
	} else
		poles -> end = NULL;

	return at;
}

/*
 * This function must be injective, so we know if the two score match
 * the poles match too.
 */
static int eval_poles (struct gr_al_poles *pole)
{
	int score;

	score = 0;
	
	score += pole -> begin != NULL ? 4 : 0;
	score += pole -> end != NULL ? 2 : 0;
	score += pole -> center != NULL ? 1 : 0;

	return score;
}

static int compare_poles (struct gr_al_poles *p1, struct gr_al_poles *p2)
{
	return eval_poles (p1) - eval_poles (p2);
}

static pdf_bool_t split_by_poles (struct gr_al_machine *dc, 
				  struct paragraph *para,
				  struct line_entry *ple,
				  struct paragraph *new_para,
				  pdf_bool_t cross,
				  pdf_bool_t *is_hard)
{
	struct line_entry *le, *nle;
	struct gr_al_poles poles; 
	enum gr_al_align_type at;
	int scdiff;

	le = ple -> next;
	if (le == NULL || cross)
		return 0;

	at = match_poles (&poles, ple, le);

	/* See if para wants initialization */
	if (para -> poles.begin == NULL && 
	    para -> poles.center == NULL && 
	    para -> poles.end == NULL) {

		if (at == GR_AT_NONE)
			return 1;

		para -> al.at = at;
		para -> poles = poles;
	}

	scdiff = compare_poles (&poles, &para -> poles);
	assert (scdiff <= 0);

	if (scdiff < 0)
		return 1;

	nle = le -> next;
	if (nle == NULL)
		return 0;
	
	match_poles (&poles, le, nle);
	scdiff = compare_poles (&poles, &para -> poles);
	if (scdiff > 0)
		return 1;
	
	return 0;
}

static int update_margin (struct gr_al_margin *dmargin,
			  struct gr_al_margin *smargin)
{
	if (smargin -> hi < dmargin -> hi)
		dmargin -> hi = smargin -> hi;

	if (smargin -> lo > dmargin -> lo)
		dmargin -> lo = smargin -> lo;

	return dmargin -> lo <= dmargin -> hi;
}

static int update_margins (struct gr_al_margins *dmargins, 
			   struct gr_al_margins *smargins)
{
	struct gr_al_margins scratch;

	scratch = *dmargins;

	if (!update_margin (&scratch.begin, &smargins -> begin) ||
	    !update_margin (&scratch.end, &smargins -> end))
		return 0;

	*dmargins = scratch;

	return 1;
}

static void mkmargin (struct gr_al_margin *margin,
		      pdf_real_t offset, pdf_real_t size)
{
	margin -> lo = offset - size;
	margin -> hi = offset;
}

static void pole2margin (struct gr_al_margin *margin, 
			 struct gr_al_pole *pole)
{
	mkmargin (margin, pole -> offset, 0);
}

static void pole2marginx (struct gr_al_margin *margin, 
			  struct gr_al_pole *pole)
{
	mkmargin (margin, pole -> offset, pole -> offset);
}

static void get_margins (struct gr_al_machine *dc,
			 struct paragraph *para,
			 struct line_entry *ple,
			 struct gr_al_margins *margins)
{
	pdf_real_t offset, size;

	if (para -> poles.begin != NULL) {
		if (para -> poles.end != NULL) {
			para -> al.at = GR_AT_JUSTIFIED;
			pole2marginx (&margins -> begin, ple -> poles.begin);
			pole2marginx (&margins -> end, ple -> poles.end);
		} else {
			para -> al.at = GR_AT_BEGIN;
			offset = ple -> poles.end -> offset;
			if (ple -> next != NULL)
				size = ple -> next -> first_word;
			else
				size = offset;

			pole2margin (&margins -> begin, ple -> poles.begin);
			mkmargin (&margins -> end, offset, size);
		}
	} else {
		if (para -> poles.end != NULL) {
			para -> al.at = GR_AT_END;
			offset = ple -> poles.begin -> offset;
			if (ple -> next != NULL)
				size = ple -> next -> first_word;
			else
				size = offset;

			mkmargin (&margins -> begin, offset, size);
			pole2margin (&margins -> end, ple -> poles.end);
		} else {
			para -> al.at = GR_AT_CENTER;
			mkmargin (&margins -> begin, 0, 0);
			mkmargin (&margins -> end, 0, 0);
		}
	}
}

static pdf_bool_t fword_is_compatible (struct paragraph *para,
				       struct line_entry *lle,
				       struct line_entry *fle)
{
	pdf_real_t trail_space;

	trail_space = 
		lle -> poles.end -> offset - 
		para -> margins.end.lo;
	return fle -> first_word >= trail_space;
}

static pdf_bool_t split_by_margins (struct gr_al_machine *dc, 
				    struct paragraph *para,
				    struct line_entry *ple,
				    struct paragraph *new_para,
				    pdf_bool_t cross,
				    pdf_bool_t *is_hard)
{
	struct gr_al_margins margins;

	if (cross)
		return 0;

	if (!para -> margins_set) {
		get_margins (dc, para, ple, &para -> margins);
		para -> margins_set = 1;
	}

	get_margins (dc, para, ple -> next, &margins);

	if (!update_margins (&para -> margins, &margins) ||
	    !fword_is_compatible (para, ple, ple -> next)) {
		new_para -> poles = para -> poles;
		return 1;
	}
		
	return 0;
}

typedef pdf_bool_t split_algorithm (struct gr_al_machine *dc,
				    struct paragraph *para,
				    struct line_entry *le,
				    struct paragraph *new_para,
				    pdf_bool_t cross,
				    pdf_bool_t *is_hard);

static void run_split_algorithm (struct gr_al_machine *dc,
				 split_algorithm sa,
				 const char *reason)
{
	struct line_entry *le, **nle;
	struct paragraph new_para;
	pdf_bool_t is_hard, cross;
	int i;

	for (i = 0; i < dc -> paras; i++) {
		memset (&new_para, 0, sizeof (struct paragraph));
		new_para.reason = reason;
		for (le = dc -> para [i].le; le -> next != NULL; 
		     le = le -> next) {
			is_hard = 0;
			if (le -> next == NULL && i < dc -> paras - 1) {
				nle = &dc -> para [i + 1].le;
				cross = 1;
			} else {
				nle = &le -> next;
				cross = 0;
			}
			if (sa (dc, &dc -> para [i], le, 
				&new_para, cross, &is_hard)) {
				dc -> para [i].hard_split = is_hard;
				if (!cross)
					split (dc, i, &new_para, nle);
				break;
			}
		}
	}
}

static void detach_unused_poles (struct gr_al_machine *dc)
{
	struct line_entry *le;
	int i;

	for (i = 0; i < dc -> paras; i++) {
		for (le = dc -> para [i].le; le != NULL; le = le -> next) {
			if (le -> poles.begin != dc -> para [i].poles.begin)
				le -> poles.begin -> occurrences--;
			if (le -> poles.end != dc -> para [i].poles.end)
				le -> poles.end -> occurrences--;
			if (le -> poles.center != dc -> para [i].poles.center)
				le -> poles.center -> occurrences--;
		}
	}
}

static enum gr_al_align_type guess_ol_at (struct paragraph *para)
{
	pdf_real_t obegin, oend, ocenter;
	struct line_entry *le;

	le = para -> le;

	obegin = le -> poles.begin -> occurrences;
	oend = le -> poles.end -> occurrences;
	ocenter = le -> poles.center -> occurrences;

	if (para -> le -> has_graph_wspace) {
		para -> poles.end = le -> poles.end;
		para -> poles.begin = le -> poles.begin;
		return GR_AT_JUSTIFIED;
	}
	
	if (ocenter > obegin && ocenter > oend) {
		para -> poles.center = le -> poles.center;
		return GR_AT_CENTER;
	}

	if (obegin >= oend) {
		para -> poles.begin = le -> poles.begin;
		return GR_AT_BEGIN;
	} else {
		para -> poles.end = le -> poles.end;
		return GR_AT_END;
	}
}

static pdf_bool_t compat (pdf_real_t ofs1, struct gr_al_pole *pole2)
{
	return pole2 != NULL && ofs1 >= pole2 -> offset;
}

static void guess_ol_margins (struct gr_al_machine *dc, int idx)
{
	struct paragraph *para, *lpara, *hpara;
	struct line_entry *le;
	pdf_real_t lofs, hofs;
	int lo, hi;

	para = &dc -> para [idx];
	le = para -> le;
	
	para -> margins.begin.hi = le -> poles.begin -> offset;
	para -> margins.end.hi = le -> poles.end -> offset;

	switch (para -> al.at) {
	case GR_AT_BEGIN:
		para -> margins.begin.lo = 0;
		lo = hi = idx;
		lofs = hofs = le -> poles.end -> offset;
		while (lo >= 0 || hi < dc -> paras) {
			lpara = &dc -> para [lo];
			if (lo >= 0 && lpara -> margins_set &&
			    lpara -> margins.end.hi <= lofs) {
				if (lpara -> poles.begin == 
				    le -> poles.begin) {
					para -> margins.end.lo = 
						lpara -> margins.end.lo;
					return;
				}
				lofs = lpara -> margins.end.hi;
			}

			hpara = &dc -> para [hi];
			if (hi < dc -> paras && hpara -> margins_set &&
			    hpara -> margins.end.hi <= hofs) {
				if (hpara -> poles.begin == 
				    le -> poles.begin &&
				    hofs >= hpara -> margins.end.hi) {
					para -> margins.end.lo = 
						hpara -> margins.end.lo;
					return;
				}
				hofs = hpara -> margins.end.hi;
			}

			lo--, hi++;
		}

		lo = hi = idx;
		while (lo >= 0 || hi < dc -> paras) {
			lpara = &dc -> para [lo];
			if (lo >= 0 && lpara -> margins_set) {

				para -> margins.end.lo = 
					lpara -> margins.end.lo;
				return;
			}

			hpara = &dc -> para [hi];
			if (hi < dc -> paras && hpara -> margins_set &&
			    hpara -> margins.end.hi <= 
			    le -> poles.end -> offset) {

				para -> margins.end.lo = 
					hpara -> margins.end.lo;
				return;
			}
			lo--, hi++;
		}

		/*
		if (max (obegin, oend) <=  2 * min (obegin, oend))
			para -> at = GR_AT_JUSTIFIED;
		*/

		break; 

	case GR_AT_END:
		para -> margins.end.lo = 0;
		lo = hi = idx;
		lofs = hofs = le -> poles.begin -> offset;
		while (lo >= 0 || hi < dc -> paras) {
			lpara = &dc -> para [lo];
			if (lo >= 0 && lpara -> margins_set) {
				if (lpara -> poles.end == le -> poles.end &&
				    compat (lofs, lpara -> poles.begin)) {
					para -> margins.begin.lo = 
						lpara -> margins.begin.lo;
					return;
				}
				if (lpara -> margins.begin.hi <= lofs)
					lofs = lpara -> margins.begin.hi;
			}

			hpara = &dc -> para [hi];
			if (hi < dc -> paras && hpara -> margins_set) {
				if (hpara -> poles.end == le -> poles.end &&
				    compat (hofs, hpara -> poles.begin)) {
					para -> margins.begin.lo = 
						hpara -> margins.begin.lo;
					return;
				}
				if (hpara -> margins.begin.hi <= hofs)
					hofs = hpara -> margins.begin.hi;
			}
			lo--, hi++;
		}

		lo = hi = idx;
		lofs = hofs = le -> poles.end -> offset;
		while (lo >= 0 || hi < dc -> paras) {
			lpara = &dc -> para [lo];
			if (lo >= 0 && lpara -> margins_set) {
				if (compat (lofs, lpara -> poles.begin)) {
					para -> margins.end.lo = 
						lpara -> margins.end.lo;
					return;
				}
				if (lpara -> margins.begin.hi <= lofs)
					lofs = lpara -> margins.begin.hi;
			}

			hpara = &dc -> para [hi];
			if (hi < dc -> paras && hpara -> margins_set) {
				if (compat (hofs, hpara -> poles.begin)) {
					para -> margins.begin.lo = 
						hpara -> margins.begin.lo;
					return;
				}
				if (hpara -> margins.begin.hi <= hofs)
					hofs = hpara -> margins.begin.hi;
			}
			lo--, hi++;
		}

		/*
		if (max (obegin, oend) <=  2 * min (obegin, oend))
			para -> at = GR_AT_JUSTIFIED;
		*/

		break;

	case GR_AT_CENTER:
		para -> margins.begin.lo = 0;
		para -> margins.end.lo = 0;
		break;

	case GR_AT_JUSTIFIED:
		assert (para -> le -> has_graph_wspace);
		para -> margins.begin.lo = para -> margins.begin.hi;
		para -> margins.end.lo = para -> margins.end.hi;
		break;

	case GR_AT_NONE:
	case GR_AT_UNKNOWN:
	case GR_AT_PROPORTIONAL:
	case GR_AT__LAST__:
		assert (0);
	}
}

static int by_len (const void *p1, const void *p2)
{
	struct paragraph *para1, *para2;
	pdf_real_t air1, air2;

	para1 = *(struct paragraph **) p1;
	para2 = *(struct paragraph **) p2;

	air1 = 	para1 -> le -> poles.begin -> offset +
		para1 -> le -> poles.end -> offset;
	air2 = 	para2 -> le -> poles.begin -> offset +
		para2 -> le -> poles.end -> offset;

	return air1 - air2;
}

static void guess_one_liners (struct gr_al_machine *dc)
{
	struct paragraph *spara [dc -> paras];
	int i, j;

	/* First of all guess their alignment */
	j = 0;
	for (i = 0; i < dc -> paras; i++) {
		if (dc -> para [i].le -> next == NULL) {
			spara [j++] = &dc -> para [i];
			dc -> para [i].hard_split |= 
				dc -> para [i].le -> has_graph_wspace;
			dc -> para [i].al.at = guess_ol_at (&dc -> para [i]);
			dc -> para [i].margins_set = 0;
		}
	}

	if (j) {
		qsort (spara, j, sizeof (struct paragraph *), by_len);

		/* Now inherit margins from paragraphs nearby */
		for (i = 0; i < j; i++) {
			guess_ol_margins (dc, spara [i] - dc -> para);
			spara [i] -> margins_set = 1;
		}
	}
}

static pdf_bool_t can_merge_last_line (struct gr_al_machine *dc,
				       int idx, enum gr_al_align_type *at)
{
	struct paragraph *para;
	struct line_entry *fle, *lle;
	pdf_bool_t align_begin, align_end;

	if (idx + 1>= dc -> paras)
		return 0;

	if (dc -> para [idx + 1].settled || 
	    dc -> para [idx + 1].solidity >= 1)
		return 0;

	para = &dc -> para [idx];
	fle = para [1].le;
	lle = para -> le;
	while (lle -> next != NULL)
		lle = lle -> next;

	if (para -> hard_split || !fword_is_compatible (para, lle, fle))
		return 0;

 	align_begin = 
		lle -> poles.begin == fle -> poles.begin &&
		compat (fle -> poles.end -> offset, lle -> poles.end);
	align_end = 
		lle -> poles.end == fle -> poles.end &&
		compat (fle -> poles.begin -> offset, lle -> poles.begin);

	if (at != NULL)
		*at = para -> al.at;

	switch (para -> al.at) {
	case GR_AT_NONE:
	case GR_AT_UNKNOWN:
		assert (0);
		break;

	case GR_AT_BEGIN:
		return align_begin;

	case GR_AT_END:
		return align_end;

	case GR_AT_JUSTIFIED:
		if (at != NULL) {
			if (align_begin)
				*at = GR_AT_BEGIN;
			else if (align_end)
				*at = GR_AT_END;
		}

		return align_begin || align_end;

	case GR_AT_CENTER:
		break;

	case GR_AT_PROPORTIONAL:
		/* Never generated */
		assert (0);
		break;

	case GR_AT__LAST__:
		assert (0);
	}

	return 0;
}


static pdf_bool_t can_merge_first_line (struct gr_al_machine *dc,
					int idx, enum gr_al_align_type *at)
{
	struct paragraph *para;
	struct line_entry *fle, *lle;
	pdf_bool_t align_begin, align_end;

	if (!idx)
		return 0;

	if (dc -> para [idx - 1].settled || 
	    dc -> para [idx - 1].solidity >= 1)
		return 0;

	para = &dc -> para [idx];
	fle = para -> le;
	lle = para [-1].le;
	while (lle -> next != NULL)
		lle = lle -> next;

	if (para [-1].hard_split || !fword_is_compatible (para, lle, fle))
		return 0;

 	align_begin = 
		labs (lle -> poles.begin -> offset - 
		      fle -> poles.begin -> offset) 
		< (dc -> heur -> max_indent * lle -> lsize);
	if (fle -> next == NULL)
		align_begin = align_begin &&
			lle -> poles.begin != fle -> poles.begin;
	else
		align_begin = align_begin && 
			compat (lle -> poles.end -> offset, fle -> poles.end);
	align_end = 
		lle -> poles.end == fle -> poles.end &&
		compat (lle -> poles.begin -> offset, fle -> poles.begin);

	if (at != NULL)
		*at = para -> al.at;

	switch (para -> al.at) {
	case GR_AT_NONE:
	case GR_AT_UNKNOWN:
		assert (0);
		break;

	case GR_AT_BEGIN:
		return align_begin;

	case GR_AT_END:
		return align_end;

	case GR_AT_JUSTIFIED:
		return align_begin && align_end;

	case GR_AT_CENTER:
		return 0;

	case GR_AT_PROPORTIONAL:
		/* Never generated */
		assert (0);
		break;

	case GR_AT__LAST__:
		assert (0);
	}

	return 0;
}

static int para_height (struct paragraph *para)
{
	struct line_entry *le;
	int height;

	height = 0;
	for (le = para -> le; le != NULL; le = le -> next)
		height++;

	return height;
}

static int para_width (struct paragraph *para)
{
	struct line_entry *le;
	pdf_real_t width;

	width = para -> le -> lwidth;
	for (le = para -> le; le != NULL; le = le -> next)
		width = min (width, le -> lwidth);

	return width;
}

static void break_para (struct gr_al_machine *dc, int *idx)
{
	while (dc -> para [*idx].le -> next != NULL) {
		split (dc, *idx, &dc -> para [*idx], 
		       &dc -> para [*idx].le -> next);
		(*idx)++;
	}

	(*idx)++;
}

static void settle_ol_para (struct gr_al_machine *dc, int idx)
{
	struct paragraph *para;

	para = &dc -> para [idx];
	para -> al.size = 1;
	para -> al.fhead = dc -> para [idx].margins.begin.hi;
	para -> al.ftail = dc -> para [idx].margins.end.hi;
	para -> al.lhead = dc -> para [idx].margins.begin.hi;
	para -> al.ltail = dc -> para [idx].margins.end.hi;
	para -> al.flush_at = para -> al.at;
	if (para -> al.flush_at == GR_AT_JUSTIFIED)
		para -> al.flush_at = GR_AT_BEGIN;
	para -> settled = 1;
}

static void apply_breaking_model (struct gr_al_machine *dc,
				  int idx, struct gr_al_model *glm)
{
	struct paragraph *para;
	int lidx;

	lidx = idx;
	break_para (dc, &lidx);
	while (idx < lidx) {
		para = &dc -> para [idx];
		para -> al = glm -> al;
		para -> margins.begin.hi = para -> al.fhead;
		para -> margins.begin.lo = para -> al.fhead;
		para -> margins.end.hi = para -> al.ftail;
		para -> margins.end.lo = para -> al.ftail;
		idx++;
	}
}

static pdf_real_t solidity (struct gr_al_machine *dc, int idx)
{
	struct paragraph *para;
	pdf_real_t sol, hr;

	para = &dc -> para [idx];

	sol = -1;
	switch (para -> al.at) {
	case GR_AT_NONE:
	case GR_AT_PROPORTIONAL:
	case GR_AT_UNKNOWN:
	case GR_AT__LAST__:
		assert (0);
		break;

	case GR_AT_BEGIN:
		sol = SN_AT_BEGIN;
		break;

	case GR_AT_END:
		sol = SN_AT_END;
		break;

	case GR_AT_CENTER:
		sol = SN_AT_CENTER;
		break;

	case GR_AT_JUSTIFIED:
		sol = SN_AT_JUSTIFIED;
	}

	if (can_merge_first_line (dc, idx, NULL))
		sol += SC_INDENT * (1 - dc -> para [idx - 1].solidity);

	hr = para_width (para) + para -> margins.begin.hi +
		para -> margins.end.hi;

	sol += SC_WIDTH * para_width (para) / hr;
	sol += SC_HEIGHT * (para_height (para) - 1);

	return sol;
}

static void jerry_build_para (struct gr_al_machine *dc, int fidx,
			      int midx, int lidx, struct gr_al_alignment *al)
{
	int i;

	if (midx != fidx && midx != lidx) {
		/* Canonical three-line para */
		if (dc -> para [fidx].poles.end == 
		    dc -> para [midx].poles.end &&
		    dc -> para [midx].poles.begin ==
		    dc -> para [lidx].poles.begin)
			al -> at = GR_AT_JUSTIFIED;
		else
			al -> at = dc -> para [midx].al.at;
	} else {
		/* One or two-line para */
		for (i = fidx - 1; i >= 0; i--) {
			if (dc -> para [i].settled &&
			    dc -> para [i].al.flush_at == 
			    dc -> para [midx].al.at) {
				al -> at = dc -> para [i].al.at;
				break;
			}
		}

		if (i < 0)
			al -> at = dc -> last_al [dc -> para [midx].al.at];
	}
}

static enum s_status settle_para (struct gr_al_machine *dc, int idx)
{
	struct gr_al_alignment *al;
	enum gr_al_align_type at;
	struct gr_al_model *glm;
	enum s_status res;
	int fidx, midx, lidx, size;

	if (can_merge_first_line (dc, idx, &at)) {
		fidx = idx - 1;
		dc -> para [fidx].al.at = at;
		dc -> para [fidx].merge_with_next = 1;
	} else
		fidx = idx;
	midx = idx;
	if (can_merge_last_line (dc, idx, &at)) {
		lidx = idx + 1;
		dc -> para [midx].merge_with_next = 1;
		dc -> para [lidx].al.at = at;
	} else
		lidx = idx;

	al = &dc -> para [fidx].al;

	al -> size = para_height (&dc -> para [midx]);
	if (fidx != midx)
		al -> size += para_height (&dc -> para [fidx]);
	if (lidx != midx)
		al -> size += para_height (&dc -> para [lidx]);
	
	al -> at = dc -> para [midx].al.at;
	al -> fhead = dc -> para [fidx].margins.begin.hi;
	al -> ftail = dc -> para [fidx].margins.end.lo;
	al -> wdir = dc -> para [midx].le -> wdir;
	al -> ldir = dc -> para [midx].le -> ldir;
	al -> flush_at = dc -> para [lidx].al.at;
	if (al -> flush_at == GR_AT_JUSTIFIED)
		al -> flush_at = GR_AT_BEGIN;
	al -> lhead = dc -> para [midx].margins.begin.hi;
	al -> ltail = dc -> para [midx].margins.end.lo;

	if (para_height (&dc -> para [midx]) <= 2) {
		glm = find_model (dc, al);
		if (glm != NULL) {
			size = al -> size;
			*al = glm -> al;
			al -> size = size;
			glm -> occurrences++;
		} else
			jerry_build_para (dc, fidx, midx, lidx, al);
	} else {
		res = add_model (dc, al);
		if (res != S_CSC_OK)
			return res;
	}

	dc -> para [fidx].settled = 1;
	dc -> para [midx].settled = 1;
	dc -> para [lidx].settled = 1;

	return S_CSC_OK;
}

static enum s_status merge (struct gr_al_machine *dc)
{
	struct gr_al_model *glm;
	pdf_real_t best_sol;
	int i, best;
	enum s_status res;

	for (i = 0; i < dc -> paras; i++) {
		dc -> para [i].solidity = solidity (dc, i); 
		if (dc -> para [i].al.at == GR_AT_CENTER)
			settle_ol_para (dc, i);
	}

	while (1) {
		best = -1;
		best_sol = -1;
		for (i = 0; i < dc -> paras; i++) {
			if (!dc -> para [i].settled &&
			    dc -> para [i].solidity > best_sol) {
				best = i;
				best_sol = dc -> para [i].solidity;
			}
		}
		if (best < 0)
			break;

		if (best_sol < 1 && dc -> para [best].le -> next != NULL) {
			glm = find_breaking_model (dc, &dc -> para [best]);
			if (glm != NULL) {
				apply_breaking_model (dc, best, glm);
				continue;
			}			
		}

		res = settle_para (dc, best);
		if (res != S_CSC_OK)
			return res;
	}

	return S_CSC_OK;
}

static void guess (struct gr_al_machine *dc)
{
	if (!dc -> paras)
		return;

	run_split_algorithm (dc, split_by_sig, "sig");
	run_split_algorithm (dc, split_by_lspace, "lspace");
	run_split_algorithm (dc, split_by_wspace, "wspace");
	run_split_algorithm (dc, split_by_poles, "poles");
	run_split_algorithm (dc, split_by_margins, "margins");
	
	detach_unused_poles (dc);
	guess_one_liners (dc);

	merge (dc);
}

static void update_grid (struct gr_al_machine *dc)
{
	struct gr_grid *grid;
	int i;

	i = 0;
	for (i = 0; i < dc -> paras; i++) {
		if (!i || !dc -> para [i - 1].merge_with_next) {
			grid = dc -> para [i].le -> grid;
			grid -> al = dc -> para [i].al;
			if (dc -> rfp != NULL)
				dump_align (dc -> rfp, &grid -> al);
		}
	}
}

static void update_last_al (struct gr_al_machine *dc)
{
	enum gr_al_align_type at;
	int i;

	for (at = 0; at < GR_AT__LAST__; at++)
		for (i = dc -> paras - 1; i >= 0; i--)
			if (dc -> para [i].al.flush_at == at) {
				dc -> last_al [at] = dc -> para [i].al.at;
				break;
			}
}

enum s_status gr_al_detect_alignment (struct gr_al_machine *dc,
				      struct gr_grid *grid, 
				      int page)
{
	enum s_status res;

	do_decad (dc);

	destroy_para_table (dc);

	res = gr_th_traverse (grid, dc -> heur, create_table, dc);
	if (res != S_CSC_OK)
		return res;

	res = create_para (dc);
	if (res != S_CSC_OK)
		return res;

	guess (dc);

	update_last_al (dc);

	if (dc -> rfp != NULL)
		fprintf (dc -> rfp, "-- %d\n", page);

	update_grid (dc);

	return S_CSC_OK;
}
