/*
 * PDFlow -- area.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <assert.h>
#include <string.h>
#include "fmt/ctab.h"
#include "grid/heur.h"
#include "grid/area.h"

static enum s_status text_copy (struct a_area *dest, struct a_area *src);
static void text_destroy (struct a_area *a);
static void text_get_direction (struct a_area *area,
				struct args_heur *heur,
				enum gs_direction *pdir,
				enum gs_direction *sdir);
static void text_get_origin (struct a_area *area,
			     enum gs_direction wdir,
			     struct args_heur *heur,
			     struct gs_point *origin,
			     pdf_bool_t *has_origin);
static void text_get_ratios (struct a_area *area, pdf_real_t *ratio,
			     pdf_real_t *badness, pdf_real_t *bad_slope,
			     struct args_output *aout);
static void text_get_output_bounds (struct a_area *area, 
				    enum gs_direction wdir,
				    pdf_real_t pspace,
				    struct args_output *aout,
				    struct cobj_rect *bounds);
static void text_query_info (struct a_area *area, enum gs_direction wdir,
			     struct a_info *info);
static enum s_status text_draw (struct a_area *area,
				struct agc_context *agc);

static enum s_status graph_copy (struct a_area *dest, struct a_area *src);
static void graph_destroy (struct a_area *a);
static void graph_get_ratios (struct a_area *area, pdf_real_t *ratio,
			      pdf_real_t *badness, pdf_real_t *bad_slope,
			      struct args_output *aout);
static enum s_status graph_draw (struct a_area *area,
				 struct agc_context *agc);

static enum s_status xo_copy (struct a_area *dest, struct a_area *src);
static void xo_destroy (struct a_area *a);
static void xo_get_ratios (struct a_area *area, pdf_real_t *ratio,
			   pdf_real_t *badness, pdf_real_t *bad_slope,
			   struct args_output *aout);
static enum s_status xo_draw (struct a_area *area,
			      struct agc_context *agc);

static enum s_status sh_copy (struct a_area *dest, struct a_area *src);
static void sh_destroy (struct a_area *a);
static void sh_get_ratios (struct a_area *area, pdf_real_t *ratio,
			   pdf_real_t *badness, pdf_real_t *bad_slope,
			   struct args_output *aout);
static enum s_status sh_draw (struct a_area *area,
			      struct agc_context *agc);

static void group_get_direction (struct a_area *area,
				 struct args_heur *heur,
				 enum gs_direction *pdir,
				 enum gs_direction *sdir);
static void group_get_origin (struct a_area *area,
			      enum gs_direction wdir,
			      struct args_heur *heur,
			      struct gs_point *origin,
			      pdf_bool_t *has_origin);
static void group_get_ratios (struct a_area *area, pdf_real_t *ratio,
			      pdf_real_t *badness, pdf_real_t *bad_slope,
			      struct args_output *aout);
static void group_query_info (struct a_area *area, enum gs_direction wdir,
			      struct a_info *info);
static void group_get_output_bounds (struct a_area *area, 
				     enum gs_direction wdir,
				     pdf_real_t pspace,
				     struct args_output *aout,
				     struct cobj_rect *bounds);

static struct a_ifc page_ifc = {
	"page", 'P', { 1., 0., 1. }, 
	{  }, NULL
};

static struct a_ifc graph_ifc = {
	"graph", 'G', { 0., 1., 0. }, 
	{ atomic : 1 }, 
	graph_copy, graph_draw, graph_get_ratios, NULL,
	NULL, NULL, NULL, graph_destroy
};

static struct a_ifc text_ifc = {
	"text", 'T', { 1., 0., 0. }, 
	{ atomic : 1 }, 
	text_copy, text_draw, text_get_ratios, text_get_direction, 
	text_get_origin, text_get_output_bounds, text_query_info, text_destroy
};

static struct a_ifc shading_ifc = {
	"shading", 'S', { 1., 0.3, 0.9 }, 
	{ atomic : 1 }, 
	sh_copy, sh_draw, sh_get_ratios, NULL, NULL, NULL, NULL, sh_destroy
};

static struct a_ifc xo_form_ifc = {
	"xo_form", 'F', { 1., 0.9, 0.3 }, 
	{ atomic : 1 }, 
	xo_copy, xo_draw, xo_get_ratios, NULL, NULL, NULL, NULL, xo_destroy
};

static struct a_ifc xo_image_ifc = {
	"xo_image", 'I', { 0.83, 1, 0 }, 
	{ atomic : 1 },
	xo_copy, xo_draw, xo_get_ratios, NULL, NULL, NULL, NULL, xo_destroy
};

static struct a_ifc group_ifc = {
	"group", 'P', { 0., 0., 1. }, 
	{ atomic : 1, logic : 1 }, 
	NULL, NULL, group_get_ratios, group_get_direction, 
	group_get_origin, group_get_output_bounds, group_query_info
};

static enum s_status insert_area (struct a_area *parent, struct a_area *area);

static enum s_status text_copy (struct a_area *dest, struct a_area *src)
{
	struct a_text_area *sta, *dta;

	sta = &src -> u.text;
	dta = &dest -> u.text;

	*dta = *sta;

	T_CALL (gs_text_ref, dta -> tstate, ref_rown (&dest -> refs));
	T_CALL (gs_graph_ref, dta -> gstate, ref_rown (&dest -> refs));

	return S_CSC_OK;
}

static void text_get_ratios (struct a_area *area, pdf_real_t *ratio,
			     pdf_real_t *badness, pdf_real_t *bad_slope,
			     struct args_output *aout)
{
	*ratio = aout -> text_ratio;
	*badness = HEUR_BADNESS_LAST_RESORT;
	*bad_slope = 0;
}

static void text_get_output_bounds (struct a_area *area, 
				    enum gs_direction wdir,
				    pdf_real_t pspace,
				    struct args_output *aout,
				    struct cobj_rect *bounds)
{
	pdf_real_t rlead, space, ascent, descent;
	struct gf_font *font;
	struct a_text_area *ta;
	struct gs_point origin;

	ta = &area -> u.text;

	font = ta -> tstate -> font -> font;

	origin.x = ta -> ginfo.t_m.a [2][0];
	origin.y = ta -> ginfo.t_m.a [2][1];

	*bounds = area -> g.rect;

	gs_revert_coords (&origin, &ta -> gstate -> ctm);
	gs_revert_coords (&origin, &ta -> tstate -> mat.t_m);
	gs_revert_rect (bounds, &ta -> gstate -> ctm);
	gs_revert_rect (bounds, &ta -> tstate -> mat.t_m);

	rlead = (font -> leading - (font -> ascent + font -> descent))*
		ta -> tstate -> font -> size;
       
	space = aout -> wspace_width > 0 ? 
		aout -> wspace_width : font -> space_width;
	space *= ta -> tstate -> font -> size;
	ascent = font -> ascent * ta -> tstate -> font -> size;
	descent = font -> descent * ta -> tstate -> font -> size;

	if (font -> wmode == CMAP_WM_HORIZONTAL) {
		bounds -> left -= pspace * space;
		if (bounds -> bottom - origin.y > descent)
			bounds -> bottom = origin.y + descent;
		if (bounds -> top - origin.y < ascent)
			bounds -> top = origin.y + ascent;
		bounds -> bottom -= rlead / 2;
		bounds -> top += rlead / 2;
	} else {
		bounds -> bottom -= pspace * space;
		if (bounds -> right - origin.x < descent)
			bounds -> right = origin.y + descent;
		if (bounds -> left - origin.x > ascent)
			bounds -> left = origin.x + ascent;
		bounds -> left -= rlead / 2;
		bounds -> right += rlead / 2;
	}

	gs_translate_rect (bounds, &ta -> tstate -> mat.t_m);
	gs_translate_rect (bounds, &ta -> gstate -> ctm);
}

static enum s_status text_draw (struct a_area *area,
				struct agc_context *agc)

{
	struct a_text_area *ta;
	enum s_status res;

	ta = &area -> u.text;

	res = agc_set_state (agc, ta -> tstate, ta -> gstate,
			     &area -> g.rect);	
	if (res != S_CSC_OK)
		return res;

	res = agc_write_char (agc, &ta -> ginfo, &ta -> paint);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static void text_get_direction (struct a_area *area,
				struct args_heur *heur,
				enum gs_direction *pdir,
				enum gs_direction *sdir)
{
	struct a_text_area *ta;

	ta = &area -> u.text;

	*pdir = ta -> primary;
	*sdir = ta -> secundary;
}

static void text_get_origin (struct a_area *area,
			     enum gs_direction wdir,
			     struct args_heur *heur,
			     struct gs_point *origin,
			     pdf_bool_t *has_origin)
{
	struct a_text_area *ta;

	ta = &area -> u.text;

	if (has_origin != NULL)
		*has_origin = 1;
	
	origin -> x = ta -> ginfo.t_m.a [2][0];
	origin -> y = ta -> ginfo.t_m.a [2][1];

	gs_translate_coords (origin, &ta -> gstate -> ctm);
}

static void text_query_info (struct a_area *area,
			     enum gs_direction wdir,
			     struct a_info *info)
{
	struct a_text_area *ta;

	ta = &area -> u.text;

	info -> hyphenized = 
		wdir == ta -> primary &&
		ct_is_hyphen (ta -> ginfo.f.unicode);
	info -> fill = ct_is_fill (ta -> ginfo.f.unicode);
	if (info -> hyphenized)
		info -> hyphen = area;
}

static void text_destroy (struct a_area *a)
{
	struct a_text_area *ta;

	ta = &a -> u.text;

	gs_text_unref (ta -> tstate, ref_rown (&a -> refs));
	gs_graph_unref (ta -> gstate, ref_rown (&a -> refs));
}

static enum s_status graph_copy (struct a_area *dest, struct a_area *src)
{
	struct a_graph_area *sga, *dga;

	sga = &src -> u.graph;
	dga = &dest -> u.graph;

	*dga = *sga;

	T_CALL (gs_graph_ref, dga -> gstate, ref_rown (&dest -> refs));

	return S_CSC_OK;
}

static void graph_destroy (struct a_area *a)
{
	struct a_graph_area *ga;

	ga = &a -> u.graph;

	gs_graph_unref (ga -> gstate, ref_rown (&a -> refs));
	gs_subpath_destroy (&ga -> sp, ref_rown (&a -> refs));
}

static void graph_get_ratios (struct a_area *area, pdf_real_t *ratio,
			      pdf_real_t *badness, pdf_real_t *bad_slope,
			      struct args_output *aout)
{
	*ratio = aout -> graph_ratio;
	*badness = 0;
	*bad_slope = aout -> graph_bad_slope;
}

static enum s_status graph_draw (struct a_area *area,
				 struct agc_context *agc)
{
	struct a_graph_area *ga;
	enum s_status res;

	ga = &area -> u.graph;

	res = agc_set_state (agc, NULL, ga -> gstate,
			     &area -> g.rect);	
	if (res != S_CSC_OK)
		return res;

	res = agc_draw_subpath (agc, &ga -> sp, &ga -> paint);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status xo_copy (struct a_area *dest, struct a_area *src)
{
	struct a_xo_area *sxa, *dxa;

	sxa = &src -> u.xo;
	dxa = &dest -> u.xo;

	*dxa = *sxa;

	T_CALL (gs_graph_ref, dxa -> gstate, ref_rown (&dest -> refs));
	T_CALL (xo_ref, dxa -> xo, ref_rown (&dest -> refs));

	return S_CSC_OK;
}

static void xo_destroy (struct a_area *a)
{
	struct a_xo_area *xa;

	xa = &a -> u.xo;

	gs_graph_unref (xa -> gstate, ref_rown (&a -> refs));
	xo_unref (xa -> xo, ref_rown (&a -> refs));
}

static void xo_get_ratios (struct a_area *area, pdf_real_t *ratio,
			   pdf_real_t *badness, pdf_real_t *bad_slope,
			   struct args_output *aout)
{
	*ratio = aout -> graph_ratio;
	*badness = 0;
	*bad_slope = aout -> graph_bad_slope;
}

static enum s_status xo_draw (struct a_area *area,
			      struct agc_context *agc)
{
	struct a_xo_area *xa;
	enum s_status res;

	xa = &area -> u.xo;

	res = agc_set_state (agc, NULL, xa -> gstate,
			     &area -> g.rect);	
	if (res != S_CSC_OK)
		return res;

	res = agc_draw_xobject (agc, xa -> xo);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status sh_copy (struct a_area *dest, struct a_area *src)
{
	struct a_sh_area *ssa, *dsa;

	ssa = &src -> u.sh;
	dsa = &dest -> u.sh;

	*dsa = *ssa;

	T_CALL (gs_graph_ref, dsa -> gstate, ref_rown (&dest -> refs));
	T_CALL (shd_ref, dsa -> sh, ref_rown (&dest -> refs));

	return S_CSC_OK;
}

static void sh_destroy (struct a_area *a)
{
	struct a_sh_area *sa;

	sa = &a -> u.sh;

	gs_graph_unref (sa -> gstate, ref_rown (&a -> refs));
	shd_unref (sa -> sh, ref_rown (&a -> refs));
}

static void sh_get_ratios (struct a_area *area, pdf_real_t *ratio,
			   pdf_real_t *badness, pdf_real_t *bad_slope,
			   struct args_output *aout)
{
	*ratio = aout -> graph_ratio;
	*badness = 0;
	*bad_slope = aout -> graph_bad_slope;
}

static enum s_status sh_draw (struct a_area *area,
			      struct agc_context *agc)
{
	struct a_sh_area *sa;
	enum s_status res;

	sa = &area -> u.sh;

	res = agc_set_state (agc, NULL, sa -> gstate,
			     &area -> g.rect);	
	if (res != S_CSC_OK)
		return res;

	res = agc_draw_shading (agc, sa -> sh);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static void group_get_ratios (struct a_area *area, pdf_real_t *ratio,
			      pdf_real_t *badness, pdf_real_t *bad_slope,
			      struct args_output *aout)
{
	pdf_real_t c_ratio, c_badness, c_bad_slope;
	struct a_area *a;

	if (area -> child == NULL) {
		*ratio = *badness = *bad_slope = 0;
		return;
	}

	*ratio = 0.01;
	*badness = 0;
	*bad_slope = 0;
	a = area -> child;
	do {
		a = a -> next;
		if (a -> ifc -> get_ratios != NULL) {
			a -> ifc -> get_ratios (a, &c_ratio, &c_badness,
						&c_bad_slope, aout);
			if (c_ratio > *ratio)
				*ratio = c_ratio;
			if (c_badness > *badness)
				*badness = c_badness;
			if (c_bad_slope > *bad_slope)
				*bad_slope = c_bad_slope;
		}
	} while (a != area -> child);
}

static void group_get_direction (struct a_area *area,
				 struct args_heur *heur,
				 enum gs_direction *pdirp,
				 enum gs_direction *sdirp)
{
	enum gs_direction pdir, sdir;
	struct heur_dir_chooser hc;
	pdf_bool_t certain;
	struct a_area *a;

	if (area -> child == NULL) {
		*pdirp = *sdirp = GS_D_UNKNOWN;
		return;
	}

	heur_dc_init (&hc, heur);

	a = area -> child;
	do {
		a = a -> next;
		a_get_direction (a, heur, &pdir, &sdir);
		heur_dc_add (&hc, pdir, sdir);
	} while (a != area -> child);

	heur_dc_choose (&hc, pdirp, sdirp, &certain);
}

static void group_get_origin (struct a_area *area,
			      enum gs_direction wdir,
			      struct args_heur *heur,
			      struct gs_point *originp,
			      pdf_bool_t *has_originp)
{
	struct gs_point origin;
	pdf_bool_t has_origin;
	struct heur_bl_chooser hbl;
	struct a_area *a;

	if (area -> child == NULL) {
		*has_originp = 0;
		return;
	}
	
	heur_bl_init (&hbl, heur, wdir);

	a = area -> child;
	do {
		a_get_origin (a, wdir, heur, &origin, &has_origin);
		if (has_origin)
			heur_bl_add (&hbl, &origin);
		a = a -> next;
	} while (a != area -> child);

	heur_bl_choose (&hbl, &area -> g.rect, originp, has_originp);
}

static void group_query_info (struct a_area *area, enum gs_direction wdir,
 			      struct a_info *info)
{
	struct a_area *a, *utmost;
	struct a_info li;
	pdf_real_t edge, r;
	gs_edge hif;
	gs_cmp cmf;

	hif = gs_get_hi_edge (wdir);
	cmf = gs_get_cmp (wdir);

	if (area -> child == NULL) {
		info -> fill = 1;
		info -> hyphenized = 0;
		return;
	}
	
	info -> fill = 1;
	utmost = a = area -> child;
	edge = hif (&utmost -> g.rect);
	do {
		a = a -> next;
		r = hif (&a -> g.rect);
		a_query_info (a, wdir, &li);

		if (cmf (r, edge) >= 0) {
			info -> hyphenized = li.hyphenized;
			info -> hyphen = li.hyphen;
		}

		info -> fill = info -> fill && li.fill;
	} while (a != area -> child);
}

static void group_get_output_bounds (struct a_area *area, 
				     enum gs_direction wdir,
				     pdf_real_t pspace,
				     struct args_output *aout,
				     struct cobj_rect *bounds)
{
	struct cobj_rect cbounds;
	struct a_area *a;

	if (area -> child == NULL) {
		*bounds = area -> g.bounds;
		return;
	}

	*bounds = area -> g.rect;
	a = area -> child;
	do {
		/* Putting pspace on any inner areas should  be an 
		 * acceptable simplification... I hope
		 * -- Note that this mirrors squeeze.c: does not 
		 * make much sense changing this and not that --
		*/
		a_get_output_bounds (a, wdir, pspace, aout, &cbounds);
		cobj_rect_union (bounds, &cbounds);
		a = a -> next;
	} while (a != area -> child);
}

enum s_status a__init (struct a_area **areap, struct ref_owner *p,
		       const char *fname, int line)
{
	struct a_area *area;

	area = malloc (sizeof (struct a_area));
	if (area == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (area, 0, sizeof (struct a_area));

	ref__init (&area -> refs, "area", p, fname, line);

	cobj_rect_set_null (&area -> g.bounds);
	cobj_rect_set_null (&area -> g.rect);

	*areap = area;

	return S_CSC_OK;
}

static enum s_status area__dup (struct a_area **areap, struct a_area *src,
				struct ref_owner *p,
				const char *fname, int line)
{
	struct a_area *area;
	enum s_status res;

	res = a__init (&area, p, fname, line);
	if (res != S_CSC_OK)
		return res;

	area -> ifc = src -> ifc;
	area -> g = src -> g;

	if (area -> ifc -> copy != NULL) {
		res = area -> ifc -> copy (area, src);
		if (res != S_CSC_OK) {
			a_unref (area, p);
			return res;
		}
	}

	*areap = area;

	return S_CSC_OK;
}

#define area_dup(dest, src, p) area__dup (dest, src, p, __FILE__, __LINE__)

static void grow_area (struct a_area *parent, struct a_area *area)
{
	cobj_rect_union (&parent -> g.bounds, &area -> g.bounds);
	cobj_rect_union (&parent -> g.rect, &area -> g.rect);
	if (area -> g.svec.x > parent -> g.svec.x)
		parent -> g.svec.x = area -> g.svec.x;
	if (area -> g.svec.y > parent -> g.svec.y)
		parent -> g.svec.y = area -> g.svec.y;
}

static void append (struct a_area *parent, struct a_area *area)
{
	if (parent -> child != NULL) {
		area -> prev = parent -> child;
		area -> next = parent -> child -> next;
		area -> prev -> next = area;
		area -> next -> prev = area;
	} else
		area -> next = area -> prev = area;
		
	parent -> child = area;	     
}

static struct a_area *get_first (struct a_area *parent)
{
	struct a_area *area;
	
	if (parent -> child == NULL)
		return NULL;

	area = parent -> child -> next;
	if (area -> next != area) {
		area -> prev -> next = area -> next;
		area -> next -> prev = area -> prev;
	} else
		parent -> child = NULL;

	area -> next = area -> prev = NULL;

	return area;
}

static void detach (struct a_area *parent, struct a_area *area)
{
	if (area -> next != area) {
		area -> next -> prev = area -> prev;
		area -> prev -> next = area -> next;
		if (parent -> child == area)
			parent -> child = area -> prev;
	} else
		parent -> child = NULL;

	area -> next = area -> prev = NULL;
}

static void replace (struct a_area *parent, 
		     struct a_area *olda, struct a_area *newa)
{
	if (parent -> child == olda)
		parent -> child = newa;

	if (olda -> next != olda) {
		newa -> next = olda -> next;
		newa -> prev = olda -> prev;
		newa -> next -> prev = newa;
		newa -> prev -> next = newa;
	} else
		newa -> next = newa -> prev = newa;

	olda -> prev = olda -> next = NULL;
}

static void add_to_area (struct a_area *parent, struct a_area *area)
{
	struct a_area *p;

	if (area -> ifc == &group_ifc) {
		for (p = get_first (area); p != NULL; p = get_first (area)) {
			add_to_area (parent, p);
			a_unref (p, ref_rown (&area -> refs));
		}
	} else {
		append (parent, area);
		a_ref (area, ref_rown (&parent -> refs));
	}
} 

static void reparent (struct a_area *parent, struct a_area *area)
{
	struct a_area *ap;

	for (ap = get_first (area); ap != NULL; ap = get_first (area)) {
		append (parent, ap);
		a_ref (ap, ref_rown (&parent -> refs));
		a_unref (ap, ref_rown (&area -> refs));
	}
}

static enum s_status insert_area_outside (struct a_area *parent,
					  struct a_area *area,
					  struct a_area *inside)
{
	replace (parent, inside, area);
	if (inside -> ifc -> flags.logic)
		reparent (area, inside);
	else
		add_to_area (area, inside);

	grow_area (area, inside);
	grow_area (parent, area);

	a_ref (area, ref_rown (&parent -> refs));
	a_unref (inside, ref_rown (&parent -> refs));

	return S_CSC_OK;
}

static enum s_status insert_area_none (struct a_area *parent, 
				       struct a_area *area)
{
	append (parent, area);
	a_ref (area, ref_rown (&parent -> refs));

	grow_area (parent, area);

	return S_CSC_OK;
}

static enum s_status insert_area_inside (struct a_area *parent,
					 struct a_area *area,
					 struct a_area *outside)
{
	enum s_status res;

	res = insert_area_none (outside, area);
	if (res != S_CSC_OK)
		return res;

	grow_area (parent, outside);

	return S_CSC_OK;
}

static enum s_status insert_area_overlap (struct a_area *parent,
					  struct a_area *area)
{
	struct a_area *outside, *p, *q;
	enum cobj_rect_overlap ov;
	enum s_status res;
	int one;

	res = a_init (&outside, ref_rown (&parent -> refs));
	if (res != S_CSC_OK)
		return res;

	outside -> ifc = &group_ifc;
	outside -> g = area -> g;
	do {
		one = 0;
		p = parent -> child;
		if (p != NULL) {
			do {
				p = p -> next;
				ov = cobj_rect_overlaps (&outside -> g.bounds,
							 &p -> g.bounds);
				if (ov == COBJ_RO_OVERLAP) {
					grow_area (outside, p);
					one++;
				}
			} while (p != parent -> child);
		}
	} while (one);

	p = parent -> child;
	if (p != NULL) {
		while (1) {
			p = p -> next;
			ov = cobj_rect_overlaps (&outside -> g.bounds, 
						 &p -> g.bounds);
			if (ov != COBJ_RO_NONE)  {
				q = p -> prev != p ? p -> prev : NULL;
				detach (parent, p);
				add_to_area (outside, p);
				a_unref (p, ref_rown (&parent -> refs));
				if (q == NULL)
					break;
				p = q;
			} else if (p == parent -> child)
				break;
		}
	}

	append (outside, area);
	a_ref (area, ref_rown (&outside -> refs));

	append (parent, outside);

	return S_CSC_OK;
}

static enum s_status insert_area (struct a_area *parent, struct a_area *area)
{
	enum cobj_rect_overlap ov;
	enum s_status res;
	struct a_area *p, *outside;

	ov = COBJ_RO_NONE;
	outside = NULL;
	if (parent -> child != NULL) {
		p = parent -> child;
		do {
			p = p -> next;
			ov = cobj_rect_overlaps (&area -> g.bounds, 
						 &p -> g.bounds);
			if (ov == COBJ_RO_INSIDE ||
			    ov == COBJ_RO_SAME ||
			    ov == COBJ_RO_OVERLAP)
				break;
			if (ov == COBJ_RO_OUTSIDE) {
				if (outside != NULL) {
					ov = COBJ_RO_OVERLAP;
					break;
				} else
					outside = p;
			}
		} while (p != parent -> child);
	}

	switch (ov) {
	case COBJ_RO_OUTSIDE:
	case COBJ_RO_NONE:
		if (outside != NULL)
			res = insert_area_outside (parent, area, outside);
		else
			res = insert_area_none (parent, area);
		break;

	case COBJ_RO_SAME:	/* So we can replace p if logic */
		res = insert_area_outside (parent, area, p);
		break;

	case COBJ_RO_INSIDE:
		res = insert_area_inside (parent, area, p);
		break;

	case COBJ_RO_OVERLAP:
		res = insert_area_overlap (parent, area);
	}
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status add_area (struct a_area *parent, struct cobj_rect *rect,
			       struct cobj_rect *bounds, struct a_ifc *ifc,
			       struct a_area *cparea, struct gs_point *svec,
			       struct a_area **areap)
{
	struct ref_owner own;
	struct a_area *area;
	enum s_status res;

	ref_init_owner (&own, "add_area");

	if (!cobj_rect_width (rect) ||
	    !cobj_rect_height (rect)) {
		if (areap != NULL)
			*areap = NULL;
		return S_CSC_OK;
	}

	res = a_init (&area, &own);
	if (res != S_CSC_OK)
		return res;

	area -> ifc = ifc;
	area -> g.rect = *rect;

	if (bounds != NULL)
		area -> g.bounds = *bounds;
	else
		area -> g.bounds = *rect;

	if (cparea != NULL) {
		cobj_rect_intersect (&area -> g.bounds,
				     &cparea -> g.bounds);
		cobj_rect_intersect (&area -> g.rect,
				     &cparea -> g.rect);
		if (cobj_rect_is_null (&area -> g.rect) ||
		    cobj_rect_is_null (&area -> g.bounds)) {
			a_unref (area, &own);
			if (areap != NULL)
				*areap = NULL;
			return S_CSC_OK;
		}
	}

	if (svec != NULL)
		area -> g.svec = *svec;
	else if (cparea != NULL)
		area -> g.svec = cparea -> g.svec;
		
	res = insert_area_none (parent, area);
	a_unref (area, &own);
	if (res != S_CSC_OK)
		return res;

	if (areap != NULL)
		*areap = area;

	return S_CSC_OK;
}

static void get_line_rect (struct gs_line *line, 
			   struct gs_graph_state *gs,
			   struct cobj_rect *rect)
{
	if (line -> a.x < line -> b.x) {
		rect -> left = line -> a.x;
		rect -> right = line -> b.x;
	} else {
		rect -> left = line -> b.x;
		rect -> right = line -> a.x;
	}

	if (line -> a.y < line -> b.y) {
		rect -> bottom = line -> a.y;
		rect -> top = line -> b.y;
	} else {
		rect -> bottom = line -> b.y;
		rect -> top = line -> a.y;
	}
}

static void get_bezier_rect (struct gs_bezier *bezier, 
 			     struct gs_graph_state *gs,
			     struct cobj_rect *rect)
{
	gs_get_bezier_rect (rect, bezier);
}

static enum s_status add_graph_area (struct a_area *parent,
				     struct a_area *cparea,
				     struct gs_paint *paint,
				     struct gs_subpath *sp,	
				     struct gs_graph_state *gs)
{
	struct cobj_rect rect, crect;
	struct gs_subpath dsp;
	struct gs_element *e;
	struct a_area *na;
	enum s_status res;
	int i;

	cobj_rect_set_null (&rect);
 	for (i = 0; i < sp -> elements; i++) {
		e = &sp -> e [i];
		switch (e -> type) {
		case GS_ET_LINE:
			get_line_rect (&e -> u.line, gs, &crect);
			break;

		case GS_ET_BEZIER:
			get_bezier_rect (&e -> u.bezier, gs, &crect);
			break;

		default:
			assert (0);
		}

		/* This avoids null rects when rendering e.g. an
		 * horizontal line. In addition it makes sense
		 * when the whidth is really thick */
		crect.left -= gs -> line_width / 2;
		crect.right += gs -> line_width / 2;
		crect.bottom -= gs -> line_width / 2;
		crect.top += gs -> line_width / 2;

		cobj_rect_union (&rect, &crect);
	}

	gs_translate_rect (&rect, &gs -> ctm);

	res = add_area (parent, &rect, NULL, 
			&graph_ifc, cparea, NULL, &na);
	if (res != S_CSC_OK)
		return res;

	if (na == NULL)
		return S_CSC_OK;

	res = gs_subpath_copy (&dsp, sp, ref_rown (&na -> refs));
	if (res != S_CSC_OK)
		return res;

	na -> u.graph.sp = dsp;
	na -> u.graph.gstate = gs;
	na -> u.graph.paint = *paint;
	
	T_CALL (gs_graph_ref, gs, ref_rown (&na -> refs));

	return S_CSC_OK;
}

static enum s_status add_text_area (struct a_area *parent,
				    struct a_area *cparea,
				    struct gs_paint *paint,
				    struct gs_subpath *sp,
				    struct args_heur *heur)
{
	struct gs_text_string string;
	struct cobj_rect bounds;
	struct gs_point vec;
	struct gf_font *font;
	struct a_area *na;
	enum s_status res;
	int i, j;

	if (!sp -> elements)
		return S_CSC_OK;

 	for (i = 0; i < sp -> elements; i++) {
		assert (sp -> e [i].type == GS_ET_TEXT);

		font = sp -> e [0].u.text.tstate -> font -> font;
		gs_get_vector (&sp -> e [i].u.text, 1, 1, &vec);
		gs_translate_vector (&vec, &sp -> e [i].u.text.gstate -> ctm);
		     
		res = gs_get_string (&sp -> e [i].u.text, &string, NULL);
		if (res != S_CSC_OK)
			return res;

		res = S_CSC_OK;
		for (j = 0; j < string.elements; j++) {
			heur_get_glyph_bounds (&string.glyphs [j].f, 
					       &bounds, &vec, heur);
			if (cobj_rect_is_null (&bounds))
				continue;

			res = add_area (parent, &string.glyphs [j].f.rect, 
					&bounds, &text_ifc, cparea, &vec,
					&na);
			if (res != S_CSC_OK)
				break;

			if (na == NULL)
				continue;

			heur_get_direction (&string, 
					    sp -> e [i].u.text.tstate,
					    &na -> u.text.primary,
					    &na -> u.text.secundary);
			na -> u.text.ginfo = string.glyphs [j];
			na -> u.text.tstate = sp -> e [i].u.text.tstate;
			na -> u.text.gstate = sp -> e [i].u.text.gstate;

			na -> u.text.paint = *paint;

			T_CALL (gs_text_ref, na -> u.text.tstate, 
				ref_rown (&na -> refs));
			T_CALL (gs_graph_ref, na -> u.text.gstate, 
				ref_rown (&na -> refs));
		}

		gs_string_destroy (&string);
		if (res != S_CSC_OK)
			return res;
	}

	return S_CSC_OK;
}

static enum s_status add_path_area (struct a_area *parent,
				    struct gs_path *path,
				    struct a_area *cparea,
				    struct gs_graph_state *gs,
				    struct gs_paint *paint,
				    struct args_heur *heur)
{
	struct gs_subpath *sp;
	enum s_status res;

	for (sp = path -> head; sp != NULL; sp = sp -> next) {
		if (sp -> elements) {
			if (sp -> e [0].type == GS_ET_TEXT)
				res = add_text_area (parent, cparea, 
						     paint, sp, heur);
			else
				res = add_graph_area (parent, cparea, 
						      paint, sp, gs);
			if (res != S_CSC_OK)
				return res;
		}
	}

	return S_CSC_OK;
}

static enum s_status intersect (struct a_area **areap, 
				struct a_area *a1, struct a_area *a2,
				struct ref_owner *p)
{
	struct a_area *parent, *ap, *aq, *area;
	struct cobj_rect bounds, rect;
	enum s_status res;

	res = a_init (&parent, p);
	if (res != S_CSC_OK)
		return res;

	if (a1 -> child == NULL || a2 -> child == NULL) {
		*areap = parent;
		return S_CSC_OK;
	}

	res = S_CSC_OK;
	ap = a1 -> child;
	do {
		ap = ap -> next;
		aq = a2 -> child;
		do {
			aq = aq -> next;
			bounds = ap -> g.bounds;
			cobj_rect_intersect (&bounds, &aq -> g.bounds);

			rect = ap -> g.rect;
			cobj_rect_intersect (&rect, &aq -> g.rect);

			if (cobj_rect_is_null (&bounds) ||
			    cobj_rect_is_null (&rect))
				continue;

			res = a_init (&area, ref_rown (&parent -> refs));
			if (res != S_CSC_OK)
				return res;

			area -> g.bounds = bounds;
			area -> g.rect = rect;

			append (parent, area);

			grow_area (parent, area);
			
		} while (aq != a2 -> child);
	} while (ap != a1 -> child);

	if (res != S_CSC_OK) {
		a_unref (parent, p);
		return res;
	}

	*areap = parent;

	return S_CSC_OK;
	
}
				
static enum s_status get_clipping_area (struct a_area **area,
					struct gs_clipping_path *cpath,
					struct cobj_rect *pgrect,
					struct ref_owner *p,
					struct args_heur *heur)
{
	struct a_area *parent, *grip, *cparea, *ap;
	enum s_status res;

	res = a_init (&parent, p);
	if (res != S_CSC_OK)
		return res;

	res = a_init (&ap, ref_rown (&parent -> refs));
	if (res != S_CSC_OK) {
		a_unref (parent, p);
		return res;
	}

        ap -> g.bounds = ap -> g.rect = 
		parent -> g.bounds = parent -> g.rect = *pgrect;
	append (parent, ap);

	res = S_CSC_OK;
	while (cpath != NULL) {
		res = a_init (&cparea, p);
		if (res != S_CSC_OK)
			break;

		res = add_path_area (cparea, cpath -> path, NULL,
				     cpath -> gstate, &gs_paint_fill, heur);
		if (res != S_CSC_OK) {
			a_unref (cparea, p);
			break;
		}

		res = intersect (&grip, parent, cparea, p);
		a_unref (cparea, p);
		if (res != S_CSC_OK)
			break;

		a_unref (parent, p);
		parent = grip;

		cpath = cpath -> next;
	}

	if (res != S_CSC_OK) {
		a_unref (parent, p);
		return res;
	}

	*area = parent;

	return S_CSC_OK;
}

static enum s_status add_shading_area (struct a_area *parent,
				       struct shd_shading *sh,
				       struct a_area *cparea,
				       struct gs_graph_state *gs)
{
	struct a_area *na;
	enum s_status res;

	res = add_area (parent, &cparea -> g.rect, &cparea -> g.bounds,
			&shading_ifc, NULL, NULL, &na);
	if (res != S_CSC_OK)
		return res;

	na -> u.sh.gstate = gs;
	na -> u.sh.sh = sh;

	T_CALL (gs_graph_ref, gs, ref_rown (&na -> refs));
	T_CALL (shd_ref, sh, ref_rown (&na -> refs));

	return S_CSC_OK;
}

static enum s_status add_xobject_area (struct a_area *parent,
				       struct xo_object *xo,
				       struct a_area *cparea,
				       struct gs_graph_state *gs)
{
	struct cobj_rect rect;
	struct a_area *na;
	enum s_status res;
	
	res = S_CSC_OK;
	switch (xo_get_subtype (xo)) {
	case XO_T_NONE:
		assert (0);
		break;

	case XO_T_PS:
		break;

	case XO_T_FORM:
		res = add_area (parent, &xo -> u.form.bbox, NULL,
				&xo_form_ifc, cparea, NULL, &na);
		break;

	case XO_T_IMAGE:
		rect.left = rect.bottom = 0;
		rect.right = rect.top = 1;
		gs_translate_rect (&rect, &gs -> ctm);
		res = add_area (parent, &rect, NULL,
				&xo_image_ifc, cparea, NULL, &na);
		break;

	default:
		res = S_CSC_NOT_SUPPORTED;
	}
	if (res != S_CSC_OK)
		return res;

	na -> u.xo.gstate = gs;
	na -> u.xo.xo = xo;

	T_CALL (gs_graph_ref, gs, ref_rown (&na -> refs));
	T_CALL (xo_ref, xo, ref_rown (&na -> refs));

	return S_CSC_OK;
}

static enum s_status add_stmt_area (struct a_area *parent,
				    struct gspg_stmt *stmt,
				    struct cobj_rect *pgrect,
				    struct args_heur *heur)
{
	struct a_area *cparea;
	enum s_status res;

	res = get_clipping_area (&cparea, stmt -> gstate -> clipping_path,
				 pgrect, ref_rown (&parent -> refs), heur);
	if (res != S_CSC_OK)
		return res;

	switch (stmt -> type) {
	case GSPG_ST_PATH:
		res = add_path_area (parent, stmt -> u.path, 
				     cparea, stmt -> gstate, &stmt -> paint, 
				     heur);
		break;

	case GSPG_ST_SHADING:
		res = add_shading_area (parent, stmt -> u.shading, 
					cparea, stmt -> gstate);
		break;

	case GSPG_ST_INLINE_IMAGE:
	case GSPG_ST_XOBJECT:
		res = add_xobject_area (parent, stmt -> u.xobject, 
					cparea, stmt -> gstate);
		break;

	default:
		assert (0);
	}

	a_unref (cparea, ref_rown (&parent -> refs));
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status get_areas (struct a_area *area, struct gspg_page *page,
				struct args_heur *heur)
{
	struct gspg_stmt *stmt;
	enum s_status res;

	for (stmt = page -> head; stmt != NULL; stmt = stmt -> next) {
		res = add_stmt_area (area, stmt, &page -> bounds, heur);
		if (res != S_CSC_OK)
			return res;
	}

	area -> g.bounds = page -> bounds;

	return S_CSC_OK;
}

enum s_status a_study_area (struct a_area **parentp, struct ref_owner *p)
{
	struct a_area *nparent, *area, *parent;
	enum s_status res;

	parent = *parentp;
	res = area_dup (&nparent, parent, p);
	if (res != S_CSC_OK)
		return res;

	for (area = get_first (parent); area != NULL; 
	     area = get_first (parent)) {
		res = insert_area (nparent, area);
		a_unref (area, ref_rown (&parent -> refs));
		if (res != S_CSC_OK)
			break;
	}

	/* Ugh, parent is not the same as before */
	if (res != S_CSC_OK) {
		a_unref (nparent, p);
		return res;
	}

	a_unref (parent, p);
	*parentp = nparent;

	return S_CSC_OK;
}

static enum s_status highlight (struct a_area *area, struct agc_context *agc,
				const char *spec)
{
	enum s_status res;
	struct a_area *a;

	if (area -> ifc != NULL && area -> child != NULL) {
		a = area -> child;
		do {
			a = a -> next;
			res = a_highlight (a, agc, spec);
			if (res != S_CSC_OK)
				return res;

		} while (a != area -> child);
	}

	if (a_in_spec (area -> ifc, spec, 'A')) {

		res = agc_rgb (agc, area -> ifc -> rgb.r, 
			       area -> ifc -> rgb.g, area -> ifc -> rgb.b);
		if (res != S_CSC_OK)
			return res;

		res = agc_ctm (agc, NULL);
		if (res != S_CSC_OK)
			return res;

		res = agc_rect (agc, a_srect (&area -> g, spec), 
				&gs_paint_stroke);
		if (res != S_CSC_OK)
			return res;
	}

	return S_CSC_OK;
}

static pdf_bool_t explicitly (char c, const char *spec, pdf_bool_t *ans)
{
	char *p;

	p = strchr (spec, c);
	if (p == NULL)
		return 0;

	*ans = p == spec || p [-1] != '!';

	return 1;
}

static pdf_bool_t is_negative_form (const char *spec)
{
	while (*spec) {
		if (*spec != '!')
			return 0;
		spec++;
		if (*spec)
			spec++;
	}

	return 1;
}

struct cobj_rect *a_srect (struct a_area_geometry *geom, const char *spec)
{
	return strchr (spec, '*') != NULL ? &geom -> bounds : &geom -> rect;
}

pdf_bool_t a_in_spec (struct a_ifc *ifc, const char *spec,
		      char group_spec)
{
	pdf_bool_t ans;

	if (spec == NULL || !*spec)
		return 1;

	if (explicitly (ifc -> spec, spec, &ans))
		return ans;

	if (explicitly (group_spec, spec, &ans))
		return ans;

	return is_negative_form (spec);
}

enum s_status a_highlight (struct a_area *area, struct agc_context *agc,
			   const char *spec)
{
	enum s_status res;

	if (area == NULL)
		return S_CSC_OK;

	res = highlight (area, agc, spec);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

enum s_status a__get_areas (struct a_area **areap, struct gspg_page *page,
			    struct args_heur *heur, struct ref_owner *p, 
			    const char *fname, int line)
{
	struct a_area *area;
	enum s_status res;

	res = a__init (&area, p, fname, line);
	if (res != S_CSC_OK)
		return res;
	
	res = get_areas (area, page, heur);
	if (res != S_CSC_OK) {
		a_unref (area, p);
		return res;
	}

	area -> ifc = &page_ifc;

	*areap = area;

	return S_CSC_OK;
}

void a_select (struct a_area *donor, struct a_area *acceptor,
	       a_selector sel, void *closure)
{
	struct a_area *area, *p;

	area = donor -> child;
	if (area == NULL)
		return;

	while (1) {
		area = area -> next;
		if (sel (area, closure)) {
			p = area -> prev != area ? 
				area -> prev : NULL;
			detach (donor, area);
			append (acceptor, area);
			grow_area (acceptor, area);
			a_ref (area, ref_rown (&acceptor -> refs));
			a_unref (area, ref_rown (&donor -> refs));
			if (p == NULL)
				break;
			area = p;
		} else if (area == donor -> child)
			break;
	}
}

enum s_status a_add_statement (struct a_area *area, struct agc_context *agc)
{
	struct a_area *a;
	enum s_status res;

	if (area -> ifc -> draw != NULL) {
		agc_set_code (agc, area -> code);
		res = area -> ifc -> draw (area, agc);
		if (res != S_CSC_OK)
			return res;
		agc_unset_code (agc);
	}

	if (area -> child != NULL) {
		
		a = area -> child;
		do {
			a = a -> next;
			res = a_add_statement (a, agc);
			if (res != S_CSC_OK)
				return res;

		} while (a != area -> child);
	}
	
	return S_CSC_OK;
}

static int ovl_selector (struct a_area *area, void *closure)
{
	enum cobj_rect_overlap ovl;

	ovl = cobj_rect_overlaps (&area -> g.rect, closure);

	return ovl == COBJ_RO_INSIDE || ovl == COBJ_RO_SAME;
}

enum s_status a__capture (struct a_area **areap, struct a_area *parent,
			  struct cobj_rect *rect, struct ref_owner *p, 
			  const char *fname, int line)
{
	struct a_area *area;	
	enum s_status res;

	res = a_init (&area, p);
	if (res != S_CSC_OK)
		return res;

	area -> ifc = &group_ifc;

	a_select (parent, area, ovl_selector, rect);
	if (area -> child == NULL) {
		a_unref (area, p);
		*areap = NULL;
		return S_CSC_OK;
	}

	if (area -> child -> next == NULL) {
		a_ref (area -> child, p);
		*areap = area -> child;
		a_unref (area, p);

		return S_CSC_OK;
	}

	*areap = area;

	return S_CSC_OK;
}

void a_get_direction (struct a_area *area,
		      struct args_heur *heur,
		      enum gs_direction *pdir,
		      enum gs_direction *sdir)
{
	*pdir = *sdir = GS_D_UNKNOWN;
	if (area -> ifc -> get_direction != NULL)
		area -> ifc -> get_direction (area, heur, pdir, sdir);
}

void a_get_origin (struct a_area *area, enum gs_direction wdir,
		   struct args_heur *heur, struct gs_point *origin, 
		   pdf_bool_t *has_origin)
{
	*has_origin = 0;
	if (area -> ifc -> get_origin != NULL)
		area -> ifc -> get_origin (area, wdir, heur, 
					   origin, has_origin);
}

void a_get_ratios (struct a_area *area, pdf_real_t *ratio,
		   pdf_real_t *badness, pdf_real_t *bad_slope,
		   struct args_output *aout)
{
	if (area -> ifc -> get_ratios == NULL) {
		*ratio = 1;
		*badness = HEUR_BADNESS_LAST_RESORT;
		*bad_slope = 0;
	} else
		area -> ifc -> get_ratios (area, ratio, badness, 
					   bad_slope, aout);
}

void a_get_output_bounds (struct a_area *area, enum gs_direction wdir,
			  pdf_real_t pspace, struct args_output *aout, 
			  struct cobj_rect *bounds)
{
	if (area -> ifc -> get_output_bounds != NULL)
		area -> ifc -> get_output_bounds (area, wdir, pspace,
						  aout, bounds);
	else
		*bounds = area -> g.bounds;
}

void a_query_info (struct a_area *area, enum gs_direction wdir,
		   struct a_info *info)
{
	if (area -> ifc -> query_info == NULL) {
		info -> hyphenized = 0;
		info -> fill = 0;
	} else
		area -> ifc -> query_info (area, wdir, info);
}

void a__ref (struct a_area *a, struct ref_owner *p, 
	     const char *fname, int line)
{
	if (a != NULL)
		ref__up (&a -> refs, p, fname, line);
}

void a_unref (struct a_area *a, struct ref_owner *p)
{
	struct a_area *ap;

	if (a == NULL || ref_down (&a -> refs, p))
		return;

	if (a -> ifc != NULL && a -> ifc -> destroy != NULL)
		a -> ifc -> destroy (a);

	for (ap = get_first (a); ap != NULL; ap = get_first (a))
		a_unref (ap, ref_rown (&a -> refs));

	ref_zero (&a -> refs);

	free (a);
}

