/*
 * PDFlow -- column.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <math.h>
#include <assert.h>
#include <string.h>

#include "grid/column.h"

struct c_closure {
	struct args_heur *heur;
};

struct gr_column_state {
	struct gr_region *reg;
	struct gr_line *first, *last;
};

static enum s_status merge_with_line (struct gr_region **regp,
				      struct gr_line *line)
{
	struct gr_region reg;
	struct gr_cell *cell;
	enum s_status res;
	
	for (cell = line -> cells; cell != NULL; cell = cell -> next) {
		reg = cell -> reg;
		gr_reg_sort (&reg);
		res = gr_reg_add (regp, &reg);
		if (res != S_CSC_OK)
			return res;
	}

	return S_CSC_OK;
}

static enum s_status column_init (struct gr_column_state *cst)
{
	cst -> reg = NULL;
	cst -> first = cst -> last = NULL;

	return S_CSC_OK;
}

static void column_destroy (struct gr_column_state *cst)
{
	gr_reg_destroy (cst -> reg);
	column_init (cst);
}

static enum s_status column_feed (struct gr_column_state *cst,
				  struct gr_line *line,
				  pdf_bool_t *found)
{
	struct gr_region *dreg;
	enum s_status res;

	*found = 0;
	if (cst -> first == NULL)
		cst -> first = line;
	cst -> last = line;

	res = gr_reg_dup (&dreg, cst -> reg);
	if (res != S_CSC_OK)
		return res;

	res = merge_with_line (&dreg, line);
	if (res != S_CSC_OK) {
		gr_reg_destroy (dreg);
		return res;
	}

	if (gr_reg_has_x_gap (dreg)) {
		gr_reg_destroy (cst -> reg);
		cst -> reg = dreg;
		cst -> last = line;
	} else if (cst -> first != cst -> last) {
		gr_reg_destroy (dreg);
		*found = 1;
	} else {
		gr_reg_destroy (dreg);
		column_destroy (cst);
	}

	return S_CSC_OK;
}

static void column_get_geom (struct a_area_geometry *g,
			     struct gr_grid *grid,
			     struct gr_column_state *cst)
{
	struct gr_line *sentinel;
	struct gr_region *reg;
	struct cobj_rect rect;
	pdf_real_t x;

	sentinel = cst -> last -> next;

	assert (cst -> reg != NULL);
	reg = cst -> reg;
	while (reg -> next != NULL)
		reg = reg -> next;

	if (grid -> wdir == GS_D_TOP_TO_BOTTOM ||
	    grid -> wdir == GS_D_BOTTOM_TO_TOP) {
		rect.bottom = cst -> first -> reg.lo;
		rect.top = cst -> last -> reg.hi;
		rect.left = cst -> reg -> lo;
		rect.right = reg -> hi;
	} else {
		rect.left = cst -> first -> reg.lo;
		rect.right = cst -> last -> reg.hi;
		rect.bottom = cst -> reg -> lo;
		rect.top = reg -> hi;
	}

	if (rect.left > rect.right) {
		x = rect.left;
		rect.left = rect.right;
		rect.right = x;
	}

	if (rect.bottom > rect.top) {
		x = rect.bottom;
		rect.bottom = rect.top;
		rect.top = x;
	}

	/* We could also derive the right bounds, but they are not
	 * used currently so why bother? */
	g -> bounds = g -> rect = rect;
	g -> svec.x = g -> svec.y = 0;
}

static enum s_status build_column (struct gr_grid *grid,
				   struct args_heur *heur,
				   struct gr_column_state *cst)
{
	struct gr_draft *draft;
	struct a_area *area;
	enum s_status res;

	res = a_init (&area, ref_rown (&grid -> refs));
	if (res != S_CSC_OK)
		return res;

	column_get_geom (&area -> g, grid, cst);

	res = gr_d_get_draft (&draft, area, grid -> wdir, grid -> ldir, heur, 
			      ref_rown (&grid -> refs));
	a_unref (area, ref_rown (&grid -> refs));
	if (res != S_CSC_OK)
		return res;

	res = gr_init (&grid, grid, draft, heur, ref_rown (&grid -> refs));
	gr_d_unref (draft, ref_rown (&grid -> refs));
	if (res != S_CSC_OK)
		return res;

	/* Otherwise this half-inited area (no ifc) may cause troubles */
	assert (grid -> ifc -> logical != GR_L_BOX);

	return S_CSC_OK;
}

static enum s_status do_build_columns (struct gr_grid *grid,
				       struct args_heur *heur)
{
	struct gr_column_state cst;
	struct gr_line *line;
	enum s_status res;
	pdf_bool_t again, found;

	if (grid -> ifc -> logical != GR_L_COLUMN)
		return S_CSC_OK;

	again = 1;
	while (again) {
		again = 0;
		res = column_init (&cst);
		if (res != S_CSC_OK)
			return res;

		for (line = grid -> d.lines; line != NULL; 
		     line = line -> next) {
			res = column_feed (&cst, line, &found);
			if (res != S_CSC_OK) {
				column_destroy (&cst);
				return res;
			}
			if (found) {
				res = build_column (grid, heur, &cst);
				if (res != S_CSC_OK) {
					column_destroy (&cst);
					return res;
				}
				again = 1;
				break;
			}
		}

		column_destroy (&cst);
	}

	return S_CSC_OK;
}

static enum s_status build_columns (struct gr_grid *grid, 
				    void *closure)
{
	struct c_closure *cc;
	enum s_status res;

	cc = closure;

	res = do_build_columns (grid, cc -> heur);
	if (res != S_CSC_OK)
		return res;

	res = gr_traverse (grid, build_columns, cc);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

enum s_status gr_c_detect_columns (struct gr_grid *grid, 
				   struct args_heur *heur)
{
	struct c_closure cc;
	enum s_status res;

	cc.heur = heur;
	res = build_columns (grid, &cc);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}
