/*
 * PDFlow -- draft.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>
#include <assert.h>

#include "grid/heur.h"
#include "grid/draft.h"

#ifndef min
#define min(a, b) ((a) < (b) ? (a) : (b))
#endif

#ifndef min
#define max(a, b) ((a) > (b) ? (a) : (b))
#endif

struct d_closure {
	struct heur_dir_chooser chooser;
	struct args_heur *heur;
};

struct f_closure {
	enum gs_direction wdir, ldir;
};

static enum s_status init_cell (struct gr_draft *draft,
				struct a_area *parent,
				struct a_area_geometry *g,
				struct args_heur *heur,
				struct gr_draft **cellp)
{
	struct gr_draft *cell;
	struct a_area *area;	
	enum s_status res;

	res = a_capture (&area, parent, &g -> rect, ref_rown (&draft -> refs));
	if (res != S_CSC_OK)
		return res;

	if (area == NULL) {
		*cellp = NULL;	
		return S_CSC_OK;
	}

	res = gr_d_init (&cell, area, heur, ref_rown (&draft -> refs));
	a_unref (area, ref_rown (&draft -> refs));
	if (res != S_CSC_OK)
		return res;

	*cellp = cell;

	return S_CSC_OK;
}

static enum s_status make_cells (struct gr_draft *draft, 
				 struct a_area *area,
				 struct args_heur *heur)
{
	struct gr_region *hreg, *vreg;
	struct gr_draft *cell;
	struct a_area_geometry g;
	int idx;
	enum s_status res;

	draft -> rows = gr_reg_size (draft -> vreg);
	draft -> cols = gr_reg_size (draft -> hreg);

	if (!draft -> rows && !draft -> cols)
		return S_CSC_OK;

	assert (draft -> rows && draft -> cols);

	if (draft -> rows == 1 && draft -> cols == 1) {
		draft -> type = GR_D_AREA;
		draft -> g = area -> g;
		draft -> area = area;
		a_ref (area, ref_rown (&draft -> refs));
		return S_CSC_OK;
	}

	draft -> cell = calloc (draft -> rows * draft -> cols,
				sizeof (struct gr_draft *));
	if (draft -> cell == NULL)
		return S_CSC_OUT_OF_MEMORY;

	idx = 0;
	g.svec.x = g.svec.y = 0;
	for (vreg = draft -> vreg; vreg != NULL; vreg = vreg -> next) {
		g.rect.bottom = vreg -> lo;
		g.rect.top = vreg -> hi;
		g.bounds.bottom = vreg -> xlo;
		g.bounds.top = vreg -> xhi;
		for (hreg = draft -> hreg; hreg != NULL; hreg = hreg -> next) {
			g.rect.left = hreg -> lo;
			g.rect.right = hreg -> hi;
			g.bounds.left = hreg -> xlo;
			g.bounds.right = hreg -> xhi;
			res = init_cell (draft, area, &g, heur, &cell);
			if (res != S_CSC_OK)
				return res;
			draft -> cell [idx] = cell;
			if (cell != NULL) {
				cobj_rect_union (&draft -> g.rect, 
						 &cell -> g.rect);
				cobj_rect_union (&draft -> g.bounds, 
						 &cell -> g.bounds);
			}
			idx++;
		}
	}

	return S_CSC_OK;
}

static void destroy_cells (struct gr_draft *draft)
{
	int i, cells;

	cells = draft -> rows * draft -> cols;
	for (i = 0; i < cells; i++) 
		gr_d_unref (draft -> cell [i], ref_rown (&draft -> refs));

	free (draft -> cell);
}

enum s_status gr_d_traverse (struct gr_draft *draft, 
			     gr_s_draft_f draftf,
			     void *closure)
{
	enum s_status res;
	int i, cells;

	if (draft -> type == GR_D_AREA)
		return S_CSC_OK;

	cells = draft -> rows * draft -> cols;
	for (i = 0; i < cells; i++) {
		if (draft -> cell [i] != NULL) {
			res = draftf (draft -> cell [i], closure);
			if (res != S_CSC_OK)
				return res;
		}
	}

	return S_CSC_OK;
}

static void apply_order (struct gr_draft *draft, 
			 struct gs_point *orgp, 
			 struct f_closure *fc)
{
	pdf_bool_t inner;

	if (fc -> wdir == GS_D_LEFT_TO_RIGHT ||
	    fc -> wdir == GS_D_RIGHT_TO_LEFT)
		inner = draft -> hin;
	else
		inner = draft -> vin;

	if (draft -> type != GR_D_AREA)
		draft -> type = inner ? GR_D_INNER : GR_D_OUTER;

	draft -> wdir = fc -> wdir;
	draft -> ldir = fc -> ldir;
}

static enum s_status fixup_dir (struct gr_draft *draft, void *closure)
{
	struct f_closure *fc;
	enum s_status res;

	fc = closure;

	if (draft -> wdir == GS_D_UNKNOWN || draft -> ldir == GS_D_UNKNOWN) {
		res = gr_d_traverse (draft, fixup_dir, fc);
		if (res != S_CSC_OK)
			return res;

		apply_order (draft, NULL, fc);
	}

	return S_CSC_OK;
}

static enum s_status set_direction (struct gr_draft *draft, void *closure)
{
	struct d_closure *dcp, dc;
	struct f_closure fc;
	struct gs_point origin;
	pdf_bool_t certain;
	enum s_status res;

	dcp = closure;

	if (draft -> type != GR_D_AREA) {

		dc = *dcp;
		heur_dc_init (&dc.chooser, dcp -> heur);
		res = gr_d_traverse (draft, set_direction, &dc);
		if (res != S_CSC_OK)
			return res;

		heur_dc_choose (&dc.chooser, &fc.wdir, &fc.ldir, &certain);

	} else {
		a_get_direction (draft -> area, dcp -> heur, 
				 &fc.wdir, &fc.ldir);
		certain = 1;
	}

	origin.x = draft -> g.rect.left;
	origin.y = draft -> g.rect.bottom;
	
	if (fc.wdir != GS_D_UNKNOWN && fc.ldir != GS_D_UNKNOWN) {
		apply_order (draft, &origin, &fc);
		res = gr_d_traverse (draft, fixup_dir, &fc);
		if (res != S_CSC_OK)
			return res;
	}

	heur_dc_add (&dcp -> chooser, fc.wdir, fc.ldir);

	return S_CSC_OK;
}

static enum s_status init (struct gr_draft *draft, struct a_area *area,
			   struct args_heur *heur)
{
	enum s_status res;

	cobj_rect_set_null (&draft -> g.rect);
	cobj_rect_set_null (&draft -> g.bounds);
	res = gr_reg_get (area, heur, &draft -> hreg, 
			  &draft -> vreg, &draft -> hin, 
			  &draft -> vin);
	if (res != S_CSC_OK)
		return res;

	/* Will be overridden when we know which direction to follow */
	draft -> type = GR_D_OUTER;

	res = make_cells (draft, area, heur);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

enum s_status gr_d__init (struct gr_draft **draftp, struct a_area *area,
			  struct args_heur *heur, struct ref_owner *p,
			  const char *fname, int line)
{
	struct gr_draft *draft;
	enum s_status res;

	draft = malloc (sizeof (struct gr_draft));
	if (draft == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (draft, 0, sizeof (struct gr_draft));
	
	ref__init (&draft -> refs, "draft", p, fname, line);
	res = init (draft, area, heur);
	if (res != S_CSC_OK) {
		gr_d_unref (draft, p);
		return res;
	}

	*draftp = draft;

	return S_CSC_OK;
}

static void set_weight (struct gr_draft *draft, struct a_area *area)
{
	pdf_real_t ave, weight, overall;
	struct gr_region *reg;

	if (draft -> ldir == GS_D_BOTTOM_TO_TOP ||
	    draft -> ldir == GS_D_TOP_TO_BOTTOM) {
		reg = draft -> vreg;
		overall = cobj_rect_height (&area -> g.bounds);
	} else {
		reg = draft -> hreg;
		overall = cobj_rect_width (&area -> g.bounds);
	}
		
	weight = ave = 0;
	while (reg != NULL) {
		ave += gr_reg_width (reg) * 
			(reg -> lo + reg -> hi) / 2;
		weight += gr_reg_width (reg);
		reg = reg -> next;
	}

	draft -> weight = (ave / weight) / overall;
}

enum s_status gr__d_get_draft (struct gr_draft **draftp, struct a_area *area, 
			       enum gs_direction wdir, enum gs_direction ldir,
			       struct args_heur *heur,
			       struct ref_owner *p, 
			       const char *fname, int line)
{
	struct f_closure fc;
	struct gr_draft *draft;
	struct d_closure dc;
	enum s_status res;

	res = gr_d__init (&draft, area, heur, p, fname, line);
	if (res != S_CSC_OK)
		return res;

	heur_dc_init (&dc.chooser, heur);
	dc.heur = heur;

	res = set_direction (draft, &dc);
	if (res != S_CSC_OK) {
		gr_d_unref (draft, p);
		return res;
	}

	fc.wdir = wdir;
	fc.ldir = ldir;
	fixup_dir (draft, &fc);

	set_weight (draft, area);

	*draftp = draft;

	return S_CSC_OK;
}

void gr_d__ref (struct gr_draft *draft, struct ref_owner *p,
		const char *fname, int line)
{
	if (draft != NULL)
		ref__up (&draft -> refs, p, fname, line);
}

void gr_d_unref (struct gr_draft *draft, struct ref_owner *p)
{
	if (draft == NULL || ref_down (&draft -> refs, p))
		return;

	gr_reg_destroy (draft -> hreg);
	gr_reg_destroy (draft -> vreg);

	if (draft -> cell != NULL)
		destroy_cells (draft);
	a_unref (draft -> area, ref_rown (&draft -> refs));

	ref_zero (&draft -> refs);

	free (draft);
}
