/*
 * PDFlow -- grid.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */
/*
 * Basic grid manipulation routines.
 * Grids are the tool we use to detect all the patterns that "meet the eye"
 * when looking at a page. The source files in this directory, therefore
 * contain all the logic to detect columns, tell how to follow the
 * text flow and so on. 
 *
 * The goal is to obtain a *logical* view of a page, starting from a
 * *physical* decomposition of its contents. If everything goes well,
 * the outcome of this process is an instance of `struct gr_grid',
 * which contains a list of lines, further decomposed into cells.
 * Lines (and cells) loosely follow the same order of the text flow.
 * In turn, each cell is a grid, and so on till we hit the bottom of this
 * tree when we encounter a special kind of 1x1 grid wrapping an area.
 *
 * I've used the term "loosely" because, in some cases, blindly sticking
 * to cell order may yield inappropriate results. An obvious example is 
 * when a floating image is encountered:
 * 
 *   <xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx>
 *   <yyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyy>
 *              <zzzzzzzzzzzzzzzzzzzzzzzz>
 *     Picture  <wwwwwwwwwwwwwwwwwwwwwwww>
 *              <tttttttttttttttttttttttt>
 *   <uuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuu>
 *
 * A partial decomposition follows:
 * +---------------------------------------+
 * | <xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx>(1) | 
 * +---------------------------------------+
 * | <yyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyy>(2) |
 * +---------------------------------------+
 * |+---------+---------------------------+|
 * || (3)     | <zzzzzzzzzzzzzzzzzzzzz>(4)|| 
 * || Picture | <wwwwwwwwwwwwwwwwwwwwwwww>||
 * ||         | <tttttttttttttttttttttttt>||
 * |+---------+---------------------------+|
 * +---------------------------------------+
 * | <uuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuu>(5) |
 * +---------------------------------------+
 *
 * Note that it is partial because cells #1, 2, 4 and 5 should contain
 * futher grids.
 * Here, the proper order is: 1, 2, 4, 5. Cell 3 is "floating" and should
 * appear nearby, without breaking text flow.
 *
 * To deal with all these complexities, we assign a "logical" meaning
 * to each grid (see gr_ifc instances below for a list).
 *
 * The algorithm (gr_get_grid) proceeds by step:
 * 1. Detect areas
 * 2. Extract physical layout (i.e. drafts): physical layouts
 *    are similar to grids, yet they do not follow the text
 *    flow (they are always bottom->top, left->rigt) and
 *    do not contain any logical meaning;
 * 3. Conversion of drafts into grids: we apply some heuristics
 *    to tell which order to follow and identify which is the
 *    logical meaning of the grid. At this stage
 *    physical cells are mapped 1:1 to logical cells;
 * 3. Column detection: this is a process meant to merge and
 *    split physical grids to detect "hidden" grids 
 *    (see column.c).
 *
 * After all this is done, thread.c takes care to extract the
 * relevant (low-level) grids in the correct order.
 *
 * This file contains only front-end functions (i.e. those
 * exported through grid.h) and their basic grid manipulation
 * routines. The "true" logic is in the other grid/ source files
 * (apart the draft->grid translation bit, which is conceptually
 * straightforward).
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <math.h>
#include <assert.h>
#include <string.h>

#include "grid/grid.h"
#include "grid/squeeze.h"
#include "grid/layout.h"
#include "grid/baseline.h"
#include "grid/column.h"
#include "grid/thread.h"
#include "grid/qi.h"
#include "grid/wspace.h"

#define ARROW_LEN 5
#define ARROW_SIZE 2.5
#define ALIGN_FLEX 2.5
#define ALIGN_FLUSH 2.5

#ifndef min
#define min(a, b) ((a) < (b) ? (a) : (b))
#endif

#ifndef max
#define max(a, b) ((a) > (b) ? (a) : (b))
#endif

/**
 * ===============
 * Grid interfaces
 * ===============
 * Once the logical meaning of a grid is detected, it assigned
 * one of these interfaces. Besides containing the logical
 * constant (GR_L_XXX), these interfaces also declare the function
 * that shall be called to squeeze (i.e. sensibly resize) that
 * grid. 
 * Some "inclusion" rules apply: a parent grid's cell may
 * only contain some kind of grids. This table lists
 * all the legal inclusions
 *				
 *    Parents
 *     VVV
 *      	Page     Box    Column  Line	Word    <<< Children
 *   Page                         X       X
 *   Box                          X
 *   Column                               X
 *   Line                 X       X              X
 *   Word
 *
 * There is some kind of trade-off between looseining rules
 * (therefore raising the number of `X'es on the diagram above)
 * and keeping them as strict as possible.
 * Keeping them strict produces a more regular output, so
 * squeezing routines are more straightfoward because they
 * will handle less cases.
 * Loosening is needed only to handle different cases in
 * a different way, i.e. when two inclusions bear different
 * meanings. It is often related to definition of new kinds 
 * of grids.
 */ 

/*
 * Word: this 1x1 grid just contains an area, has no line/cells
 * internal structure and its only way to squeeze is to resize
 * (not to reflow). In most cases they are actually what
 * their name promises, i.e. word (the letters being represented
 * by areas). Picutures, elaborate formulas and artwork, however
 * are `words' too.
 */
static struct gr_ifc word_ifc = {
	GR_L_WORD, {
		"word", 'w', { 1, 0.7, 1 }, 
		{ atomic : 0, logic : 1 }, NULL 
	}, gr_sq_homog_probe, gr_sq_homog_squeeze, gr_th_word_thread
};

/*
 * Line: lines are 1xN grids that usually (but not necessarily)
 * contain words. An example of line that does not
 * contain words is the example at the start of this file:
 *  +---------+---------------------------+
 *  | (3)     | <zzzzzzzzzzzzzzzzzzzzz>(4)| 
 *  | Picture | <wwwwwwwwwwwwwwwwwwwwwwww>|
 *  |         | <tttttttttttttttttttttttt>|
 *  +---------+---------------------------+
 * Lines may contain words, columns and even boxes
 * (i.e. they can contain virtually everything but lines).
 * Lines can only be contained in columns.
 *
 * When squeezeing, lines have different options: they
 * might resize if their contents allow to do so (think of
 * a line that only contains pictures). Typically, however
 * they achieve better results by reflowing, i.e. merging
 * their contents with other lines. In the example above,
 * it seems likely that line <yyy> will merge with <zzz>.
 * 
 * An important restriction of lines is that gaps between its
 * cells is limited. If one ore more gaps exceed a 
 * (content-sensitive) limit, the line grid is split into two
 * or more grids. The reason for that is to properly 
 * detect text columns:
 *
 *  xxxx xx xxx    aaaaaaa aaa
 *  yyy y yy yy    bb bbb bbbb
 *  zzz zzz zzz    ccccc ccccc
 * 
 * Thanks to this limit <xxx> won't be merged into a single
 * line with <aaa>.
 */
static struct gr_ifc line_ifc = {
	GR_L_LINE, {
		"line", 'l', { 0.4, 0.4, 0.4 }, 
		{ atomic : 0, logic : 1 }, NULL
	}, NULL, NULL, gr_th_line_thread
};

/*
 * Column: an Nx1 grid ALWAYS containing lines. Since
 * line gaps can't be too large, the following example:
 *
 *  xxxx xx xxx    aaaaaaa aaa
 *  yyy y yy yy    bb bbb bbbb
 *  zzz zzz zzz    ccccc ccccc
 * 
 * will be decomposed into TWO columns grouped 
 * be a box. Each column will contain three lines.
 * 
 * Columns may appear inside lines or boxes.
 */
static struct gr_ifc column_ifc = {
	GR_L_COLUMN, {
		"column", 'c', { 0.5, 1, 0.5 }, 
		{ atomic : 0, logic : 1 }, NULL
	}, gr_sq_homog_probe, gr_sq_homog_squeeze, gr_th_column_thread
};

/*
 * Box: a box is a "catch-all" NxM grid, i.e. any container
 * that does not respect any of the restrictions above.
 * Most important cases are:
 * 1. Column groups;
 * 2. Equations and their reference number
 * Though it may seem clumsy (and perhaps is!), it can only
 * contain columns, the reference number of an equation
 * will be a 1x1 column grid containing a 1x1 line containing
 * a 1x1 word grid enclosing the areas.
 */
static struct gr_ifc box_ifc = {
	GR_L_BOX, {
		"box", 'b', { 1, 0.7, 0.7 }, 
		{ atomic : 0, logic : 1 }, NULL 
	}, gr_sq_homog_probe, gr_sq_homog_squeeze, gr_th_box_thread
};

/*
 * Page: if present, it can only appear as the top level grid.
 * This grid features special code to handle the overall layout
 * of the page, like headers, footers and so on. 
 */
static struct gr_ifc page_ifc = {
	GR_L_PAGE, {
		"page", 'p', { 0, 0, 0 }, 
		{ atomic : 0, logic : 1 }, NULL 
	}, NULL, NULL, gr_th_page_thread
};

/*
 * ===================
 * Grid implementation
 * ===================
 */

/* highlight() closure arg */
struct h_closure {
	struct agc_context *agc;
	struct gspg_page *page;
	const char *spec;
};

/* 
 * Iterators are structures that contain all the info needed
 * to run through the cells of a grid in a predefined order
 * (left->right, right->left, top->bottom, bottom->top)
 */
struct iter_data {
	/* List of regions covered in current direction 
	 * Note that the order will always bottom->top 
	 * or left->right and won't necessary reflect 
	 * the order in which the cells will be traversed
	 * (e.g. it may be rigt->left) */
	struct gr_region *reg;

	/* The actual direction */
	enum gs_direction dir;

	/* The counterclockwise angle of the direction
	 * w.r.t the X-axis (left->right). */
	pdf_integer_t rotate;

	/* Iterator data */
	int start, stop, inc, size;

	/* Function telling the lower/upper edge of the cell's
	 * rect. E.g. in left->right mode, lo_edge will
	 * return the left edge. Note that if the direction
	 * is right->left or top->bottom, the lo_edge
	 * will be greater than the hi_edge */
	gs_edge lo_edge;
	gs_edge hi_edge;
};

static void destroy_line (struct gr_grid *grid, struct gr_line *line);

/**
 * =================
 * Utility functions
 * =================
 */
/*
 * A simple traverse function. The order will reflect the
 * loose text flow order.
 */
enum s_status gr_traverse (struct gr_grid *grid, gr_grid_f gridf,
			   void *closure)
{
	struct gr_line *line;
	struct gr_cell *cell;
	enum s_status res;

	if (grid -> ifc -> logical == GR_L_WORD)
		return S_CSC_OK;

	for (line = grid -> d.lines; line != NULL; line = line -> next)
		for (cell = line -> cells; cell != NULL; cell = cell -> next) {
			res = gridf (cell -> grid, closure);
			if (res != S_CSC_OK)
				return res;
		}

	return S_CSC_OK;
}

enum s_status gr_traverse_c (struct gr_grid *grid, gr_cell_f cellf,
			     void *closure)
{
	struct gr_line *line;
	struct gr_cell *cell;
	enum s_status res;

	if (grid -> ifc -> logical == GR_L_WORD)
		return S_CSC_OK;

	for (line = grid -> d.lines; line != NULL; line = line -> next)
		for (cell = line -> cells; cell != NULL; cell = cell -> next) {
			res = cellf (cell, closure);
			if (res != S_CSC_OK)
				return res;
		}

	return S_CSC_OK;
}

/*
 * Tell whether the grid is a word or not. Of course this
 * is needed to access the gr_grid.d union properly.
 */
pdf_bool_t gr_is_word (struct gr_grid *grid)
{
	return grid -> ifc == &word_ifc;
}

/**
 * ==================
 * Highlight function
 * ==================
 * A simple function that traverses the grid to highlight their border
 * to see how grid/ is decomposing the document.
 */
static void get_arrow (enum gs_direction dir, enum gs_direction oth_dir, 
		       struct cobj_rect *rect,
		       struct gs_point *a, struct gs_point *b, 
		       struct gs_point *c)
{
	pdf_real_t hsize, vsize;
	pdf_real_t hlen, vlen;
	pdf_real_t ox, oy;

	hlen = vlen = ARROW_LEN;
	hsize = vsize = ARROW_SIZE;


	/* Set an impossible default, because it must not be used */
	ox = oy = -1;

	switch (oth_dir) {
	case GS_D_LEFT_TO_RIGHT:
		ox = rect -> left;
		break;

	case GS_D_RIGHT_TO_LEFT:
		ox = rect -> right;
		break;

	case GS_D_BOTTOM_TO_TOP:
		oy = rect -> bottom;
		break;

	case GS_D_TOP_TO_BOTTOM:
		oy = rect -> top;
		break;

	case GS_D_UNKNOWN:
	case GS_D__LAST__:
		assert (0);
	}

	switch (dir) {
	case GS_D_LEFT_TO_RIGHT:
		a -> x = rect -> right;
		a -> y = oy;
		*b = *c = *a;
		b -> x -= hlen;
		c -> x -= hlen;
		b -> y += vsize;
		c -> y -= vsize;
		break;

	case GS_D_RIGHT_TO_LEFT:
		a -> x = rect -> right;
		a -> y = oy;
		*b = *c = *a;
		b -> x += hlen;
		c -> x += hlen;
		b -> y += vsize;
		c -> y -= vsize;
		break;

	case GS_D_BOTTOM_TO_TOP:
		a -> x = ox;
		a -> y = rect -> top;
		*b = *c = *a;
		b -> y -= vlen;
		c -> y -= vlen;
		b -> x += hsize;
		c -> x -= hsize;
		break;

	case GS_D_TOP_TO_BOTTOM:
		a -> x = ox;
		a -> y = rect -> bottom;
		*b = *c = *a;
		b -> y += vlen;
		c -> y += vlen;
		b -> x += hsize;
		c -> x -= hsize;
		break;

	case GS_D_UNKNOWN:
	case GS_D__LAST__:
		assert (0);
	}
}

static void get_al_arrow (enum gs_direction dir, struct cobj_rect *rect,
			  struct gs_point *a, struct gs_point *b, 
			  struct gs_point *c, pdf_bool_t flush,
			  pdf_bool_t internal)
{
	pdf_real_t flex, ofs;

	flex = internal ? ALIGN_FLEX : -ALIGN_FLEX;
	ofs = flush ? ALIGN_FLUSH : 0;

	switch (dir) {
	case GS_D_LEFT_TO_RIGHT:
		a -> x = b -> x = c -> x = rect -> left - ofs;
		c -> x += flex;
		
		a -> y = rect -> bottom;
		b -> y = rect -> top;
		c -> y = (rect -> top + rect -> bottom) / 2;
		break;

	case GS_D_RIGHT_TO_LEFT:
		a -> x = b -> x = c -> x = rect -> right + ofs;
		c -> x += flex;
		
		a -> y = rect -> bottom;
		b -> y = rect -> top;
		c -> y = (rect -> top + rect -> bottom) / 2;
		break;

	case GS_D_BOTTOM_TO_TOP:
		a -> y = b -> y = c -> y = rect -> bottom - ofs;
		c -> y += flex;
		
		a -> x = rect -> left;
		b -> x = rect -> right;
		c -> x = (rect -> left + rect -> right) / 2;
		break;

	case GS_D_TOP_TO_BOTTOM:
		a -> y = b -> y = c -> y = rect -> top + ofs;
		c -> y += flex;
		
		a -> x = rect -> left;
		b -> x = rect -> right;
		c -> x = (rect -> left + rect -> right) / 2;
		break;		

	case GS_D_UNKNOWN:
	case GS_D__LAST__:
		assert (0);
	}
}


static enum s_status draw_al_begin (struct gr_grid *grid,
				    struct agc_context *agc,
				    pdf_bool_t flush,
				    pdf_bool_t internal)
{
	struct gs_point a, b, c;
	enum s_status res;

	get_al_arrow (grid -> wdir, &grid -> g.rect, 
		      &a, &b, &c, flush, internal);

	if (flush) {
		res = agc_rgb (agc, 0, 0, 0);
		if (res != S_CSC_OK)
			return res;
	} else {
		res = agc_rgb (agc, 1, 0, 0);
		if (res != S_CSC_OK)
			return res;
	}

	res = agc_triangle (agc, &a, &b, &c, &gs_paint_stroke);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status draw_al_end (struct gr_grid *grid,
				  struct agc_context *agc,
				  pdf_bool_t flush,
				  pdf_bool_t internal)
{
	struct gs_point a, b, c;
	enum s_status res;

	get_al_arrow (gs_dir_neg (grid -> wdir), &grid -> g.rect, 
		      &a, &b, &c, flush, !internal);

	if (flush) {
		res = agc_rgb (agc, 0, 0, 0);
		if (res != S_CSC_OK)
			return res;
	} else {
		res = agc_rgb (agc, 1, 0, 0);
		if (res != S_CSC_OK)
			return res;
	}

	res = agc_triangle (agc, &a, &b, &c, &gs_paint_stroke);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status draw_align_fat (struct gr_grid *grid,
				     enum gr_al_align_type at,
				     pdf_bool_t flush,
				     struct agc_context *agc)
{
	enum s_status res;

	switch (at) {
	case GR_AT_BEGIN:
		res = draw_al_begin (grid, agc, flush, 1);
		if (res != S_CSC_OK)
			return res;
		break;

	case GR_AT_END:
		res = draw_al_end (grid, agc, flush, 1);
		if (res != S_CSC_OK)
			return res;
		break;

	case GR_AT_CENTER:
		res = draw_al_begin (grid, agc, flush, 0);
		if (res != S_CSC_OK)
			return res;

		res = draw_al_end (grid, agc, flush, 0);
		if (res != S_CSC_OK)
			return res;
		break;

	case GR_AT_JUSTIFIED:
		res = draw_al_begin (grid, agc, flush, 1);
		if (res != S_CSC_OK)
			return res;

		res = draw_al_end (grid, agc, flush, 1);
		if (res != S_CSC_OK)
			return res;
		break;

	case GR_AT_PROPORTIONAL:
		res = draw_al_begin (grid, agc, flush, 1);
		if (res != S_CSC_OK)
			return res;

		res = draw_al_end (grid, agc, flush, 0);
		if (res != S_CSC_OK)
			return res;
		break;

	case GR_AT_NONE:
		return S_CSC_OK;

	case GR_AT_UNKNOWN:
	case GR_AT__LAST__:
		assert (0);
	}

	return S_CSC_OK;
}

static enum s_status draw_align (struct gr_grid *grid,
				 struct agc_context *agc)
{
	enum s_status res;

	res = draw_align_fat (grid, grid -> al.at, 0, agc);
	if (res != S_CSC_OK)
		return res;

	res = draw_align_fat (grid, grid -> al.flush_at, 1, agc);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status draw_grid_specials (struct gr_grid *grid,
					 struct agc_context *agc,
					 const char *spec)
{
	struct gs_point a, b, c;
	enum s_status res;

	if (spec == NULL || strchr (spec, '^') != NULL) {
		get_arrow (grid -> wdir, grid -> ldir, &grid -> g.rect, 
			   &a, &b, &c);
		res = agc_triangle (agc, &a, &b, &c, &gs_paint_stroke_fill);
		if (res != S_CSC_OK)
			return res;

		get_arrow (grid -> ldir, grid -> wdir, &grid -> g.rect, 
			   &a, &b, &c);
		res = agc_triangle (agc, &a, &b, &c, &gs_paint_stroke);
		if (res != S_CSC_OK)
			return res;
	}

	if (spec == NULL || strchr (spec, '@') != NULL) {
		res = draw_align (grid, agc);
		if (res != S_CSC_OK)
			return res;
	}

	return S_CSC_OK;
}

static enum s_status highlight (struct gr_grid *grid, void *closure)
{
	struct h_closure *hc;
	struct a_ifc *ifc;
	enum s_status res;

	hc = closure;

	res = draw_grid_specials (grid, hc -> agc, hc -> spec);
	if (res != S_CSC_OK)
		return res;

	if (grid -> ifc -> logical == GR_L_WORD)
		return a_highlight (grid -> d.area, hc -> agc, hc -> spec);

	ifc = &grid -> ifc -> ifc;

	if (a_in_spec (ifc, hc -> spec, 'g')) {
		res = agc_rgb (hc -> agc, ifc -> rgb.r, 
			       ifc -> rgb.g, ifc -> rgb.b);
		if (res != S_CSC_OK)
			return res;

		res = agc_rgb_s (hc -> agc, ifc -> rgb.r, 
				 ifc -> rgb.g, ifc -> rgb.b, 0);
		if (res != S_CSC_OK)
			return res;

		res = agc_ctm (hc -> agc, NULL);
		if (res != S_CSC_OK)
			return res;

		res = agc_rect (hc -> agc, a_srect (&grid -> g, hc -> spec), 
				&gs_paint_stroke);
		if (res != S_CSC_OK)
			return res;
	}

	return gr_traverse (grid, highlight, closure);
} 

enum s_status gr_highlight (struct gr_grid *grid, struct agc_context *agc,
			    const char *spec, struct gspg_page *page)
{
	struct h_closure hc;

	hc.agc = agc;
	hc.page = page;
	hc.spec = spec;

	return highlight (grid, &hc);
}

/**
 * ========================================
 * Draft->Grid conversion (lines and cells)
 * ========================================
 */

/*
 * Init the iterator data according to a given direction.
 */
static void get_iter_data (struct gr_draft *draft, 
			   enum gs_direction dir, int rows, 
			   int cols, struct iter_data *id)
{
	id -> dir = dir;
	id -> reg = NULL;
	id -> lo_edge = gs_get_lo_edge (dir);
	id -> hi_edge = gs_get_hi_edge (dir);

	switch (dir) {
	case GS_D_LEFT_TO_RIGHT:
		if (draft != NULL)
			id -> reg = draft -> hreg;
		id -> rotate = 0;
		id -> start = 0;
		id -> stop = cols;
		id -> inc = 1;
		id -> size = cols;
		break;

	case GS_D_RIGHT_TO_LEFT:
		if (draft != NULL)
			id -> reg = draft -> hreg;
		id -> rotate = 180;
		id -> start = cols - 1;
		id -> stop = -1;
		id -> inc = -1;
		id -> size = cols;
		break;

	case GS_D_BOTTOM_TO_TOP:
		if (draft != NULL)
			id -> reg = draft -> vreg;
		id -> rotate = 90;
		id -> start = 0;
		id -> stop = rows * cols;
		id -> inc = cols;
		id -> size = rows;
		break;

	case GS_D_TOP_TO_BOTTOM:
		if (draft != NULL)
			id -> reg = draft -> vreg;
		id -> rotate = 270;
		id -> start = (rows - 1) * cols;
		id -> stop = -cols;
		id -> inc = -cols;
		id -> size = rows;
		break;

	case GS_D_UNKNOWN:
	case GS_D__LAST__:
		assert (0);
	}
}

/*
 * Cast a geometry object onto a region. Note that
 * the region's lo field is greater than the hi field
 * when traversing the grid leftwise or bottomwise.
 */ 
static void geom_to_reg (struct gr_region *reg,
			 struct a_area_geometry *g,
			 struct iter_data *id)
{
	reg -> lo = id -> lo_edge (&g -> rect);
	reg -> hi = id -> hi_edge (&g -> rect);
	reg -> xlo = id -> lo_edge (&g -> bounds);
	reg -> xhi = id -> hi_edge (&g -> bounds);
}

/*
 * Init a geometry object from the line and the cell 
 * regions.
 */
static void reg_to_geom (struct a_area_geometry *g,
			 struct gr_region *lreg,
			 struct gr_region *wreg,
			 pdf_integer_t rotate)
{
	struct gr_region *hreg, *vreg;

	if (rotate % 180) {
		hreg = lreg;
		vreg = wreg;
	} else {
		hreg = wreg;
		vreg = lreg;
	}

	g -> rect.left = min (hreg -> lo, hreg -> hi);
	g -> rect.right = max (hreg -> lo, hreg -> hi);
	g -> rect.bottom = min (vreg -> lo, vreg -> hi);
	g -> rect.top = max (vreg -> lo, vreg -> hi);

	g -> bounds.left = min (hreg -> xlo, hreg -> xhi);
	g -> bounds.right = max (hreg -> xlo, hreg -> xhi);
	g -> bounds.bottom = min (vreg -> xlo, vreg -> xhi);
	g -> bounds.top = max (vreg -> xlo, vreg -> xhi);
}

/* Init a cell from the draft's cell */
static enum s_status init_cell (struct gr_grid *grid,
				struct gr_draft *draft, 
				struct iter_data *wid,
				int index,
				struct args_heur *heur,
				struct a_area_geometry *ugeom,
				struct gr_cell **cellp)
{
	struct gr_cell *cell;
	struct gr_draft *scell;
	enum s_status res;

	scell = draft -> cell [index];
	if (scell == NULL) {
		*cellp = NULL;
		return S_CSC_OK;
	}

	cell = malloc (sizeof (struct gr_cell));
	if (cell == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (cell, 0, sizeof (struct gr_cell));

	geom_to_reg (&cell -> reg, &scell -> g, wid);
	if (ugeom != NULL) {
		cobj_rect_union (&ugeom -> rect, &scell -> g.rect);
		cobj_rect_union (&ugeom -> bounds, &scell -> g.bounds);
	}

	res = gr_init (&cell -> grid, grid, scell, heur, 
		       ref_rown (&grid -> refs));
	if (res != S_CSC_OK) {
		free (cell);
		return res;
	}

	*cellp = cell;

	return S_CSC_OK;
}

/* Init a cell given its region */
static enum s_status init_cell_reg (struct gr_grid *grid,
				    struct gr_region *wreg,
				    struct args_heur *heur,
				    struct gr_cell **cellp)
{
	struct gr_cell *cell;
	enum s_status res;

	cell = malloc (sizeof (struct gr_cell));
	if (cell == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (cell, 0, sizeof (struct gr_cell));

	cell -> reg = *wreg;

	res = gr_init (&cell -> grid, grid, NULL, heur, 
		       ref_rown (&grid -> refs));
	if (res != S_CSC_OK) {
		free (cell);
		return res;
	}

	*cellp = cell;

	return S_CSC_OK;
}

static void destroy_cell (struct gr_grid *grid, struct gr_cell *cell)
{
	gr_unref (cell -> grid, ref_rown (&grid -> refs));
	free (cell);
}

/* Init a line from the draft's line */
static enum s_status init_line (struct gr_grid *grid,
				struct gr_draft *draft, 
				struct iter_data *lid,
				struct iter_data *wid,
				int lidx,
				struct args_heur *heur,
				struct gr_line **linep)
{
	struct a_area_geometry ugeom;
	struct gr_cell **cellp;
	struct gr_line *line;
	int widx;
	enum s_status res;

	line = malloc (sizeof (struct gr_line));
	if (line == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (line, 0, sizeof (struct gr_line));

	cobj_rect_set_null (&ugeom.rect);
	cobj_rect_set_null (&ugeom.bounds);

	cellp = &line -> cells;
	for (widx = wid -> start; widx != wid -> stop; widx += wid -> inc) {
		res = init_cell (grid, draft, wid, lidx + widx, 
				 heur, &ugeom, cellp);
		if (res != S_CSC_OK) {
			destroy_line (grid, line);
			break;
		}
		if (*cellp != NULL)
			cellp = &(*cellp) -> next;
	}

	geom_to_reg (&line -> reg, &ugeom, lid);
	if (line -> cells == NULL) {
		free (line);
		line = NULL;
	}

	*linep = line;

	return S_CSC_OK;
}

/* Init a line given its region */
static enum s_status init_line_reg (struct gr_grid *grid,
				    struct gr_region *lreg,
				    struct gr_region *wreg,
				    struct args_heur *heur,
				    struct gr_line **linep)
{
	struct gr_cell **cellp;
	struct gr_line *line;
	enum s_status res;

	line = malloc (sizeof (struct gr_line));
	if (line == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (line, 0, sizeof (struct gr_line));

	line -> reg = *lreg;

	cellp = &line -> cells;
	while (wreg != NULL) {
		res = init_cell_reg (grid, wreg, heur, cellp);
		if (res != S_CSC_OK) {
			destroy_line (grid, line);
			break;
		}
		if (*cellp != NULL)
			cellp = &(*cellp) -> next;
		wreg = wreg -> next;
	}

	if (line -> cells == NULL) {
		free (line);
		line = NULL;
	}

	*linep = line;

	return S_CSC_OK;
}

static void destroy_line (struct gr_grid *grid, struct gr_line *line)
{
	struct gr_cell *cell;

	while (line -> cells != NULL) {
		cell = line -> cells;
		line -> cells = cell -> next;
		destroy_cell (grid, cell);
	}

	free (line);
}

/**
 * ==============================
 * Draft->Grid conversion (tools)
 * ==============================
 * Right after creating a new grid, apply_inclusions() is called to make
 * sure that it can appear as a cell of its parent grid. If it can't,
 * then it is wrapped into 1x1 grids making sure that the inclusion
 * rule table is respected.
 * The only exception is the line/line (forbidden) inclusion. This 
 * inconsistency is removed at parent level, since the most obviuos
 * way to solve it is to merge its children lines into itself.
 * Note however that if a parent line discovers that one of
 * its children lines's text flow directions do not match with its
 * own, that line is not merged. Rather the child line will be wrapped
 * into a column.
 */
static enum s_status init_wrap (struct gr_grid *grid,
				struct gr_grid *child,
				struct gr_ifc *ifc)
{
	struct iter_data wid, lid;

	grid -> ldir = child -> ldir;
	grid -> wdir = child -> wdir;
	grid -> lines = 1;
	grid -> words = 1;
	grid -> ifc = ifc;
	grid -> g = child -> g;

	get_iter_data (NULL, grid -> wdir, 1, 1, &wid);
	get_iter_data (NULL, grid -> ldir, 1, 1, &lid);

	grid -> d.lines = malloc (sizeof (struct gr_line));
	if (grid -> d.lines == NULL)
		return S_CSC_OUT_OF_MEMORY;

	grid -> d.lines -> next = NULL;
	geom_to_reg (&grid -> d.lines -> reg, &grid -> g, &lid);

	grid -> d.lines -> cells = malloc (sizeof (struct gr_cell));
	if (grid -> d.lines -> cells == NULL)
		return S_CSC_OUT_OF_MEMORY;

	grid -> d.lines -> cells -> next = NULL;
	grid -> d.lines -> cells -> grid = child;
	geom_to_reg (&grid -> d.lines -> cells -> reg, &grid -> g, &wid);

	gr_ref (child, ref_rown (&grid -> refs));

	return S_CSC_OK;
}

static enum s_status wrap (struct gr_grid *parent,
			   struct gr_grid **gridp,
			   struct gr_ifc *ifc,
			   struct ref_owner *p)
{
	struct gr_grid *grid;
	static enum s_status res;

	if (parent != NULL)
		assert (p == ref_rown (&parent -> refs));

	res = gr_init (&grid, parent, NULL, NULL, p);
	if (res != S_CSC_OK)
		return res;

	res = init_wrap (grid, *gridp, ifc);
	if (res != S_CSC_OK) {
		gr_unref (grid, p);
		return res;
	}

	gr_unref (*gridp, p);
	*gridp = grid;		  

	return S_CSC_OK;
}

static enum s_status wrap_into_line (struct gr_grid *parent,
				     struct gr_grid **gridp,
				     struct ref_owner *p)
{
	enum s_status res;

	assert ((*gridp) -> ifc -> logical != GR_L_LINE);

	res = wrap (parent, gridp, &line_ifc, p);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status wrap_into_column (struct gr_grid *parent,
				       struct gr_grid **gridp,
				       struct ref_owner *p)
{
	enum s_status res;

	if ((*gridp) -> ifc -> logical != GR_L_LINE) {
		res = wrap_into_line (parent, gridp, p);
		if (res != S_CSC_OK)
			return res;
	}
		
	res = wrap (parent, gridp, &column_ifc, p);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status apply_inclusions (struct gr_grid *parent,
				       struct gr_grid **gridp)
{
	enum s_status res;

	/* To make compiler happy */
	res = S_CSC_OK;

	/* At this stage, any grid may be at top-level */
	if (parent == NULL)
		return S_CSC_OK;

	switch (parent -> ifc -> logical) {
	case GR_L_BOX:
		res = wrap_into_column (parent, gridp, 
					ref_rown (&parent -> refs));
		break;
		
	case GR_L_COLUMN:
		res = wrap_into_line (parent, gridp, 
				      ref_rown (&parent -> refs));
		break;
		
	case GR_L_LINE:
		/* Tolerate this. Will be done once for all 
		 * by simplify_lines at parent level */
		res = S_CSC_OK;
		break;

	case GR_L_WORD:
	case GR_L_PAGE:
	case GR_L_UNDEFINED:
		assert (0);
	}
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static void simplify (struct gr_grid *grid, struct gr_cell **tcellp)
{
	struct gr_grid *bgrid;
	struct gr_line *bline;
	struct gr_cell **bcellp, *xcell;

	bgrid = (*tcellp) -> grid;
	bline = bgrid -> d.lines;
	assert (bline -> next == NULL);
	for (bcellp = &bline -> cells; *bcellp != NULL; 
	     bcellp = &(*bcellp) -> next) {
		gr_ref ((*bcellp) -> grid, ref_rown (&grid -> refs));
		gr_unref ((*bcellp) -> grid, ref_rown (&bgrid -> refs));
	}
	
	xcell = *tcellp;
	*bcellp = xcell -> next;
	*tcellp = bline -> cells;
	bline -> cells = NULL;

	grid -> words += bgrid -> words - 1;

	destroy_cell (grid, xcell);
}

static enum s_status simplify_lines (struct gr_grid *grid)
{
	struct gr_cell **cellp;
	struct gr_line *line;
	struct gr_grid *bgrid;
	enum s_status res;

	line = grid -> d.lines;
	cellp = &line -> cells;
	while (*cellp != NULL) {
		bgrid = (*cellp) -> grid;
		if (bgrid -> ifc -> logical != GR_L_LINE)
			cellp = &(*cellp) -> next;
		else if (bgrid -> wdir == grid -> wdir &&
			 bgrid -> ldir == grid -> ldir)
			simplify (grid, cellp);
		else {
			res = wrap_into_column (grid, &(*cellp) -> grid,
						ref_rown (&grid -> refs));
			if (res != S_CSC_OK)
				return res;
		}
	}

	return S_CSC_OK;
}

/**
 * ======================
 * Draft->Grid conversion
 * ======================
 * Though quite a simple step, this process is somehow complicated
 * by the fact that we have to convert physical layout (always
 * bottom->top, left->righ) to the text-flow layout. The heuristics
 * are actually in heur.c so what we have to do is to use its
 * routines to detect the proper order and assemble lines
 * and cells by traversing the draft accordingly.
 * 
 * The output grid shall comply with all the rules above, so:
 *	- GR_D_AREA cells alway become words;
 *	- GR_D_INNER cells (i.e. those which do NOT contain larger
 *		gaps) become columns containing lines.
 *	- GR_D_OUTER cells become boxes.
 * Inclusion rules are applied after init_area/init_inner/init_outer
 * area called.
 */
static void common_init (struct gr_grid *grid, 
			 struct a_area_geometry *g,
			 struct iter_data *lid,
			 struct iter_data *wid,
			 struct args_heur *heur)
{
	pdf_real_t gspace;

	grid -> g = *g;
	grid -> wdir = wid -> dir;
	grid -> ldir = lid -> dir;
	grid -> lines = lid -> size;
	grid -> words = wid -> size;
	grid -> fill.ridx = -1;

	gspace = 
		grid -> ldir == GS_D_LEFT_TO_RIGHT ||
		grid -> ldir == GS_D_RIGHT_TO_LEFT ?
		cobj_rect_width (&grid -> g.rect) : 
		cobj_rect_height (&grid -> g.rect);

	grid -> w_graph_space = heur -> w_graph_space * gspace;

	gs_origin_init (&grid -> origin, &g -> rect, 
			grid -> wdir, grid -> ldir);
}

static enum s_status init_area (struct gr_grid *grid, 
				struct gr_draft *draft,
				struct args_heur *heur,
				struct iter_data *lid, 
				struct iter_data *wid)
{
	grid -> ifc = &word_ifc;
	grid -> d.area = draft -> area;
	a_ref (grid -> d.area, ref_rown (&grid -> refs));

	return S_CSC_OK;
}

static pdf_bool_t line_limits (struct gr_region *lreg, struct gr_region *wreg,
			       struct gr_draft **cell, struct iter_data *lid, 
			       struct iter_data *wid)
{
	int i, j;

	for (i = wid -> start; i != wid -> stop; i += wid -> inc)
		if (cell [i] != NULL)
			break;
	if (i == wid -> stop)
		return 0;

	lreg -> lo = lid -> lo_edge (&cell [i] -> g.rect);
	lreg -> xlo = lid -> lo_edge (&cell [i] -> g.bounds);
	lreg -> hi = lid -> hi_edge (&cell [i] -> g.rect);
	lreg -> xhi = lid -> hi_edge (&cell [i] -> g.bounds);

	wreg -> lo = wid -> lo_edge (&cell [i] -> g.rect);
	wreg -> xlo = wid -> lo_edge (&cell [i] -> g.bounds);

	for (j = i; j != wid -> stop; j += wid -> inc)
		if (cell [j] != NULL)
			i = j;

	wreg -> hi = wid -> hi_edge (&cell [i] -> g.rect);
	wreg -> xhi = wid -> hi_edge (&cell [i] -> g.bounds);

	wreg -> next = NULL;

	return 1;
}

static enum s_status init_single_line (struct gr_grid *grid, 
				       struct gr_draft *draft,
				       struct args_heur *heur,
				       struct iter_data *lid, 
				       struct iter_data *wid)
{
	struct gr_line **linep;
	enum s_status res;
	int lidx;

	grid -> ifc = &line_ifc;

	linep = &grid -> d.lines;
	for (lidx = lid -> start; lidx != lid -> stop; lidx += lid -> inc) {
		res = init_line (grid, draft, lid, wid, lidx, heur, linep);
		if (res != S_CSC_OK)
			return res;
		if (*linep != NULL)
			linep = &(*linep) -> next;
	}
	assert (grid -> d.lines != NULL);

	res = simplify_lines (grid);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}


static enum s_status init_inner (struct gr_grid *grid, 
				 struct gr_draft *draft,
				 struct args_heur *heur,
				 struct iter_data *lid, 
				 struct iter_data *wid)
{
	struct a_area_geometry g;
	struct iter_data llid;
	struct gr_line *line, **linep;
	struct gr_cell *cell;
	struct gr_region lreg, wreg;
	enum s_status res;
	int lidx;

	if (grid -> lines == 1)
		return init_single_line (grid, draft, heur, lid, wid);

	grid -> ifc = &column_ifc;

	llid = *lid;
	llid.reg = NULL;	/* To make sure we don't touch it */
	llid.size = 1;
	llid.inc = 1;

	linep = &grid -> d.lines;
	for (lidx = lid -> start; lidx != lid -> stop; lidx += lid -> inc) {

		llid.start = lidx;
		llid.stop = lidx + llid.inc;

		/* Empty lines can appear when the column is part
		 * of a box. 
		 *
		 *  xxxx xx xxx    aaaaaaa aaa
		 *  yyy y yy yy    			<<- empty line
		 *  zzz zzz zzz    ccccc ccccc
		 *	
		 */
		if (!line_limits (&lreg, &wreg, draft -> cell + lidx, 
				  lid, wid))
			continue;

		res = init_line_reg (grid, &lreg, &wreg, heur, linep);
		if (res != S_CSC_OK)
			return res;

		line = *linep;
		if (line == NULL)
			continue;

		cell = line -> cells;

		reg_to_geom (&g, &line -> reg, &cell -> reg, wid -> rotate);
		common_init (cell -> grid, &g, &llid, wid, heur);
		res = init_inner (cell -> grid, draft, heur, &llid, wid);
		if (res != S_CSC_OK)
			return res;

		linep = &line -> next;
	}

	return S_CSC_OK;
}

static enum s_status init_outer (struct gr_grid *grid, 
				 struct gr_draft *draft,
				 struct args_heur *heur,
				 struct iter_data *lid, 
				 struct iter_data *wid)
{
	struct a_area_geometry g;
	struct iter_data wwid;
	struct gr_line *line;
	struct gr_cell *cell;
	struct gr_region creg;
	enum s_status res;
	int widx;

	grid -> ifc = &box_ifc;
	gr_reg_limits (&creg, lid -> reg);
	res = init_line_reg (grid, &creg, wid -> reg, heur, &line);
	if (res != S_CSC_OK)
		return res;

	grid -> d.lines = line;

	cell = line -> cells;
	wwid = *wid;
	wwid.reg = NULL;	/* To make sure we don't touch it */
	wwid.size = 1;
	wwid.inc = 1;
	for (widx = wid -> start; widx != wid -> stop; widx += wid -> inc) {
		wwid.start = widx;
		wwid.stop = widx + wwid.inc;
		reg_to_geom (&g, &line -> reg, &cell -> reg, wwid.rotate);
		common_init (cell -> grid, &g, lid, &wwid, heur);
		res = init_inner (cell -> grid, draft, heur, lid, &wwid);
		if (res != S_CSC_OK)
			return res;

		/* Note that since we are not a line, line/line inclusion
		 * (not handled by apply_inclusions) is not done */
		res = apply_inclusions (grid, &cell -> grid);
		if (res != S_CSC_OK)
			return res;

		cell = cell -> next;
	}

	return S_CSC_OK;
}

static enum s_status init (struct gr_grid *grid, 
			   struct gr_grid *parent,
			   struct gr_draft *draft,
			   struct args_heur *heur)
{
	struct iter_data wid, lid;
	enum s_status res;

	get_iter_data (draft, draft -> wdir,
		       draft -> rows, draft -> cols, &wid);
	get_iter_data (draft, draft -> ldir, 
		       draft -> rows, draft -> cols, &lid);

	grid -> weight = draft -> weight;

	common_init (grid, &draft -> g, &lid, &wid, heur);

	res = S_CSC_OK;
	switch (draft -> type) {
	case GR_D_AREA:
		res = init_area (grid, draft, heur, &lid, &wid);
		break;

	case GR_D_INNER:
 		res = init_inner (grid, draft, heur, &lid, &wid);
		break;

	case GR_D_OUTER:
		res = init_outer (grid, draft, heur, &lid, &wid);
	}
	if (res != S_CSC_OK)
		return res;

	res = apply_inclusions (parent, &grid);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status init_top_level (struct gr_grid **gridp,
				     struct cobj_rect *page_rect,
				     struct args_margins *amys,
				     struct args_heur *heur,
				     struct ref_owner *p)
{
	struct gr_grid *grid;
	pdf_real_t tlspace;
	enum s_status res;

	if ((*gridp) -> ifc -> logical == GR_L_WORD ||
	    (*gridp) -> ifc -> logical == GR_L_LINE) {
		res = wrap_into_column (NULL, gridp, p);
		if (res != S_CSC_OK)
			return res;
	}

	grid = *gridp;
	
	grid -> ifc = &page_ifc;
	res = gr_layout_setup_grid (grid, page_rect, amys, heur);
	if (res != S_CSC_OK)
		return res;

	tlspace = 
		grid -> ldir == GS_D_LEFT_TO_RIGHT ||
		grid -> ldir == GS_D_RIGHT_TO_LEFT ?
		cobj_rect_width (page_rect) : cobj_rect_height (page_rect);
	grid -> pinfo.graph_space = heur -> graph_space * tlspace;	
	
	return S_CSC_OK;
}

enum s_status gr__init (struct gr_grid **gridp, struct gr_grid *parent,
			struct gr_draft *draft, struct args_heur *heur, 
			struct ref_owner *p, const char *fname, int line)
{
	struct gr_grid *grid;
	enum s_status res;

	grid = malloc (sizeof (struct gr_grid));
	if (grid == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (grid, 0, sizeof (struct gr_grid));

	ref__init (&grid -> refs, "grid", p, fname, line);

	if (draft != NULL) {
		res = init (grid, parent, draft, heur);
		if (res != S_CSC_OK) {
			gr_unref (grid, p);
			return res;
		}
	}

	*gridp = grid;

	return S_CSC_OK;
}

static enum s_status post_process (struct gr_grid **gridp,
				   struct gr_al_machine *dc,
				   struct cobj_rect *page_rect,
				   struct args_margins *amys,
				   struct args_heur *heur,
				   int page,
				   struct ref_owner *p)
{
	struct gr_grid *grid;
	enum s_status res;

       	res = gr_c_detect_columns (*gridp, heur);
	if (res != S_CSC_OK)
		return res;

	res = init_top_level (gridp, page_rect, amys, heur, p);
	if (res != S_CSC_OK)
		return res;

	grid = *gridp;

	res = gr_qi_classify_lines (grid, heur);
	if (res != S_CSC_OK)
		return res;

	res = gr_ws_fix_spaces (grid, heur);
	if (res != S_CSC_OK)
		return res;

	res = gr_al_detect_alignment (dc, grid, page);
	if (res != S_CSC_OK)
		return res;

	res = gr_bl_fix_baselines (grid, heur);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

enum s_status gr__get_grid (struct gr_grid **gridp, struct gspg_page *page,
			    struct gr_al_machine *dc, 
			    struct args_margins *amys,
			    struct args_heur *heur, struct ref_owner *p, 
			    const char *fname, int line)
{
	enum gs_direction wdir, ldir;
	struct a_area *area;
	struct gr_draft *draft;
	struct gr_grid *grid;
	enum s_status res;

	res = a__get_areas (&area, page, heur, p, fname, line);
	if (res != S_CSC_OK)
		return res;

	res = a_study_area (&area, p);
	if (res != S_CSC_OK) {
		a_unref (area, p);
		return res;
	}

	heur_rotate_to_dir (heur, page -> rotate, &wdir, &ldir);

	res = gr_d_get_draft (&draft, area, wdir, ldir, heur, p);
	a_unref (area, p);
	if (res != S_CSC_OK)
		return res;

	res = gr__init (&grid, NULL, draft, heur, p, fname, line);
	gr_d_unref (draft, p);
	if (res != S_CSC_OK)
		return res;

	res = post_process (&grid, dc, &page -> bounds, amys, 
			    heur, page -> number, p);
	if (res != S_CSC_OK) {
		gr_unref (grid, p);
		return res;
	}

	*gridp = grid;

	return S_CSC_OK;
}

void gr__ref (struct gr_grid *gr, struct ref_owner *p, 
	      const char *fname, int line)
{
	if (gr != NULL)
		ref__up (&gr -> refs, p, fname, line);
}

void gr_unref (struct gr_grid *grid, struct ref_owner *p)
{
	struct gr_line *line;

	if (grid == NULL || ref_down (&grid -> refs, p))
		return;

	if (grid -> ifc -> logical != GR_L_WORD) {
		while (grid -> d.lines != NULL) {
			line = grid -> d.lines;
			grid -> d.lines = line -> next;
			destroy_line (grid, line);
		} 
	} else
		a_unref (grid -> d.area, ref_rown (&grid -> refs));

	ref_zero (&grid -> refs);

	free (grid);
}

