/*
 * PDFlow -- heur.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <ctype.h>
#include <assert.h>
#include <string.h>
#include <math.h>

#include "grid/heur.h"

#ifndef min
#define min(a, b) ((a) < (b) ? (a) : (b))
#endif

#ifndef max
#define max(a, b) ((a) > (b) ? (a) : (b))
#endif

void heur_get_glyph_bounds (struct gf_glyph_info *gi, 
			    struct cobj_rect *bounds,
			    struct gs_point *vec,
			    struct args_heur *heur)
{
	pdf_real_t dx, dy;

	if (heur != NULL) {
		dx = fabs (vec -> x * heur -> word_glue);
		dy = fabs (vec -> y * heur -> word_glue);
		
		dx += fabs (vec -> y * heur -> glyph_glue);
		dy += fabs (vec -> x * heur -> glyph_glue);
	} else
		dx = dy = 0;

	*bounds = gi -> rect;
	if (gi -> unicode >= 128 || !isblank (gi -> unicode)) {
		bounds -> left -= dx;
		bounds -> right += dx;
		bounds -> bottom -= dy;
		bounds -> top += dy;
	} else
		cobj_rect_set_null (bounds);
}

void heur_extend_cols (pdf_real_t lo, pdf_real_t hi, pdf_real_t svc,
		       pdf_real_t *xlo, pdf_real_t *xhi,
		       struct args_heur *heur)
{
	*xlo = lo - svc * heur -> column_glue;
	*xhi = hi + svc * heur -> column_glue;
}

/* Western (latin): left to right, top to bottom */
static void latin (struct gs_text_string *gs, pdf_real_t vx, pdf_real_t vy,
		   enum gs_direction *pdir, enum gs_direction *sdir)
{
	if (labs (vx) > labs (vy)) {
		if (vx > 0) {
			*pdir = GS_D_LEFT_TO_RIGHT;
			*sdir = GS_D_TOP_TO_BOTTOM;
		} else {
			*pdir = GS_D_RIGHT_TO_LEFT;
			*sdir = GS_D_BOTTOM_TO_TOP;
		}
	} else {
		if (vy < 0) {
			*pdir = GS_D_TOP_TO_BOTTOM;
			*sdir = GS_D_RIGHT_TO_LEFT;
		} else {
			*pdir = GS_D_BOTTOM_TO_TOP;
			*sdir = GS_D_LEFT_TO_RIGHT;
		}
	}
}

/* Chinese, Japanese, Korean, vertical writing mode. I'm assuming
 * top to bottom, right to left */
static void cjk (struct gs_text_string *gs, pdf_real_t vx, pdf_real_t vy,
		   enum gs_direction *pdir, enum gs_direction *sdir)
{
	if (labs (vx) > labs (vy)) {
		if (vx > 0) {
			*pdir = GS_D_LEFT_TO_RIGHT;
			*sdir = GS_D_TOP_TO_BOTTOM;
		} else {
			*pdir = GS_D_RIGHT_TO_LEFT;
			*sdir = GS_D_BOTTOM_TO_TOP;
		}
	} else {
		if (vy < 0) {
			*pdir = GS_D_TOP_TO_BOTTOM;
			*sdir = GS_D_RIGHT_TO_LEFT;
		} else {
			*pdir = GS_D_BOTTOM_TO_TOP;
			*sdir = GS_D_LEFT_TO_RIGHT;
		}
	}
}

void heur_get_direction (struct gs_text_string *gs, struct gs_text_state *tws,
			 enum gs_direction *pdir, enum gs_direction *sdir)
{
	pdf_real_t vx, vy;
	int idx;

	idx = tws -> font -> font -> wmode == CMAP_WM_HORIZONTAL ? 0 : 1;
	vx = tws -> mat.t_m.a [0][idx];
	vy = tws -> mat.t_m.a [1][idx];

	/* I'd really love to add a case for right-to-left writings 
	*  but I don't know how to recognize them */
	if (tws -> font -> font -> wmode == CMAP_WM_HORIZONTAL)
		latin (gs, vx, vy, pdir, sdir);
	else
		cjk (gs, vx, vy, pdir, sdir);
}

void heur_dc_init (struct heur_dir_chooser *hc, struct args_heur *heur)
{
	memset (hc, 0, sizeof (struct heur_dir_chooser));

	hc -> heur = heur;
}

void heur_dc_add (struct heur_dir_chooser *hc, 
		  enum gs_direction pdir, enum gs_direction sdir)
{
	hc -> pstat [pdir]++;
	hc -> sstat [sdir]++;
}

static void choose (int *stat, enum gs_direction *dir, int *certain)
{
	enum gs_direction cur;
	int i;

	cur = GS_D_UNKNOWN;
	for (i = 0; i < GS_D__LAST__; i++)
		if (stat [i] && i != cur) {
			if (cur != GS_D_UNKNOWN) {
				*certain = 0;
				if (stat [i] > stat [cur])
					cur = i;
			} else
				cur = i;
		}

	*dir = cur;
}

static void primary_from_rotate (enum gs_direction *pdir, pdf_integer_t rot)
{
	switch (rot % 360) {
	case 0:
	default:
		*pdir = GS_D_LEFT_TO_RIGHT;
		break;

	case 90:
		*pdir = GS_D_TOP_TO_BOTTOM;
		break;

	case 180:
		*pdir = GS_D_RIGHT_TO_LEFT;
		break;

	case 270:
		*pdir = GS_D_BOTTOM_TO_TOP;
		break;
	}
}

static void secundary_from_primary (enum gs_direction *sdir, 
				    enum gs_direction pdir)
{
	switch (pdir) {
	case GS_D_LEFT_TO_RIGHT:
		*sdir = GS_D_TOP_TO_BOTTOM;
		break;

	case GS_D_TOP_TO_BOTTOM:
		*sdir = GS_D_RIGHT_TO_LEFT;
		break;

	case GS_D_RIGHT_TO_LEFT:
		*sdir = GS_D_BOTTOM_TO_TOP;
		break;

	case GS_D_BOTTOM_TO_TOP:
		*sdir = GS_D_LEFT_TO_RIGHT;
		break;

	default:
		assert (0);
	}
}

void heur_dc_choose (struct heur_dir_chooser *hc,
		     enum gs_direction *pdir,
		     enum gs_direction *sdir,
		     pdf_bool_t *certain)
{
	*certain = 0;

	choose (hc -> pstat, pdir, certain);
	choose (hc -> sstat, sdir, certain);	
}

static pdf_real_t bl (struct heur_bl_chooser *hbl, int idx)
{
	return hbl -> blstat [idx].tval / hbl -> blstat [idx].occurrences;
}

void heur_bl_init (struct heur_bl_chooser *hbl, struct args_heur *heur,
		   enum gs_direction pdir)
{
	hbl -> heur = heur;
	hbl -> idx = 0;
	hbl -> pdir = pdir;
	hbl -> horiz = 
		pdir == GS_D_TOP_TO_BOTTOM || 
		pdir == GS_D_BOTTOM_TO_TOP;
}

void heur_bl_add (struct heur_bl_chooser *hbl, struct gs_point *porigin)
{
	pdf_real_t origin;
	int i;

	origin = hbl -> horiz ? porigin -> x : porigin -> y;

	for (i = 0; i < hbl -> idx; i++)
		if (labs (origin - bl (hbl, i)) < 
		    hbl -> heur -> baseline_tolerance)
			break;

	if (i < hbl -> idx) {
		hbl -> blstat [i].tval += origin;
		hbl -> blstat [i].occurrences++;
	} else if (hbl -> idx < HEUR_BL_MAX) {
		hbl -> blstat [hbl -> idx].tval = origin;
		hbl -> blstat [hbl -> idx].occurrences = 1;
		hbl -> idx++;
	}
}


void heur_bl_choose (struct heur_bl_chooser *hbl,
		     struct cobj_rect *rect,
		     struct gs_point *origin, 
		     pdf_bool_t *has_origin)
{
	int i, best, total;

	gs_origin_half_init (origin, rect, hbl -> pdir);

	best = total = 0;
	for (i = 0; i < hbl -> idx; i++) {
		if (hbl -> blstat [i].occurrences > 
		    hbl -> blstat [best].occurrences)
			best = i;
		total += hbl -> blstat [i].occurrences;
	}

	if (total && hbl -> blstat [best].occurrences >= total / 2) {
		*has_origin = 1;
		if (hbl -> horiz)
			origin -> x = bl (hbl, best);
		else
			origin -> y = bl (hbl, best);
	} else
		*has_origin = 0;
}

void heur_hls_init (struct heur_hls_chooser *hls, 
		    struct args_heur *heur)
{
	memset (hls, 0, sizeof (struct heur_hls_chooser));
	hls -> tol = heur -> w_tolerance;
	hls -> graph = heur -> w_graph_space;
}

void heur_hls_add (struct heur_hls_chooser *hls, pdf_real_t space)
{
	if (!hls -> c.occurrences ||
	    labs (space - hls -> c.tval / hls -> c.occurrences) < hls -> tol) {
		hls -> c.occurrences++;
		hls -> c.tval += space;
	}
}

void heur_hls_choose (struct heur_hls_chooser *hls,
		      pdf_real_t space, pdf_real_t *spu)
{
	*spu = space / (hls -> c.tval / hls -> c.occurrences);
}

pdf_bool_t heur_col_break (struct args_heur *heur, struct gr_region *preg, 
			   struct gr_region *nreg)
{
	return 	preg != NULL && nreg != NULL &&
		preg -> xhi < nreg -> xlo;
}

void heur_rotate_to_dir (struct args_heur *heur, 
			 pdf_integer_t rotate, 
			 enum gs_direction *pdir,
			 enum gs_direction *sdir)
{
	primary_from_rotate (pdir, rotate);
	secundary_from_primary (sdir, *pdir);
}
