/*
 * PDFlow -- layout.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */
/*
 * This file (assisted by heur.c, of course) handles all the 
 * tricky logic that is needed to grok the overall layout of the page. 
 * Unfortunately, this is not only a delicate process, but indispensible 
 * to produce an acceptable output (think of a document where the header
 * is not detected and is continually placed inside of the body text). 
 * 
 * What is done here is to:
 * 	+ Detect headers and footers;
 *	+ Detect chapter boundaries and cover pages;
 *	+ Detect footnotes.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>
#include <assert.h>

#include "grid/layout.h"

#ifndef min
#define min(a, b) ((a) < (b) ? (a) : (b))
#endif

#ifndef max
#define max(a, b) ((a) > (b) ? (a) : (b))
#endif

#define DELTA 10

void gr_layout_init (struct gr_layout *gly, struct args_heur *heur)
{
	memset (gly, 0, sizeof (struct gr_layout));

	gly -> heur = heur;
}

static enum s_status append_page (struct gr_layout_el *gel, pdf_integer_t page)
{
	pdf_integer_t *p;

	gel -> occurrences++;
	while (gel -> pgsz < gel -> occurrences) {
		p = realloc (gel -> pages, 
			     sizeof (pdf_integer_t) * (gel -> pgsz + DELTA));
		if (p == NULL)
			return S_CSC_OUT_OF_MEMORY;
		
		gel -> pages = p;
		gel -> pgsz += DELTA;
	}

	gel -> pages [gel -> occurrences - 1] = page;

	return S_CSC_OK;
}

static enum s_status add_to_els (struct gr_layout_els *gels,
				 struct gr_region *reg,
				 struct args_heur *heur,
				 pdf_integer_t page)
{
	struct gr_layout_el *p, *nel;
	pdf_real_t dlo, dhi;
	enum s_status res;

	/* Remove spurious lines I've found in some docs */
	if (reg -> hi - reg -> lo < heur -> hf_tolerance)
		return S_CSC_OK;

	for (p = gels -> head; p != NULL; p = p -> next) {
		dlo = p -> reg.lo - reg -> lo;
		if (labs (dlo) < heur -> hf_tolerance)
			dlo = 0;

		dhi = p -> reg.hi - reg -> hi;
		if (labs (dhi) < heur -> hf_tolerance)
			dhi = 0;

		if (dlo > 0 || (!dlo && dhi > 0))
			break;

		if (!dlo && !dhi)
			return append_page (p, page);
	}

	nel = malloc (sizeof (struct gr_layout_el));
	if (nel == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (nel, 0, sizeof (struct gr_layout_el));

	res = append_page (nel, page);
	if (res != S_CSC_OK) {
		free (p);
		return res;
	}

	nel -> reg = *reg;

	nel -> next = p;
	nel -> prev = p != NULL ? p -> prev : gels -> tail;

	if (nel -> next != NULL)
		nel -> next -> prev = nel;
	else
		gels -> tail = nel;

	if (nel -> prev != NULL)
		nel -> prev -> next = nel;
	else
		gels -> head = nel;

	return S_CSC_OK;
}

static enum s_status learn_dir (struct gr_layout_els *upgel, 
				struct gr_layout_els *downgel, 
				struct gr_region *reg,
				pdf_real_t lo, pdf_real_t hi,
				struct args_heur *heur,
				pdf_integer_t page)
{
	enum s_status res;
	pdf_real_t hit, lot;

	hit = hi - (hi - lo) * heur -> header_thresh;
	lot = lo + (hi - lo) * heur -> footer_thresh;

	while (reg != NULL) {
		if (reg -> lo >= hit)
			res = add_to_els (downgel, reg, heur, page);
		else if (reg -> hi <= lot)
			res = add_to_els (upgel, reg, heur, page);
		else
			res = S_CSC_OK;

		if (res != S_CSC_OK)
			return res;

		reg = reg -> next;
	}

	return S_CSC_OK;
}

static enum s_status learn (struct gr_layout *gly, struct gr_draft *draft,
			    struct gspg_page *page)
{
	enum s_status res;

	res = learn_dir (&gly -> west, &gly -> east, draft -> hreg, 
			 page -> bounds.left, page -> bounds.right,
			 gly -> heur, gly -> samples);
	if (res != S_CSC_OK)
		return res;

	res = learn_dir (&gly -> south, &gly -> north, draft -> vreg,
			 page -> bounds.bottom, page -> bounds.top,
			 gly -> heur, gly -> samples);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

enum s_status gr_layout_learn (struct gr_layout *gly, struct gspg_page *page)
{
	enum gs_direction wdir, ldir;
	struct gr_draft *draft;
	struct a_area *area;
	struct ref_owner own;
	enum s_status res;

	ref_init_owner (&own, "gr_layout_learn");

	res = a_get_areas (&area, page, gly -> heur, &own);
	if (res != S_CSC_OK)
		return res;

	heur_rotate_to_dir (gly -> heur, page -> rotate, &wdir, &ldir);
	
	res = gr_d_get_draft (&draft, area, wdir, ldir, gly -> heur, &own);
	a_unref (area, &own);
	if (res != S_CSC_OK)
		return res;

	gly -> weight += draft -> weight;

	res = learn (gly, draft, page);
	gr_d_unref (draft, &own);
	if (res != S_CSC_OK)
		return res;

	gly -> samples++;

	return S_CSC_OK;
}

static void delete_page (struct gr_layout_els *gels, pdf_integer_t page)
{
	struct gr_layout_el *gel;
	int i;

	for (gel = gels -> head; gel != NULL; gel = gel -> next) {
		for (i = 0; i < gel -> occurrences; i++) {
			if (gel -> pages [i] > page)
				break;
			if (gel -> pages [i] == page) {
				gel -> occurrences--;
				if (i + 1 < gel -> occurrences)
					memmove (&gel -> pages [i], 
						 &gel -> pages [i + 1],
						 (gel -> occurrences - i) *
						 sizeof (pdf_integer_t));
				break;
			}
		}
	}
}

static void apply_selection (struct gr_layout_els *glys,
			     struct gr_layout_el *gel,
			     pdf_bool_t *pages)
{
	pdf_integer_t pbuf [gel -> occurrences];
	int i, n;

	n = gel -> occurrences;

	memcpy (pbuf, gel -> pages, 
		gel -> occurrences * sizeof (pdf_integer_t));

	for (i = 0; i < n; i++) {
		delete_page (glys, pbuf [i]);
		pages [pbuf [i]] = 1;
	}
}

static pdf_bool_t contains_pages (struct gr_layout_el *gel,
				  struct gr_layout_el *prev)
{
	int i, j;

	if (prev == NULL)
		return 1;

	if (prev -> occurrences > gel -> occurrences)
		return 0;

	j = 0;
	for (i = 0; i < prev -> occurrences; i++) {
		while (j < gel -> occurrences) {
			if (gel -> pages [j] == gel -> pages [i])
				break;
			if (gel -> pages [j] > gel -> pages [j])
				return 0;
			j++;
		}
	}

	return 1;
}

/*
 * This is the real algorithm. It is done at several passes: each
 * pass aims at discovering margins for a different page layout.
 * If a margin has already been selected, it is simply skipped. 
 * We start from the first margin and go on till the number of
 * occurrences becomes lower. However, if we touch the last line
 * in the region we do not return any margin at all since probably
 * the page model does not have any header/footer.
 */
static void select_region_d (struct gr_layout_el **gelp, pdf_real_t *margin,
			     struct gr_layout_els *gels,
			     struct gr_layout *gly, pdf_integer_t samples,
			     pdf_bool_t up)
{
	struct gr_layout_el *gel, *last, *next;
	pdf_real_t gap, height;
	int thresh;

	thresh = samples * gly -> heur -> hf_percentage;
	*gelp = NULL;
	*margin = up ? 0 : -1;
	last = next = NULL;

	for (gel = up ? gels -> head : gels -> tail; gel != NULL; gel = next) {

		next = up ? gel -> next : gel -> prev;

		if (gel -> selected)
			continue;

		if (gel -> occurrences < thresh ||
		    !contains_pages (gel, last))
			break;

		*margin = up ? gel -> reg.hi : gel -> reg.lo;
		gel -> selected = 1;

		last = gel;
		
		if (next == NULL)
			break;

		height = gel -> reg.hi - gel -> reg.lo;
		gap = up ? 
			next -> reg.lo - gel -> reg.hi :
			gel -> reg.lo - next -> reg.hi;

		if (gap > gly -> heur -> hf_gap * height)
			break;
	}

	if (next != NULL)
		*gelp = last;
}

static void select_region (struct gr_layout *gly, 
			   struct args_margins *amys,
			   struct gr_layout_els *gels,
			   enum args_margin_type ltype,
			   pdf_bool_t *pages,
			   pdf_integer_t samples)
{
	pdf_integer_t quorum;
	struct gr_layout_el *gel;
	pdf_real_t margin;
	pdf_bool_t up;

	up = ltype == ARGS_LD_SOUTH || ltype == ARGS_LD_WEST;

	quorum = min (amys -> quorum, gly -> samples);

	while (amys -> margins < amys -> max_margins) {
		select_region_d (&gel, &margin, gels, gly, samples, up);
		if (gel == NULL || gel -> occurrences < quorum)
			break;

		samples -= gel -> occurrences;
		apply_selection (gels, gel, pages);

		amys -> margin [amys -> margins].type = ltype;
		amys -> margin [amys -> margins].value = margin;

		amys -> margins++;
	}
}

void gr_layout_apply (struct gr_layout *gly, struct args_margins *amys)
{
	pdf_bool_t pages [gly -> samples];
	pdf_integer_t midx;
	int i, samples;

	memset (pages, 0, sizeof (pages));
	samples = gly -> samples;

	if (!amys -> skip_w != -1)
		amys -> weight = gly -> weight / gly -> samples;

	while (1) {
		midx = amys -> margins;
		select_region (gly, amys, &gly -> north, 
			       ARGS_LD_NORTH, pages, samples);
		select_region (gly, amys, &gly -> east, 
			       ARGS_LD_EAST, pages, samples);
		select_region (gly, amys, &gly -> west, 
			       ARGS_LD_WEST, pages, samples);
		select_region (gly, amys, &gly -> south, 
			       ARGS_LD_SOUTH, pages, samples);
		if (amys -> margins == midx)
			break;

		samples = gly -> samples;
		for (i = 0; i < gly -> samples; i++) {
			if (pages [i] == 1) {
				delete_page (&gly -> north, i);
				delete_page (&gly -> east, i);
				delete_page (&gly -> west, i);
				delete_page (&gly -> south, i);
				pages [i] = -1;
			}
			if (pages [i])
				samples--;
		}
	}
}

static struct gr_grid *first (struct gr_line *line)
{
	struct gr_cell *cell;

	for (cell = line -> cells; cell != NULL; cell = cell -> next)
		if (cell -> next == NULL)
			return cell -> grid;

	return NULL;
}

static struct gr_grid *last (struct gr_line *line)
{
	return line -> cells != NULL ? line -> cells -> grid : NULL;
}

static void def_margin (struct gr_region *reg, enum args_margin_type mt,
			struct cobj_rect *page_rect, struct cobj_rect *tl_rect)
{
	pdf_real_t hspace, vspace;

	hspace = min (tl_rect -> bottom - page_rect -> bottom,
		      page_rect -> top - tl_rect -> top);
	vspace = min (tl_rect -> left - page_rect -> left,
		      page_rect -> right - tl_rect -> right);

	switch (mt) {
	case ARGS_LD_NORTH:
		reg -> lo = reg -> hi = page_rect -> top - vspace;
		break;

	case ARGS_LD_EAST:
		reg -> lo = reg -> hi = hspace;
		break;

	case ARGS_LD_WEST:
		reg -> lo = reg -> hi = page_rect -> right - hspace;
		break;

	case ARGS_LD_SOUTH:
		reg -> lo = reg -> hi = hspace;
		break;
	}
}

static void detect_margin (struct gr_grid *grid,
			   struct cobj_rect *page_rect,
			   struct args_margins *amys,
			   struct args_heur *heur,
			   enum args_margin_type hmt,
			   enum args_margin_type fmt)
{
	struct gr_page_info *pinfo;
	struct gr_region hreg, freg;
	struct gr_line *line;
	int i;

	pinfo = &grid -> pinfo;

	def_margin (&hreg, hmt, page_rect, &grid -> g.rect);
	def_margin (&freg, fmt, page_rect, &grid -> g.rect);

	pinfo -> first = NULL;
	pinfo -> last = NULL;
	for (line = grid -> d.lines; line != NULL; line = line -> next) {
		for (i = 0; i < amys -> margins; i++) {
			if (amys -> margin [i].type == hmt &&
			    labs (amys -> margin [i].value - 
				  line -> reg.hi) < heur -> hf_tolerance) {
				hreg = line -> reg;
				pinfo -> header.last = last (line);
				pinfo -> first = line -> next;
			}
				
			if (amys -> margin [i].type == fmt &&
			    labs (amys -> margin [i].value - 
				  line -> reg.lo) < heur -> hf_tolerance) {
				freg = line -> reg;
				pinfo -> footer.last = first (line);
			}
		}

		if (pinfo -> footer.last != NULL)
			break;

		pinfo -> last = line;
	}		

	pinfo -> header.space = pinfo -> first != NULL ? 
		labs (hreg.hi - pinfo -> first -> reg.lo) : 0;

	pinfo -> footer.space = pinfo -> last != NULL ? 
		labs (pinfo -> last -> reg.hi - freg.lo) : 0;

	pinfo -> header.pos = hreg.hi;
	pinfo -> footer.pos = freg.lo;

	if (pinfo -> first == NULL)
		pinfo -> first = grid -> d.lines;

	if (pinfo -> header.space > pinfo -> footer.space) {
		pinfo -> header.space -= 
			gr_reg_width (&hreg) * heur -> hf_gap;
		pinfo -> footer.space += 
			gr_reg_width (&freg) * heur -> hf_gap;
	} else {
		pinfo -> header.space += 
			gr_reg_width (&hreg) * heur -> hf_gap;
		pinfo -> footer.space -= 
			gr_reg_width (&freg) * heur -> hf_gap;
	}
}

static void detect_page_type (struct gr_grid *grid,
			      struct args_margins *amys,
			      struct args_heur *heur)
{
	struct gr_page_info *pinfo;
	pdf_bool_t light, heavy;

	pinfo = &grid -> pinfo;

	if (pinfo -> first == NULL) {
		pinfo -> type = GR_PT_EMPTY;
		return;
	}

	heavy = (amys -> weight - grid -> weight) >= heur -> weight_tolerance;
	light = (amys -> weight - grid -> weight) <= -heur -> weight_tolerance;
	if (grid -> ldir == GS_D_LEFT_TO_RIGHT ||
	    grid -> ldir == GS_D_BOTTOM_TO_TOP) {
		if (heavy)
			pinfo -> type = GR_PT_SECTION_END;
		else if (light)
			pinfo -> type = GR_PT_SECTION_START;
		else
			pinfo -> type = GR_PT_ORDINARY;
	} else {
		if (light)
			pinfo -> type = GR_PT_SECTION_END;
		else if (heavy)
			pinfo -> type = GR_PT_SECTION_START;
		else
			pinfo -> type = GR_PT_ORDINARY;
	}
}

enum s_status gr_layout_setup_grid (struct gr_grid *grid,
				    struct cobj_rect *page_rect,
				    struct args_margins *amys,
				    struct args_heur *heur)
{
	switch (grid -> ldir) {
	case GS_D_LEFT_TO_RIGHT:
		detect_margin (grid, page_rect, amys, heur, 
			       ARGS_LD_EAST, ARGS_LD_WEST);
		break;

	case GS_D_RIGHT_TO_LEFT:
		detect_margin (grid, page_rect, amys, heur, 
			       ARGS_LD_WEST, ARGS_LD_EAST);
		break;

	case GS_D_BOTTOM_TO_TOP:
		detect_margin (grid, page_rect, amys, heur, 
			       ARGS_LD_SOUTH, ARGS_LD_NORTH);
		break;

	case GS_D_TOP_TO_BOTTOM:
		detect_margin (grid, page_rect, amys, heur, 
			       ARGS_LD_NORTH, ARGS_LD_SOUTH);
		break;

	case GS_D_UNKNOWN:
	case GS_D__LAST__:
		assert (0);
	}

	detect_page_type (grid, amys, heur);

	return S_CSC_OK;
}

static void destroy_el (struct gr_layout_els *gels)
{
	struct gr_layout_el *p;

	while (gels -> head != NULL) {
		p = gels -> head;
		gels -> head = p -> next;
		if (p -> pages != NULL)
			free (p -> pages);
		free (p);
	}

	gels -> tail = NULL;
}

void gr_layout_destroy (struct gr_layout *gly)
{
	destroy_el (&gly -> north);
	destroy_el (&gly -> east);
	destroy_el (&gly -> west);
	destroy_el (&gly -> south);
}
