/*
 * PDFlow -- lc.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <assert.h>
#include <string.h>

#include "grid/thread.h"
#include "grid/lc.h"

enum option_type {

	OTT_SAME_LINE_SAME_PAGE,
	OTT_NEXT_LINE_SAME_PAGE,
	OTT_SAME_LINE_NEXT_PAGE,
	OTT_NEXT_LINE_NEXT_PAGE,

	OTT__LAST__

};

/* Unused, hide warning */
#if 0
static void line_cell__ref (struct gr_line_cell *lcl, struct ref_owner *p,
			    const char *fname, int line)
{
	if (lcl != NULL)
		ref__up (&lcl -> refs, p, fname, line);
}
#define line_cell_ref(lcl, p) line_cell__ref (lcl, p, __FILE__, __LINE__)
#endif

static void line_cell_unref (struct gr_line_cell *lcl, struct ref_owner *p)
{
	if (lcl == NULL || ref_down (&lcl -> refs, p))
		return;

	agc_unref (lcl -> agc, ref_rown (&lcl -> refs));
	gr_unref (lcl -> grid, ref_rown (&lcl -> refs));

	ref_zero (&lcl -> refs);

	free (lcl);
}

static enum s_status line_cell__new (struct gr_line_cell **lclp,
				     struct gr_line_composer *lc,
				     struct agc_context *agc,
				     struct gr_squeeze_result *sqr,
				     struct gr_grid *grid,
				     struct td_space *wspace,
				     struct ref_owner *p,
				     const char *fname, int line)
{
	struct gr_line_cell *lcl;

	lcl = malloc (sizeof (struct gr_line_cell));
	if (lcl == NULL)
		return S_CSC_OUT_OF_MEMORY;

	memset (lcl, 0, sizeof (struct gr_line_cell));

	ref__init (&lcl -> refs, "line_cell", p, fname, line);

	lcl -> agc = agc;
	if (sqr != NULL) {
		lcl -> rect = sqr -> rect;
		lcl -> bounds = sqr -> bounds;
		lcl -> origin = sqr -> origin;
	} else {
		cobj_rect_set_null (&lcl -> rect);
		cobj_rect_set_null (&lcl -> bounds);
		lcl -> origin.x = lcl -> origin.y = 0;
	}
	lcl -> grid = grid;
	lcl -> wspace = *wspace;

	agc_ref (agc, ref_rown (&lcl -> refs));
	gr_ref (grid, ref_rown (&lcl -> refs));

	*lclp = lcl;

	return S_CSC_OK;
}
#define line_cell_new(lclp, lc, agc, sqr, wspace, grid, p) \
line_cell__new (lclp, lc, agc, sqr, wspace, grid, p, __FILE__, __LINE__)

static enum s_status lc_init (struct gr_line_composer *lc, gr_eject eject,
			      void *ejclosure, struct gr_lc_param *lcp,
			      struct cobj_rect *rect)
{
	enum s_status res;

	res = td_pool_new (&lc -> tdp, rect, ref_rown (&lc -> refs));
	if (res != S_CSC_OK)
		return res;

	lc -> r.lavail = *rect;
	lc -> eject = eject;
	lc -> ejclosure = ejclosure;
	lc -> lcp = *lcp;

	return S_CSC_OK;
}

enum s_status gr_lc__new (struct gr_line_composer **lcpp, gr_eject eject,
			  void *ejclosure, struct gr_lc_param *lcp,
			  struct cobj_rect *rect, 
			  struct ref_owner *p, const char *fname, int line)
{
	struct gr_line_composer *lc;
	enum s_status res;

	lc = malloc (sizeof (struct gr_line_composer));
	if (lc == NULL)
		return S_CSC_OUT_OF_MEMORY;

	memset (lc, 0, sizeof (struct gr_line_composer));

	ref__init (&lc -> refs, "line_composer", p, fname, line);

	res = lc_init (lc, eject, ejclosure, lcp, rect);
	if (res != S_CSC_OK) {
		gr_lc_unref (lc, p);
		return res;
	}

	*lcpp = lc;

	return S_CSC_OK;
}

void gr_lc__ref (struct gr_line_composer *lc, struct ref_owner *p,
		 const char *fname, int line)
{
	if (lc != NULL)
		ref__up (&lc -> refs, p, fname, line);
}

static void destroy_cells (struct gr_line_composer *lc)
{
	struct gr_line_cell *cell;

	cell = lc -> cells;
	if (cell != NULL) {
		lc -> cells = cell -> next;
		cell -> next = NULL;

		while (lc -> cells != NULL) {
			cell = lc -> cells;
			lc -> cells = cell -> next;
			
			line_cell_unref (cell, ref_rown (&lc -> refs));
		}

		lc -> cells = NULL;
	}
}

void gr_lc_unref (struct gr_line_composer *lc, struct ref_owner *p)
{
	if (lc == NULL || ref_down (&lc -> refs, p))
		return;

	destroy_cells (lc);

	td_pool_unref (lc -> tdp, ref_rown (&lc -> refs));

	ref_zero (&lc -> refs);

	free (lc);
}

static void xchg (pdf_real_t *a, pdf_real_t *b)
{
	pdf_real_t x;

	x = *a;
	*a = *b;
	*b = x;
}

static void reset_rects (struct gr_line_rects *lcr,
			 struct td_pool *tdp,
			 pdf_bool_t first_line,
			 struct args_output *aout)
{
	td_pool_avail (tdp, &lcr -> lavail);
	lcr -> ascent = lcr -> descent = 0;
	lcr -> blanks = lcr -> strict = 0;
	if (first_line) {
		lcr -> dhead = lcr -> al.fhead * aout -> text_ratio;
		lcr -> dtail = lcr -> al.ftail * aout -> text_ratio;
	} else {
		lcr -> dhead = lcr -> al.lhead * aout -> text_ratio;
		lcr -> dtail = lcr -> al.ltail * aout -> text_ratio;
	}
	
	if (lcr -> al.wdir == GS_D_LEFT_TO_RIGHT ||
	    lcr -> al.wdir == GS_D_RIGHT_TO_LEFT)
		lcr -> line_size = cobj_rect_width (&lcr -> lavail);
	else
		lcr -> line_size = cobj_rect_height (&lcr -> lavail);
	
	switch (lcr -> al.at) {
	case GR_AT_BEGIN:
		lcr -> dtail = 0;
		break;
		
	case GR_AT_CENTER:
		lcr -> dhead = 0;
		lcr -> dtail = 0;
		break;

	case GR_AT_JUSTIFIED:
		break;

	case GR_AT_END:
		lcr -> dhead = 0;
		break;

	case GR_AT_PROPORTIONAL:
		lcr -> ratio = lcr -> dhead / (lcr -> dtail + lcr -> dhead);
		lcr -> dhead = 0;
		lcr -> dtail = 0;
		break;

	case GR_AT_NONE:
	case GR_AT_UNKNOWN:
	case GR_AT__LAST__:
		assert (0);
	}

	lcr -> strict += lcr -> dhead + lcr -> dtail;

	switch (lcr -> al.wdir) {
	case GS_D_LEFT_TO_RIGHT:
		lcr -> lavail.left += lcr -> dhead; 
		lcr -> lavail.right -= lcr -> dtail; 
		break;

	case GS_D_RIGHT_TO_LEFT:
		lcr -> lavail.right -= lcr -> dhead; 
		lcr -> lavail.left += lcr -> dtail; 
		break;

	case GS_D_BOTTOM_TO_TOP:
		lcr -> lavail.bottom += lcr -> dhead; 
		lcr -> lavail.top -= lcr -> dtail; 
		break;

	case GS_D_TOP_TO_BOTTOM:
		lcr -> lavail.top -= lcr -> dhead; 
		lcr -> lavail.bottom += lcr -> dtail; 
		break;

	case GS_D_UNKNOWN:
	case GS_D__LAST__:
		assert (0);
	}

	if (lcr -> lavail.right < lcr -> lavail.left ||
	    lcr -> lavail.top < lcr -> lavail.bottom) {
		if (lcr -> al.at == GR_AT_PROPORTIONAL) {
			lcr -> al.at = GR_AT_JUSTIFIED;
			lcr -> al.flush_at = GR_AT_BEGIN;
		}
		td_pool_avail (tdp, &lcr -> lavail);
		lcr -> dhead = lcr -> dtail = 0;
	}
}

static void update_rects (struct gr_line_composer *lc, 
			  struct gr_squeeze_result *sqr)
{
	pdf_real_t strict, blanks, ascent, descent;

	switch (lc -> r.al.wdir) {
	case GS_D_LEFT_TO_RIGHT:
	case GS_D_RIGHT_TO_LEFT:
		strict = cobj_rect_width (&sqr -> rect);
		blanks = cobj_rect_width (&sqr -> bounds) - strict;
		ascent = sqr -> bounds.top - sqr -> origin.y;
		descent = sqr -> origin.y - sqr -> bounds.bottom;

		if (lc -> r.al.wdir == GS_D_RIGHT_TO_LEFT) {
			xchg (&ascent, &descent);
			lc -> r.lavail.right -= strict + blanks;
		} else
			lc -> r.lavail.left += strict + blanks;
		break;

	case GS_D_TOP_TO_BOTTOM:
	case GS_D_BOTTOM_TO_TOP:
		strict = cobj_rect_height (&sqr -> rect);
		blanks = cobj_rect_height (&sqr -> bounds) - strict;
		ascent = sqr -> origin.x - sqr -> bounds.left;
		descent = sqr -> bounds.right - sqr -> origin.x;
		if (lc -> r.al.wdir == GS_D_TOP_TO_BOTTOM) {
			xchg (&ascent, &descent);
			lc -> r.lavail.top -= strict + blanks; 
		} else
			lc -> r.lavail.bottom += strict + blanks;
		break;

	default:
		assert (0);
	}

	if (ascent > lc -> r.ascent)
		lc -> r.ascent = ascent;
	if (descent > lc -> r.descent)
		lc -> r.descent = descent;

	lc -> r.blanks += lc -> r.last_word_blank = blanks;
	lc -> r.strict += strict;
}

static void get_spaces (struct gr_line_rects *lcr, 
			pdf_real_t cspace, pdf_real_t gspace,
			pdf_real_t *head, pdf_real_t *tail,
			pdf_real_t *extra, pdf_real_t *gextra,
			pdf_bool_t hard)
{
	pdf_real_t space;

	space = lcr -> line_size - lcr -> strict - 
		(lcr -> blanks - lcr -> last_word_blank);
	if (space < 0)
		space = 0;

	if (gspace) {
		*gextra = space / gspace;
		space = 0;
	} else
		*gextra = 0;

	switch (hard ? lcr -> al.flush_at : lcr -> al.at) {
	case GR_AT_BEGIN:
		*head = lcr -> dhead;
		*tail = space;
		*extra = 0;
		return;

	case GR_AT_CENTER:
		*head = *tail = space / 2;
		*extra = 0;
		return;
		
	case GR_AT_JUSTIFIED:
		*head = lcr -> dhead;
		*tail = lcr -> dtail;
		*extra = cspace > 0 ? space / cspace : 0;
		return;

	case GR_AT_END:
		*head = space;
		*tail = lcr -> dtail;
		*extra = 0;
		return;

	case GR_AT_PROPORTIONAL:
		*head = lcr -> ratio * space;
		*tail = (1 - lcr -> ratio) * space;
		*extra = 0;
		return;

	case GR_AT_NONE:
	case GR_AT_UNKNOWN:
		assert (0);
	}

	assert (0);
}

static enum s_status flush_line (struct gr_line_composer *lc,
				 pdf_bool_t hard)
{
	struct td_line_status tdl;
	pdf_real_t head, tail, extra, gextra;
	struct gr_line_cell *cell;
	pdf_real_t gspace, cspace;
	enum s_status res;
	pdf_bool_t last;

	if (lc -> cells == NULL)
		return S_CSC_OK;

	cell = lc -> cells;
	gspace = cspace = 0;
	do {
		last = cell == lc -> cells;

		cell = cell -> next;
		if (cell -> wspace.type == TD_ST_GRAPHICAL)
			gspace += cell -> wspace.amount;
		else if (!last)
			cspace += cell -> wspace.amount;
	} while (cell != lc -> cells);

	get_spaces (&lc -> r, cspace, gspace, 
		    &head, &tail, &extra, &gextra, hard);

	res = td_float_bol (&tdl, lc -> tdf, lc -> r.ascent, lc -> r.descent);
	if (res != S_CSC_OK)
		return res;

	res = S_CSC_OK;
	cell = lc -> cells;

	res = td_float_space (&tdl, head);
	if (res != S_CSC_OK)
		return res;

	do {
		cell = cell -> next;

		if (cell -> grid != NULL) {
			res = td_float_write (&tdl, cell -> agc, 
					      &cell -> bounds, 
					      &cell -> origin);
			if (res != S_CSC_OK)
				break;
		}

		if (cell -> wspace.type == TD_ST_GRAPHICAL)
			cspace = gextra * cell -> wspace.amount;
		else
			cspace = extra;

		res = td_float_space (&tdl, cspace);
		if (res != S_CSC_OK)
			break;

	} while (cell != lc -> cells);

	if (res == S_CSC_OK)
		res = td_float_eol (&tdl);
	if (res != S_CSC_OK)
		return res;

	destroy_cells (lc);
	reset_rects (&lc -> r, lc -> tdp, 0, lc -> lcp.aout);

	return S_CSC_OK;
}

static enum s_status flush_par (struct gr_line_composer *lc, pdf_real_t space)
{
	struct td_space tds;
	enum s_status res;

	if (lc -> tdf == NULL)
		return S_CSC_OK;

	res = flush_line (lc, 1);
	if (res != S_CSC_OK)
		return res;

	tds.type = TD_ST_GRAPHICAL;
	tds.amount = space;

	td_float_done (lc -> tdp, lc -> tdf, space ? &tds : NULL);

	return S_CSC_OK;
}

static enum s_status change_page (struct gr_line_composer *lc)
{
	struct agc_context *agc;
	struct cobj_rect rect;
	enum s_status res;

	res = td_pool_finalize (lc -> tdp, &agc, ref_rown (&lc -> refs));
	if (res != S_CSC_OK)
		return res;

	res = lc -> eject (agc, &rect, lc -> ejclosure);
	agc_unref (agc, ref_rown (&lc -> refs));
	if (res != S_CSC_OK)
		return res;

	lc -> dirty = 0;

	res = td_float_alloc (&lc -> tdf, lc -> tdp, lc -> r.al.wdir, 
			      lc -> r.al.ldir, NULL);
	if (res != S_CSC_OK)
		return res;

	lc -> page_flushed = 1;
	
	return S_CSC_OK;
}

static enum s_status flush_page (struct gr_line_composer *lc, pdf_real_t space)
{
	enum s_status res;

	if (!lc -> dirty)
		return S_CSC_OK;

	res = flush_par (lc, space);
	if (res != S_CSC_OK)
		return res;

	res = change_page (lc);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

enum s_status gr_lc_finalize (struct gr_line_composer *lc)
{
	enum s_status res;

	res = flush_page (lc, 0);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status unhyphen (struct gr_line_composer *lc,
			       struct gr_line_cell *pcell, 
			       struct gr_line_cell *ncell)
{
	pdf_real_t dwidth, dheight;
	enum s_status res;

	agc_disable_code (pcell -> agc, GR_AGCC_HYPHEN);
	dwidth = cobj_rect_width (&pcell -> rect);
	dheight = cobj_rect_height (&pcell -> rect);
	res = agc_measure (pcell -> agc, &pcell -> rect);
	if (res != S_CSC_OK)
		return res;
	dwidth -= cobj_rect_width (&pcell -> rect);
	dheight -= cobj_rect_height (&pcell -> rect);

	pcell -> wspace.type = TD_ST_ORDINARY;
	pcell -> wspace.amount = 0;

	switch (lc -> r.al.wdir) {
	case GS_D_LEFT_TO_RIGHT:
		lc -> r.blanks -= labs (pcell -> bounds.right -
					pcell -> rect.right);
		ncell -> origin.x += labs (ncell -> bounds.left -
					   ncell -> rect.left);
		lc -> r.strict -= dwidth;
		pcell -> bounds.right = pcell -> rect.right;
		ncell -> bounds.left = ncell -> rect.left;
		break;

	case GS_D_RIGHT_TO_LEFT:
		lc -> r.blanks -= labs (pcell -> bounds.left -
					pcell -> rect.left);
		ncell -> origin.x += labs (ncell -> bounds.right -
					   ncell -> rect.right);
		lc -> r.strict -= dwidth;
		pcell -> bounds.left = pcell -> rect.left;
		ncell -> bounds.right = ncell -> rect.right;
		break;

	case GS_D_BOTTOM_TO_TOP:
		lc -> r.blanks -= labs (pcell -> bounds.top -
					pcell -> rect.top);
		lc -> r.strict -= dheight;
		ncell -> origin.y += labs (ncell -> bounds.bottom -
					   ncell -> rect.bottom);
		pcell -> bounds.top = pcell -> rect.top;
		ncell -> bounds.bottom = ncell -> rect.bottom;
		break;

	case GS_D_TOP_TO_BOTTOM:
		lc -> r.blanks -= labs (pcell -> bounds.bottom -
					pcell -> rect.bottom);
		lc -> r.strict -= dheight;
		ncell -> origin.y += labs (ncell -> bounds.top -
					   ncell -> rect.top);
		pcell -> bounds.bottom = pcell -> rect.bottom;
		ncell -> bounds.top = ncell -> rect.top;
		break;
		
	case GS_D_UNKNOWN:
	case GS_D__LAST__:
		assert (0);
	}

	return S_CSC_OK;
}

static enum s_status append_cell (struct gr_line_composer *lc,
				  struct agc_context *agc,
				  struct gr_squeeze_result *sqr,
				  struct gr_grid *grid,
				  struct td_space *wspace)
{
	struct gr_line_cell *lcl;
	enum s_status res;

	res = line_cell_new (&lcl, lc, agc, sqr, grid, wspace,
			     ref_rown (&lc -> refs));
	if (res != S_CSC_OK)
		return res;

	if (lc -> cells != NULL) {
		if (grid != NULL && lc -> cells -> grid != NULL &&
		    lc -> cells -> grid -> hyphen != NULL) {
			res = unhyphen (lc, lc -> cells, lcl);
			if (res != S_CSC_OK) {
				line_cell_unref (lcl, ref_rown (&lc -> refs));
				return res;
			}
		}

		lcl -> next = lc -> cells -> next;
		lc -> cells -> next = lcl;
	} else
		lc -> cells = lcl -> next = lcl;

	lc -> cells = lcl;

	return S_CSC_OK;
}

static enum s_status do_update_line (struct gr_line_composer *lc,
				     struct gr_squeeze_result *sqr,
				     struct agc_context *agc,
				     struct gr_grid *grid,
				     struct td_space *wspace)
{
	enum s_status res;

	lc -> dirty = 1;

	res = append_cell (lc, agc, sqr, grid, wspace);
	if (res != S_CSC_OK)
		return res;

	update_rects (lc, sqr);

	if (sqr -> badness > lc -> badness)
		lc -> badness = sqr -> badness;

	return S_CSC_OK;
}

static enum s_status update_line (struct gr_line_composer *lc,
				  struct gr_squeeze_result *sqr,
				  struct agc_context *agc,
				  struct gr_grid *grid,
				  struct td_space *wspace,
				  enum option_type ott)
{	
	enum s_status res;

	if (ott == OTT_NEXT_LINE_SAME_PAGE ||
	    ott == OTT_NEXT_LINE_NEXT_PAGE) {
		res = flush_line (lc, 0);
		if (res != S_CSC_OK)
			return res;
	}

	if (ott == OTT_SAME_LINE_NEXT_PAGE ||
	    ott == OTT_NEXT_LINE_NEXT_PAGE) {
		res = change_page (lc);
		if (res != S_CSC_OK)
			return res;

		reset_rects (&lc -> r, lc -> tdp, grid -> al.at != GR_AT_NONE, 
			     lc -> lcp.aout);
	}

	
	res = do_update_line (lc, sqr, agc, grid, wspace);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status change_par (struct gr_line_composer *lc,
				 struct gr_al_alignment *al,
				 struct td_space *space)
{
	enum s_status res;

	res = flush_par (lc, 0);
	if (res != S_CSC_OK)
		return res;

	space -> amount *= lc -> lcp.aout -> space_ratio;
	if (space -> type == TD_ST_GRAPHICAL)
		space -> amount *= lc -> lcp.aout -> graph_space_el;
		
	res = td_float_alloc (&lc -> tdf, lc -> tdp, al -> wdir, 
			      al -> ldir, space);
	if (res != S_CSC_OK)
		return res;

	lc -> r.al = *al;
	reset_rects (&lc -> r, lc -> tdp, 1, lc -> lcp.aout);

	return S_CSC_OK;
}

static enum s_status handle_fill_char (struct gr_line_composer *lc,
				       struct gr_grid *cell)
{
	struct td_space wspace;
	enum s_status res;
	
	/* All processing is done when the first fill char is 
	 * encountered */
	if (cell -> fill.ridx > 0)
		return S_CSC_OK;

	wspace.type = TD_ST_GRAPHICAL;
	wspace.amount = 1;

	res = append_cell (lc, NULL, NULL, NULL, &wspace);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status add_cell (struct gr_line_composer *lc, 
			       struct gr_grid *cell,
			       struct td_space *wspace,
			       struct td_space *lspace)
{
	struct gr_squeeze_option opt [OTT__LAST__];
	struct agc_context *agc;
	enum option_type ott, best;
	pdf_real_t pspace;
	enum s_status res;

	if (cell -> al.at != GR_AT_NONE) {
		res = change_par (lc, &cell -> al, lspace);
		if (res != S_CSC_OK)
			return res;
	}

	if (cell -> fill.ridx >= 0)
		return handle_fill_char (lc, cell);

	res = agc_float_new (&agc, ref_rown (&lc -> refs));
	if (res != S_CSC_OK)
		return res;

	if (wspace -> type == TD_ST_ORDINARY)
		pspace = wspace -> amount;
	else
		pspace = 0;

	for (ott = 0; ott < OTT__LAST__; ott++) {
		switch (ott) {
		case OTT_SAME_LINE_SAME_PAGE:
			opt [ott].avail = lc -> r.lavail;
			break;

		case OTT_NEXT_LINE_SAME_PAGE:
			td_pool_avail (lc -> tdp, &opt [ott].avail);
			break;

		case OTT_SAME_LINE_NEXT_PAGE:
			td_pool_line_avail (lc -> tdp, &opt [ott].avail);
			break;

		case OTT_NEXT_LINE_NEXT_PAGE:
			opt [ott].avail = lc -> tdp -> total;
			break;

		case OTT__LAST__:
			assert (0);
		}
	}
	
	cell -> ifc -> probe (cell, pspace, lc -> lcp.aout, OTT__LAST__, opt);
	opt [OTT_NEXT_LINE_SAME_PAGE].sqr.badness += HEUR_BADNESS_NEW_LINE;
	opt [OTT_NEXT_LINE_NEXT_PAGE].sqr.badness += HEUR_BADNESS_NEW_LINE;

	opt [OTT_SAME_LINE_NEXT_PAGE].sqr.badness += HEUR_BADNESS_NEW_PAGE;
	opt [OTT_NEXT_LINE_NEXT_PAGE].sqr.badness += HEUR_BADNESS_NEW_PAGE;

	best = 0;
	for (ott = 1; ott < OTT__LAST__; ott++)
		if (opt [ott].sqr.badness < opt [best].sqr.badness)
			best = ott;

	res = cell -> ifc -> squeeze (cell, agc, &opt [best]);
	if (res != S_CSC_OK) {
		agc_unref (agc, ref_rown (&lc -> refs));
		return res;
	}
	
	res = update_line (lc, &opt [best].sqr, agc, cell, wspace, best);
	agc_unref (agc, ref_rown (&lc -> refs));
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status feed (struct gr_grid *grid, struct gr_th_info *info,
			   void *closure)
{
	struct gr_line_composer *lc;
	enum s_status res;

	lc = closure;

	/* If this assertion fails, there's some incongruence
	 * between squeeze and thread */
	assert (grid -> ifc -> squeeze != NULL);

	res = add_cell (lc, grid, &grid -> wspace, &info -> lspace);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

enum s_status gr_lc_feed (struct gr_line_composer *lc, struct gr_grid *grid)
{
	enum s_status res;

	if (grid -> pinfo.type == GR_PT_COVER ||
	    grid -> pinfo.type == GR_PT_SECTION_START ||
	    grid -> pinfo.type == GR_PT_EMPTY) {
		res = flush_page (lc, 0);
		if (res != S_CSC_OK)
			return res;
	}

	lc -> page_flushed = 0;

	res = gr_th_traverse (grid, lc -> lcp.heur, feed, lc);
	if (res != S_CSC_OK)
		return res;

	if (!lc -> page_flushed ||
	    grid -> pinfo.type == GR_PT_COVER ||
	    grid -> pinfo.type == GR_PT_SECTION_END) {
		res = flush_page (lc, grid -> pinfo.footer.space);
		if (res != S_CSC_OK)
			return res;
	}

	return S_CSC_OK;
}

