/*
 * PDFlow -- region.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "grid/heur.h"
#include "grid/region.h"

#ifndef min
#define min(a, b) ((a) < (b) ? (a) : (b))
#endif

#ifndef max
#define max(a, b) ((a) > (b) ? (a) : (b))
#endif

static enum s_status append_region (struct gr_region **regp,
				    struct gr_region *spec)
{
	struct gr_region *reg;

	reg = malloc (sizeof (struct gr_region));
	if (reg == NULL)
		return S_CSC_OUT_OF_MEMORY;

	*reg = *spec;

	reg -> next = *regp;

	*regp = reg;

	return S_CSC_OK;
}

int gr_reg_size (struct gr_region *reg)
{
	int i;

	i = 0;
	while (reg != NULL) {
		i++;
		reg = reg -> next;
	}

	return i;
}

pdf_real_t gr_reg_width (struct gr_region *reg)
{
	return labs (reg -> hi - reg -> lo);
}

void gr_reg_destroy (struct gr_region *reg)
{
	struct gr_region *p;

	while (reg != NULL) {
		p = reg;
		reg = p -> next;
		free (p);
	}
}

void gr_reg_limits (struct gr_region *limits, struct gr_region *reg)
{
	limits -> lo = reg -> lo;
	limits -> xlo = reg -> xlo;

	while (reg -> next != NULL)
		reg = reg -> next;

	limits -> hi = reg -> hi;
	limits -> xhi = reg -> xhi;

	limits -> next = NULL;
}

enum s_status gr_reg_dup (struct gr_region **dreg, 
			  struct gr_region *sreg)
{
	struct gr_region **r, *p;

	for (r = dreg, p = sreg; p != NULL; r = &(*r) -> next, p = p -> next) {
		*r = malloc (sizeof (struct gr_region));
		if (*r == NULL)
			break;
		**r = *p;
	}

	*r = NULL;

	if (p != NULL) {
		gr_reg_destroy (*dreg);
		return S_CSC_OUT_OF_MEMORY;
	}

	return S_CSC_OK;
}

void gr_reg_sort (struct gr_region *reg)
{
	pdf_real_t x;

	if (reg -> lo > reg -> hi) {
		x = reg -> lo;
		reg -> lo = reg -> hi;
		reg -> hi = x;
	}

	if (reg -> xlo > reg -> xhi) {
		x = reg -> xlo;
		reg -> xlo = reg -> xhi;
		reg -> xhi = reg -> xlo;
	}
}

enum s_status gr_reg_add (struct gr_region **reg, 
			  struct gr_region *spec)
{
	struct gr_region *p, *q;
	pdf_real_t lo, hi, xlo, xhi;

	while (*reg != NULL && (*reg) -> hi < spec -> lo)
		reg = &(*reg) -> next;

	if (*reg == NULL || (*reg) -> lo > spec -> hi)
		return append_region (reg, spec);

	p = *reg;
	lo = min (p -> lo, spec -> lo);
	hi = max (p -> hi, spec -> hi);
	xlo = min (p -> xlo, spec -> xlo);
	xhi = max (p -> xhi, spec -> xhi);

	while (p -> next != NULL && p -> next -> lo <= hi) {
		q = p -> next;
		if (q -> hi > hi)
			hi = q -> hi;
		if (q -> xhi > xhi)
			xhi = q -> xhi;
		p -> next = q -> next;
		free (q);
	}

	p -> lo = lo;
	p -> hi = hi;
	p -> xlo = xlo;
	p -> xhi = xhi;

	return S_CSC_OK;
}

static enum s_status gr_reg_add_d (struct gr_region **reg, 
				   pdf_real_t lo, pdf_real_t hi, 
				   pdf_real_t svc, struct args_heur *heur)
{
	struct gr_region spec;

	spec.lo = lo;
	spec.hi = hi;

	heur_extend_cols (lo, hi, svc, &spec.xlo, &spec.xhi, heur);

	return gr_reg_add (reg, &spec);
}

pdf_bool_t gr_reg_has_x_gap (struct gr_region *reg)
{
	if (reg == NULL)
		return 0;

	while (reg -> next != NULL) {
		if (reg -> xhi < reg -> next -> xlo)
			return 1;
		reg = reg -> next;
	}

	return 0;
}

static void collapse_inner_gaps (struct gr_region **regp)
{
	struct gr_region *p;

	if ((*regp) == NULL)
		return;

	while ((*regp) -> next != NULL) {
		if ((*regp) -> xhi >= (*regp) -> next -> xlo) {
			p = (*regp) -> next;
			(*regp) -> hi = p -> hi;
			(*regp) -> xhi = p -> xhi;
			(*regp) -> next = p -> next;
			free (p);
		} else
			regp = &(*regp) -> next;
	}

}

enum s_status gr_reg_get (struct a_area *parent,
			  struct args_heur *heur,
			  struct gr_region **hregp, 
			  struct gr_region **vregp,
			  pdf_bool_t *hin, pdf_bool_t *vin)
{
	struct gr_region *hreg, *vreg;
	struct a_area *area;
	enum s_status res;

	area = parent -> child;
	if (area == NULL)
		return S_CSC_OK;

	res = S_CSC_OK;
	hreg = vreg = NULL;
	do {
		area = area -> next;
		
		res = gr_reg_add_d (&hreg, 
				    area -> g.bounds.left,
				    area -> g.bounds.right,
				    area -> g.svec.x, heur);
		if (res != S_CSC_OK)
			break;
		res = gr_reg_add_d (&vreg, 
				    area -> g.bounds.bottom,
				    area -> g.bounds.top,
				    area -> g.svec.y, heur);
		if (res != S_CSC_OK)
			break;
	} while (area != parent -> child);
	if (res != S_CSC_OK) {
		gr_reg_destroy (hreg);
		gr_reg_destroy (vreg);
		return res;
	}
	
	if (gr_reg_has_x_gap (hreg)) {
		collapse_inner_gaps (&hreg);
		*hin = 0;
	} else
		*hin = 1;

	if (gr_reg_has_x_gap (vreg)) {
		collapse_inner_gaps (&vreg);
		*vin = 0;
	} else
		*vin = 1;

	*hregp = hreg;
	*vregp = vreg;

	return S_CSC_OK;
}

