/* 
 * PDFlow -- squeeze.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */
/*
 * This file mirrors the contents of thread.c: for each grid type
 * that may be treated as an "atomic type", there's an implementation
 * that squeezes its contents. This is meant to provide support
 * to lc.c, that traverses the grid structures through 
 * gr_th_traverse and, for each atomic grid, calls its squeeze
 * function and attaches the results to the output agc.
 *
 * Currently we only support homogeneous squeezeing, i.e. we
 * just resize, making sure that the resulting grid does not
 * become too small. This leaves out some important cases:
 * - Tables: they appear as word areas (because of the graphic
 *           frame that makes them appear as a single area.
 *	     When badness becomes too high, we should "open"
 *	     them and treat them as grids;
 * - Boxes/columns: they should be reflowed in most cases, so
 *	     usage of box_squeeze et al. should be restricted to
 *	     inline formulas and similar "inline" stuff.
 *	     Deciding whether to resize or reflow, however should
 *	     be up to thread.c and does not regard this file.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <assert.h>
#include <string.h>

#include "grid/lc.h"
#include "grid/squeeze.h"
#include "grid/thread.h"

#define EPSILON 1e-5

static void combine (struct gr_grid *grid,
		     pdf_real_t *ratio, pdf_real_t *badness, 
		     pdf_real_t *bad_slope, struct cobj_rect *bounds,
		     pdf_real_t tratio, pdf_real_t tbadness,
		     pdf_real_t tbad_slope, struct cobj_rect *tbounds)
{
	if (*ratio < tratio) {
		*ratio = tratio;
		*badness = tbadness;
		*bad_slope = tbad_slope;
	}

	cobj_rect_union (bounds, tbounds);
}

static void get_homog_info (struct gr_grid *grid, pdf_real_t pspace, 
			    pdf_real_t *ratio, pdf_real_t *badness, 
			    pdf_real_t *bad_slope, struct cobj_rect *bounds, 
			    struct args_output *aout)
{
	struct gr_line *line;
	struct gr_cell *cell;
	struct cobj_rect tbounds;
	pdf_real_t tratio, tbadness, tslope;

	if (grid -> ifc -> logical != GR_L_WORD) {
		cobj_rect_set_null (&tbounds);
		cobj_rect_set_null (bounds);
		*ratio = EPSILON;
		*badness = *bad_slope = 0;
		for (line = grid -> d.lines; line != NULL; line = line -> next)
			for (cell = line -> cells; cell != NULL; 
			     cell = cell -> next) {
				/* Putting pspace on any inner grids should
				   be an acceptable simplification... I hope
				   -- Note that this mirrors area.c: does
				   not make much sense changing this and
				   not that --
				 */
				get_homog_info (cell -> grid, pspace, &tratio,
						&tbadness, &tslope, 
						&tbounds, aout);
				combine (cell -> grid,
					 ratio, badness, bad_slope, bounds,
					 tratio, tbadness, tslope, &tbounds);
			}
	} else {
		a_get_ratios (grid -> d.area, ratio, badness, bad_slope, aout);
		a_get_output_bounds (grid -> d.area, grid -> wdir, pspace,
				     aout, bounds);
	}
}

static enum s_status draw (struct gr_grid *grid, struct agc_context *agc,
			   struct gr_squeeze_result *sqr)
{
	struct gr_line *line;
	struct gr_cell *cell;
	enum s_status res;

	if (grid -> ifc -> logical != GR_L_WORD) {
		for (line = grid -> d.lines; line != NULL; line = line -> next)
			for (cell = line -> cells; cell != NULL; 
			     cell = cell -> next) {
				res = draw (cell -> grid, agc, sqr);
				if (res != S_CSC_OK)
					return res;
			}
	} else {
		res = a_add_statement (grid -> d.area, agc);
		if (res != S_CSC_OK)
			return res;
	}

	return S_CSC_OK;
}

void gr_sq_homog_probe (struct gr_grid *grid, pdf_real_t pspace,
			struct args_output *aout,
			int options, struct gr_squeeze_option opt [])
{
	pdf_real_t width, height, awidth, aheight, wratio, hratio, ratio;
	pdf_real_t nat_ratio, badness, bad_slope;
	struct gr_squeeze_result *sqr;
	int i;

	for (i = 0; i < options; i++) {

		sqr = &opt [i].sqr;

		get_homog_info (grid, pspace, &nat_ratio, &badness, 
				&bad_slope, &sqr -> bounds, aout);

		width = cobj_rect_width (&sqr -> bounds);
		height = cobj_rect_height (&sqr -> bounds);

		awidth = cobj_rect_width (&opt [i].avail);
		aheight = cobj_rect_height (&opt [i].avail);

		wratio = awidth / width;
		hratio = aheight / height;
		ratio = wratio < hratio ? wratio : hratio;

		if (nat_ratio < ratio) { 

			width = cobj_rect_width (&grid -> g.rect);
			height = cobj_rect_height (&grid -> g.rect);

			ratio = nat_ratio;

			sqr -> rect.left = 0;
			sqr -> rect.right = nat_ratio * width;
			sqr -> rect.bottom = 0;
			sqr -> rect.top = nat_ratio * height;
			sqr -> badness = 0;
		} else {
			sqr -> rect.left = 0;
			sqr -> rect.right = ratio * width;
			sqr -> rect.bottom = 0;
			sqr -> rect.top = ratio * height;

			sqr -> badness = badness + 
				(nat_ratio - ratio) * bad_slope;
		}

		sqr -> origin.x = (grid -> origin.x - 
				   grid -> g.rect.left) * ratio;
		sqr -> origin.y = (grid -> origin.y - 
				   grid -> g.rect.bottom) * ratio;

		sqr -> bounds.left -= grid -> g.rect.left;
		sqr -> bounds.right -= grid -> g.rect.left;
		sqr -> bounds.bottom -= grid -> g.rect.bottom;
		sqr -> bounds.top -= grid -> g.rect.bottom;

		sqr -> bounds.left *= ratio;
		sqr -> bounds.right *= ratio;
		sqr -> bounds.bottom *= ratio;
		sqr -> bounds.top *= ratio;
	}
}

enum s_status gr_sq_homog_squeeze (struct gr_grid *grid, 
				   struct agc_context *agc,
				   struct gr_squeeze_option *opt)
{
	enum s_status res;
	
	res = agc_translate_like (agc, &grid -> g.rect, &opt -> sqr.rect);
	if (res != S_CSC_OK)
		return res;

	res = draw (grid, agc, &opt -> sqr);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}
