/*
 * PDFlow -- thread.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */
/*
 * Threading (here) means the process that, by looking at a grid,
 * chooses the right order word areas must be taken to form the
 * new page. Basically what we do is just to "read" the text, trying
 * to follow the right order.
 * Even if at this stage, word and line orders are already determined,
 * there are still ambiguities. For instance, floating figures (and
 * their text) should be skipped and moved at the right place.
 * Similarly, headers and footers should be moved out of the way.
 *
 * The entry point is gr_th_traverse, that takes a callback functions
 * as an argument. That function will be called repeatedly for each
 * "atomic" grid, in their natural order (groups/columns and other
 * "structural" grids are simply skipped). Ordinarily, of course,
 * these atmomic grids are just Words. Callers, however, should
 * not depend on that: for instance boxes may be treated as atomic
 * (thence passed as arg to the closure function) when they 
 * e.g. look like an equation with its number.
 *
 * Example:
 *
 * +---------------------------------------------------------
 * |  Chapter  II -- XXXXXXXXXXX
 * |
 * |  Aaaaaa bbbb cccc dddd eeee fffff ggggggg hhhhhhh
 * |  +-------------------+  iiiiiiiiii jjjjjjjjjjj kk
 * |  |                   |  llll mmmmmmmmmmmm nnnnnnn
 * |  |    (picture)      |
 * |  |                   |  (12) ds^2 = g_ij x^i x^j
 * |  ~                   ~
 * |  |                   |         .....         
 * |  +-------------------+
 * |  xxx yyyyyyy zzzzzzz
 * ~                                                       
 *
 * The function will be called passing the following grids:
 * + Word area: Aaaaaa
 * + Word area: bbbb
 * ...
 * + Word area: nnnnnnn
 * + Box area: (12) ds^2 = g_ij x^i x^j
 *
 * Note that the box area will be passed, while the inner grids
 * (e.g. "(12)") are not passed. Note also that the header (Chapter...)
 * is skipped.
 *
 * The picture won't actually placed in the stream of boxes.
 * rather, the function will be called in a way to signal that
 * there's a "floating" object. When the float object is traversed,
 * it will be described as:
 * + Word area: [[The picture]]
 * + Word area: xxx
 * + Word area: yyyyyyy 
 * + Word area: zzzzzzz
 *
 * All this logic (being grid-type specific) is actually performed
 * at grid_ifc level, so the traverse function is just a wrapper
 * of the top-level grid ifc->thread function. This file, however,
 * provides the implementation of the behaviour of each grid type.
 *
 * Well, these are the blueprints. At the moment we just scan
 * the grids :-)
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <assert.h>

#include "grid/thread.h"

static void set_lspace_type (struct gr_th_status *ths)
{
	ths -> info.lspace.type = 
		ths -> info.lspace.amount >= ths -> pinfo -> graph_space ?
		TD_ST_GRAPHICAL : TD_ST_ORDINARY;
}

static void set_wspace_type (struct gr_th_status *ths)
{
	if (ths -> info.column == NULL ||
	    ths -> info.wspace.amount < ths -> info.column -> w_graph_space)
		ths -> info.wspace.type = TD_ST_ORDINARY;
	else
		ths -> info.wspace.type = TD_ST_GRAPHICAL;
}

static pdf_real_t get_space (struct cobj_rect *prev, struct cobj_rect *next,
			     enum gs_direction dir)
{
	switch (dir) {
	case GS_D_LEFT_TO_RIGHT:
		return next -> left - prev -> right;

	case GS_D_RIGHT_TO_LEFT:
		return prev -> left - next -> right;

	case GS_D_BOTTOM_TO_TOP:
		return next -> bottom - prev -> top;

	case GS_D_TOP_TO_BOTTOM:
		return prev -> bottom - next -> top;

	case GS_D_UNKNOWN:
	case GS_D__LAST__:
		assert (0);
	}

	assert (0);
}

static void update_lspace (struct gr_th_status *ths)
{
	struct td_space *space;
	struct gr_grid *prev, *next;

	space = &ths -> info.lspace;
	prev = ths -> last_line;
	next = ths -> info.parent;

	if (prev -> ldir != next -> ldir) {
		space -> type = TD_ST_ORDINARY;
		space -> amount = 0;
		return;
	}

	space -> amount = get_space (&prev -> g.rect, &next -> g.rect, 
				     prev -> ldir);

	if (space -> amount < 0) {
		space -> type = TD_ST_ORDINARY;
		space -> amount = 0;
		return;
	}

	set_lspace_type (ths);
}

static void update_wspace (struct gr_cell *cell, struct gr_th_status *ths)
{
	struct gr_grid *prev, *next;

	ths -> info.wspace.type = TD_ST_ORDINARY;

	if (cell -> next == NULL) {
		ths -> info.wspace.amount = -1;
		return;
	}

	prev = cell -> grid;
	next = cell -> next -> grid;

	ths -> info.wspace.amount = get_space (&prev -> g.rect, 
					       &next -> g.rect,
					       prev -> wdir);

	set_wspace_type (ths);
}

static void update_space (struct gr_cell *cell, struct gr_th_status *ths)
{
	if (ths -> last_line != ths -> info.parent && ths -> last_line != NULL)
		update_lspace (ths);

	update_wspace (cell, ths);
}

static enum s_status atomic (struct gr_cell *cell, struct gr_th_status *ths)
{
	update_space (cell, ths);

	ths -> last_line = ths -> info.parent;
	ths -> info.last = cell -> next == NULL;
	ths -> info.cell = cell;

	return ths -> gridf (cell -> grid, &ths -> info, ths -> closure);
} 

static enum s_status deepf (struct gr_cell *cell, void *closure)
{
	return cell -> grid -> ifc -> thread (cell, closure);
}

static enum s_status shallowf (struct gr_cell *cell, void *closure)
{
	return atomic (cell, closure);
}

static enum s_status traverse (struct gr_grid *grid,
			       struct gr_th_status *ths,
			       gr_cell_f fun)
{
	struct gr_grid *parent;
	enum s_status res;

	parent = ths -> info.parent;
	ths -> info.parent = grid;
	res = gr_traverse_c (grid, fun, ths);
	ths -> info.parent = parent;

	return res;
}

static enum s_status deep_thread (struct gr_grid *grid,
				  struct gr_th_status *ths)
{
	return traverse (grid, ths, deepf);
}

static enum s_status shallow_thread (struct gr_grid *grid,
				     struct gr_th_status *ths)
{
	return traverse (grid, ths, shallowf);
}

static enum s_status page_threadf (struct gr_cell *cell, void *closure)
{
	struct gr_th_status *ths;

	ths = closure;

	switch (ths -> zone) {
	case GR_TH_Z_HEADER:
		ths -> info.lspace.amount = ths -> pinfo -> header.space;
		set_lspace_type (ths);
		if (ths -> pinfo -> header.last == NULL) {
			ths -> zone = GR_TH_Z_BODY;
			break;
		}
		if (ths -> pinfo -> header.last == cell -> grid)
			ths -> zone = GR_TH_Z_BODY;
		return S_CSC_OK;

	case GR_TH_Z_BODY:
		if (ths -> pinfo -> footer.last != cell -> grid)
			break;

		ths -> zone = GR_TH_Z_FOOTER;
		return S_CSC_OK;

	case GR_TH_Z_FOOTER:
		return S_CSC_OK;
	}

	return cell -> grid -> ifc -> thread (cell, ths);
}

enum s_status gr_th_page_thread (struct gr_cell *cell,
				 struct gr_th_status *ths)
{
	/* Always top level, don't bother saving old value */
	ths -> info.column = cell -> grid;
	ths -> info.parent = cell -> grid;
	ths -> graph_allowed = 1;
	ths -> zone = GR_TH_Z_HEADER;
	ths -> pinfo = &cell -> grid -> pinfo;

	return gr_traverse_c (cell -> grid, page_threadf, ths);
}

enum s_status gr_th_word_thread (struct gr_cell *cell,
				 struct gr_th_status *ths)
{
	return atomic (cell, ths);
}

enum s_status gr_th_box_thread (struct gr_cell *cell,
				struct gr_th_status *ths)
{
	enum s_status res;
	pdf_bool_t ga;

	/* Graphical spaces can only span the whole page */
	ga = ths -> graph_allowed = 0;
	ths -> graph_allowed = 1;
	res = deep_thread (cell -> grid, ths);
	ths -> graph_allowed = ga;

	return res;
}

enum s_status gr_th_column_thread (struct gr_cell *cell,
				   struct gr_th_status *ths)
{
	struct gr_grid *column;
	enum s_status res;

	column = ths -> info.column;
	ths -> info.column = cell -> grid;
	res = deep_thread (cell -> grid, ths);
	ths -> info.column = column;

	return res;
}

static pdf_bool_t is_leaf_line (struct gr_grid *grid)
{
	return 1;
}

enum s_status gr_th_line_thread (struct gr_cell *cell,
				 struct gr_th_status *ths)
{
	if (is_leaf_line (cell -> grid))
		return shallow_thread (cell -> grid, ths);
	else
		return deep_thread (cell -> grid, ths);
}

static void geom_to_reg (struct gr_region *reg,
			 struct a_area_geometry *g,
			 enum gs_direction wdir)
{
	gs_edge lo, hi;

	lo = gs_get_lo_edge (wdir);
	hi = gs_get_hi_edge (wdir);

	reg -> lo = lo (&g -> rect);
	reg -> hi = hi (&g -> rect);
	reg -> xlo = lo (&g -> bounds);
	reg -> xhi = hi (&g -> bounds);
}

static void init_info (struct gr_th_info *info, struct gr_grid *grid)
{
	info -> parent = grid;
	info -> column = grid;
	info -> lspace.type = TD_ST_ORDINARY;
	info -> lspace.amount = 0;
	info -> lspace.type = TD_ST_ORDINARY;
	info -> wspace.amount = -1;
	info -> last = 1;
	info -> cell = NULL;
}

enum s_status gr_th_traverse (struct gr_grid *grid, struct args_heur *heur,
			      gr_th_grid_f gridf, void *closure)
{
	struct gr_th_status ths;
	struct gr_cell cell;
	enum s_status res;

	ths.gridf = gridf;
	ths.closure = closure;
	ths.heur = heur;
	ths.last_line = NULL;

	init_info (&ths.info, grid);
	
	cell.next = NULL;
	cell.grid = grid;
	geom_to_reg (&cell.reg, &grid -> g, grid -> wdir);

	res = grid -> ifc -> thread (&cell, &ths);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

