/*
 * PDFlow -- hash.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>
#include <stdlib.h>
#include <stdint.h>

#include "hash.h"

#ifdef HAVE_LIBSSL
#include <openssl/md5.h>

struct hash_status {
	MD5_CTX ctxt;
};

enum s_status hash_init (struct hash_status **hsp)
{
	struct hash_status *hs;

	hs = malloc (sizeof (struct hash_status));
	if (hs == NULL)
		return S_CSC_OUT_OF_MEMORY;

	if (!MD5_Init (&hs -> ctxt)) {
		free (hs);
		return S_CSC_OUT_OF_MEMORY;
	}

	*hsp = hs;

	return S_CSC_OK;
}

enum s_status hash_update (struct hash_status *hs, 
			   const void *data, unsigned len)
{
	struct hash drop;

	if (!MD5_Update (&hs -> ctxt, data, len)) {
		MD5_Final (drop.data, &hs -> ctxt);
		free (hs);
		return S_CSC_OUT_OF_MEMORY;
	}

	return S_CSC_OK;
}

enum s_status hash_final (struct hash_status *hs, struct hash *data)
{
	int res;

	res = MD5_Final (data -> data, &hs -> ctxt);
	free (hs);
	
	return res ? S_CSC_OK : S_CSC_OUT_OF_MEMORY;
}

#else

struct hash_status {
	int pos;
	uint8_t buf [HASH_SIZE];
};

enum s_status hash_init (struct hash_status **hsp)
{
	struct hash_status *hs;

	hs = malloc (sizeof (struct hash_status));
	if (hs == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (hs, 0, sizeof (struct hash_status));

	*hsp = hs;

	return S_CSC_OK;
}

enum s_status hash_update (struct hash_status *hs,
			   const void *data, unsigned len)
{
	const uint8_t *p;
	int i;

	p = data;
	for (i = 0; i < len; i++) {
		hs -> buf [hs -> pos] ^= p [i];
		hs -> pos = (hs -> pos + 1) % HASH_SIZE;
	}

	return S_CSC_OK;
}

enum s_status hash_final (struct hash_status *hs, struct hash *data)
{
	memcpy (data, hs -> buf, HASH_SIZE);
	free (hs);

	return S_CSC_OK;
}

#endif


