/*
 * PDFlow -- btree.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifndef btree_h

#define btree_h

#include <stdlib.h>
#include <stdarg.h>

#include "refs.h"
#include "status.h"

typedef void bt_tree_data_t;
typedef void bt_node_data_t;
typedef void bt_data_t;
typedef void bt_key_t;

enum bt_node_type {

	BT_NT_INTERMEDIATE,

	BT_NT_LEAF

};

struct bt_node {

	ref_t refs;

	enum bt_node_type type;

	struct bt_node *parent;

	bt_node_data_t *ndata;

	unsigned size, used;

	bt_key_t *key;

	union bt_ptr {
		bt_data_t *data;

		struct bt_node *child;
	} *p;
};

struct bt_param {

	size_t key_size;

	unsigned min_size;

	bt_tree_data_t *tdata;

	int (*compare) (bt_key_t *key1, bt_key_t *key2);

	void (*destroy_tree_data) (bt_tree_data_t *tdata);

	enum s_status (*create_node_data) (bt_tree_data_t *tdata, 
					   bt_node_data_t *parent,
					   bt_node_data_t **ndata);

	void (*destroy_node_data) (bt_tree_data_t *tdata, 
				   bt_node_data_t *ndata);

	void (*destroy_data) (bt_tree_data_t *tdata, bt_data_t *data);
};

struct bt_tree {

	ref_t refs;

	struct bt_node *root;

	struct bt_param param;

};

struct bt_imp_data {

	enum bt_node_type type;

	unsigned size;

	struct bt_node_data *ndata;

	bt_data_t **data;

	bt_key_t *key;
};

struct bt_exp_data {

	enum bt_node_type type;

	unsigned size;

	struct bt_node_data *ndata, *parent, **child;

	bt_key_t *lo, *hi;

	bt_data_t **data;
};

struct bt_cursor {

	ref_t refs;

	struct bt_tree *tree;

	struct bt_node *node;

	int idx;
};

typedef void (*bt_rekey_f) (bt_node_data_t *data, bt_key_t *key, 
			    int seqno, void *closure);

typedef enum s_status (*bt_traverse_f) (bt_node_data_t *ndata, 
					bt_data_t *data, 
					va_list ap);

struct bt_imp_info {
	enum s_status (*import) (bt_tree_data_t *tdata,
				 bt_node_data_t *parent,
				 unsigned idx,
				 struct bt_imp_data *imps,
				 void *closure);
};

struct bt_exp_info {
	enum s_status (*export) (bt_tree_data_t *tdata,
				 struct bt_exp_data *exps,
				 va_list ap);
};

enum s_status bt_init (struct bt_tree **tree, struct bt_param *param,
		       struct ref_owner *p);
void bt_ref (struct bt_tree *tree, struct ref_owner *p);
void bt_unref (struct bt_tree *tree, struct ref_owner *p);

enum s_status bt_import (struct bt_tree *tree, struct bt_imp_info *ii,
			 void *closure);
enum s_status bt_export (struct bt_tree *tree, struct bt_exp_info *ei, ...);

void bt_rekey (struct bt_tree *tree, bt_rekey_f rfun, void *closure);
void bt_rekey_from (struct bt_tree *tree, struct bt_cursor *pc,
		    bt_rekey_f rfun, void *closure);

enum s_status bt_traverse (struct bt_tree *tree, bt_traverse_f tfun, ...);

enum s_status bt_insert (struct bt_tree *tree, bt_key_t *key, 
			 bt_data_t *data, struct bt_cursor *pc);
enum s_status bt_get (struct bt_tree *tree, bt_key_t *key, bt_data_t **data);
bt_node_data_t *bt_get_node_data (struct bt_cursor *pc);
enum s_status bt_delete (struct bt_tree *tree, bt_key_t *key);
			 
enum s_status bt_cursor_init (struct bt_tree *tree, struct bt_cursor **pc,
			      struct ref_owner *p, int *eof);
int bt_cursor_goto (struct bt_cursor *pc, bt_key_t *key);
int bt_cursor_first (struct bt_cursor *pc);
int bt_cursor_last (struct bt_cursor *pc);
int bt_cursor_prev (struct bt_cursor *pc);
int bt_cursor_next (struct bt_cursor *pc);
void bt_cursor_ref (struct bt_cursor *pc, struct ref_owner *p);
void bt_cursor_unref (struct bt_cursor *pc, struct ref_owner *p);

bt_data_t *bt_cursor_get (struct bt_cursor *pc);
bt_key_t *bt_cursor_get_key (struct bt_cursor *pc);

#endif
