/*
 * PDFlow -- file.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifndef file_h

#define file_h

#include <stdint.h>
#include <stdarg.h>
#include <stdio.h>

#include "status.h"

#if defined(_WIN32) || defined(WIN32)
#include <windows.h>
#define F_EMULATE_MMAP
#endif

#define F_PRINTF(fmt, p) __attribute__ ((format (printf, fmt, p)))

typedef int pdf_bool_t;
typedef long pdf_integer_t, pdf_oid_t, pdf_gen_t, pdf_ofs_t;
typedef unsigned long pdf_unsigned_t;
typedef double pdf_real_t;

struct f_file {
	enum f_file_type {
		F_T_FILE,
		F_T_WRITER,
		F_T_REGION,
		F_T_BUFFER
	} type;

	char *fname;
	const uint8_t *pdf, *end;
	int refcount, unlink;
	union {
		FILE *fp;
		int fd;

#ifdef F_EMULATE_MMAP
		struct {
			HANDLE hfile;
			HANDLE hmap;
		} wm;
#endif

		struct {
			struct f_file *parent;
			void *data;
			void (*deallocator) (void *p);
		} reg;

		struct {
			const uint8_t *eob;
			int increment;
		} buffer;
	} c;
};

struct f_ptr {
	struct f_file *file;
	const uint8_t *cptr;
};

extern struct f_ptr f_null;

enum s_status f_open (const char *fname, struct f_file **pfile);
enum s_status f_creat (const char *fname, struct f_file **pfile);
enum s_status f_append (struct f_file *file, struct f_file **pfile);
enum s_status f_reg_open (struct f_file *parent, const uint8_t *begin, 
			  const uint8_t *end, void *data,
			  void (*deallocator) (void *p),
			  struct f_file **pfile);
enum s_status f_buffer_open (const char *name, struct f_file **pfile);
enum s_status f_write (struct f_file *file, 
		       const char *fmt, ...) F_PRINTF (2, 3);
enum s_status f_writeln (struct f_file *file, 
			 const char *fmt, ...) F_PRINTF (2, 3);
enum s_status f_newline (struct f_file *file);
enum s_status f_dump (struct f_file *dest, struct f_file *src);
enum s_status f_raw (struct f_file *dest, const void *data, size_t size);
enum s_status f_tell (struct f_file *file, pdf_ofs_t *pos);

void f_unlink_on_close (struct f_file *file);

#define F_EX(f, arg...) \
do { \
	res = f (arg); \
	if (res != S_CSC_OK) \
		return res; \
} while (0)
#define F_WRITE(arg...)   F_EX(f_write, arg)
#define F_WRITELN(arg...) F_EX(f_writeln, arg)
#define F_TELL(arg...)    F_EX(f_tell, arg)
#define F_NEWLINE(arg...) F_EX(f_newline, arg)

void f_ref (struct f_file *file);
void f_unref (struct f_file *file);

int f_ptr_init (struct f_ptr *fptr, struct f_file *file, long ofs);
void f_ptr_eof (struct f_ptr *fptr, struct f_file *file);
int f_ptr_valid (struct f_ptr *fptr);
int f_ptr_test (struct f_ptr *fptr, long s);
int f_ptr_seek (struct f_ptr *fptr, long s);
size_t f_ptr_avail (struct f_ptr *fptr);

enum s_status f_verror (const struct f_ptr *fptr, const char *fmt, va_list ap);
enum s_status f_error (const struct f_ptr *fptr, 
		       const char *fmt, ...) F_PRINTF (2, 3);

#endif
