/*
 * PDFlow -- obj.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifndef obj_h

#define obj_h

#include <stdint.h>

#include "file.h"

#define OBJ_STORE_SIZE 40

enum obj_type {

	OT_NULL,

	OT_BOOLEAN,
	OT_INTEGER,
	OT_REAL,
	OT_NAME,
	OT_STRING,
	OT_REGION,
	OT_ARRAY,
	OT_DICTIONARY,
	OT_REFERENCE,
	OT_XREF,
	OT_DEFINITION,
	OT_STREAM,
	OT_KEYWORD,
	OT_OBJ_PTR,

	OT_ANY,

	OT_ALIVE

};

struct obj_bstr {
	unsigned size;
	char *s;
};

struct obj_region;
struct obj_region_allocator {

	enum s_status (*dup) (const struct obj_region *sreg, 
			      struct obj_region *dreg);
	void (*deallocate) (struct obj_region *obj_region);

};

struct obj_region {
	const uint8_t *begin, *end;
	struct obj_region_allocator *allocator;
	void *ddata;
};

struct obj_dictionary {
	struct obj_dict_node *head;
};

struct obj_array {
	struct obj_array_node *head;
};

struct obj_reference {

	pdf_oid_t oid;
	pdf_gen_t gnum;

};

struct obj_xref {

	int valid;
	
	union {
		pdf_ofs_t next;
		pdf_ofs_t ofs;
	} u;

	pdf_gen_t gnum;
};

struct obj_definition {
	struct obj *obj;
};

struct obj_stream {
	struct obj_dictionary dictionary;
	struct obj_region region;
	struct obj_region decoded;
};

struct obj_alive;
struct obj_alive_allocator {
	enum s_status (*dup) (const struct obj_alive *sal, 
			      struct obj_alive *dal);
	void (*deallocate) (struct obj_alive *obj_alive);

};

struct obj_alive {

	void *ptr;

	struct obj_alive_allocator allocator;
};

struct obj_keyword {

	const char *name;

	void *data;
	
};

struct obj {

	enum obj_type type;

	struct obj_reference ref;

	struct f_ptr ptr;

	union {

		pdf_bool_t bool;
		pdf_integer_t integer;
		pdf_real_t real;
		struct obj_bstr string;
		struct obj_region region;
		struct obj_dictionary dictionary;
		struct obj_array array;
		struct obj_xref xref;
		struct obj_definition definition;
		struct obj_stream stream;
		struct obj_alive alive;
		struct obj_keyword *keyword;
		struct obj_keyword oth_keyword;
		struct obj_ptr *obj_ptr;

	} val;
};

struct obj_ptr {

	int refs;

	struct obj obj;

};

struct obj_dict_node {

	struct obj_dict_node *next;
	struct obj_bstr key;
	struct obj obj;

};

struct obj_array_node {

	struct obj_array_node *next;
	struct obj obj;

};

struct obj_store {

	struct obj_store *next;
		
	struct obj obj [OBJ_STORE_SIZE];
	
	int level;

};

extern const struct obj obj_null, obj_null_dictionary;
extern const struct obj obj_null_array, obj_null_name;

void obj_set_null (struct obj *obj, struct f_ptr *ptr);
void obj_set_boolean (struct obj *obj, struct f_ptr *ptr, pdf_bool_t val);
void obj_set_real (struct obj *obj, struct f_ptr *ptr, pdf_real_t val);
void obj_set_integer (struct obj *obj, struct f_ptr *ptr, pdf_integer_t val);
void obj_set_string (struct obj *obj, struct f_ptr *ptr, 
		     struct obj_bstr *string);
void obj_set_name (struct obj *obj, struct f_ptr *ptr, 
		   struct obj_bstr *string);
void obj_set_region (struct obj *obj, struct f_ptr *ptr, 
		     struct obj_region *region);
void obj_set_dictionary (struct obj *obj, struct f_ptr *ptr, 
			 const struct obj_dictionary *dict);
void obj_set_array (struct obj *obj, struct f_ptr *ptr, 
		    const struct obj_array *array);
void obj_set_reference (struct obj *obj, const struct f_ptr *ptr, 
			pdf_oid_t oid, pdf_gen_t gnum);
void obj_copy_reference (struct obj *obj, const struct obj *robj);
void obj_set_xref (struct obj *obj, struct f_ptr *ptr, 
		   pdf_ofs_t f1, pdf_gen_t gnum, int valid);
enum s_status obj_set_definition (struct obj *obj, struct f_ptr *ptr, 
				  struct obj *robj);
void obj_set_stream (struct obj *obj, struct f_ptr *ptr, 
		     struct obj_dictionary *dict, struct obj_region *region);
void obj_set_alive (struct obj *obj, struct f_ptr *ptr,
		    struct obj_alive_allocator *allocator, void *p);
void obj_set_keyword (struct obj *obj, struct f_ptr *ptr,
		      struct obj_keyword *keyword);
void obj_init_dictionary (struct obj_dictionary *dict);
void obj_init_dict_node (struct obj_dict_node *dnode,
			 struct obj *key, struct obj *obj);
void obj_dict_node_destroy (struct obj_dict_node **dn);
void obj_dict_destroy (struct obj_dictionary *dict);
struct obj_dict_node *obj_dict_find_node (struct obj_dictionary *dict,
					  struct obj_bstr *name);
enum s_status obj_dict_dup (const struct obj_dictionary *sdict,
			    struct obj_dictionary *ddict);
enum s_status obj_xfer_dict (struct obj_dictionary *dict, 
			     struct obj_dict_node *dnode);
enum s_status obj_xfer_t_dict (struct obj_dictionary *dict, 
			       struct obj_dict_node *dnode);
enum s_status obj_add_dict (struct obj_dictionary *dict,
			    const char *key, struct obj *val);
enum s_status obj_add_dict_s (struct obj_dictionary *dict,
			      struct obj *kobj, struct obj *val);
void obj_dict_merge (struct obj_dictionary *dest, struct obj_dictionary *src);

void obj_init_xref (struct obj_xref *ref, pdf_ofs_t f1, 
		    pdf_gen_t gnum, int valid);

void obj_init_array (struct obj_array *array);
enum s_status obj_xfer_array (struct obj_array *array, struct obj *obj);
enum s_status obj_xfer_t_array (struct obj_array *array, struct obj *obj,
				struct obj_array_node **anp);
void obj_array_destroy (struct obj_array *array);
enum s_status obj_array_dup (const struct obj_array *sarray,
			     struct obj_array *darray);
void obj_array_node_destroy (struct obj_array_node **an);

void obj_init_file_region (struct obj_region *region,
			   const void *begin, const void *end, 
			   struct f_file *file);
void obj_init_malloc_region (struct obj_region *region,
			     const void *begin, const void *end);
enum s_status obj_region_file (struct obj *reg, struct f_file **f);

void obj_region_destroy (struct obj_region *region);
enum s_status obj_region_dup (const const struct obj_region *sregion,
			      struct obj_region *dregion);
void obj_alive_destroy (struct obj_alive *alive);
enum s_status obj_alive_dup (const struct obj_alive *sal,
			     struct obj_alive *dal);

enum s_status obj_stream_file (struct obj *stream, struct f_file **f);
void obj_stream_destroy (struct obj_stream *stream);
void obj_definition_unwrap (struct obj_definition *def, struct obj *obj);

enum s_status obj_navigate (struct obj *obj, 
			    enum s_status (*f) (struct obj *o, void *data),
			    void *data);
void obj_unwrap (struct obj *obj, void *val);
void obj_wrap (void *val, enum obj_type ot, 
	       struct f_ptr *ptr, struct obj *obj);

int obj_region_equals (struct obj_region *reg1, struct obj_region *reg2);
int obj_array_equals (struct obj_array *array1, struct obj_array *array2);
int obj_dict_equals (struct obj_dictionary *dict1, 
		     struct obj_dictionary *dict2);
int obj_xref_equals (struct obj_xref *xref1, struct obj_xref *xref2);

int obj_equals (struct obj *obj1, struct obj *obj2);
int obj_same (const struct obj *obj1, const struct obj *obj2);
int obj_exists (const struct obj *obj);

struct obj_array *obj_get_rw_array (struct obj *obj);
struct obj_dictionary *obj_get_rw_dictionary (struct obj *obj);

const char *obj_tag (enum obj_type ot);
int obj_is_scalar (enum obj_type ot);

int obj_string_equal (const struct obj_bstr *str, const char *s);
int obj_string_equal_s (const struct obj_bstr *str1, 
			const struct obj_bstr *str2);
int obj_string_compare (const struct obj_bstr *str1, 
			const char *str2);
int obj_string_compare_s (const struct obj_bstr *str1, 
			  const struct obj_bstr *str2);
enum s_status obj_string_to_c (const struct obj_bstr *str, char **s);
enum s_status obj_string_from_c (const char *s, struct obj_bstr *str);
enum s_status obj_string_dup (const struct obj_bstr *src, 
			      struct obj_bstr *dest);
void obj_string_destroy (struct obj_bstr *str);

enum s_status obj_ptr_dup (struct obj *sobj, struct obj *dobj);
enum s_status obj_ptr_deref (struct obj *sobj, struct obj *dobj);
struct obj *obj_ptr_get (struct obj *ptr);

void obj_unload (struct obj *obj);
void obj_destroy (struct obj *obj);
enum s_status obj_dup1 (struct obj *obj);
enum s_status obj_dup (const struct obj *sobj, struct obj *dobj);

void obj_store_init (struct obj_store *lobj);
enum s_status obj_store_add (struct obj_store *lobj, struct obj *obj);
int obj_mark (struct obj_store *lobj);
enum s_status obj_rollback (struct obj_store *lobj, int p, enum s_status res);
enum s_status obj_cleanup (struct obj_store *lobj, enum s_status res);
void obj_commit (struct obj_store *lobj);

int obj_can_cast (const struct obj *obj, enum obj_type ot);
int obj_cast (struct obj *obj, enum obj_type ot);

enum s_status obj_verror (const struct obj *obj, const char *fmt, va_list ap);
enum s_status obj_error (const struct obj *obj, const char *fmt, ...);

#endif
