/*
 * PDFlow -- area.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifndef area_h

#define area_h

#include "args.h"
#include "types/cobj.h"
#include "graph/agc.h"
#include "graph/gspg.h"

struct a_area;

typedef int (*a_selector) (struct a_area *area, void *closure);

struct a_ifc_flags {
	int atomic : 1;
	int logic  : 1;
};

struct a_info {
	int hyphenized : 1;
	int fill : 1;

	struct a_area *hyphen;
};

struct a_ifc {
	const char *name;
	const char spec;
	struct {
		pdf_real_t r;
		pdf_real_t g;
		pdf_real_t b;
	} rgb;

	struct a_ifc_flags flags;

	enum s_status (*copy) (struct a_area *dest, struct a_area *src);
	
	enum s_status (*draw) (struct a_area *area,
			       struct agc_context *agc);
	
	void (*get_ratios) (struct a_area *area, pdf_real_t *ratio, 
			    pdf_real_t *badness, pdf_real_t *bad_slope,
			    struct args_output *aout);

	void (*get_direction) (struct a_area *area,
			       struct args_heur *heur,
			       enum gs_direction *pdir,
			       enum gs_direction *sdir);

	void (*get_origin) (struct a_area *area, 
			    enum gs_direction wdir,
			    struct args_heur *heur,
			    struct gs_point *origin,
			    pdf_bool_t *has_origin);

	void (*get_output_bounds) (struct a_area *area,
				   enum gs_direction wdir,
				   pdf_real_t pspace,
				   struct args_output *aout,
				   struct cobj_rect *bounds);

	void (*query_info) (struct a_area *area, enum gs_direction wdir,
			    struct a_info *info);

	void (*destroy) (struct a_area *a);
};

struct a_text_area {
	
	struct gs_glyph_info ginfo;

	enum gs_direction primary, secundary;

	struct gs_text_state *tstate;
	struct gs_graph_state *gstate;

	struct gs_paint paint;
};

struct a_graph_area {

	struct gs_subpath sp;

	struct gs_graph_state *gstate;

	struct gs_paint paint;
};

struct a_xo_area {

	struct gs_graph_state *gstate;

	struct xo_object *xo;
};

struct a_sh_area {

	struct gs_graph_state *gstate;

	struct shd_shading *sh;
};

union a_area_data {

	struct a_text_area text;

	struct a_graph_area graph;

	struct a_xo_area xo;

	struct a_sh_area sh;
};

struct a_area_geometry {
	struct cobj_rect bounds, rect;
	struct gs_point svec;
};

struct a_area {

	ref_t refs;

	struct a_area *next, *prev;
	struct a_area *child;

	struct a_ifc *ifc;

	agc_code_t code;

	struct a_area_geometry g;

	union a_area_data u;
};

enum s_status a__init (struct a_area **areap, struct ref_owner *p,
		       const char *fname, int line);
#define a_init(area, p) a__init (area, p, __FILE__, __LINE__)

enum s_status a__get_areas (struct a_area **bareap, struct gspg_page *page,
			    struct args_heur *heur, struct ref_owner *p, 
			    const char *fname, int line);
#define a_get_areas(area, page, heur, p) \
a__get_areas (area, page, heur, p, __FILE__, __LINE__)

enum s_status a__capture (struct a_area **areap, struct a_area *parent,
			  struct cobj_rect *rect, struct ref_owner *p, 
			  const char *fname, int line);
#define a_capture(area, parent, rect, p) \
a__capture (area, parent, rect, p, __FILE__, __LINE__)

enum s_status a_study_area (struct a_area **parent, struct ref_owner *p);

void a__ref (struct a_area *a, struct ref_owner *p, 
	     const char *fname, int line);
#define a_ref(area, p) a__ref (area, p, __FILE__, __LINE__)

void a_unref (struct a_area *a, struct ref_owner *p);

enum s_status a_dup (struct a_area *sa, struct a_area **da, 
		     struct ref_owner *p);

pdf_bool_t a_in_spec (struct a_ifc *ifc, const char *spec,
		      char group_spec);

struct cobj_rect *a_srect (struct a_area_geometry *geom, const char *spec);

enum s_status a_highlight (struct a_area *area, struct agc_context *agc,
			   const char *spec);

void a_select (struct a_area *donor, struct a_area *acceptor,
	       a_selector sel, void *closure);

enum s_status a_add_statement (struct a_area *area, struct agc_context *agc);

void a_get_direction (struct a_area *area,
		      struct args_heur *heur,
		      enum gs_direction *pdir,
		      enum gs_direction *sdir);

void a_get_origin (struct a_area *area, enum gs_direction wdir,
		   struct args_heur *heur, struct gs_point *origin, 
		   pdf_bool_t *has_origin);
void a_get_ratios (struct a_area *area, pdf_real_t *ratio,
		   pdf_real_t *badness, pdf_real_t *bad_slope,
		   struct args_output *aout);
void a_get_output_bounds (struct a_area *area, enum gs_direction wdir,
			  pdf_real_t pspace, struct args_output *aout, 
			  struct cobj_rect *bounds);
void a_query_info (struct a_area *area, enum gs_direction wdir,
		   struct a_info *info);

#endif
