/*
 * PDFlow -- status.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */
/*
 * Status codes. They are made of two distinct parts: a system code
 * and the error code itself. Though often general-purpose codes
 * are good enough (e.g. out of memory, internal error...), sometimes
 * subsystems may choose to define their own status codes. 
 */

#ifndef status_h

#define status_h

#define S_OFS 16		/* 64K possible codes */

#ifdef __GNUC__
#define S_PRINTF_LIKE(s, a) __attribute__ ((format (printf, s, a)))
#else
#define S_PRINTF_LIKE(s, a)
#endif

#define S_STATUS(sys, code) (((sys) << S_OFS) | (code))
#define S_SYS(status)	    ((status) >> S_OFS)
#define S_CODE(status)	    ((status) & ((1 << S_OFS) - 1))

#define S_ERRNO S_STATUS(S_SYS_OS, errno)

#if defined(_WIN32) || defined(WIN32)
#define S_WIN_ERROR S_STATUS(S_SYS_WIN32, GetLastError ())
#endif

enum s_system {

	S_SYS_COMMON,	/* Common codes */
	S_SYS_INDIRECT,	/* Indirect codes */
	S_SYS_OS,	/* Operating system (code==errno) */
#ifdef HAVE_LIBCURL
	S_SYS_CURL,	/* libcurl errors */
#endif

#ifdef HAVE_FREETYPE2
	S_SYS_FREETYPE,
#endif

#if defined(_WIN32) || defined(WIN32)
	S_SYS_WIN32,
#endif

	S_SYS__LAST__	/* Must be last */

};

enum s_code {

	S_CCC_OK,
	S_CCC_OUT_OF_MEMORY,			
	S_CCC_INTERNAL_ERROR,
	S_CCC_PARSE_ERROR,
	S_CCC_NOT_SUPPORTED,
	S_CCC_USER_CANCEL,
	
	S_CCC_REGISTRY_FULL,
	S_CCC_ENTRY_NOT_FOUND,
	S_CCC_ENTRY_ALREADY_PRESENT,
	S_CCC_BAD_ENTRY
};

enum s_status {	/* Common codes. Others are legal too, but not listed here */
	
	S_CSC_OK = S_STATUS (S_SYS_COMMON, S_CCC_OK),
	S_CSC_OUT_OF_MEMORY = S_STATUS (S_SYS_COMMON, S_CCC_OUT_OF_MEMORY),
	S_CSC_INTERNAL_ERROR = S_STATUS (S_SYS_COMMON, S_CCC_INTERNAL_ERROR),
	S_CSC_PARSE_ERROR = S_STATUS (S_SYS_COMMON, S_CCC_PARSE_ERROR),
	S_CSC_NOT_SUPPORTED = S_STATUS (S_SYS_COMMON, S_CCC_NOT_SUPPORTED),
	S_CSC_USER_CANCEL = S_STATUS (S_SYS_COMMON, S_CCC_USER_CANCEL),

	S_CSC_REGISTRY_FULL = S_STATUS (S_SYS_COMMON, S_CCC_REGISTRY_FULL),
	S_CSC_ENTRY_NOT_FOUND = S_STATUS (S_SYS_COMMON, S_CCC_ENTRY_NOT_FOUND),
	S_CSC_ENTRY_ALREADY_PRESENT = 
		S_STATUS (S_SYS_COMMON, S_CCC_ENTRY_ALREADY_PRESENT),
	S_CSC_BAD_ENTRY = S_STATUS (S_SYS_COMMON, S_CCC_BAD_ENTRY)
};

/* The structure that must be provided by each subsystem that can
 * generate a status code. Structures must be registered by using
 * s_register_system */
struct s_system_ifc {

	const char *name;

	enum s_system sys;

	const char *(*get_string) (enum s_code code);
};

struct pl_state;

/* Make indirect code. Indirect codes are used when in addition
 * to an already defined status code, a non-default string
 * should be displayed. For instance, when returning "parse error"
 * it's nice to tell what kind of failure happened. Indirect codes
 * are not permanent: they are reset each time a new indirect code 
 * is generated in the same thread (that's more or less like errno, so
 * it should be ok for most purposes) */
enum s_status s_make_indirect (enum s_status major, const char *format, ...)
     S_PRINTF_LIKE (2, 3);

/* Return string associated to a given status code */
const char *s_get_string (enum s_status status);

/* Get the real code. Normally it's equal to the original code, however
 * when using an indirect code it needs some lookup */
enum s_status s_get_code (enum s_status status);

/* Status registry */
enum s_status s_register (struct s_system_ifc *sys);
enum s_status s_unregister (struct s_system_ifc *sys);

#endif
