/*
 * PDFlow -- main.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <locale.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>

#include "app/conv.h"
#include "types/filespec.h"
#include "types/tree.h"
#include "pdf/pdfi.h"
#include "pdf/pdfc.h"
#include "pdfdoc.h"

enum s_status s_init (void);
void s_cleanup (void);

enum s_status ct_init (void);
void ct_cleanup (void);

enum s_status psi_init (void);
void psi_cleanup (void);

enum s_status gff_init (void);
void gff_cleanup (void);

enum s_status ps_init (void);
void ps_cleanup (void);

enum s_status fspec_init (void);
void fspec_cleanup (void);

enum s_status gsg_init (void);
void gsg_cleanup (void);

enum s_status gc_init (void);
void gc_cleanup (void);


static struct {
	enum s_status (*init) (void);
	void (*cleanup) (void);
} ctors [] = {
	{ s_init, s_cleanup },
	{ ct_init, ct_cleanup },
	{ gff_init, gff_cleanup },
	{ ps_init,  ps_cleanup },
	{ psi_init, psi_cleanup },
	{ fspec_init, fspec_cleanup },
	{ gsg_init, gsg_cleanup },
	{ gc_init, gc_cleanup },
	{ NULL }
};

static void set_locale (void)
{
	setlocale (LC_ALL, "C.UTF-8");
}

static void cleanup_i (int limit)
{
	while (limit--)
		ctors [limit].cleanup ();
}

static enum s_status init (void)
{
	enum s_status res;
	int i;

	for (i = 0; ctors [i].init != NULL; i++) {
		res = ctors [i].init ();
		if (res != S_CSC_OK) {
			cleanup_i (i);
			return res;
		}
	}

	return S_CSC_OK;
}

static void cleanup (void)
{
	int i;

	i = 0;
	while (ctors [i].init != NULL)
		i++;

	cleanup_i (i);
}

int main (int argc, char *argv [])
{
	struct args args;
	struct v_pref pref;
	enum s_status res;
	int i;

	set_locale ();

	res = init ();
	if (res != S_CSC_OK) {
 		fprintf (stderr, "%s\n", s_get_string (res));
		return 1;
	}

	memset (&pref, 0, sizeof (struct v_pref));
	res = args_init (&args, argc, argv);
	if (res != S_CSC_OK) {
		cleanup ();
		if (res != S_CSC_USER_CANCEL) {
			fprintf (stderr, "%s\n", s_get_string (res));
			return 1;
		} else
			return 0;
	}

	for (i = 0; i < args.files; i++) {
		if (args.debug.rfp != NULL)
			fprintf (args.debug.rfp, "# %s\n", args.file [i].src);

		res = conv_run (args.file [i].src, args.file [i].dest, &args);
		if (res != S_CSC_OK)	
			fprintf (stderr, "%s\n", s_get_string (res));
	}	

	args_destroy (&args);
	
	ref_shutdown ();

	cleanup ();

	return 0;
}
