/*
 * PDFlow -- pdfc.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <assert.h>
#include <string.h>

#include "graph/gstree.h"
#include "graph/grdtree.h"
#include "fmt/pdf-lex.h"
#include "pdf/pdfw.h"
#include "pdf/pdfc.h"

struct pdfc_cstate {

	struct ref_owner owner;

	struct t_tag lftag;

	struct gc_colour_state *lscst, *lncst;

};

struct pdfc_cpath {

	struct pdfc_cpath *next;

	struct gs_path *path;

	struct pdfc_cstate cstate;

};

struct pdfc_compiler {

	struct ref_owner owner;

	struct xref_db *rdb;

	struct xref_bind robj;

	struct v_selection *sel;

	struct pdfw_writer *pdfw;

	struct rd_resdict *rd;

	struct ob_bridge *ob;

	struct oc_properties *oc;

	struct gs_context *gc;

	struct gs_graph_state *cgs;

	struct gs_text_state *ctws;

	struct gst_tree *gst;

	struct grd_tree *ft;

	struct grd_tree *shdt;

	struct grd_tree *xot;

	struct grd_tree *patt;

	struct grd_tree *cspt;

	struct pdfc_cpath *cpath;

	struct pdfc_cstate cstate;
};

static void destroy_cstate (struct pdfc_cstate *ct)
{
	t_tag_destroy (&ct -> lftag);
	gc_cstate_unref (ct -> lscst, &ct -> owner);
	gc_cstate_unref (ct -> lncst, &ct -> owner);
	
	memset (ct, 0, sizeof (struct pdfc_cstate));
}

static enum s_status init_cstate (struct pdfc_cstate *dct,
				  struct pdfc_cstate *sct)
{
	enum s_status res;

	ref_init_owner (&dct -> owner, "cstate");

	if (sct != NULL) {
		res = t_tag_dup (&sct -> lftag, &dct -> lftag);
		if (res != S_CSC_OK)
			return res;

		dct -> lscst = sct -> lscst;
		gc_cstate_ref (dct -> lscst, &dct -> owner);

		dct -> lncst = sct -> lncst;
		gc_cstate_ref (dct -> lncst, &dct -> owner);
	}

	return S_CSC_OK;
}

static void replace_cstate (struct pdfc_cstate *dct,
			    struct pdfc_cstate *sct)
{
	destroy_cstate (dct);

	*dct = *sct;

	gc_cstate_ref (dct -> lscst, &dct -> owner);
	gc_cstate_ref (dct -> lncst, &dct -> owner);

	gc_cstate_unref (sct -> lscst, &sct -> owner);
	gc_cstate_unref (sct -> lncst, &sct -> owner);
}

static enum s_status cpath_push (struct pdfc_compiler *pdc,
				 struct gs_path *path)
{
	struct pdfc_cpath *cpath;
	enum s_status res;

	cpath = malloc (sizeof (struct pdfc_cpath));
	if (cpath == NULL)
		return S_CSC_OUT_OF_MEMORY;

	res = init_cstate (&cpath -> cstate, &pdc -> cstate);
	if (res != S_CSC_OK) {
		free (cpath);
		return res;
	}

	cpath -> path = path;
	cpath -> next = pdc -> cpath;
	pdc -> cpath = cpath;

	gs_path_ref (path, &pdc -> owner);

	return S_CSC_OK;
}

static int cpath_pop (struct pdfc_compiler *pdc)
{
	struct pdfc_cpath *cpath;
	struct gs_path *path;

	if (pdc -> cpath == NULL)
		return 0;

	cpath = pdc -> cpath;
	path = cpath -> path;

	destroy_cstate (&pdc -> cstate);

	replace_cstate (&pdc -> cstate, &cpath -> cstate);	
	
	pdc -> cpath = cpath -> next;
	free (cpath);

	gs_path_unref (path, &pdc -> owner);

	return 1;
}

static void cpath_empty (struct pdfc_compiler *pdc)
{
	while (pdc -> cpath != NULL)
		cpath_pop (pdc);
}

static enum s_status init (struct pdfc_compiler *pdc,
			   struct cat_catalog *cat,
			   struct ob_bridge *ob,
			   struct v_selection *sel,
			   struct ref_owner *p)
{
	enum s_status res;

	res = init_cstate (&pdc -> cstate, NULL);
	if (res != S_CSC_OK)
		return res;
	
	res = T_CALL (rd_new, &pdc -> rd, pdc -> rdb, p);
	if (res != S_CSC_OK)
		return res;

	res = gst_new (&pdc -> gst, pdc -> rdb, sel, p);
	if (res != S_CSC_OK)
		return res;

	res = grd_new (&pdc -> ft, "f", pdc -> rdb, 
		       ob, &gf_cache_closure, sel, p);
	if (res != S_CSC_OK)
		return res;

	res = grd_new (&pdc -> shdt, "shd", pdc -> rdb, ob, NULL, sel, p);
	if (res != S_CSC_OK)
		return res;

	res = grd_new (&pdc -> xot, "xo", pdc -> rdb, ob, NULL, sel, p);
	if (res != S_CSC_OK)
		return res;

	res = grd_new (&pdc -> patt, "pt", pdc -> rdb, ob, NULL, sel, p);
	if (res != S_CSC_OK)
		return res;

	res = grd_new (&pdc -> cspt, "cs", pdc -> rdb, ob, NULL, sel, p);
	if (res != S_CSC_OK)
		return res;

	pdc -> oc = cat -> oc_properties;
	pdc -> sel = sel;
	pdc -> ob = ob;

	xref_bind_init (&pdc -> robj);
	xref_bind_data (&pdc -> robj, rd_bkey, pdc -> rd);
	xref_bind_data (&pdc -> robj, oc_properties_bkey, pdc -> oc);

	
	return S_CSC_OK;
}

enum s_status pdfc_create (struct cat_catalog *cat,
			   struct ob_bridge *ob,
			   struct v_selection *sel,
			   struct pdfc_compiler **pdcp)
{
	struct pdfc_compiler *pdc;
	enum s_status res;

	pdc = malloc (sizeof (struct pdfc_compiler));
	if (pdc == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (pdc, 0, sizeof (struct pdfc_compiler));

	ref_init_owner (&pdc -> owner, "compiler");

	pdc -> rdb = cat -> t.rdb;

	res = init (pdc, cat, ob, sel, &pdc -> owner);
	if (res != S_CSC_OK) {
		pdfc_destroy (pdc);
		return res;
	}

	*pdcp = pdc;

	return S_CSC_OK;
}

void pdfc_destroy (struct pdfc_compiler *pdc)
{
	cpath_empty (pdc);
	destroy_cstate (&pdc -> cstate);
	rd_unref (pdc -> rd, &pdc -> owner);
	gst_unref (pdc -> gst, &pdc -> owner);
	grd_unref (pdc -> ft, &pdc -> owner);
	grd_unref (pdc -> shdt, &pdc -> owner);
	grd_unref (pdc -> xot, &pdc -> owner);
	grd_unref (pdc -> patt, &pdc -> owner);
	grd_unref (pdc -> cspt, &pdc -> owner);
	gs_context_unref (pdc -> gc, &pdc -> owner);
	gs_graph_unref (pdc -> cgs, &pdc -> owner);
	gs_text_unref (pdc -> ctws, &pdc -> owner);
	pdfw_destroy (pdc -> pdfw);

	free (pdc);
}

static enum s_status serialize_dash_pattern (struct pdfc_compiler *pdc,
					     struct gs_graph_state *gs,
					     struct gs_graph_state *tgs)
{
	struct gs_dash_pattern *dp, *tdp, *ndp;
	struct ref_owner *p;
	enum s_status res;
	struct obj obj;

	dp = gs -> dash_pattern;
	tdp = tgs -> dash_pattern;
	p = gs_graph_owner (pdc -> gc);

	if (tdp == NULL) {
		res = pdfw_append (pdc -> pdfw,
				   PDFW_OP_DASH_PATTERN, 
				   &obj_null_array, (pdf_integer_t) 0);
		if (res != S_CSC_OK)
			return res;

		gs_dash_pattern_unref (dp, p);
		gs -> dash_pattern = NULL;

		return S_CSC_OK;
	}

	res = gs_dash_pattern_sync (pdc -> rdb, NULL, pdc -> sel,
				    &obj, NULL, tdp);
	if (res != S_CSC_OK)
		return res;

	res = T_CALL (gs_dash_pattern_init, &ndp, pdc -> rdb, 
		      NULL, &obj, p);
	obj_destroy (&obj);
	if (res != S_CSC_OK)
		return res;

	res = xref_array_unmap_ps (ndp -> t.rdb, NULL, pdc -> sel, 
				   &ndp -> dash_array, &obj, 0, 
				   &xref_real_pstable, NULL);
	if (res != S_CSC_OK) {
		gs_dash_pattern_unref (ndp, p);
		return res;
	}

	res = pdfw_append (pdc -> pdfw, PDFW_OP_DASH_PATTERN, 
			   &obj, tdp -> dash_phase);
	obj_destroy (&obj);
	if (res != S_CSC_OK) {
		gs_dash_pattern_unref (ndp, p);
		return res;
	}

	gs_dash_pattern_unref (dp, p);
	
	gs -> dash_pattern = ndp;

	return S_CSC_OK;
}

static enum s_status serialize_rendering_intent (struct pdfc_compiler *pdc,
						 struct gs_graph_state *gs,
						 struct gs_graph_state *tgs)
{
	struct fun_enum *ri, *tri;
	struct xref_field sfd;
	enum s_status res;
	struct obj obj;

	ri = &gs -> rendering_intent;
	tri = &tgs -> rendering_intent;

	xref_init_field (&sfd, tri, gs_rendering_intent_desc);
	res = cobj_enum_pstable.serializer (pdc -> rdb, NULL, pdc -> sel,
					    &obj, NULL, &sfd);
	if (res != S_CSC_OK)
		return res;

	res = pdfw_append (pdc -> pdfw,
			   PDFW_OP_RENDERING_INTENT, &obj);
	obj_destroy (&obj);
	if (res != S_CSC_OK)
		return res;

	gs -> rendering_intent = *tri;

	return S_CSC_OK;
}

static enum s_status align_gop (struct pdfc_compiler *pdc,
				struct gs_graph_state *tgs)
{
	struct gs_graph_state *gs;
	enum s_status res;

	res = gs_graph (pdc -> gc, GS_GO_OP, &gs);
	if (res != S_CSC_OK)
		return res;

	if (obj_string_equal_s (&gs -> key.opkey, &tgs -> key.opkey))
		return S_CSC_OK;

	if (memcmp (&gs -> ctm, &tgs -> ctm, sizeof (gs -> ctm))) {
		res = pdfw_append (pdc -> pdfw, 
				   PDFW_OP_SET_CTM, &tgs -> ctm);
		gs -> ctm = tgs -> ctm;
	}

	if (gs -> line_width != tgs -> line_width) {
		res = pdfw_append (pdc -> pdfw, 
				   PDFW_OP_LINE_WIDTH, tgs -> line_width);
		if (res != S_CSC_OK)
			return res;

		gs -> line_width = tgs -> line_width;
	}

	if (gs -> line_cap != tgs -> line_cap) {
		res = pdfw_append (pdc -> pdfw, 
				   PDFW_OP_LINE_CAP, tgs -> line_cap);
		if (res != S_CSC_OK)
			return res;

		gs -> line_cap = tgs -> line_cap;
	}

	if (gs -> line_join != tgs -> line_join) {
		res = pdfw_append (pdc -> pdfw, 
				   PDFW_OP_LINE_JOIN, tgs -> line_join);
		if (res != S_CSC_OK)
			return res;

		gs -> line_join = tgs -> line_join;
	}

	if (gs -> miter_limit != tgs -> miter_limit) {
		res = pdfw_append (pdc -> pdfw, 
				   PDFW_OP_MITER_LIMIT, tgs -> miter_limit);
		if (res != S_CSC_OK)
			return res;

		gs -> miter_limit = tgs -> miter_limit;
	}

	if (!gs_dash_pattern_eq (gs -> dash_pattern, 
				 tgs -> dash_pattern)) {
		res = serialize_dash_pattern (pdc, gs, tgs);
		if (res != S_CSC_OK)
			return res;
	}

	if (!gs_rendering_intent_eq (&gs -> rendering_intent, 
				     &tgs -> rendering_intent)) {
		res = serialize_rendering_intent (pdc, gs, tgs);
		if (res != S_CSC_OK)
			return res;
	}

	return gs_graph_update_key (gs, GS_GO_OP);
}

static enum s_status generate_colour (struct pdfc_compiler *pdc,
				      struct pt_pattern *tpat,
				      pdf_real_t *val, int n,
				      struct xref_obj_array *ma)

{
	struct obj obj;
	enum s_status res;
	int i;

	ma -> n = n;
	if (tpat != NULL)
		ma -> n++;

	if (ma -> n) {
		ma -> data = calloc (ma -> n, sizeof (struct obj));
		if (ma -> data == NULL)
			return S_CSC_OUT_OF_MEMORY;

		for (i = 0; i < n; i++)
			obj_set_real (&ma -> data [i], &f_null, val [i]);
	} else
		ma -> data = NULL;

	if (tpat != NULL) {
		res = pt_sync (NULL, &pdc -> robj, pdc -> sel, 
			       &obj, NULL, tpat);
		if (res != S_CSC_OK) {
			if (ma -> data != NULL)
				free (ma -> data);
			return res;
		}

		res = grd_lookup (pdc -> patt, &tpat -> t.tag, 
				  &obj, tpat -> t.rdb, &ma -> data [n], 
				  NULL, NULL);
		obj_destroy (&obj);
		if (res != S_CSC_OK) {
			if (ma -> data != NULL)
				free (ma -> data);
			return res;
		}
	}

	return S_CSC_OK;
}

static enum s_status align_colour_space (struct pdfc_compiler *pdc,
					 struct gc_colour_state **cst,
					 struct gc_colour_state *tcst,
					 struct xref_obj_array *ma,
					 pdf_bool_t stroke)
{
	struct obj obj, name, csobj;
	enum s_status res;

	switch (gc_get_family (tcst -> sp)) {
	case GC_CSF_DEVICE_GRAY:
		res = pdfw_append (pdc -> pdfw, PDFW_OP_SET_GRAY,
				   stroke, ma);
		if (res != S_CSC_OK)
			break;

		res = gs_set_device_colour (pdc -> gc, pdc -> rd, cst, 
					    gc_get_family (tcst -> sp));
		break;
				   
	case GC_CSF_DEVICE_RGB:
		res = pdfw_append (pdc -> pdfw, PDFW_OP_SET_RGB,
				   stroke, ma);
		if (res != S_CSC_OK)
			break;

		res = gs_set_device_colour (pdc -> gc, pdc -> rd, cst, 
					    gc_get_family (tcst -> sp));
		break;

	case GC_CSF_DEVICE_CMYK:
		res = pdfw_append (pdc -> pdfw, PDFW_OP_SET_CMYK,
				   stroke, ma);
		if (res != S_CSC_OK)
			break;

		res = gs_set_device_colour (pdc -> gc, pdc -> rd, cst, 
					    gc_get_family (tcst -> sp));
		break;

	default:
		if (!v_supported (pdc -> sel -> pref, 1, 1)) {
			res = s_make_indirect (S_CSC_PARSE_ERROR, 
					       "Need PDF 1.1 to use this "
					       "colour space");
			break;
		}
			
		v_upgrade (pdc -> sel, 1, 1);
		res = gc_cspace_sync (NULL, &pdc -> robj, pdc -> sel,
				      &obj, NULL, tcst -> sp);
		if (res != S_CSC_OK)
			break;

		res = grd_lookup (pdc -> cspt, &tcst -> sp -> t.tag,
				  &obj, tcst -> sp -> t.rdb, 
				  &name, &csobj, NULL);
		obj_destroy (&obj);
		if (res != S_CSC_OK)
			break;

		res = pdfw_append (pdc -> pdfw, PDFW_OP_SET_CSPACE,
				   stroke, &name);
		obj_destroy (&name);
		if (res != S_CSC_OK) {
			obj_destroy (&csobj);
			break;
		}

		res = gc_set_colour_state (&csobj, pdc -> rdb, pdc -> rd,
					   cst, gs_graph_owner (pdc -> gc));
		obj_destroy (&csobj);
		if (res != S_CSC_OK)
			break;

		res = pdfw_append (pdc -> pdfw, PDFW_OP_SET_COLOUR,
				   stroke, ma);
		break;
	}
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status align_colour_components (struct pdfc_compiler *pdc,
					      struct gc_colour_state **cst,
					      struct gc_colour_state *tcst,
					      struct xref_obj_array *ma,
					      pdf_bool_t stroke)
{
	if (!v_supported (pdc -> sel -> pref, 1, 1))
		return align_colour_space (pdc, cst, tcst, ma, stroke);

	v_upgrade (pdc -> sel, 1, 1);
	return pdfw_append (pdc -> pdfw, PDFW_OP_SET_COLOUR, stroke, ma);
}

static enum s_status do_align_colour_state (struct pdfc_compiler *pdc,
					    struct gc_colour_state **cst,
					    struct gc_colour_state *tcst,
					    struct pt_pattern *tpat,
					    pdf_real_t *val, int n,
					    pdf_bool_t stroke,
					    pdf_bool_t update_cspace)
{
	struct xref_obj_array ma;
	enum s_status res;

	if (tpat != NULL) {
		if (!v_supported (pdc -> sel -> pref, 1, 2))
			return s_make_indirect (S_CSC_PARSE_ERROR, 
					       "Need PDF 1.2 to use "
					       "patterned spaces");
		v_upgrade (pdc -> sel, 1, 2);
	}

	res = generate_colour (pdc, tpat, val, n, &ma);
	if (res != S_CSC_OK)
		return res;

	if (update_cspace)
		res = align_colour_space (pdc, cst, tcst, &ma, stroke);
	else
		res = align_colour_components (pdc, cst, tcst, &ma, stroke);

	xref_array_map_rel_ps (&ma, &pdc -> owner, &xref_obj_pstable, NULL);
	if (res != S_CSC_OK)
		return res;

	res = gc_select_colour (cst, tpat, val, gs_graph_owner (pdc -> gc));
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status align_colour_state (struct pdfc_compiler *pdc,
					 struct gc_colour_state **cst,
					 struct gc_colour_state *tcst,
					 struct gc_colour_state **lcst,
					 pdf_bool_t stroke)
{
	struct pt_pattern *tpat;
	pdf_bool_t cspace, cstate;
	pdf_real_t *val;
	enum s_status res;
	int n;

	if (*lcst == tcst || tcst == NULL)
		return S_CSC_OK;

	if (*lcst != NULL) {
		cspace = gc_cspace_equal ((*lcst) -> sp, tcst -> sp);
		cstate = gc_cstate_equal (*lcst, tcst);
	} else
		cspace = cstate = 0;

	gc_cstate_unref (*lcst, &pdc -> cstate.owner);
	*lcst = tcst;
	gc_cstate_ref (*lcst, &pdc -> cstate.owner);

	if (cspace && cstate)
		return S_CSC_OK;

	res = gc_get_colour (tcst, &tpat, &val, &n, &pdc -> owner);
	if (res != S_CSC_OK)
		return res;

	res = do_align_colour_state (pdc, cst, tcst, tpat, 
				     val, n, stroke, !cspace);
	if (val != NULL)
		free (val);
	pt_unref (tpat, &pdc -> owner);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}
					 
static enum s_status align_colour (struct pdfc_compiler *pdc,
				   struct gs_graph_state *tgs,
				   struct gs_paint *paint)
{
	struct gs_graph_state *gs;
	enum s_status res;

	res = gs_graph (pdc -> gc, GS_GO_COLOUR, &gs);
	if (res != S_CSC_OK)
		return res;

	gs = pdc -> gc -> state -> gstate;

	if (paint != NULL && paint -> stroke) {
		res = align_colour_state (pdc, &gs -> stroking_colour_state,
					  tgs -> stroking_colour_state, 
					  &pdc -> cstate.lscst, 1);
		if (res != S_CSC_OK)
			return res;
	}

	if (paint != NULL && paint -> fill) {
		res = align_colour_state (pdc, &gs -> nonstroking_colour_state,
					  tgs -> nonstroking_colour_state, 
					  &pdc -> cstate.lncst, 0);
		if (res != S_CSC_OK)
			return res;
	}

	return S_CSC_OK;
}

static enum s_status serialize_font (struct pdfc_compiler *pdc,
				     struct gs_font **fontp,
				     struct gs_font *tfont)
{
	struct obj obj, name, xobj;
	struct gs_font *nfont;
	struct gf_font *ffont;
	struct t_type **cfont;
	enum s_status res;

	res = gf_sync (NULL, &pdc -> robj, pdc -> sel, 
		       &obj, NULL, tfont -> font);
	if (res != S_CSC_OK)
		return res;

	res = grd_lookup (pdc -> ft, &tfont -> font -> t.tag, &obj,
			  tfont -> t.rdb, &name, &xobj, &cfont);
	obj_destroy (&obj);
	if (res != S_CSC_OK)
		return res;

	res = pdfw_append (pdc -> pdfw, 
			   PDFW_OP_FONT, &name, tfont -> size);
	obj_destroy (&name);
	if (res != S_CSC_OK) {
		obj_destroy (&xobj);
		return res;
	}

	res = T_CALL (gs_font_new, &nfont, pdc -> rdb,
		      gs_text_owner (pdc -> gc));
	if (res != S_CSC_OK) {
		obj_destroy (&xobj);
		return res;
	}

	if (*cfont == NULL) {
		res = T_CALL (gf_init, &ffont, pdc -> rdb, &pdc -> robj,
			      &xobj, ref_rown (&nfont -> t.refs));
		obj_destroy (&xobj);
		if (res != S_CSC_OK) {
			gs_font_unref (nfont, gs_text_owner (pdc -> gc));
			return res;
		}
		*cfont = &ffont -> t;
		T_CALL (gf_ref, ffont, ref_rown (&pdc -> ft -> refs));
	} else {
		obj_destroy (&xobj);
		ffont = *(struct gf_font **) cfont;
		T_CALL (gf_ref, ffont, ref_rown (&nfont -> t.refs));
	}

	gs_font_unref (*fontp, gs_text_owner (pdc -> gc));

	nfont -> font = ffont;
	nfont -> size = tfont -> size;
	
	*fontp = nfont;

	return S_CSC_OK;
}

static enum s_status serialize_mat (struct pdfc_compiler *pdc,
				    struct gs_text_state *tws,
				    struct gs_text_state *ttws)
{
	struct cobj_matrix *tm, *ttm;
	struct gs_mat *mat, *tmat;
	enum s_status res;

	mat = &tws -> mat;
	tmat = &ttws -> mat;
	
	tm = &mat -> t_m;
	ttm = &tmat -> t_m;
	
	res = pdfw_append (pdc -> pdfw, PDFW_OP_TEXT_MATRIX,
			   ttws -> leading, ttm);
	if (res != S_CSC_OK)
		return res;

	mat -> t_m = mat -> t_lm = tmat -> t_m;
	tws -> leading = ttws -> leading;

	return S_CSC_OK;
}

static enum s_status align_top (struct pdfc_compiler *pdc,
				struct gs_text_state *ttws)
{
	struct gs_text_state *tws;
	enum s_status res;

	res = gs_text (pdc -> gc, GS_GO_OP, &tws);
	if (res != S_CSC_OK)
		return res;

	/* t_lm is not protected by opkey */
	if (memcmp (&ttws -> mat.t_m, &tws -> mat.t_m,
	    sizeof (struct cobj_matrix))) {
		res = serialize_mat (pdc, tws, ttws);
		if (res != S_CSC_OK)
			return res;
	}

	if (obj_string_equal_s (&ttws -> key.opkey, &tws -> key.opkey))
		return S_CSC_OK;

	if (tws -> char_space != ttws -> char_space) {
		res = pdfw_append (pdc -> pdfw,
				   PDFW_OP_CHAR_SPACE, ttws -> char_space);
		if (res != S_CSC_OK)
			return res;

		tws -> char_space = ttws -> char_space;
	}

	if (tws -> word_space != ttws -> word_space) {
		res = pdfw_append (pdc -> pdfw,
				   PDFW_OP_WORD_SPACE, ttws -> word_space);
		if (res != S_CSC_OK)
			return res;

		tws -> word_space = ttws -> word_space;
	}

	if (tws -> scale != ttws -> scale) {
		res = pdfw_append (pdc -> pdfw,
				   PDFW_OP_SCALE, ttws -> scale);
		if (res != S_CSC_OK)
			return res;

		tws -> scale = ttws -> scale;
	}

	if (tws -> leading != ttws -> leading) {
		res = pdfw_append (pdc -> pdfw,
				   PDFW_OP_SCALE, ttws -> leading);
		if (res != S_CSC_OK)
			return res;

		tws -> leading = ttws -> leading;
	}

	if (tws -> rise != ttws -> rise) {
		res = pdfw_append (pdc -> pdfw,
				   PDFW_OP_RISE, ttws -> rise);
		if (res != S_CSC_OK)
			return res;

		tws -> rise = ttws -> rise;
	}

	if (ttws -> font != NULL && 
	    (tws -> font == NULL ||
	     !t_tag_same (&pdc -> cstate.lftag, 
			  &ttws -> font -> font -> t.tag) ||
	     tws -> font -> size != ttws -> font -> size)) {
		res = serialize_font (pdc, &tws -> font, ttws -> font);
		if (res != S_CSC_OK)
			return res;

		t_tag_destroy (&pdc -> cstate.lftag);

		res = t_tag_dup (&ttws -> font -> font -> t.tag, 
				 &pdc -> cstate.lftag);
		if (res != S_CSC_OK)
			return res;
	}

	return gs_text_update_key (tws, GS_GO_OP);
}

static enum s_status align_dict (struct pdfc_compiler *pdc,
				 struct gs_graph_state *tgs,
				 struct gs_text_state *ttws)
{
	struct gs_graph_state *gs, *ltgs;
	struct gs_text_state *tws, *lttws;
	struct gs_tags tags;
	struct obj sobj;
	enum s_status res;
	struct obj name;

	res = gs_graph (pdc -> gc, GS_GO_DICT, &gs);
	if (res != S_CSC_OK)
		return res;

	res = gs_graph_update_key (gs, GS_GO_DICT);
	if (res != S_CSC_OK)
		return res;

	res = gs_text (pdc -> gc, GS_GO_DICT, &tws);
	if (res != S_CSC_OK)
		return res;

	res = gs_text_update_key (tws, GS_GO_DICT);
	if (res != S_CSC_OK)
		return res;

	if (!gs_state_dkey_compare (&gs -> key, &tgs -> key) &&
	    !gs_state_dkey_compare (&tws -> key, &ttws -> key))
		return S_CSC_OK;

	res = T_CALL (gs_graph_cross, &ltgs, tgs, NULL, 
		      pdc -> ob, &pdc -> owner);
	if (res != S_CSC_OK)
		return res;
	res = T_CALL (gs_text_cross, &lttws, ttws, NULL, 
		      pdc -> ob, &pdc -> owner);
	if (res != S_CSC_OK) {
		gs_graph_unref (ltgs, &pdc -> owner);
		return res;
	}

	res = gs_graph_update_key (ltgs, GS_GO_DICT);
	if (res == S_CSC_OK)
		res = gs_text_update_key (lttws, GS_GO_DICT);
	if (res == S_CSC_OK)
		res = gst_lookup (pdc -> gst, ltgs, lttws, 
				  &name, &sobj, &tags);
	gs_graph_unref (ltgs, &pdc -> owner);
	gs_text_unref (lttws, &pdc -> owner);
	if (res != S_CSC_OK)
		return res;

	res = pdfw_append (pdc -> pdfw, PDFW_OP_GS, &name);
	obj_destroy (&name);
	if (res != S_CSC_OK) {
		gs_tags_destroy (&tags);
		obj_destroy (&sobj);
		return res;
	}

	res = gs_context_set_dictionary (pdc -> gc, &sobj, &tags);
	obj_destroy (&sobj);
	if (res != S_CSC_OK) {
		gs_tags_destroy (&tags);
		return res;
	}

	gs = pdc -> gc -> state -> gstate;
	tws = pdc -> gc -> state -> tstate;

	res = gs_graph_update_key (gs, GS_GO_DICT);
	if (res != S_CSC_OK)
		return res;

	res = gs_text_update_key (tws, GS_GO_DICT);
	if (res != S_CSC_OK)
		return res;

#ifdef DEBUG
	assert (!gs_state_dkey_compare (&gs -> key, &tgs -> key) &&
		!gs_state_dkey_compare (&tws -> key, &ttws -> key));
#endif

	return S_CSC_OK;
}

static enum s_status align_gstate (struct pdfc_compiler *pdc,
				   struct gs_graph_state *gs,
				   struct gs_paint *paint)
{
	enum s_status res;

	gs_graph_check_key (gs);
	if (pdc -> cgs != gs) {
		res = align_gop (pdc, gs);
		if (res != S_CSC_OK)
			return res;

		res = align_colour (pdc, gs, paint);
		if (res != S_CSC_OK)
			return res;

		gs_graph_unref (pdc -> cgs, &pdc -> owner);
		pdc -> cgs = NULL;
		if (res != S_CSC_OK)
			return res;

		pdc -> cgs = gs;
		T_CALL (gs_graph_ref, pdc -> cgs, &pdc -> owner);
	}

	return S_CSC_OK;
}

static enum s_status align_tstate (struct pdfc_compiler *pdc,
				   struct gs_text_state *ttws)
{
	enum s_status res;

	gs_text_check_key (ttws);

	/* t_m is the only element that is allowed to change */
	if (pdc -> ctws != ttws ||
	    memcmp (&ttws -> mat.t_m, &gs_ro_text (pdc -> gc) -> mat.t_m,
	    sizeof (struct cobj_matrix))) {
		res = align_top (pdc, ttws);
		if (res != S_CSC_OK)
			return res;

		gs_text_unref (pdc -> ctws, &pdc -> owner);
		pdc -> ctws = NULL;
		if (res != S_CSC_OK)
			return res;

		pdc -> ctws = ttws;
		T_CALL (gs_text_ref, pdc -> ctws, &pdc -> owner);
	} 

	return S_CSC_OK;
}

static enum s_status text (struct pdfc_compiler *pdc, 
			   struct gs_text *text,
			   struct gs_paint *paint)
{
	struct gs_text_state *tws;
	struct cobj_matrix mat;
	enum s_status res;

	res = align_dict (pdc, text -> gstate, text -> tstate);
	if (res != S_CSC_OK)
		return res;

	res = align_tstate (pdc, text -> tstate);
	if (res != S_CSC_OK)
		return res;

	res = align_gstate (pdc, text -> gstate, paint);
	if (res != S_CSC_OK)
		return res;

	res = gs_get_string (text, NULL, &mat);
	if (res != S_CSC_OK)
		return res;

	res = pdfw_append (pdc -> pdfw, PDFW_OP_DRAW_TEXT, 
			   &text -> text, &mat);
	if (res != S_CSC_OK)
		return res;

	tws = pdc -> gc -> state -> tstate;

	tws -> mat.t_m = mat;
	res = gs_text_update_key (tws, GS_GO_DICT);

	return S_CSC_OK;
}

static enum s_status compile_element (struct pdfc_compiler *pdc,
				      struct gs_element *e,
				      struct gs_paint *paint)
{
	enum s_status res;

	switch (e -> type) {
	case GS_ET_LINE:
		res = pdfw_append (pdc -> pdfw,
				   PDFW_OP_DRAW_LINE, 
				   e -> u.line.a.x, e -> u.line.a.y, 
				   e -> u.line.b.x, e -> u.line.b.y);
		break;

	case GS_ET_BEZIER:
		res = pdfw_append (pdc -> pdfw,
				   PDFW_OP_DRAW_BEZIER, 
				   e -> u.bezier.a.x, e -> u.bezier.a.y, 
				   e -> u.bezier.b.x, e -> u.bezier.b.y,
				   e -> u.bezier.ca.x, e -> u.bezier.ca.y,
				   e -> u.bezier.cb.x, e -> u.bezier.cb.y);
		if (res != S_CSC_OK)
			return res;
		break;

	case GS_ET_TEXT:
		res = text (pdc, &e -> u.text, paint);
		if (res != S_CSC_OK)
			return res;
		break;
	}

	return S_CSC_OK;
}

static enum s_status enter_text (struct pdfc_compiler *pdc,
				 struct gs_paint *paint,
				 enum gs_fill_rule cfrule)
{
	struct gs_text_state *tws;
	enum gs_render_type rt;
	enum s_status res;

	tws = pdc -> gc -> state -> tstate;
	tws -> mat.t_m = tws -> mat.t_lm = cobj_eye;

	res = pdfw_append (pdc -> pdfw, PDFW_OP_BT);
	if (res != S_CSC_OK)
		return res;
		
	gs_paint_to_render (paint, &rt, cfrule);
	if (tws -> render != rt) {
		res = pdfw_append (pdc -> pdfw, PDFW_OP_RENDER, 
				   (pdf_integer_t) rt);
		if (res != S_CSC_OK)
			return res;

		tws -> render = rt;
	}

	return S_CSC_OK;
}

static enum s_status compile_path (struct pdfc_compiler *pdc,
				   struct gs_path *path,
				   struct gs_paint *paint,
				   enum gs_fill_rule cfrule)
{
	struct gs_subpath *sp;
	enum s_status res;
	int i, text, has_path;

	for (sp = path -> head; sp != NULL; sp = sp -> next) {
		for (i = 0; i < sp -> elements; i++)
			if (sp -> e [i].type == GS_ET_TEXT)
				break;
		if (i != sp -> elements)
			break;
	}

	text = sp != NULL;

	if (text) {
		res = enter_text (pdc, paint, cfrule);
		if (res != S_CSC_OK)
			return res;
	}	

	has_path = 0;
	for (sp = path -> head; sp != NULL; sp = sp -> next) {
		for (i = 0; i < sp -> elements; i++) {

			has_path |= sp -> e [i].type != GS_ET_TEXT;
			res = compile_element (pdc, &sp -> e [i], paint);
			if (res != S_CSC_OK)
				return res;
		}
	}

	if (has_path) {
		if (cfrule != GS_FR_NONE) {
			res = pdfw_append (pdc -> pdfw, 
					   PDFW_OP_ADD_CLIPPING_PATH, 
					   (pdf_integer_t) cfrule);
			if (res != S_CSC_OK)
				return res;
		}

		if (paint != NULL)
			res = pdfw_append (pdc -> pdfw,
					   PDFW_OP_CLOSE_PATH, 
					   (pdf_integer_t) paint -> fill,
					   (pdf_integer_t) paint -> stroke,
					   (pdf_integer_t) paint -> rule);
		else
			res = pdfw_append (pdc -> pdfw,
					   PDFW_OP_CLOSE_PATH,
					   (pdf_integer_t) 0,
					   (pdf_integer_t) 0,
					   (pdf_integer_t) 0);
		if (res != S_CSC_OK)
			return res;
	}

	if (text) {
		res = pdfw_append (pdc -> pdfw, PDFW_OP_ET);
		if (res != S_CSC_OK)
			return res;
	}	

	return S_CSC_OK;
}

static enum s_status compile_shading (struct pdfc_compiler *pdc,
				      struct shd_shading *shd,
				      struct gs_graph_state *gss,
				      struct gs_text_state *tws)
{	
	struct obj name, obj;
	enum s_status res;

	res = shd_sync (NULL, &pdc -> robj, pdc -> sel, &obj, NULL, shd);
	if (res != S_CSC_OK)
		return res;

	res = grd_lookup (pdc -> shdt, &shd -> t.tag, &obj, 
			  shd -> t.rdb, &name, NULL, NULL);
	obj_destroy (&obj);
	if (res != S_CSC_OK)
		return res;

	res = pdfw_append (pdc -> pdfw, PDFW_OP_SHADING, &name);
	obj_destroy (&name);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status compile_xobject (struct pdfc_compiler *pdc,
				      struct xo_object *xo,
				      struct gs_graph_state *gss,
				      struct gs_text_state *tws)
{	
	struct obj name, obj;
	enum s_status res;

	res = xo_sync (NULL, &pdc -> robj, pdc -> sel, &obj, NULL, xo);
	if (res != S_CSC_OK)
		return res;

	res = grd_lookup (pdc -> xot, &xo -> t.tag, &obj,
			  xo -> t.rdb, &name, NULL, NULL);
	obj_destroy (&obj);
	if (res != S_CSC_OK)
		return res;

	res = pdfw_append (pdc -> pdfw, PDFW_OP_XOBJECT, &name);
	obj_destroy (&name);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status write_inline (struct pdfw_writer *pdfw, struct obj *obj)
{
	enum s_status res;

	if (obj -> type == OT_NULL)
		return S_CSC_OK;

	res = gspg_apply_shorthands (&obj -> val.stream.dictionary);
	if (res != S_CSC_OK)
		return res;

	res = pdfw_append (pdfw, PDFW_OP_INLINE_IMAGE, obj);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status compile_inline (struct pdfc_compiler *pdc,
				     struct xo_object *xo,
				     struct gs_graph_state *gss,
				     struct gs_text_state *tws)
{	
	struct xo_object *lxo;
	struct obj obj;
	enum s_status res;

	res = T_CALL (xo_cross, &lxo, xo, &pdc -> robj, 
		      pdc -> ob, &pdc -> owner);
	if (res != S_CSC_OK)
		return res;

	res = obj_dup (&lxo -> t.obj, &obj);
	xo_unref (lxo, &pdc -> owner);
	if (res != S_CSC_OK)
		return res;
		
	res = write_inline (pdc -> pdfw, &obj);
	obj_destroy (&obj);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status push_path (struct pdfc_compiler *pdc,
				struct gs_clipping_path *fpath,
				struct gs_clipping_path *lpath,
				struct gs_paint *paint)
{
	enum s_status res;

	if (fpath == lpath)
		return S_CSC_OK;

	res = push_path (pdc, fpath -> next, lpath, NULL);
	if (res != S_CSC_OK)
		return res;
	
	res = cpath_push (pdc, fpath -> path);
	if (res != S_CSC_OK)
		return res;

	res = pdfw_append (pdc -> pdfw, PDFW_OP_PUSH);
	if (res != S_CSC_OK) {
		cpath_pop (pdc);
		return res;
	}

	res = gs_push (pdc -> gc);
	if (res != S_CSC_OK) {
		cpath_pop (pdc);
		return res;
	}

	res = align_dict (pdc, fpath -> gstate, fpath -> tstate);
	if (res != S_CSC_OK)
		return res;

	res = align_gstate (pdc, fpath -> gstate, paint);
	if (res != S_CSC_OK)
		return res;

	res = compile_path (pdc, fpath -> path, NULL, fpath -> rule);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}
				

static enum s_status align_clipping_path (struct pdfc_compiler *pdc,
					  struct gs_graph_state *gs,
					  struct gs_path *path,
					  struct gs_paint *paint,
					  int *accepted)
{
	struct gs_clipping_path *p;
	enum s_status res;
	int eos;

	p = NULL;
	while (pdc -> cpath != NULL) {
		if (gs != NULL) {
			for (p = gs -> clipping_path; p != NULL; p = p -> next)
				if (p -> path == pdc -> cpath -> path)
					break;
			if (p != NULL)
				break;
		}

		res = gs_pop (pdc -> gc, &eos);
		if (res != S_CSC_OK)
			return res;

		res = pdfw_append (pdc -> pdfw, PDFW_OP_POP, 
				   &pdc -> gc -> state -> gstate -> ctm);
		if (res != S_CSC_OK)
			return res;

		cpath_pop (pdc);
	}

	if (gs != NULL && gs -> clipping_path != NULL) {
		if (gs -> clipping_path == p ||
		    gs -> clipping_path -> path != path)
			paint = NULL;
		if (accepted != NULL)
			*accepted = paint != NULL;
		res = push_path (pdc, gs -> clipping_path, p, paint);
		if (res != S_CSC_OK)
			return res;
	}

	return S_CSC_OK;
}

static enum s_status clear_stack (struct pdfc_compiler *pdc)
{
	return align_clipping_path (pdc, NULL, NULL, NULL, NULL);
}

static enum s_status compile_stmt (struct pdfc_compiler *pdc,
				   struct gspg_stmt **stmtp)
{
	struct gspg_stmt *stmt, *nstmt;
	struct gs_path *path;
	int accepted;
	enum s_status res;

	stmt = *stmtp;

	nstmt = stmt -> next;
	if (nstmt != NULL && nstmt -> type == GSPG_ST_PATH)
		path = stmt -> u.path;
	else
		path = NULL;

	if (!stmt -> no_cpath) {
		accepted = 0;
		res = align_clipping_path (pdc, stmt -> gstate, path, 
					   &stmt -> paint, &accepted);
		if (res != S_CSC_OK)
			return res;
		if (accepted)
			*stmtp = stmt -> next;
	} else {
		res = clear_stack (pdc);
		if (res != S_CSC_OK)
			return res;
	}

	res = align_dict (pdc, stmt -> gstate, stmt -> tstate);
	if (res != S_CSC_OK)
		return res;

	res = align_gstate (pdc, stmt -> gstate, &stmt -> paint);
	if (res != S_CSC_OK)
		return res;

	switch (stmt -> type) {
	case GSPG_ST_PATH:
		res = compile_path (pdc, stmt -> u.path, 
				    &stmt -> paint, GS_FR_NONE);
		break;
	case GSPG_ST_SHADING:
		res = compile_shading (pdc, stmt -> u.shading,
				       stmt -> gstate, stmt -> tstate);
		break;

	case GSPG_ST_XOBJECT:
		res = compile_xobject (pdc, stmt -> u.xobject,
				       stmt -> gstate, stmt -> tstate);
		break;

	case GSPG_ST_INLINE_IMAGE:
		res = compile_inline (pdc, stmt -> u.xobject,
				      stmt -> gstate, stmt -> tstate);
	}
	if (res != S_CSC_OK)
		return res;
		
	return S_CSC_OK;
}

enum s_status pdfc_compile (struct pdfc_compiler *pdc,
			    struct gspg_page *input,
			    struct pg_page *output)
{
	struct f_file *file;
	struct gspg_stmt *stmt;
	struct obj_region sreg;
	enum s_status res;

	f_buffer_open ("pdfc stream", &file);

	gs_context_unref (pdc -> gc, &pdc -> owner);
	
	res = T_CALL (gs_context_new, &pdc -> gc, pdc -> rdb, &pdc -> owner);
	if (res != S_CSC_OK)
		return res;

	res = pdfw_create (file, pdc -> sel, &pdc -> pdfw);
	if (res != S_CSC_OK)
		return res;

	destroy_cstate (&pdc -> cstate);

	cpath_empty (pdc);

	for (stmt = input -> head; stmt != NULL; stmt = stmt -> next) {
		res = compile_stmt (pdc, &stmt);
		if (res != S_CSC_OK) {
			f_unref (file);
			return res;
		}
	}

	res = clear_stack (pdc);
	if (res != S_CSC_OK)
		return res;

	res = pdfw_flush (pdc -> pdfw);
	if (res != S_CSC_OK) {
		f_unref (file);
		return res;
	}

	pdfw_destroy (pdc -> pdfw);
	pdc -> pdfw = NULL;

	obj_init_file_region (&sreg, file -> pdf, file -> end, file);
	f_unref (file);
	res = pg_page_store_contents (output, &sreg, &pdc -> robj, pdc -> sel);
	if (res != S_CSC_OK)
		return res;

	obj_region_destroy (&sreg);

	pg_page_set_resdict (output, pdc -> rd);

	return S_CSC_OK;
}

enum s_status pdfc_finalize (struct pdfc_compiler *pdc)
{
	enum s_status res;

	res = gst_rdict (pdc -> gst, &pdc -> rd -> ext_g_state);
	if (res != S_CSC_OK)
		return res;

	res = grd_rdict (pdc -> ft, &pdc -> rd -> font);
	if (res != S_CSC_OK)
		return res;

	res = grd_rdict (pdc -> shdt, &pdc -> rd -> shading);
	if (res != S_CSC_OK)
		return res;

	res = grd_rdict (pdc -> xot, &pdc -> rd -> xobject);
	if (res != S_CSC_OK)
		return res;

	res = grd_rdict (pdc -> patt, &pdc -> rd -> pattern);
	if (res != S_CSC_OK)
		return res;

	res = grd_rdict (pdc -> cspt, &pdc -> rd -> colour_space);
	if (res != S_CSC_OK)
		return res;
	
	return S_CSC_OK;
}
