/*
 * PDFlow -- pdfw.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <assert.h>
#include <string.h>

#include "fmt/pdf-lex.h"
#include "graph/gstate.h"
#include "pdf/pdfw.h"
#include "pdf/pdfo.h"

#define EPSILON 0
#define EEQ(a, b) (labs ((a) - (b)) < EPSILON)
#define ENE(a, b) (!EEQ (a, b))

enum opcode_arg {

	OPA_END = 0,

	OPA_BOOL,
	OPA_INTEGER,
	OPA_REAL,
	OPA_OBJ,
	OPA_MATRIX,
	OPA_ARRAY
};

struct pdfw_opcode;
struct pdfw_opcode_desc;

typedef enum s_status (*pdfw_handler_t) (struct pdfw_writer *pdfw,
					 struct pdfw_opcode *opc);

static enum s_status def_handler (struct pdfw_writer *pdfw,
				  struct pdfw_opcode *opc);

static enum s_status emit (struct pdfw_writer *pdfw, 
			   struct pdfw_opcode *opc);
static enum s_status h_set_ctm (struct pdfw_writer *pdfw,
				struct pdfw_opcode *opc);
static enum s_status h_text_matrix (struct pdfw_writer *pdfw,
				    struct pdfw_opcode *opc);
static enum s_status h_text_bt (struct pdfw_writer *pdfw,
				struct pdfw_opcode *opc);
static enum s_status h_draw_line (struct pdfw_writer *pdfw,
				  struct pdfw_opcode *opc);
static enum s_status h_draw_bezier (struct pdfw_writer *pdfw,
				    struct pdfw_opcode *opc);
static enum s_status h_close_path (struct pdfw_writer *pdfw,
				   struct pdfw_opcode *opc);
static enum s_status h_inline_image (struct pdfw_writer *pdfw,
				     struct pdfw_opcode *opc);

static enum s_status h_set_gray (struct pdfw_writer *pdfw,
				 struct pdfw_opcode *opc);
static enum s_status h_set_rgb (struct pdfw_writer *pdfw,
				struct pdfw_opcode *opc);
static enum s_status h_set_cmyk (struct pdfw_writer *pdfw,
				 struct pdfw_opcode *opc);

static enum s_status h_set_cspace (struct pdfw_writer *pdfw,
				   struct pdfw_opcode *opc);
static enum s_status h_set_colour (struct pdfw_writer *pdfw,
				   struct pdfw_opcode *opc);

static enum s_status h_draw_text (struct pdfw_writer *pdfw,
				  struct pdfw_opcode *opc);

static enum s_status h_pop (struct pdfw_writer *pdfw,
			    struct pdfw_opcode *opc);

static enum s_status h_add_clipping_path (struct pdfw_writer *pdfw,
					  struct pdfw_opcode *opc);
				      
struct pdfw_writer {

	struct ref_owner owner;

	struct v_selection *sel;

	struct pdfw_opcode *opc;

	struct pdfw_opcode **tail;

	struct f_file *dest;

	struct cobj_matrix ctm, tm, tlm;

	struct pdfo_optimizer *pdfo;

	pdf_real_t leading;
	
	int open;
	struct gs_point cp;
};

struct pdfw_opcode_desc {

	enum pdfw_opcode_type type;

	const char *name;

	enum opcode_arg args [PDFW_ARG_MAX + 1];

	pdfw_handler_t handler;
};

static struct pdfw_opcode_desc op_desc [] = {
	{ PDFW_OP_SET_CTM, "cm", { OPA_MATRIX }, h_set_ctm },
	{ PDFW_OP_LINE_WIDTH, "w", { OPA_REAL } },
	{ PDFW_OP_LINE_CAP, "J", { OPA_INTEGER } },
	{ PDFW_OP_LINE_JOIN, "j", { OPA_INTEGER } },
	{ PDFW_OP_MITER_LIMIT, "M", { OPA_REAL } },
	{ PDFW_OP_DASH_PATTERN, "d", { OPA_OBJ, OPA_INTEGER } },
	{ PDFW_OP_RENDERING_INTENT, "ri", { OPA_OBJ } },

	{ PDFW_OP_FONT, "Tf", { OPA_OBJ, OPA_REAL } },
	{ PDFW_OP_TEXT_MATRIX, "Tm", { OPA_REAL, OPA_MATRIX }, h_text_matrix },
	{ PDFW_OP_CHAR_SPACE, "Tc", { OPA_REAL } },
	{ PDFW_OP_WORD_SPACE, "Tw", { OPA_REAL } },
	{ PDFW_OP_SCALE, "Tz", { OPA_REAL } },
	{ PDFW_OP_LEADING, "TL", { OPA_REAL } },
	{ PDFW_OP_RENDER, "Tr", { OPA_INTEGER } },
	{ PDFW_OP_RISE, "Ts", { OPA_REAL } },

	{ PDFW_OP_GS, "gs", { OPA_OBJ } },

	{ PDFW_OP_BT, "BT", {  }, h_text_bt },
	{ PDFW_OP_ET, "ET", {  } },

	{ PDFW_OP_DRAW_LINE, "l", { OPA_REAL, OPA_REAL,
				    OPA_REAL, OPA_REAL }, h_draw_line },
	{ PDFW_OP_DRAW_BEZIER, "c", { OPA_REAL, OPA_REAL,
				      OPA_REAL, OPA_REAL,
				      OPA_REAL, OPA_REAL,
				      OPA_REAL, OPA_REAL }, h_draw_bezier },

	{ PDFW_OP_DRAW_TEXT, "TJ", { OPA_OBJ, OPA_MATRIX }, h_draw_text },

	{ PDFW_OP_CLOSE_PATH, "n", { 
		OPA_INTEGER, OPA_INTEGER, OPA_INTEGER }, h_close_path},

	{ PDFW_OP_SHADING, "sh", { OPA_OBJ } },
	{ PDFW_OP_XOBJECT, "Do", { OPA_OBJ } },
	{ PDFW_OP_INLINE_IMAGE, "BI", { OPA_OBJ }, h_inline_image },

	{ PDFW_OP_SET_GRAY, "G", { OPA_BOOL, OPA_ARRAY }, h_set_gray },
	{ PDFW_OP_SET_RGB, "RG", { OPA_BOOL, OPA_ARRAY }, h_set_rgb },
	{ PDFW_OP_SET_CMYK, "K", { OPA_BOOL, OPA_ARRAY }, h_set_cmyk },

	{ PDFW_OP_SET_CSPACE, "CS", { OPA_BOOL, OPA_OBJ }, h_set_cspace },
	{ PDFW_OP_SET_COLOUR, "SCN", { OPA_BOOL, OPA_ARRAY }, h_set_colour },

	{ PDFW_OP_PUSH, "q" },
	{ PDFW_OP_POP, "Q", { OPA_MATRIX }, h_pop },

	{ PDFW_OP_ADD_CLIPPING_PATH, "W", { OPA_INTEGER }, 
	  h_add_clipping_path },
	
	{ -1, NULL }
};

static struct pdfw_opcode_desc opd_cm = { -1, "cm", { 
	OPA_REAL, OPA_REAL, OPA_REAL, OPA_REAL, OPA_REAL, OPA_REAL } };
static struct pdfw_opcode_desc opd_T_ = { -1, "T*" };
static struct pdfw_opcode_desc opd_TD = { -1, "TD", { OPA_REAL, OPA_REAL } };
static struct pdfw_opcode_desc opd_Td = { -1, "Td", { OPA_REAL, OPA_REAL } };
static struct pdfw_opcode_desc opd_Tm = { -1, "Tm", { 
	OPA_REAL, OPA_REAL, OPA_REAL, OPA_REAL, OPA_REAL, OPA_REAL } };
static struct pdfw_opcode_desc opd_BT = { -1, "BT" };

static struct pdfw_opcode_desc opd_m = { -1, "m", { OPA_REAL, OPA_REAL } };
static struct pdfw_opcode_desc opd_l = { -1, "l", { OPA_REAL, OPA_REAL } };

static struct pdfw_opcode_desc opd_v = { -1, "v", { 
	OPA_REAL, OPA_REAL, OPA_REAL, OPA_REAL } };
static struct pdfw_opcode_desc opd_y = { -1, "y", {
	OPA_REAL, OPA_REAL, OPA_REAL, OPA_REAL } };
static struct pdfw_opcode_desc opd_c = { -1, "c", {
	OPA_REAL, OPA_REAL, OPA_REAL, OPA_REAL, OPA_REAL, OPA_REAL } };

static struct pdfw_opcode_desc opd_B_ = { -1, "B*" };
static struct pdfw_opcode_desc opd_B = { -1, "B" };
static struct pdfw_opcode_desc opd_f_ = { -1, "f*" };
static struct pdfw_opcode_desc opd_f = { -1, "f" };
static struct pdfw_opcode_desc opd_S = { -1, "S" };
static struct pdfw_opcode_desc opd_n = { -1, "n" };
static struct pdfw_opcode_desc opd_CS = { -1, "CS", { OPA_OBJ } };
static struct pdfw_opcode_desc opd_cs = { -1, "cs", { OPA_OBJ } };

static struct pdfw_opcode_desc opd_Q = { -1, "Q" };

static struct pdfw_opcode_desc opd_W = { -1, "W" };
static struct pdfw_opcode_desc opd_W_ = { -1, "W*" };

static struct pdfw_opcode_desc opd_Tj = { -1, "Tj", { OPA_OBJ } };
static struct pdfw_opcode_desc opd_TJ = { -1, "TJ", { OPA_OBJ } };

static enum s_status dup_array (struct xref_obj_array *sma,
				struct xref_obj_array *dma)
{
	enum s_status res;
	int i;

	dma -> n = XREF_A_SIZE (sma);
	if (!dma -> n) {
		dma -> data = NULL;
		return S_CSC_OK;
	}

	dma -> data = calloc (dma -> n, sizeof (struct obj));
	if (dma -> data == NULL)
		return S_CSC_OUT_OF_MEMORY;

	for (i = 0; i < dma -> n; i++) {
		res = obj_dup (&sma -> data [i], &dma -> data [i]);
		if (res != S_CSC_OK) {
			xref_array_map_rel_ps (dma, NULL, 
					       &xref_obj_pstable, NULL);
			return res;
		}
	}

	return S_CSC_OK;
}

static enum s_status vinit_opcode (struct pdfw_opcode *opc, 
				   struct pdfw_opcode_desc *opd,
				   va_list ap)
{
	enum s_status res;
	struct obj *obj;
	int i;

	opc -> opd = opd;

	for (i = 0; opd -> args [i] != OPA_END; i++) {
		obj = &opc -> arg [i];
		switch (opd -> args [i]) {
		case OPA_BOOL:
			obj_set_boolean (obj, &f_null,
					 va_arg (ap, pdf_bool_t));
			break;

		case OPA_INTEGER:
			obj_set_integer (obj, &f_null,
					 va_arg (ap, pdf_integer_t));
			break;

		case OPA_REAL:
			obj_set_real (obj, &f_null,
				      va_arg (ap, pdf_real_t));
			break;

		case OPA_OBJ:
			res = obj_dup (va_arg (ap, struct obj *), obj);
			if (res != S_CSC_OK)
				return res;

			break;

  		case OPA_MATRIX:
			opc -> matrix = *(va_arg (ap, struct cobj_matrix *));
			break;

		case OPA_ARRAY:
			res = dup_array (va_arg (ap, struct xref_obj_array *),
					 &opc -> array);
			if (res != S_CSC_OK)
				return res;
			break;

		case OPA_END:
			assert (0);
		}
	}

	opc -> args = i;

	return S_CSC_OK;
}

static void destroy_opcode (struct pdfw_opcode *opc, int dyn)
{
	int i;

	for (i = 0; i < opc -> args; i++)
		obj_destroy (&opc -> arg [i]);

	xref_array_map_rel_ps (&opc -> array, NULL, &xref_obj_pstable, NULL);
	if (dyn)
		free (opc);
}

static enum s_status emit_varg (struct pdfw_writer *pdfw, 
				struct pdfw_opcode_desc *opd, va_list ap)
{
	struct pdfw_opcode opc;
	enum s_status res;

	memset (&opc, 0, sizeof (struct pdfw_opcode));

	res = vinit_opcode (&opc, opd, ap);
	if (res != S_CSC_OK)
		return res;

	res = emit (pdfw, &opc);
	destroy_opcode (&opc, 0);
	if (res != S_CSC_OK)
		return res;
	
	return res;
}

static enum s_status emit_arg (struct pdfw_writer *pdfw, 
			       struct pdfw_opcode_desc *opd, ...)
{
	enum s_status res;
	va_list ap;

	va_start (ap, opd);
	res = emit_varg (pdfw, opd, ap);
	va_end (ap);
	return res;
}

static enum s_status init (struct pdfw_writer *pdfw, 
			   struct v_selection *sel,
			   struct f_file *dest)
{
	enum s_status res;

	ref_init_owner (&pdfw -> owner, "pdfw");

	pdfw -> tail = &pdfw -> opc;
	pdfw -> dest = dest;
	pdfw -> sel = sel;

	f_ref (pdfw -> dest);

	res = pdfo_create (pdfw, &pdfw -> pdfo);
	if (res != S_CSC_OK)
		return res;

	pdfw -> ctm = cobj_eye;

	return S_CSC_OK;
}

enum s_status pdfw_create (struct f_file *dest,
			   struct v_selection *sel,
			   struct pdfw_writer **pdfwp)
{
	struct pdfw_writer *pdfw;
	enum s_status res;

	pdfw = malloc (sizeof (struct pdfw_writer));
	if (pdfw == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (pdfw, 0, sizeof (struct pdfw_writer));

	res = init (pdfw, sel, dest);
	if (res != S_CSC_OK) {
		pdfw_destroy (pdfw);
		return res;
	}
	
	*pdfwp = pdfw;

	return S_CSC_OK;
}

void pdfw_destroy (struct pdfw_writer *pdfw)
{
	struct pdfw_opcode *opc;

	if (pdfw == NULL)
		return;

	if (pdfw -> pdfo != NULL)
		pdfo_destroy (pdfw -> pdfo);

	while (pdfw -> opc != NULL) {
		opc = pdfw -> opc;
		pdfw -> opc = opc -> next;

		destroy_opcode (opc, 1);
	}

	f_unref (pdfw -> dest);

	free (pdfw);
}

static enum s_status h_set_ctm (struct pdfw_writer *pdfw,
				struct pdfw_opcode *opc)
{
	struct cobj_matrix tm;
	enum s_status res;

	cobj_matrix_div (&opc -> matrix, &pdfw -> ctm, &tm, 1);
	res = emit_arg (pdfw, &opd_cm, 
			tm.a [0][0], tm.a [0][1],
			tm.a [1][0], tm.a [1][1],
			tm.a [2][0], tm.a [2][1]);
	if (res != S_CSC_OK)
		return res;

	pdfw -> ctm = opc -> matrix;

	return S_CSC_OK;
}

static enum s_status h_text_matrix (struct pdfw_writer *pdfw,
				    struct pdfw_opcode *opc)
{
	struct cobj_matrix *tm, *tlm, *ttm;
	struct cobj_matrix dm;
	pdf_real_t leading;
	pdf_real_t tx, ty;
	enum s_status res;

	tm = &pdfw -> tm;
	tlm = &pdfw -> tlm;
	ttm = &opc -> matrix;
	leading = opc -> arg [0].val.real;

	if (memcmp (tm, ttm, sizeof (struct cobj_matrix))) {

		cobj_matrix_div (ttm, tlm, &dm, 1);

		tx = dm.a [2][0];
		ty = dm.a [2][1];

		if (ENE (tm -> a [0][0], ttm -> a [0][0]) ||
		    ENE (tm -> a [0][1], ttm -> a [0][1]) ||
		    ENE (tm -> a [1][0], ttm -> a [1][0]) ||
		    ENE (tm -> a [1][1], ttm -> a [1][1]))
			res = emit_arg (pdfw, &opd_Tm,
					ttm -> a [0][0], ttm -> a [0][1],
					ttm -> a [1][0], ttm -> a [1][1],
					ttm -> a [2][0], ttm -> a [2][1]);
		else if (EEQ (tx, 0) && EEQ (ty, -pdfw -> leading))
			res = emit_arg (pdfw, &opd_T_);
		else if (ty == -leading) {
			res = emit_arg (pdfw, &opd_TD, tx, ty);
			if (res == S_CSC_OK)
				pdfw -> leading = -ty;
		} else 
			res = emit_arg (pdfw, &opd_Td, tx, ty);

		if (res != S_CSC_OK)
			return res;

		*tm = *tlm = *ttm;
	}

	if (pdfw -> leading != leading) {
		res = emit_arg (pdfw, &op_desc [PDFW_OP_LEADING], leading);
		if (res != S_CSC_OK)
			return res;

		pdfw -> leading = leading;
	}

	return S_CSC_OK;
}

static enum s_status h_text_bt (struct pdfw_writer *pdfw,
				struct pdfw_opcode *opc)
{
	pdfw -> leading = 0;
	pdfw -> tm = pdfw -> tlm = cobj_eye;

	return emit_arg (pdfw, &opd_BT);
}

static enum s_status h_draw_line (struct pdfw_writer *pdfw,
				  struct pdfw_opcode *opc)
{
	struct gs_point a, b;
	enum s_status res;

	a.x = opc -> arg [0].val.real;
	a.y = opc -> arg [1].val.real;
	b.x = opc -> arg [2].val.real;
	b.y = opc -> arg [3].val.real;

	if (pdfw -> open && gs_point_same (&a, &pdfw -> cp)) {
		res = emit_arg (pdfw, &opd_l, b.x, b.y);
		if (res != S_CSC_OK)
			return res;
		pdfw -> cp = b;
	} else if (pdfw -> open && gs_point_same (&b, &pdfw -> cp)) {
		res = emit_arg (pdfw, &opd_l, a.x, a.y);
		if (res != S_CSC_OK)
			return res;
		pdfw -> cp = a;
	} else {
		res = emit_arg (pdfw, &opd_m, a.x, a.y);
		if (res != S_CSC_OK)
			return res;
		pdfw -> cp = a;

		res = emit_arg (pdfw, &opd_l, b.x, b.y);
		if (res != S_CSC_OK)
			return res;
		pdfw -> cp = b;
	}

	pdfw -> open = 1;

	return S_CSC_OK;
}

static enum s_status h_draw_bezier (struct pdfw_writer *pdfw,
				    struct pdfw_opcode *opc)
{
	struct gs_point a, b, ca, cb, x;
	enum s_status res;

	a.x = opc -> arg [0].val.real;
	a.y = opc -> arg [1].val.real;
	b.x = opc -> arg [2].val.real;
	b.y = opc -> arg [3].val.real;

	ca.x = opc -> arg [4].val.real;
	ca.y = opc -> arg [5].val.real;
	cb.x = opc -> arg [6].val.real;
	cb.y = opc -> arg [7].val.real;

	if (pdfw -> open && gs_point_same (&a, &pdfw -> cp))
		;
	else if (pdfw -> open && gs_point_same (&b, &pdfw -> cp)) {
		x = b, a = b, x = a;
		x = cb, ca = cb, x = ca;
	} else {
		res = emit_arg (pdfw, &opd_m, a.x, a.y);
		if (res != S_CSC_OK)
			return res;
		pdfw -> cp = a;
	}
	pdfw -> open = 1;

	if (gs_point_same (&a, &ca))
		res = emit_arg (pdfw, &opd_v, cb.x, cb.y, b.x, b.y);
	else if (gs_point_same (&b, &cb))
		res = emit_arg (pdfw, &opd_y, ca.x, ca.y, b.x, b.y);
	else
		res = emit_arg (pdfw, &opd_c, ca.x, ca.y, 
				cb.x, cb.y, b.x, b.y);
	if (res != S_CSC_OK)	
		return res;

	pdfw -> cp = b;

	return S_CSC_OK;
}

static enum s_status h_draw_text (struct pdfw_writer *pdfw,
				  struct pdfw_opcode *opc)
{
	struct pdfw_opcode_desc *desc;
	enum s_status res;

	desc = opc -> arg [0].type == OT_STRING ? &opd_Tj : &opd_TJ;

	res = emit_arg (pdfw, desc, &opc -> arg [0]);
	if (res != S_CSC_OK)
		return res;
	
	pdfw -> tm = opc -> matrix;

	return S_CSC_OK;
}

static enum s_status h_close_path (struct pdfw_writer *pdfw,
				   struct pdfw_opcode *opc)
{
	pdf_bool_t fill, stroke;
	struct pdfw_opcode_desc *opd;
	enum gs_fill_rule rule;
	enum s_status res;

	fill = opc -> arg [0].val.integer;
	stroke = opc -> arg [1].val.integer;
	rule = opc -> arg [2].val.integer;
 	pdfw -> open = 0;

	if (fill) {
		if (stroke)
			opd = rule == GS_FR_WINDING ? &opd_B : &opd_B_;
		else
			opd = rule == GS_FR_WINDING ? &opd_f : &opd_f_;
	} else {
		if (stroke)
			opd = &opd_S;
		else
			opd = &opd_n;
	}

	res = emit_arg (pdfw, opd);
	if (res != S_CSC_OK)	
		return res;

	return S_CSC_OK;
}

static enum s_status h_inline_image (struct pdfw_writer *pdfw,
				   struct pdfw_opcode *opc)
{
	struct obj_dict_node *dn;
	struct obj_region *reg;
	struct obj *obj, name;
	enum s_status res;
	size_t len;

	obj = &opc -> arg [0];

	res = f_writeln (pdfw -> dest, "BI");
	if (res != S_CSC_OK)
		return res;

	for (dn = obj -> val.stream.dictionary.head; 
	     dn != NULL; dn = dn -> next) {
		obj_set_name (&name, &f_null, &dn -> key);
		res = pl_serialize (pdfw -> dest, &name);
		if (res != S_CSC_OK)
			return res;

		res = f_write (pdfw -> dest, " ");
		if (res != S_CSC_OK)
			return res;

		res = pl_serialize (pdfw -> dest, &dn -> obj);
		if (res != S_CSC_OK)
			return res;

		res = f_newline (pdfw -> dest);
		if (res != S_CSC_OK)
			return res;
	}

	res = f_write (pdfw -> dest, "ID ");
	if (res != S_CSC_OK)
		return res;

	reg = obj -> val.stream.region.begin != NULL ? 
		&obj -> val.stream.region : 
		&obj -> val.stream.decoded;
	len = reg -> end - reg -> begin;
	if (len && 
	    (reg -> end [-1] == '\n' || reg -> end [-1] == '\r'))
		len--;

	res = f_raw (pdfw -> dest, reg -> begin, len);
	if (res != S_CSC_OK)
		return res;

	res = f_newline (pdfw -> dest);
	if (res != S_CSC_OK)
		return res;

	res = f_writeln (pdfw -> dest, "EI");
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status array_dump (struct pdfw_writer *pdfw,
				 struct xref_obj_array *ma, 
				 const char *op)
{
	enum s_status res;
	int i;

	for (i = 0; i < XREF_A_SIZE (ma); i++) {
		res = pl_serialize (pdfw -> dest, &ma -> data [i]);
		if (res != S_CSC_OK)
			return res;

		res = f_write (pdfw -> dest, " ");
		if (res != S_CSC_OK)
			return res;
	}

	res = f_writeln (pdfw -> dest, "%s", op);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status h_set_gray (struct pdfw_writer *pdfw,
				 struct pdfw_opcode *opc)
{
	return array_dump (pdfw, &opc -> array, 
			   opc -> arg [0].val.bool ? "G" : "g");
}

static enum s_status h_set_rgb (struct pdfw_writer *pdfw,
				struct pdfw_opcode *opc)
{
	return array_dump (pdfw, &opc -> array, 
			   opc -> arg [0].val.bool ? "RG" : "rg");
}

static enum s_status h_set_cmyk (struct pdfw_writer *pdfw,
				 struct pdfw_opcode *opc)
{
	return array_dump (pdfw, &opc -> array, 
			   opc -> arg [0].val.bool ? "K" : "k");
}

static enum s_status h_set_cspace (struct pdfw_writer *pdfw,
				   struct pdfw_opcode *opc)
{
	struct pdfw_opcode_desc *opd;

	opd = opc -> arg [0].val.bool ? &opd_CS : &opd_cs;

	return emit_arg (pdfw, opd, &opc -> arg [1]);
}

static enum s_status h_set_colour (struct pdfw_writer *pdfw,
				   struct pdfw_opcode *opc)
{
	const char *op;
	int n;

	n = XREF_A_SIZE (&opc -> array);
	if (n && opc -> array.data [n - 1].type == OT_NAME) {
		if (opc -> arg [0].val.bool)
			op = "SCN";
		else
			op = "scn";
	} else {
		if (opc -> arg [0].val.bool)
			op = "SC";
		else
			op = "sc";
	}

	return array_dump (pdfw, &opc -> array, op);
}

static enum s_status h_pop (struct pdfw_writer *pdfw,
			    struct pdfw_opcode *opc)
{
	pdfw -> ctm = opc -> matrix;

	return emit_arg (pdfw, &opd_Q);
}

static enum s_status h_add_clipping_path (struct pdfw_writer *pdfw,
					  struct pdfw_opcode *opc)
{
	struct pdfw_opcode_desc *opd;
	enum gs_fill_rule rule;

	rule = opc -> arg [0].val.integer;

	switch (rule) {

	case GS_FR_WINDING:
		opd = &opd_W;
		break;

	case GS_FR_EVEN_ODD:
		opd = &opd_W_;
		break;

	default:
		assert (0);
	}

	return emit_arg (pdfw, opd);
}

static enum s_status def_handler (struct pdfw_writer *pdfw,
				  struct pdfw_opcode *opc)
{
	enum s_status res;
	int i;

	for (i = 0; i < opc -> args; i++) {

		res = pl_serialize (pdfw -> dest, &opc -> arg [i]);
		if (res != S_CSC_OK)
			return res;

		res = f_write (pdfw -> dest, " ");
		if (res != S_CSC_OK)
			return res;
	}

	res = f_writeln (pdfw -> dest, "%s", opc -> opd -> name);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status emit (struct pdfw_writer *pdfw, struct pdfw_opcode *opc)
{
	pdfw_handler_t h;

	h = opc -> opd -> handler;
	if (h == NULL)
		h = def_handler;

	return h (pdfw, opc);
}

enum s_status pdfw_ll_flush (struct pdfw_writer *pdfw)
{
	struct pdfw_opcode *opc;
	enum s_status res;

	while (pdfw -> opc != NULL) {
		opc = pdfw -> opc;
		pdfw -> opc = opc -> next;
		if (pdfw -> opc == NULL)
			pdfw -> tail = &pdfw -> opc;

		res = emit (pdfw, opc);
		destroy_opcode (opc, 1);

		if (res != S_CSC_OK)
			return res;
	}

	return S_CSC_OK;
}

enum s_status pdfw_append (struct pdfw_writer *pdfw,
			   enum pdfw_opcode_type type, ...)
{
	struct pdfw_opcode_desc *opd;
	struct pdfw_opcode *opc;
	enum s_status res;
	va_list ap;

	opd = &op_desc [type];
#ifdef DEBUG
	assert (opd -> type == type);
#endif	

	if (pdfw -> sel -> pref -> skip_optimizer) {
		va_start (ap, type);
		res = emit_varg (pdfw, opd, ap);
		va_end (ap);
		return res;
	}

	opc = malloc (sizeof (struct pdfw_opcode));
	if (opc == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (opc, 0, sizeof (struct pdfw_opcode));

	va_start (ap, type);
	res = vinit_opcode (opc, opd, ap);
	va_end (ap);

	if (res != S_CSC_OK) {
		destroy_opcode (opc, 1);
		return res;
	}

	*pdfw -> tail = opc;
	pdfw -> tail = &opc -> next;

	res = pdfo_flush (pdfw -> pdfo, &pdfw -> opc, &pdfw -> tail, 0);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

enum s_status pdfw_flush (struct pdfw_writer *pdfw)
{
	enum s_status res;

	if (!pdfw -> sel -> pref -> skip_optimizer) {
		res = pdfo_flush (pdfw -> pdfo, &pdfw -> opc, 
				  &pdfw -> tail, 1);
		if (res != S_CSC_OK)
			return res;
	} else
		assert (pdfw -> opc == NULL);

	return S_CSC_OK;
}
