/*
 * PDFlow -- cid.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <fcntl.h>
#include <stdio.h>

#include "fmt/ctab.h"
#include "ps/psi.h"
#include "psi_private.h"
#include "types/cmap.h"

#define CMAP_SIGNATURE "%!PS-Adobe-3.0 Resource-CMap"
#define CMAP_BEGINRESOURCE "%%BeginResource"

static enum s_status tr_usecmap (struct m_stack *stack,
				 struct obj *obj, int token,
				 void *closure);
static enum s_status tr_usefont (struct m_stack *stack,
				 struct obj *obj, int token,
				 void *closure);
static enum s_status tr_begincodespacerange (struct m_stack *stack,
					     struct obj *obj, int token,
					     void *closure);
static enum s_status tr_endcodespacerange (struct m_stack *stack,
					   struct obj *obj, int token,
					   void *closure);
static enum s_status tr_begincmap (struct m_stack *stack,
				   struct obj *obj, int token,
				   void *closure);
static enum s_status tr_endcmap (struct m_stack *stack,
				 struct obj *obj, int token,
				 void *closure);
static enum s_status tr_beginbfchar (struct m_stack *stack,
				     struct obj *obj, int token,
				     void *closure);
static enum s_status tr_endbfchar (struct m_stack *stack,
				   struct obj *obj, int token,
				   void *closure);
static enum s_status tr_beginbfrange (struct m_stack *stack,
				      struct obj *obj, int token,
				      void *closure);
static enum s_status tr_endbfrange (struct m_stack *stack,
				    struct obj *obj, int token,
				    void *closure);
static enum s_status tr_begincidchar (struct m_stack *stack,
				      struct obj *obj, int token,
				      void *closure);
static enum s_status tr_endcidchar (struct m_stack *stack,
				    struct obj *obj, int token,
				    void *closure);
static enum s_status tr_begincidrange (struct m_stack *stack,
				       struct obj *obj, int token,
				       void *closure);
static enum s_status tr_endcidrange (struct m_stack *stack,
				     struct obj *obj, int token,
				     void *closure);
static enum s_status tr_beginnotdefrange (struct m_stack *stack,
					  struct obj *obj, int token,
					  void *closure);
static enum s_status tr_endnotdefrange (struct m_stack *stack,
					struct obj *obj, int token,
					void *closure);

static enum s_status cid_init (struct m_stack *stack,
			       struct psi_closure *closure);


static struct obj_keyword keywords [] = {
	{ "usecmap", tr_usecmap },
	{ "usefont", tr_usefont },
	{ "begincodespacerange", tr_begincodespacerange },
	{ "endcodespacerange", tr_endcodespacerange },
	{ "begincmap", tr_begincmap },
	{ "endcmap", tr_endcmap },
	{ "beginbfchar", tr_beginbfchar },
	{ "endbfchar", tr_endbfchar },
	{ "beginbfrange", tr_beginbfrange },
	{ "endbfrange", tr_endbfrange },
	{ "begincidchar", tr_begincidchar },
	{ "endcidchar", tr_endcidchar },
	{ "begincidrange", tr_begincidrange },
	{ "endcidrange", tr_endcidrange },
	{ "beginnotdefrange", tr_beginnotdefrange },
	{ "endnotdefrange", tr_endnotdefrange },
	{ NULL }
};

struct psi_procset psi_cmap_procset = {
	"CIDInit", keywords, cid_init
};

static enum s_status cid_init (struct m_stack *stack,
			       struct psi_closure *closure)
{
	enum s_status res;
	struct obj obj;

	obj_set_null (&obj, &f_null);
	res = psi_push (stack, &obj, TOK_DICTIONARY);
	if (res != S_CSC_OK)
		return res;
	
	return S_CSC_OK;
}

static enum s_status get_cmap (struct psi_closure *pc,
			       struct psi_ccmap **cmap)
{
	struct psi_interpreter *pi;

	pi = pc -> interpreter;
	if (pi -> ccmap == NULL)
		return psi_error (pc, "No cmap here");

	*cmap = pi -> ccmap;

	return S_CSC_OK;
}

static enum s_status get_compiled_ccmap (struct psi_interpreter *pi,
					 struct obj *name,
					 struct psi_ccmap **cmap)
{
	struct psi_ccmap *p;

	for (p = pi -> resources -> ccmaps; p != NULL; p = p -> next)
		if (obj_string_equal_s (&name -> val.string, 
					&p -> cmap -> cmap_name.val.string))
			break;

	*cmap = p;

	return S_CSC_OK;
}


static enum s_status get_ccmap_by_name (struct psi_interpreter *pi,
					struct obj *name,
					struct psi_ccmap **cmap)
{
	struct psi_file_resource *fr;
	struct psi_interpreter *cpi;
	struct f_file *file;
	enum s_status res;

	res = get_compiled_ccmap (pi, name, cmap);
	if (res != S_CSC_OK)
		return res;

	if (*cmap != NULL)
		return S_CSC_OK;

	fr = psi_find_file_resource (pi -> resources -> cmapres, name);
	if (fr == NULL)
		return S_CSC_OK;

	res = f_open (fr -> path, &file);
	if (res != S_CSC_OK)
		return S_CSC_OK;

	res = psi_clone (pi, &cpi);
	if (res != S_CSC_OK) {
		f_unref (file);
		return res;
	}

	res = psi_add_cmap (cpi, file, NULL, NULL);
	f_unref (file);
	psi_destroy (cpi);
	if (res != S_CSC_OK)
		return res;

	res = get_compiled_ccmap (pi, name, cmap);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

enum s_status psi_cid_get_cmap_by_name (struct psi_interpreter *pi,
					struct obj *name,
					struct ref_owner *p,
					struct cmap **cmap)
{
	struct psi_ccmap *ccmap;
	enum s_status res;

	res = get_ccmap_by_name (pi, name, &ccmap);
	if (res != S_CSC_OK)
		return res;

	if (ccmap == NULL) {
		*cmap = NULL;
		return S_CSC_OK;
	}

	T_CALL (cmap_ref, ccmap -> cmap, p);
	*cmap = ccmap -> cmap;

	return S_CSC_OK;
}

void psi_cid_get_last_cmap (struct psi_interpreter *pi,
			    struct ref_owner *p,
			    struct cmap **cmapp)
{
	struct psi_ccmap *cmap;

	cmap = pi -> resources -> ccmaps;

	if (cmap != NULL)
		T_CALL (cmap_ref, cmap -> cmap, p);

	*cmapp = cmap -> cmap;
}

static enum s_status tr_usecmap (struct m_stack *stack,
				 struct obj *obj, int token,
				 void *closure)
{
	struct psi_closure *pc;
	enum s_status res;
	struct psi_ccmap *cmap, *bcmap;
	struct cmap_table *t, *nt;
	struct obj nobj;

	pc = closure;
	res = get_cmap (pc, &cmap);
	if (res != S_CSC_OK)
		return res;

	res = psi_safe_pop (stack, closure, 1, &nobj);
	if (res != S_CSC_OK)
		return res;

	if (nobj.type != OT_NAME) {
		res = obj_error (&nobj, "Expected name");
		obj_destroy (&nobj);
		return res;
	}

	res = get_ccmap_by_name (pc -> interpreter, &nobj, &bcmap);
	obj_destroy (&nobj);
	if (res != S_CSC_OK)
		return res;

	for (t = bcmap -> tables; t != NULL; t = t -> next) {
		res = cmap_dup_table (t, &nt);
		if (res != S_CSC_OK)
			return res;

		nt -> next = cmap -> tables;
		cmap -> tables = nt;
	}

	return S_CSC_OK;
}

static enum s_status tr_usefont (struct m_stack *stack,
				 struct obj *obj, int token,
				 void *closure)
{
	enum s_status res;
	struct obj nobj;

	res = psi_safe_pop (stack, closure, 1, &nobj);
	if (res != S_CSC_OK)
		return res;

	if (nobj.type != OT_INTEGER) {
		res = obj_error (&nobj, "Expected integer");
		obj_destroy (&nobj);
		return res;
	}

	if (nobj.val.integer) {
		res = obj_error (&nobj, "PDF supports font 0 only");
		obj_destroy (&nobj);
		return res;
	}

	return S_CSC_OK;
}

static enum s_status begin_marker (struct m_stack *stack,
				   struct obj *obj, int token,
				   void *closure)
{
	enum s_status res;
	struct obj nobj;

	res = psi_safe_pop (stack, closure, 1, &nobj);
	if (res != S_CSC_OK)
		return res;

	if (nobj.type != OT_INTEGER) {
		res = obj_error (&nobj, "Expected integer");
		obj_destroy (&nobj);
		return res;
	}

	res = m_push (stack, obj, token);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status tr_begincodespacerange (struct m_stack *stack,
					     struct obj *obj, int token,
					     void *closure)
{
	return begin_marker (stack, obj, token, closure);
}

static enum s_status tr_endcodespacerange (struct m_stack *stack,
					   struct obj *obj, int token,
					   void *closure)
{
	struct psi_closure *pc;
	struct obj sobj, eobj;
	struct cmap_table *table;
	struct psi_ccmap *cmap;
	enum s_status res;

	pc = closure;
	res = get_cmap (pc, &cmap);
	if (res != S_CSC_OK)
		return res;

	while (1) {
		res = psi_safe_pop (stack, closure, 1, &eobj);
		if (res != S_CSC_OK)
			return res;

		if (eobj.type == OT_KEYWORD)
			break;

		if (eobj.type != OT_STRING) {
			res = obj_error (&eobj, "Expected string");
			obj_destroy (&eobj);
			return res;
		}

		res = psi_safe_pop (stack, closure, 1, &sobj);
		if (res != S_CSC_OK) {
			obj_destroy (&eobj);
			return res;
		}

		if (sobj.type != OT_STRING) {
			res = obj_error (&sobj, "Expected string");
			obj_destroy (&eobj);
			obj_destroy (&sobj);
			return res;
		}

		res = cmap_create_table (&table, &sobj, &eobj);
		obj_destroy (&eobj);
		obj_destroy (&sobj);
		if (res != S_CSC_OK)
			return res;

		table -> next = cmap -> tables;
		cmap -> tables = table;		
	}

	if (eobj.val.keyword -> data != tr_begincodespacerange)
		return obj_error (&eobj, "begin/end codespacerange mismatch");

	return S_CSC_OK;
}

static enum s_status tr_begincmap (struct m_stack *stack,
				   struct obj *obj, int token,
				   void *closure)
{
	struct psi_closure *pc;
	struct psi_interpreter *pi;
	struct psi_ccmap *ccmap;

	pc = closure;
	pi = pc -> interpreter;

	ccmap = malloc (sizeof (struct psi_ccmap));
	if (ccmap == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (ccmap, 0, sizeof (struct psi_ccmap));

	ref_init_owner (&ccmap -> owner, "ccmap");

	ccmap -> next = pi -> ccmap;
	pi -> ccmap = ccmap;

	return S_CSC_OK;
}

static enum s_status tr_endcmap (struct m_stack *stack,
				 struct obj *obj, int token,
				 void *closure)
{
	return S_CSC_OK;
}

static enum s_status tr_beginbfchar (struct m_stack *stack,
				     struct obj *obj, int token,
				     void *closure)
{
	return begin_marker (stack, obj, token, closure);
}

static uint32_t string_to_code (struct obj_bstr *s)
{
	return ct_utf16_to_unicode (s -> s, s -> size);
}

static enum s_status find_slot (struct psi_closure *pc,
				struct obj *fsobj,
				struct obj *lsobj,
				uint32_t **code)
{
	struct psi_ccmap *cmap;
	struct cmap_table *t;
	struct obj_bstr *fstr, *lstr;
	uint32_t *p;
	enum s_status res;

	res = get_cmap (pc, &cmap);
	if (res != S_CSC_OK)
		return res;

	fstr = &fsobj -> val.string;
	lstr = &lsobj -> val.string;
	
	if (!fstr -> size ||
	    fstr -> size != lstr -> size || 
	    obj_string_compare_s (fstr, lstr) > 0)
		return obj_error (fsobj, "Bad bounds");

	for (t = cmap -> tables; t != NULL; t = t -> next) {
		p = cmap_find_slot (t, fstr, lstr);
		if (p != NULL)
			break;
	}

	if (t == NULL)
		return obj_error (fsobj, "Mapping outside charmap");

	*code = p;

	return S_CSC_OK;
}

static enum s_status map_char (struct psi_closure *pc,
			       struct obj *sobj, uint32_t code)
{
	enum s_status res;
	uint32_t *dest;

	res = find_slot (pc, sobj, sobj, &dest);
	if (res != S_CSC_OK)
		return res;

	*dest = code;

	return S_CSC_OK;
}

static enum s_status tr_endbfchar (struct m_stack *stack,
				   struct obj *obj, int token,
				   void *closure)
{
	struct psi_closure *pc;
	struct obj sobj, dobj;
	struct psi_ccmap *cmap;
	uint32_t code;
	enum s_status res;

	pc = closure;
	res = get_cmap (pc, &cmap);
	if (res != S_CSC_OK)
		return res;

	while (1) {
		res = psi_safe_pop (stack, closure, 1, &dobj);
		if (res != S_CSC_OK)
			return res;

		if (dobj.type == OT_KEYWORD)
			break;

		if (dobj.type == OT_NAME)
			code = ct_name_to_unicode_s (&dobj.val.string);
		else if (dobj.type == OT_STRING)
			code = string_to_code (&dobj.val.string);
		else {
			res = obj_error (&dobj, "Expected string");
			obj_destroy (&dobj);
			return res;
		}
		obj_destroy (&dobj);

		res = psi_safe_pop (stack, closure, 1, &sobj);
		if (res != S_CSC_OK)
			return res;

		if (sobj.type != OT_STRING) {
			res = obj_error (&sobj, "Expected string");
			obj_destroy (&sobj);
			return res;
		}

		res = map_char (pc, &sobj, code);
		obj_destroy (&sobj);
		if (res != S_CSC_OK)
			return res;
	}

	if (dobj.val.keyword -> data != tr_beginbfchar)
		return obj_error (&dobj, "begin/end bfchar mismatch");

	return S_CSC_OK;
}

static enum s_status tr_beginbfrange (struct m_stack *stack,
				      struct obj *obj, int token,
				      void *closure)
{
	return begin_marker (stack, obj, token, closure);
}

static enum s_status map_chars (struct psi_closure *pc,
			        struct obj *fsobj, struct obj *lsobj,
				struct obj *dobj)
{
	uint32_t fcode, lcode, code, tcode;
	struct obj_array_node *an;
	enum s_status res;
	uint32_t *dest;
	int i;

	res = find_slot (pc, fsobj, lsobj, &dest);
	if (res != S_CSC_OK)
		return res;

	fcode = string_to_code (&fsobj -> val.string);
	lcode = string_to_code (&lsobj -> val.string);
	if (dobj -> type == OT_ARRAY) {
		an = dobj -> val.array.head;
		i = 0;
		for (code = fcode; code <= lcode; code++) {
			if (an == NULL)
				break;
			if (an -> obj.type == OT_NAME)
				tcode = ct_name_to_unicode_s 
					(&an -> obj.val.string);
			else if (an -> obj.type == OT_STRING)
				tcode = string_to_code (&an -> obj.val.string);
			else
				return psi_error (pc, "Bad bf specs");
			dest [i++] = tcode;
			an = an -> next;
		}
	} else {
		if (dobj -> type == OT_STRING)
			tcode = string_to_code (&dobj -> val.string);
		else
			tcode = dobj -> val.integer;
		i = 0;
		for (code = fcode; code <= lcode; code++)
			dest [i++] = tcode++;
	}	
			
	return S_CSC_OK;
}

static enum s_status map_unmapped_chars (struct psi_closure *pc,
					 struct obj *fsobj, struct obj *lsobj,
					 uint32_t code)
{
	uint32_t fcode, lcode;
	enum s_status res;
	uint32_t *dest;
	int i;

	res = find_slot (pc, fsobj, lsobj, &dest);
	if (res != S_CSC_OK)
		return res;

	fcode = string_to_code (&fsobj -> val.string);
	lcode = string_to_code (&lsobj -> val.string);
	i = 0;
	for (code = fcode; code <= lcode; code++, i++)
		if (!dest [i])
			dest [i] = code;
			
	return S_CSC_OK;
}

static enum s_status tr_endbfrange (struct m_stack *stack,
				    struct obj *obj, int token,
				    void *closure)
{
	struct psi_closure *pc;
	struct obj fsobj, lsobj, dobj;
	struct psi_ccmap *cmap;
	enum s_status res;

	pc = closure;
	res = get_cmap (pc, &cmap);
	if (res != S_CSC_OK)
		return res;

	while (1) {
		res = psi_safe_pop (stack, closure, 1, &dobj);
		if (res != S_CSC_OK)
			return res;

		if (dobj.type == OT_KEYWORD)
			break;

		if (obj_ptr_get (&dobj) -> type != OT_ARRAY &&
		    dobj.type != OT_STRING) {
			res = obj_error (&dobj, "Expected string/array");
			obj_destroy (&dobj);
			return res;
		}

		res = psi_safe_pop (stack, closure, 2, &lsobj, &fsobj);
		if (res != S_CSC_OK)
			return res;

		if (lsobj.type != OT_STRING) 
			res = obj_error (&lsobj, "Expected string");
		else if (fsobj.type != OT_STRING)
			res = obj_error (&fsobj, "Expected string");
		else
			res = S_CSC_OK;

		if (res != S_CSC_OK) {
			obj_destroy (&fsobj);
			obj_destroy (&lsobj);
			obj_destroy (&dobj);
			return res;
		}

		res = map_chars (pc, &fsobj, &lsobj, obj_ptr_get (&dobj));
		obj_destroy (&fsobj);
		obj_destroy (&lsobj);
		obj_destroy (&dobj);
		if (res != S_CSC_OK)
			return res;
	}

	if (dobj.val.keyword -> data != tr_beginbfrange)
		return obj_error (&dobj, "begin/end bfrange mismatch");

	return S_CSC_OK;
}

static enum s_status tr_begincidchar (struct m_stack *stack,
				      struct obj *obj, int token,
				      void *closure)
{
	return begin_marker (stack, obj, token, closure);
}

static enum s_status tr_endcidchar (struct m_stack *stack,
				    struct obj *obj, int token,
				    void *closure)
{
	struct psi_closure *pc;
	struct obj sobj, dobj;
	struct psi_ccmap *cmap;
	uint32_t code;
	enum s_status res;

	pc = closure;
	res = get_cmap (pc, &cmap);
	if (res != S_CSC_OK)
		return res;

	while (1) {
		res = psi_safe_pop (stack, closure, 1, &dobj);
		if (res != S_CSC_OK)
			return res;

		if (dobj.type == OT_KEYWORD)
			break;

		if (dobj.type == OT_NAME)
			code = ct_name_to_unicode_s (&dobj.val.string);
		else if (dobj.type == OT_INTEGER) {
			code = dobj.val.integer;
			obj_destroy (&dobj);
		} else {
			res = obj_error (&dobj, "Expected string");
			obj_destroy (&dobj);
			return res;
		}

		res = psi_safe_pop (stack, closure, 1, &sobj);
		if (res != S_CSC_OK)
			return res;

		if (sobj.type != OT_STRING) {
			res = obj_error (&sobj, "Expected string");
			obj_destroy (&sobj);
			return res;
		}

		res = map_char (pc, &sobj, code);
		obj_destroy (&sobj);
		if (res != S_CSC_OK)
			return res;
	}

	if (dobj.val.keyword -> data != tr_begincidchar)
		return obj_error (&dobj, "begin/end cidchar mismatch");

	return S_CSC_OK;
}

static enum s_status tr_begincidrange (struct m_stack *stack,
				       struct obj *obj, int token,
				       void *closure)
{
	return begin_marker (stack, obj, token, closure);
}

static enum s_status tr_endcidrange (struct m_stack *stack,
				     struct obj *obj, int token,
				     void *closure)
{
	struct psi_closure *pc;
	struct obj fsobj, lsobj, dobj;
	struct psi_ccmap *cmap;
	enum s_status res;

	pc = closure;
	res = get_cmap (pc, &cmap);
	if (res != S_CSC_OK)
		return res;

	while (1) {
		res = psi_safe_pop (stack, closure, 1, &dobj);
		if (res != S_CSC_OK)
			return res;

		if (dobj.type == OT_KEYWORD)
			break;

		if (obj_ptr_get (&dobj) -> type != OT_ARRAY &&
		    dobj.type != OT_INTEGER) {
			res = obj_error (&dobj, "Expected integer/array");
			obj_destroy (&dobj);
			return res;
		}

		res = psi_safe_pop (stack, closure, 2, &lsobj, &fsobj);
		if (res != S_CSC_OK)
			return res;

		if (fsobj.type != OT_STRING)
			res = obj_error (&lsobj, "Expected string");
		else if (lsobj.type != OT_STRING)
			res = obj_error (&lsobj, "Expected string");
		else
			res = S_CSC_OK;
		if (res != S_CSC_OK) {
			obj_destroy (&fsobj);
			obj_destroy (&lsobj);
			obj_destroy (&dobj);
			return res;
		}

		res = map_chars (pc, &fsobj, &lsobj, obj_ptr_get (&dobj));
		obj_destroy (&fsobj);
		obj_destroy (&lsobj);
		obj_destroy (&dobj);
		if (res != S_CSC_OK)
			return res;
	}

	if (dobj.val.keyword -> data != tr_begincidrange)
		return obj_error (&dobj, "begin/end cidrange mismatch");

	return S_CSC_OK;
}

static enum s_status tr_beginnotdefrange (struct m_stack *stack,
					  struct obj *obj, int token,
					  void *closure)
{
	return begin_marker (stack, obj, token, closure);
}

static enum s_status tr_endnotdefrange (struct m_stack *stack,
					struct obj *obj, int token,
					void *closure)
{
	struct psi_closure *pc;
	struct obj fsobj, lsobj, dobj;
	uint32_t code;
	struct psi_ccmap *cmap;
	enum s_status res;

	pc = closure;
	res = get_cmap (pc, &cmap);
	if (res != S_CSC_OK)
		return res;

	while (1) {
		res = psi_safe_pop (stack, closure, 1, &dobj);
		if (res != S_CSC_OK)
			return res;

		if (dobj.type == OT_KEYWORD)
			break;

		if (dobj.type != OT_INTEGER) {
			res = obj_error (&dobj, "Expected integer");
			obj_destroy (&dobj);
			return res;
		}

		code = dobj.val.integer;

		res = psi_safe_pop (stack, closure, 2, &lsobj, &fsobj);
		if (res != S_CSC_OK)
			return res;

		if (lsobj.type != OT_STRING)
			res = obj_error (&lsobj, "Expected string");
		else if (fsobj.type != OT_STRING)
			res = obj_error (&fsobj, "Expected string");
		else
			res = S_CSC_OK;

		if (res != S_CSC_OK) {
			obj_destroy (&lsobj);
			obj_destroy (&fsobj);
			return res;
		}

		res = map_unmapped_chars (pc, &fsobj, &lsobj, code);
		obj_destroy (&fsobj);
		obj_destroy (&lsobj);
		if (res != S_CSC_OK)
			return res;
	}

	if (dobj.val.keyword -> data != tr_beginnotdefrange)
		return obj_error (&dobj, "begin/end notdef mismatch");

	return S_CSC_OK;
}

enum s_status psi_cid_def_cmap (struct m_stack *stack,
				struct psi_closure *closure)
{
	struct psi_closure *pc;
	struct psi_ccmap *cmap;
	struct psi_interpreter *pi;
	struct obj iobj, dobj, kobj;
	enum s_status res;

	pc = closure;
	pi = pc -> interpreter;
	res = get_cmap (pc, &cmap);
	if (res != S_CSC_OK)
		return res;

	res = psi_safe_pop (stack, closure, 2, &iobj, &kobj);
	if (res != S_CSC_OK)
		return res;

	res = obj_ptr_deref (&iobj, &dobj);
	if (res != S_CSC_OK) {
		obj_destroy (&iobj);
		obj_destroy (&kobj);
		return res;
	}

	res = m_push (stack, &iobj, TOK_DICTIONARY);
	if (res != S_CSC_OK) {
		obj_destroy (&iobj);
		obj_destroy (&dobj);
		obj_destroy (&kobj);
		return res;
	}

	res = T_CALL (cmap_init, &cmap -> cmap, NULL, NULL, 
		      obj_ptr_get (&dobj), &cmap -> owner);
	obj_destroy (&dobj);
	obj_destroy (&kobj);
	if (res != S_CSC_OK)
		return res;

	pi -> ccmap = cmap -> next;

	cmap -> next = pi -> resources -> ccmaps;
	pi -> resources -> ccmaps = cmap;

	return S_CSC_OK;
}

static void destroy_cmap (struct psi_ccmap *cmap)
{
	struct cmap_table *table;

	while (cmap -> tables != NULL) {
		table = cmap -> tables;
		cmap -> tables = table -> next;

		cmap_destroy_table (table);
	}

	obj_destroy (&cmap -> dict);
	cmap_unref (cmap -> cmap, &cmap -> owner);

	free (cmap);
}

void psi_cid_destroy_cmap (struct psi_interpreter *pi)
{
	struct psi_ccmap *cmap;

	while (pi -> ccmap != NULL) {
		cmap = pi -> ccmap;
		pi -> ccmap = cmap -> next;
		destroy_cmap (cmap);
	}
}

void psi_cid_destroy_cmap_res (struct psi_resources *res)
{
	struct psi_ccmap *cmap;

	while (res -> ccmaps != NULL) {
		cmap = res -> ccmaps;
		res -> ccmaps = cmap -> next;
		destroy_cmap (cmap);
	}

	psi_destroy_file_resources (&res -> cmapres);
}

static enum s_status test_cmap (const char *fname, char **name)
{
	struct f_file *file;
	enum s_status res;
	char *line, *p, *q;

	res = f_open (fname, &file);
	if (res != S_CSC_OK)
		return S_CSC_OK;

	if (memcmp (file -> pdf, CMAP_SIGNATURE, strlen (CMAP_SIGNATURE))) {
		f_unref (file);
		return S_CSC_OK;
	}
	
	res = pl_get_header (file, CMAP_BEGINRESOURCE, &line);
	f_unref (file);
	if (res != S_CSC_OK || line == NULL)
		return res;

	p = strchr (line, '(');
	q = strrchr (line, ')');
	if (strncmp (line, "CMap", 4) ||
	    p == NULL || q == NULL || q < p) {
		free (line);
		return S_CSC_OK;
	}

	p++;
	*name = malloc (q - p + 1);
	if (*name == NULL) {
		free (line);
		return S_CSC_OUT_OF_MEMORY;
	}

	memcpy (*name, p, q - p);
	(*name) [q - p] = 0;
	free (line);

	return S_CSC_OK;
}

#if defined(_WIN32) || defined(WIN32)
static int readdir_r (DIR *dirp, struct dirent *entry, struct dirent **result)
{
	*result = readdir (dirp);
	if (*result == NULL)
		*entry = **result;

	return 0;
}
#endif

enum s_status psi_cid_add_cmap_dir (struct psi_resources *rsrc,
				    const char *path)
{
	struct dirent entry, *re;
	enum s_status res;
	char *fname, *name;
	DIR *dir;

	dir = opendir (path);
	if (dir == NULL)
		return S_CSC_OK;

	res = S_CSC_OK;
	while (!readdir_r (dir, &entry, &re) && re != NULL) {
		fname = malloc (strlen (path) + 1 + strlen (entry.d_name) + 1);
		if (fname == NULL)
			return S_CSC_OUT_OF_MEMORY;
		strcpy (fname, path);
		strcat (fname, "/");
		strcat (fname, entry.d_name);

		name = NULL;
		res = test_cmap (fname, &name);
		if (res != S_CSC_OK) {
			free (fname);
			break;
		}

		if (name != NULL) {
			res = psi_add_file_resource (&rsrc -> cmapres, 
						     name, fname);
			free (fname);
			free (name);
			if (res != S_CSC_OK)
				break;
		} else
			free (fname);
	}
	closedir (dir);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}
