/*
 * PDFlow -- core.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "psi_private.h"

static enum s_status tr_and (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure);
static enum s_status tr_bitshift (struct m_stack *stack,
				  struct obj *obj, int token,
				  void *closure);
static enum s_status tr_eq (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);
static enum s_status tr_ge (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);
static enum s_status tr_gt (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);
static enum s_status tr_le (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);
static enum s_status tr_lt (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);
static enum s_status tr_ne (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);
static enum s_status tr_not (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure);
static enum s_status tr_or (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);

static enum s_status tr_xor (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure);
static enum s_status tr_if (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);
static enum s_status tr_ifelse (struct m_stack *stack,
				struct obj *obj, int token,
				void *closure);
static enum s_status tr_copy (struct m_stack *stack,
			      struct obj *obj, int token,
			      void *closure);
static enum s_status tr_dup (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure);
static enum s_status tr_exch (struct m_stack *stack,
			      struct obj *obj, int token,
			      void *closure);
static enum s_status tr_index (struct m_stack *stack,
			       struct obj *obj, int token,
			       void *closure);
static enum s_status tr_pop (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure);
static enum s_status tr_roll (struct m_stack *stack,
			      struct obj *obj, int token,
			      void *closure);

static enum s_status tr_begin (struct m_stack *stack,
			       struct obj *obj, int token,
			       void *closure);
static enum s_status tr_end (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure);
static enum s_status tr_currentdict (struct m_stack *stack,
				     struct obj *obj, int token,
				     void *closure);
static enum s_status tr_dict (struct m_stack *stack,
			      struct obj *obj, int token,
			      void *closure);
static enum s_status tr_def (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure);
static enum s_status tr_findresource (struct m_stack *stack,
				      struct obj *obj, int token,
				      void *closure);
static enum s_status tr_defineresource (struct m_stack *stack,
					struct obj *obj, int token,
					void *closure);

static enum s_status tr_res_find_procset (struct m_stack *stack,
					 struct psi_closure *closure);

static struct obj_keyword keywords [] = {
	{ "and", tr_and },
	{ "bitshift", tr_bitshift },
	{ "eq", tr_eq },
	{ "ge", tr_ge },
	{ "gt", tr_gt },
	{ "le", tr_le },
	{ "lt", tr_lt },
	{ "ne", tr_ne },
	{ "not", tr_not },
	{ "or", tr_or },
	{ "xor", tr_xor },
	{ "if", tr_if },
	{ "ifelse", tr_ifelse },
	{ "copy", tr_copy },
	{ "dup", tr_dup },
	{ "exch", tr_exch },
	{ "index", tr_index },
	{ "pop", tr_pop },
	{ "roll", tr_roll },

	{ "begin", tr_begin },
	{ "end", tr_end },
	{ "currentdict", tr_currentdict },
	{ "dict", tr_dict },
	{ "def", tr_def },
	{ "findresource", tr_findresource },
	{ "defineresource", tr_defineresource },
	{ NULL }
};

struct psi_procset psi_core_procset = {
	NULL, keywords
};

static struct {
	const char *name;
	enum s_status (*find) (struct m_stack *stack,
			       struct psi_closure *closure);
	enum s_status (*define) (struct m_stack *stack,
				 struct psi_closure *closure);

} resources [] = {
	{ "ProcSet", tr_res_find_procset, NULL },
	{ "CMap", NULL, psi_cid_def_cmap },
	{ NULL }
};

static struct psi_procset *procsets [] = {
	&psi_cmap_procset,
	NULL
};

static void set_log_integer (struct obj *obj, struct psi_closure *pc,
			     pdf_integer_t n)
{
	obj_set_integer (obj, &pc -> ps.ptr, n);
}

static void set_boolean (struct obj *obj, struct psi_closure *pc,
			 pdf_bool_t b)
{
	obj_set_boolean (obj, &pc -> ps.ptr, b);
}

static enum s_status tr_and (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure)
{
	struct obj n1, n2, n3;
	enum s_status res;

	res = psi_safe_pop (stack, closure, 2, &n2, &n1);
	if (res != S_CSC_OK)
		return res;

	if (n1.type == OT_INTEGER && n2.type == OT_INTEGER) {
		set_log_integer (&n3, closure, 
				 n1.val.integer & n2.val.integer);
		return m_push (stack, &n3, TOK_INTEGER);
	}

	if (n1.type == OT_BOOLEAN && n2.type == OT_BOOLEAN) {
		set_boolean (&n3, closure, 
			     n1.val.bool && n2.val.bool);
		return m_push (stack, &n3, TOK_BOOLEAN);
	}

	obj_destroy (&n1);
	obj_destroy (&n2);

	return psi_error (closure, "Bad operand types");
}

static enum s_status tr_bitshift (struct m_stack *stack,
				  struct obj *obj, int token,
				  void *closure)
{
	struct obj n1, n2, n3;
	pdf_unsigned_t u1;
	enum s_status res;

	res = psi_safe_pop (stack, closure, 2, &n2, &n1);
	if (res != S_CSC_OK)
		return res;

	if (n1.type != OT_INTEGER || n2.type != OT_INTEGER) {
		obj_destroy (&n1);
		obj_destroy (&n2);
		return psi_error (closure, "Bad operand types");
	}

	u1 = n1.val.integer;

	if (n2.val.integer > 0)
		set_log_integer (&n3, closure, u1 << n2.val.integer);
	else
		set_log_integer (&n3, closure, u1 >> -n2.val.integer);

	return m_push (stack, &n3, TOK_INTEGER);
}

static enum s_status tr_eq (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct obj n1, n2, n3;
	enum s_status res;

	res = psi_safe_pop (stack, closure, 2, &n2, &n1);
	if (res != S_CSC_OK)
		return res;

	if (n1.type == OT_BOOLEAN && n2.type == OT_BOOLEAN) {
		set_boolean (&n3, closure, n1.val.bool == n2.val.bool);
		return m_push (stack, &n3, TOK_BOOLEAN);
	}

	if (n1.type == OT_INTEGER && n2.type == OT_INTEGER) {
		set_boolean (&n3, closure, n1.val.bool == n2.val.bool);
		return m_push (stack, &n3, TOK_BOOLEAN);
	}

	if (n1.type == OT_ALIVE && n2.type == OT_ALIVE) {
		set_boolean (&n3, closure, m_compare_blocks (&n1, &n2));
		return m_push (stack, &n3, TOK_BOOLEAN);
	}

	if (!obj_cast (&n1, OT_REAL) || !obj_cast (&n2, OT_REAL)) {
		set_boolean (&n3, closure, 0);
		return m_push (stack, &n3, TOK_BOOLEAN);
	}

	set_boolean (&n3, closure, n1.val.real == n2.val.real);

	return m_push (stack, &n3, TOK_BOOLEAN);
}

static enum s_status tr_ge (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct obj n1, n2, n3;
	enum s_status res;

	res = psi_safe_pop (stack, closure, 2, &n2, &n1);
	if (res != S_CSC_OK)
		return res;

	if (n1.type == OT_INTEGER && n2.type == OT_INTEGER) {
		set_boolean (&n3, closure, n1.val.bool >= n2.val.bool);
		return m_push (stack, &n3, TOK_BOOLEAN);
	}

	if (!obj_cast (&n1, OT_REAL) || !obj_cast (&n2, OT_REAL)) {
		set_boolean (&n3, closure, 0);
		return m_push (stack, &n3, TOK_BOOLEAN);
	}

	set_boolean (&n3, closure, n1.val.real >= n2.val.real);

	return m_push (stack, &n3, TOK_BOOLEAN);
}

static enum s_status tr_gt (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct obj n1, n2, n3;
	enum s_status res;

	res = psi_safe_pop (stack, closure, 2, &n2, &n1);
	if (res != S_CSC_OK)
		return res;

	if (n1.type == OT_INTEGER && n2.type == OT_INTEGER) {
		set_boolean (&n3, closure, n1.val.bool > n2.val.bool);
		return m_push (stack, &n3, TOK_BOOLEAN);
	}

	if (!obj_cast (&n1, OT_REAL) || !obj_cast (&n2, OT_REAL)) {
		set_boolean (&n3, closure, 0);
		return m_push (stack, &n3, TOK_BOOLEAN);
	}

	set_boolean (&n3, closure, n1.val.real > n2.val.real);

	return m_push (stack, &n3, TOK_BOOLEAN);
}

static enum s_status tr_le (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct obj n1, n2, n3;
	enum s_status res;

	res = psi_safe_pop (stack, closure, 2, &n2, &n1);
	if (res != S_CSC_OK)
		return res;

	if (n1.type == OT_INTEGER && n2.type == OT_INTEGER) {
		set_boolean (&n3, closure, n1.val.bool <= n2.val.bool);
		return m_push (stack, &n3, TOK_BOOLEAN);
	}

	if (!obj_cast (&n1, OT_REAL) || !obj_cast (&n2, OT_REAL)) {
		set_boolean (&n3, closure, 0);
		return m_push (stack, &n3, TOK_BOOLEAN);
	}

	set_boolean (&n3, closure, n1.val.real <= n2.val.real);

	return m_push (stack, &n3, TOK_BOOLEAN);
}

static enum s_status tr_lt (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct obj n1, n2, n3;
	enum s_status res;

	res = psi_safe_pop (stack, closure, 2, &n2, &n1);
	if (res != S_CSC_OK)
		return res;

	if (n1.type == OT_INTEGER && n2.type == OT_INTEGER) {
		set_boolean (&n3, closure, n1.val.bool < n2.val.bool);
		return m_push (stack, &n3, TOK_BOOLEAN);
	}

	if (!obj_cast (&n1, OT_REAL) || !obj_cast (&n2, OT_REAL)) {
		set_boolean (&n3, closure, 0);
		return m_push (stack, &n3, TOK_BOOLEAN);
	}

	set_boolean (&n3, closure, n1.val.real < n2.val.real);

	return m_push (stack, &n3, TOK_BOOLEAN);
}

static enum s_status tr_ne (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct obj n1, n2, n3;
	enum s_status res;

	res = psi_safe_pop (stack, closure, 2, &n2, &n1);
	if (res != S_CSC_OK)
		return res;

	if (n1.type == OT_BOOLEAN && n2.type == OT_BOOLEAN) {
		set_boolean (&n3, closure, n1.val.bool != n2.val.bool);
		return m_push (stack, &n3, TOK_BOOLEAN);
	}

	if (n1.type == OT_INTEGER && n2.type == OT_INTEGER) {
		set_boolean (&n3, closure, n1.val.bool != n2.val.bool);
		return m_push (stack, &n3, TOK_BOOLEAN);
	}

	if (n1.type == OT_ALIVE && n2.type == OT_ALIVE) {
		set_boolean (&n3, closure, !m_compare_blocks (&n1, &n2));
		return m_push (stack, &n3, TOK_BOOLEAN);
	}

	if (!obj_cast (&n1, OT_REAL) || !obj_cast (&n2, OT_REAL)) {
		set_boolean (&n3, closure, 1);
		return m_push (stack, &n3, TOK_BOOLEAN);
	}

	set_boolean (&n3, closure, n1.val.real != n2.val.real);

	return m_push (stack, &n3, TOK_BOOLEAN);
}

static enum s_status tr_not (struct m_stack *stack,
			     struct obj *dobj, int token,
			     void *closure)
{
	struct obj obj;
	enum s_status res;

	res = psi_safe_pop (stack, closure, 1, &obj);
	if (res != S_CSC_OK)
		return res;

	if (obj.type == OT_INTEGER) {
		set_log_integer (&obj, closure, ~obj.val.integer);
		return m_push (stack, &obj, TOK_INTEGER);
	}

	if (obj.type == OT_BOOLEAN) {
		set_boolean (&obj, closure, !obj.val.bool);
		return m_push (stack, &obj, TOK_BOOLEAN);
	}

	obj_destroy (&obj);

	return psi_error (closure, "Bad operand type");
}

static enum s_status tr_or (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct obj n1, n2, n3;
	enum s_status res;

	res = psi_safe_pop (stack, closure, 2, &n2, &n1);
	if (res != S_CSC_OK)
		return res;

	if (n1.type == OT_INTEGER && n2.type == OT_INTEGER) {
		set_log_integer (&n3, closure, 
				 n1.val.integer | n2.val.integer);
		return m_push (stack, &n3, TOK_INTEGER);
	}

	if (n1.type == OT_BOOLEAN && n2.type == OT_BOOLEAN) {
		set_boolean (&n3, closure, 
			     n1.val.bool || n2.val.bool);
		return m_push (stack, &n3, TOK_BOOLEAN);
	}

	obj_destroy (&n1);
	obj_destroy (&n2);

	return psi_error (closure, "Bad operand types");
}

static enum s_status tr_xor (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure)
{
	struct obj n1, n2, n3;
	enum s_status res;

	res = psi_safe_pop (stack, closure, 2, &n2, &n1);
	if (res != S_CSC_OK)
		return res;

	if (n1.type == OT_INTEGER && n2.type == OT_INTEGER) {
		set_log_integer (&n3, closure, 
				 n1.val.integer ^ n2.val.integer);
		return m_push (stack, &n3, TOK_INTEGER);
	}

	if (n1.type == OT_BOOLEAN && n2.type == OT_BOOLEAN) {
		set_boolean (&n3, closure, 
			     n1.val.bool ^ n2.val.bool);
		return m_push (stack, &n3, TOK_BOOLEAN);
	}

	obj_destroy (&n1);
	obj_destroy (&n2);

	return psi_error (closure, "Bad operand types");
}

static enum s_status tr_if (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct obj cond, block;
	struct psi_closure *pc;
	enum s_status res;

	res = psi_safe_pop (stack, closure, 2, &block, &cond);
	if (res != S_CSC_OK)
		return res;

	pc = closure;

	if (block.type != OT_ALIVE || cond.type != OT_BOOLEAN) {
		obj_destroy (&cond);
		obj_destroy (&block);
		return psi_error (closure, "Bad operand types");
	}

	if (cond.val.bool)
		res = m_eval_block (pc -> interpreter -> state, &block, pc);
	else
		res = S_CSC_OK;
		
	obj_destroy (&block);
	return res;
}

static enum s_status tr_ifelse (struct m_stack *stack,
				struct obj *obj, int token,
				void *closure)
{
	struct obj cond, block1, block2;
	struct psi_closure *pc;
	enum s_status res;

	res = psi_safe_pop (stack, closure, 3, &block2, &block1, &cond);
	if (res != S_CSC_OK)
		return res;

	pc = closure;

	if (block1.type != OT_ALIVE || 
	    block2.type != OT_ALIVE || 
	    cond.type != OT_BOOLEAN) {
		obj_destroy (&cond);
		obj_destroy (&block1);
		obj_destroy (&block2);
		return psi_error (closure, "Bad operand types");
	}

	if (cond.val.bool)
		res = m_eval_block (pc -> interpreter -> state, &block1, pc);
	else
		res = m_eval_block (pc -> interpreter -> state, &block2, pc);
		
	obj_destroy (&block1);
	obj_destroy (&block2);

	return res;
}

static enum s_status tr_copy (struct m_stack *stack,
			      struct obj *obj, int token,
			      void *closure)
{
	enum s_status res;
	struct obj nobj;

	res = psi_safe_pop (stack, closure, 1, &nobj);
	if (nobj.type != OT_INTEGER) {
		obj_destroy (&nobj);
		return obj_error (&nobj, "Integer expected");
	}

	if (nobj.val.integer < 0)
		return obj_error (&nobj, "Nonnegative integer expected");

	if (!nobj.val.integer)
		return S_CSC_OK;

	return m_copy (stack, obj, nobj.val.integer);
}

static enum s_status tr_dup (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure)
{
	return m_copy (stack, obj, 1);
}

static enum s_status tr_exch (struct m_stack *stack,
			      struct obj *obj, int token,
			      void *closure)
{
	return m_exch (stack, obj);
}

static enum s_status tr_index (struct m_stack *stack,
			       struct obj *obj, int token,
			       void *closure)
{
	enum s_status res;
	struct obj nobj;

	res = psi_safe_pop (stack, closure, 1, &nobj);
	if (nobj.type != OT_INTEGER) {
		obj_destroy (&nobj);
		return obj_error (&nobj, "Integer expected");
	}

	if (nobj.val.integer < 0)
		return obj_error (&nobj, "Nonnegative integer expected");

	return m_index (stack, obj, nobj.val.integer);
}

static enum s_status tr_pop (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure)
{
	enum s_status res;
	struct obj nobj;

	res = psi_safe_pop (stack, closure, 1, &nobj);
	if (res != S_CSC_OK)
		return res;

	obj_destroy (&nobj);

	return S_CSC_OK;
}


static enum s_status tr_roll (struct m_stack *stack,
			      struct obj *obj, int token,
			      void *closure)
{
	enum s_status res;
	struct obj nobj, jobj;

	res = psi_safe_pop (stack, closure, 2, &nobj, &jobj);
	if (nobj.type != OT_INTEGER || jobj.type != OT_INTEGER) {
		obj_destroy (&nobj);
		obj_destroy (&jobj);
		return obj_error (obj, "Integer expected");
	}

	if (!nobj.val.integer || !jobj.val.integer)
		return S_CSC_OK;

	return m_roll (stack, obj, nobj.val.integer, jobj.val.integer);
}

static enum s_status tr_res_find_procset (struct m_stack *stack,
					  struct psi_closure *pc)
{
	enum s_status res;
	struct obj obj;
	int i;

	res = psi_safe_pop (stack, pc, 1, &obj);
	if (res != S_CSC_OK)
		return res;

	if (obj.type != OT_NAME) {
		res = obj_error (&obj, "Expected procset name");
		obj_destroy (&obj);
		return res;
	}

	for (i = 0; procsets [i] != NULL; i++) {
		if (obj_string_equal (&obj.val.string, procsets [i] -> name))
			break;
	}

	obj_destroy (&obj);
	if (procsets [i] == NULL)
		return psi_error (pc, "Unsupported procset");

	res = psi_add_procset (pc -> interpreter, procsets [i]);
	if (res != S_CSC_OK)
		return res;

	if (procsets [i] -> init != NULL) {
		res = procsets [i] -> init (stack, pc);
		if (res != S_CSC_OK)
			return res;
	}

	return S_CSC_OK;
}


static enum s_status tr_begin (struct m_stack *stack,
			       struct obj *obj, int token,
			       void *closure)
{
	struct psi_interpreter *pi;
	struct psi_closure *pc;
	struct obj dobj;
	enum s_status res;

	pc = closure;
	pi = pc -> interpreter;

	res = psi_safe_pop (stack, pc, 1, &dobj);
	if (res != S_CSC_OK)
		return res;

	if (!obj_can_cast (obj_ptr_get (&dobj), OT_DICTIONARY)) {
		res = obj_error (&dobj, "Not a dictionary");
		obj_destroy (&dobj);
		return res;
	}

	res = psi_user_push (pi, &dobj);
	obj_destroy (&dobj);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status tr_end (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure)
{
	struct psi_interpreter *pi;
	struct psi_closure *pc;

	pc = closure;
	pi = pc -> interpreter;

	if (!psi_user_pop (pi))
		return obj_error (obj, "User dict stack underflow");

	return S_CSC_OK;
}

static enum s_status tr_currentdict (struct m_stack *stack,
				     struct obj *obj, int token,
				     void *closure)
{
	struct psi_interpreter *pi;
	struct psi_closure *pc;
	struct obj *dobj, nobj;
	enum s_status res;

	pc = closure;
	pi = pc -> interpreter;

	if (!psi_user_top (pi, &dobj))
		return obj_error (obj, "User dict stack underflow");

	res = obj_dup (dobj, &nobj);
	if (res != S_CSC_OK)
		return res;

	res = m_push (stack, &nobj, TOK_DICTIONARY);
	if (res != S_CSC_OK) {
		obj_destroy (&nobj);
		return res;
	}

	return S_CSC_OK;
}

static enum s_status tr_dict (struct m_stack *stack,
			      struct obj *obj, int token,
			      void *closure)
{
	struct psi_interpreter *pi;
	struct psi_closure *pc;
	struct obj dobj, nobj;
	enum s_status res;

	pc = closure;
	pi = pc -> interpreter;

	res = psi_safe_pop (stack, pc, 1, &dobj);
	if (res != S_CSC_OK)
		return res;

	if (!obj_can_cast (obj_ptr_get (&dobj), OT_INTEGER)) {
		res = obj_error (&dobj, "Not an integer");
		obj_destroy (&dobj);
		return res;
	}

	obj_set_null (&dobj, &f_null);

	res = obj_ptr_dup (&dobj, &nobj);
	if (res != S_CSC_OK)
		return res;

	res = psi_push (stack, &nobj, TOK_DICTIONARY);
	obj_destroy (&nobj);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
	
}

static enum s_status tr_def (struct m_stack *stack,
			     struct obj *nobj, int token,
			     void *closure)
{
	struct psi_dict_stack_node *us;
	struct psi_interpreter *pi;
	struct psi_closure *pc;
	struct obj obj, kobj;
	enum s_status res;

	pc = closure;
	pi = pc -> interpreter;

	res = psi_safe_pop (stack, pc, 1, &obj);
	if (res != S_CSC_OK)
		return res;

	res = psi_safe_pop (stack, pc, 1, &kobj);
	if (res != S_CSC_OK) {
		obj_destroy (&obj);
		return res;
	}

	if (kobj.type != OT_NAME) {
		res = obj_error (&kobj, "Name expected");
		obj_destroy (&kobj);
		obj_destroy (&obj);
		return res;
	}

	us = pi -> user_stack;
	res = obj_add_dict_s (obj_get_rw_dictionary (&us -> obj), &kobj, &obj);
	obj_destroy (&kobj);
	obj_destroy (&obj);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status tr_defineresource (struct m_stack *stack,
					struct obj *nobj, int token,
					void *closure)
{
	enum s_status res;
	struct obj obj;
	int i;

	res = psi_safe_pop (stack, closure, 1, &obj);
	if (res != S_CSC_OK)
		return res;

	if (obj.type != OT_NAME) {
		res = obj_error (&obj, "Expected resource name");
		obj_destroy (&obj);
		return res;
	}

	for (i = 0; resources [i].name != NULL; i++) {
		if (obj_string_equal (&obj.val.string, resources [i].name))
			break;
	}

	obj_destroy (&obj);
	if (resources [i].name == NULL || resources [i].define == NULL)
		return psi_error (closure, "Unsupported resource");

	res = resources [i].define (stack, closure);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}


static enum s_status tr_findresource (struct m_stack *stack,
				      struct obj *nobj, int token,
				      void *closure)
{
	enum s_status res;
	struct obj obj;
	int i;

	res = psi_safe_pop (stack, closure, 1, &obj);
	if (res != S_CSC_OK)
		return res;

	if (obj.type != OT_NAME) {
		res = obj_error (&obj, "Expected resource name");
		obj_destroy (&obj);
		return res;
	}

	for (i = 0; resources [i].name != NULL; i++) {
		if (obj_string_equal (&obj.val.string, resources [i].name))
			break;
	}

	obj_destroy (&obj);
	if (resources [i].name == NULL || resources [i].find == NULL)
		return psi_error (closure, "Unsupported resource");

	res = resources [i].find (stack, closure);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

