/*
 * PDFlow -- math.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#define _GNU_SOURCE

#include <math.h>
#include <stdlib.h>

#include "ps/psi.h"
#include "psi_private.h"

#ifndef M_PIl
#define M_PIl M_PI
#endif

static enum s_status tr_abs (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure);
static enum s_status tr_add (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure);
static enum s_status tr_atan (struct m_stack *stack,
			      struct obj *obj, int token,
			      void *closure);
static enum s_status tr_ceiling (struct m_stack *stack,
				 struct obj *obj, int token,
				 void *closure);
static enum s_status tr_cos (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure);
static enum s_status tr_cvi (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure);
static enum s_status tr_cvr (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure);
static enum s_status tr_div (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure);
static enum s_status tr_exp (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure);
static enum s_status tr_floor (struct m_stack *stack,
			       struct obj *obj, int token,
			       void *closure);
static enum s_status tr_idiv (struct m_stack *stack,
			      struct obj *obj, int token,
			      void *closure);
static enum s_status tr_ln (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);
static enum s_status tr_log (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure);
static enum s_status tr_mod (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure);
static enum s_status tr_mul (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure);
static enum s_status tr_neg (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure);
static enum s_status tr_round (struct m_stack *stack,
			       struct obj *obj, int token,
			       void *closure);
static enum s_status tr_sin (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure);
static enum s_status tr_sub (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure);
static enum s_status tr_truncate (struct m_stack *stack,
				  struct obj *obj, int token,
				  void *closure);

static struct obj_keyword keywords [] = {
	{ "abs", tr_abs },
	{ "add", tr_add },
	{ "atan", tr_atan },
	{ "ceiling", tr_ceiling },
	{ "cos", tr_cos },
	{ "cvi", tr_cvi },
	{ "cvr", tr_cvr },
	{ "div", tr_div },
	{ "exp", tr_exp },
	{ "floor", tr_floor },
	{ "idiv", tr_idiv },
	{ "ln", tr_ln },
	{ "log", tr_log },
	{ "mod", tr_mod },
	{ "mul", tr_mul },
	{ "neg", tr_neg },
	{ "round", tr_round },
	{ "sin", tr_sin },
	{ "sub", tr_sub },
	{ "truncate", tr_truncate },
	{ NULL }
};

struct psi_procset psi_math_procset = {
	NULL, keywords
};

static void set_integer (struct obj *obj, struct psi_closure *pc,
			 pdf_real_t n)
{
	pdf_integer_t i;

	i = (pdf_integer_t) n;
	if (i == n)
		obj_set_integer (obj, &pc -> ps.ptr, i);
	else
		obj_set_real (obj, &pc -> ps.ptr, n);
}

static void set_real (struct obj *obj, struct psi_closure *pc,
		      pdf_real_t n)
{
	obj_set_real (obj, &pc -> ps.ptr, n);
}

static enum s_status tr_abs (struct m_stack *stack,
			     struct obj *tobj, int token,
			     void *closure)
{
	struct obj obj;
	enum s_status res;
	int tt;

	res = psi_safe_pop (stack, closure, 1, &obj);
	if (res != S_CSC_OK)
		return res;

	switch (obj.type) {
	case OT_INTEGER:
		tt = TOK_INTEGER;
		if (obj.val.integer >= 0)	
			break;

		set_integer (&obj, closure, -(pdf_real_t) obj.val.integer);

		tt = obj.type == OT_INTEGER ? TOK_INTEGER : TOK_REAL;
		break;

	case OT_REAL:
		tt = TOK_REAL;
		if (obj.val.real < 0)
			set_real (&obj, closure, -obj.val.real);
		break;

	default:
		return obj_error (&obj, "Bad operand type");
	}

	return m_push (stack, &obj, tt);
}

static enum s_status tr_add (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure)
{
	struct obj n1, n2, n3;
	enum s_status res;

	res = psi_safe_pop (stack, closure, 2, &n2, &n1);
	if (res != S_CSC_OK)
		return res;

	if (n1.type == OT_INTEGER && n2.type == OT_INTEGER) {
		set_integer (&n3, closure,
			     ((pdf_real_t) n1.val.integer) + 
			     n2.val.integer);
		return m_push (stack, &n3, TOK_INTEGER);
	}

	if (!obj_cast (&n1, OT_REAL) || !obj_cast (&n2, OT_REAL)) {
		obj_destroy (&n1);
		obj_destroy (&n2);
		return psi_error (closure, "Bad operand types");
	}

	set_real (&n3, closure, n1.val.real + n2.val.real);

	return m_push (stack, &n3, TOK_REAL);
}

static enum s_status tr_atan (struct m_stack *stack,
			      struct obj *obj, int token,
			      void *closure)
{
	struct obj num, den, deg;
	enum s_status res;

	res = psi_safe_pop (stack, closure, 2, &den, &num);
	if (res != S_CSC_OK)
		return res;
	if (!obj_cast (&num, OT_REAL) ||
	    !obj_cast (&den, OT_REAL)) {
		obj_destroy (&num);
		obj_destroy (&den);
		return psi_error (closure, "Bad operand type");
	}

	if (num.val.real == 0 && den.val.real == 0) 
		return psi_error (closure, "atan(0/0) requested");

	if (den.val.real != 0) {
		set_real (&deg, closure, atan (num.val.real / den.val.real));
		deg.val.real *= 180 / M_PIl;
	} else if (num.val.real > 0)
		set_real (&deg, closure, 90);
	else
		set_real (&deg, closure, -90);

	if (den.val.real < 0) {
		if (deg.val.real > 0)
			deg.val.real = deg.val.real + 180;
		else
			deg.val.real = deg.val.real - 180;
	}

	if (deg.val.real < 0)
		deg.val.real += 360;

	return m_push (stack, &deg, TOK_REAL);
}

static enum s_status tr_ceiling (struct m_stack *stack,
				 struct obj *dobj, int token,
				 void *closure)
{
	struct obj obj;
	enum s_status res;
	int tt;

	res = psi_safe_pop (stack, closure, 1, &obj);
	if (res != S_CSC_OK)
		return res;

	switch (obj.type) {
	case OT_INTEGER:
		tt = TOK_INTEGER;
		break;

	case OT_REAL:
		tt = TOK_REAL;
		set_real (&obj, closure, ceil (obj.val.real));
		break;

	default:
		return obj_error (&obj, "Bad operand type");
	}

	return m_push (stack, &obj, tt);
}

static enum s_status tr_cos (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure)
{
	struct obj arg;
	enum s_status res;

	res = psi_safe_pop (stack, closure, 1, &arg);
	if (res != S_CSC_OK)
		return res;

	if (!obj_cast (&arg, OT_REAL)) {
		obj_destroy (&arg);
		return obj_error (&arg, "Bad operand type");
	}

	set_real (&arg, closure, cos (arg.val.real * M_PIl / 180));

	return m_push (stack, &arg, TOK_REAL);
}

static enum s_status tr_cvi (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure)
{
	struct obj arg;
	enum s_status res;

	res = psi_safe_pop (stack, closure, 1, &arg);
	if (res != S_CSC_OK)
		return res;

	if (!obj_cast (&arg, OT_INTEGER)) {
		obj_destroy (&arg);
		return obj_error (&arg, "Bad operand type");
	}

	return m_push (stack, &arg, TOK_INTEGER);
}

static enum s_status tr_cvr (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure)
{
	struct obj arg;
	enum s_status res;

	res = psi_safe_pop (stack, closure, 1, &arg);
	if (res != S_CSC_OK)
		return res;

	if (!obj_cast (&arg, OT_REAL)) {
		obj_destroy (&arg);
		return obj_error (&arg, "Bad operand type");
	}

	return m_push (stack, &arg, TOK_REAL);
}

static enum s_status tr_div (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure)
{
	struct obj n1, n2, n3;
	enum s_status res;

	res = psi_safe_pop (stack, closure, 2, &n2, &n1);
	if (res != S_CSC_OK)
		return res;

	if (!obj_cast (&n1, OT_REAL) || !obj_cast (&n2, OT_REAL)) {
		obj_destroy (&n1);
		obj_destroy (&n2);
		return psi_error (closure, "Bad operand types");
	}

	set_real (&n3, closure, n1.val.real / n2.val.real);

	if (n3.val.real == HUGE_VAL || n3.val.real == -HUGE_VAL)
		return psi_error (closure, "Divide by zero");

	return m_push (stack, &n3, TOK_REAL);
}

static enum s_status tr_exp (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure)
{
	struct obj base, exp, ans;
	enum s_status res;

	res = psi_safe_pop (stack, closure, 2, &exp, &base);
	if (res != S_CSC_OK)
		return res;

	if (!obj_cast (&base, OT_REAL) || !obj_cast (&exp, OT_REAL)) {
		obj_destroy (&base);
		obj_destroy (&exp);
		return psi_error (closure, "Bad operand types");
	}

	set_real (&ans, closure, pow (base.val.real, exp.val.real));

	if (ans.val.real == HUGE_VAL || ans.val.real == -HUGE_VAL)
		return psi_error (closure, "Domain error");

	return m_push (stack, &ans, TOK_REAL);
}

static enum s_status tr_floor (struct m_stack *stack,
			       struct obj *dobj, int token,
			       void *closure)
{
	struct obj obj;
	enum s_status res;
	int tt;

	res = psi_safe_pop (stack, closure, 1, &obj);
	if (res != S_CSC_OK)
		return res;

	switch (obj.type) {
	case OT_INTEGER:
		tt = TOK_INTEGER;
		break;

	case OT_REAL:
		tt = TOK_REAL;
		set_real (&obj, closure, floor (obj.val.real));
		break;

	default:
		return obj_error (&obj, "Bad operand type");
	}

	return m_push (stack, &obj, tt);
}

static enum s_status tr_idiv (struct m_stack *stack,
			      struct obj *obj, int token,
			      void *closure)
{
	struct obj n1, n2, n3;
	enum s_status res;

	res = psi_safe_pop (stack, closure, 2, &n2, &n1);
	if (res != S_CSC_OK)
		return res;

	if (n1.type != OT_INTEGER || n2.type != OT_INTEGER) {
		obj_destroy (&n1);
		obj_destroy (&n2);
		return psi_error (closure, "Bad operand types");
	}

	if (!n2.val.integer)
		return psi_error (closure, "Divide by zero");

	set_integer (&n3, closure, n1.val.integer / n2.val.integer);

	return m_push (stack, &n3, TOK_INTEGER);
}

static enum s_status tr_ln (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct obj arg;
	enum s_status res;

	res = psi_safe_pop (stack, closure, 1, &arg);
	if (res != S_CSC_OK)
		return res;

	if (!obj_cast (&arg, OT_REAL)) {
		obj_destroy (&arg);
		return obj_error (&arg, "Bad operand type");
	}

	if (arg.val.real <= 0)
		return psi_error (closure, "Domain error");

	set_real (&arg, closure, log (arg.val.real));

	return m_push (stack, &arg, TOK_REAL);
}

static enum s_status tr_log (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure)
{
	struct obj arg;
	enum s_status res;

	res = psi_safe_pop (stack, closure, 1, &arg);
	if (res != S_CSC_OK)
		return res;

	if (!obj_cast (&arg, OT_REAL)) {
		obj_destroy (&arg);
		return obj_error (&arg, "Bad operand type");
	}

	if (arg.val.real <= 0)
		return psi_error (closure, "Domain error");

	set_real (&arg, closure, log (arg.val.real) / log (10));

	return m_push (stack, &arg, TOK_REAL);
}

static enum s_status tr_mod (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure)
{
	struct obj n1, n2, n3;
	enum s_status res;

	res = psi_safe_pop (stack, closure, 2, &n2, &n1);
	if (res != S_CSC_OK)
		return res;

	if (n1.type != OT_INTEGER || n2.type != OT_INTEGER) {
		obj_destroy (&n1);
		obj_destroy (&n2);
		return psi_error (closure, "Bad operand types");
	}

	if (n2.val.integer <= 0)
		return psi_error (closure, "Divide by zero");

	set_integer (&n3, closure, n1.val.integer % n2.val.integer);

	return m_push (stack, &n3, TOK_INTEGER);
}

static enum s_status tr_mul (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure)
{
	struct obj n1, n2, n3;
	enum s_status res;

	res = psi_safe_pop (stack, closure, 2, &n2, &n1);
	if (res != S_CSC_OK)
		return res;

	if (n1.type == OT_INTEGER && n2.type == OT_INTEGER) {
		set_integer (&n3, closure,
			     ((pdf_real_t) n1.val.integer) *
			     n2.val.integer);
		return m_push (stack, &n3, TOK_INTEGER);
	}

	if (!obj_cast (&n1, OT_REAL) || !obj_cast (&n2, OT_REAL)) {
		obj_destroy (&n1);
		obj_destroy (&n2);
		return psi_error (closure, "Bad operand types");
	}

	set_real (&n3, closure, n1.val.real * n2.val.real);

	return m_push (stack, &n3, TOK_REAL);
}

static enum s_status tr_neg (struct m_stack *stack,
			     struct obj *dobj, int token,
			     void *closure)
{	
	struct obj obj;
	enum s_status res;
	int tt;

	res = psi_safe_pop (stack, closure, 1, &obj);
	if (res != S_CSC_OK)
		return res;

	switch (obj.type) {
	case OT_INTEGER:
		tt = TOK_INTEGER;

		set_integer (&obj, closure, -(pdf_real_t) obj.val.integer);

		tt = obj.type == OT_INTEGER ? TOK_INTEGER : TOK_REAL;
		break;

	case OT_REAL:
		tt = TOK_REAL;
		set_real (&obj, closure, -obj.val.real);
		break;

	default:
		return obj_error (&obj, "Bad operand type");
	}

	return m_push (stack, &obj, tt);
}

static enum s_status tr_round (struct m_stack *stack,
			       struct obj *dobj, int token,
			       void *closure)
{
	struct obj obj;
	enum s_status res;
	double df, dc;
	int tt; 

	res = psi_safe_pop (stack, closure, 1, &obj);
	if (res != S_CSC_OK)
		return res;

	switch (obj.type) {
	case OT_INTEGER:
		tt = TOK_INTEGER;
		break;

	case OT_REAL:
		tt = TOK_REAL;
		df = fabs (obj.val.real - floor (obj.val.real));
		dc = fabs (ceil (obj.val.real) - obj.val.real);
		if (df < dc)
			set_real (&obj, closure, floor (obj.val.real));
		else
			set_real (&obj, closure, ceil (obj.val.real));
		break;

	default:
		return obj_error (&obj, "Bad operand type");
	}

	return m_push (stack, &obj, tt);
}

static enum s_status tr_sin (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure)
{
	struct obj arg;
	enum s_status res;

	res = psi_safe_pop (stack, closure, 1, &arg);
	if (res != S_CSC_OK)
		return res;

	if (!obj_cast (&arg, OT_REAL)) {
		obj_destroy (&arg);
		return obj_error (&arg, "Bad operand type");
	}

	set_real (&arg, closure, sin (arg.val.real * M_PIl / 180));

	return m_push (stack, &arg, TOK_REAL);
}

static enum s_status tr_sub (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure)
{
	struct obj n1, n2, n3;
	enum s_status res;

	res = psi_safe_pop (stack, closure, 2, &n2, &n1);
	if (res != S_CSC_OK)
		return res;

	if (n1.type == OT_INTEGER && n2.type == OT_INTEGER) {
		set_integer (&n3, closure,
			     ((pdf_real_t) n1.val.integer) -
			     n2.val.integer);
		return m_push (stack, &n3, TOK_INTEGER);
	}

	if (!obj_cast (&n1, OT_REAL) || !obj_cast (&n2, OT_REAL)) {
		obj_destroy (&n1);
		obj_destroy (&n2);
		return psi_error (closure, "Bad operand types");
	}

	set_real (&n3, closure, n1.val.real - n2.val.real);

	return m_push (stack, &n3, TOK_REAL);
}

static enum s_status tr_truncate (struct m_stack *stack,
				  struct obj *dobj, int token,
				  void *closure)
{
	struct obj obj;
	enum s_status res;
	int tt;

	res = psi_safe_pop (stack, closure, 1, &obj);
	if (res != S_CSC_OK)
		return res;

	switch (obj.type) {
	case OT_INTEGER:
		tt = TOK_INTEGER;
		break;

	case OT_REAL:
		tt = TOK_REAL;
		set_real (&obj, closure, trunc (obj.val.real));
		break;

	default:
		return obj_error (&obj, "Bad operand type");
	}

	return m_push (stack, &obj, tt);
}

