/*
 * PDFlow -- psi.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#define _GNU_SOURCE

#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "psi_private.h"

static enum s_status tr_push_ptr (struct m_stack *stack,
				  struct obj *obj, int token,
				  void *closure);

static enum s_status tr_integer (struct m_stack *stack,
				 struct obj *obj, int token,
				 void *closure);
static enum s_status tr_real (struct m_stack *stack,
			      struct obj *obj, int token,
			      void *closure);

static enum s_status tr_dict_begin (struct m_stack *stack,
				    struct obj *obj, int token,
				    void *closure);
static enum s_status tr_dict_end (struct m_stack *stack,
				  struct obj *obj, int token,
				  void *closure);
static enum s_status tr_array_begin (struct m_stack *stack,
				     struct obj *obj, int token,
				     void *closure);
static enum s_status tr_array_end (struct m_stack *stack,
				   struct obj *obj, int token,
				   void *closure);
static enum s_status tr_null (struct m_stack *stack,
			      struct obj *obj, int token,
			      void *closure);
static enum s_status tr_true (struct m_stack *stack,
			      struct obj *obj, int token,
			      void *closure);
static enum s_status tr_false (struct m_stack *stack,
			       struct obj *obj, int token,
			       void *closure);

static enum s_status tr_oth_keyword (struct m_stack *stack,
				     struct obj *obj, int token,
				     void *closure);

static enum s_status tr_eof (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure);

static struct m_block_desc bdesc = { TOK_LCURLY, TOK_RCURLY };

static struct m_rule rules [] = {
	{ TOK_NULL, tr_null },
	{ TOK_PLUS, NULL },
	{ TOK_MINUS, NULL },
	{ TOK_BOOLEAN, NULL },
	{ TOK_TRUE, tr_true },
	{ TOK_FALSE, tr_false },
	{ TOK_INTEGER, tr_integer },
	{ TOK_REAL, tr_real },
	{ TOK_EOF, tr_eof },

	{ TOK_DICT_BEGIN, tr_dict_begin },
	{ TOK_DICT_END, tr_dict_end },
	{ TOK_ARRAY_BEGIN, tr_array_begin },
	{ TOK_ARRAY_END, tr_array_end },
	{ TOK_NAME, NULL },
	{ TOK_LITERAL, NULL },
	{ TOK_HEX, NULL },

	{ TOK_OTH_KEYWORD, tr_oth_keyword },
	{ -1 }
};

static struct psi_interpreter template;

enum s_status psi_error (struct psi_closure *pc, const char *fmt, ...) 
{
	enum s_status res;
	va_list ap;

	va_start (ap, fmt);
	res = pl_verror (&pc -> ps, fmt, ap);
	va_end (ap);

	return res;
}

enum s_status psi_user_push (struct psi_interpreter *pi, struct obj *dict)
{
	struct psi_dict_stack_node *sdn;
	enum s_status res;

	sdn = malloc (sizeof (struct psi_dict_stack_node));
	if (sdn == NULL)
		return S_CSC_OUT_OF_MEMORY;

	assert (dict -> type == OT_OBJ_PTR);

	res = obj_dup (dict, &sdn -> obj);
	if (res != S_CSC_OK) {
		free (sdn);
		return res;

	}

	sdn -> next = pi -> user_stack;
	pi -> user_stack = sdn;
	
	return S_CSC_OK;
}

int psi_user_pop (struct psi_interpreter *pi)
{
	struct psi_dict_stack_node *sdn;

	sdn = pi -> user_stack;
	if (sdn == NULL)
		return 0;

	pi -> user_stack = sdn -> next;
	obj_destroy (&sdn -> obj);
	free (sdn);
	
	return 1;
}

int psi_user_top (struct psi_interpreter *pi, struct obj **obj)
{
	struct psi_dict_stack_node *sdn;

	sdn = pi -> user_stack;
	if (sdn == NULL)
		return 0;

	*obj = &pi -> user_stack -> obj;
	
	return 1;
}

enum s_status psi_push (struct m_stack *stack, struct obj *obj,
			int token)
{
	enum s_status res;
	struct obj nobj;

	res = obj_ptr_dup (obj, &nobj);
	if (res != S_CSC_OK)
		return res;

	res = m_push (stack, &nobj, token);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;	
}	

static enum s_status tr_push_ptr (struct m_stack *stack,
				  struct obj *obj, int token,
				  void *closure)
{
	enum s_status res;

	res = psi_push (stack, obj, token);
	obj_destroy (obj);
	if (res != S_CSC_OK)
		return res;
	
	return S_CSC_OK;
}

static enum s_status tr_real (struct m_stack *stack,
			      struct obj *nobj, int token,
			      void *closure)
{
	struct obj *obj, sobj;
	int tt, neg;

	neg = 0;
	obj = m_top (stack, &tt);
	if (obj != NULL) {
		if (tt == TOK_PLUS || tt == TOK_MINUS) {
			m_pop (stack, &sobj, &tt);
			if (tt == TOK_MINUS)
				neg = 1;
		}	
	}

	if (neg)
		nobj -> val.real = -nobj -> val.real;

	return m_push (stack, nobj, token);
}

static enum s_status tr_integer (struct m_stack *stack,
				 struct obj *nobj, int token,
				 void *closure)
{
	struct obj *obj, sobj;
	int tt, neg;

	neg = 0;
	obj = m_top (stack, &tt);
	if (obj != NULL) {
		if (tt == TOK_PLUS || tt == TOK_MINUS) {
			m_pop (stack, &sobj, &tt);
			if (tt == TOK_MINUS)
				neg = 1;
		}
	}

	if (neg)
		nobj -> val.integer = -nobj -> val.integer;
	else if (nobj -> val.integer < 0)
		return obj_error (nobj, "Integer too large");

	return m_push (stack, nobj, token);
}

enum s_status psi_safe_pop (struct m_stack *stack, void *closure,
			    int nops, ...)
{
	va_list ap;
	int i, j;
	int tt;

	va_start (ap, nops);
	for (i = 0; i < nops; i++)
		if (!m_pop (stack, va_arg (ap, struct obj *), &tt))
			break;
	va_end (ap);

	if (i == nops)
		return S_CSC_OK;

	va_start (ap, nops);
	for (j = 0; j < i; j++)
		obj_destroy (va_arg (ap, struct obj *));
	va_end (ap);

	return psi_error (closure, "Stack underflow");
}

static void set_boolean (struct obj *obj, struct psi_closure *pc,
			 pdf_bool_t b)
{
	obj_set_boolean (obj, &pc -> ps.ptr, b);
}

static enum s_status tr_true (struct m_stack *stack,
			      struct obj *obj, int token,
			      void *closure)
{
	struct obj true;

	set_boolean (&true, closure, 1);

	return m_push (stack, &true, TOK_BOOLEAN);
}

static enum s_status tr_false (struct m_stack *stack,
			       struct obj *obj, int token,
			       void *closure)
{
	struct obj false;

	set_boolean (&false, closure, 0);

	return m_push (stack, &false, TOK_BOOLEAN);
}

static enum s_status tr_oth_keyword (struct m_stack *stack,
				     struct obj *obj, int token,
				     void *closure)
{
	struct psi_dict_stack_node *sn;
	struct obj_dictionary *dict;
	struct obj_dict_node *dn;
	struct psi_interpreter *pi;
	struct obj *dobj, nobj;
	struct psi_closure *pc;
	enum s_status res;

	pc = closure;
	pi = pc -> interpreter;

	dn = NULL;
	for (sn = pi -> user_stack; sn != NULL; sn = sn -> next) {
		dobj = obj_ptr_get (&sn -> obj);

		if (dobj -> type == OT_NULL)
			continue;

		dict = &dobj -> val.dictionary;
		dn = obj_dict_find_node (dict, &obj -> val.string);
		if (dn != NULL)
			break;
	}

	if (dn == NULL)
		dn = obj_dict_find_node (&pi -> sys_dict, &obj -> val.string);

	if (dn == NULL) {
		res = obj_error (obj, "Undefined object");
		obj_destroy (obj);
		return res;
	}
	obj_destroy (obj);
	
	obj = obj_ptr_get (&dn -> obj);

	switch (obj -> type) {
	case OT_ALIVE:
		res = m_eval_block (pi -> state, obj, pc);
		break;

	case OT_KEYWORD:
		res = ((m_trigger) obj -> val.keyword -> data) 
			(stack, &dn -> obj, token, pc);
		break;

	default:
		res = obj_dup (&dn -> obj, &nobj);
		if (res != S_CSC_OK)
			break;

		res = m_push (stack, &nobj, token);
		if (res != S_CSC_OK)
			obj_destroy (&nobj);
	}

	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}


static enum s_status tr_eof (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure)
{
	return psi_error (closure, "Unexpected end of file");
}


static enum s_status tr_dict_begin (struct m_stack *stack,
				    struct obj *obj, int token,
				    void *closure)
{
	m_open (stack);

	return m_push (stack, obj, token);
}

static enum s_status tr_dict_end (struct m_stack *stack,
				  struct obj *nobj, int token,
				  void *closure)
{
	struct obj_dictionary dictionary;
	struct obj_dict_node dnode;
	struct obj kobj, vobj, dobj;
	enum s_status res;
	int tt;

	obj_init_dictionary (&dictionary);
	while (m_pop (stack, &vobj, &tt)) {
		if (tt == TOK_DICT_BEGIN)
			break;

		if (!m_pop (stack, &kobj, &tt)) {
			obj_destroy (&vobj);
			obj_dict_destroy (&dictionary);
			return psi_error (closure, "Bad dictionary");
		}

		if (tt != TOK_NAME) {
			obj_destroy (&kobj);
			obj_destroy (&vobj);
			obj_dict_destroy (&dictionary);
			return psi_error (closure, "Bad dictionary");
		}

		obj_init_dict_node (&dnode, &kobj, &vobj);

		res = obj_xfer_dict (&dictionary, &dnode);
		if (res != S_CSC_OK) {
			obj_dict_destroy (&dictionary);
			return res;
		}
	}
	
	if (tt != TOK_DICT_BEGIN) {
		obj_dict_destroy (&dictionary);
		return psi_error (closure, "Unmatched '>>'");
	}

	obj_set_dictionary (&dobj, &vobj.ptr, &dictionary);
	res = tr_push_ptr (stack, &dobj, TOK_DICTIONARY, closure);
	if (res != S_CSC_OK) {
		obj_destroy (&dobj);
		return res;
	}

	m_close (stack);

	return S_CSC_OK;
}

static enum s_status tr_array_begin (struct m_stack *stack,
				     struct obj *obj, int token,
				     void *closure)
{
	m_open (stack);

	return m_push (stack, obj, token);
}

static enum s_status tr_array_end (struct m_stack *stack,
				   struct obj *nobj, int token,
				   void *closure)
{
	struct obj_array array;
	struct obj obj, aobj;
	enum s_status res;
	int tt;

	obj_init_array (&array);
	while (m_pop (stack, &obj, &tt)) {
		if (tt == TOK_ARRAY_BEGIN)
			break;

		res = obj_xfer_array (&array, &obj);
		if (res != S_CSC_OK) {
			obj_array_destroy (&array);
			return res;
		}
	}
	
	if (tt != TOK_ARRAY_BEGIN) {
		obj_array_destroy (&array);
		return psi_error (closure, "Unmatched ']'");
	}

	obj_set_array (&aobj, &obj.ptr, &array);
	res = tr_push_ptr (stack, &aobj, TOK_ARRAY, closure);
	if (res != S_CSC_OK) {
		obj_destroy (&aobj);
		return res;
	}

	m_close (stack);

	return S_CSC_OK;
}

static enum s_status tr_null (struct m_stack *stack,
			      struct obj *obj, int token,
			      void *closure)
{
	obj -> type = OT_NULL;

	return m_push (stack, obj, token);
}

enum s_status psi_add_procset (struct psi_interpreter *pi,
			       struct psi_procset *ps)
{
	struct obj_keyword *kw;
	struct obj obj;
	enum s_status res;

	for (kw = ps -> keywords; kw -> name != NULL; kw++) {
		obj_set_keyword (&obj, &f_null, kw);

		res = obj_add_dict (&pi -> sys_dict, kw -> name, &obj);
		if (res != S_CSC_OK)
			return res;

		obj_destroy (&obj);
	}

	return S_CSC_OK;
}

static void clear_procsets (struct psi_interpreter *pi)
{
	obj_dict_destroy (&pi -> sys_dict);
}

static enum s_status init_procsets (struct psi_interpreter *pi)
{
	enum s_status res;

	res = psi_add_procset (pi, &psi_core_procset);
	if (res != S_CSC_OK)
		return res;

	res = psi_add_procset (pi, &psi_math_procset);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static void destroy_file_resource (struct psi_file_resource *fr)
{
	if (fr -> path != NULL)
		free (fr -> path);
	if (fr -> name != NULL)
		free (fr -> name);

	free (fr);
}

enum s_status psi_add_file_resource (struct psi_file_resource **list,
				     char *name, char *path)
{
	struct psi_file_resource *fr;

	fr = malloc (sizeof (struct psi_file_resource));
	if (fr == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (fr, 0, sizeof (struct psi_file_resource));
	fr -> name = strdup (name);
	fr -> path = strdup (path);
	if (fr -> name == NULL || fr -> path == NULL) {
		destroy_file_resource (fr);
		return S_CSC_OUT_OF_MEMORY;
	}

	fr -> next = *list;
	*list = fr;

	return S_CSC_OK;
}

struct psi_file_resource *psi_find_file_resource (struct psi_file_resource *l,
						  struct obj *name)
{
	struct psi_file_resource *fr;

	for (fr = l; fr != NULL; fr = fr -> next)
		if (obj_string_equal (&name -> val.string, fr -> name))
			break;

	return fr;
}

void psi_destroy_file_resources (struct psi_file_resource **list)
{
	struct psi_file_resource *fr;

	while (*list != NULL) {
		fr = *list;
		*list = fr -> next;
		
		destroy_file_resource (fr);
	}
}

static enum s_status init_user_stack (struct psi_interpreter *pi)
{
	struct obj obj, dobj;
	enum s_status res;

	while (psi_user_pop (pi))
		;

	obj_set_null (&obj, &f_null);
	res = obj_ptr_dup (&obj, &dobj);
	if (res != S_CSC_OK)
		return res;

	res = psi_user_push (pi, &dobj);
	obj_destroy (&dobj);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status eval (struct psi_interpreter *pi,
			   struct f_file *file, 
			   int argin, struct obj *in,
			   int argout, struct obj *out, int curly)
{
	enum pl_token_type tt;
	struct psi_closure pc;
	struct f_ptr fptr;
	enum pl_token_type eoftt;
	enum s_status res;
	int n, outer;

	f_ptr_init (&fptr, file, 0);

	pc.interpreter = pi;
	pl_init (&pc.ps, &fptr);	

	clear_procsets (pi);
	res = init_procsets (pi);
	if (res != S_CSC_OK)
		return res;

	res = init_user_stack (pi);
	if (res != S_CSC_OK)
		return res;

	res = m_open_machine (pi -> state);
	if (res != S_CSC_OK)
		return res;

	res = m_push_stack (pi -> state, in, TOK_REAL, argin);
	if (res != S_CSC_OK) {
		m_close_machine (pi -> state);
		return res;
	}
	
	outer = 0;
	if (curly) {
		res = pl_lex (&pc.ps, &tt);
		if (res != S_CSC_OK) {
			m_close_machine (pi -> state);
			return res;
		}

		if (tt != TOK_LCURLY) {
			m_close_machine (pi -> state);
			return psi_error (&pc, "No open curly brace");
		}

		outer++;
		eoftt = TOK_RCURLY;
	} else
		eoftt = TOK_EOF;

	while (1) {
		res = pl_lex (&pc.ps, &tt);
		if (res != S_CSC_OK) {
			m_close_machine (pi -> state);
			return res;
		}
		if (outer && tt == eoftt)
			break;
		outer = !m_run_machine_cyc (pi -> state, &pc.ps.token, 
					    tt, &res, &pc);
		if (res != S_CSC_OK) {
			m_close_machine (pi -> state);
			return res;
		}
	} 
	if (res != S_CSC_OK) {
		obj_destroy (&pc.ps.token);
		m_close_machine (pi -> state);
		return res;
	}

	n = argout;
	res = m_dump_stack (pi -> state, out, &n);
	m_close_machine (pi -> state);
	if (res != S_CSC_OK)
		return res;

	if (n != argout)
		return psi_error (&pc, "Output stack does not match args");

	for (n = 0; n < argout; n++) {
		if (!obj_cast (&out [n], OT_REAL))
			return psi_error (&pc, "Bad object type");
	}

	return S_CSC_OK;
}

enum s_status psi_eval (struct psi_interpreter *pi,
			struct f_file *file)
{
	return eval (pi, file, 0, NULL, 0, NULL, 0);
}

enum s_status psi_add_cmap (struct psi_interpreter *pi,
			    struct f_file *file,
			    struct ref_owner *p,
			    struct cmap **cmap)
{
	enum s_status res;

	res = eval (pi, file, 0, NULL, 0, NULL, 0);
	if (res != S_CSC_OK)
		return res;

	if (cmap != NULL)
		psi_cid_get_last_cmap (pi, p, cmap);

	return S_CSC_OK;
}

enum s_status psi_add_cmap_dir (struct psi_interpreter *pi,
				const char *dir)
{
	return psi_cid_add_cmap_dir (pi -> resources, dir);
}

enum s_status psi_get_cmap (struct psi_interpreter *pi,
			    struct obj *name,
			    struct ref_owner *p,
			    struct cmap **cmap)
{
	return psi_cid_get_cmap_by_name (pi, name, p, cmap);
}

enum s_status psi_eval_f (struct psi_interpreter *pi,
			  struct f_file *file, 
			  int argin, struct obj *in,
			  int argout, struct obj *out)
{
	return eval (pi, file, argin, in, argout, out, 1);
}

static enum s_status init_resources (struct psi_resources **prsp)
{
	struct psi_resources *prs;

	prs = malloc (sizeof (struct psi_resources));
	if (prs == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (prs, 0, sizeof (struct psi_resources));

	prs -> refs = 1;
	
	*prsp = prs;

	return S_CSC_OK;
}

static void ref_resources (struct psi_resources *prs)
{
	prs -> refs++;
}

static void unref_resources (struct psi_resources *prs)
{
	if (!--prs -> refs) {
		psi_cid_destroy_cmap_res (prs);
		free (prs);
	}
}

enum s_status psi_create (struct psi_interpreter **pi)
{
	struct psi_interpreter *p;
	enum s_status res;

	p = malloc (sizeof (struct psi_interpreter));
	if (p == NULL)
		return S_CSC_OUT_OF_MEMORY;

	*p = template;

	res = init_resources (&p -> resources);
	if (res != S_CSC_OK) {
		free (p);
		return res;
	}

	*pi = p;

	return S_CSC_OK;
}

void psi_destroy (struct psi_interpreter *pi)
{
	while (psi_user_pop (pi))
		;
	
	clear_procsets (pi);

	psi_cid_destroy_cmap (pi);
	unref_resources (pi -> resources);

	free (pi);
}

enum s_status psi_clone (struct psi_interpreter *spi,
			 struct psi_interpreter **dpi)
{
	struct psi_interpreter *pi;
	enum s_status res;

	res = psi_create (&pi);
	if (res != S_CSC_OK)
		return res;

	unref_resources (pi -> resources);
	pi -> resources = spi -> resources;
	ref_resources (pi -> resources);

	*dpi = pi;

	return S_CSC_OK;
}

enum s_status psi_init (void)
{
	enum s_status res;

	res = m_create_machine (rules, TOK__LAST__, &bdesc, 1,
				TOK_EOF, &template.mach);
	if (res != S_CSC_OK)
		return res;

	res = m_start_machine (template.mach, &template.state);
	if (res != S_CSC_OK) {
		m_destroy_machine (template.mach);
		return res;
	}

	return S_CSC_OK;
}

void psi_cleanup (void)
{
	m_stop_machine (template.state);
	m_destroy_machine (template.mach);
}



