/*
 * PDFlow -- refs.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "refs.h"

#ifdef DEBUG_REFS

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>

#define DELTA 10

//#define LOG
//#define VLOG
//#define OOI "area"
//#define CHAIN

static ref_t *refs;

#ifdef LOG
static struct ref_owner *wown;
static int ref_op;
#endif

#ifdef LOG 
static void r_log (ref_t *r, struct ref_owner *p, const char *op,
		   const char *fname, int line)
{
#ifdef VLOG
	int i;
#endif
	if (wown != NULL && &r -> myself != wown && p != wown)
		return;

#ifdef OOI
	if (p -> name != NULL &&
	    strcmp (p -> name, OOI) && strcmp (r -> myself.name, OOI))
		return;
#endif
	
	fprintf (stderr, ":p%p(%s)->:r%p(%s):%s=%d:%s:%d:%d\n", 
		 p, p -> name, &r -> myself, r -> myself.name, op, 
		 r -> n, fname, line, ref_op);
	ref_op++;
#ifdef VLOG
	for (i = 0; i < r -> n; i++) {
		fprintf (stderr, "   %s[%p] (%s:%d)\n", 
			 r -> table [i].p -> name, r -> table [i].p,
			 r -> table [i].fname, r -> table [i].line);
	}
#endif

#ifdef CHAIN
	fprintf (stderr, "### CHAIN\n");
	for (r = refs; r != NULL; r = r -> next)
		fprintf (stderr, " c%p\n", &r -> myself);
#endif
}
#else
#define r_log(r, p, op, fname, line)
#endif

static void moan (int end, ref_t *r, const char *msg, ...)
{
	va_list ap;
	int i;

	va_start (ap, msg);
	vfprintf (stderr, msg, ap);
	va_end (ap);

	fprintf (stderr, "\nObject %s[%p] is pointed to by:\n", 
		 r -> myself.name, &r -> myself);
	for (i = 0; i < r -> n; i++) {
		fprintf (stderr, "   %s[%p] (%s:%d)\n", 
			 r -> table [i].p -> name, r -> table [i].p,
			 r -> table [i].fname, r -> table [i].line);
	}

	if (end)
		abort ();
}

void ref__init (ref_t *r, const char *name, struct ref_owner *p,
		const char *fname, int line)
{
	r -> next = refs;
	refs = r;

	r -> myself.type = REF_OT_REF;
	r -> myself.name = name;

	r -> n = r -> cap = 0;
	r -> table = NULL;

	ref__up (r, p, fname, line);
}

int ref_dump (ref_t *r)
{
	int i, found, dead;
	ref_t **p, *q, *s;

	p = &refs;
	found = dead = 0;
	s = q = NULL;
	while (*p != NULL) {
		if (*p == r) {
			found++;
			*p = r -> next;
			continue;
		} 
		for (i = 0; i < (*p) -> n; i++)
			if ((*p) -> table [i].p == ref_rown (r)) {
				if (!dead)
					fprintf (stderr, 
						 "Object %s[%p] is "
						 "pointing:\n",
						 ref_rown (r) -> name, 
						 ref_rown (r));
				fprintf (stderr, "+ %s[%p] (%s:%d)\n", 
					 ref_rown (*p) -> name, 
					 ref_rown (*p),
					 (*p) -> table [i].fname, 
					 (*p) -> table [i].line);
				dead++;
			}
		/* These pointers are useful to debug when ref_zero is
		 * not called */
		s = q;
		q = *p;
		p = &(*p) -> next;
	}

	return dead ? -1 : found;
}

void ref_zero (ref_t *r)
{
	int found;

	if (r -> n)
		moan (1, r, "Destroy reference when still pointed to by");

	found = ref_dump (r);

	if (found < 0)
		abort ();
	if (found != 1)
		moan (1, r, "Found %d link references", found);

	if (r -> table != NULL)
		free (r -> table);
}

int ref_val (ref_t *r)
{
	return r -> n;
}

void ref__up (ref_t *r, struct ref_owner *p, const char *fname, int line)
{
	r_log (r, p, "+", fname, line);
	if (r -> n >= r -> cap) {
		r -> table = realloc (r -> table, 
				      sizeof (struct ref_ptr) * 
				      (r -> n + DELTA));
		assert (r -> table != NULL);
	}

	r -> table [r -> n].p = p;
	r -> table [r -> n].fname = fname;
	r -> table [r -> n].line = line;
	r -> n++;
}

int ref_down (ref_t *r, struct ref_owner *p)
{
	int i;

	r_log (r, p, "-", "", 0);
	if (p == NULL) {
		if (r -> n != 1)
			moan (1, r, "Wildcard DOWN on referenced obj");

		return r -> n = 0;
	}

	if (!r -> n)
		moan (1, r, "DOWN on a nonexisting object");

	for (i = 0; i < r -> n; i++) {
		if (r -> table [i].p == p)
			break;
	}

	if (i == r -> n)
		moan (1, r, "Unmatched DOWN request by %s[%p]", 
		      p -> name, p);

	r -> n--;
	if (i < r -> n)
		memmove (&r -> table [i], &r -> table [i + 1],
			 sizeof (struct ref_ptr) * (r -> n - i));

	return r -> n;
}

void ref_init_owner (struct ref_owner *p, const char *name)
{
	p -> type = REF_OT_ENDPOINT;
	p -> name = name;
}

struct ref_owner *ref_rown (ref_t *r)
{
	return &r -> myself;
}

void ref_shutdown (void)
{
	ref_t *r;

	if (refs == NULL)
		return;

	fprintf (stderr, "Some objects have not been deallocated:");
	for (r = refs; r != NULL; r = r -> next)
		moan (0, r, "###########################\n");

	abort ();
}

#endif
