/*
 * PDFlow -- cmap.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <string.h>

#include "version.h"
#include "types/cmap.h"

#define XREF_IMPLEMENTATION struct cmap
#define CMAP
#include "types/cmap.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct cmap_system_info
#define CMAP_SYSTEM_INFO
#include "types/cmap.tab.h"
#undef XREF_IMPLEMENTATION

static void fini (struct cmap *cmap, struct ref_owner *p);

T_IMPLEMENTATIONS (cmap_system_info, cmap_system_info, 
		   &cmap_system_info_field);
T_IMPLEMENTATIONS_LOC (cmap, cmap, &cmap_field, NULL, NULL, fini);


static void fini (struct cmap *cmap, struct ref_owner *p)
{
	struct cmap_table *t;

	while (cmap -> tables != NULL) {
		t = cmap -> tables;
		cmap -> tables = t -> next;

		cmap_destroy_table (t);
	}

	cmap_unref (cmap -> base, ref_rown (&cmap -> t.refs));
}

static enum s_status identity (struct cmap **cmapp, struct xref_db *rdb,
			       struct ref_owner *p, enum cmap_wmode wmode,
			       const char *fname, int line)
{
	struct obj first, last;
	struct obj_bstr s;
	char fstr [2], lstr [2];
	struct cmap *cmap;
	enum s_status res;
	int i;

	res = cmap_new (&cmap, rdb, p, fname, line);
	if (res != S_CSC_OK)
		return res;

	fstr [0] = fstr [1] = 0;
	lstr [0] = lstr [1] = 0xff;

	s.size = 2, s.s = fstr;
	obj_set_string (&first, &f_null, &s);
	
	s.size = 2, s.s = lstr;
	obj_set_string (&last, &f_null, &s);

	res = cmap_create_table (&cmap -> tables, &first, &last);
	if (res != S_CSC_OK) {
		cmap_unref (cmap, p);
		return res;
	}

	for (i = 0; i <= 0xffff; i++)
		cmap -> tables -> code [i] = i;

	cmap -> wmode = wmode;

	*cmapp = cmap;

	return S_CSC_OK;
}

enum s_status cmap__build_identity_h (struct cmap **cmap, struct xref_db *rdb,
				      struct ref_owner *p, const char *fname,
				      int line)
{
	return identity (cmap, rdb, p, CMAP_WM_HORIZONTAL, fname, line);
}

enum s_status cmap__build_identity_v (struct cmap **cmap, struct xref_db *rdb,
				      struct ref_owner *p, const char *fname,
				      int line)
{
	return identity (cmap, rdb, p, CMAP_WM_VERTICAL, fname, line);
}

static int matchlen (struct cmap_table *table, struct obj_bstr *s)
{
	int i, n;

	n = table -> blen < s -> size ? table -> blen : s -> size;
	for (i = 0; i < n; i++)
		if (s -> s [i] < table -> lower [i] ||
		    s -> s [i] > table -> upper [i])
			break;
	return i;
}

uint32_t cmap_get_cid (struct cmap *cmap, struct obj_bstr *s, size_t *len)
{
	struct cmap_table *t;
	struct obj_bstr bs;
	uint32_t *slot;
	int i, match;

	for (t = cmap -> tables; t != NULL; t = t -> next) {
		bs = *s;
		if (bs.size < t -> blen)
			continue;
		bs.size = t -> blen;
		slot = cmap_find_slot (t, &bs, &bs);
		if (slot != NULL) {
			*len = t -> blen;
			return *slot;
		}
	}

	match = 0;
	for (t = cmap -> tables; t != NULL; t = t -> next) {
		i = matchlen (t, s);
		if (i > match)
			match = i;
	}

	*len = 5;
	for (t = cmap -> tables; t != NULL; t = t -> next) {
		if (matchlen (t, s) == match && t -> blen < *len)
			*len = t -> blen;
	}

	return 0;
}

enum s_status cmap_create_table (struct cmap_table **tablep,
				 struct obj *start, struct obj *end)
{
	struct obj_bstr *sstr, *estr;
	struct cmap_table *table;
	int i;

	if (start -> type != OT_STRING ||
	    end -> type != OT_STRING ||
	    start -> val.string.size != end -> val.string.size ||
	    start -> val.string.size < 1 ||
	    start -> val.string.size > 4 ||
	    obj_string_compare_s (&start -> val.string,
				  &end -> val.string) > 0 ||
	    !start -> val.string.size)
		return obj_error (start, "Bad range");

	sstr = &start -> val.string;
	estr = &end -> val.string;

	table = malloc (sizeof (struct cmap_table));
	if (table == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (table, 0, sizeof (struct cmap_table));

	table -> blen = sstr -> size;
	table -> size = 1;
	for (i = 0; i < table -> blen; i++) {
		table -> lower [i] = sstr -> s [i];
		table -> upper [i] = estr -> s [i];
		table -> size *= (table -> upper [i] -
				  table -> lower [i] + 1);
	}

	table -> code = calloc (table -> size, sizeof (uint32_t));
	if (table -> code == NULL) {
		cmap_destroy_table (table);
		return S_CSC_OUT_OF_MEMORY;
	}

	*tablep = table;

	return S_CSC_OK;
}

enum s_status cmap_dup_table (struct cmap_table *stable,
			      struct cmap_table **dtable)
{
	struct cmap_table *t;

	t = malloc (sizeof (struct cmap_table));
	if (t == NULL)
		return S_CSC_OUT_OF_MEMORY;
	*t = *stable;

	t -> code = calloc (t -> size, sizeof (uint32_t));
	if (t == NULL) {
		free (t);
		return S_CSC_OUT_OF_MEMORY;
	}

	memcpy (t -> code, stable -> code, t -> size * sizeof (uint32_t));

	*dtable = t;

	return S_CSC_OK;
}

void cmap_destroy_table (struct cmap_table *table)
{
	if (table -> code != NULL)
		free (table -> code);
	free (table);
}

uint32_t *cmap_find_slot (struct cmap_table *table, 
			  struct obj_bstr *first, struct obj_bstr *last)
{
	uint8_t fidx, lidx;
	size_t ofs, lrow;
	int i;

	if (first -> size != table -> blen)
		return NULL;

	ofs = 0;
	lrow = 0;
	fidx = lidx = 0;
	for (i = 0; i < table -> blen; i++) {
		ofs *= lrow;
		lrow = table -> upper [i] - table -> lower [i] + 1;

		if (lrow != 256 && fidx != lidx)
			return NULL;

		fidx = first -> s [i];
		lidx = last -> s [i];
		if (fidx < table -> lower [i] || 
		    lidx > table -> upper [i])
			return NULL;
		ofs += fidx - table -> lower [i];
	}

	return &table -> code [ofs];
}
