/*
 * PDFlow -- cobj.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>

#include "types/cobj.h"

#define D_PREFIX_WIDTH  2
#define D_YEAR_WIDTH    4
#define D_MONTH_WIDTH   2
#define D_DAY_WIDTH     2
#define D_HOUR_WIDTH    2
#define D_MIN_WIDTH     2
#define D_SEC_WIDTH     2
#define D_TZ_SIGN_WIDTH 1
#define D_TZ_HOUR_WIDTH 2
#define D_TZ_SEP_WIDTH  1
#define D_TZ_MIN_WIDTH  2

#define D_WIDTH (D_PREFIX_WIDTH + D_YEAR_WIDTH + D_MONTH_WIDTH +	  \
		 D_DAY_WIDTH + D_HOUR_WIDTH + D_MIN_WIDTH + D_SEC_WIDTH + \
		 D_TZ_SIGN_WIDTH + D_TZ_HOUR_WIDTH + D_TZ_SEP_WIDTH +	  \
		 D_TZ_MIN_WIDTH)


static enum s_status enum_parser (struct xref_db *rdb, struct xref_bind *robj, 
				  struct ref_owner *p, struct obj *src, 
				  struct xref_field *fd,
				  const char *fname, int line);
static enum s_status date_parser (struct xref_db *rdb, struct xref_bind *robj, 
				  struct ref_owner *p, struct obj *src, 
				  struct xref_field *fd,
				  const char *fname, int line);
static enum s_status rect_parser (struct xref_db *rdb, struct xref_bind *robj, 
				  struct ref_owner *p, struct obj *src, 
				  struct xref_field *fd,
				  const char *fname, int line);
static enum s_status matrix_parser (struct xref_db *rdb, 
				    struct xref_bind *robj, 
				    struct ref_owner *p, 
				    struct obj *src, 
				    struct xref_field *fd,
				    const char *fname, int line);
static enum s_status id_parser (struct xref_db *rdb, struct xref_bind *robj, 
				struct ref_owner *p, struct obj *src, 
				struct xref_field *fd,
				const char *fname, int line);

static enum s_status enum_ser (struct xref_db *rdb, struct xref_bind *robj,
			       struct v_selection *sel, struct obj *dest, 
			       enum xref_changed *changed, 
			       struct xref_field *fd);
static enum s_status date_ser (struct xref_db *rdb, struct xref_bind *robj,
			       struct v_selection *sel, struct obj *dest, 
			       enum xref_changed *changed, 
			       struct xref_field *fd);
static enum s_status rect_ser (struct xref_db *rdb, struct xref_bind *robj,
			       struct v_selection *sel, struct obj *dest, 
			       enum xref_changed *changed, 
			       struct xref_field *fd);
static enum s_status matrix_ser (struct xref_db *rdb, struct xref_bind *robj,
				 struct v_selection *sel, struct obj *dest, 
				 enum xref_changed *changed, 
				 struct xref_field *fd);
static enum s_status matrix9_ser (struct xref_db *rdb, struct xref_bind *robj,
				 struct v_selection *sel, struct obj *dest, 
				 enum xref_changed *changed, 
				  struct xref_field *fd);
static enum s_status id_ser (struct xref_db *rdb, struct xref_bind *robj,
			     struct v_selection *sel, struct obj *dest, 
			     enum xref_changed *changed, 
			     struct xref_field *fd);
static void enum_funref (struct ref_owner *p, struct xref_field *fd);
static void id_funref (struct ref_owner *p, struct xref_field *fd);

struct cobj_matrix cobj_eye = {
	{ { 1, 0, 0 },
	  { 0, 1, 0 },
	  { 0, 0, 1 } }
};

struct xref_pstable cobj_enum_pstable = {
	sizeof (struct cobj_enum),
	enum_parser, enum_ser, enum_funref
};

struct xref_pstable cobj_date_pstable = {
	sizeof (struct cobj_date),
	date_parser, date_ser, NULL
};

struct xref_pstable cobj_rect_pstable = {
	sizeof (struct cobj_rect),
	rect_parser, rect_ser, NULL
};

struct xref_pstable cobj_matrix_pstable = {
	sizeof (struct cobj_matrix),
	matrix_parser, matrix_ser, NULL
};

struct xref_pstable cobj_matrix9_pstable = {
	sizeof (struct cobj_matrix),
	matrix_parser, matrix9_ser, NULL
};

struct xref_pstable cobj_id_pstable = {
	sizeof (struct cobj_id),
	id_parser, id_ser, id_funref
};

static enum s_status enum_get_special (struct cobj_enum_desc *desc,
				       int *defvalp, int *psvalp, int *unkvalp,
				       struct xref_pstable **psp)
{
	struct cobj_enum_desc *dp;
	struct xref_pstable *ps;
	int defval, unkval, psval;

	ps = NULL;
	defval = psval = unkval = -1;
	for (dp = desc; dp -> name != NULL; dp++) {
		if (!strcmp (dp -> name, "/Default"))
			defval = dp -> value;
		else if (!strcmp (dp -> name, "/Unknown"))
			unkval = dp -> value;
		else if (!strcmp (dp -> name, "/PSTable")) {
			psval = dp -> value;
			ps = dp -> data;
		} else if (*dp -> name == '/')
			return s_make_indirect (S_CSC_INTERNAL_ERROR,
						"Bad special enum %s",
						dp -> name);
	}

	if (defval < 0)
		defval = 0;

	if (defvalp != NULL)
		*defvalp = defval;

	if (unkvalp != NULL)
		*unkvalp = unkval;

	if (psvalp != NULL)
		*psvalp = psval;

	if (psp != NULL)
		*psp = ps;

	return S_CSC_OK;
}

static enum s_status enum_parser (struct xref_db *rdb, struct xref_bind *robj, 
				  struct ref_owner *p, struct obj *src, 
				  struct xref_field *fd,
				  const char *fname, int line)
{
	struct cobj_enum_desc *desc;
	struct xref_field sfd;
	struct xref_pstable *ps;
	struct cobj_enum *coe;
	enum s_status res;
	int defval, psval, unkval;

	desc = fd -> data;
	coe = fd -> field;

	res = enum_get_special (desc, &defval, &psval, &unkval, &ps);
	if (res != S_CSC_OK)
		return res;

	if (src -> type == OT_NULL) {
		coe -> type = defval;
		return S_CSC_OK;
	}

	if (src -> type != OT_NAME) {
		if (ps == NULL)
			return obj_error (src, "Expected name (for enum)");

		xref_init_field (&sfd, &coe -> t, NULL);

		res = ps -> parser (rdb, robj, p, src, &sfd, fname, line);
		if (res != S_CSC_OK)
			return res;

		coe -> type = psval;

		return S_CSC_OK;
	}

	while (desc -> name != NULL) {
		if (obj_string_equal (&src -> val.string, desc -> name)) {
			coe -> type = desc -> value;
			return S_CSC_OK;
		}
		desc++;
	}

	if (desc -> name == NULL) {
		if (unkval < 0)
			return obj_error (src, "Unknown enum value");
		coe -> type = unkval;
		res = obj_dup (src, &coe -> unkval); 
		if (res != S_CSC_OK)
			return res;
	}

	return S_CSC_OK;
}

static enum s_status enum_ser (struct xref_db *rdb, struct xref_bind *robj,
			       struct v_selection *sel, struct obj *dest, 
			       enum xref_changed *changed, 
			       struct xref_field *fd)
{
	struct cobj_enum_desc *desc, *dp;
	struct xref_field sfd;
	struct xref_pstable *ps;
	struct cobj_enum *ce;
	struct obj_bstr str;
	int defval, psval, unkval;
	enum s_status res;

	ce = fd -> field;
	desc = fd -> data;
	enum_get_special (desc, &defval, &psval, &unkval, &ps);
	if (ce -> type == defval && fd -> state == &xref_optional_tag)
		return S_CSC_OK;

	for (dp = desc; dp -> name != NULL; dp++)
		if (dp -> value == ce -> type)
			break;

	if (dp -> name == NULL)
		return s_make_indirect (S_CSC_INTERNAL_ERROR,
					"Bad enum requested");

	if (sel != NULL) {
		if (!v_supported (sel -> pref, dp -> major, dp -> minor))
			return s_make_indirect 
				(S_CSC_NOT_SUPPORTED,
				 "Need at least PDF %d.%d to store this",
				 dp -> major, dp -> minor);

		v_upgrade (sel, dp -> major, dp -> minor);
	}

		
	if (ce -> type == psval) {
		xref_init_field (&sfd, &ce -> t, NULL);

		res = ps -> serializer (rdb, robj, sel, dest, changed, &sfd);
		if (res != S_CSC_OK)
			return res;
			
	} else if (ce -> type == unkval) {
		res = obj_dup (&ce -> unkval, dest);
		if (res != S_CSC_OK)
			return res;
	} else {
		res = obj_string_from_c (dp -> name, &str);
		if (res != S_CSC_OK)
			return res;
		obj_set_name (dest, &f_null, &str);
	}

	return S_CSC_OK;
}

static void enum_funref (struct ref_owner *p, struct xref_field *fd)
{
	struct xref_pstable *ps;
	struct xref_field sfd;
	struct cobj_enum *ce;
	int psval;

	ce = fd -> field;

	if (ce -> t != NULL) {
		enum_get_special (fd -> data, NULL, &psval, NULL, &ps);
		if (psval == ce -> type) {
			xref_init_field (&sfd, &ce -> t, NULL);
			ps -> funref (p, &sfd);
			ce -> t = NULL;
		}
	}
	obj_destroy (&ce -> unkval);
}

static enum s_status rd_field (struct obj *src, const char *buf,
			       int *i, int width, int min, int max,
			       const char *fname, int *dest)
{
	int j;

	if (!buf [*i]) {
		*dest = min;
		return S_CSC_OK;
	}

	*dest = 0;
	for (j = 0; j < width; j++) {
		if (!buf [*i])
			return obj_error (src, "%s length is wrong", fname);

		if (!isdigit (buf [*i]))
			return obj_error (src, "%s is not a number", fname);
		*dest *= 10;
		*dest += buf [*i] - '0';
		(*i)++;
	}

	if (*dest < min || *dest > max)
		return obj_error (src, "%s out of range", fname);

	return S_CSC_OK;
}

static enum s_status date_parser (struct xref_db *rdb, struct xref_bind *robj, 
				  struct ref_owner *p, struct obj *src, 
				  struct xref_field *fd,
				  const char *fname, int line)
{
	char buf [D_WIDTH + 1];
	enum s_status res;
	struct cobj_date *ddate;
	struct tm tm;
	int sign;
	int i;

	ddate = fd -> field;
	if (src -> type == OT_NULL) {
		memset (ddate, 0, sizeof (struct cobj_date));
		return S_CSC_OK;
	}

	if (src -> type != OT_STRING)
		return obj_error (src, "Expected string (for date)");

	if (src -> val.string.size < D_PREFIX_WIDTH + D_YEAR_WIDTH ||
	    memchr (src -> val.string.s, 0, src -> val.string.size) != NULL)
		return obj_error (src, "Does not look like a date");

	memset (buf, 0, sizeof (buf));
	memcpy (buf, src -> val.string.s, src -> val.string.size);

	memset (&tm, 0, sizeof (struct tm));
	tm.tm_isdst = 0;

	i = 0;
	if (memcmp (buf + i, "D:", D_PREFIX_WIDTH))
		return obj_error (src, "Bad date prefix");

	i += D_PREFIX_WIDTH;
	res = rd_field (src, buf, &i, D_YEAR_WIDTH, 0, 9999, 
			"year", &tm.tm_year);
	if (res != S_CSC_OK)
		return res;
	tm.tm_year += 1900;

	res = rd_field (src, buf, &i, D_MONTH_WIDTH, 1, 12, 
			"month", &tm.tm_mon);
	if (res != S_CSC_OK)
		return res;
	tm.tm_mon--;

	res = rd_field (src, buf, &i, D_DAY_WIDTH, 1, 31, 
			"day", &tm.tm_mday);
	if (res != S_CSC_OK)
		return res;

	res = rd_field (src, buf, &i, D_HOUR_WIDTH, 0, 23, 
			"hour", &tm.tm_hour);
	if (res != S_CSC_OK)
		return res;

	res = rd_field (src, buf, &i, D_MIN_WIDTH, 0, 59, 
			"minute", &tm.tm_min);
	if (res != S_CSC_OK)
		return res;

	res = rd_field (src, buf, &i, D_SEC_WIDTH, 0, 59, 
			"seconds", &tm.tm_sec);
	if (res != S_CSC_OK)
		return res;
	

	if (buf [i] == '+') {
		i++;
		sign = 1;
	} else if (buf [i] == '-') {
		i++;
		sign = -1;
	} else if (buf [i] == 'Z') {
		i++;
		sign = 0;
	} else if (!buf [i])
		sign = 0;
	else
		return obj_error (src, "Bad UT relationship");

	res = rd_field (src, buf, &i, D_TZ_HOUR_WIDTH, 0, 23, 
			"tz hours", &ddate -> tz_hour);
	if (res != S_CSC_OK)
		return res;

	if (buf [i] == '\'')
		i++;
	else if (buf [i])
		return obj_error (src, "Bad UT separator");

	res = rd_field (src, buf, &i, D_TZ_MIN_WIDTH, 0, 59, 
			"tz minutes", &ddate -> tz_minutes);
	if (res != S_CSC_OK)
		return res;
	
	ddate -> tz_hour *= sign;
	ddate -> tz_minutes *= sign;

	ddate -> time = mktime (&tm);

	return S_CSC_OK;
}

static enum s_status date_ser (struct xref_db *rdb, struct xref_bind *robj,
			       struct v_selection *sel, struct obj *dest, 
			       enum xref_changed *changed, 
			       struct xref_field *fd)
{
	struct cobj_date *date;
	char buf [D_WIDTH + 1];
	struct obj_bstr str;
	struct tm *tm;
	enum s_status res;

	date = fd -> field;

	if (date -> time <= 0)
		return S_CSC_OK;

	tm = gmtime (&date -> time);
	snprintf (buf, sizeof (buf), 
		  "D:%.04d%.02d%.02d%.02d%.02d%.02d%c%.02d'%.02d",
		  tm -> tm_year + 1900, tm -> tm_mon + 1, tm -> tm_mday,
		  tm -> tm_hour, tm -> tm_min, tm -> tm_sec,
		  date -> tz_hour > 0 ? '+' : date -> tz_hour < 0 ? '-' : 'Z',
		  abs (date -> tz_hour), abs (date -> tz_minutes));

	res = obj_string_from_c (buf, &str);
	if (res != S_CSC_OK)
		return res;

	obj_set_string (dest, &f_null, &str);

	return S_CSC_OK;
}

void cobj_date_now (struct cobj_date *date)
{
	struct timezone tz;
	struct timeval tv;

	if (!gettimeofday (&tv, &tz)) {
		date -> time = tv.tv_sec;
		date -> tz_hour = tz.tz_minuteswest / 60;
		date -> tz_minutes = tz.tz_minuteswest % 60;
	} else {
		memset (date, 0, sizeof (struct cobj_date));
		time (&date -> time);
	}
}

#define IDX(x, y) (3 * x + y)

static enum s_status matrix_parser (struct xref_db *rdb, 
				    struct xref_bind *robj, 
				    struct ref_owner *p, 
				    struct obj *src, 
				    struct xref_field *fd,
				    const char *fname, int line)
{
	struct cobj_matrix *dest;
	struct xref_real_array ma;
	enum s_status res;

	dest = fd -> field;

	if (src -> type != OT_NULL &&
	    src -> type != OT_ARRAY)
		return obj_error (src, "Expecting matrix array");

	memset (dest, 0, sizeof (struct cobj_matrix));
		
	res = xref_array_map_ps (rdb, NULL, NULL, src, &ma, 
				 "Matrix element", &xref_real_pstable, NULL);
	if (res != S_CSC_OK)
		return res;

	switch (XREF_A_SIZE (&ma)) {
	case 0:
		cobj_matrix_set_null (dest);
		break;

	case 6:
		dest -> a [0][0] = ma.data [0];
		dest -> a [0][1] = ma.data [1];
		dest -> a [1][0] = ma.data [2];
		dest -> a [1][1] = ma.data [3];
		dest -> a [2][0] = ma.data [4];
		dest -> a [2][1] = ma.data [5];
		break;

	case 9:
		dest -> a [0][0] = ma.data [0];
		dest -> a [0][1] = ma.data [1];
		dest -> a [0][2] = ma.data [2];
		dest -> a [1][0] = ma.data [3];
		dest -> a [1][1] = ma.data [4];
		dest -> a [1][2] = ma.data [5];
		dest -> a [2][0] = ma.data [6];
		dest -> a [2][1] = ma.data [7];
		dest -> a [2][2] = ma.data [8];
		break;

	default:
		xref_array_map_rel_ps (&ma, NULL, &xref_real_pstable, NULL);

		return obj_error (src, "Bad number of elements in matrix");
	}

	xref_array_map_rel_ps (&ma, NULL, &xref_real_pstable, NULL);

	return S_CSC_OK;
}

static enum s_status parse_rect (struct xref_real_array *src,
				 struct cobj_rect *dest,
				 struct obj *refobj)
{
	pdf_real_t xr;

	if (XREF_A_IS_NULL (src)) {
		cobj_rect_set_null (dest);
		return S_CSC_OK;
	}

	if (XREF_A_SIZE (src) != 4)
		return obj_error (refobj, "Rect should contain four numbers");

	dest -> left = src -> data [0];
	dest -> top = src -> data [1];
	dest -> right = src -> data [2];
	dest -> bottom = src -> data [3];

	if (dest -> left > dest -> right) {
		xr = dest -> left;
		dest -> left = dest -> right;
		dest -> right = xr;
	}

	if (dest -> bottom > dest -> top) {
		xr = dest -> top;
		dest -> top = dest -> bottom;
		dest -> bottom = xr;
	}

	return S_CSC_OK;
}

static enum s_status rect_parser (struct xref_db *rdb, struct xref_bind *robj, 
				  struct ref_owner *p, struct obj *src, 
				  struct xref_field *fd,
				  const char *fname, int line)
{
	struct xref_real_array ma;
	enum s_status res;

	res = xref_array_map_ps (rdb, NULL, p, src, &ma, 
				 "Rect", &xref_real_pstable, NULL);
	if (res != S_CSC_OK)
		return res;

	res = parse_rect (&ma, fd -> field, src);
	xref_array_map_rel_ps (&ma, p, &xref_real_pstable, NULL);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status rect_ser (struct xref_db *rdb, struct xref_bind *robj,
			       struct v_selection *sel, struct obj *dest, 
			       enum xref_changed *changed, 
			       struct xref_field *fd)
{
	struct xref_real_array ma;
	pdf_real_t vec [4];
	struct cobj_rect *rect;
	enum s_status res;

	rect = fd -> field;
	if (rect == NULL || cobj_rect_is_null (rect))
		return S_CSC_OK;

	ma.n = 4;
	ma.data = vec;

	ma.data [0] = rect -> left;
	ma.data [1] = rect -> bottom;
	ma.data [2] = rect -> right;
	ma.data [3] = rect -> top;

	res = xref_array_unmap_ps (rdb, NULL, sel, &ma, 
				   dest, 0, &xref_real_pstable, NULL);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status matrix_ser (struct xref_db *rdb, struct xref_bind *robj,
				 struct v_selection *sel, struct obj *dest, 
				 enum xref_changed *changed, 
				 struct xref_field *fd)
{
	struct xref_real_array ma;
	pdf_real_t vec [6];
	struct cobj_matrix *matrix;
	enum s_status res;

	matrix = fd -> field;
	if (matrix == NULL || cobj_matrix_is_null (matrix))
		return S_CSC_OK;

	ma.n = 6;
	ma.data = vec;

	ma.data [0] = matrix -> a [0][0];
	ma.data [1] = matrix -> a [0][1];
	ma.data [2] = matrix -> a [1][0];
	ma.data [3] = matrix -> a [1][1];
	ma.data [4] = matrix -> a [2][0];
	ma.data [5] = matrix -> a [2][1];
	res = xref_array_unmap_ps (rdb, NULL, sel, &ma, 
				   dest, 0, &xref_real_pstable, NULL);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status matrix9_ser (struct xref_db *rdb, struct xref_bind *robj,
				  struct v_selection *sel, struct obj *dest, 
				  enum xref_changed *changed, 
				  struct xref_field *fd)
{
	struct xref_real_array ma;
	pdf_real_t vec [9];
	struct cobj_matrix *matrix;
	enum s_status res;

	matrix = fd -> field;
	if (matrix == NULL || cobj_matrix_is_null (matrix))
		return S_CSC_OK;

	ma.n = 9;
	ma.data = vec;

	ma.data [0] = matrix -> a [0][0];
	ma.data [1] = matrix -> a [0][1];
	ma.data [2] = matrix -> a [0][2];
	ma.data [3] = matrix -> a [1][0];
	ma.data [4] = matrix -> a [1][1];
	ma.data [5] = matrix -> a [1][2];
	ma.data [6] = matrix -> a [2][0];
	ma.data [7] = matrix -> a [2][1];
	ma.data [8] = matrix -> a [2][2];
	res = xref_array_unmap_ps (rdb, NULL, sel, &ma, 
				   dest, 0, &xref_real_pstable, NULL);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

void cobj_rect_set_null (struct cobj_rect *rect)
{
	rect -> top = 1;
	rect -> bottom = -1;
	rect -> left = 1;
	rect -> right = -1;
}

int cobj_rect_is_null (struct cobj_rect *rect)
{
	return 	rect -> left == 1 && rect -> right == -1;
}

static enum cobj_rect_overlap line_overlap (pdf_real_t a1, pdf_real_t a2,
					    pdf_real_t b1, pdf_real_t b2)
{
	if (a1 < b1) {
		if (a2 < b1)
			return COBJ_RO_NONE;
		else if (a2 < b2)
			return COBJ_RO_OVERLAP;
		else
			return COBJ_RO_OUTSIDE;
	} else if (a1 == b1) {
		if (a2 < b2)
			return COBJ_RO_INSIDE;
		else if (a2 == b2)
			return COBJ_RO_SAME;
		else
			return COBJ_RO_OUTSIDE;
	} else if (a1 <= b2) {
		if (a2 <= b2)
			return COBJ_RO_INSIDE;
		else 
			return COBJ_RO_OVERLAP;
	} else
		return COBJ_RO_NONE;
}

enum cobj_rect_overlap cobj_rect_overlaps (struct cobj_rect *rect1, 
					   struct cobj_rect *rect2)
{
	enum cobj_rect_overlap ox, oy;

	if (cobj_rect_is_null (rect1) || cobj_rect_is_null (rect2))
		return COBJ_RO_NONE;

	ox = line_overlap (rect1 -> left, rect1 -> right,
			   rect2 -> left, rect2 -> right);
	oy = line_overlap (rect1 -> bottom, rect1 -> top,
			   rect2 -> bottom, rect2 -> top);

	if (ox == COBJ_RO_NONE || oy == COBJ_RO_NONE)
		return COBJ_RO_NONE;

	if (ox == oy)
		return ox;
	else if (ox == COBJ_RO_SAME)
		return oy;
	else if (oy == COBJ_RO_SAME)
		return ox;
	else
		return COBJ_RO_OVERLAP;
}

void cobj_rect_intersect (struct cobj_rect *drect, struct cobj_rect *srect)
{
	if (cobj_rect_is_null (drect) || cobj_rect_is_null (srect)) {
		cobj_rect_set_null (drect);
		return;
	}
		
	if (srect -> top < drect -> top)
		drect -> top = srect -> top;
	if (srect -> left > drect -> left)
		drect -> left = srect -> left;

	if (srect -> bottom > drect -> bottom)
		drect -> bottom = srect -> bottom;
	if (srect -> right < drect -> right)
		drect -> right = srect -> right;

	if (drect -> left > drect -> right ||
	    drect -> bottom > drect -> top)
		cobj_rect_set_null (drect);
}

void cobj_rect_sub (struct cobj_rect *drect, struct cobj_rect *srect)
{
	if (srect -> right != drect -> right)
		drect -> left = srect -> right;
	else if (srect -> bottom != drect -> bottom)
		drect -> top = srect -> bottom;
	else if (srect -> left != drect -> left)
		drect -> right = srect -> left;
	else 
		drect -> bottom = srect -> top;
}

void cobj_rect_add_point (struct cobj_rect *drect, pdf_real_t x, 
			  pdf_real_t y)
{
	if (cobj_rect_is_null (drect)) {
		drect -> left = drect -> right = x;
		drect -> top = drect -> bottom = y;
	} else {
		if (drect -> left > x)
			drect -> left = x;
		if (drect -> right < x)
			drect -> right = x;
		if (drect -> bottom > y)
			drect -> bottom = y;
		if (drect -> top < y)
			drect -> top = y;
	}
}

pdf_real_t cobj_rect_width (struct cobj_rect *rect)
{
	return cobj_rect_is_null (rect) ? 0 : rect -> right - rect -> left;
}

pdf_real_t cobj_rect_height (struct cobj_rect *rect)
{
	return cobj_rect_is_null (rect) ? 0 : rect -> top - rect -> bottom;
}

void cobj_rect_union (struct cobj_rect *drect, struct cobj_rect *srect)
{
	if (cobj_rect_is_null (drect)) {
		*drect = *srect;
		return;
	}

	if (cobj_rect_is_null (srect))
		return;

	if (srect -> top > drect -> top)
		drect -> top = srect -> top;
	if (srect -> left < drect -> left)
		drect -> left = srect -> left;

	if (srect -> bottom < drect -> bottom)
		drect -> bottom = srect -> bottom;
	if (srect -> right > drect -> right)
		drect -> right = srect -> right;
}

void cobj_matrix_set_null (struct cobj_matrix *mat)
{
	mat -> a [2][2] = -1;
}

int cobj_matrix_is_null (struct cobj_matrix *mat)
{
	return mat -> a [2][2] == -1;
}

void cobj_matrix_mul (struct cobj_matrix *a, struct cobj_matrix *b,
		      struct cobj_matrix *ans, int red)
{
	ans -> a [0][0] = 
		a -> a [0][0] * b -> a [0][0] +
		a -> a [0][1] * b -> a [1][0];
	ans -> a [0][1] = 
		a -> a [0][0] * b -> a [0][1] +
		a -> a [0][1] * b -> a [1][1];
	ans -> a [1][0] = 
		a -> a [1][0] * b -> a [0][0] +
		a -> a [1][1] * b -> a [1][0];
	ans -> a [1][1] = 
		a -> a [1][0] * b -> a [0][1] +
		a -> a [1][1] * b -> a [1][1];
	ans -> a [2][0] = 
		a -> a [2][0] * b -> a [0][0] +
		a -> a [2][1] * b -> a [1][0] +
		                b -> a [2][0];
	ans -> a [2][1] = 
		a -> a [2][0] * b -> a [0][1] +
		a -> a [2][1] * b -> a [1][1] +
		                b -> a [2][1];
	if (!red) {
		ans -> a [0][0] += a -> a [0][2] * b -> a [2][0];
		ans -> a [0][1] += a -> a [0][2] * b -> a [2][1];
		ans -> a [1][0] += a -> a [1][2] * b -> a [2][0];
		ans -> a [1][1] += a -> a [1][2] * b -> a [2][1];
		ans -> a [2][0] += a -> a [2][2] * b -> a [2][0];
		ans -> a [2][1] += a -> a [2][2] * b -> a [2][1];

		ans -> a [0][2] = 
			a -> a [0][0] * b -> a [0][2] +
			a -> a [0][1] * b -> a [1][2] +
			a -> a [0][2] * b -> a [1][2];

		ans -> a [1][2] = 
			a -> a [1][0] * b -> a [0][2] +
			a -> a [1][1] * b -> a [1][2] +
			a -> a [1][2] * b -> a [1][2];

		ans -> a [2][2] = 
			a -> a [2][0] * b -> a [0][2] +
			a -> a [2][1] * b -> a [1][2] +
			a -> a [2][2] * b -> a [1][2];
		
	} else {
		ans -> a [0][2] = ans -> a [1][2] = 0;
		ans -> a [2][2] = 1;
	}
}

void cobj_matrix_inv (struct cobj_matrix *a, struct cobj_matrix *ans, int red)
{
	pdf_real_t det;

	if (red) {
		det = 	a -> a [0][0] * a -> a [1][1] -
			a -> a [0][1] * a -> a [1][0];
		ans -> a [0][0] =   a -> a [1][1] / det;
		ans -> a [0][1] = - a -> a [0][1] / det;
		ans -> a [1][0] = - a -> a [1][0] / det;
		ans -> a [1][1] =   a -> a [0][0] / det;
		ans -> a [2][0] =   (a -> a [1][0] * a -> a [2][1] -
				     a -> a [1][1] * a -> a [2][0]) / det;
		ans -> a [2][1] = - (a -> a [0][0] * a -> a [2][1] -
				     a -> a [0][1] * a -> a [2][0]) / det;
		ans -> a [0][2] = ans -> a [1][2] = 0;
		ans -> a [2][2] = 1;
	} else {
		det = 	a -> a [0][0] * a -> a [1][1] * a -> a [2][2]
			+ a -> a [0][1] * a -> a [1][2] * a -> a [2][0]
			+ a -> a [0][2] * a -> a [1][0] * a -> a [2][1]
			- a -> a [0][0] * a -> a [1][2] * a -> a [2][1]
			- a -> a [0][1] * a -> a [1][0] * a -> a [2][2]
			- a -> a [0][2] * a -> a [1][1] * a -> a [2][0];

		ans -> a [0][0] =   (a -> a [1][1] * a -> a [2][2] -
				     a -> a [1][2] * a -> a [2][1]) / det;
		ans -> a [0][1] = - (a -> a [0][1] * a -> a [2][2] -
				     a -> a [0][2] * a -> a [2][1]) / det;
		ans -> a [0][2] =   (a -> a [0][1] * a -> a [1][2] -
				     a -> a [0][2] * a -> a [1][1]) / det;
		ans -> a [1][0] = - (a -> a [1][0] * a -> a [2][2] -
				     a -> a [1][2] * a -> a [2][0]) / det;
		ans -> a [1][1] =   (a -> a [0][0] * a -> a [2][2] -
				     a -> a [0][2] * a -> a [2][0]) / det;
		ans -> a [1][2] = - (a -> a [0][0] * a -> a [1][2] -
				     a -> a [0][2] * a -> a [1][0]) / det;
		ans -> a [2][0] =   (a -> a [1][0] * a -> a [2][1] -
				     a -> a [1][1] * a -> a [2][0]) / det;
		ans -> a [2][1] = - (a -> a [0][0] * a -> a [2][1] -
				     a -> a [0][1] * a -> a [2][0]) / det;
		ans -> a [2][2] =   (a -> a [0][0] * a -> a [1][1] -
				     a -> a [0][1] * a -> a [1][0]) / det;
	}
}

/* A * B^-1 */
void cobj_matrix_div (struct cobj_matrix *a, struct cobj_matrix *b,
		      struct cobj_matrix *ans, int red)
{
	struct cobj_matrix binv;

	cobj_matrix_inv (b, &binv, red);
	cobj_matrix_mul (a, &binv, ans, red);
}

static enum s_status id_parser (struct xref_db *rdb, struct xref_bind *robj, 
				struct ref_owner *p, struct obj *src, 
				struct xref_field *fd,
				const char *fname, int line)
{
	struct xref_obj_array ma;
	struct cobj_id *id;
	enum s_status res;

	id = fd -> field;
	
	res = xref_array_map_ps (rdb, NULL, p, src, &ma, 
				 "ID", &xref_obj_pstable, NULL);
	if (res != S_CSC_OK)
		return res;

	if (XREF_A_IS_NULL (&ma))
		return S_CSC_OK;

	if (XREF_A_SIZE (&ma) == 2 &&
	    ma.data [0].type == OT_STRING &&
	    ma.data [1].type == OT_STRING) {
		id -> id1 = ma.data [0].val.string;
		id -> id2 = ma.data [1].val.string;

		obj_set_null (&ma.data [0], &f_null);
		obj_set_null (&ma.data [1], &f_null);

		res = S_CSC_OK;
	} else
		res = obj_error (src, "Bad ID array");

	xref_array_map_rel_ps (&ma, p, &xref_obj_pstable, NULL);

	return res;
}

static enum s_status id_ser (struct xref_db *rdb, struct xref_bind *robj,
			     struct v_selection *sel, struct obj *dest, 
			     enum xref_changed *changed, struct xref_field *fd)
{
	struct xref_obj_array ma;
	struct obj vec [2];
	struct cobj_id *id;
	enum s_status res;

	id = fd -> field;
	if (id == NULL || !id -> id1.s || !id -> id2.s)
		return S_CSC_OK;

	ma.n = 2;
	ma.data = vec;

	obj_set_string (&ma.data [0], &f_null, &id -> id1);
	obj_set_string (&ma.data [1], &f_null, &id -> id2);

	res = xref_array_unmap_ps (rdb, NULL, sel, &ma, 
				   dest, 0, &xref_obj_pstable, NULL);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static void id_funref (struct ref_owner *p, struct xref_field *fd)
{
	struct cobj_id *id;

	id = fd -> field;

	obj_string_destroy (&id -> id1);
	obj_string_destroy (&id -> id2);
}

enum s_status cobj_id_cross (struct cobj_id *id, struct cobj_id *sid,
			     struct ob_bridge *ob)
{
	struct obj_bstr id1, id2;
	enum s_status res;

	res = obj_string_dup (&sid -> id1, &id1);
	if (res != S_CSC_OK)
		return res;

	res = obj_string_dup (&sid -> id2, &id2);
	if (res != S_CSC_OK) {
		obj_string_destroy (&id1);
		return res;
	}

	obj_string_destroy (&id -> id1);
	obj_string_destroy (&id -> id2);

	id -> id1 = id1;
	id -> id2 = id2;

	return S_CSC_OK;
}

void cobj_translate (struct cobj_rect *rect, struct cobj_matrix *mat)
{
	struct cobj_rect drect;

	drect.left = 
		mat -> a [0][0] * rect -> left +
		mat -> a [1][0] * rect -> top +
		mat -> a [2][0];

	drect.top = 
		mat -> a [0][1] * rect -> left +
		mat -> a [1][1] * rect -> top +
		mat -> a [2][1];

	drect.right = 
		mat -> a [0][0] * rect -> right +
		mat -> a [1][0] * rect -> bottom +
		mat -> a [2][0];

	drect.bottom = 
		mat -> a [0][1] * rect -> right +
		mat -> a [1][1] * rect -> bottom +
		mat -> a [2][1];

	*rect = drect;
}
