/*
 * PDFlow -- colour.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>
#include <stdlib.h>

#include "colour_pst.h"
#include "types/pattern.h"

static struct obj def_cspace;

struct gc_colour_space_ifc {

	const char *defname;

	struct xref_field *fd;

	int nested;

	int (*get_components) (struct gc_colour_space *sp);

	enum s_status (*init) (struct gc_colour_space *sp,
			       struct gc_colour_state *cst,
			       struct xref_db *rdb,
			       struct rd_resdict *rd);

	void (*select) (struct gc_colour_space *sp,
			struct gc_colour_state *cst, 
			pdf_real_t *val);

	void (*retrieve) (struct gc_colour_space *sp,
			  struct gc_colour_state *cst, 
			  pdf_real_t *val);

	enum s_status (*dup) (struct gc_colour_space *sp,
			      struct gc_colour_state *src,
			      struct gc_colour_state *dest);

	void (*destroy) (struct gc_colour_space *sp,
			 struct gc_colour_state *cst);
};

static enum s_status dyn_ifc (enum xref_op op, struct xref_field *fd, 
			      void *data, int *stop);

static int csi_null_get_components (struct gc_colour_space *sp);
static int csi_device_gray_get_components (struct gc_colour_space *sp);
static int csi_device_rgb_get_components (struct gc_colour_space *sp);
static int csi_device_cmyk_get_components (struct gc_colour_space *sp);
static int csi_cal_gray_get_components (struct gc_colour_space *sp);
static int csi_cal_rgb_get_components (struct gc_colour_space *sp);
static int csi_lab_get_components (struct gc_colour_space *sp);
static int csi_icc_based_get_components (struct gc_colour_space *sp);
static int csi_pattern_get_components (struct gc_colour_space *sp);
static int csi_indexed_get_components (struct gc_colour_space *sp);
static int csi_separation_get_components (struct gc_colour_space *sp);
static int csi_device_n_get_components (struct gc_colour_space *sp);

static enum s_status csi_null_init (struct gc_colour_space *sp,
				    struct gc_colour_state *cst,
				    struct xref_db *rdb,
				    struct rd_resdict *rd);
static enum s_status csi_device_gray_init (struct gc_colour_space *sp,
					   struct gc_colour_state *cst,
					   struct xref_db *rdb,
					   struct rd_resdict *rd);
static enum s_status csi_device_rgb_init (struct gc_colour_space *sp,
					  struct gc_colour_state *cst,
					  struct xref_db *rdb,
					  struct rd_resdict *rd);
static enum s_status csi_device_cmyk_init (struct gc_colour_space *sp,
					   struct gc_colour_state *cst,
					   struct xref_db *rdb,
					   struct rd_resdict *rd);
static enum s_status csi_cal_gray_init (struct gc_colour_space *sp,
					struct gc_colour_state *cst,
					struct xref_db *rdb,
					struct rd_resdict *rd);
static enum s_status csi_cal_rgb_init (struct gc_colour_space *sp,
				       struct gc_colour_state *cst,
				       struct xref_db *rdb,
				       struct rd_resdict *rd);
static enum s_status csi_lab_init (struct gc_colour_space *sp,
				   struct gc_colour_state *cst,
				   struct xref_db *rdb,
				   struct rd_resdict *rd);
static enum s_status csi_icc_based_init (struct gc_colour_space *sp,
					 struct gc_colour_state *cst,
					 struct xref_db *rdb,
					 struct rd_resdict *rd);
static enum s_status csi_pattern_init (struct gc_colour_space *sp,
				       struct gc_colour_state *cst,
				       struct xref_db *rdb,
				       struct rd_resdict *rd);
static enum s_status csi_indexed_init (struct gc_colour_space *sp,
				       struct gc_colour_state *cst,
				       struct xref_db *rdb,
				       struct rd_resdict *rd);
static enum s_status csi_separation_init (struct gc_colour_space *sp,
					  struct gc_colour_state *cst,
					  struct xref_db *rdb,
					  struct rd_resdict *rd);
static enum s_status csi_device_n_init (struct gc_colour_space *sp,
					struct gc_colour_state *cst,
					struct xref_db *rdb,
					struct rd_resdict *rd);

static void csi_null_select (struct gc_colour_space *sp,
			     struct gc_colour_state *cst, 
			     pdf_real_t *val);
static void csi_device_gray_select (struct gc_colour_space *sp,
				    struct gc_colour_state *cst, 
				    pdf_real_t *val);
static void csi_device_rgb_select (struct gc_colour_space *sp,
				   struct gc_colour_state *cst, 
				   pdf_real_t *val);
static void csi_device_cmyk_select (struct gc_colour_space *sp,
				    struct gc_colour_state *cst, 
				    pdf_real_t *val);
static void csi_cal_gray_select (struct gc_colour_space *sp,
				 struct gc_colour_state *cst, 
				 pdf_real_t *val);
static void csi_cal_rgb_select (struct gc_colour_space *sp,
				struct gc_colour_state *cst, 
				pdf_real_t *val);
static void csi_lab_select (struct gc_colour_space *sp,
			    struct gc_colour_state *cst, 
			    pdf_real_t *val);
static void csi_icc_based_select (struct gc_colour_space *sp,
				  struct gc_colour_state *cst, 
				  pdf_real_t *val);
static void csi_pattern_select (struct gc_colour_space *sp,
				struct gc_colour_state *cst, 
				pdf_real_t *val);
static void csi_indexed_select (struct gc_colour_space *sp,
				struct gc_colour_state *cst, 
				pdf_real_t *val);
static void csi_separation_select (struct gc_colour_space *sp,
				   struct gc_colour_state *cst, 
				   pdf_real_t *val);
static void csi_device_n_select (struct gc_colour_space *sp,
				 struct gc_colour_state *cst, 
				 pdf_real_t *val);

static void csi_null_retrieve (struct gc_colour_space *sp,
			       struct gc_colour_state *cst, 
			       pdf_real_t *val);
static void csi_device_gray_retrieve (struct gc_colour_space *sp,
				      struct gc_colour_state *cst, 
				      pdf_real_t *val);
static void csi_device_rgb_retrieve (struct gc_colour_space *sp,
				     struct gc_colour_state *cst, 
				     pdf_real_t *val);
static void csi_device_cmyk_retrieve (struct gc_colour_space *sp,
				      struct gc_colour_state *cst, 
				      pdf_real_t *val);
static void csi_cal_gray_retrieve (struct gc_colour_space *sp,
				   struct gc_colour_state *cst, 
				   pdf_real_t *val);
static void csi_cal_rgb_retrieve (struct gc_colour_space *sp,
				  struct gc_colour_state *cst, 
				  pdf_real_t *val);
static void csi_lab_retrieve (struct gc_colour_space *sp,
			      struct gc_colour_state *cst, 
			      pdf_real_t *val);
static void csi_icc_based_retrieve (struct gc_colour_space *sp,
				    struct gc_colour_state *cst, 
				    pdf_real_t *val);
static void csi_pattern_retrieve (struct gc_colour_space *sp,
				  struct gc_colour_state *cst, 
				  pdf_real_t *val);
static void csi_indexed_retrieve (struct gc_colour_space *sp,
				  struct gc_colour_state *cst, 
				  pdf_real_t *val);
static void csi_separation_retrieve (struct gc_colour_space *sp,
				     struct gc_colour_state *cst, 
				     pdf_real_t *val);
static void csi_device_n_retrieve (struct gc_colour_space *sp,
				   struct gc_colour_state *cst, 
				   pdf_real_t *val);

static enum s_status csi_null_dup (struct gc_colour_space *sp,
				   struct gc_colour_state *src,
				   struct gc_colour_state *dest);
static enum s_status csi_pattern_dup (struct gc_colour_space *sp,
				      struct gc_colour_state *src,
				      struct gc_colour_state *dest);
static enum s_status csi_device_n_dup (struct gc_colour_space *sp,
				       struct gc_colour_state *src,
				       struct gc_colour_state *dest);

static void csi_null_destroy (struct gc_colour_space *sp,
			      struct gc_colour_state *cst);
static void csi_pattern_destroy (struct gc_colour_space *sp,
				 struct gc_colour_state *cst);
static void csi_device_n_destroy (struct gc_colour_space *sp,
				  struct gc_colour_state *cst);

static struct cobj_enum_desc cspace_desc [] = {
	{ GC_CSF_DEVICE_GRAY, "DeviceGray" },
	{ GC_CSF_DEVICE_RGB, "DeviceRGB" },
	{ GC_CSF_DEVICE_CMYK, "DeviceCMYK" },

	{ GC_CSF_CAL_GRAY, "CalGray" },
	{ GC_CSF_CAL_RGB, "CalRGB" },
	{ GC_CSF_LAB, "Lab" },
	{ GC_CSF_ICC_BASED, "ICCBased" },

	{ GC_CSF_PATTERN, "Pattern" },
	{ GC_CSF_INDEXED, "Indexed" },
	{ GC_CSF_SEPARATION, "Separation" },
	{ GC_CSF_DEVICE_N, "DeviceN" },

	{ GC_CSF_OTHER, "/Unknown" },

	{ -1, NULL }
};

#define XREF_IMPLEMENTATION struct gc_colour_space
#define GC_STUB
#include "types/colour.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct gc_colour_space
#define GC_COLOUR_SPACE
#include "types/colour.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION <none>
#define GC_D_CAL_GRAY
#include "types/colour.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION <none>
#define GC_D_CAL_RGB
#include "types/colour.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION <none>
#define GC_D_LAB
#include "types/colour.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION <none>
#define GC_D_ICC_BASED
#include "types/colour.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct gc_d_pattern
#define GC_D_PATTERN
#include "types/colour.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct gc_d_indexed
#define GC_D_INDEXED
#include "types/colour.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct gc_d_separation
#define GC_D_SEPARATION
#include "types/colour.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct gc_d_device_n
#define GC_D_DEVICE_N
#include "types/colour.tab.h"
#undef XREF_IMPLEMENTATION

static struct gc_colour_space_ifc csi_null = {
	NULL, NULL, 0,

	csi_null_get_components,
	csi_null_init,
	csi_null_select,
	csi_null_retrieve,
	csi_null_dup,
	csi_null_destroy
};

static struct gc_colour_space_ifc csi_device_gray = {
	"DefaultGray", NULL, 0,

	csi_device_gray_get_components,
	csi_device_gray_init,
	csi_device_gray_select,
	csi_device_gray_retrieve,
	csi_null_dup,
	csi_null_destroy	
};

static struct gc_colour_space_ifc csi_device_rgb = {
	"DefaultRGB", NULL, 0,

	csi_device_rgb_get_components,
	csi_device_rgb_init,
	csi_device_rgb_select,
	csi_device_rgb_retrieve,
	csi_null_dup,
};

static struct gc_colour_space_ifc csi_device_cmyk = {
	"DefaultCMYK", NULL, 0,

	csi_device_cmyk_get_components,
	csi_device_cmyk_init,
	csi_device_cmyk_select,
	csi_device_cmyk_retrieve,
	csi_null_dup,
	csi_null_destroy
};

static struct gc_colour_space_ifc csi_cal_gray = {
	NULL, &gc_d_cal_gray_field, 1,

	csi_cal_gray_get_components,
	csi_cal_gray_init,
	csi_cal_gray_select,
	csi_cal_gray_retrieve,
	csi_null_dup
};
 
static struct gc_colour_space_ifc csi_cal_rgb = {
	NULL, &gc_d_cal_rgb_field, 1,

	csi_cal_rgb_get_components,
	csi_cal_rgb_init,
	csi_cal_rgb_select,
	csi_cal_rgb_retrieve,
	csi_null_dup,
	csi_null_destroy
};

static struct gc_colour_space_ifc csi_lab  = {
	NULL, &gc_d_lab_field, 1,

	csi_lab_get_components,
	csi_lab_init,
	csi_lab_select,
	csi_lab_retrieve,
	csi_null_dup,
	csi_null_destroy
};

static struct gc_colour_space_ifc csi_icc_based = {
	NULL, &gc_d_icc_based_field, 1,

	csi_icc_based_get_components,
	csi_icc_based_init,
	csi_icc_based_select,
	csi_icc_based_retrieve,
	csi_null_dup,
	csi_null_destroy
};

static struct gc_colour_space_ifc csi_pattern = {
	NULL, &gc_d_pattern_field, 0,

	csi_pattern_get_components,
	csi_pattern_init,
	csi_pattern_select,
	csi_pattern_retrieve,
	csi_pattern_dup,
	csi_pattern_destroy
};

static struct gc_colour_space_ifc csi_indexed = {
	NULL, &gc_d_indexed_field, 0,

	csi_indexed_get_components,
	csi_indexed_init,
	csi_indexed_select,
	csi_indexed_retrieve,
	csi_null_dup,
	csi_null_destroy
};

static struct gc_colour_space_ifc csi_separation = {
	NULL, &gc_d_separation_field, 0,

	csi_separation_get_components,
	csi_separation_init,
	csi_separation_select,
	csi_separation_retrieve,
	csi_null_dup,
	csi_null_destroy
};

static struct gc_colour_space_ifc csi_device_n = {
	NULL, &gc_d_device_n_field, 0,

	csi_device_n_get_components,
	csi_device_n_init,
	csi_device_n_select,
	csi_device_n_retrieve,
	csi_device_n_dup,
	csi_device_n_destroy
};

static struct gc_colour_space_ifc *ifcs [] = {
	&csi_null, 
	&csi_device_gray, &csi_device_rgb, &csi_device_cmyk,
	&csi_cal_gray, &csi_cal_rgb, &csi_lab, &csi_icc_based, 
	&csi_pattern, &csi_indexed, &csi_separation, &csi_device_n, 
	NULL
};

T_SUGAR_IMPLEMENTATIONS (gc_cspace, gc_colour_space)

/* -- NULL colour -- */
static int csi_null_get_components (struct gc_colour_space *sp)
{
	return 0;
}

static enum s_status csi_null_init (struct gc_colour_space *sp,
				    struct gc_colour_state *cst,
				    struct xref_db *rdb,
				    struct rd_resdict *rd)
{
	return S_CSC_OK;
}

static void csi_null_select (struct gc_colour_space *sp,
			     struct gc_colour_state *cst, 
			     pdf_real_t *val)
{
	/* empty */
}

static void csi_null_retrieve (struct gc_colour_space *sp,
			       struct gc_colour_state *cst, 
			       pdf_real_t *val)
{
	/* empty */
}

static void csi_null_destroy (struct gc_colour_space *sp,
			      struct gc_colour_state *cst)
{
	/* empty */
}

static enum s_status csi_null_dup (struct gc_colour_space *sp,
				   struct gc_colour_state *src,
				   struct gc_colour_state *dest)
{
	return S_CSC_OK;
}

/* -- Device gray colour -- */
static int csi_device_gray_get_components (struct gc_colour_space *sp)
{
	return 1;
}

static enum s_status csi_device_gray_init (struct gc_colour_space *sp,
					   struct gc_colour_state *cst,
					   struct xref_db *rdb,
					   struct rd_resdict *rd)
{
	cst -> u.device_gray.gray = 0;

	return S_CSC_OK;
}

static void csi_device_gray_select (struct gc_colour_space *sp,
				    struct gc_colour_state *cst, 
				    pdf_real_t *val)
{
	cst -> u.device_gray.gray = val [0];
}

static void csi_device_gray_retrieve (struct gc_colour_space *sp,
				      struct gc_colour_state *cst, 
				      pdf_real_t *val)
{
	val [0] = cst -> u.device_gray.gray;
}

/* -- Device RGB colour -- */
static int csi_device_rgb_get_components (struct gc_colour_space *sp)
{
	return 3;
}

static enum s_status csi_device_rgb_init (struct gc_colour_space *sp,
					  struct gc_colour_state *cst,
					  struct xref_db *rdb,
					  struct rd_resdict *rd)
{
	cst -> u.device_rgb.red = 0;
	cst -> u.device_rgb.blue = 0;
	cst -> u.device_rgb.green = 0;

	return S_CSC_OK;
}

static void csi_device_rgb_select (struct gc_colour_space *sp,
				   struct gc_colour_state *cst, 
				   pdf_real_t *val)
{
	cst -> u.device_rgb.red = val [0];
	cst -> u.device_rgb.green = val [1];
	cst -> u.device_rgb.blue = val [2];
}

static void csi_device_rgb_retrieve (struct gc_colour_space *sp,
				     struct gc_colour_state *cst, 
				     pdf_real_t *val)
{
	val [0] = cst -> u.device_rgb.red;
	val [1] = cst -> u.device_rgb.green;
	val [2] = cst -> u.device_rgb.blue;
}

/* -- Device CMYK colour -- */
static int csi_device_cmyk_get_components (struct gc_colour_space *sp)
{
	return 4;
}

static enum s_status csi_device_cmyk_init (struct gc_colour_space *sp,
					   struct gc_colour_state *cst,
					   struct xref_db *rdb,
					   struct rd_resdict *rd)
{
	cst -> u.device_cmyk.cyan = 0;
	cst -> u.device_cmyk.magenta = 0;
	cst -> u.device_cmyk.yellow = 0;
	cst -> u.device_cmyk.black = 1;
	
	return S_CSC_OK;
}

static void csi_device_cmyk_select (struct gc_colour_space *sp,
				    struct gc_colour_state *cst, 
				    pdf_real_t *val)
{
	cst -> u.device_cmyk.cyan = val [0];
	cst -> u.device_cmyk.magenta = val [1];
	cst -> u.device_cmyk.yellow = val [2];
	cst -> u.device_cmyk.black = val [3];
}

static void csi_device_cmyk_retrieve (struct gc_colour_space *sp,
				      struct gc_colour_state *cst, 
				      pdf_real_t *val)
{
	val [0] = cst -> u.device_cmyk.cyan;
	val [1] = cst -> u.device_cmyk.magenta;
	val [2] = cst -> u.device_cmyk.yellow;
	val [3] = cst -> u.device_cmyk.black;
}

/* -- Cal gray colour -- */
static int csi_cal_gray_get_components (struct gc_colour_space *sp)
{
	return 1;
}

static enum s_status csi_cal_gray_init (struct gc_colour_space *sp,
					struct gc_colour_state *cst,
					struct xref_db *rdb,
					struct rd_resdict *rd)
{
	cst -> u.cal_gray.a = 0;

	return S_CSC_OK;
}

static void csi_cal_gray_select (struct gc_colour_space *sp,
				 struct gc_colour_state *cst, 
				 pdf_real_t *val)
{
	cst -> u.cal_gray.a = val [0];
}

static void csi_cal_gray_retrieve (struct gc_colour_space *sp,
				   struct gc_colour_state *cst, 
				   pdf_real_t *val)
{
	val [0] = cst -> u.cal_gray.a;
}

/* -- Cal RGB colour -- */
static int csi_cal_rgb_get_components (struct gc_colour_space *sp)
{
	return 3;
}

static enum s_status csi_cal_rgb_init (struct gc_colour_space *sp,
				       struct gc_colour_state *cst,
				       struct xref_db *rdb,
				       struct rd_resdict *rd)
{
	cst -> u.cal_rgb.a = 0;
	cst -> u.cal_rgb.b = 0;
	cst -> u.cal_rgb.c = 0;

	return S_CSC_OK;
}

static void csi_cal_rgb_select (struct gc_colour_space *sp,
				struct gc_colour_state *cst, 
				pdf_real_t *val)
{
	cst -> u.cal_rgb.a = val [0];
	cst -> u.cal_rgb.b = val [1];
	cst -> u.cal_rgb.c = val [2];
}

static void csi_cal_rgb_retrieve (struct gc_colour_space *sp,
				  struct gc_colour_state *cst, 
				  pdf_real_t *val)
{
	val [0] = cst -> u.cal_rgb.a;
	val [1] = cst -> u.cal_rgb.b;
	val [2] = cst -> u.cal_rgb.c;
}

/* -- Lab colour -- */
static int csi_lab_get_components (struct gc_colour_space *sp)
{
	return 3;
}

static enum s_status csi_lab_init (struct gc_colour_space *sp,
				   struct gc_colour_state *cst,
				   struct xref_db *rdb,
				   struct rd_resdict *rd)
{
	cst -> u.lab.l = 0;
	cst -> u.lab.a = 0;
	cst -> u.lab.b = 0;

	if (cst -> u.lab.a < sp -> d.cie.amin)
		cst -> u.lab.a = sp -> d.cie.amin;
	else if (cst -> u.lab.a > sp -> d.cie.amax)
		cst -> u.lab.a = sp -> d.cie.amax;

	if (cst -> u.lab.b < sp -> d.cie.bmin)
		cst -> u.lab.b = sp -> d.cie.bmin;
	else if (cst -> u.lab.b > sp -> d.cie.bmax)
		cst -> u.lab.b = sp -> d.cie.bmax;

	return S_CSC_OK;
}

static void csi_lab_select (struct gc_colour_space *sp,
			    struct gc_colour_state *cst, 
			    pdf_real_t *val)
{
	cst -> u.lab.l = val [0];
	cst -> u.lab.a = val [1];
	cst -> u.lab.b = val [2];
}

static void csi_lab_retrieve (struct gc_colour_space *sp,
			      struct gc_colour_state *cst, 
			      pdf_real_t *val)
{
	val [0] = cst -> u.lab.l;
	val [1] = cst -> u.lab.a;
	val [2] = cst -> u.lab.b;
}

/* -- ICC based colour -- */
static int csi_icc_based_get_components (struct gc_colour_space *sp)
{
	return sp -> d.icc.n;
}

static enum s_status csi_icc_based_init (struct gc_colour_space *sp,
					 struct gc_colour_state *cst,
					 struct xref_db *rdb,
					 struct rd_resdict *rd)
{
	int i;

	for (i = 0; i < sp -> d.icc.n; i++) {
		cst -> u.icc.v [i] = 0;
		if (cst -> u.icc.v [i] < sp -> d.icc.min [i])
			cst -> u.icc.v [i] = sp -> d.icc.min [i];
		else if (cst -> u.icc.v [i] > sp -> d.icc.max [i])
			cst -> u.icc.v [i] = sp -> d.icc.max [i];
	}

	return S_CSC_OK;
}

static void csi_icc_based_select (struct gc_colour_space *sp,
				  struct gc_colour_state *cst, 
				  pdf_real_t *val)
{
	memcpy (cst -> u.icc.v, val, 
		sizeof (pdf_real_t) * sp -> d.icc.n);
}

static void csi_icc_based_retrieve (struct gc_colour_space *sp,
				    struct gc_colour_state *cst, 
				    pdf_real_t *val)
{
	memcpy (val, cst -> u.icc.v, 
		sizeof (pdf_real_t) * sp -> d.icc.n);
}

/* -- Pattern colour -- */
static int csi_pattern_get_components (struct gc_colour_space *sp)
{
	struct gc_colour_space *tsp;

	tsp = sp -> d.pattern.csp;
	if (tsp == NULL)
		return 0;

	return tsp -> ifc -> get_components (tsp);
}

static enum s_status csi_pattern_init (struct gc_colour_space *sp,
				       struct gc_colour_state *cst,
				       struct xref_db *rdb,
				       struct rd_resdict *rd)
{
	struct gc_colour_space *tsp;

	tsp = sp -> d.pattern.csp;
	if (tsp == NULL)
		return S_CSC_OK;

	return tsp -> ifc -> init (tsp, cst, rdb, rd);
}

static void csi_pattern_select (struct gc_colour_space *sp,
				struct gc_colour_state *cst, 
				pdf_real_t *val)
{
	struct gc_colour_space *tsp;

	tsp = sp -> d.pattern.csp;
	if (tsp == NULL)
		return;

	tsp -> ifc -> select (tsp, cst, val);
}

static void csi_pattern_retrieve (struct gc_colour_space *sp,
				  struct gc_colour_state *cst, 
				  pdf_real_t *val)
{
	struct gc_colour_space *tsp;

	tsp = sp -> d.pattern.csp;
	if (tsp == NULL)
		return;

	tsp -> ifc -> retrieve (tsp, cst, val);
}

static void csi_pattern_destroy (struct gc_colour_space *sp,
				 struct gc_colour_state *cst)
{
	struct gc_colour_space *tsp;

	tsp = sp -> d.pattern.csp;
	if (tsp == NULL)
		return;

	return tsp -> ifc -> destroy (tsp, cst);
}

static enum s_status csi_pattern_dup (struct gc_colour_space *sp,
				      struct gc_colour_state *src,
				      struct gc_colour_state *dest)
{
	struct gc_colour_space *tsp;

	tsp = sp -> d.pattern.csp;
	if (tsp == NULL)
		return S_CSC_OK;

	return tsp -> ifc -> dup (tsp, src, dest);
}

/* -- Indexed colour -- */
static int csi_indexed_get_components (struct gc_colour_space *sp)
{
	return 1;
}

static enum s_status csi_indexed_init (struct gc_colour_space *sp,
				       struct gc_colour_state *cst,
				       struct xref_db *rdb,
				       struct rd_resdict *rd)
{
	cst -> u.indexed.index = 0;

	return S_CSC_OK;
}

static void csi_indexed_select (struct gc_colour_space *csp, 
				struct gc_colour_state *cst, 
				pdf_real_t *val)
{
	cst -> u.indexed.index = val [0];
}

static void csi_indexed_retrieve (struct gc_colour_space *csp, 
				  struct gc_colour_state *cst, 
				  pdf_real_t *val)
{
	val [0] = cst -> u.indexed.index;
}

/* -- Separation colour -- */
static int csi_separation_get_components (struct gc_colour_space *sp)
{
	return 1;
}

static enum s_status csi_separation_init (struct gc_colour_space *sp,
					  struct gc_colour_state *cst,
					  struct xref_db *rdb,
					  struct rd_resdict *rd)
{
	cst -> u.separation.value = 1;

	return S_CSC_OK;
}

static void csi_separation_select (struct gc_colour_space *csp,
				   struct gc_colour_state *cst, 
				   pdf_real_t *val)
{
	cst -> u.separation.value = val [0];
}

static void csi_separation_retrieve (struct gc_colour_space *csp,
				     struct gc_colour_state *cst, 
				     pdf_real_t *val)
{
	val [0] = cst -> u.separation.value;
}

/* -- Device N colour -- */
static int csi_device_n_get_components (struct gc_colour_space *sp)
{
	return sp -> d.device_n.n;
}

static enum s_status csi_device_n_init (struct gc_colour_space *sp,
					struct gc_colour_state *cst,
					struct xref_db *rdb,
					struct rd_resdict *rd)
{
 	pdf_real_t *value;
	int i;

	value = calloc (sp -> d.device_n.n, sizeof (pdf_real_t));
	if (value == NULL)
		return S_CSC_OUT_OF_MEMORY;

	for (i = 0; i < sp -> d.device_n.n; i++)
		value [i] = 1;

	if (cst -> u.device_n.value != NULL)
		free (cst -> u.device_n.value);

	cst -> u.device_n.value = value;

	return S_CSC_OK;
}

static void csi_device_n_select (struct gc_colour_space *sp,
				 struct gc_colour_state *cst, 
				 pdf_real_t *val)
{
	memcpy (cst -> u.device_n.value, val, 
		sizeof (pdf_real_t) * sp -> d.device_n.n);
}

static void csi_device_n_retrieve (struct gc_colour_space *sp,
				   struct gc_colour_state *cst, 
				   pdf_real_t *val)
{
	memcpy (val, cst -> u.device_n.value, 
		sizeof (pdf_real_t) * sp -> d.device_n.n);
}

static void csi_device_n_destroy (struct gc_colour_space *sp,
				  struct gc_colour_state *cst)
{
	struct gc_f_device_n *dn;

	dn = &cst -> u.device_n;

	if (dn -> value != NULL)
		free (dn -> value);
}

static enum s_status csi_device_n_dup (struct gc_colour_space *sp,
				       struct gc_colour_state *src,
				       struct gc_colour_state *dest)
{
	struct gc_f_device_n *sdn, *ddn;
 	pdf_real_t *value;
	size_t size;

	sdn = &src -> u.device_n;
	ddn = &dest -> u.device_n;
	
	size = src -> sp -> d.device_n.n * sizeof (pdf_real_t);
	value = malloc (size);
	if (value == NULL)
		return S_CSC_OUT_OF_MEMORY;

	ddn -> value = value;
	memcpy (ddn -> value, sdn -> value, size);

	return S_CSC_OK;
}

static void get_parsing_coords (struct gc_colour_space *csp, 
				struct xref_field **fd, void **base)
{
	if (csp -> ifc != NULL && 
	    csp -> ifc -> fd != NULL &&
	    csp -> ifc -> nested) {
		*fd = &gc_cspace_field;
		*base = csp;
	} else {
		*fd = csp -> ifc -> fd;
		*base = &csp -> d;
	}
}

static enum s_status dyn_ifc (enum xref_op op, struct xref_field *fd, 
			      void *data, int *stop)
{
	struct gc_colour_space *csp;
	
	csp = data;

	if (csp -> ifc == NULL) {
		*stop = 1;
		return S_CSC_OK;
	}
		
	*fd = *csp -> ifc -> fd;

	return S_CSC_OK;
}

static enum s_status init_cspace (struct gc_colour_space *csp, 
				  struct xref_db *rdb, 
				  struct xref_bind *robj,
				  struct obj *obj,
				  const char *fname, int line)
{
	struct rd_resdict *rd;
	struct xref_field *fd;
	struct obj rname;
	void *base;
	enum s_status res;

	while (1) {
		res = xref__fields_parse (rdb, robj, ref_rown (&csp -> t.refs),
					  obj, &gc_stub_field, csp,
					  fname, line);
		if (res != S_CSC_OK)
			return res;

		if (csp -> type.type != GC_CSF_OTHER)
			break;

		xref_fields_destroy (&gc_stub_field, csp, 
				     ref_rown (&csp -> t.refs));

		if (robj != NULL) {
			rd = xref_bind_get (robj, rd_bkey);
			res = rd_lookup (rd, obj, &rd -> colour_space, 
					 &rname, NULL, NULL);
			if (res != S_CSC_OK)
				return res;

			obj_destroy (obj);
			*obj = rname;
		}
	}

	csp -> ifc = ifcs [csp -> type.type];

	get_parsing_coords (csp, &fd, &base);

	if (fd != NULL) {
		res = xref__fields_parse (rdb, robj, ref_rown (&csp -> t.refs),
					  obj, fd, base, fname, line);
		if (res != S_CSC_OK)
			return res;
	}
				
	return S_CSC_OK;
}

enum s_status gc_cspace_init (struct gc_colour_space **cspp, 
			      struct xref_db *rdb, 
			      struct xref_bind *robj, 
			      const struct obj *obj, 
			      struct ref_owner *p,
			      const char *fname, int line)
{
	struct gc_colour_space *csp;
	enum s_status res;

	if (obj != NULL && obj -> type == OT_NULL) {
		*cspp = NULL;
		return S_CSC_OK;
	}

	if (obj == NULL)
		obj = &def_cspace;

	if (obj -> type != OT_ARRAY &&
	    obj -> type != OT_NAME &&
	    obj -> type != OT_NULL)
		return obj_error (obj, "Bad colour space type");

	csp = malloc (sizeof (struct gc_colour_space));
	if (csp == NULL)
		return S_CSC_OK;
	memset (csp, 0, sizeof (struct gc_colour_space));

	ref_init (&csp -> t.refs, "colour_space", p);
	csp -> t.rdb = rdb;
	csp -> ifc = NULL;
	xref_ref_db (rdb, ref_rown (&csp -> t.refs));

	res = obj_dup (obj, &csp -> t.obj);
	if (res != S_CSC_OK) {
		gc_cspace_unref (csp, p);
		return res;
	}

	res = init_cspace (csp, rdb, robj, &csp -> t.obj, fname, line);
	if (res != S_CSC_OK) {
		gc_cspace_unref (csp, p);
		return res;
	}

	*cspp = csp;

	return S_CSC_OK;
}

void gc_cspace_unref (struct gc_colour_space *csp, struct ref_owner *p)
{
	struct xref_field *fd;
	void *base;

	if (csp == NULL || ref_down (&csp -> t.refs, p))
		return;

	get_parsing_coords (csp, &fd, &base);
	if (fd != NULL)
		xref_fields_destroy (fd, base, ref_rown (&csp -> t.refs));

	xref_unref_db (csp -> t.rdb, ref_rown (&csp -> t.refs));

	t_tag_destroy (&csp -> t.tag);
	obj_destroy (&csp -> t.obj);

	ref_zero (&csp -> t.refs);

	free (csp);
}

enum s_status gc_cspace_sync (struct xref_db *rdb, struct xref_bind *robj,
			      struct v_selection *sel, struct obj *obj,
			      enum xref_changed *changed, 
			      struct gc_colour_space *csp)
{
	struct xref_field *cfd;
	enum xref_changed c;
	enum s_status res;
	void *base;

	if (csp == NULL) {
		obj_set_null (obj, &f_null);
		return S_CSC_OK;
	}

	if (changed == NULL) {
		c = XREF_C_UNKNOWN;
		changed = &c;
	}

	get_parsing_coords (csp, &cfd, &base);
	if (cfd != NULL) {
		res = xref_fields_ser (rdb, robj, sel, &csp -> t.obj, 
				       changed, cfd, base);
		if (res != S_CSC_OK)
			return res;
	}

	if (obj != NULL) {
		res = obj_dup (&csp -> t.obj, obj);
		if (res != S_CSC_OK)
			return res;
	}

	return S_CSC_OK;
}

static enum s_status set_default_cspace (struct rd_resdict *rd,
					 struct xref_db *rdb,
					 struct gc_colour_state *cst)
{
	struct gc_colour_space_ifc *ifc;
	enum s_status res;
	struct obj obj;
	int pres;

	ifc = cst -> sp -> ifc;
	if (ifc -> defname == NULL)
		return S_CSC_OK;

	res = rd_lookup_s (rd, ifc -> defname, 
			   &rd -> colour_space, &obj, NULL, &pres);
	if (res != S_CSC_OK)
		return res;

	if (!pres)
		return S_CSC_OK;

	res = rd_get_cspace (rd, rdb, &obj, &cst -> defsp, 
			     ref_rown (&cst -> refs));
	obj_destroy (&obj);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}
					 
/* -- cstate -- */

enum s_status gc_set_colour_state (struct obj *name, 
				   struct xref_db *rdb,
				   struct rd_resdict *rd,
				   struct gc_colour_state **cstp,
				   struct ref_owner *p)
{
	struct gc_colour_state *cst;
	enum s_status res;
	
	cst = malloc (sizeof (struct gc_colour_state));
	if (cst == NULL)
		return S_CSC_OK;

	memset (cst, 0, sizeof (struct gc_colour_state));

	ref_init (&cst -> refs, "colour_state", p);

	res = rd_get_cspace (rd, rdb, name, &cst -> sp, 
			     ref_rown (&cst -> refs));
	if (res != S_CSC_OK) {
		gc_cstate_unref (cst, p);
		return res;
	}

	res = set_default_cspace (rd, rdb, cst);
	if (res != S_CSC_OK) {
		gc_cstate_unref (cst, p);
		return res;
	}

	res = cst -> sp -> ifc -> init (cst -> sp, cst, rdb, rd);
	if (res != S_CSC_OK) {
		gc_cstate_unref (cst, p);
		return res;
	}

	gc_cstate_unref (*cstp, p);
	*cstp = cst;

	return S_CSC_OK;
}

int gc_get_components (struct gc_colour_state *cst)
{
	return cst -> sp -> ifc -> get_components (cst -> sp);
}

enum s_status gc_select_colour (struct gc_colour_state **cstp,
				struct pt_pattern *pattern, 
				pdf_real_t *val,
				struct ref_owner *p)
{
	struct gc_colour_state *cst;
	enum s_status res;

	cst = malloc (sizeof (struct gc_colour_state));
	if (cst == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (cst, 0, sizeof (struct gc_colour_state));

	ref_init (&cst -> refs, "colour_state", p);

	cst -> sp = (*cstp) -> sp;
	cst -> defsp = (*cstp) -> defsp;

	res = cst -> sp -> ifc -> dup (cst -> sp, *cstp, cst);
	if (res != S_CSC_OK) {
		free (cst);
		return res;
	}

	T_CALL (gc_cspace_ref, cst -> sp, ref_rown (&cst -> refs));
	if (cst -> defsp != NULL)
		T_CALL (gc_cspace_ref, cst -> defsp, ref_rown (&cst -> refs));
		
	if (gc_is_pattern (cst)) {
		T_CALL (pt_ref, pattern, ref_rown (&cst -> refs));
		cst -> pattern = pattern;
	}

	cst -> sp -> ifc -> select (cst -> sp, cst, val);
	
	gc_cstate_unref (*cstp, p);
	*cstp = cst;

	return S_CSC_OK;
}

enum s_status gc_get_colour (struct gc_colour_state *cst,
			     struct pt_pattern **pattern, 
			     pdf_real_t **val, int *components,
			     struct ref_owner *p)
{
	pdf_real_t *v;

	v = calloc (gc_get_components (cst), sizeof (pdf_real_t));
	if (v == NULL)
		return S_CSC_OUT_OF_MEMORY;

	cst -> sp -> ifc -> retrieve (cst -> sp, cst, v);

	*components = gc_get_components (cst);
	*val = v;
	*pattern = cst -> pattern;

	T_CALL (pt_ref, cst -> pattern, p);

	return S_CSC_OK;
}

int gc_cspace_equal (const struct gc_colour_space *csp1, 
		     const struct gc_colour_space *csp2)
{
	if (csp1 -> ifc != csp2 -> ifc)
		return 0;

	return t_tag_same (&csp1 -> t.tag, &csp2 -> t.tag);
}

int gc_cstate_equal (const struct gc_colour_state *cst1,
		     const struct gc_colour_state *cst2)
{
	if (!gc_cspace_equal (cst1 -> sp, cst2 -> sp))
		return 0;

	if (memcmp (&cst1 -> u, &cst2 -> u, sizeof (cst1 -> u)))
		return 0;

	if (cst1 -> pattern == NULL && cst2 -> pattern == NULL)
		return 1;

	if (cst1 -> pattern != NULL || cst2 -> pattern != NULL)
		return 0;

	return t_tag_same (&cst1 -> pattern -> t.tag,
			   &cst2 -> pattern -> t.tag);
}

enum gc_colour_space_family gc_get_family (struct gc_colour_space *csp)
{
	return csp -> type.type;
}

int gc_is_pattern (struct gc_colour_state *cst)
{
	return cst -> sp -> ifc == &csi_pattern;
}

static enum s_status cstate_init (struct gc_colour_state *cst,
				  struct xref_db *rdb)
{
	enum s_status res;

	res = T_CALL (gc_cspace_new, &cst -> sp, rdb,
		      ref_rown (&cst -> refs));
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

enum s_status gc_cstate__init (struct gc_colour_state **cst,
			       struct xref_db *rdb,
			       struct ref_owner *p,
			       const char *fname, int line)
{
	struct gc_colour_state *ans;
	enum s_status res;

	ans = malloc (sizeof (struct gc_colour_state));
	if (ans == NULL)
		return S_CSC_OUT_OF_MEMORY;

	memset (ans, 0, sizeof (struct gc_colour_state));

	ref__init (&ans -> refs, "colour_state", p, fname, line);

	res = cstate_init (ans, rdb);
	if (res != S_CSC_OK) {
		gc_cstate_unref (ans, p);
		return res;
	}

	*cst = ans;

	return S_CSC_OK;
}

void gc_cstate__ref (struct gc_colour_state *cst, struct ref_owner *p,
		     const char *fname, int line)
{
	if (cst != NULL)
		ref__up (&cst -> refs, p, fname, line);
}

void gc_cstate_unref (struct gc_colour_state *cst, struct ref_owner *p)
{
	if (cst == NULL || ref_down (&cst -> refs, p))
		return;

	gc_cspace_unref (cst -> defsp, ref_rown (&cst -> refs));
	gc_cspace_unref (cst -> sp, ref_rown (&cst -> refs));
	pt_unref (cst -> pattern, ref_rown (&cst -> refs));

	ref_zero (&cst -> refs);

	free (cst);
}

enum s_status gc_init (void)
{
	struct obj_bstr str;
	enum s_status res;

	res = obj_string_from_c ("DeviceGray", &str);
	if (res != S_CSC_OK)
		return res;

	obj_set_name (&def_cspace, &f_null, &str);
	
	return S_CSC_OK;
}

void gc_cleanup (void)
{
	obj_destroy (&def_cspace);
}
