/*
 * PDFlow -- colour_pst.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */
/*
 * This file implements the pstables needed to bridge between the
 * abstract structures defined in colour.h and arrays an dictionaries
 * defined by the Adobe specification. In some cases we need to bridge
 * the top level structure (i.e. the array specification) because its
 * format is too weird (e.g. DeviceN) or too simple (e.g. CIE spaces).
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <string.h>

#include "colour_pst.h"

static enum s_status cie_cf_parser (struct xref_db *rdb, 
				    struct xref_bind *robj, 
				    struct ref_owner *p, 
				    struct obj *src, 
				    struct xref_field *fd,
				    const char *fname, int line);
static enum s_status cie_cf_ser (struct xref_db *rdb, struct xref_bind *robj,
				 struct v_selection *sel, struct obj *dest, 
				 enum xref_changed *changed, 
				 struct xref_field *fd,
				 enum gc_colour_space_family family);
static enum s_status cal_gray_cf_ser (struct xref_db *rdb, 
				      struct xref_bind *robj,
				      struct v_selection *sel, 
				      struct obj *dest, 
				      enum xref_changed *changed,
				      struct xref_field *fd);
static enum s_status cal_rgb_cf_ser (struct xref_db *rdb, 
				     struct xref_bind *robj,
				     struct v_selection *sel, 
				     struct obj *dest, 
				     enum xref_changed *changed,
				     struct xref_field *fd);
static enum s_status lab_cf_ser (struct xref_db *rdb, struct xref_bind *robj,
				 struct v_selection *sel, struct obj *dest, 
				 enum xref_changed *changed, 
				 struct xref_field *fd);

static enum s_status icc_cf_parser (struct xref_db *rdb, 
				    struct xref_bind *robj, 
				    struct ref_owner *p, 
				    struct obj *src, 
				    struct xref_field *fd,
				    const char *fname, int line);
static enum s_status icc_cf_ser (struct xref_db *rdb, struct xref_bind *robj,
				 struct v_selection *sel, struct obj *dest, 
				 enum xref_changed *changed, 
				 struct xref_field *fd);
static void icc_cf_funref (struct ref_owner *p, struct xref_field *fd);

static enum s_status device_n_cf_parser (struct xref_db *rdb, 
					 struct xref_bind *robj, 
					 struct ref_owner *p, 
					 struct obj *src, 
					 struct xref_field *fd,
					 const char *fname, int line);
static enum s_status device_n_cf_ser (struct xref_db *rdb, 
				      struct xref_bind *robj,
				      struct v_selection *sel, 
				      struct obj *dest, 
				      enum xref_changed *changed, 
				      struct xref_field *fd);
static void device_n_cf_funref (struct ref_owner *p, struct xref_field *fd);

static enum s_status lookup_cf_parser (struct xref_db *rdb, 
				       struct xref_bind *robj, 
				       struct ref_owner *p, 
				       struct obj *src, 
				       struct xref_field *fd,
				       const char *fname, int line);
static enum s_status lookup_cf_ser (struct xref_db *rdb, 
				    struct xref_bind *robj,
				    struct v_selection *sel, 
				    struct obj *dest, 
				    enum xref_changed *changed, 
				    struct xref_field *fd);
static void lookup_cf_funref (struct ref_owner *p, struct xref_field *fd);

static struct cobj_enum_desc subtype_desc [] = {
	{ GC_DNT_DEVICE_N, "DeviceN" },
	{ GC_DNT_NCHANNEL, "NChannel"},
	{ -1, NULL }
};

/* Bridge structures */
/* Low level (dictionaries & Co) */
struct pst_cie {
#define PST_CIE
#include "colour_pst.tab.h"
};

struct pst_icc {
#define PST_ICC
#include "colour_pst.tab.h"
};

struct pst_devn_mixing_hints {
#define PST_DEVN_MIXING_HINTS
#include "colour_pst.tab.h"
};

struct pst_devn_process {
#define PST_DEVN_PROCESS
#include "colour_pst.tab.h"
};

struct pst_devn_attr {
#define PST_DEVN_ATTR
#include "colour_pst.tab.h"
};

struct pst_device_n {
#define PST_DEVICE_N
#include "colour_pst.tab.h"
};

#define PST_CIE
#define XREF_IMPLEMENTATION struct pst_cie
#include "colour_pst.tab.h"
#undef XREF_IMPLEMENTATION

#define PST_ICC
#define XREF_IMPLEMENTATION struct pst_icc
#include "colour_pst.tab.h"
#undef XREF_IMPLEMENTATION

#define PST_DEVN_PROCESS
#define XREF_IMPLEMENTATION struct pst_devn_process
#include "colour_pst.tab.h"
#undef XREF_IMPLEMENTATION

#define PST_DEVN_MIXING_HINTS
#define XREF_IMPLEMENTATION struct pst_devn_mixing_hints
#include "colour_pst.tab.h"
#undef XREF_IMPLEMENTATION

#define PST_DEVN_ATTR
#define XREF_IMPLEMENTATION struct pst_devn_attr
#include "colour_pst.tab.h"
#undef XREF_IMPLEMENTATION

#define PST_DEVICE_N
#define XREF_IMPLEMENTATION struct pst_device_n
#include "colour_pst.tab.h"
#undef XREF_IMPLEMENTATION

struct xref_pstable gc_cf_cal_gray_pstable = {
	sizeof (struct gc_d_cie),
	cie_cf_parser, cal_gray_cf_ser, NULL
};

struct xref_pstable gc_cf_cal_rgb_pstable = {
	sizeof (struct gc_d_cie),
	cie_cf_parser, cal_rgb_cf_ser, NULL
};

struct xref_pstable gc_cf_lab_pstable = {
	sizeof (struct gc_d_cie),
	cie_cf_parser, lab_cf_ser, NULL
};

struct xref_pstable gc_cf_icc_based_pstable = {
	sizeof (struct gc_d_icc_based),
	icc_cf_parser, icc_cf_ser, icc_cf_funref
};

struct xref_pstable gc_cf_device_n_pstable = {
	sizeof (struct gc_d_icc_based),
	device_n_cf_parser, device_n_cf_ser, device_n_cf_funref
};

struct xref_pstable gc_lookup_pstable = {
	sizeof (struct obj_bstr),
	lookup_cf_parser, lookup_cf_ser, lookup_cf_funref
};

static enum s_status vec_parse (struct xref_real_array *sa, pdf_real_t da [], 
				int size, int opt, int ext, pdf_real_t defv, 
				struct obj *refobj, const char *name)
{
	int i;

	if (XREF_A_IS_NULL (sa) && opt) {
		for (i = 0; i < size; i++)
			da [i] = defv;
	} else if (XREF_A_SIZE (sa) == size) {
		for (i = 0; i < size; i++)
			da [i] = sa -> data [i];
	} else
		return obj_error (refobj, "Bad %s array size", name);

	return S_CSC_OK;
}

static enum s_status vec_copy (pdf_real_t sa [], struct xref_real_array *da, 
			       int size, int opt, pdf_real_t defv)
{
	int i;

	if (opt) {
		for (i = 0; i < size; i++)
			if (sa [i] != defv)
				break;

		if (i == size) {
			XREF_A_SET_NULL (da);
			return S_CSC_OK;
		}
	}

	da -> data = calloc (size, sizeof (pdf_real_t));
	if (da -> data == NULL)
		return S_CSC_OUT_OF_MEMORY;
	da -> n = size;

	for (i = 0; i < size; i++)
		da -> data [i] = sa [i];

	return S_CSC_OK;
}

static enum s_status cie_init (struct pst_cie *src, struct gc_d_cie *dest,
			       struct obj *refobj)
{
	enum s_status res;

	res = vec_parse (&src -> white_point, dest -> wp, 3, 0, 0, 0.,
			 refobj, "WhitePoint");
	if (res != S_CSC_OK)
		return res;

	res = vec_parse (&src -> black_point, dest -> bp, 3, 1, 0, 0.,
			 refobj, "BlackPoint");
	if (res != S_CSC_OK)
		return res;

	res = vec_parse (&src -> black_point, dest -> bp, 3, 1, 1, 1.,
			 refobj, "Gamma");
	if (res != S_CSC_OK)
		return res;

	dest -> matrix = src -> matrix;

	if (XREF_A_IS_NULL (&src -> range)) {
		dest -> amin = dest -> bmin = -100;
		dest -> amax = dest -> bmax = +100;
	} else if (XREF_A_SIZE (&src -> range) == 4) {
		dest -> amin = src -> range.data [0];
		dest -> amax = src -> range.data [1];
		dest -> bmin = src -> range.data [2];
		dest -> bmax = src -> range.data [3];
	} else
		return obj_error (refobj, "Bad Range array");

	return S_CSC_OK;
}

static enum s_status cie_cf_parser (struct xref_db *rdb, 
				    struct xref_bind *robj, 
				    struct ref_owner *p, 
				    struct obj *src, 
				    struct xref_field *fd,
				    const char *fname, int line)
{
	struct pst_cie cie;
	enum s_status res;

	memset (&cie, 0, sizeof (struct pst_cie));
	res = xref__fields_parse (rdb, robj, p, src, 
				  &pst_cie_field, &cie, fname, line);
	if (res != S_CSC_OK)
		return res;

	res = cie_init (&cie, fd -> field, src);

	xref_fields_destroy (&pst_cie_field, &cie, p);

	return res;
}

static enum s_status cie_copy (struct gc_d_cie *src, struct pst_cie *dest,
			       enum gc_colour_space_family family)
{
	enum s_status res;

	res = vec_copy (src -> wp, &dest -> white_point, 3, 0, 0.);
	if (res != S_CSC_OK)
		return res;

	res = vec_copy (src -> bp, &dest -> black_point, 3, 1, 0.);
	if (res != S_CSC_OK)
		return res;

	if (family == GC_CSF_CAL_GRAY)
		res = vec_copy (src -> gamma, &dest -> gamma, 1, 1, 1.);
	else if (family == GC_CSF_CAL_RGB)
		res = vec_copy (src -> gamma, &dest -> gamma, 3, 1, 1.);
	else
		res = S_CSC_OK;
	if (res != S_CSC_OK)
		return res;

	if (family == GC_CSF_CAL_RGB)
		dest -> matrix = src -> matrix;

	XREF_A_SET_NULL (&dest -> range);
	if (family == GC_CSF_LAB) {
		if (src -> amin != -100 || src -> amax != +100 ||
		    src -> bmin != -100 || src -> bmax != +100) {
			dest -> range.data = calloc (4, sizeof (pdf_real_t));
			if (dest -> range.data == NULL)
				return S_CSC_OUT_OF_MEMORY;
			dest -> range.n = 4;
			dest -> range.data [0] = src -> amin;
			dest -> range.data [1] = src -> amax;
			dest -> range.data [2] = src -> bmin;
			dest -> range.data [3] = src -> bmax;
		}
	}
	
	return S_CSC_OK;
}

static enum s_status cie_cf_ser (struct xref_db *rdb, struct xref_bind *robj,
				 struct v_selection *sel, struct obj *dest, 
				 enum xref_changed *changed,
				 struct xref_field *fd, 
				 enum gc_colour_space_family family)
{
	struct pst_cie cie;
	enum s_status res;

	memset (&cie, 0, sizeof (struct pst_cie));
	res = cie_copy (fd -> field, &cie, family);
	if (res == S_CSC_OK)
		res = xref_fields_ser (rdb, robj, sel, dest, 
				       changed, &pst_cie_field, &cie);

	xref_fields_destroy (&pst_cie_field, &cie, NULL);

	return S_CSC_OK;
}

static enum s_status cal_gray_cf_ser (struct xref_db *rdb, 
				      struct xref_bind *robj,
				      struct v_selection *sel, 
				      struct obj *dest, 
				      enum xref_changed *changed,
				      struct xref_field *fd)
{
	return cie_cf_ser (rdb, robj, sel, dest, changed, fd, GC_CSF_CAL_GRAY);
}

static enum s_status cal_rgb_cf_ser (struct xref_db *rdb, 
				     struct xref_bind *robj,
				     struct v_selection *sel, 
				     struct obj *dest, 
				     enum xref_changed *changed,
				     struct xref_field *fd)
{
	return cie_cf_ser (rdb, robj, sel, dest, changed, fd, GC_CSF_CAL_RGB);
}

static enum s_status lab_cf_ser (struct xref_db *rdb, struct xref_bind *robj,
				 struct v_selection *sel, struct obj *dest, 
				 enum xref_changed *changed, 
				 struct xref_field *fd)
{
	return cie_cf_ser (rdb, robj, sel, dest, changed, fd, GC_CSF_LAB);
}


static enum s_status icc_init (struct pst_icc *src, 
			       struct gc_d_icc_based *dest,
			       struct obj *refobj,
			       struct xref_db *rdb, struct xref_bind *robj,
			       struct ref_owner *p)
{
	enum s_status res;
	int i;

	if (src -> n != 1 && src -> n != 3 && src -> n != 4)
		return obj_error (refobj, "Bad N value for ICC space");

	dest -> n = src -> n;
	if (src -> alternate != NULL) {
		dest -> alternate = src -> alternate;
		T_CALL (gc_cspace_ref, dest -> alternate, p);
	}

	if (XREF_A_IS_NULL (&src -> range)) {
		for (i = 0; i < dest -> n; i++) {
			dest -> min [i] = 0;
			dest -> max [i] = 1;
		}
	} else if (XREF_A_SIZE (&src -> range) == 2 * dest -> n) {
		for (i = 0; i < dest -> n; i++) {
			dest -> min [i] = src -> range.data [2 * i];
			dest -> max [i] = src -> range.data [2 * i + 1];
		}
	} else
		return obj_error (refobj, "Bad Range array");

	res = obj_dup (&src -> metadata, &dest -> metadata);
	if (res != S_CSC_OK)
		return res;

	if (src -> data.encoded.begin != NULL &&
	    src -> data.decoded.begin == NULL)
		return obj_error (refobj, "Can't decode ICC data");

	res = obj_region_dup (&src -> data.decoded, &dest -> data);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

enum s_status icc_copy (struct gc_d_icc_based *src, struct pst_icc *dest,
			struct xref_db *rdb, struct xref_bind *robj,
			struct v_selection *sel, struct ref_owner *p)
{
	enum s_status res;
	int i;

	dest -> n = src -> n;

	if (src -> alternate != NULL) {
		dest -> alternate = src -> alternate;
		T_CALL (gc_cspace_ref, dest -> alternate, p);
	}

	for (i = 0; i < dest -> n; i++)
		if (src -> min [i] != 0 || src -> max [i] != 1)
			break;

	if (i < dest -> n) {
		dest -> range.data = calloc (2 * dest -> n, 
					     sizeof (pdf_real_t));
		if (dest -> range.data == NULL)
			return S_CSC_OUT_OF_MEMORY;
		dest -> range.n = 2 * dest -> n;
		for (i = 0; i < dest -> n; i++) {
			dest -> range.data [2 * i] = src -> min [i];
			dest -> range.data [2 * i + 1] = src -> max [i];
		}
	} else
		dest -> range.n = -1;

	res = obj_dup (&src -> metadata, &dest -> metadata);
	if (res != S_CSC_OK)
		return res;

	memset (&dest -> data, 0, sizeof (struct xref_stream));

	res = obj_region_dup (&src -> data, &dest -> data.decoded);
	if (res != S_CSC_OK)
		return res;

	res = obj_region_dup (&src -> data, &dest -> data.encoded);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status icc_cf_parser (struct xref_db *rdb, 
				    struct xref_bind *robj, 
				    struct ref_owner *p, 
				    struct obj *src, 
				    struct xref_field *fd,
				    const char *fname, int line)
{
	struct pst_icc icc;
	enum s_status res;

	memset (&icc, 0, sizeof (struct pst_icc));
	res = xref__fields_parse (rdb, robj, p, src, 
				  &pst_icc_field, &icc, fname, line);
	if (res != S_CSC_OK)
		return res;

	res = icc_init (&icc, fd -> field, src, rdb, robj, p);

	xref_fields_destroy (&pst_icc_field, &icc, p);
	if (res != S_CSC_OK) {
		icc_cf_funref (p, fd);
		return res;
	}

	return S_CSC_OK;
}

static enum s_status icc_cf_ser (struct xref_db *rdb, struct xref_bind *robj,
				 struct v_selection *sel, struct obj *dest, 
				 enum xref_changed *changed, 
				 struct xref_field *fd)
{
	struct pst_icc icc;
	struct ref_owner p;
	enum s_status res;

	ref_init_owner (&p, "icc_cf_ser");

	memset (&icc, 0, sizeof (struct pst_icc));
	res = icc_copy (fd -> field, &icc, rdb, robj, sel, &p);
	if (res == S_CSC_OK)
		res = xref_fields_ser (rdb, robj, sel, dest,
				       changed, &pst_icc_field, &icc);

	xref_fields_destroy (&pst_icc_field, &icc, &p);

	return S_CSC_OK;
}

static void icc_funref (struct ref_owner *p, struct gc_d_icc_based *icc)
{
	obj_destroy (&icc -> metadata);
	obj_region_destroy (&icc -> data);
	gc_cspace_unref (icc -> alternate, p);
	memset (icc, 0, sizeof (struct gc_d_icc_based));
}

static void icc_cf_funref (struct ref_owner *p, struct xref_field *fd)
{
	icc_funref (p, fd -> field);
}

static enum s_status parse_process (struct pst_devn_process *src, 
				    struct gc_d_device_n *dest,
				    struct obj *refobj,
				    struct ref_owner *p)
{
	struct obj *obj;
	int i, j;

	if (src -> color_space != NULL) {
 		dest -> process_space = src -> color_space;
		T_CALL (gc_cspace_ref, dest -> process_space, p);
	}

	for (i = 0; i < dest -> n; i++) {
		for (j = 0; j < XREF_A_SIZE (&src -> components); j++) {
			obj = &src -> components.data [i];
			if (obj_string_equal (&obj -> val.string, 
					      dest -> col [i].name))
				break;
			j++;
		}

		dest -> col [i].pidx = j;
	}

	return S_CSC_OK;
}

static enum s_status parse_colorants (struct xref_db *rdb,
				      struct xref_bind *robj,
				      struct obj *src, 
				      struct gc_d_device_n *dest,
				      struct obj *refobj,
				      struct ref_owner *p)
{
	enum s_status res;
	struct obj obj;
	int i;

	for (i = 0; i < dest -> n; i++) {
		res = xref_dict_resolve	(rdb, src, dest -> col [i].name, 
					 OT_ANY, &obj);
		if (res != S_CSC_OK)
			return res;

		if (obj.type == OT_NULL)
			continue;

		res = T_CALL (gc_cspace_init,
			      &dest -> col [i].sep, rdb, robj, &obj, p);
		if (res != S_CSC_OK)
			return res;
	}

	return S_CSC_OK;
}

static enum s_status parse_solidities (struct xref_db *rdb,
				       struct xref_bind *robj,
				       struct obj *src, 
				       struct gc_d_device_n *dest,
				       struct obj *refobj,
				       struct ref_owner *p)
{
	pdf_real_t def_sol;
	struct obj obj;
	enum s_status res;
	int i;

	res = xref_dict_resolve (rdb, src, "Default", OT_REAL, &obj);
	if (res != S_CSC_OK) 
		return res;

	def_sol = obj.type != OT_NULL ? obj.val.real : 0;
	for (i = 0; i < dest -> n; i++) {
		res = xref_dict_resolve (rdb, src, dest -> col [i].name, 
					 OT_REAL, &obj);
		if (res != S_CSC_OK)
			break;

		dest -> col [i].solidity = obj.type != OT_NULL ?
			obj.val.real : def_sol;
	}

	return S_CSC_OK;
}

static enum s_status parse_porder (struct xref_db *rdb,
				   struct xref_bind *robj,
				   struct xref_obj_array *src, 
				   struct gc_d_device_n *dest,
				   struct obj *refobj,
				   struct ref_owner *p)
{
	int i, j;

	for (i = 0; i < dest -> n; i++) {
		for (j = 0; j < XREF_A_SIZE (src); j++) {
			if (obj_string_equal (&src -> data [j].val.string, 
					      dest -> col [i].name));
				break;
			j++;
		}

		dest -> col [i].printing_order = j;
	}
	
	return S_CSC_OK;
}

static enum s_status parse_dot_gain (struct xref_db *rdb,
				     struct xref_bind *robj,
				     struct obj *src, 
				     struct gc_d_device_n *dest,
				     struct obj *refobj,
				     struct ref_owner *p)
{
	struct obj obj;
	enum s_status res;
	int i;

	res = xref_dict_resolve	(rdb, src, "Default", OT_ANY, &obj);
	if (res != S_CSC_OK)
		return res;

	if (obj.type != OT_NULL) {
		res = T_CALL (fun_init, &dest -> def_dg, rdb, robj, &obj, p);
		if (res != S_CSC_OK)
			return res;
	}

	for (i = 0; i < dest -> n; i++) {
		res = xref_dict_resolve (rdb, src, dest -> col [i].name,
					 OT_ANY, &obj);
		if (res != S_CSC_OK)
			break;

		if (obj.type == OT_NULL)
			continue;

		res = T_CALL (fun_init, &dest -> col [i].dot_gain, 
			      rdb, robj, &obj, p);
		if (res != S_CSC_OK)
			return res;
	}

	return S_CSC_OK;
}

static enum s_status parse_mixing_hints (struct xref_db *rdb,
					 struct xref_bind *robj,
					 struct pst_devn_mixing_hints *src, 
					 struct gc_d_device_n *dest,
					 struct obj *refobj,
					 struct ref_owner *p)
{
	enum s_status res;

	res = parse_solidities (rdb, robj, &src -> solidities, 
				dest, refobj, p);
	if (res != S_CSC_OK)
		return res;

	res = parse_porder (rdb, robj, &src -> printing_order, 
			    dest, refobj, p);
	if (res != S_CSC_OK)
		return res;

	res = parse_dot_gain (rdb, robj, &src -> dot_gain, 
			      dest, refobj, p); 
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status parse_attributes (struct xref_db *rdb,
				       struct xref_bind *robj,
				       struct pst_devn_attr *src, 
				       struct gc_d_device_n *dest,
				       struct obj *refobj,
				       struct ref_owner *p)
{
	enum s_status res;

	dest -> type = src -> subtype.type;

	res = parse_colorants (rdb, robj, &src -> colorants, dest, refobj, p);
	if (res != S_CSC_OK)
		return res;

	res = parse_process (&src -> process, dest, refobj, p);
	if (res != S_CSC_OK)
		return res;

	res = parse_mixing_hints (rdb, robj, &src -> mixing_hints, 
				  dest, refobj, p);
	if (res != S_CSC_OK)
		return res;

	return res;
}

static enum s_status device_n_init (struct xref_db *rdb,
				    struct xref_bind *robj,
				    struct pst_device_n *src, 
				    struct gc_d_device_n *dest,
				    struct obj *refobj,
				    struct ref_owner *p)
{
	enum s_status res;
	int i;

	if (src -> alternate_space != NULL) {
 		dest -> alternate_space = src -> alternate_space;
		T_CALL (gc_cspace_ref, dest -> alternate_space, p);
	}

	if (src -> tint_transform != NULL) {
		dest -> tint_transform = src -> tint_transform;
		T_CALL (fun_ref, dest -> tint_transform, p);
	}

	if (!XREF_A_SIZE (&src -> names))
		return obj_error (refobj, "Names array is empty");

	dest -> n = XREF_A_SIZE (&src -> names);
	dest -> col = calloc (dest -> n, sizeof (struct gc_d_device_n_col));
	if (dest -> col == NULL)
		return S_CSC_OUT_OF_MEMORY;

	for (i = 0; i < XREF_A_SIZE (&src -> names); i++) {
		res = obj_string_to_c (&src -> names.data [i].val.string, 
				       &dest -> col [i].name);
		if (res != S_CSC_OK)
			return S_CSC_OK;
	}
	
	res = parse_attributes (rdb, robj, &src -> attributes, 
				dest, refobj, p);
	if (res != S_CSC_OK)
		return res;

	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status devn_copy (struct gc_d_device_n *src, 
				struct pst_device_n *dest)
{
	return S_CSC_NOT_SUPPORTED;
}

static enum s_status device_n_cf_parser (struct xref_db *rdb, 
					 struct xref_bind *robj, 
					 struct ref_owner *p, 
					 struct obj *src, 
					 struct xref_field *fd,
					 const char *fname, int line)
{
	struct pst_device_n bs;
	enum s_status res;

	memset (&bs, 0, sizeof (struct pst_device_n));
	res = xref__fields_parse (rdb, robj, p, src, 
				  &pst_device_n_field, &bs, fname, line);
	if (res != S_CSC_OK)
		return res;

	res = device_n_init (rdb, robj, &bs, fd -> field, src, p);
	xref_fields_destroy (&pst_device_n_field, &bs, p);
	if (res != S_CSC_OK) {
		device_n_cf_funref (p, fd);
		return res;
	}

	return res;
}

static enum s_status device_n_cf_ser (struct xref_db *rdb, 
				      struct xref_bind *robj,
				      struct v_selection *sel, 
				      struct obj *dest,
				      enum xref_changed *changed,
				      struct xref_field *fd)
{
	struct pst_device_n bs;
	enum s_status res;

	memset (&bs, 0, sizeof (struct pst_device_n));
	res = devn_copy (fd -> field, &bs);
	if (res == S_CSC_OK)
		res = xref_fields_ser 
			(rdb, robj, sel, dest, 
			 changed, &pst_device_n_field, &bs);

	xref_fields_destroy (&pst_device_n_field, &bs, NULL);

	return S_CSC_OK;
}

static void device_n_cf_funref (struct ref_owner *p, 
				struct xref_field *fd)
{
	struct gc_d_device_n *dn;
	int i;

	dn = fd -> field;

	if (dn -> col != NULL) {
		for (i = 0; i < dn -> n; i++) {
			free (dn -> col [i].name);
			if (dn -> col [i].sep != NULL)
				gc_cspace_unref (dn -> col [i].sep, p);
			if (dn -> col [i].dot_gain != NULL)
				fun_unref (dn -> col [i].dot_gain, p);
		}

		free (dn -> col);
	}

	if (dn -> tint_transform != NULL)
		fun_unref (dn -> tint_transform, p);

	if (dn -> def_dg != NULL)
		fun_unref (dn -> tint_transform, p);

	if (dn -> process_space != NULL)
		gc_cspace_unref (dn -> process_space, p);

	if (dn -> alternate_space != NULL)
		gc_cspace_unref (dn -> alternate_space, p);

	memset (dn, 0, sizeof (struct gc_d_device_n));
}

static enum s_status lookup_cf_parser (struct xref_db *rdb, 
				       struct xref_bind *robj, 
				       struct ref_owner *p, 
				       struct obj *src, 
				       struct xref_field *fd,
				       const char *fname, int line)
{
	struct obj_bstr *lookup;
	enum s_status res;
	size_t size;
	void *data;

	lookup = fd -> field;
	if (src == NULL || src -> type == OT_NULL)
		return S_CSC_OK;

	switch (src -> type) {
	case OT_STRING:
		res = obj_string_dup (&src -> val.string, lookup);
		if (res != S_CSC_OK)
			return res;

		break;

	case OT_STREAM:
		res = xref_resolve_stream (rdb, src);
		if (res != S_CSC_OK)
			return res;

		size = 	src -> val.stream.decoded.end -
			src -> val.stream.decoded.begin;
		data = malloc (size);
		if (data == NULL)
			return S_CSC_OK;

		memcpy (data, src -> val.stream.decoded.begin, size);
		lookup -> size = size;
		lookup -> s = data;
		break;

	default:
		return obj_error (src, 
				  "Lookup table not a string nor a stream");
	}
	    
	return res;
}

static enum s_status lookup_cf_ser (struct xref_db *rdb, 
				    struct xref_bind *robj,
				    struct v_selection *sel, 
				    struct obj *dest, 
				    enum xref_changed *changed, 
				    struct xref_field *fd)
{
	struct obj_bstr *lookup;
	struct obj_bstr str;
	enum s_status res;

	lookup = fd -> field;
	res = obj_string_dup (lookup, &str);
	if (res != S_CSC_OK)
		return res;

	obj_set_string (dest, &f_null, &str);

	return S_CSC_OK;
}

static void lookup_cf_funref (struct ref_owner *p, struct xref_field *fd)
{
	obj_string_destroy (fd -> field);
}
