/*
 * PDFlow -- filespec.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if defined (_WIN32) || defined (WIN32)
#define CURL_STATICLIB
#endif

#ifdef HAVE_LIBCURL
#include <curl/curl.h>
#endif

#ifdef HAVE_LIBSSL
#include <openssl/md5.h>
#endif

#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>

#include "types/filespec.h"

#ifdef HAVE_LIBCURL
#define CHUNK 65536
#endif

static enum s_status dyn_ifc (enum xref_op op, struct xref_field *fd, 
			      void *data, int *stop);

static enum s_status ini (struct fspec_specification *fs, 
			  struct xref_db *rdb, struct xref_bind *robj,
			  struct ref_owner *p);
static void fini (struct fspec_specification *fs, struct ref_owner *p);

static enum s_status fspec_string_parser (struct xref_db *rdb, 
					  struct xref_bind *robj, 
					  struct ref_owner *p, 
					  struct obj *src, 
					  struct xref_field *fd,
					  const char *fname, int line);
static enum s_status fspec_string_ser (struct xref_db *rdb, 
				       struct xref_bind *robj,
				       struct v_selection *sel, 
				       struct obj *dest, 
				       enum xref_changed *changed,
				       struct xref_field *fd);
static void fspec_string_funref (struct ref_owner *p, 
				 struct xref_field *fd);

struct xref_pstable fspec_string_pstable = {
	sizeof (struct fspec_specification),
	fspec_string_parser, fspec_string_ser, fspec_string_funref
};

static struct cobj_enum_desc fs_desc [] = {
	{ FS_FT_URI, "Uri" },

	{ FS_FT_DEFAULT, "/Default" },

	{ -1, NULL }
};

#define XREF_IMPLEMENTATION struct fspec_mac
#define FSPEC_MAC
#include "types/filespec.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct fspec_params
#define FSPEC_PARAMS
#include "types/filespec.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct fspec_embedded
#define FSPEC_EMBEDDED
#include "types/filespec.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct fspec_ef
#define FSPEC_EF
#include "types/filespec.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct fspec_rf
#define FSPEC_RF
#include "types/filespec.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct fspec_specification
#define FSPEC_DICTIONARY
#include "types/filespec.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION
#define FSPEC_STRING
#include "types/filespec.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION
#define FSPEC_ROOT
#include "types/filespec.tab.h"
#undef XREF_IMPLEMENTATION

T_IMPLEMENTATIONS_LOC (fspec_fs, fspec_specification, &fspec_root,
		       ini, NULL, fini)

#ifdef HAVE_LIBCURL
static enum s_status make_error (CURLcode errornum);
static const char *get_string (enum s_code code);
#endif

#ifdef HAVE_LIBCURL
static struct s_system_ifc ifc ={
	"libcurl",
	S_SYS_CURL,
	get_string
};

static CURL *chandle;

#endif

static enum s_status fspec_error (struct obj_bstr *fname,
				  const char *fmt,...)
{
	enum s_status res;
	char buf [1024];
	va_list ap;

	va_start (ap, fmt);
	vsnprintf (buf, sizeof (buf), fmt, ap);
	buf [sizeof (buf) - 1] = 0;
	res = s_make_indirect (S_CSC_PARSE_ERROR,
			       "[%.*s]: %s", 
			       fname -> size, fname -> s, buf);
	va_end (ap);
	
	return res;
}

static enum s_status dyn_ifc (enum xref_op op, struct xref_field *fd, 
			      void *data, int *stop)
{
	struct fspec_specification *fs;

	fs = data;

	switch (op) {
	case XREF_OP_PARSE:
		switch (fs -> t.obj.type) {
		case OT_STRING:
			*fd = fspec_string_field;

		case OT_NULL:
		case OT_DICTIONARY:
			*fd = fspec_dictionary_field;
			break;

		default:
			return obj_error (&fs -> t.obj,
					  "Bad filespec type: %s",
					  obj_tag (fs -> t.obj.type));
		}
		break;
		
	case XREF_OP_SERIALIZE:
	case XREF_OP_DESTROY:
		*fd = fspec_dictionary_field;
	}

	return S_CSC_OK;
}

static enum s_status ini (struct fspec_specification *fs, 
			  struct xref_db *rdb, struct xref_bind *robj,
			  struct ref_owner *p)
{
	enum s_status res;

	res = ts_strip_lang (&fs -> uf, &fs -> ufo);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static void fini (struct fspec_specification *fs, struct ref_owner *p)
{
	if (fs -> file != NULL)
		f_unref (fs -> file);
	if (fs -> filename != NULL)
		free (fs -> filename);
	obj_string_destroy (&fs -> descname);
	obj_destroy (&fs -> ufo);
}

static enum s_status fspec_string_parser (struct xref_db *rdb, 
					  struct xref_bind *robj, 
					  struct ref_owner *p, 
					  struct obj *src, 
					  struct xref_field *fd,
					  const char *fname, int line)
{
	struct fspec_specification *fs;

	fs = fd -> field;

	return obj_dup (src, &fs -> f);
}

static enum s_status fspec_string_ser (struct xref_db *rdb, 
				       struct xref_bind *robj,
				       struct v_selection *sel, 
				       struct obj *dest, 
				       enum xref_changed *changed,
				       struct xref_field *fd)
{
	assert (0);
}

static void fspec_string_funref (struct ref_owner *p, 
				 struct xref_field *fd)
{
	assert (0);
}

static enum s_status demangle_default (struct obj_bstr *string, 
				       char sep, int winquirk, 
				       char **out)
{
	const char *p, *sentinel;
	char *buffer, *q;

	buffer = malloc (string -> size + 2);
	if (buffer == NULL)
		return S_CSC_OUT_OF_MEMORY;

	p = string -> s;
	q = buffer;
	sentinel = p + string -> size;
	if (winquirk && p < sentinel && *p == '/') {
		/* Absolute path */
		p++;
		if (p < sentinel && *p != '/') {
			/* Share or drive */
			if (p + 1 < sentinel && p [1] == '/') {
				/* Drive */
 				*(q++) = *p;
				*(q++) = ':';
				p += 2;
			} else {
				/* Share */
 				*(q++) = '\\';
 				*(q++) = '\\';
			}
		}
	}

	while (p < sentinel) {
		if (*p == '\\' && p + 1 < sentinel)
			*q = *(++p);
		else if (*p == '/')
			*q = sep;
		else
			*q = *p;
		q++, p++;
	}

	*q = 0;
	*out = buffer;

	return S_CSC_OK;
}

static enum s_status fix_name (struct fspec_specification *fs, 
			       enum fspec_type *type,
			       enum fspec_filesys *fsys)
{
	struct obj *fsstring, dobj;
	int sep, winquirk;
	enum s_status res;
	char *fname;

#if defined (_WIN32) ||	defined (WIN32)
	*fsys = FS_FT_DOS;
	fsstring = &fs -> dos;
	winquirk = 1;
	sep = '\\';
#elif defined (__APPLE__)
	*fsys = FS_FT_MAC;
	fsstring = &fs -> mac; 
	winquirk = 0;
	sep = ':';
#else
	*fsys = FS_FT_UNIX;
	fsstring = &fs -> un_x; 
	winquirk = 0;
	sep = '/';
#endif
	if (fsstring -> type == OT_NULL) {
		if (fs -> ufo.type != OT_NULL) {
			*fsys = FS_FT_UF;
			fsstring = &fs -> ufo;
		} else {
			*fsys = FS_FT_F;
			fsstring = &fs -> f;
		}
		*type = fs -> fs.type;
	} else
		*type = FS_FT_DEFAULT;

	if (*type == FS_FT_DEFAULT) {
		res = demangle_default (&fsstring -> val.string, 
					sep, winquirk, &fname);
		if (res != S_CSC_OK)
			return res;
	} else {
		res = obj_string_to_c (&fsstring -> val.string, &fname);
		if (res != S_CSC_OK)
			return res;
	}

	if (!ts_is_null (&fs -> desc)) {
		res = ts_strip_lang (&fs -> desc, &dobj);
		if (res != S_CSC_OK) {
			free (fname);
			return res;
		}
	}

	if (dobj.type == OT_NULL) {
		res = obj_string_from_c (fname, &fs -> descname);
		if (res != S_CSC_OK) {
			free (fname);
			return res;
		}
	} else
		fs -> descname = dobj.val.string;

	fs -> filename = fname;

	return S_CSC_OK;
}

static enum s_status verify (struct obj_region *reg, 
			     struct obj *csum, int *v)
{
#ifdef HAVE_LIBSSL
	unsigned char buf [MD5_DIGEST_LENGTH];
	MD5_CTX ctxt;

	*v = 0;
	if (csum -> val.string.size != MD5_DIGEST_LENGTH)
		return obj_error (csum, "Checksum not %d bytes long",
				  MD5_DIGEST_LENGTH);

	if (!MD5_Init (&ctxt))
		return S_CSC_OUT_OF_MEMORY;

	if (!MD5_Update (&ctxt, reg -> begin, reg -> end - reg -> begin)) {
		MD5_Final (buf, &ctxt);
		return S_CSC_OUT_OF_MEMORY;
	}
	
	if (!MD5_Final (buf, &ctxt))
		return S_CSC_OUT_OF_MEMORY;

	*v = !memcmp (buf, csum -> val.string.s, MD5_DIGEST_LENGTH);

	return S_CSC_OK;
#else
	*v = 1;

	return S_CSC_OK;
#endif
}

static enum s_status get_embedded (struct fspec_specification *fs,
				   struct obj_bstr *fname,
				   struct f_file **f,
				   struct fspec_info *info,
				   enum fspec_filesys fsys)
{
	struct fspec_embedded *efd;
	enum s_status res;
	int valid;

	*f = NULL;
	switch (fsys) {
	case FS_FT_F:
		efd = &fs -> ef.f;
		break;

	case FS_FT_UF:
		efd = &fs -> ef.uf;
		break;

	case FS_FT_DOS:
		efd = &fs -> ef.dos;
		break;

	case FS_FT_MAC:
		efd = &fs -> ef.mac;
		break;

	case FS_FT_UNIX:
		efd = &fs -> ef.un_x;
		break;

	default:
		assert (0);
	}

	if (efd -> contents.decoded.begin == NULL)
		return fspec_error (fname, "Can't decode file spec");

	if (efd -> params.checksum.type != OT_NULL) {
		res = verify (&efd -> contents.decoded, 
			      &efd -> params.checksum, &valid);
		if (res == S_CSC_OK && !valid)
			res = fspec_error (fname, "Checksum error");
		if (res != S_CSC_OK)
			return res;
	}


	if (efd -> subtype.type != OT_NULL) {
		res = obj_string_to_c (&efd -> subtype.val.string, 
				       &info -> subtype);
		if (res != S_CSC_OK)
			return res;
	}

	info -> mac = efd -> params.mac;
	res = obj_dup (&efd -> params.mac.res_fork, &info -> mac.res_fork);
	if (res != S_CSC_OK) {
		free (info -> subtype);
		return res;
	}

	return S_CSC_OK;
}

static enum s_status get_related (struct fspec_specification *fs, int *relno, 
				  struct fspec_related_info **rinfo,
				  enum fspec_filesys fsys)
{
	struct xref_obj_array *rfd;
	struct obj rfobj;
	enum s_status res;
	int i;

	rfd = NULL;
	switch (fsys) {
	case FS_FT_F:
		rfd = &fs -> rf.f;
		break;

	case FS_FT_UF:
		rfd = &fs -> rf.uf;
		break;

	case FS_FT_DOS:
		rfd = &fs -> rf.dos;
		break;

	case FS_FT_MAC:
		rfd = &fs -> rf.mac;
		break;

	case FS_FT_UNIX:
		rfd = &fs -> rf.un_x;
	}
	
	*relno = 0;
	if (!XREF_A_SIZE (rfd))
		return S_CSC_OK;

	if (XREF_A_SIZE (rfd) & 0)
		return obj_error (&fs -> t.obj, 
				  "Odd number of elements in ER array");

	*relno = XREF_A_SIZE (rfd) / 2;
	*rinfo = calloc (*relno, sizeof (struct fspec_related_info));
	if (*rinfo == NULL)
		return S_CSC_OUT_OF_MEMORY;

	res = S_CSC_OK;
	for (i = 0; i < *relno; i++) {
		res = obj_string_to_c (&rfd -> data [2 * i].val.string,
				       &(*rinfo) [i].fname);
		if (res != S_CSC_OK)
			break;

		res = xref_resolve (fs -> t.rdb, 
				    &rfd -> data [2 * i + 1], OT_STREAM,
				    &rfobj, NULL, "related file");
		if (res != S_CSC_OK)
			break;

		res = get_embedded (fs, &rfd -> data [2 * i].val.string,
				    &(*rinfo) [i].file,
				    &(*rinfo) [i].info, fsys);
		obj_destroy (&rfobj);
		if (res != S_CSC_OK)
			break;
	}

	if (res != S_CSC_OK) {	
		fspec_related_info_destroy (*rinfo, *relno);
		return res;
	}

	return S_CSC_OK;
}

#ifdef HAVE_LIBCURL
struct write_f_data {
	void *buf;
	size_t rd, avail;
};

static size_t write_f (void *ptr, size_t size, size_t nmemb, void *stream)
{
	struct write_f_data *data;
	size_t toread, nblock;
	void *p;

	data = stream;
	toread = size * nmemb;
	if (!toread)
		return 0;

	if (data -> buf == NULL) {
		data -> avail = CHUNK * (1 + toread / CHUNK);
		data -> buf = malloc (data -> avail);
		data -> rd = 0;
		if (data -> buf == NULL)
			return 0;
	} else {
		nblock = CHUNK * (1 + (data -> rd + toread) / CHUNK);
		p = realloc (data -> buf, nblock);
		if (p == NULL) {
			free (data -> buf);
			return 0;
		}
		data -> buf = p;
		data -> avail = nblock;
	}

	memcpy (data -> buf + data -> rd, ptr, toread);
	data -> rd += toread;

	return toread;
}

static CURLcode setup_once (CURL *curl)
{
	CURLcode code;

	code = curl_easy_setopt (curl, CURLOPT_NOPROGRESS, 1L);
	if (code)
		return code;

	code = curl_easy_setopt (curl, CURLOPT_FILETIME, 1L);
	if (code)
		return code;

	code = curl_easy_setopt (curl, CURLOPT_WRITEFUNCTION, write_f);
	if (code)
		return code;

	code = curl_easy_setopt (curl, CURLOPT_HTTPAUTH, CURLAUTH_ANY);
	if (code)
		return code;

	code = curl_easy_setopt (curl, CURLOPT_PROXYAUTH, CURLAUTH_ANY);
	if (code)
		return code;

	return curl_easy_setopt (curl, CURLOPT_NETRC, 1L);

} 

static CURLcode setup_url (CURL *curl, const char *filename,
			   struct write_f_data *data)
{
	CURLcode code;

	code = curl_easy_setopt (curl, CURLOPT_WRITEDATA, data);
	if (code)
		return code;

	return curl_easy_setopt (curl, CURLOPT_URL, filename);
} 
#endif

static enum s_status download (const char *filename,
			       void **tdata, size_t *size,
			       struct cobj_date *date)
{
#ifdef HAVE_LIBCURL
	struct write_f_data data;
	enum s_status res;
	long tdate;

	data.buf = NULL;
	data.rd = 0;
				 
	res = make_error (setup_url (chandle, filename, &data));
	if (res != S_CSC_OK)
		return res;

	res = make_error (curl_easy_perform (chandle));
	if (res != S_CSC_OK)
		return res;

	res = curl_easy_getinfo (chandle, CURLINFO_FILETIME, &tdate);
	if (res != S_CSC_OK || tdate < 0)
		tdate = 0;

	date -> time = tdate;
	date -> tz_hour = 0;
	date -> tz_minutes = 0;

	*tdata = data.buf;
	*size = data.rd;

	return S_CSC_OK;
		
#else
	return s_make_indirect (S_CSC_NOT_SUPPORTED,
				"libCurl has not been compiled in");
#endif
}

static enum s_status get_external (struct fspec_specification *fs,
				   enum fspec_type type,
				   struct f_file **f, 
				   struct fspec_info *info)
{
	enum s_status res;
	struct stat st;
	void *tdata;
	size_t size;

	switch (type) {
	case FS_FT_DEFAULT:
		res = f_open (fs -> filename, f);
		if (res != S_CSC_OK)
			return res;

		if (stat (fs -> filename, &st)) {
			f_unref (*f);
			return S_ERRNO;
		}

		info -> size = st.st_size;
		info -> mod_date.time = st.st_mtime;

		break;

	case FS_FT_URI:
		size = 0;
		tdata = NULL;
		res = download (fs -> filename, &tdata, &size, 
				&info -> mod_date);
		if (res != S_CSC_OK)
			return res;

		res = f_reg_open (fs -> t.obj.ptr.file, tdata, 
				  tdata + size, tdata, free, f);
		if (res != S_CSC_OK) {
			free (tdata);
			return res;
		}
		break;
	}

	return S_CSC_OK;
}

enum s_status fspec_get (struct fspec_specification *fs, 
			 struct f_file **f, struct fspec_info *finfo,
			 struct fspec_related_info **rinfo, int *relno)
{
	enum s_status res;
	struct fspec_info sinfo, *info;

	if (fs -> file != NULL && finfo == NULL && 
	    relno == NULL && rinfo == NULL) {
		f_ref (fs -> file);
		*f = fs -> file;
		return S_CSC_OK;
	}

	info = finfo != NULL ? finfo : &sinfo;
	memset (info, 0, sizeof (struct fspec_info));

	if (fs -> filename == NULL) {
		res = fix_name (fs, &fs -> type, &fs -> fsys);
		if (res != S_CSC_OK)
			return res;
	}

	res = get_embedded (fs, &fs -> descname, f, info, fs -> fsys);
	if (res != S_CSC_OK)
		return res;

	if (*f == NULL) {
		res = get_external (fs, fs -> type, f, info);
		if (res != S_CSC_OK)
			return res;

		if (relno != NULL)
			*relno = 0;
		if (rinfo != NULL)
			*rinfo = NULL;
	} else if (rinfo != NULL) {
		res = get_related (fs, relno, rinfo, fs -> fsys);
		if (res != S_CSC_OK) {
			fspec_info_destroy (info);
			f_unref (*f);
			return res;
		}
	}

	if (!fs -> v) {
		f_ref (*f);
		fs -> file = *f;
	}

	if (finfo == NULL)
		fspec_info_destroy (&sinfo);

	return S_CSC_OK;
}

void fspec_info_destroy (struct fspec_info *info)
{
	if (info -> subtype != NULL)
		free (info -> subtype);

	obj_destroy (&info -> mac.res_fork);
}

void fspec_related_info_destroy (struct fspec_related_info *info, int relno)
{
	int i;

	for (i = 0; i < relno; i++) {
		if (info [i].fname != NULL)
			free (info [i].fname);
		if (info [i].file != NULL)
			f_unref (info [i].file);
		fspec_info_destroy (&info [i].info);
	}

	free (info);
}

#ifdef HAVE_LIBCURL
static const char *get_string (enum s_code code)
{
	return curl_easy_strerror (code);
}

static enum s_status make_error (CURLcode errornum)
{
	if (!errornum)
		return S_CSC_OK;

	return S_STATUS (S_SYS_CURL, errornum);
}
#endif

enum s_status fspec_init (void)
{
#ifdef HAVE_LIBCURL
	enum s_status res;

	res = s_register (&ifc);
	if (res != S_CSC_OK)
		return res;

	res = make_error (curl_global_init (CURL_GLOBAL_ALL));
	if (res != S_CSC_OK) {
		s_unregister (&ifc);
		return res;
	}

	chandle = curl_easy_init ();
	if (chandle == NULL) {
		curl_global_cleanup ();
		s_unregister (&ifc);
		return S_CSC_OUT_OF_MEMORY;;
	}

	res = make_error (setup_once (chandle));
	if (res != S_CSC_OK) {
		curl_easy_cleanup (chandle);
		curl_global_cleanup ();
		s_unregister (&ifc);
		return res;
	}
#endif

	return S_CSC_OK;
}

void fspec_cleanup (void)
{
#ifdef HAVE_LIBCURL
	curl_easy_cleanup (chandle);
	curl_global_cleanup ();
	s_unregister (&ifc);
#endif
}
