/*
 * PDFlow -- font.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_FREETYPE2
#include <ft2build.h>
#include FT_FREETYPE_H
#include <freetype/ftcid.h>
#endif

#include <assert.h>
#include <string.h>

#include "fmt/ctab.h"
#include "fmt/pdf-lex.h"
#include "types/font.h"

/* Circular reference; unavoidable */
#include "pdf/pdfi.h"
#include "ps/psi.h"

#include "std14/times-roman.h"
#include "std14/helvetica.h"
#include "std14/courier.h"
#include "std14/symbol.h"
#include "std14/times-bold.h"
#include "std14/helvetica-bold.h"
#include "std14/courier-bold.h"
#include "std14/zapfdingbats.h"
#include "std14/times-italic.h"
#include "std14/helvetica-oblique.h"
#include "std14/courier-oblique.h"
#include "std14/times-bolditalic.h"
#include "std14/helvetica-boldoblique.h"
#include "std14/courier-boldoblique.h"

#define LEADING_FACTOR 1.2

#ifdef HAVE_FREETYPE2
static const char *ft2_get_string (enum s_code code);

struct s_system_ifc sys = { 
	"freetype_2", S_SYS_FREETYPE, ft2_get_string
};
#endif

static enum s_status f_ini (struct gf_font *font, struct xref_db *rdb,
			    struct xref_bind *robj, struct ref_owner *p);
static void f_fini (struct gf_font *font, struct ref_owner *p);

static enum s_status dyn_ifc (enum xref_op op, struct xref_field *fd, 
			      void *data, int *stop);

enum gf_flags {
	GF_F_FIXED_PITCH = 	1 <<  0,
	GF_F_SERIF = 		1 <<  1,
	GF_F_SYMBOLIC = 	1 <<  2,
	GF_F_SCRIPT = 		1 <<  3,
	GF_F_NONSYBOLIC = 	1 <<  5,
	GF_F_ITALIC = 		1 <<  6,
	GF_F_ALL_CAP = 		1 << 16,
	GF_F_SMALL_CAP = 	1 << 17,
	GF_F_FORCE_BOLD =	1 << 18
};

static struct cobj_enum_desc subtype_desc [] = {
	{ GF_FT_TYPE0, "Type0" },
	{ GF_FT_TYPE1, "Type1" },
	{ GF_FT_MMTYPE1, "MMType1" },
	{ GF_FT_TYPE3, "Type3" },
	{ GF_FT_TRUE_TYPE, "TrueType" },
	{ GF_FT_CID_FONT_TYPE0, "CIDFontType0" },
	{ GF_FT_CID_FONT_TYPE2, "CIDFontType2" },
	{ 0, NULL }
};

static struct cobj_enum_desc cid_to_gid_map_desc [] = {
	{ GF_C2G_IDENTITY, "Identity" },
	{ GF_C2G_STREAM, "/PSTable", 1, 0, &xref_obj_pstable },

	{ GF_C2G_IDENTITY, "/Default" },
	{ 0, NULL }
};

#define GF_CT_ET_DICTIONARY -1
#define GF_CT_ET_NONE -2

static struct cobj_enum_desc encoding_desc [] = {

	{ CT_ET_MAC_ROMAN, "MacRomanEncoding" },
	{ CT_ET_WIN_ANSI, "WinAnsiEncoding" },
	{ CT_ET_MAC_EXPERT, "MacExpertEncoding" },
	
	{ GF_CT_ET_DICTIONARY, "/PSTable", 1, 0, &gf_encoding_pstable },

	{ GF_CT_ET_NONE, "/Default" },
	
	{ 0, NULL }

};

static struct gf_builtin {
	const char *name;
	struct gf_glyph_info *info;
} builtin [] = {
	{ "Times-Roman", gi_times_roman },
	{ "Helvetica", gi_helvetica },
	{ "Courier", gi_courier },
	{ "Symbol", gi_symbol },
	{ "Times-Bold", gi_times_bold },
	{ "Helvetica-Bold", gi_helvetica_bold },
	{ "Courier-Bold", gi_courier_bold },
	{ "ZapfDingbats", gi_zapfdingbats },
	{ "Times-Italic", gi_times_italic },
	{ "Helvetica-Oblique", gi_helvetica_oblique },
	{ "Courier-Oblique", gi_courier_oblique },
	{ "Times-BoldItalic", gi_times_bolditalic },
	{ "Helvetica-BoldOblique", gi_helvetica_boldoblique },
	{ "Courier-BoldOblique", gi_courier_boldoblique },
	{ NULL }
};

static enum s_status type0_init (struct gf_font *font, 
				 struct xref_bind *robj);
static void type0_destroy (struct gf_font *font);
static struct gf_glyph_info *type0_get_glyph (struct gf_font *font,
					      uint8_t *string, size_t limit);

static enum s_status simple_init (struct gf_font *font, 
				  struct xref_bind *robj);
static void simple_destroy (struct gf_font *font);
static struct gf_glyph_info *simple_get_glyph (struct gf_font *font,
					       uint8_t *string, size_t limit);

static enum s_status type3_init (struct gf_font *font, 
				 struct xref_bind *robj);
static void type3_destroy (struct gf_font *font);
static struct gf_glyph_info *type3_get_glyph (struct gf_font *font,
					      uint8_t *string, size_t limit);

static struct gf_glyph_info null_glyph = { { }, 0, 0, 1, 0 };

#define XREF_IMPLEMENTATION struct gf_font
#define GF_FONT
#include "types/font.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct gf_encoding
#define GF_ENCODING
#include "types/font.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct gf_simple_font
#define GF_SIMPLE_FONT
#include "types/font.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct gf_type0_font
#define GF_TYPE0_FONT
#include "types/font.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct gf_type3_font
#define GF_TYPE3_FONT
#include "types/font.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct gf_cid_font
#define GF_CID_FONT
#include "types/font.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct gf_font_desc
#define GF_FONT_DESC
#include "types/font.tab.h"
#undef XREF_IMPLEMENTATION

T_IMPLEMENTATIONS_LOC (gf, gf_font, &gf_font_field, f_ini, NULL, f_fini);
T_IMPLEMENTATIONS (gf_encoding, gf_encoding, &gf_encoding_field);
T_IMPLEMENTATIONS (gf_desc, gf_font_desc, &gf_font_desc_field);


struct gf_font_ifc {
	struct xref_field *fd;
	enum s_status (*init) (struct gf_font *font, 
			       struct xref_bind *robj);
	void (*destroy) (struct gf_font *font);
	struct gf_glyph_info *(*get_glyph) (struct gf_font *font,
					    uint8_t *string, size_t limit);
};

static struct gf_font_ifc type0_ifc = {
	&gf_type0_font_field,
	type0_init, type0_destroy, type0_get_glyph
};

static struct gf_font_ifc simple_ifc = {
	&gf_simple_font_field,
	simple_init, simple_destroy, simple_get_glyph
};

static struct gf_font_ifc type3_ifc = {
	&gf_type3_font_field,
	type3_init, type3_destroy, type3_get_glyph
};

static struct gf_font_ifc cid_ifc = {
	&gf_cid_font_field
};


struct gf_binding {
	enum gf_font_type type;
	struct gf_font_ifc *ifc;
};

static struct gf_binding ifcs [] = { 
	{ GF_FT_TYPE0, &type0_ifc },
	{ GF_FT_TYPE1, &simple_ifc },
	{ GF_FT_MMTYPE1, &simple_ifc },
	{ GF_FT_TRUE_TYPE, &simple_ifc },
	{ GF_FT_TYPE3, &type3_ifc },
	{ GF_FT_CID_FONT_TYPE0, &cid_ifc },
	{ GF_FT_CID_FONT_TYPE2, &cid_ifc },
	{ 0, NULL }
};

static enum s_status dyn_ifc (enum xref_op op, struct xref_field *fd, 
			      void *data, int *stop)
{
	struct gf_font *f;
	int i;

	f = data;

	if (f -> ifc == NULL) {
		for (i = 0; ifcs [i].ifc != NULL; i++)
			if (f -> subtype.type == ifcs [i].type)
				break;

		f -> ifc = ifcs [i].ifc;
	}

	if (f -> ifc == NULL) {
		switch (op) {
		case XREF_OP_PARSE:
			return obj_error (&f -> t.obj, 
					  "Unsupported font type: %d",
					  f -> subtype.type);
			break;

		case XREF_OP_SERIALIZE:
			return s_make_indirect 
				(S_CSC_INTERNAL_ERROR,
				 "Trying to serialize font type %d",
				 f -> subtype.type);

		case XREF_OP_DESTROY:
			/* Can only happen if no initialization
			 * has been done, otherwise the interface 
			 * should have already been set
			 */
			*stop = 1;
			return S_CSC_OK;
		}
	}

	*fd = *f -> ifc -> fd;

	return S_CSC_OK;
}

static enum s_status f_ini (struct gf_font *font, struct xref_db *rdb,
			    struct xref_bind *robj, struct ref_owner *p)
{
	enum s_status res;

	if (font -> ifc -> init != NULL && font -> ifc -> init != NULL) {
		font -> wmode = CMAP_WM_HORIZONTAL;
		res = font -> ifc -> init (font, robj);
		if (res != S_CSC_OK)
			return res;
	}
		
	return S_CSC_OK;
}

static void f_fini (struct gf_font *font, struct ref_owner *p)
{
	if (font -> ifc != NULL && font -> ifc -> destroy != NULL)
		font -> ifc -> destroy (font);
}

struct gf_glyph_info *gf_get_glyph_info (struct gf_font *font,
					 uint8_t *string, size_t limit)
{
	if (font -> ifc -> get_glyph == NULL)
		return &null_glyph;

	return font -> ifc -> get_glyph (font, string, limit);
}

static enum s_status decode_differences (struct ct_encoding_table *e,
					 struct xref_obj_array *diff)
{
	enum s_status res;
	struct obj *obj;
	char *name;
	int i, code;

	code = -1;
	for (i = 0; i < XREF_A_SIZE (diff); i++) {
		obj = &diff -> data [i];
		switch (obj -> type) {
		case OT_NAME:
			if (code < 0)
				return obj_error (obj, "Expecting integer");

			res = obj_string_to_c (&obj -> val.string, &name);
			if (res != S_CSC_OK)
				return res;

			free (e -> e [code].name);
			e -> e [code].name = name;
			e -> e [code].unicode = ct_name_to_unicode (name);

			if (strlen (name) > e -> namelen)
				e -> namelen = strlen (name);

			code++;
			break;

		case OT_INTEGER:
			code = diff -> data [i].val.integer;
			break;

		default:
			return obj_error (obj, "Expecting integer or name");
		}
	}

	return S_CSC_OK;
}

static enum s_status fill_encoding_table (struct ct_encoding_table *e,
					  struct gf_encoding_enum *ee,
					  enum ct_encoding_type bie,
					  enum ct_encoding_type fille)
{
	struct ct_encoding_table fillt;
	struct gf_encoding *enc;
	enum s_status res;
	int i;

	if (ee -> type == GF_CT_ET_DICTIONARY) {
		enc = ee -> t;
		if (enc -> base_encoding.type != GF_CT_ET_NONE)
			bie = enc -> base_encoding.type;
	} else if (ee -> type != GF_CT_ET_NONE) {
		enc = NULL;
		bie = ee -> type;
	} else
		enc = ee -> t;

	if (bie == GF_CT_ET_NONE)
		return obj_error (&enc -> t.obj, "No built-in encoding");

	res = ct_fill_encoding (e, bie);
	if (res != S_CSC_OK)
		return res;

	if (enc != NULL) {
		res = decode_differences (e, &enc -> differences);
		if (res != S_CSC_OK) {
			ct_destroy_encoding (e);
			return res;
		}
	}

	if (ee -> type == GF_CT_ET_DICTIONARY && fille != GF_CT_ET_NONE) {
		res = ct_fill_encoding (&fillt, fille);
		if (res != S_CSC_OK) {
			ct_destroy_encoding (e);
			return res;
		}
		for (i = 0; i < 256; i++) {
			if (e -> e [i].name == NULL) {
				e -> e [i] = fillt.e [i];
				fillt.e [i].name = NULL;
			}
		}
		ct_destroy_encoding (&fillt);
	}

	return S_CSC_OK;
}

static void table_destroy (struct gf_table *table)
{
	if (table -> info != NULL)
		free (table -> info);
}

static struct gf_glyph_info *table_get_glyph (struct gf_table *table,
					      uint8_t *string, size_t limit)
{
	struct gf_glyph_info *gi;
	pdf_integer_t idx;

	idx = *string - table -> first;
	gi = idx >= 0 && idx < table -> size ? 
		&table -> info [idx] : &table -> unmapped;
	gi -> size = 1;

	return gi;
}

static void get_space_width (struct gf_font *font, struct gf_font_desc *fd,
			     struct gf_table *table)
{
	pdf_integer_t i;

	/* A reasonable default */
	font -> space_width = 0.2;

	for (i = 0; i < table -> size; i++)
		if (table -> info [i].unicode == ' ') {
			font -> space_width = table -> info [i].wx;
			break;
		}
}

static void get_leading (struct gf_font *font, struct gf_font_desc *fd,
			 struct gf_table *table)
{
	/* A reasonable default */
	font -> ascent = table -> unmapped.rect.top;
	font -> descent = table -> unmapped.rect.bottom;

	if (fd != NULL) {
		if (fd -> ascent)
			font -> ascent = fd -> ascent / 1000;
		if (fd -> descent)
			font -> descent = fd -> descent / 1000;
		if (fd -> leading)
			font -> leading = fd -> leading;
		else
			font -> leading = (font -> ascent + font -> descent) 
				* LEADING_FACTOR;
	}
}

static enum s_status map_std14 (struct gf_simple_font *sf,
				struct gf_table *table,
				struct obj *ref)
{
	struct ct_encoding_table et;
	const char *name;
	enum s_status res;
	int i, j, k;

	for (i = 0; builtin [i].name != NULL; i++) 
		if (obj_string_equal (&sf -> base_font.val.string, 
				      builtin [i].name))
			break;

	if (builtin [i].name == NULL)
		return obj_error (ref, 
				  "Missing mandatory fields in simple font");
	
	table -> info = calloc (256, sizeof (struct gf_glyph_info));
	table -> first = 0;
	table -> size = 256;

	if (sf -> encoding.type != GF_CT_ET_NONE) {
		res = fill_encoding_table (&et, &sf -> encoding,
					   CT_ET_STANDARD, GF_CT_ET_NONE);
		if (res != S_CSC_OK) {
			free (table -> info);
			return res;
		}

		for (j = 0; j < 256; j++) {
			name = ct_char_to_name (j, CT_ET_STANDARD);
			if (name == NULL)
				continue;
			for (k = 0; k < 256; k++) {
				if (et.e [k].name == NULL)
					continue;
				if (!strcmp (et.e [k].name, name)) {
					table -> info [k] = 
						builtin [i].info [j];
					break;
				}
			}
		}

		ct_destroy_encoding (&et);
	} else
		memcpy (table -> info, builtin [i].info, 
			256 * sizeof (struct gf_glyph_info));


	return S_CSC_OK;
}

#ifdef HAVE_FREETYPE2
#undef __FTERRORS_H__
#define FT_ERRORDEF(e, v, s )  { e, s },
#define FT_ERROR_START_LIST     {
#define FT_ERROR_END_LIST       { 0, NULL } };

static const struct {
	int code;
	const char *msg;
} ft_errors [] = 
#include FT_ERRORS_H
;

static const char *ft2_get_string (enum s_code code)
{
	int i;

	for (i = 0; ft_errors [i].msg != NULL; i++)
		if (ft_errors [i].code == code)
			return ft_errors [i].msg;

	return "FreeType2 error";
}

static enum s_status ft_error (FT_Error error)
{
	return S_STATUS (S_SYS_FREETYPE, error);
}

static enum s_status get_region (struct gf_font *font,
				 struct gf_font_desc *fd,
				 struct obj_region **reg)
{
	enum s_status res;
	struct obj *obj;

	if (fd -> font_file.type == OT_STREAM)
		obj = &fd -> font_file;
	else if (fd -> font_file2.type == OT_STREAM)
		obj = &fd -> font_file2;
	else if (fd -> font_file3.type == OT_STREAM)
		obj = &fd -> font_file3;
	else
		return obj_error (&font -> t.obj, "No font file");

	res = xref_resolve_stream (font -> t.rdb, obj);
	if (res != S_CSC_OK)
		return res;

	*reg = &obj -> val.stream.decoded;

	return S_CSC_OK;
}

typedef enum s_status (*ft_selector) (FT_Face face,
				      struct ct_encoding_table *et,
				      FT_ULong code, FT_UInt idx, int *oidx);

static enum s_status fill_one_glyph (FT_Face face, struct gf_table *table,
				     struct ct_encoding_table *et,
				     int code, int idx, ft_selector sel)
{
	struct gf_glyph_info *gi;
	FT_GlyphSlot glyph;
	enum s_status res;
	double em_size;
	int i;

	res = sel (face, et, code, idx, &i);
	if (res != S_CSC_OK)
		return res;

	if (i < table -> first || 
	    i >= table -> first + table -> size)
		return S_CSC_OK;

	em_size = face -> units_per_EM;

	glyph = face -> glyph;
	gi = &table -> info [i - table -> first];

	gi -> rect.left = 
		glyph -> metrics.horiBearingX / em_size;
	gi -> rect.right = 
		(glyph -> metrics.horiBearingX +
		 glyph -> metrics.width) / em_size;
	gi -> rect.top = 
		glyph -> metrics.horiBearingY / em_size;
	gi -> rect.bottom = 
		(glyph -> metrics.horiBearingY - 
		 glyph -> metrics.height) / em_size;

	gi -> wx = glyph -> metrics.horiAdvance / em_size;
	gi -> wy = 0;
	gi -> size = 1;
	gi -> unicode = et -> e [i].unicode;

	return S_CSC_OK;
}


static enum s_status fill_ft2 (struct gf_font *font,
			       struct gf_simple_font *sf,
			       FT_Face face,
			       struct gf_table *table,
			       struct ct_encoding_table *et,
			       ft_selector sel)
{
	enum s_status res;
	FT_Error error;
	FT_ULong code;
	FT_UInt idx;
	int done [face -> num_glyphs + 1];

	memset (done, 0, sizeof (done));

	error = 0;
	for (code = FT_Get_First_Char (face, &idx); idx;
	     code = FT_Get_Next_Char (face, code, &idx)) {

		error = FT_Load_Glyph (face, idx, FT_LOAD_NO_SCALE);
		if (error)
			return ft_error (error);

		res = fill_one_glyph (face, table, et, code, idx, sel);
		if (res != S_CSC_OK)
			return res;
		
		done [idx] = 1;
	}

	for (idx = 1; idx <= face -> num_glyphs; idx++) {
		if (!done [idx]) {
			error = FT_Load_Glyph (face, idx, FT_LOAD_NO_SCALE);
			if (error)
				continue;

			res = fill_one_glyph (face, table, et, 0, idx, sel);
			if (res != S_CSC_OK)
				return res;
		}
	}

	return S_CSC_OK;
}

static enum s_status tt_unicode_sel (FT_Face face,
				     struct ct_encoding_table *et,
				     FT_ULong code, FT_UInt idx, int *oidx)
{
	int i;

	for (i = 0; i < 256; i++)
		if (et -> e [i].unicode == code)
			break;

	*oidx = i;

	return S_CSC_OK;
}

static enum s_status tt_symbolic_sel (FT_Face face,
				      struct ct_encoding_table *et,
				      FT_ULong code, FT_UInt idx, 
				      int *oidx)
{
	*oidx = code & 0xff;

	return S_CSC_OK;
}

static enum s_status tt_mac_sel (FT_Face face,
				 struct ct_encoding_table *et,
				 FT_ULong code, FT_UInt idx, int *oidx)
{
	int i;

	code = ct_char_to_unicode (code, CT_ET_MAC_ROMAN);
	for (i = 0; i < 256; i++)
		if (et -> e [i].unicode == code)
			break;

	*oidx = i;

	return S_CSC_OK;
}

static enum s_status fill_truetype_ft2 (struct gf_font *font,
					struct gf_simple_font *sf,
					FT_Face face, 
					struct gf_table *table)
{
	struct ct_encoding_table et;
	enum ct_encoding_type bie;
	enum s_status res;
	FT_Error error;
	ft_selector sel;
	int i, ucmap, macmap;

	ucmap = macmap = -1;
	bie = GF_CT_ET_NONE;
	for (i = 0; i < face -> num_charmaps; i++) {
		switch (face -> charmaps [i] -> encoding) {
		case FT_ENCODING_UNICODE:
			ucmap = i;
			break;

		case FT_ENCODING_APPLE_ROMAN:
			bie = CT_ET_MAC_OS_ROMAN;
			macmap = i;
			break;

		default:
			break;
		}
	}

	if (ucmap >= 0) {
		i = ucmap;
		sel = tt_unicode_sel;
	} else if (macmap >= 0) {
		i = macmap;
		sel = tt_mac_sel;
	} else {
		i = -1;
		sel = tt_symbolic_sel;
	}

	if (sf -> font_descriptor -> flags & GF_F_SYMBOLIC ||
	    sf -> encoding.type == GF_CT_ET_NONE) {
		sel = tt_symbolic_sel;
	} 

	if (i >= 0) {
		error = FT_Set_Charmap (face, face -> charmaps [i]);
		if (error)
			return ft_error (error);
	}

	res = fill_encoding_table (&et, &sf -> encoding, bie, CT_ET_STANDARD);
	if (res != S_CSC_OK)
		return res;

	res = fill_ft2 (font, sf, face, table, &et, sel);
	ct_destroy_encoding (&et);
	if (res != S_CSC_OK)
		return res;


	return S_CSC_OK;	
}

static enum s_status type1_sel (FT_Face face,
				struct ct_encoding_table *et,
				FT_ULong code, FT_UInt idx, int *oidx)
{
	char buf [et -> namelen + 1];
	FT_Error error;
	int i;

	error = FT_Get_Glyph_Name (face, idx, buf, sizeof (buf));
	if (error)
		return ft_error (error);

	for (i = 0; i < 256; i++)
		if (et -> e [i].name != NULL &&
		    !strcmp (et -> e [i].name, buf))
			break;

	*oidx = i;

	return S_CSC_OK;
}
				
static enum s_status fill_type1_ft2 (struct gf_font *font,
				     struct gf_simple_font *sf,
				     FT_Face face, 
				     struct gf_table *table)
{
	struct ct_encoding_table et;
	enum ct_encoding_type bie;
	enum s_status res;
	FT_Error error;
	int i, uni, notuni;

	uni = notuni = -1;
	for (i = 0; i < face -> num_charmaps; i++) 
		if (face -> charmaps [i] -> encoding == FT_ENCODING_UNICODE)
			uni = i;
		else
			notuni = i;
	if (uni >= 0) {
		error = FT_Set_Charmap (face, face -> charmaps [uni]);
		if (error)
			return ft_error (error);
	} 

	if (notuni >= 0) {
		switch (face -> charmaps [notuni] -> encoding) {
		case FT_ENCODING_ADOBE_STANDARD:
			bie = CT_ET_STANDARD;
			break;

		case FT_ENCODING_ADOBE_EXPERT:
			bie = CT_ET_MAC_EXPERT;
			break;

		default:
			bie = GF_CT_ET_NONE;
		}
	} else
		bie = GF_CT_ET_NONE;

	res = fill_encoding_table (&et, &sf -> encoding, bie, GF_CT_ET_NONE);
	if (res != S_CSC_OK)
		return res;

	res = fill_ft2 (font, sf, face, table, &et, type1_sel);
	ct_destroy_encoding (&et);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;	
}

static enum s_status fill_table_ft2 (struct gf_font *font,
				     struct gf_simple_font *sf,
				     struct gf_font_desc *fd,
				     struct gf_table *table)
{
	struct obj_region *reg;
	enum s_status res;
	FT_Face face;
	FT_Error error;
	FT_Library lib;

	res = get_region (font, fd, &reg);
	if (res != S_CSC_OK)
		return res;

	error = FT_Init_FreeType (&lib);
	if (error)
		return ft_error (error);

	error = FT_New_Memory_Face (lib, reg -> begin,
				    reg -> end - reg -> begin, 0, &face);
	if (error) {
		FT_Done_FreeType (lib);
		return ft_error (error);
	}

	if (font -> subtype.type == GF_FT_TRUE_TYPE)
		res = fill_truetype_ft2 (font, sf, face, table);
	else
		res = fill_type1_ft2 (font, sf, face, table);
	FT_Done_Face (face);
	FT_Done_FreeType (lib);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status fill_one_glyph_t0 (FT_Face face, int idx,
					struct gf_t0_entry *t0e,
					enum cmap_wmode wmode,
					size_t size)
{
	struct gf_glyph_info *gi;
	FT_GlyphSlot glyph;
	double em_size;
	FT_UInt cid;

	glyph = face -> glyph;
	em_size = face -> units_per_EM;

	/* I don't know why, but it seems that the only way to
	 * obtain a CID from the GID is to assume they are the
	 * same: the following function returns something else... */
	/*
	FT_Get_CID_From_Glyph_Index (face, idx, &cid);
	*/
	cid = idx;
	t0e -> cid = cid;

	gi = &t0e -> info;

	switch (wmode) {
	case CMAP_WM_HORIZONTAL:
		gi -> rect.left = 
			glyph -> metrics.horiBearingX / em_size;
		gi -> rect.right = 
			(glyph -> metrics.horiBearingX +
			 glyph -> metrics.width) / em_size;
		gi -> rect.top = 
			glyph -> metrics.horiBearingY / em_size;
		gi -> rect.bottom = 
			(glyph -> metrics.horiBearingY - 
			 glyph -> metrics.height) / em_size;

		gi -> wx = glyph -> metrics.horiAdvance / em_size;
		gi -> wy = 0;
		break;

	case CMAP_WM_VERTICAL:
		gi -> rect.left = 
			glyph -> metrics.vertBearingX / em_size;
		gi -> rect.right = 
			(glyph -> metrics.vertBearingX +
			 glyph -> metrics.width) / em_size;
		gi -> rect.top = 
			glyph -> metrics.vertBearingY / em_size;
		gi -> rect.bottom = 
			(glyph -> metrics.vertBearingY - 
			 glyph -> metrics.height) / em_size;

		gi -> wx = 0;
		gi -> wy = glyph -> metrics.vertAdvance / em_size;
		break;
	}

	gi -> size = size;
	gi -> unicode = 0;

	return S_CSC_OK;
}

static enum s_status fill_ft2_t0 (struct gf_font *font,
				  struct gf_type0_font *t0, FT_Face face,
				  enum cmap_wmode wmode, size_t size)
{
	int done [face -> num_glyphs + 1];
	enum s_status res;
	FT_Error error;
	FT_ULong code;
	FT_UInt idx;
	int i;

	memset (done, 0, sizeof (done));

	error = 0;

	i = 0;
	for (code = FT_Get_First_Char (face, &idx); idx;
	     code = FT_Get_Next_Char (face, code, &idx)) {

		error = FT_Load_Glyph (face, idx, FT_LOAD_NO_SCALE);
		if (error)
			return ft_error (error);

		res = fill_one_glyph_t0 (face, idx, &t0 -> t0e [i],
					 wmode, size);
		if (res != S_CSC_OK)
			return res;

		done [idx] = 1;

		i++;
	}

	for (idx = 1; idx <= face -> num_glyphs; idx++) {
		if (!done [idx]) {
			error = FT_Load_Glyph (face, idx, FT_LOAD_NO_SCALE);
			if (error)
				continue;

			res = fill_one_glyph_t0 (face, idx, &t0 -> t0e [i],
						 wmode, size);
			if (res != S_CSC_OK)
				return res;

			i++;
		}
	}

	return S_CSC_OK;
}

static enum s_status ft_fill_type0 (struct gf_font *font,
				    struct gf_type0_font *t0,
				    struct gf_cid_font *descf,
				    enum cmap_wmode wmode)
{
	struct obj_region *reg;
	FT_Face face;
	FT_Error error;
	FT_Library lib;
	enum s_status res;

	res = get_region (font, descf -> font_descriptor, &reg);
	if (res != S_CSC_OK)
		return res;

	error = FT_Init_FreeType (&lib);
	if (error)
		return ft_error (error);

	error = FT_New_Memory_Face (lib, reg -> begin,
				    reg -> end - reg -> begin, 0, &face);
	if (error) {
		FT_Done_FreeType (lib);
		return ft_error (error);
	}

	t0 -> size = face -> num_glyphs;
	if (!t0 -> size) {
		FT_Done_Face (face);
		FT_Done_FreeType (lib);
		return obj_error (&font -> t.obj, "No glyphs!");
	}
	t0 -> t0e = calloc (t0 -> size, sizeof (struct gf_t0_entry));
	if (t0 == NULL) {
		FT_Done_Face (face);
		FT_Done_FreeType (lib);
		return S_CSC_OUT_OF_MEMORY;
	}

	res = fill_ft2_t0 (font, t0, face, wmode, 1);
	FT_Done_Face (face);
	FT_Done_FreeType (lib);
	if (res != S_CSC_OK)
		return res;

	if (descf -> cid_to_gid_map.type == GF_C2G_STREAM) {
		res = xref_resolve_stream (font -> t.rdb, 
					   &descf -> cid_to_gid_map.t);
		if (res != S_CSC_OK)
			return res;

		t0 -> c2g = &descf -> cid_to_gid_map.t.val.stream.decoded;
	}

	return S_CSC_OK;	
}

#endif

static int t0e_compare (const void *t0e1, const void *t0e2)
{
	const struct gf_t0_entry *t1, *t2;

	t1 = t0e1;
	t2 = t0e2;

	return t1 -> cid - t2 -> cid;
}

static void get_default_rect (struct cobj_rect *rect,
			      struct gf_font_desc *fd,
			      pdf_real_t width)
{
	rect -> left = 0;
	rect -> right = width / 1000;
	rect -> bottom = fd -> descent / 1000;
	rect -> top = fd -> ascent / 1000;
}

static enum s_status fill_table_default (struct gf_font *font,
					 struct gf_simple_font *sf,
					 struct gf_font_desc *fd,
					 struct gf_table *table)
{
	struct ct_encoding_table et;
	enum s_status res;
	int idx;

	if (!table -> size)
		return S_CSC_OK;

	res = fill_encoding_table (&et, &sf -> encoding, 
				   GF_CT_ET_NONE, GF_CT_ET_NONE);
	if (res != S_CSC_OK)
		return res;

	for (idx = 0; idx < table -> size; idx++) {
		get_default_rect (&table -> info [idx].rect, fd,
				  sf -> widths.data [idx]);
		table -> info [idx].size = 1;
		table -> info [idx].wx = sf -> widths.data [idx] / 1000;
		table -> info [idx].wy = 0;
		table -> info [idx].unicode = 
			et.e [idx + table -> first].unicode;
	}
	
	ct_destroy_encoding (&et);

	return S_CSC_OK;
}

static enum s_status build_named_encoding (struct gf_font *font,
					   struct obj *name,
					   struct psi_interpreter *pi,
					   struct cmap **cmap)
{
	if (obj_string_equal (&name -> val.string, "Identity-H"))
		return cmap_build_identity_h (cmap, font -> t.rdb,
					      ref_rown (&font -> t.refs));
	else if (obj_string_equal (&name -> val.string, "Identity-V"))
		return cmap_build_identity_v (cmap, font -> t.rdb,
					      ref_rown (&font -> t.refs));
	else
		return psi_get_cmap (pi, name, 
				     ref_rown (&font -> t.refs), cmap);
}

static enum s_status load_stream_cmap (struct gf_font *font,
				       struct obj *stream,
				       struct psi_interpreter *pi,	
				       struct cmap **cmap)
{
	struct f_file *file;
	enum s_status res;

	res = xref_resolve_stream (font -> t.rdb, stream);
	if (res != S_CSC_OK)
		return res;

	res = obj_stream_file (stream, &file);
	if (res != S_CSC_OK)
		return res;

	res = psi_add_cmap (pi, file, ref_rown (&font -> t.refs), cmap);
	f_unref (file);
	if (res != S_CSC_OK)
		return res;
	
	return S_CSC_OK;
}

static enum s_status build_stream_encoding (struct gf_font *font,
					    struct obj *stream,
					    struct psi_interpreter *pi,
					    struct cmap **cmap)
{
	enum s_status res;
	struct obj pcmap;

	res = xref_dict_resolve (font -> t.rdb, stream,
				 "UseCMap", OT_ANY, &pcmap);
	if (res != S_CSC_OK)
		return res;
	if (pcmap.type == OT_STREAM) {
		res = load_stream_cmap (font, &pcmap, pi, NULL);
		obj_destroy (&pcmap);
		if (res != S_CSC_OK)
			return res;
	}

	res = load_stream_cmap (font, stream, pi, cmap);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status build_cmap (struct gf_font *font,
				 struct obj *spec,
				 struct psi_interpreter *pi,
				 struct cmap **cmap)
{
	enum s_status res;

	switch (spec -> type) {
	case OT_NAME:
		res = build_named_encoding (font, spec, pi, cmap);
		break;

	case OT_STREAM:
		res = build_stream_encoding (font, spec, pi, cmap);
		break;

	default:
		return obj_error (spec, "Bad encoding type");
	}
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status fill_to_unicode_map (struct gf_font *font,
					  struct gf_table *table,
					  struct obj *obj)
{
	struct psi_interpreter *psi;
	struct cmap *cmap;
	enum s_status res;
	struct obj_bstr s;
	uint8_t cid;
	size_t i, len;

	if (obj -> type == OT_NULL)
		return S_CSC_OK;

	res = psi_create (&psi);
	if (res != S_CSC_OK)
		return res;

	res = build_cmap (font, obj, psi, &cmap);
	psi_destroy (psi);
	if (res != S_CSC_OK)
		return res;

	if (cmap == NULL)
		return S_CSC_OK;

	s.size = 1;
	s.s = (char *) &cid;
	cid = table -> first;
	for (i = 0; i < table -> size; i++, cid++)
		table -> info [i].unicode = cmap_get_cid (cmap, &s, &len);

	cmap_unref (cmap, ref_rown (&font -> t.refs));

	return S_CSC_OK;
}

static void fill_t0_h (struct gf_t0_entry *toe, struct gf_font_desc *fd,
		       pdf_real_t width)
{
	get_default_rect (&toe -> info.rect, fd, width);
	toe -> info.size = 1;
	toe -> info.wx = width / 1000;
	toe -> info.wy = 0;
	toe -> info.unicode = 0;
}

static enum s_status default_fill_type0_h (struct gf_font *font,
					   struct gf_type0_font *t0,
					   struct gf_cid_font *descf)
{
	struct obj_array_node *an;
	struct gf_font_desc *fd;
	struct xref_obj_array *w;
	size_t size;
	int i, idx, cid;

	w = &descf -> w;

	fd = descf -> font_descriptor;

	if (XREF_A_IS_NULL (w) || !XREF_A_SIZE (w))
		return S_CSC_OK;

	size = 0;
	for (i = 0; i < XREF_A_SIZE (w); i++) {
		if (w -> data [i].type != OT_INTEGER ||
		    i + 1 >= XREF_A_SIZE (w))
			return obj_error (&font -> t.obj, "Bad W array");
		i++;
		switch (w -> data [i].type) {
		case OT_ARRAY:
			for (an = w -> data [i].val.array.head;
			     an != NULL; an = an -> next) {
				if (!obj_cast (&an -> obj, OT_REAL))
					return obj_error (&font -> t.obj, 
							  "Bad W array");
						
				size++;
			}
			break;
		case OT_INTEGER:
			size += w -> data [i].val.integer -
				w -> data [i - 1].val.integer + 1;
			i++;
			if (i >= XREF_A_SIZE (w) ||
			    !obj_cast (&w -> data [i], OT_REAL))
				return obj_error (&font -> t.obj, 
						  "Bad W array");
			break;

		default:
			return obj_error (&font -> t.obj, "Bad W array");
		}
	}

	t0 -> size = size;
	t0 -> t0e = calloc (size, sizeof (struct gf_t0_entry));
	if (t0 -> t0e == NULL)
		return S_CSC_OUT_OF_MEMORY;

	idx = 0;
	for (i = 0; i < XREF_A_SIZE (w); i++) {
		cid = w -> data [i].val.integer;
		i++;
		switch (w -> data [i].type) {
		case OT_ARRAY:
			for (an = w -> data [i].val.array.head;
			     an != NULL; an = an -> next) {
				fill_t0_h (&t0 -> t0e [idx], fd, 
					   an -> obj.val.real);
				t0 -> t0e [idx].cid = cid;
				idx++, cid++;
			}
			break;
		case OT_INTEGER:
			i++;
			while (cid <= w -> data [i - 1].val.integer) {
				fill_t0_h (&t0 -> t0e [idx], fd, 
					   w -> data [i].val.real);
				t0 -> t0e [idx].cid = cid;
				cid++, idx++;
			}
			break;

		default:	
			assert (0);
		}
	}	

	assert (idx == size);

	return S_CSC_OK;
}

static void fill_t0_v (struct gf_t0_entry *toe, struct gf_font_desc *fd,
		       pdf_real_t wy, pdf_real_t vx, pdf_real_t vy)
{
	toe -> info.rect = fd -> font_bbox;
	if (vx < 0)
		vx = (toe -> info.rect.right - toe -> info.rect.left) / 2;
	toe -> info.rect.left /= 1000;
	toe -> info.rect.right /= 1000;
	toe -> info.rect.top = 0;
	toe -> info.rect.bottom = -wy / 1000;

	toe -> info.rect.left -= vx / 1000;
	toe -> info.rect.right -= vx / 1000;
	toe -> info.rect.top -= vy / 1000;
	toe -> info.rect.bottom -= vy / 1000;

	toe -> info.size = 1;
	toe -> info.wx = 0;
	toe -> info.wy = wy / 1000;
	toe -> info.unicode = 0;
}

static enum s_status default_fill_type0_v (struct gf_font *font,
					   struct gf_type0_font *t0,
					   struct gf_cid_font *descf)
{
	struct obj_array_node *an;
	struct gf_font_desc *fd;
	struct xref_obj_array *w;
	pdf_real_t w1, w2, w3;
	size_t size, asize;
	int i, j, idx, cid;

	w = &descf -> w2;

	fd = descf -> font_descriptor;

	if (XREF_A_IS_NULL (w) || !XREF_A_SIZE (w))
		return S_CSC_OK;

	size = 0;
	for (i = 0; i < XREF_A_SIZE (w); i++) {
		if (w -> data [i].type != OT_INTEGER ||
		    i + 1 >= XREF_A_SIZE (w))
			return obj_error (&font -> t.obj, "Bad W2 array");
		i++;
		switch (w -> data [i].type) {
		case OT_ARRAY:
			asize = 0;
			for (an = w -> data [i].val.array.head;
			     an != NULL; an = an -> next) {
				if (!obj_cast (&an -> obj, OT_REAL))
					return obj_error (&font -> t.obj, 
							  "Bad W2 array");
						
				asize++;
			}

			if (asize % 3)
				return obj_error (&font -> t.obj, 
						  "Bad W2 array");
			size += asize / 3;
			break;

		case OT_INTEGER:
			size += w -> data [i].val.integer -
				w -> data [i - 1].val.integer + 1;
			for (j = 0; j < 3; j++) {
				i++;
				if (i >= XREF_A_SIZE (w) ||
				    !obj_cast (&w -> data [i], OT_REAL))
					return obj_error (&font -> t.obj, 
							  "Bad W2 array");
			}
			break;

		default:
			return obj_error (&font -> t.obj, "Bad W2 array");
		}
	}

	t0 -> size = size;
	t0 -> t0e = calloc (size, sizeof (struct gf_t0_entry));
	if (t0 -> t0e == NULL)
		return S_CSC_OUT_OF_MEMORY;

	idx = 0;
	for (i = 0; i < XREF_A_SIZE (w); i++) {
		cid = w -> data [i].val.integer;
		i++;
		switch (w -> data [i].type) {
		case OT_ARRAY:
			for (an = w -> data [i].val.array.head;
			     an != NULL; an = an -> next) {
				w1 = an -> obj.val.real;
				an = an -> next;

				w2 = an -> obj.val.real;
				an = an -> next;

				w3 = an -> obj.val.real;

				fill_t0_v (&t0 -> t0e [idx], fd, w1, w2, w3);

				t0 -> t0e [idx].cid = cid;
				idx++, cid++;
			}
			break;
		case OT_INTEGER:
			w1 = w -> data [++i].val.real;
			w2 = w -> data [++i].val.real;
			w3 = w -> data [++i].val.real;
			while (cid <= w -> data [i - 3].val.integer) {
				fill_t0_v (&t0 -> t0e [idx], fd, 
					   w1, w2, w3);
				t0 -> t0e [idx].cid = cid;
				cid++, idx++;
			}
			break;

		default:	
			assert (0);
		}
	}	

	assert (idx == size);

	return S_CSC_OK;
}

static enum s_status fill_to_unicode_map_t0 (struct gf_font *font,
					     struct gf_type0_font *t0,
					     struct obj *obj,
					     struct cmap *ocmap)
{
	struct psi_interpreter *psi;
	struct obj_bstr ucs2map;
	struct cmap *cmap;
	enum s_status res;
	struct obj_bstr s, *registry, *ordering;
	char *ucs2name;
	struct obj ucs2obj;
	uint8_t cid;
	size_t i, len;

	if (obj -> type == OT_NULL) {
		registry = &ocmap -> cid_system_info.registry.val.string;
		ordering = &ocmap -> cid_system_info.ordering.val.string;

		/* Buggy cmaps exists ... */
		if (!registry -> size || !ordering -> size)
			return S_CSC_OK;

		len = 	registry -> size + 1 +
			ordering -> size + 1 +
			3 + 1;
		ucs2name = malloc (len);
		if (ucs2name == NULL)
			return S_CSC_OUT_OF_MEMORY;

		ucs2map.s = ucs2name;
		ucs2map.size = len;

		snprintf (ucs2name, len, "%.*s-%.*s-UCS2",
			  registry -> size, registry -> s,
			  ordering -> size, ordering -> s);
		
		obj_set_name (&ucs2obj, &f_null, &ucs2map);
		obj = &ucs2obj;
	} else
		ucs2name = NULL;

	res = psi_create (&psi);
	if (res != S_CSC_OK) {
		if (ucs2name != NULL)
			free (ucs2name);
		return res;
	}

	res = build_cmap (font, obj, psi, &cmap);
	if (ucs2name != NULL)
		free (ucs2name);
	psi_destroy (psi);
	if (res != S_CSC_OK)
		return res;

	if (cmap == NULL)
		return S_CSC_OK;

	s.size = 1;
	s.s = (char *) &cid;
	for (i = 0; i < t0 -> size; i++) {
		cid = t0 -> t0e [i].cid;
		t0 -> t0e [i].info.unicode = cmap_get_cid (cmap, &s, &len);
	}

	cmap_unref (cmap, ref_rown (&font -> t.refs));

	return S_CSC_OK;
}

static void t0_get_space_width (struct gf_font *font, 
				struct gf_type0_font *t0,
				enum cmap_wmode wmode)
{
	pdf_integer_t i, count;
	pdf_real_t ws, w;

	w = 0;
	for (i = 0; i < t0 -> size; i++)
		if (t0 -> t0e [i].info.unicode == ' ') {
			w = wmode == CMAP_WM_HORIZONTAL ? 
				t0 -> t0e [i].info.wx :
				t0 -> t0e [i].info.wy;
			font -> space_width = w;
			break;
		}
			
	if (!font -> space_width) {
		ws = count = 0;
		for (i = 0; i < t0 -> size; i++)
			w = wmode == CMAP_WM_HORIZONTAL ? 
				t0 -> t0e [i].info.wx :
				t0 -> t0e [i].info.wy;
			if (w) {
				ws += w;
				count++;
			}
		font -> space_width = ws / count;
	}
}

static void t0_get_leading (struct gf_font *font, 
			    struct gf_type0_font *t0,
			    enum cmap_wmode wmode)
{
	if (wmode == CMAP_WM_HORIZONTAL) {
		font -> leading = 
			cobj_rect_height (&t0 -> def_entry.info.rect);
		font -> ascent = t0 -> def_entry.info.rect.top;
		font -> descent = t0 -> def_entry.info.rect.bottom;
	} else {
		font -> leading = 
			cobj_rect_width (&t0 -> def_entry.info.rect);
		font -> ascent = t0 -> def_entry.info.rect.left;
		font -> descent = t0 -> def_entry.info.rect.right;
	}

	font -> leading *= LEADING_FACTOR;
}

static enum s_status type0_init (struct gf_font *font, 
				 struct xref_bind *robj)
{
	struct psi_interpreter *psi;
	struct gf_type0_font *t0;
	struct gf_cid_font *cf;
	enum cmap_wmode wmode;
	enum s_status res;

	t0 = &font -> u.type0;
	if (XREF_A_SIZE (&t0 -> descendant_fonts) != 1)
		return obj_error (&font -> t.obj, 
				  "Bad descendant fonts array");

	switch (t0 -> descendant_fonts.data [0] -> subtype.type) {
	case GF_FT_CID_FONT_TYPE0:
	case GF_FT_CID_FONT_TYPE2:
		break;

	default:
		return obj_error (&font -> t.obj, 
				  "Descendant font is not a CID font");
	}

	cf = &t0 -> descendant_fonts.data [0] -> u.cid;
	if (cf -> cid_to_gid_map.type == GF_C2G_STREAM) {
		if (cf -> cid_to_gid_map.t.type != OT_STREAM)
			return obj_error (&cf -> cid_to_gid_map.t,
					  "Bad CID2GIDMap table type");

		res = xref_resolve_stream (font -> t.rdb,
					   &cf -> cid_to_gid_map.t);
		if (res != S_CSC_OK)
			return res;
	}

	res = psi_create (&psi);
	if (res != S_CSC_OK)
		return res;

	res = build_cmap (font, &t0 -> encoding, psi, &t0 -> cmap);
	psi_destroy (psi);
	if (res != S_CSC_OK)
		return res;

	if (t0 -> cmap == NULL) {
		if (t0 -> encoding.type == OT_NAME)
			return obj_error (&font -> t.obj, 
					  "Can't find %.*s charmap",
					  t0 -> encoding.val.string.size,
					  t0 -> encoding.val.string.s);
		else
			return obj_error (&font -> t.obj, 
					  "Charmap out of sync");
	}


	wmode = t0 -> cmap -> wmode;
	switch (wmode) {
	case CMAP_WM_HORIZONTAL:
		fill_t0_h (&t0 -> def_entry, cf -> font_descriptor, cf -> dw);
		break;

	case CMAP_WM_VERTICAL:
		if (XREF_A_SIZE (&cf -> dw2) == 2)
			fill_t0_v (&t0 -> def_entry, cf -> font_descriptor, 
				   cf -> dw2.data [1], 
				   -1,
				   cf -> dw2.data [0]);
		else
			fill_t0_v (&t0 -> def_entry, 
				   cf -> font_descriptor, -1000, -1, 880);
		break;
	}

#ifdef HAVE_FREETYPE2
	res = ft_fill_type0 (font, t0, cf, wmode);
	if (res != S_CSC_OK) {
#endif
		switch (wmode) {
		case CMAP_WM_HORIZONTAL:
			res = default_fill_type0_h (font, t0, cf);
			break;

		case CMAP_WM_VERTICAL:
			res = default_fill_type0_v (font, t0, cf);
			break;
		}
		if (res != S_CSC_OK)
			return res;
#ifdef HAVE_FREETYPE2
	}
#endif

	qsort (t0 -> t0e, t0 -> size, sizeof (struct gf_t0_entry),
	       t0e_compare);

	res = fill_to_unicode_map_t0 (font, t0, &t0 -> to_unicode,
				      t0 -> cmap);
	if (res != S_CSC_OK)
		return res;

	font -> wmode = wmode;
	
	t0_get_space_width (font, t0, wmode);
	t0_get_leading (font, t0, wmode);

	return S_CSC_OK;
}

static void type0_destroy (struct gf_font *font)
{
	struct gf_type0_font *t0;
	
	t0 = &font -> u.type0;

	cmap_unref (t0 -> cmap, ref_rown (&font -> t.refs));

	if (t0 -> c2g != NULL)
		free (t0 -> c2g);

	if (t0 -> t0e != NULL)
		free (t0 -> t0e);
}

static struct gf_glyph_info *type0_get_glyph (struct gf_font *font,
					      uint8_t *string, size_t limit)
{
	struct gf_type0_font *t0;
	struct obj_bstr str;
	uint32_t cid, i;
	const uint8_t *gidp;

	t0 = &font -> u.type0;
	str.size = limit;
	str.s = (char *) string;
	cid = cmap_get_cid (t0 -> cmap, &str, &limit);

	if (t0 -> c2g != NULL) {
		gidp = t0 -> c2g -> begin + 2 * cid;
		if (gidp < t0 -> c2g -> end)
			cid = 	((gidp [0] & 255) << 8) |
				((gidp [1] & 255));
	}

	for (i = 0; i < t0 -> size; i++)
		if (t0 -> t0e [i].cid >= cid)
			break;
	
	if (i == t0 -> size || t0 -> t0e [i].cid != cid) {
		t0 -> def_entry.info.size = limit;
		return &t0 -> def_entry.info;
	}

	t0 -> t0e [i].info.size = limit;

	return &t0 -> t0e [i].info;
}

static enum s_status simple_init (struct gf_font *font, 
				  struct xref_bind *robj)
{
	struct gf_simple_font *sf;
	struct gf_font_desc *fd;
	struct gf_table *table;
	enum s_status res;

	sf = &font -> u.simple;
	table = &sf -> table;
	fd = sf -> font_descriptor;

	if (fd == NULL || XREF_A_IS_NULL (&sf -> widths) ||
	    sf -> first_char < 0 || sf -> last_char < 0) {
		res = map_std14 (sf, table, &font -> t.obj);
		if (res != S_CSC_OK)
			return res;

		get_space_width (font, NULL, table);
		get_leading (font, NULL, table);

		return S_CSC_OK;
	}

	table -> first = sf -> first_char;
	table -> size = sf -> last_char - sf -> first_char + 1;
	if (table -> size < 0)
		table -> size = 0;

	if (XREF_A_SIZE (&sf -> widths) != table -> size)
		return obj_error (&font -> t.obj, "Font width mismatch");

	get_default_rect (&table -> unmapped.rect, fd, fd -> missing_width);
	table -> unmapped.size = 1;
	table -> unmapped.wx = table -> unmapped.rect.right / 1000;
	table -> unmapped.wy = 0;
	table -> unmapped.unicode = 0;

	table -> info = calloc (table -> size, 
				sizeof (struct gf_glyph_info));
	if (table -> info == NULL)
		return S_CSC_OUT_OF_MEMORY;

#ifdef HAVE_FREETYPE2
	res = fill_table_ft2 (font, sf, fd, table);
	if (res != S_CSC_OK)
		res = fill_table_default (font, sf, fd, table);
	if (res != S_CSC_OK)
		return res;
#else
	res = fill_table_default (font, sf, fd, table);
	if (res != S_CSC_OK)
		return res;
#endif

	if (sf -> to_unicode.type != OT_NULL) {
		res = fill_to_unicode_map (font, table, &sf -> to_unicode);
		if (res != S_CSC_OK)
			return res;
	}	

	get_space_width (font, fd, table);
	get_leading (font, fd, table);

	return S_CSC_OK;
}

static void simple_destroy (struct gf_font *font)
{
	table_destroy (&font -> u.simple.table);
}

static struct gf_glyph_info *simple_get_glyph (struct gf_font *font,
					       uint8_t *string, size_t limit)
{

	return table_get_glyph (&font -> u.simple.table, string, limit);
}

static enum s_status get_t3_glyph_rect (struct xref_db *rdb,
					struct rd_resdict *rd, 
					struct gf_type3_font *t3,
					const char *name,
					struct gf_glyph_info *unmapped,
					struct cobj_rect *rect,
					pdf_real_t *wx, 
					pdf_real_t *wy)
{
	enum s_status res;
	struct obj obj;

	if (name != NULL) {
		res = xref_dict_resolve (rdb, &t3 -> char_procs,
					 name, OT_STREAM, &obj);
		if (res != S_CSC_OK)
			return res;
	} else
		obj_set_null (&obj, &f_null);

	if (obj.type == OT_NULL) {
		*wx = *wy = 0;
		*rect = unmapped -> rect;
		return S_CSC_OK;
	}

	res = xref_resolve_stream (rdb, &obj);
	if (res != S_CSC_OK) {
		obj_destroy (&obj);
		return res;
	}

	res = pdfi_get_area (rd, &obj, rect, wx, wy);
	obj_destroy (&obj);
	if (res != S_CSC_OK)
		return res;

	cobj_translate (rect, &t3 -> font_matrix);

	return S_CSC_OK;
}

static enum s_status type3_init (struct gf_font *font, struct xref_bind *robj)
{
	struct ct_encoding_table et;
	struct cobj_rect rect;
	struct gf_type3_font *t3;
	struct gf_table *table;
	struct gf_font_desc *fd;
	struct rd_resdict *rd;
	pdf_real_t wx, wy;
	char *name;
	pdf_integer_t idx;
	enum s_status res;

	t3 = &font -> u.type3;
	table = &t3 -> table;
	fd = t3 -> font_descriptor;

	table -> first = t3 -> first_char;
	table -> size = t3 -> last_char - t3 -> first_char + 1;
	if (table -> size < 0)
		table -> size = 0;

	res = fill_encoding_table (&et, &t3 -> encoding, 
				   GF_CT_ET_NONE, GF_CT_ET_NONE);
	if (res != S_CSC_OK)
		return res;

	rect = t3 -> font_bbox;
	rect.left = 0;
	rect.right = fd != NULL ? fd -> missing_width : 0;

	cobj_translate (&rect, &t3 -> font_matrix);

	table -> unmapped.rect = rect;
	table -> unmapped.wx = rect.right;
	table -> unmapped.wy = 0;
	table -> unmapped.size = 1;
	table -> unmapped.unicode = 0;

	rd = t3 -> resources != NULL ? 
		t3 -> resources : xref_bind_get (robj, rd_bkey);

	if (table -> size) {
		table -> info = calloc (table -> size, 
					sizeof (struct gf_glyph_info));
		if (table -> info == NULL) {
			ct_destroy_encoding (&et);
			return S_CSC_OUT_OF_MEMORY;
		}

		for (idx = 0; idx < table -> size; idx++) {
			name = et.e [idx].name;
			res = get_t3_glyph_rect (font -> t.rdb, rd, t3, name, 
						 &table -> unmapped, &rect,
						 &wx, &wy);
			if (res != S_CSC_OK) {
				ct_destroy_encoding (&et);
				return res;
			}
			table -> info [idx].wx = wx;
			table -> info [idx].wy = wy;
			table -> info [idx].rect = rect;
			table -> info [idx].size = 1;
			table -> info [idx].unicode = et.e [idx].unicode;
		}
	}

	ct_destroy_encoding (&et);

	if (t3 -> to_unicode.type != OT_NULL) {
		res = fill_to_unicode_map (font, table, &t3 -> to_unicode);
		if (res != S_CSC_OK)
			return res;
	}

	get_space_width (font, fd, table);
	get_leading (font, fd, table);

	return S_CSC_OK;
}

static void type3_destroy (struct gf_font *font)
{
	table_destroy (&font -> u.type3.table);
}

static struct gf_glyph_info *type3_get_glyph (struct gf_font *font,
					      uint8_t *string, size_t limit)
{
	return table_get_glyph (&font -> u.type3.table, string, limit);
}

enum s_status gff_init (void)
{
#ifdef HAVE_FREETYPE2
	return s_register (&sys); 
#else
	return S_CSC_OK;
#endif
}

void gff_cleanup (void)
{
#ifdef HAVE_FREETYPE2
	s_unregister (&sys); 
#endif
}
