/*
 * PDFlow -- function.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#define _ISOC99_SOURCE

#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "hash.h"
#include "fmt/fields.h"
#include "ps/psi.h"
#include "types/function.h"

static enum s_status dyn_ifc (enum xref_op op, struct xref_field *fd, 
			      void *data, int *stop);

static enum s_status fun_region_parser (struct xref_db *rdb, 
					struct xref_bind *robj, 
					struct ref_owner *p, 
					struct obj *src, 
					struct xref_field *fd,
					const char *fname, int line);
static enum s_status fun_region_ser (struct xref_db *rdb, 
				     struct xref_bind *robj,
				     struct v_selection *sel, 
				     struct obj *dest, 
				     enum xref_changed *changed,
				     struct xref_field *fd);
static void fun_region_funref (struct ref_owner *p, 
			       struct xref_field *fd);

static enum s_status sampled_init (struct fun_function *func,
				   struct xref_db *rdb, struct obj *obj);
static enum s_status sampled_exec (struct fun_function *func,
				   struct fun_request *request);
static enum s_status sampled_get_checksum (struct fun_function *func, 
					   struct hash_status *fs);
static void sampled_destroy (struct fun_function *func);

static enum s_status exponential_exec (struct fun_function *func,
				       struct fun_request *request);
static enum s_status exponential_get_checksum (struct fun_function *func, 
					       struct hash_status *fs);

static enum s_status stitching_init (struct fun_function *func,
				     struct xref_db *rdb, struct obj *obj);
static enum s_status stitching_exec (struct fun_function *func,
				     struct fun_request *request);
static enum s_status stitching_get_checksum (struct fun_function *func, 
					     struct hash_status *fs);

static enum s_status postscript_init (struct fun_function *func,
				      struct xref_db *rdb, struct obj *obj);
static enum s_status postscript_exec (struct fun_function *func,
				      struct fun_request *request);
static enum s_status postscript_get_checksum (struct fun_function *func, 
					      struct hash_status *fs);
static void postscript_destroy (struct fun_function *func);

struct fun_ifc {
	enum fun_type type;

	struct xref_field *fd;

	enum s_status (*init) (struct fun_function *func,
			       struct xref_db *rdb, struct obj *obj);

	enum s_status (*exec) (struct fun_function *func,
			       struct fun_request *request);

	enum s_status (*get_checksum) (struct fun_function *func, 
				       struct hash_status *ctxt);
	
	void (*destroy) (struct fun_function *func);
};

struct xref_pstable fun_region_pstable = {
	sizeof (struct fun_region),
	fun_region_parser, fun_region_ser, fun_region_funref
};

#define XREF_IMPLEMENTATION struct fun_function
#define FUN_FUNCTION
#include "types/function.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct fun_sampled
#define FUN_SAMPLED
#include "types/function.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct fun_exponential
#define FUN_EXPONENTIAL
#include "types/function.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct fun_stitching
#define FUN_STITCHING
#include "types/function.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct fun_postscript
#define FUN_POSTSCRIPT
#include "types/function.tab.h"
#undef XREF_IMPLEMENTATION

static struct fun_ifc sampled_ifc = {
	FUN_T_SAMPLED, &fun_sampled_field,
	sampled_init, sampled_exec, sampled_get_checksum, sampled_destroy
};
static struct fun_ifc exponential_ifc = {
	FUN_T_EXPONENTIAL, &fun_exponential_field,
	NULL, exponential_exec, exponential_get_checksum, NULL
};
static struct fun_ifc stitching_ifc = {
	FUN_T_STITCHING, &fun_stitching_field,
	stitching_init, stitching_exec, stitching_get_checksum, NULL
};
static struct fun_ifc postscript_ifc = {
	FUN_T_POSTSCRIPT, &fun_postscript_field, postscript_init, 
	postscript_exec, postscript_get_checksum, postscript_destroy
};

static struct fun_ifc *ifcs [] = {
	&sampled_ifc, &exponential_ifc, &stitching_ifc, &postscript_ifc,
	NULL
};

T_IMPLEMENTATIONS (fun, fun_function, &fun_function_field);

static enum s_status parse_region (struct xref_real_array *src,
				   struct fun_region *dest,
				   struct obj *refobj)
{
	int i;

	if (XREF_A_SIZE (src) & 1)
		return obj_error (refobj, "Odd number of bounds");

	dest -> m = XREF_A_SIZE (src) / 2;
	if (!dest -> m) {
		dest -> bound = NULL;
		return S_CSC_OK;
	}
		
	dest -> bound = calloc (dest -> m, sizeof (struct fun_bound));
	if (dest -> bound == NULL)
		return S_CSC_OUT_OF_MEMORY;

	for (i = 0; i < dest -> m; i++) {
		dest -> bound [i].lower = src -> data [2 * i];
		dest -> bound [i].upper = src -> data [2 * i + 1];
	}

	return S_CSC_OK;
}

static enum s_status fun_region_parser (struct xref_db *rdb, 
					struct xref_bind *robj, 
					struct ref_owner *p, 
					struct obj *src, 
					struct xref_field *fd,
					const char *fname, int line)
{
	struct xref_real_array ma;
	enum s_status res;

	res = xref__array_map_ps (rdb, robj, p, src, &ma, 
				  "Region", &xref_real_pstable, NULL,
				  fname, line);
	if (res != S_CSC_OK)
		return res;

	res = parse_region (&ma, fd -> field, src);
	xref_array_map_rel_ps (&ma, p, &xref_real_pstable, NULL);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status region_copy (struct fun_region *src,
				  struct xref_real_array *dest)
{
	int i;

	dest -> n = 2 * src -> m;
	dest -> data = calloc (dest -> n, sizeof (pdf_real_t));
	if (dest -> data == NULL)
		return S_CSC_OUT_OF_MEMORY;

	for (i = 0; i < src -> m; i++) {
		dest -> data [2 * i] = src -> bound [i].lower;
		dest -> data [2 * i + 1] = src -> bound [i].upper;
	}

	return S_CSC_OK;
}

static enum s_status region_get_checksum (struct fun_region *reg, 
					  struct hash_status *fs)
{
	enum s_status res;
	int i;

	res = hash_update (fs, &reg -> m, sizeof (pdf_integer_t));
	if (res != S_CSC_OK)
		return res;

	for (i = 0; i < reg -> m; i++) {
		res = hash_update (fs, &reg -> bound [i].lower, 
				   sizeof (pdf_real_t));
		if (res != S_CSC_OK)
			return res;

		res = hash_update (fs, &reg -> bound [i].upper, 
				   sizeof (pdf_real_t));
		if (res != S_CSC_OK)
			return res;
	}

	return S_CSC_OK;
}

static enum s_status fun_region_ser (struct xref_db *rdb, 
				     struct xref_bind *robj,
				     struct v_selection *sel, 
				     struct obj *dest, 
				     enum xref_changed *changed,
				     struct xref_field *fd)
{
	struct xref_real_array ma;
	struct fun_region *reg;
	enum s_status res;

	reg = fd -> field;
	if (!reg -> m)
		return S_CSC_OK;

	res = region_copy (reg, &ma);
	if (res != S_CSC_OK)
		return res;

	res = xref_array_unmap_ps (rdb, robj, sel, &ma, dest, 0,
				   &xref_real_pstable, NULL);
	free (ma.data);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static void fun_region_funref (struct ref_owner *p, 
			       struct xref_field *fd)
{
	struct fun_region *reg;

	reg = fd -> field;
	if (reg -> bound != NULL)
		free (reg -> bound);
}

static enum s_status dyn_ifc (enum xref_op op, struct xref_field *fd, 
			      void *data, int *stop)
{
	struct fun_function *f;
	int i;

	f = data;

	if (f -> ifc == NULL) {
		for (i = 0; ifcs [i] != NULL; i++)
			if (f -> function_type == ifcs [i] -> type)
				break;

		f -> ifc = ifcs [i];
	}

	if (f -> ifc == NULL) {
		switch (op) {
		case XREF_OP_PARSE:
			return obj_error (&f -> t.obj, 
					  "Unsupported function type: %lu",
					  f -> function_type);
			break;

		case XREF_OP_SERIALIZE:
			return s_make_indirect 
				(S_CSC_INTERNAL_ERROR,
				 "Trying to serialize function type %lu",
				 f -> function_type);

		case XREF_OP_DESTROY:
			/* Can only happen if no initialization
			 * has been done, otherwise the interface 
			 * should have already been set
			 */
			*stop = 1;
			return S_CSC_OK;
		}
	}

	*fd = *f -> ifc -> fd;

	return S_CSC_OK;
}

enum s_status fun_exec (struct fun_function *func, struct fun_request *request)
{
	enum s_status res;
	pdf_integer_t i;

	for (i = 0; i < func -> domain.m; i++) {
		if (request -> in [i] < func -> domain.bound [i].lower)
			request -> in [i] = func -> domain.bound [i].lower;
		else if (request -> in [i] > func -> domain.bound [i].upper)
			request -> in [i] = func -> domain.bound [i].upper;
	}		

	res = func -> ifc -> exec (func, request);
	if (res != S_CSC_OK)
		return res;

	if (!func -> range.m)
		return S_CSC_OK;

	for (i = 0; i < func -> range.m; i++) {
		if (request -> out [i] < func -> range.bound [i].lower)
			request -> out [i] = func -> range.bound [i].lower;
		else if (request -> out [i] > func -> range.bound[i].upper)
			request -> out [i] = func -> range.bound [i].upper;
	}		

	return S_CSC_OK;
}

static const uint8_t valid_sizes [] = {1, 2, 4, 8, 12, 16, 24}; 

static enum s_status load_samples (struct obj *refobj,
				   const struct obj_region *reg,
				   pdf_integer_t bps,
				   pdf_integer_t *samples,
				   size_t scount)
{
	uint32_t smask, dmask;
	const uint8_t *sptr;
	uint8_t u8;
	size_t i;

	if (bps <= 0 || bps > 255)
		return obj_error (refobj, "Bad samples size (%d)", bps);

	if (memchr (valid_sizes, bps, sizeof (valid_sizes)) == NULL)
		return obj_error (refobj, "Bad samples size (%d)", bps);

	smask = dmask = u8 = 0;
	i = -1;
	sptr = reg -> begin;
	while (1) {
		if (!dmask) {
			if (++i >= scount)
				break;

			dmask = 1 << (bps - 1);
			samples [i] = 0;
		}
		if (!smask) {
			if (sptr >= reg -> end)
				return obj_error (refobj, "Stream too short");
			
			smask = 0x80;
			u8 = *(sptr++);
		}

		if (u8 & smask)
			samples [i] |= dmask;

		smask >>= 1;
		dmask >>= 1;
	}

	return S_CSC_OK;
}
				   
static enum s_status sampled_init (struct fun_function *func,
				   struct xref_db *rdb, struct obj *obj)
{
	struct fun_sampled *data;
	enum s_status res;
	size_t scount;
	int i;

	data = &func -> u.sampled;

	if (obj -> type != OT_STREAM)
		return obj_error (obj, "Sample func must be a stream");
	if (func -> domain.m <= 0)
		return obj_error (obj, "Sample func needs a range");
	if (func -> domain.m != XREF_A_SIZE (&data -> size))
		return obj_error (obj, "Bad Size array size");

	scount = 1;
	for (i = 0; i < func -> domain.m; i++)
		scount *= data -> size.data [i];

	data -> scount = scount;
	data -> samples = calloc (scount * func -> range.m, 
				  sizeof (pdf_integer_t));
	if (data -> samples == NULL)
		return S_CSC_OUT_OF_MEMORY;

	res = load_samples (obj, &obj -> val.stream.decoded,
			    data -> bits_per_sample, data -> samples, 
			    scount * func -> range.m);
	if (res != S_CSC_OK) {
		free (data -> samples);
		return res;
	}

	return S_CSC_OK;
}

static void interpolate (pdf_real_t *x, 
			 pdf_real_t x1, pdf_real_t x2,
			 pdf_real_t y1, pdf_real_t y2)
{
	*x = y1 + ((*x - x1) * (y2 - y1) / (x2 -x1));
}

static pdf_real_t u0 (struct fun_function *func,
		      struct fun_sampled *data,
		      pdf_integer_t *p)
{
	pdf_integer_t i, idx;

	idx = 0;
	for (i = func -> domain.m - 1; i >= 0; i--) {

		if (p [i] < 0 || 
		    p [i] >= data -> size.data [i])
			return 0;

		idx += p [i];
		if (i)
			idx *= data -> size.data [i - 1];
	}

	return data -> samples [idx];
}

static pdf_real_t u1 (struct fun_function *func,
		      struct fun_sampled *data,
		      pdf_integer_t i,
		      pdf_integer_t *p)
{
	pdf_integer_t dp [func -> domain.m];

	memcpy (dp, p, sizeof (dp));
	dp [i]++;
	if (dp [i] < 0 || dp [i] >= data -> size.data [i])
		return 0;

	return u0 (func, data, dp) - u0 (func, data, p);
}

static pdf_real_t u2 (struct fun_function *func,
		      struct fun_sampled *data,
		      pdf_integer_t i,
		      pdf_integer_t j,
		      pdf_integer_t *p)
{
	pdf_integer_t dp [2][func -> domain.m];
	
	memcpy (dp [0], p, sizeof (dp [0]));
	memcpy (dp [1], p, sizeof (dp [1]));
	dp [0][i]++;
	dp [1][j]++;
	if (dp [0][i] < 0 || dp [0][i] >= data -> size.data [i] ||
	    dp [1][j] < 0 || dp [1][j] >= data -> size.data [j])
		return 0;

	return (u1 (func, data, i, dp [1]) - 
		u1 (func, data, i, p));
}

static pdf_real_t u3 (struct fun_function *func,
		      struct fun_sampled *data,
		      pdf_integer_t i,
		      pdf_integer_t j,
		      pdf_integer_t k,
		      pdf_integer_t *p)
{
	pdf_integer_t dp [3][func -> domain.m];

	memcpy (dp [0], p, sizeof (dp [0]));
	memcpy (dp [1], p, sizeof (dp [1]));
	memcpy (dp [2], p, sizeof (dp [2]));
	dp [0][i]++;
	dp [1][j]++;
	dp [2][k]--;
	if (dp [0][i] < 0 || dp [0][i] >= data -> size.data [i] ||
	    dp [1][j] < 0 || dp [1][j] >= data -> size.data [j] ||
	    dp [2][k] < 0 || dp [2][k] >= data -> size.data [k])
		return 0;

	return (u2 (func, data, i, j, dp [2]) - 
		u2 (func, data, i, j, p));
}

static int fact3 (pdf_integer_t i, pdf_integer_t j, pdf_integer_t k)
{
	return i != j ?
		j != k ? 1 : 2 :
		j != k ? 2 : 6;
}

static int fact2 (pdf_integer_t i, pdf_integer_t j)
{
	return i != j ? 1 : 2;
}

static enum s_status linear_interp (struct fun_function *func,
				    struct fun_sampled *data,
				    struct fun_request *request,
				    int dofs)
{
	pdf_integer_t p [func -> domain.m];
	pdf_real_t ans, coef, mon [2];
	pdf_integer_t i;

	for (i = 0; i < func -> domain.m; i++)
		p [i] = request -> in [i];

	ans = 0;
	mon [0] = 1;
	for (i = 0; i < func -> domain.m; i++) {
		mon [1] = request -> in [i] - p [i];
		coef = u1 (func, data, i, p);
		ans += coef * mon [1];
	}

	coef = u0 (func, data, p);
	ans += coef * mon [0];

	request -> out [dofs] = ans;

	return S_CSC_OK;
}

static enum s_status spline_interp (struct fun_function *func,
				    struct fun_sampled *data,
				    struct fun_request *request,
				    int dofs)
{
	pdf_integer_t p [func -> domain.m];
	pdf_real_t ans, coef, mon [4];
	pdf_integer_t i, j, k;

	for (i = 0; i < func -> domain.m; i++)
		p [i] = request -> in [i];

	ans = 0;
	mon [0] = 1;
	for (i = 0; i < func -> domain.m; i++) {
		mon [1] = request -> in [i] - p [i];
		for (j = i; j < func -> domain.m; j++) {
			mon [2] = mon [1] * (request -> in [j] - p [j]);
			for (k = j; k < func -> domain.m; k++) {
				mon [3] = mon [2] * 
					(request -> in [k] - p [k]);
				coef = u3 (func, data, i, j, k, p);
				ans += coef * mon [3] / fact3 (i, j, k);
			}
			coef = u2 (func, data, i, j, p);
			ans += coef * mon [2] / fact2 (i, j);
		}
		coef = u1 (func, data, i, p);
		ans += coef * mon [1];
	}

	coef = u0 (func, data, p);
	ans += coef * mon [0];

	request -> out [dofs] = ans;

	return S_CSC_OK;
}

static enum s_status sampled_exec (struct fun_function *func,
				   struct fun_request *request)
{
	struct fun_sampled *data;
	pdf_integer_t i, max_out, iofs;
	enum s_status res;

	data = &func -> u.sampled;

	if (data -> encode.m) 
		for (i = 0; i < func -> domain.m; i++)
			interpolate (&request -> in [i],
				     func -> domain.bound [i].lower,
				     func -> domain.bound [i].upper,
				     data -> encode.bound [i].lower,
				     data -> encode.bound [i].upper);

	for (i = 0; i < func -> domain.m; i++)
		if (request -> in [i] < 0)
			request -> in [i] = 0;
		else if (request -> in [i] > (data -> size.data [i] - 1))
			request -> in [i] = data -> size.data [i] - 1;

	iofs = 0;
	for (i = 0; i < func -> range.m; i++) {
		if (data -> order == 1)
			res = linear_interp (func, data + iofs, request, i);
		else
			res = spline_interp (func, data + iofs, request, i);
		if (res != S_CSC_OK)
			return res;
		iofs += data -> scount;
	}

	max_out = (1 << data -> bits_per_sample) - 1;
	if (data -> decode.m) 
		for (i = 0; i < func -> range.m; i++)
			interpolate (&request -> out [i],
				     0, max_out,
				     data -> decode.bound [i].lower,
				     data -> decode.bound [i].upper);


	return S_CSC_OK;
}

static enum s_status sampled_get_checksum (struct fun_function *func, 
					   struct hash_status *fs)
{
	struct fun_sampled *data;
	enum s_status res;
	int i;

	data = &func -> u.sampled;

	res = hash_update (fs, &data -> size.n, sizeof (data -> size.n));
	if (res != S_CSC_OK)
		return res;

	for (i = 0; i < XREF_A_SIZE (&data -> size); i++) {
		res = hash_update (fs, &data -> size.data [i], 
				   sizeof (pdf_integer_t));
		if (res != S_CSC_OK)
			return res;
	}

	res = hash_update (fs, &data -> bits_per_sample, sizeof (pdf_integer_t));
	if (res != S_CSC_OK)
		return res;

	res = hash_update (fs, &data -> order, sizeof (pdf_integer_t));
	if (res != S_CSC_OK)
		return res;

	res = region_get_checksum (&data -> encode, fs);
	if (res != S_CSC_OK)
		return res;

	res = region_get_checksum (&data -> decode, fs);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static void sampled_destroy (struct fun_function *func)
{
	struct fun_sampled *data;

	data = &func -> u.sampled;

	if (data -> samples != NULL)
		free (data -> samples);
}

static enum s_status exponential_exec (struct fun_function *func,
				       struct fun_request *request)
{
	struct fun_exponential *data;
	pdf_real_t ans, c0, c1;
	pdf_integer_t i;

	data = &func -> u.exponential;
	for (i = 0; i < func -> range.m; i++) {
		c0 = !XREF_A_IS_NULL (&data -> c0) ? data -> c0.data [i] : 0;
		c1 = !XREF_A_IS_NULL (&data -> c1) ? data -> c1.data [i] : 1;
		ans = c0;
		ans += pow (request -> in [i], data -> n) * (c1 - c0);
		if (ans == HUGE_VAL || ans == NAN)
			return s_make_indirect (S_CSC_PARSE_ERROR,
						"Function overflow");
		request -> out [i] = ans;
	}

	return S_CSC_OK;
}

static enum s_status exponential_get_checksum (struct fun_function *func, 
					       struct hash_status *fs)
{
	struct fun_exponential *data;
	enum s_status res;
	int i;

	data = &func -> u.exponential;
	res = hash_update (fs, &data -> c0.n, sizeof (data -> c0.n));
	if (res != S_CSC_OK)
		return res;

	for (i = 0; i < XREF_A_SIZE (&data -> c0); i++) {
		res = hash_update (fs, &data -> c0.data [i], 
				   sizeof (pdf_real_t));
		if (res != S_CSC_OK)
			return res;
	}

	res = hash_update (fs, &data -> c1.n, sizeof (data -> c0.n));
	for (i = 0; i < XREF_A_SIZE (&data -> c1); i++) {
		res = hash_update (fs, &data -> c1.data [i], 
				   sizeof (pdf_real_t));
		if (res != S_CSC_OK)
			return res;
	}

	res = hash_update (fs, &data -> n, sizeof (pdf_integer_t));
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status stitching_init (struct fun_function *func,
				     struct xref_db *rdb, struct obj *obj)
{
	if (obj -> type != OT_DICTIONARY)
		return obj_error (obj, "Stitching func must be a dictionary");
	if (func -> domain.m != 1)
		return obj_error (obj, "Stitching func takes one arg");
	if (func -> range.m > 0 && func -> range.m != 1)
		return obj_error (obj, "Stitching func is mondimensional");

	return S_CSC_OK;
}

static enum s_status stitching_exec (struct fun_function *func,
				     struct fun_request *request)
{
	struct fun_stitching *data;
	struct fun_bound b;
	int i;

	data = &func -> u.stitching;
	for (i = 0; i < XREF_A_SIZE (&data -> functions); i++) {
		if (request -> in [0] < data -> bounds.data [i])
			break;
	}

	b.lower = i == 0 ? 
		func -> domain.bound -> lower : data -> bounds.data [i - 1];
	b.upper = i == XREF_A_SIZE (&data -> functions) - 1 ? 
		func -> domain.bound -> upper : data -> bounds.data [i];

	interpolate (&request -> in [0],
		     b.lower,
		     b.upper,
		     data -> encode.bound [i].lower,
		     data -> encode.bound [i].upper);

	return fun_exec (data -> functions.data [i], request);
}

static enum s_status stitching_get_checksum (struct fun_function *func, 
					     struct hash_status *fs)
{
	struct fun_stitching *data;
	enum s_status res;
	int i;

	data = &func -> u.stitching;

	res = hash_update (fs, &data -> functions.n, sizeof (data -> functions.n));
	if (res != S_CSC_OK)
		return res;
	for (i = 0; i < XREF_A_SIZE (&data -> functions); i++) {
		res = func -> ifc -> get_checksum 
			(data -> functions.data [i], fs);
		if (res != S_CSC_OK)
			return res;
	}

	res = hash_update (fs, &data -> bounds.n, sizeof (data -> bounds.n));
	if (res != S_CSC_OK)
		return res;

	res = hash_update (fs, &data -> bounds.n, sizeof (data -> bounds.n));
	for (i = 0; i < XREF_A_SIZE (&data -> bounds); i++) {
		res = hash_update (fs, &data -> bounds.data [i], 
				   sizeof (pdf_real_t));
		if (res != S_CSC_OK)
			return res;
	}

	res = region_get_checksum (&data -> encode, fs);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status postscript_init (struct fun_function *func,
				      struct xref_db *rdb, struct obj *obj)
{
	struct fun_postscript *data;
	enum s_status res;

	data = &func -> u.postscript;
	if (obj -> type != OT_STREAM)
		return obj_error (obj, "Postscipt func must be a stream");
	if (func -> range.m <= 0)
		return obj_error (obj, "Postscript func needs a range");

	res = f_reg_open (obj -> ptr.file, 
			  obj -> val.stream.decoded.begin, 
			  obj -> val.stream.decoded.end,
			  NULL, NULL,
			  &data -> file);
	if (res != S_CSC_OK)
		return res;

	res = psi_create (&data -> pi);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status postscript_exec (struct fun_function *func,
				      struct fun_request *request)
{
	struct fun_postscript *data;
	struct obj in [func -> domain.m], out [func -> range.m];
	enum s_status res;
	pdf_integer_t i;

	data = &func -> u.postscript;

	for (i = 0; i < func -> domain.m; i++)
		obj_set_real (&in [i], &func -> t.obj.ptr, request -> in [i]);

	res = psi_eval_f (data -> pi, data -> file, 
			  func -> domain.m, in, func -> range.m, out);
	if (res != S_CSC_OK)
		return res;

	for (i = 0; i < func -> range.m; i++) {
		obj_cast (&out [i], OT_REAL);
		request -> out [i] = out [i].val.real;
	}

	return S_CSC_OK;
}

static enum s_status postscript_get_checksum (struct fun_function *func, 
					      struct hash_status *fs)
{
	struct fun_postscript *data;
	enum s_status res;

	data = &func -> u.postscript;
	if (data -> file != NULL && data -> file -> pdf != NULL) {
		res = hash_update (fs, data -> file -> pdf,
				   data -> file -> end - data -> file -> pdf);
		if (res != S_CSC_OK)
			return res;
	}

	return S_CSC_OK;
}

static void postscript_destroy (struct fun_function *func)
{
	struct fun_postscript *data;

	data = &func -> u.postscript;
	if (data != NULL) {
		if (data -> file != NULL)
			f_unref (data -> file);

		if (data -> pi != NULL)
			psi_destroy (data -> pi);
	}
}

enum s_status fun_get_checksum (struct fun_function *func, struct hash *cks)
{
	struct hash_status *ctxt;
	enum s_status res;

	if (func == NULL) {
		memset (cks, 0, sizeof (struct hash));
		return S_CSC_OK;
	}

	res = hash_init (&ctxt);
	if (res != S_CSC_OK)
		return res;

	res = func -> ifc -> get_checksum (func, ctxt);
	if (res != S_CSC_OK)
		return res;

	res = hash_final (ctxt, cks);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}
