/*
 * PDFlow -- oc.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>
#include <stdlib.h>

#include "types/cobj.h"
#include "types/oc.h"

static enum s_status group_list_entry_parser (struct xref_db *rdb, 
					      struct xref_bind *robj,
					      struct ref_owner *p,
					      struct obj *src,
					      struct xref_field *fd,
					      const char *fname, 
					      int line);
static enum s_status group_list_entry_ser (struct xref_db *rdb, 
					   struct xref_bind *robj,
					   struct v_selection *sel,
					   struct obj *dest,
					   enum xref_changed *changed,
					   struct xref_field *fd);
static void group_list_entry_funref (struct ref_owner *p,
				     struct xref_field *fd);

static enum s_status reference_parser (struct xref_db *rdb, 
				       struct xref_bind *robj,
				       struct ref_owner *p,
				       struct obj *src,
				       struct xref_field *fd,
				       const char *fname,
				       int line);
static enum s_status reference_ser (struct xref_db *rdb, 
				    struct xref_bind *robj,
				    struct v_selection *sel,
				    struct obj *dest,
				    enum xref_changed *changed,
				    struct xref_field *fd);
static void reference_funref (struct ref_owner *p,
			      struct xref_field *fd);

static enum s_status parse_ve (struct xref_db *rdb, 
			       struct oc_properties *ocp,
			       struct ref_owner *p,
			       struct obj *src,
			       struct oc_ve **vep,
			       const char *fname, int line);
static enum s_status serialize_ve (struct xref_db *rdb, 
				   struct xref_bind *robj,
				   struct v_selection *sel,
				   struct oc_ve *ve, struct obj *dest);
static void release_ve (struct ref_owner *p, struct oc_ve *ve);

static enum s_status ve_parser (struct xref_db *rdb, 
				struct xref_bind *robj,
				struct ref_owner *p,
				struct obj *src,
				struct xref_field *fd,
				const char *fname, int line);
static enum s_status ve_ser (struct xref_db *rdb, 
			     struct xref_bind *robj,
			     struct v_selection *sel,
			     struct obj *dest,
			     enum xref_changed *changed,
			     struct xref_field *fd);
static void ve_funref (struct ref_owner *p,
		       struct xref_field *fd);

const char *oc_properties_bkey = "oc_properties bind key";

static struct cobj_enum_desc intent_desc [] = {

	{ OC_I_VIEW, "View" },
	{ OC_I_DESIGN, "Design" },

	{ -1, NULL }

};

static struct cobj_enum_desc base_state_desc [] = {

	{ OC_CFG_BS_ON, "ON" },
	{ OC_CFG_BS_OFF, "OFF" },
	{ OC_CFG_BS_UNCHANGED, "Unchanged" },

	{ OC_CFG_BS_ON, "/Default" },
	{ -1, NULL }

};

static struct cobj_enum_desc state_desc [] = {

	{ OC_S_ON, "ON" },
	{ OC_S_OFF, "OFF" },

	{ OC_S_OFF, "/Default" },

	{ -1, NULL }

};

static struct cobj_enum_desc utype_desc [] = {

	{ OC_U_UT_INDIVIDUAL, "Ind" },
	{ OC_U_UT_TITLE, "Ttl" },
	{ OC_U_UT_ORGANIZATION, "Org" },

	{ -1, NULL }
};

static struct cobj_enum_desc petype_desc [] = {

	{ OC_U_PET_HEADER_FOOTER, "HF" },
	{ OC_U_PET_FOREGROUND, "FG" },
	{ OC_U_PET_BACKGROUND, "BG" },
	{ OC_U_PET_LOGO, "L" },

	{ -1, NULL }
};

static struct cobj_enum_desc list_mode_desc [] = {

	{ OC_LM_ALL_PAGES, "AllPages" },
	{ OC_LM_VISIBLE_PAGES, "VisiblePages" },

	{ OC_LM_ALL_PAGES, "/Default" },
	{ -1, NULL }

};

static struct cobj_enum_desc p_desc [] = {

	{ OC_DP_ALL_ON, "AllOn" },
	{ OC_DP_ANY_ON, "AnyOn" },
	{ OC_DP_ALL_OFF, "AllOff" },
	{ OC_DP_ANY_OFF, "AnyOff" },
	{ -1, NULL }

};

static struct cobj_enum_desc op_desc [] = {

	{ OC_DO_NOT, "Not" },
	{ OC_DO_AND, "And" },
	{ OC_DO_OR, "Or" },
	{ -1, NULL }

};

static struct cobj_enum_desc event_desc [] = {

	{ OC_ET_VIEW, "View" },
	{ OC_ET_PRINT, "Print" },
	{ OC_ET_EXPORT, "Export" },
	{ -1, NULL }

};

struct xref_pstable oc_group_list_entry_pstable = {
	sizeof (struct oc_group_list_entry),
	group_list_entry_parser,
	group_list_entry_ser,
	group_list_entry_funref
};

struct xref_pstable oc_reference_pstable = {
	sizeof (struct oc_reference),
	reference_parser,
	reference_ser,
	reference_funref
};

struct xref_pstable oc_ve_pstable = {
 	sizeof (struct oc_ve), ve_parser, ve_ser, ve_funref
};

#define XREF_IMPLEMENTATION struct oc_group
#define OC_GROUP
#include "types/oc.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct oc_directory
#define OC_DIRECTORY
#include "types/oc.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct oc_properties
#define OC_PROPERTIES
#include "types/oc.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct oc_config
#define OC_CONFIG
#include "types/oc.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct oc_u_creator_info
#define OC_U_CREATOR_INFO
#include "types/oc.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct oc_u_language
#define OC_U_LANGUAGE
#include "types/oc.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct oc_u_export
#define OC_U_EXPORT
#include "types/oc.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct oc_u_zoom
#define OC_U_ZOOM
#include "types/oc.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct oc_u_print
#define OC_U_PRINT
#include "types/oc.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct oc_u_view
#define OC_U_VIEW
#include "types/oc.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct oc_u_user
#define OC_U_USER
#include "types/oc.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct oc_u_page_element
#define OC_U_PAGE_ELEMENT
#include "types/oc.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct oc_usage
#define OC_USAGE
#include "types/oc.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct oc_app_dir
#define OC_APP_DIR
#include "types/oc.tab.h"
#undef XREF_IMPLEMENTATION

T_IMPLEMENTATIONS (oc_group, oc_group, &oc_group_field);
T_IMPLEMENTATIONS (oc_directory, oc_directory, &oc_directory_field);
T_IMPLEMENTATIONS (oc_config, oc_config, &oc_config_field);
T_IMPLEMENTATIONS (oc_properties, oc_properties, &oc_properties_field);
T_IMPLEMENTATIONS (oc_app_dir, oc_app_dir, &oc_app_dir_field);
T_IMPLEMENTATIONS (oc_u_creator_info, oc_u_creator_info, 
		   &oc_u_creator_info_field);
T_IMPLEMENTATIONS (oc_u_language, oc_u_language, &oc_u_language_field);
T_IMPLEMENTATIONS (oc_u_export, oc_u_export, &oc_u_export_field);
T_IMPLEMENTATIONS (oc_u_zoom, oc_u_zoom, &oc_u_zoom_field);
T_IMPLEMENTATIONS (oc_u_print, oc_u_print, &oc_u_print_field);
T_IMPLEMENTATIONS (oc_u_view, oc_u_view, &oc_u_view_field);
T_IMPLEMENTATIONS (oc_u_user, oc_u_user, &oc_u_user_field);
T_IMPLEMENTATIONS (oc_u_page_element, oc_u_page_element, 
		   &oc_u_page_element_field);
T_IMPLEMENTATIONS (oc_usage, oc_usage, &oc_usage_field);

static struct oc_group *find_group (struct oc_properties *ocp,
				    struct obj_reference *ref,
				    struct ref_owner *p,
				    const char *fname, int line)
{
	int i;

	for (i = 0; i < XREF_A_SIZE (&ocp -> groups); i++)
		if (ocp -> groups.data [i] -> t.obj.ref.oid == ref -> oid &&
		    ocp -> groups.data [i] -> t.obj.ref.gnum == ref -> gnum) {
			oc_group_ref (ocp -> groups.data [i], p,
				      fname, line);
			return ocp -> groups.data [i];
		}

	return NULL;
}

static enum s_status lookup_dir (struct oc_directory **ocdp, 
				 struct xref_db *rdb, struct xref_bind *robj,
				 struct oc_properties *ocp, struct obj *obj,
				 struct ref_owner *op, const char *fname,
				 int line)
{
	struct oc_directory *ocd, **p;
	enum s_status res;

	for (ocd = ocp -> directory; ocp != NULL; ocd = ocd -> next) {
		if (ocd -> t.obj.ref.oid == obj -> ref.oid &&
		    ocd -> t.obj.ref.gnum == obj -> ref.gnum) {
			*ocdp = ocd;
			oc_directory_ref (ocd, op, fname, line);
			return S_CSC_OK;
		}
	}

	res = oc_directory_init (&ocd, rdb, robj, obj, op, fname, line);
	if (res != S_CSC_OK)
		return res;

	for (p = &ocp -> directory; *p != NULL; p = &(*p) -> next)
		;

	(*p) -> next = ocd;
	oc_directory_ref (ocd, op, fname, line);

	*ocdp = ocd;

	return S_CSC_OK;
}

static enum s_status group_list_entry_parser (struct xref_db *rdb, 
					      struct xref_bind *robj,
					      struct ref_owner *p,
					      struct obj *src,
					      struct xref_field *fd,
					      const char *fname, int line)
{
	struct oc_group_list_entry *e;
	struct oc_properties *ocp;
	struct oc_group *ocg;
	enum s_status res;

	e = fd -> field;

	ocp = xref_bind_get (robj, oc_properties_bkey);

	switch (src -> type) {
	case OT_STRING:
		e -> type = OC_GL_ET_TAG;
		res = ts_init (&e -> u.tag, rdb, src);
		if (res != S_CSC_OK)
			return res;
		break;

	case OT_DICTIONARY:
		ocg = find_group (ocp, &src -> ref, p, fname, line);
		if (ocg != NULL) {
			e -> type = OC_GL_ET_GROUP;
			e -> u.group = ocg;
		} else {
			e -> type = OC_GL_ET_UNRESOLVED;
			e -> u.ref = src -> ref;
		}

		break;

	case OT_ARRAY:
		e -> type = OC_GL_ET_GROUP_LIST;
		res = xref_array_map_ps (rdb, robj, p, src,
					 &e -> u.list, "OC group list",
					 &oc_group_list_entry_pstable, NULL);
		if (res != S_CSC_OK)
			return S_CSC_OUT_OF_MEMORY;

		break;

	default:
		return obj_error (src, 
				  "Unexpected OC config object type");
	}

	return S_CSC_OK;
}

static enum s_status group_list_entry_ser (struct xref_db *rdb, 
					   struct xref_bind *robj,
					   struct v_selection *sel,
					   struct obj *dest,
					   enum xref_changed *changed,
					   struct xref_field *fd)
{
	struct oc_group_list_entry *e;
	struct xref_field sfd;
	enum s_status res;

	e = fd -> field;

	sfd = *fd;
	switch (e -> type) {
	case OC_GL_ET_UNRESOLVED:
		break;

	case OC_GL_ET_TAG:
		sfd.field = &e -> u.tag;
		res = ts_ser (rdb, robj, sel, dest, changed, &sfd);
		if (res != S_CSC_OK)
			return res;
		break;

	case OC_GL_ET_GROUP:
		sfd.field = e -> u.group;
		res = oc_group_ser (rdb, robj, sel, dest, changed, &sfd);
		if (res != S_CSC_OK)
			return res;
		break;

	case OC_GL_ET_GROUP_LIST:
		res = xref_array_unmap_ps (rdb, robj, sel, &e -> u.list,
					   dest, 0, 
					   &oc_group_list_entry_pstable, NULL);
		if (res != S_CSC_OK)
			return res;
		break;
	}

	return S_CSC_OK;
}

static void group_list_entry_funref (struct ref_owner *p,
				     struct xref_field *fd)
{
	struct oc_group_list_entry *e;

	e = fd -> field;

	switch (e -> type) {
	case OC_GL_ET_UNRESOLVED:
		break;

	case OC_GL_ET_TAG:
		ts_destroy (&e -> u.tag);
		break;

	case OC_GL_ET_GROUP:
		oc_group_unref (e -> u.group, p);
		break;

	case OC_GL_ET_GROUP_LIST:
		xref_array_map_rel_ps (&e -> u.list, p,
				       &oc_group_list_entry_pstable, NULL);
		break;
	}
}

static enum s_status reference_parser (struct xref_db *rdb, 
				       struct xref_bind *robj,
				       struct ref_owner *p,
				       struct obj *src,
				       struct xref_field *fd,
				       const char *fname, int line)
{
	struct oc_reference *ocr;
	struct oc_properties *ocp;
	enum s_status res;
	struct obj type;

	ocr = fd -> field;
	ocp = xref_bind_get (robj, oc_properties_bkey);
	if (src -> type == OT_NULL) {
		ocr -> type = OC_RT_NONE;
		return S_CSC_OK;
	}

	res = xref_dict_resolve (rdb, src, "Type", OT_NAME, &type);
	if (res != S_CSC_OK)
		return res;

	if (obj_string_equal (&type.val.string, "OCG")) {
		ocr -> u.group = find_group (ocp, &src -> ref, p, fname, line);
		ocr -> type = OC_RT_GROUP;
	} else if (obj_string_equal (&type.val.string, "OCMD")) {
		res = lookup_dir (&ocr -> u.directory, rdb, robj, ocp, 
				  src, p, fname, line);
		if (res != S_CSC_OK) {
			obj_destroy (&type);
			return res;
		}
		ocr -> type = OC_RT_DIRECTORY;
	} else
		return obj_error (src, "Bad reference type");

	obj_destroy (&type);

	return S_CSC_OK;
}

static enum s_status reference_ser (struct xref_db *rdb, 
				    struct xref_bind *robj,
				    struct v_selection *sel,
				    struct obj *dest,
				    enum xref_changed *changed,
				    struct xref_field *fd)
{
	struct oc_reference *ocr;

	ocr = fd -> field;

	switch (ocr -> type) {
	case OC_RT_NONE:
	case OC_RT_UNRESOLVED:
		break;

	case OC_RT_GROUP:
		obj_copy_reference (dest, &ocr -> u.group -> t.obj);
		break;

	case OC_RT_DIRECTORY:
		obj_copy_reference (dest, &ocr -> u.directory -> t.obj);
	}

	return S_CSC_OK;
}

static void reference_funref (struct ref_owner *p,
			      struct xref_field *fd)
{
	struct oc_reference *ocr;

	ocr = fd -> field;

	switch (ocr -> type) {
	case OC_RT_NONE:
	case OC_RT_UNRESOLVED:
		break;

	case OC_RT_GROUP:
		oc_group_unref (ocr -> u.group, p);
		break;

	case OC_RT_DIRECTORY:
		oc_directory_unref (ocr -> u.directory, p);
	}
}

static enum s_status parse_directory_op (struct obj *obj,
					 enum oc_directory_op *op)
{
	int i;

	if (obj -> type != OT_NAME)
		return obj_error (obj, "VE operator is not a name");

	for (i = 0; op_desc [i].name != NULL; i++)
		if (obj_string_equal (&obj -> val.string, op_desc [i].name)) {
			*op = op_desc [i].value;
			return S_CSC_OK;
		}

	return obj_error (obj, "Unknown VE operator");
}

static enum s_status ser_directory_op (enum oc_directory_op op,
				       struct obj *obj)
{
	struct obj_bstr str;
	enum s_status res;
	int i;

	for (i = 0; op_desc [i].name != NULL; i++) {
		if (op_desc [i].value == op) {
			res = obj_string_from_c (op_desc [i].name, &str);
			if (res != S_CSC_OK)
				return res;

			obj_set_string (obj, &f_null, &str);
		}
	}

	return s_make_indirect (S_CSC_INTERNAL_ERROR, "Bad operator");
}

static enum s_status do_parse_ve (struct xref_db *rdb,
				  struct oc_properties *ocp,
				  struct ref_owner *p,
				  struct obj *refobj,
				  struct xref_obj_array *ma,
				  struct oc_ve *ve,
				  const char *fname, int line)
{
	enum oc_directory_op op;
	enum s_status res;
	int i;

	i = 0;
	if (XREF_A_SIZE (ma) < i)
		return obj_error (refobj, "VE array is too short");

	ve -> entry = calloc (XREF_A_SIZE (ma), sizeof (struct oc_ve));
	if (ve -> entry == NULL)
		return S_CSC_OUT_OF_MEMORY;

	res = parse_directory_op (&ma -> data [i], &op);
	if (res != S_CSC_OK)
		return res;

	if (op == OC_DO_NOT) {
		ve -> negated = 1;

		i++;
		if (XREF_A_SIZE (ma) < i)
			return obj_error (refobj, "VE array is too short");

		if (res != S_CSC_OK)
			return res;

		res = parse_directory_op (&ma -> data [i], &op);
		if (res != S_CSC_OK)
			return res;

		if (op == OC_DO_NOT)
			return obj_error (refobj, "Bad VE array");
	}
	
	ve -> op = op;
	ve -> entries = XREF_A_SIZE (ma) - i;

	while (i < XREF_A_SIZE (ma)) {
		if (ma -> data [i].type == OT_ARRAY) {
			ve -> entry [i].type = OC_VT_VE;
			ve -> entry [i].u.ve = malloc (sizeof (struct oc_ve));
			if (ve -> entry [i].u.ve == NULL)
				return S_CSC_OUT_OF_MEMORY;
			res = parse_ve (rdb, ocp, p, &ma -> data [i], 
					&ve -> entry [i].u.ve,
					fname, line);
			if (res != S_CSC_OK)
				return res;
		} else {
			ve -> entry [i].u.group = 
				find_group (ocp, &ma -> data [i].ref, p,
					    fname, line);
			ve -> entry [i].type = 
				ve -> entry [i].u.group != NULL ?
				OC_VT_GROUP : OC_VT_UNRESOLVED;
		} 
		i++;
	}

	return S_CSC_OK;
}

static enum s_status parse_ve (struct xref_db *rdb, 
			       struct oc_properties *ocp,
			       struct ref_owner *p,
			       struct obj *src,
			       struct oc_ve **vep,
			       const char *fname, int line)
{
	struct xref_obj_array ma;
	struct oc_ve *ve;
	enum s_status res;

	ve = malloc (sizeof (struct oc_ve));
	if (ve == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (ve, 0, sizeof (struct oc_ve));

	res = xref_array_map_ps (rdb, NULL, NULL, src, &ma, "VE expression",
				 &xref_obj_pstable, NULL);
	if (res != S_CSC_OK) {
		free (ve);
		return res;
	}

	res = do_parse_ve (rdb, ocp, p, src, &ma, ve, fname, line);
	xref_array_map_rel_ps (&ma, p, &xref_obj_pstable, NULL);
	if (res != S_CSC_OK) {
		release_ve (p, ve);
		return res;
	}

	*vep = ve;

	return S_CSC_OK;
}

static enum s_status ve_parser (struct xref_db *rdb, 
				struct xref_bind *robj,
				struct ref_owner *p,
				struct obj *src,
				struct xref_field *fd,
				const char *fname, int line)
{
	struct oc_properties *ocp;
	enum s_status res;
	struct oc_ve *ve;

	ocp = xref_bind_get (robj, oc_properties_bkey);
	res = parse_ve (rdb, ocp, p, src, &ve, fname, line);
	if (res != S_CSC_OK)
		return res;

	fd -> field = ve;

	return S_CSC_OK;
}

static enum s_status do_serialize_ve (struct xref_db *rdb, 
				      struct xref_bind *robj,
				      struct v_selection *sel,
				      struct oc_ve *ve,
				      struct xref_obj_array *ma)
{
	struct oc_ve_entry *ocve;
	enum s_status res;
	int i, j;
	
	i = 0;
	if (ve -> negated) {
		res = ser_directory_op (OC_DO_NOT, &ma -> data [i]);
		if (res != S_CSC_OK)
			return res;
		i++;
	}

	res = ser_directory_op (ve -> op, &ma -> data [i]);
	if (res != S_CSC_OK)
		return res;
	i++;

	for (j = 0; j < ve -> entries; j++) {
		ocve = &ve -> entry [j];
		switch (ocve -> type) {
		case OC_VT_UNRESOLVED:
			i--;
			break;

		case OC_VT_GROUP:
			obj_copy_reference (&ma -> data [j], 
					    &ocve -> u.group -> t.obj);
			break;

		case OC_VT_VE:
			res = serialize_ve (rdb, robj, sel, ocve -> u.ve,
					    &ma -> data [j]);
			if (res != S_CSC_OK)
				return res;
		}
		i++;
	}

	return S_CSC_OK;
}

static enum s_status serialize_ve (struct xref_db *rdb, 
				   struct xref_bind *robj,
				   struct v_selection *sel,
				   struct oc_ve *ve, struct obj *dest)
{
	struct xref_obj_array ma;
	enum s_status res;

	ma.n = ve -> entries - 1;
	if (ve -> negated)
		ma.n--;

	ma.data = calloc (ma.n, sizeof (struct obj));
	if (ma.data == NULL)
		return S_CSC_OUT_OF_MEMORY;

	res = do_serialize_ve (rdb, robj, sel, ve, &ma);
	if (res != S_CSC_OK) {
		xref_array_map_rel_ps (&ma, NULL, &xref_obj_pstable, NULL);
		return res;
	}

	res = xref_array_unmap_ps (rdb, robj, sel, &ma, dest, 0, 
				   &xref_obj_pstable, NULL);
	xref_array_map_rel_ps (&ma, NULL, &xref_obj_pstable, NULL);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status ve_ser (struct xref_db *rdb, 
			     struct xref_bind *robj,
			     struct v_selection *sel,
			     struct obj *dest,
			     enum xref_changed *changed,
			     struct xref_field *fd)
{
	return serialize_ve (rdb, robj, sel, fd -> field, dest);
}

static void release_ve (struct ref_owner *p, struct oc_ve *ve)
{
	struct oc_ve_entry *ocve;
	int i;
	
	if (ve -> entry != NULL) {
		for (i = 0; i < ve -> entries; i++) {
			ocve = &ve -> entry [i];
			switch (ocve -> type) {
			case OC_VT_UNRESOLVED:
			case OC_VT_GROUP:
				break;

			case OC_VT_VE:
				release_ve (p, ocve -> u.ve);
			}
		}
		free (ve -> entry);
	}

	free (ve);
}

static void ve_funref (struct ref_owner *p,
		       struct xref_field *fd)
{
	release_ve (p, fd -> field);
}

