/*
 * PDFlow -- page.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>
#include <stdlib.h>

#include "fmt/fields.h"
#include "types/page.h"
#include "types/cobj.h"

#define DEFAULT_NODE_SIZE 40

#define XREF_IMPLEMENTATION struct pg_inheritable
#define PG_INHERITABLE
#include "types/page.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct pg_node
#define PG_NODE
#include "types/page.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct pg_page
#define PG_PAGE
#include "types/page.tab.h"
#undef XREF_IMPLEMENTATION

static void node_fini (struct pg_node *node, struct ref_owner *p);
static void page_fini (struct pg_page *page, struct ref_owner *p);

T_IMPLEMENTATIONS_LOC (pg_node, pg_node, &node_field, NULL, NULL, node_fini)
T_IMPLEMENTATIONS_LOC (pg_page, pg_page, &page_field, NULL, NULL, page_fini)
T_SUGAR_IMPLEMENTATIONS (pg_pages, pg_pages)


static int t_compare (bt_key_t *key1, bt_key_t *key2);

static void t_destroy_node_data (bt_tree_data_t *tdata, 
				 bt_node_data_t *ndata);

static enum s_status t_create_node_data (bt_tree_data_t *tdata, 
					 bt_node_data_t *parent,
					 bt_node_data_t **ndata);

static void t_destroy_data (bt_tree_data_t *tdata, bt_data_t *data);

static void pages_add_delnode (struct pg_pages *pages, struct pg_node *node);

static struct bt_param tree_param = {
	sizeof (pdf_integer_t), 
	DEFAULT_NODE_SIZE,
	NULL,
	t_compare,
	NULL,
	t_create_node_data,
	t_destroy_node_data,
	t_destroy_data
};

static struct obj *cast_imp_closure (void *closure)
{
	return closure;
}

static void *uncast_imp_closure (struct obj *obj)
{
	return obj;
}

static struct pg_pages *cast_tree_data (bt_tree_data_t *tdata)
{
	return (struct pg_pages *) tdata;
}

static bt_tree_data_t *uncast_tree_data (struct pg_pages *p)
{
	return (struct bt_imp_data_t *) p;
}

static struct pg_node *cast_node_data (bt_node_data_t *ndata)
{
	return (struct pg_node *) ndata;
}

static bt_node_data_t *uncast_node_data (struct pg_node *node)
{
	return (bt_node_data_t *) node;
}

static struct pg_page *cast_data (bt_data_t *data)
{
	return (struct pg_page *) data;
}

static bt_data_t *uncast_data (struct pg_page *page)
{
	return (bt_data_t *) page;
}

static pdf_integer_t *cast_key (bt_key_t *key)
{
	return (pdf_integer_t *) key;
}

static int t_compare (bt_key_t *key1, bt_key_t *key2)
{
	return	*cast_key (key1) - *cast_key (key2);
}

static void init_inherit (struct pg_inheritable *inh, 
			  struct pg_inheritable *sinh, 
			  struct pg_node *parent,
			  struct t_tag *tag,
			  struct ref_owner *p)
{
	struct pg_inheritable *inhp;

	if (sinh -> resources != NULL && 
	    !t_tag_is_valid (&sinh -> resources -> t.tag))
		sinh -> resources -> t.tag = *tag;

	inhp = parent != NULL ? &parent -> inh : NULL;
	*inh = *sinh;

	if (inhp != NULL) {
		if (inh -> resources == NULL &&
		    inhp -> resources != NULL)
			inh -> resources = inhp -> resources;

		if (cobj_rect_is_null (&inh -> media_box))
			inh -> media_box = inhp -> media_box;

		if (cobj_rect_is_null (&inh -> crop_box))
			inh -> crop_box = inhp -> crop_box;

		if (inh -> rotate.type == OT_NULL)
			inh -> rotate = inhp -> rotate;
	}

	if (inh -> resources != NULL)
		T_CALL (rd_ref, inh -> resources, p);
}

static void push_inh (struct pg_inheritable *inh, 
		      struct pg_inheritable *sinh,
		      struct ref_owner *p,
		      struct pg_inheritable **child, 
		      struct pg_inheritable **schild, 
		      struct ref_owner **cp,
		      unsigned n)
{
	pdf_bool_t resdict, media_box, crop_box, rotate;
	struct pg_inheritable cinh;
	int i;

	if (!n)
		return;

	for (i = 0; i < n; i++) {
		if (schild [i] -> resources != NULL)
			rd_unref (schild [i] -> resources, cp [i]);
		*schild [i] = *child [i];
		if (schild [i] -> resources != NULL)
			T_CALL (rd_ref, schild [i] -> resources, cp [i]);
	}

	cinh = *schild [0];

	resdict = 1;
	media_box = 1;
	crop_box = 1;
	rotate = 1;

	for (i = 1; i < n; i++) {
		resdict &= 
			cinh.resources == NULL ||
			schild [i] -> resources == NULL ||
			cinh.resources == schild [i] -> resources;
		if (resdict && cinh.resources == NULL)
			cinh.resources = schild [i] -> resources;
		media_box &= !memcmp (&cinh.media_box,
				      &schild [i] -> media_box,
				      sizeof (struct cobj_rect));
		crop_box &= !memcmp (&cinh.crop_box,
				     &schild [i] -> crop_box,
				     sizeof (struct cobj_rect));
		rotate &= obj_equals (&cinh.rotate, &schild [i] -> rotate);
	}

	if (inh -> resources != NULL)
		rd_unref (inh -> resources, p);
	*inh = *sinh = cinh;
	if (cinh.resources != NULL) {
		T_CALL (rd_ref, inh -> resources, p);
		T_CALL (rd_ref, sinh -> resources, p);
	}

	for (i = 0; i < n; i++) {
		if (resdict) {
			if (schild [i] -> resources != NULL)
				rd_unref (schild [i] -> resources, cp [i]);
			schild [i] -> resources = NULL;
		}
		if (media_box)
			cobj_rect_set_null (&schild [i] -> media_box);
		if (crop_box)
			cobj_rect_set_null (&schild [i] -> crop_box);
		if (rotate)
			obj_set_null (&schild [i] -> rotate, &f_null);
	}
}

static void destroy_inherit (struct pg_inheritable *inh, struct ref_owner *p)
{
	xref_fields_destroy (&inh_field, inh, p);
}

static enum s_status pg_node_init_p (struct pg_node **nodep, 
				     struct xref_db *rdb, struct obj *obj,
				     struct pg_node *parent, 
				     struct ref_owner *p)
{
	struct pg_node *node;
	enum s_status res;

	res = T_CALL (pg_node_init, &node, rdb, NULL, obj, p);
	if (res != S_CSC_OK)
		return res;

	init_inherit (&node -> inh, &node -> sinh, parent, 
		      &node -> t.tag, ref_rown (&node -> t.refs));

	*nodep = node;

	return S_CSC_OK;
}

static void node_fini (struct pg_node *node, struct ref_owner *p)
{
	destroy_inherit (&node -> inh, ref_rown (&node -> t.refs));
}

static void page_fini (struct pg_page *page, struct ref_owner *p)
{
	destroy_inherit (&page -> inh, ref_rown (&page -> t.refs));
}

static enum s_status t_create_node_data (bt_tree_data_t *tdata, 
					 bt_node_data_t *parent,
					 bt_node_data_t **ndata)
{
	struct pg_pages *pages;
	struct pg_node *node, *pnode;
	enum s_status res;

	pages = cast_tree_data (tdata);
	pnode = cast_node_data (parent);
	
	res = pg_node_init_p (&node, pages -> t.rdb, NULL, pnode,
			      ref_rown (&pages -> t.refs));
	if (res != S_CSC_OK)
		return res;

	*ndata = uncast_node_data (node);

	return S_CSC_OK;
}

static void t_destroy_node_data (bt_tree_data_t *tdata, 
				 bt_node_data_t *ndata)
{
	struct pg_pages *pages;
	struct pg_node *node;

	pages = cast_tree_data (tdata);

	if (ndata != NULL) {
		node = cast_node_data (ndata);
		if (obj_exists (&node -> t.obj)) 
			pages_add_delnode (pages, node);
		else
			pg_node_unref (node, ref_rown (&pages -> t. refs));
	}
}

static void t_destroy_data (bt_tree_data_t *tdata, bt_data_t *data)
{
	struct pg_pages *pages;

	pages = cast_tree_data (tdata);

	if (data != NULL)
		pg_page_unref (cast_data (data),
			       ref_rown (&pages -> t.refs));
}

static enum s_status pg_page_init_p (struct pg_page **pagep, 
				     struct xref_db *rdb, struct obj *obj,
				     struct pg_node *parent, 
				     struct ref_owner *p)
{
	struct pg_page *page;
	enum s_status res;

	res = T_CALL (pg_page_init, &page, rdb, NULL, obj, p);
	if (res != S_CSC_OK)
		return res;

	init_inherit (&page -> inh, &page -> sinh, parent,
		      &page -> t.tag, ref_rown (&page -> t.refs));

	*pagep = page;

	return S_CSC_OK;
}

static enum s_status build_node_kids (struct xref_obj_array *array, 
				      struct pg_node **kids,
				      unsigned count,
				      pdf_integer_t *tkids)
{
	int i;

	array -> n = count;
	array -> data = calloc (count, sizeof (struct obj));
	if (array -> data == NULL)
		return S_CSC_OUT_OF_MEMORY;


	*tkids = 0;
	for (i = 0; i < count; i++) {
		*tkids += kids [i] -> count;
		obj_copy_reference (&array -> data [i], &kids [i] -> t.obj);
	}

	return S_CSC_OK;
}

static enum s_status build_leaf_kids (struct xref_obj_array *array, 
				      struct pg_page **kids,
				      unsigned count,
				      pdf_integer_t *tkids)
{
	int i;

	array -> n = count;
	array -> data = calloc (count, sizeof (struct obj));
	if (array -> data == NULL)
		return S_CSC_OUT_OF_MEMORY;


	for (i = 0; i < count; i++)
		obj_copy_reference (&array -> data [i], &kids [i] -> t.obj);

	*tkids = count;

	return S_CSC_OK;
}

static enum s_status pages_export_f (bt_tree_data_t *tdata,
				     struct bt_exp_data *exps,
				     va_list ap)
{
	struct xref_db *rdb;
	struct v_selection *sel;
	enum s_status res;
	struct pg_node *node, *pnode;
	struct xref_obj_array karray;
	struct pg_page *page;
	struct obj *dest;
	enum xref_changed *changed;
	int i;

	rdb = va_arg (ap, struct xref_db *);
	sel = va_arg (ap, struct v_selection *);

	node = cast_node_data (exps -> ndata);
	pnode = cast_node_data (exps -> parent);
	
	obj_destroy (&node -> parent);
	if (pnode != NULL) {
		dest = NULL;
		changed = NULL;
		obj_copy_reference (&node -> parent, &pnode -> t.obj);
	} else {
		dest = va_arg (ap, struct obj *);
		changed = va_arg (ap, enum xref_changed *);
		obj_set_null (&node -> parent, &f_null);
	}
	
	if (exps -> type == BT_NT_LEAF) {
		for (i = 0; i < exps -> size; i++) {
			page = cast_data (exps -> data [i]);
			obj_copy_reference (&page -> parent, &node -> t.obj);
			res = pg_page_sync (rdb, NULL, sel, NULL, NULL, page);
			if (res != S_CSC_OK)
				return res;
		}
	}

	if (exps -> type == BT_NT_LEAF)
		res = build_leaf_kids (&karray,
				       (struct pg_page **) exps -> data,
				       exps -> size, &node -> count);
	else
		res = build_node_kids (&karray, 
				       (struct pg_node **) exps -> child,
				       exps -> size, &node -> count);

	if (res != S_CSC_OK)
		return res;

	xref_array_map_rel_ps (&node -> kids, ref_rown (&node -> t.refs),
			       &xref_dictionary_pstable, NULL);
	*(struct xref_obj_array *) &node -> kids = karray;

	res = pg_node_sync (rdb, NULL, sel, dest, changed, node);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status pages_push_inh_f (bt_tree_data_t *tdata,
				       struct bt_exp_data *exps,
				       va_list ap)
{
	struct pg_inheritable *inhp [exps -> size], *sinhp [exps -> size];
	struct ref_owner *cp [exps -> size];
	struct pg_node *node, *nchild;
	struct pg_page *pchild;
	int i;

	node = cast_node_data (exps -> ndata);
	if (exps -> type == BT_NT_LEAF)
		for (i = 0; i < exps -> size; i++) {
			pchild = cast_data (exps -> data [i]);
			inhp [i] = &pchild -> inh;
			sinhp [i] = &pchild -> sinh;
			cp [i] = ref_rown (&pchild -> t.refs);
		}
	else
		for (i = 0; i < exps -> size; i++) {
			nchild = cast_node_data (exps -> child [i]);
			inhp [i] = &nchild -> inh;
			sinhp [i] = &nchild -> sinh;
			cp [i] = ref_rown (&nchild -> t.refs);
		}
	
	push_inh (&node -> inh, &node -> sinh, 
		  ref_rown (&node -> t.refs), 
		  inhp, sinhp, cp, exps -> size);

	return S_CSC_OK;
}

static enum s_status sync_nodes_f (bt_node_data_t *ndata, 
				   bt_data_t *data, 
				   va_list ap)
{
	struct xref_db *rdb;
	enum s_status res;
	struct obj *obj;

	rdb = va_arg (ap, struct xref_db *);

	obj = data != NULL ? 
		&cast_data (data) -> t.obj :
		&cast_node_data (ndata) -> t.obj;

	if (!obj_exists (obj)) {
		res = xref_register_object (rdb, obj);
		if (res != S_CSC_OK)
			return res;
	}
		
	return S_CSC_OK;
}

static enum s_status sync_nodes (struct pg_pages *pages)
{
	struct pg_node *node;
	enum s_status res;

	res = bt_traverse (pages -> tree, sync_nodes_f, pages -> t.rdb);
	if (res != S_CSC_OK)
		return res;

	while (pages -> delnodes != NULL) {
		node = pages -> delnodes;
		pages -> delnodes = node -> next;
		xref_remove_object (pages -> t.rdb, &node -> t.obj);
		pg_node_unref (node, ref_rown (&pages -> t.refs));
	}

	return S_CSC_OK;
}

enum s_status pg_pages_sync (struct xref_db *rdb, struct xref_bind *robj,
			     struct v_selection *sel, struct obj *dest, 
			     enum xref_changed *changed, 
			     struct pg_pages *pages)
{
	struct bt_exp_info ei;
	enum s_status res;

	res = sync_nodes (pages);
	if (res != S_CSC_OK)
		return res;

	ei.export = pages_push_inh_f;
	res = bt_export (pages -> tree, &ei);
	if (res != S_CSC_OK)
		return res;

	ei.export = pages_export_f;
	res = bt_export (pages -> tree, &ei, rdb, sel, dest, changed);
	if (res != S_CSC_OK)
		return res;

	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static void pages_add_delnode (struct pg_pages *pages, struct pg_node *node)
{
	node -> next = pages -> delnodes;
	pages -> delnodes = node;
}

static void page_ins_rekey_f (bt_node_data_t *data, bt_key_t *key, 
			      int seqno, void *closure)
{
	if (seqno)
		(*cast_key (key))++;
}

enum s_status pg_page_insert (struct pg_page **pagep, 
			      struct pg_pages *pages,
			      struct bt_cursor *pc)
{
	pdf_integer_t *lpc;
	struct pg_node *parent;
	struct pg_page *page;
	enum s_status res;
	int count;

	page = pg_get (pc);
	parent = bt_get_node_data (pc);

	res = pg_page_init_p (&page, pages -> t.rdb, NULL,
			      parent, ref_rown (&pages -> t.refs));
	if (res != S_CSC_OK)
		return res;

	lpc = cast_key (bt_cursor_get_key (pc));
	count = lpc != NULL ? *lpc : 0;
	count++;
	bt_rekey_from (pc -> tree, pc, page_ins_rekey_f, NULL);
	
	res = bt_insert (pc -> tree, &count, uncast_data (page), pc);
	if (res != S_CSC_OK) {
		pg_page_unref (page, ref_rown (&pages -> t.refs));
		return res;
	}

	*pagep = page;

	return S_CSC_OK;
}

struct pg_page *pg_get (struct bt_cursor *pc)
{
	return cast_data (bt_cursor_get (pc));
}

/* -- Pages -- */
static enum s_status is_leaf (struct xref_db *rdb, struct obj *obj, 
			      pdf_bool_t *leaf)
{
	return xref_dict_is (rdb, obj, "Page", leaf);
}

static enum s_status import_leaf (struct xref_db *rdb,
				  struct obj *obj,
				  size_t size,
				  struct pg_node *parent,
				  struct bt_imp_data *imps,
				  struct pg_pages *pages)
{
	struct pg_page **page;
	struct ref_owner *p;
	enum s_status res;
	int i;

	p = ref_rown (&pages -> t.refs);

	page = calloc (size, sizeof (struct pg_page *));
	if (page == NULL)
		return S_CSC_OUT_OF_MEMORY;

	res = S_CSC_OK;
	for (i = 0; i < size; i++) {
		res = pg_page_init_p (&page [i], rdb, &obj [i], parent, p);
		if (res != S_CSC_OK)
			break;
	}

	if (res != S_CSC_OK) {
		while (--i >= 0)
			pg_page_unref (page [i], p);
		free (page);
		return res;
	}
	
	imps -> type = BT_NT_LEAF;
	imps -> size = size;
	imps -> ndata = NULL;
	imps -> data = (bt_data_t **) page;
	imps -> key = 0;

	return S_CSC_OK;
}

static enum s_status import_node (struct xref_db *rdb,
				  struct obj *obj,
				  struct pg_node *parent,
				  struct bt_imp_data *imps,
				  struct pg_pages *pages)
{
	struct pg_node *node;
	enum s_status res;
	int i, leaf;

	res = pg_node_init_p (&node, rdb, obj, parent, 
			      ref_rown (&pages -> t.refs));
	if (res != S_CSC_OK)
		return res;

	imps -> type = BT_NT_INTERMEDIATE;
	imps -> size = XREF_A_SIZE (&node -> kids);
	imps -> ndata = uncast_node_data (node);
	imps -> data = NULL;
	imps -> key = 0;

	/* This is an optimization, so we can afford skipping it on error */
	for (i = 0; i < imps -> size; i++) {
		if (is_leaf (rdb, &node -> kids.data [i], &leaf) != S_CSC_OK)
			break;

		if (!leaf)
			break;
	}
	if (i == imps -> size) {
		res = import_leaf (rdb, node -> kids.data, imps -> size,
				   node, imps, pages);
		pg_node_unref (node, ref_rown (&pages -> t.refs));
		if (res != S_CSC_OK)
			return res;
	}

	return S_CSC_OK;
}

static void page_rekey_f (bt_node_data_t *data, bt_key_t *key, 
			  int seqno, void *closure)
{
	*cast_key (key) = seqno;
}

static enum s_status pages_import_f (bt_tree_data_t *tdata,
				     bt_node_data_t *ndata,
				     unsigned idx,
				     struct bt_imp_data *imps,
				     void *closure)
{
	struct pg_node *parent;
	struct pg_pages *pages;
	struct obj *obj, *root;
	enum s_status res;
	int leaf;

	root = cast_imp_closure (closure);
	parent = cast_node_data (ndata);
	pages = cast_tree_data (tdata);

	if (parent != NULL)
		obj = &parent -> kids.data [idx];
	else
		obj = root;
		
	if (obj -> type != OT_NULL) {
		res = is_leaf (pages -> t.rdb, obj, &leaf);
		if (res != S_CSC_OK)
			return res;
	} else
		leaf = 1;

	res = leaf ?
		import_leaf (pages -> t.rdb, obj, 1, parent, imps, pages) :
		import_node (pages -> t.rdb, obj, parent, imps, pages);

	return res;
}

static enum s_status init_pages (struct pg_pages *pages, struct xref_db *rdb,
				 struct obj *obj)
{
	struct bt_imp_info ii;
	struct bt_param param;
	struct ref_owner *p;
	enum s_status res;

	p = ref_rown (&pages -> t.refs);

	param = tree_param;

	param.tdata = uncast_tree_data (pages);
	res = bt_init (&pages -> tree, &param, p);
	if (res != S_CSC_OK)
		return res;

	if (obj -> type != OT_NULL) {
		ii.import = pages_import_f;

		res = bt_import (pages -> tree, &ii, 
				 uncast_imp_closure (obj));
		if (res != S_CSC_OK)
			return res;

		bt_rekey (pages -> tree, page_rekey_f, NULL);
	}

	return S_CSC_OK;
}

enum s_status pg_pages_init (struct pg_pages **pagesp, struct xref_db *rdb, 
			     struct xref_bind *robj, const struct obj *obj, 
			     struct ref_owner *p, const char *fname, 
			     int line)
{
	struct pg_pages *pages;
	enum s_status res;

	if (obj != NULL && obj -> type == OT_NULL) {
		*pagesp = NULL;
		return S_CSC_OK;
	}

	if (obj == NULL)
		obj = &obj_null;

	if (obj -> type != OT_DICTIONARY &&
	    obj -> type != OT_NULL)
		return obj_error (obj, "Bad page object type");

	pages = malloc (sizeof (struct pg_pages));
	if (pages == NULL)
		return S_CSC_OK;
	memset (pages, 0, sizeof (struct pg_pages));

	ref_init (&pages -> t.refs, "pages", p);
	pages -> t.rdb = rdb;
	xref_ref_db (rdb, ref_rown (&pages -> t.refs));

	res = obj_dup (obj, &pages -> t.obj);
	if (res != S_CSC_OK) {
		pg_pages_unref (pages, p);
		return res;
	}

	res = init_pages (pages, rdb, &pages -> t.obj);
	if (res != S_CSC_OK) {
		pg_pages_unref (pages, p);
		return res;
	}

	*pagesp = pages;

	return S_CSC_OK;
}

void pg_pages_unref (struct pg_pages *pages, struct ref_owner *p)
{
	struct pg_node *node;

	if (pages == NULL || ref_down (&pages -> t.refs, p))
		return;

	if (pages -> tree != NULL)
		bt_unref (pages -> tree, ref_rown (&pages -> t.refs));

	while (pages -> delnodes != NULL) {
		node = pages -> delnodes;
		pages -> delnodes = node -> next;
		pg_node_unref (node, ref_rown (&pages -> t.refs));
	}

	xref_unref_db (pages -> t.rdb, ref_rown (&pages -> t.refs));

	obj_destroy (&pages -> t.obj);

	ref_zero (&pages -> t.refs);

	free (pages);
}

enum s_status pg_page_get_contents (struct pg_pages *pages,
				    struct pg_page *page,
				    struct pg_contents *pgc)
{
	enum s_status res;
	int i;

	pgc -> elements = XREF_A_SIZE (&page -> contents);
	pgc -> file = calloc (pgc -> elements, sizeof (struct f_file *));
	if (pgc -> file == NULL)
		return S_CSC_OUT_OF_MEMORY;

	res = S_CSC_OK;
	for (i = 0; i < pgc -> elements; i++) {
		res = xref_resolve_stream (page -> t.rdb, 
					   &page -> contents.data [i]);
		if (res != S_CSC_OK)
			break;

		res = obj_stream_file (&page -> contents.data [i], 
				       &pgc -> file [i]);
		if (res != S_CSC_OK)
			break;
	}

	if (res != S_CSC_OK) {
		pg_page_destroy_contents (pgc);
		return res;
	}

	return S_CSC_OK;
}

void pg_page_destroy_contents (struct pg_contents *pgc)
{
	int i;

	for (i = 0; i < pgc -> elements; i++) {
		if (pgc -> file [i] != NULL)
			f_unref (pgc -> file [i]);
	}

	free (pgc -> file);
}

void pg_page_set_resdict (struct pg_page *page,
			  struct rd_resdict *rd)
{
	rd_unref (page -> inh.resources, ref_rown (&page -> t.refs));
	page -> inh.resources = rd;
	T_CALL (rd_ref, page -> inh.resources, ref_rown (&page -> t.refs));
}

enum s_status pg_page_xfer_contents (struct pg_page *page,
				     struct obj_region *reg,
				     struct xref_bind *robj,
				     struct v_selection *sel)
{
	struct xref_obj_array ma;
	struct xref_stream xs;
	enum s_status res;

	ma.n = 1;
	ma.data = calloc (1, sizeof (struct obj));
	if (ma.data == NULL)
		return S_CSC_OUT_OF_MEMORY;

	memset (&xs, 0, sizeof (struct xref_stream));

	xs.decoded = *reg;
	memset (&xs.encoded, 0, sizeof (struct obj_region));

	res = xref_stream_ser (page -> t.rdb, robj, sel, ma.data, &xs);
	if (res != S_CSC_OK) {
		xref_array_map_rel_ps (&ma, ref_rown (&page -> t.refs),
				       &xref_obj_pstable, 0);
		return res;
	}

	xref_array_map_rel_ps (&page -> contents, ref_rown (&page -> t.refs),
			       &xref_obj_pstable, 0);

	page -> contents = ma;

	return S_CSC_OK;
}

enum s_status pg_page_store_contents (struct pg_page *page,
				      struct obj_region *reg,
				      struct xref_bind *robj,
				      struct v_selection *sel)
{
	enum s_status res;
	int i;

	res = pg_page_xfer_contents (page, reg, robj, sel);
	if (res != S_CSC_OK)
		return res;

	for (i = 0; i < XREF_A_SIZE (&page -> contents); i++) {
		res = xref_serialize_obj (page -> t.rdb, 
					  &page -> contents.data [i]);
		if (res != S_CSC_OK)
			return res;

		obj_unload (&page -> contents.data [i]);
	}

	return S_CSC_OK;
}

pdf_integer_t pg_get_rotate (struct pg_page *page)
{
	return page -> inh.rotate.type == OT_INTEGER ?
		page -> inh.rotate.val.integer : 0;
}
