/*
 * PDFlow -- resdict.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>
#include <stdlib.h>

#include "version.h"
#include "types/resdict.h"
#include "types/colour.h"
#include "types/pattern.h"
#include "types/xobj.h"

#define XREF_IMPLEMENTATION struct rd_resdict
#include "types/resdict.tab.h"

const char *rd_bkey = "rd_resdict bind key";

T_IMPLEMENTATIONS (rd, rd_resdict, &fields);

static void init_robj (struct xref_bind *robj, struct rd_resdict *rd,
		       struct oc_properties *ocp)
{
	xref_bind_init (robj);
	xref_bind_data (robj, rd_bkey, rd);
	xref_bind_data (robj, oc_properties_bkey, ocp);
}

enum s_status rd_lookup (struct rd_resdict *rd, struct obj *name,
			 struct obj *coll, struct obj *val, 
			 struct t_tag *tag, int *pres)
{
	enum s_status res;
	char *key;

	res = obj_string_to_c (&name -> val.string, &key);
	if (res != S_CSC_OK)
		return res;

	res = rd_lookup_s (rd, key, coll, val, tag, pres);
	free (key);

	if (res != S_CSC_OK)
		return res;
				 
	return S_CSC_OK;
}

enum s_status rd_lookup_s (struct rd_resdict *rd, const char *name,
			   struct obj *coll, struct obj *val, 
			   struct t_tag *tag, int *pres)
{
	enum s_status res;

	if (rd != NULL) {
		res = xref_dict_resolve (rd -> t.rdb, coll, name, OT_ANY, val);
		if (res != S_CSC_OK)
			return res;
	} else
		obj_set_null (val, &f_null);

	if (pres != NULL)
		*pres = val -> type != OT_NULL;
	else if (val -> type == OT_NULL)
		return obj_error (coll, "Missing resdict key: %s", name);

	if (tag != NULL) {
		if (!obj_exists (val)) {
			res = obj_string_from_c (name, &tag -> name);
			if (res != S_CSC_OK) {
				obj_destroy (val);
				return res;
			}
			tag -> ref = rd -> t.tag.ref;
			tag -> rid = rd -> t.tag.rid;
		} else {
			memset (&tag -> name, 0, sizeof (struct obj_bstr));
			tag -> ref = val -> ref;
			tag -> rid = rd -> t.tag.rid;
		}
	}


	return S_CSC_OK;
}

enum s_status rd_get_ext_g_state (struct rd_resdict *rd, struct obj *name,
				  struct obj *dict)
{
	enum s_status res;
	int pres;

	if (rd == NULL || rd -> ext_g_state.type == OT_NULL)
		return obj_error (name, "No ExtGState dictionary");

	pres = 1;
	res = rd_lookup (rd, name, &rd -> ext_g_state, dict, NULL, &pres);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

enum s_status rd_get_cspace (struct rd_resdict *rd, struct xref_db *rdb,
			     struct obj *name,
			     struct gc_colour_space **space, 
			     struct ref_owner *p)
{
	struct xref_bind robj;
	struct t_tag tag;
	enum s_status res;

	init_robj (&robj, rd, NULL);

	if (rd != NULL && name -> type == OT_NAME) {
		tag.rid = rd -> t.tag.rid;
		tag.ref = rd -> t.tag.ref;
		res = obj_string_dup (&name -> val.string, &tag.name);
		if (res != S_CSC_OK)
			return res;
	} else {
		tag.rid = xref_get_id (rdb);
		tag.ref = name -> ref;
		memset (&tag.name, 0, sizeof (struct obj_bstr));
	}

	res = T_CALL (gc_cspace_init, space, rdb, &robj, name, p);
	if (res != S_CSC_OK) {
		obj_string_destroy (&tag.name);
		return res;
	}

	if (t_tag_is_valid (&(*space) -> t.tag)) {
		obj_string_destroy (&tag.name);
		return S_CSC_OK;
	}
	
	(*space) -> t.tag = tag;

	return S_CSC_OK;
}

enum s_status rd_get_pattern (struct rd_resdict *rd, struct obj *name,
			      struct pt_pattern **pattern, struct ref_owner *p)
{
	struct xref_bind robj;
	struct t_tag tag;
	enum s_status res;
	struct obj obj;

	if (rd == NULL || rd -> pattern.type == OT_NULL)
 		return obj_error (name, "No pattern dictionary");

	res = rd_lookup (rd, name, &rd -> pattern, &obj, &tag, NULL);
	if (res != S_CSC_OK)
		return res;

	init_robj (&robj, rd, NULL);

	res = T_CALL (pt_init, pattern, rd -> t.rdb, &robj, &obj, p);
	obj_destroy (&obj);
	if (res != S_CSC_OK)
		return res;

	(*pattern) -> t.tag = tag;

	return S_CSC_OK;
}

enum s_status rd_get_shading (struct rd_resdict *rd, struct obj *name,
			      struct shd_shading **shd, struct ref_owner *p)
{
	struct xref_bind robj;
	struct t_tag tag;
	enum s_status res;
	struct obj obj;

	if (rd == NULL || rd -> shading.type == OT_NULL)
		return obj_error (name, "No shading dictionary");

	res = rd_lookup (rd, name, &rd -> shading, &obj, &tag, NULL);
	if (res != S_CSC_OK)
		return res;

	init_robj (&robj, rd, NULL);

	res = T_CALL (shd_init, shd, rd -> t.rdb, &robj, &obj, p);
	obj_destroy (&obj);
	if (res != S_CSC_OK)
		return res;

	(*shd) -> t.tag = tag;
	
	return S_CSC_OK;
}

enum s_status rd_get_xobject (struct rd_resdict *rd, struct obj *name,
			      struct oc_properties *ocp, 
			      struct xo_object **xo, struct ref_owner *p)
{
	struct xref_bind robj;
	struct t_tag tag;
	enum s_status res;
	struct obj obj;

	if (rd == NULL || rd -> xobject.type == OT_NULL)
		return obj_error (name, "No xobject dictionary");

	res = rd_lookup (rd, name, &rd -> xobject, &obj, &tag, NULL);
	if (res != S_CSC_OK)
		return res;

	init_robj (&robj, rd, ocp);

	res = T_CALL (xo_init, xo, rd -> t.rdb, &robj, &obj, p);
	obj_destroy (&obj);
	if (res != S_CSC_OK)
		return res;

	(*xo) -> t.tag = tag;
	
	return S_CSC_OK;
}

enum s_status rd_get_font (struct rd_resdict *rd, struct obj *name,
			   struct gf_font **gf, struct ref_owner *p)
{
	struct xref_bind robj;
	enum s_status res;
	struct t_tag tag;
	struct obj obj;

	if (rd == NULL || rd -> font.type == OT_NULL)
		return obj_error (name, "No font dictionary");

	init_robj (&robj, rd, NULL);

	res = rd_lookup (rd, name, &rd -> font, &obj, &tag, NULL);
	if (res != S_CSC_OK)
		return res;

	res = T_CALL (gf_init, gf, rd -> t.rdb, &robj, &obj, p);
	obj_destroy (&obj);
	if (res != S_CSC_OK)
		return res;

	(*gf) -> t.tag = tag;
	
	return S_CSC_OK;
}

