/*
 * PDFlow -- tree.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <string.h>

#include "fmt/fields.h"
#include "types/tree.h"

#define DEFAULT_NODE_SIZE 40

static void t_destroy_node_data (bt_tree_data_t *tdata, 
				 bt_node_data_t *ndata);

static enum s_status t_create_node_data (bt_tree_data_t *tdata, 
					 bt_node_data_t *parent,
					 bt_node_data_t **ndata);

static void t_destroy_data (bt_tree_data_t *tdata, bt_data_t *data);

static int num_compare (bt_key_t *key1, bt_key_t *key2);
static int name_compare (bt_key_t *key1, bt_key_t *key2);

static struct bt_param tree_param = {
	sizeof (struct obj *), 
	DEFAULT_NODE_SIZE,
	NULL,
	NULL,
	NULL,
	t_create_node_data,
	t_destroy_node_data,
	t_destroy_data
};

struct tree_ifc {
	enum obj_type type;
	struct xref_field *field;
	int (*compare) (bt_key_t *key1, bt_key_t *key2);
};

struct tr_data {
	struct obj key;
	struct obj value;
};

struct tr_node {

	struct obj obj;

	struct tr_node *next;

#include "tree.tab.h"

};

#define TR_PAIRS "Nums"
#define TR_FIELD tr_num_tree_field
#define TR_FIELD_DESC tr_num_tree_field_desc
#define XREF_IMPLEMENTATION struct tr_node
#include "tree.tab.h"
#undef XREF_IMPLEMENTATION

#define TR_PAIRS "Names"
#define TR_FIELD tr_name_tree_field
#define TR_FIELD_DESC tr_name_tree_field_desc
#define XREF_IMPLEMENTATION struct tr_node
#include "tree.tab.h"
#undef XREF_IMPLEMENTATION

T_SUGAR_IMPLEMENTATIONS (tr_num, tr_tree)
T_SUGAR_IMPLEMENTATIONS (tr_name, tr_tree)

static struct tree_ifc num_ifc = {
	OT_INTEGER,
	&tr_num_tree_field,
	num_compare
};

static struct tree_ifc name_ifc = {
	OT_NAME,
	&tr_name_tree_field,
	name_compare
};

static struct obj *cast_imp_closure (void *closure)
{
	return closure;
}

static void *uncast_imp_closure (struct obj *obj)
{
	return obj;
}

static struct tr_tree *cast_tree_data (bt_tree_data_t *tdata)
{
	return (struct tr_tree *) tdata;
}

static bt_tree_data_t *uncast_tree_data (struct tr_tree *t)
{
	return (struct bt_imp_data_t *) t;
}

static struct tr_node *cast_node_data (bt_node_data_t *ndata)
{
	return (struct tr_node *) ndata;
}

static bt_node_data_t *uncast_node_data (struct tr_node *node)
{
	return (bt_node_data_t *) node;
}

static struct tr_data *cast_data (bt_data_t *data)
{
	return (struct tr_data *) data;
}

static bt_data_t *uncast_data (struct tr_data *data)
{
	return (bt_data_t *) data;
}

static void add_delnode (struct tr_tree *tree, struct tr_node *node)
{
	node -> next = tree -> delnodes;
	tree -> delnodes = node;
}

static void destroy_node (struct tr_tree *tree, struct tr_node *node)
{
	xref_fields_destroy (tree -> ifc -> field, node,
			     ref_rown (&tree -> t.refs));
	obj_destroy (&node -> obj);
	free (node);
}

static void t_destroy_node_data (bt_tree_data_t *tdata, 
				 bt_node_data_t *ndata)
{
	struct tr_tree *tree;
	struct tr_node *node;

	tree = cast_tree_data (tdata);

	if (ndata != NULL) {
		node = cast_node_data (ndata);
		if (obj_exists (&node -> obj)) 
			add_delnode (tree, node);
		else
			destroy_node (tree, node);
	}
}

static enum s_status t_create_node_data (bt_tree_data_t *tdata, 
					 bt_node_data_t *parent,
					 bt_node_data_t **ndata)
{
	struct tr_node *node;

	node = malloc (sizeof (struct tr_node));
	if (node == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (node, 0, sizeof (struct tr_node));

	*ndata = uncast_node_data (node);

	return S_CSC_OUT_OF_MEMORY;
}

static void t_destroy_data (bt_tree_data_t *tdata, bt_data_t *d)
{
	struct tr_data *data;

	data = cast_data (d);

	obj_destroy (&data -> key);
	obj_destroy (&data -> value);

	free (data);
}

static void tree_unref (struct tr_tree *tree, struct ref_owner *p)
{
	struct tr_node *node;

	if (tree == NULL || ref_down (&tree -> t.refs, p))
		return;

	if (tree -> tree != NULL)
		bt_unref (tree -> tree, ref_rown (&tree -> t.refs));

	while (tree -> delnodes != NULL) {
		node = tree -> delnodes;
		tree -> delnodes = node -> next;
		destroy_node (tree, node);
	}

	xref_unref_db (tree -> t.rdb, ref_rown (&tree -> t.refs));

	obj_destroy (&tree -> t.obj);

	ref_zero (&tree -> t.refs);

	free (tree);
}

static enum s_status import_node (struct xref_db *rdb,
				  struct obj *obj,
				  struct tr_node *node,
				  struct bt_imp_data *imps,
				  struct tr_tree *tree)
{
	imps -> type = BT_NT_INTERMEDIATE;
	imps -> size = XREF_A_SIZE (&node -> kids);
	imps -> ndata = NULL;
	imps -> data = NULL;
	imps -> key = 0;

	return S_CSC_OK;
}

static enum s_status import_leaf (struct xref_db *rdb,
				  struct obj *obj,
				  struct tr_node *node,
				  struct bt_imp_data *imps,
				  struct tr_tree *tree)
{
	struct obj *kobj, *dobj;
	struct tr_data **data;
	struct obj **keys;
	enum s_status res;
	int i;

	if (XREF_A_SIZE (&node -> pairs) & 1)
		return obj_error (obj, "Bad number of elements in pairs");

	imps -> type = BT_NT_LEAF;
	imps -> size = XREF_A_SIZE (&node -> pairs) / 2;
	imps -> ndata = NULL;

	for (i = 0; i < imps -> size; i++) {
		kobj = &node -> pairs.data [i * 2];
		if (kobj -> type != tree -> ifc -> type)
			return obj_error (kobj, "Bad key type");
	}
			
	keys = calloc (imps -> size, sizeof (struct obj *));
	if (keys == NULL)
		return S_CSC_OUT_OF_MEMORY;

	data = calloc (imps -> size, sizeof (struct tr_data *));
	if (data == NULL) {
		free (keys);
		return S_CSC_OUT_OF_MEMORY;
	}

	res = S_CSC_OK;
	for (i = 0; i < imps -> size; i++) {
		kobj = &node -> pairs.data [i * 2];
		dobj = &node -> pairs.data [i * 2 + 1];

		data [i] = malloc (sizeof (struct tr_data));
		if (data [i] == NULL) {
			res = S_CSC_OUT_OF_MEMORY;
			break;
		}

		keys [i] = &data [i] -> key;
		data [i] -> key = *kobj;
		data [i] -> value = *dobj;
		obj_set_null (kobj, &f_null);
		obj_set_null (dobj, &f_null);
	}
	
	if (res != S_CSC_OK) {
		for (i = 0; i < imps -> size; i++) {
			obj_destroy (&data [i] -> key);
			obj_destroy (&data [i] -> value);
		}
		free (keys);
		free (data);

		return res;
	}

	imps -> key = keys;
	imps -> data = (bt_data_t **) data;

	return S_CSC_OK;
}

static enum s_status tree_import_f (bt_tree_data_t *tdata,
				    bt_node_data_t *ndata,
				    unsigned idx,
				    struct bt_imp_data *imps,
				    void *closure)
{
	struct tr_node *node, *parent;
	struct tr_tree *tree;
	struct obj *obj, *root;
	enum s_status res;
	int leaf;

	root = cast_imp_closure (closure);
	parent = cast_node_data (ndata);
	tree = cast_tree_data (tdata);
	
	if (parent != NULL)
		obj = &parent -> kids.data [idx];
	else
		obj = root;

	node = malloc (sizeof (struct tr_node));
	if (node == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (node, 0, sizeof (struct tr_node));

	res = obj_dup (obj, &node -> obj);
	if (res != S_CSC_OK) {
		destroy_node (tree, node);
		return res;
	}

	res = xref_fields_parse (tree -> t.rdb, NULL, 
				 ref_rown (&tree -> t.refs), obj,
				 tree -> ifc -> field, node);
	if (res != S_CSC_OK) {
		destroy_node (tree, node);
		return res;
	}

	leaf = XREF_A_IS_NULL (&node -> kids);

	res = leaf ?
		import_leaf (tree -> t.rdb, obj, node, imps, tree) :
		import_node (tree -> t.rdb, obj, node, imps, tree);

	if (res != S_CSC_OK) {
		destroy_node (tree, node);
		return res;
	}

	imps -> ndata = uncast_node_data (node);

	return res;
}

static enum s_status init_tree (struct tr_tree *tree, struct xref_db *rdb,
				struct obj *obj)
{
	struct bt_imp_info ii;
	struct bt_param param;
	struct ref_owner *p;
	enum s_status res;

	p = ref_rown (&tree -> t.refs);

	param = tree_param;

	param.tdata = uncast_tree_data (tree);
	param.compare = tree -> ifc -> compare;
	res = bt_init (&tree -> tree, &param, p);
	if (res != S_CSC_OK)
		return res;

	if (obj -> type != OT_NULL) {
		ii.import = tree_import_f;

		res = bt_import (tree -> tree, &ii, 
				 uncast_imp_closure (obj));
		if (res != S_CSC_OK)
			return res;
	}

	return S_CSC_OK;
}

static enum s_status tree_init (struct tr_tree **treep, struct xref_db *rdb, 
				struct xref_bind *robj, const struct obj *obj, 
				struct ref_owner *p, struct tree_ifc *ifc,
				const char *fname, int line)
{
	struct tr_tree *tree;
	enum s_status res;

	if (obj != NULL && obj -> type == OT_NULL) {
		*treep = NULL;
		return S_CSC_OK;
	}

	if (obj == NULL)
		obj = &obj_null;

	if (obj -> type != OT_DICTIONARY &&
	    obj -> type != OT_NULL)
		return obj_error (obj, "Bad tree object type");

	tree = malloc (sizeof (struct tr_tree));
	if (tree == NULL)
		return S_CSC_OK;
	memset (tree, 0, sizeof (struct tr_tree));

	ref__init (&tree -> t.refs, "tree", p, fname, line);
	tree -> t.rdb = rdb;
	tree -> ifc = ifc;
	xref_ref_db (rdb, ref_rown (&tree -> t.refs));

	res = obj_dup (obj, &tree -> t.obj);
	if (res != S_CSC_OK) {
		tree_unref (tree, p);
		return res;
	}

	res = init_tree (tree, rdb, &tree -> t.obj);
	if (res != S_CSC_OK) {
		tree_unref (tree, p);
		return res;
	}

	*treep = tree;

	return S_CSC_OK;
}

static enum s_status export (struct xref_obj_array *kids, 
			     struct xref_obj_array *pairs,
			     struct xref_obj_array *limits,
			     struct bt_exp_data *exps)
{
	struct tr_data *data;
	struct tr_node *n;
	enum s_status res;
	int i;

	if (exps -> type == BT_NT_INTERMEDIATE) {
		kids -> data = calloc (exps -> size, sizeof (struct obj));
		if (kids -> data == NULL)
			return S_CSC_OUT_OF_MEMORY;
		kids -> n = exps -> size;

		for (i = 0; i < exps -> size; i++) {
			n = cast_node_data (exps -> child [i]);
			obj_copy_reference (&kids -> data [i], &n -> obj);
		}

		limits -> data = calloc (2, sizeof (struct obj));
		if (limits -> data == NULL)
			return S_CSC_OUT_OF_MEMORY;
		limits -> n = 2;

		res = obj_dup (exps -> lo, &limits -> data [0]);
		if (res != S_CSC_OK)
			return res;

		res = obj_dup (exps -> hi, &limits -> data [1]);
		if (res != S_CSC_OK)
			return res;
	} 

	if (exps -> type == BT_NT_LEAF) {
		pairs -> data = calloc (2 * exps -> size, sizeof (struct obj));
		if (kids -> data == NULL)
			return S_CSC_OUT_OF_MEMORY;
		pairs -> n = 2 * exps -> size;

		for (i = 0; i < exps -> size; i++) {
			data = uncast_data (exps -> data [i]);
			res = obj_dup (&data -> key, 
				       &pairs -> data [2 * i]);
			if (res != S_CSC_OK)
				return res;
			res = obj_dup (&data -> value, 
				       &pairs -> data [2 * i + 1]);
			if (res != S_CSC_OK)
				return res;
		}
	}

	return S_CSC_OK;
}

static enum s_status export_f (bt_tree_data_t *tdata,
			       struct bt_exp_data *exps,
			       va_list ap)
{
	struct xref_db *rdb;
	struct v_selection *sel;
	enum s_status res;
	struct tr_node *node;
	struct tr_tree *tree;
	enum xref_serialize ser;
	struct xref_obj_array kids, pairs, limits;
	enum xref_changed changed;

	rdb = va_arg (ap, struct xref_db *);
	sel = va_arg (ap, struct v_selection *);
	ser = va_arg (ap, enum xref_serialize);

	tree = cast_tree_data (tdata);
	node = cast_node_data (exps -> ndata);
	
	memset (&kids, 0, sizeof (struct xref_obj_array));
	memset (&pairs, 0, sizeof (struct xref_obj_array));
	memset (&limits, 0, sizeof (struct xref_obj_array));

	res = export (&kids, &pairs, &limits, exps);
	if (res != S_CSC_OK) {
		xref_array_map_rel_ps (&kids, NULL, 
				       &xref_obj_pstable, NULL);
		xref_array_map_rel_ps (&pairs, NULL, 
				       &xref_obj_pstable, NULL);
		xref_array_map_rel_ps (&limits, NULL, 
				       &xref_obj_pstable, NULL);
		return res;
	}

	xref_array_map_rel_ps (&node -> kids, NULL, 
			       &xref_obj_pstable, NULL);
	xref_array_map_rel_ps (&node -> pairs, NULL, 
			       &xref_obj_pstable, NULL);
	xref_array_map_rel_ps (&node -> limits, NULL, 
			       &xref_obj_pstable, NULL);

	node -> kids = kids;
	node -> pairs = pairs;
	node -> limits = limits;
	
	return xref_fields_ser (rdb, NULL, sel, &node -> obj, 
				&changed, tree -> ifc -> field, node);
}

enum s_status tree_ser (struct xref_db *rdb, struct xref_bind *robj,
			struct v_selection *sel, struct obj *obj,
			enum xref_changed *changed, struct tr_tree *tree)
{
	struct tr_node *node;
	struct bt_exp_info ei;
	enum s_status res;

	ei.export = export_f;

	res = bt_export (tree -> tree, &ei, rdb, sel, XREF_S_REFERENCE);
	if (res != S_CSC_OK)
		return res;

	if (tree -> tree -> root != NULL) {
		node = cast_node_data (tree -> tree -> root -> ndata);
		res = xref_fields_ser (rdb, robj, sel, &node -> obj, changed, 
				       tree -> ifc -> field, node);
		if (res != S_CSC_OK)
			return res;

		res = obj_dup (&node -> obj, obj);
		if (res != S_CSC_OK)
			return res;
	} else
		return S_CSC_OK;

	while (tree -> delnodes != NULL) {
		node = tree -> delnodes;
		tree -> delnodes = node -> next;
		xref_remove_object (tree -> t.rdb, &node -> obj);
		destroy_node (tree, node);
	}

	return S_CSC_OK;
}

static enum s_status lookup (struct tr_tree *tree, 
			     struct obj *key, struct obj *val)
{
	enum s_status res;
	struct tr_data *data;
	bt_data_t *d;

	res = bt_get (tree -> tree, (bt_key_t *) key, &d);
	if (res != S_CSC_OK)
		return res;

	if (d == NULL) {
		obj_set_null (val, &f_null);
		return S_CSC_OK;
	}

	data = cast_data (d);

	return obj_dup (&data -> value, val);
}

static int num_compare (bt_key_t *key1, bt_key_t *key2)
{
	struct obj *o1, *o2;

	o1 = (struct obj *) key1;
	o2 = (struct obj *) key2;

	return o1 -> val.integer - o2 -> val.integer;
}

enum s_status tr_num_init (struct tr_tree **treep, struct xref_db *rdb,
			   struct xref_bind *robj, const struct obj *obj,
			   struct ref_owner *p, const char *fname,
			   int line)
{
	return tree_init (treep, rdb, robj, obj, p, &num_ifc, fname, line);
}

void tr_num_unref (struct tr_tree *tree, struct ref_owner *p)
{
	tree_unref (tree, p);
}

enum s_status tr_num_sync (struct xref_db *rdb, struct xref_bind *robj,
			   struct v_selection *sel, struct obj *obj,
			   enum xref_changed *changed, struct tr_tree *tree)
{
	return tree_ser (rdb, robj, sel, obj, changed, tree);
}

enum s_status tr_num_lookup (struct tr_tree *tree,
			     pdf_integer_t key, struct obj *val)
{
	struct obj kobj;

	obj_set_integer (&kobj, &f_null, key);

	return lookup (tree, &kobj, val);
}

enum s_status tr_num_walk (struct tr_tree *tree, 
			   struct obj *key, struct obj *val, 
			   struct tr_cursor *pos);

static int name_compare (bt_key_t *key1, bt_key_t *key2)
{
	struct obj *o1, *o2;
	int smaller, v;

	o1 = (struct obj *) key1;
	o2 = (struct obj *) key2;

	smaller = o1 -> val.string.size < o2 -> val.string.size ?
		o1 -> val.string.size : o2  -> val.string.size;

	v = memcmp (o1 -> val.string.s, o2 -> val.string.s, smaller);
	if (!v)
		v = o1 -> val.string.size - o2 -> val.string.size;

	return v;
}

enum s_status tr_name_init (struct tr_tree **treep, struct xref_db *rdb,
			    struct xref_bind *robj, const struct obj *obj,
			    struct ref_owner *p, const char *fname,
			    int line)
{
	return tree_init (treep, rdb, robj, obj, p, &name_ifc, fname, line);
}

void tr_name_unref (struct tr_tree *tree, struct ref_owner *p)
{
	tree_unref (tree, p);
}

enum s_status tr_name_sync (struct xref_db *rdb, struct xref_bind *robj,
			    struct v_selection *sel, struct obj *obj,
			    enum xref_changed *changed, struct tr_tree *tree)
{
	return tree_ser (rdb, robj, sel, obj, changed, tree);
}

enum s_status tr_name_lookup (struct tr_tree *tree,
			      struct obj_bstr key, struct obj *val)
{
	struct obj kobj;

	obj_set_name (&kobj, &f_null, &key);

	return lookup (tree, &kobj, val);
}

