/*
 * PDFlow -- txtstring.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <string.h>

#include "fmt/ctab.h"
#include "types/txtstring.h"

static void ts_funref (struct ref_owner *p, struct xref_field *fd);

struct xref_pstable ts_pstable = {
	sizeof (struct ts_string),
	ts_parser, ts_ser, ts_funref
};

static const char *next_token (const char *p, const char *e,
			       struct ts_token *token)
{
	while (p < e) {
		p = memchr (p, 0, e - p);
		if (p == NULL)
			break;

		p++;

		if (p >= e || *p != 0x1b)
			continue;
		p++;

		if (p + 2 >= e)
			continue;
		if (token != NULL) {
			token -> lang [0] = p [0];
			token -> lang [1] = p [1];
		}
		p += 2;

		if (p + 2 > e)
			continue;
		if (*p) {
			token -> country [0] = p [0];
			token -> country [1] = p [1];
			p += 2;
			
			if (p + 2 > e)
				continue;
		} else {
			token -> country [0] = 0;
			token -> country [1] = 0;
		}			

		if (*p || p [1] != 0x1b)
			continue;

		return p + 2;
	}

	return NULL;
}

static enum s_status init_pdf (struct ts_string *ts, 
			       struct xref_db *rdb, 
			       struct obj_bstr *string)
{
	ts -> encoding = TS_E_PDF_DOC;

	ts -> token = calloc (1, sizeof (struct ts_token));
	if (ts -> token == NULL)
		return S_CSC_OUT_OF_MEMORY;

	ts -> tokens = 1;

	ts -> token [0].s = malloc (string -> size + 1);
	if (ts -> token [0].s == NULL)
		return S_CSC_OUT_OF_MEMORY;
	
	memcpy (ts -> token [0].s, string -> s, string -> size);
	ts -> token [0].s [string -> size] = 0;
	ts -> token [0].len = string -> size;

	return S_CSC_OK;
}


static enum s_status init_utf (struct ts_string *ts, 
			       struct xref_db *rdb, 
			       struct obj_bstr *string)
{
	const char *p, *q, *e;
	size_t size;
	int i;

	ts -> encoding = TS_E_UTF16BE;

	p = string -> s;
	e = string -> s + string -> size;

	i = 0;
	do {
		i++;
		p = next_token (p, e, NULL);
	} while (p != NULL);

	ts -> token = calloc (i, sizeof (struct ts_token));
	if (ts -> token == NULL)
		return S_CSC_OUT_OF_MEMORY;

	ts -> tokens = i;
	
	i = 0;
	p = string -> s;
	do {
		q = next_token (p, e, &ts -> token [i]);
		size = q == NULL ? e - p : q - p;

		ts -> token [i].s = malloc (size);
		if (ts -> token [i].s == NULL)
			return S_CSC_OUT_OF_MEMORY;

		memcpy (ts -> token [i].s, p, size);
		ts -> token [i].len = size;

		p = q;

	} while (p != NULL);

	return S_CSC_OK;
}

enum s_status ts_init (struct ts_string *ts, struct xref_db *rdb,
		       struct obj *obj)
{
	struct obj_bstr string;
	enum s_status res;

	if (obj -> type == OT_NULL)
		return S_CSC_OK;
	else if (obj -> type != OT_STRING)
		return obj_error (obj, "Bad text string object string");

	res = obj_dup (obj, &ts -> obj);
	if (res != S_CSC_OK)
		return res;

	string = obj -> val.string;
	if (string.size >= 2 &&
	    string.s [0] == '\xfe' && string.s [1] == '\xff') {
		string.s += 2;
		string.size -= 2;
		res = init_utf (ts, rdb, &string);
	} else
		res = init_pdf (ts, rdb, &string);

	if (res != S_CSC_OK) {
		obj_destroy (&ts -> obj);
		return res;
	}

	return S_CSC_OK;
}

enum s_status ts_init_c (struct ts_string *ts, const char *s)
{
	struct obj_bstr bs;
	struct obj obj;
	enum s_status res;

	res = obj_string_from_c (s, &bs);
	if (res != S_CSC_OK)
		return res;

	obj_set_string (&obj, &f_null, &bs);
	res = ts_init (ts, NULL, &obj);
	obj_destroy (&obj);

	return res;
}

enum s_status ts_update (struct ts_string *ts, struct hash_status *hs)
{
	enum s_status res;

	if (ts -> obj.type == OT_NULL)
		return S_CSC_OK;

	res = hash_update (hs, ts -> obj.val.string.s, 
			   ts -> obj.val.string.size);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

void ts_destroy (struct ts_string *ts)
{
	int i;

	for (i = 0; i < ts -> tokens; i++) {
		if (ts -> token [i].s != NULL)
			free (ts -> token [i].s);
	}

	if (ts -> token != NULL)
		free (ts -> token);

	obj_destroy (&ts -> obj);
}

enum s_status ts_strip_lang (struct ts_string *src, 
			     struct obj *dest)
{
	char *p, *q, *conv;
	struct obj_bstr s;
	enum s_status res;
	int i, len;

	p = NULL;
	res = S_CSC_OK;
	len = 0;
	for (i = 0; i < src -> tokens; i++) {
		if (src -> encoding == TS_E_UTF16BE)
			res = ct_utf16_to_utf8 (src -> token [i].s,
						src -> token [i].len, &conv);
		else
			res = ct_pdfdoc_to_utf8 (src -> token [i].s, &conv);
		if (res != S_CSC_OK)
			return res;

		if (!*conv) {
			free (conv);
			continue;
		}

		q = realloc (p, len + strlen (conv) + 1);
		if (q == NULL) {
			free (conv);
			res = S_CSC_OUT_OF_MEMORY;
			break;
		}

		p = q;
		strcpy (p + len, conv);
		len += strlen (conv);
		free (conv);
	}

	if (p == NULL) {
		obj_set_null (dest, &src -> obj.ptr);
		return S_CSC_OK;
	}

	if (res != S_CSC_OK) {
		if (p != NULL)
			free (p);
		return res;
	}

	s.s = p;
	s.size = strlen (p);
	
	obj_set_string (dest, &src -> obj.ptr, &s);

	return S_CSC_OK;
}

int ts_is_null (struct ts_string *ts)
{
	return ts -> obj.type == OT_NULL;
}

enum s_status ts_parser (struct xref_db *rdb, struct xref_bind *robj,
			 struct ref_owner *p, struct obj *src,
			 struct xref_field *fd, const char *fname,
			 int line)
{
	return ts_init (fd -> field, rdb, src);
}

enum s_status ts_ser (struct xref_db *rdb, struct xref_bind *robj,
		      struct v_selection *sel, struct obj *obj,
		      enum xref_changed *changed, struct xref_field *fd)
{
	struct ts_string *ts;

	ts = fd -> field;

	return obj_dup (&ts -> obj, obj);
}

static void ts_funref (struct ref_owner *p, struct xref_field *fd)
{
	ts_destroy (fd -> field);
}

enum s_status ts_cross (struct ts_string *ts, struct ts_string *sts,
			struct ob_bridge *ob)
{
	struct ts_string nts;
	struct obj obj;
	enum s_status res;

	memset (&nts, 0, sizeof (struct ts_string));

	if (sts -> obj.type != OT_NULL) {
		res = ob_lookup (ob, &sts -> obj, &obj);
		if (res != S_CSC_OK)
			return res;
	} else
		obj_set_null (&obj, &f_null);

	res = ts_init (&nts, ob -> drdb, &obj);
	obj_destroy (&obj);
	if (res != S_CSC_OK)
		return res;

	*ts = nts;

	return S_CSC_OK;
}

