/*
 * PDFlow -- types.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <assert.h>
#include <string.h>
#include <stdlib.h>

#include "types/types.h"

enum s_status t_mandatory (struct xref_field_desc *desc, void *data)
{
	return S_CSC_OK;
}

enum s_status t_init (struct t_type **tp, const char *name,
		      struct xref_db *rdb, struct xref_bind *robj,
		      const struct obj *obj, struct ref_owner *p, 
		      size_t size, struct xref_field *fd,
		      t_ini ini, t_ini ini0, t_ini ini00, t_fini fini,
		      const char *xkey, const char *fname, int line)
{
	struct xref_cache *cache;
	struct ref_owner *owner;
	struct t_type *t;
	enum s_status res;

	if (obj != NULL && obj -> type == OT_NULL) {
		*tp = NULL;
		return S_CSC_OK;
	}

	if (rdb != NULL) {
		cache = xref_cache_get (rdb, xkey);
		t = xref_cache_find (cache, obj);
		if (t != NULL) {
			t_ref (t, p, fname, line);
			*tp = t;
			return S_CSC_OK;
		}
	} else
		cache = NULL;


	if (obj == NULL) {
		obj = &obj_null;
		if (ini0 != NULL)
			ini = ini0;
	}

	t = malloc (size);
	if (t == NULL)
		return S_CSC_OUT_OF_MEMORY;
	memset (t, 0, size);

	ref__init (&t -> refs, name, p, fname, line);

	t -> rdb = rdb;
	t -> tag.rid = xref_get_id (rdb);
	t -> tag.ref = obj -> ref;
	xref_ref_db (rdb, ref_rown (&t -> refs));
	
	res = obj_dup (obj, &t -> obj);
	if (res == S_CSC_OK && ini00 != NULL)
		res = ini00 (t, rdb, robj, p);
	if (res == S_CSC_OK && fd != NULL)
		res = xref__fields_parse (rdb, robj, ref_rown (&t -> refs),
					  &t -> obj, fd, t, fname, line);
	if (res == S_CSC_OK && ini != NULL)
		res = ini (t, rdb, robj, p);

	if (res != S_CSC_OK) {
		t_unref (t, p, fd, fini);
		return res;
	}

	owner = xref_cache_put (cache, obj, t);
	if (owner != NULL)
		T_CALL (t_ref, t, owner);

	*tp = t;
	
	return S_CSC_OK;
}

void t_ref (struct t_type *t, struct ref_owner *p, 
	    const char *fname, int line)
{
	if (t != NULL)
		ref__up (&t -> refs, p, fname, line);
}

void t_unref (struct t_type *t, struct ref_owner *p, 
	      struct xref_field *fd, t_fini fini)
{
	if (t == NULL)
		return;

	if (ref_down (&t -> refs, p))
		return;

	if (fini != NULL)
		fini (t, p);

	if (fd != NULL)
		xref_fields_destroy (fd, t, ref_rown (&t -> refs));

	xref_unref_db (t -> rdb, ref_rown (&t -> refs));

	t_tag_destroy (&t -> tag);
	obj_destroy (&t -> obj);

	ref_zero (&t -> refs);

	free (t);
}

enum s_status t_sync (struct xref_db *rdb, struct xref_bind *robj,
		      struct v_selection *sel, struct obj *dest, 
		      struct t_type *t, enum xref_changed *changed, 
		      struct xref_field *fd)
{
	enum xref_changed c;
	enum s_status res;
	struct obj *sobj, obj;

	if (t == NULL) {
		obj_set_null (dest, &f_null);
		return S_CSC_OK;
	}

	if (changed == NULL) {
		c = XREF_C_UNKNOWN;
		changed = &c;
	}

	obj_set_null (&obj, &f_null);
	if (fd != NULL) {
		sobj = rdb == t -> rdb ? &t -> obj : &obj;
		res = xref_fields_ser (rdb, robj, sel, sobj, changed, fd, t);
		if (res != S_CSC_OK)
			return res;
	} else {
		*changed = XREF_C_UNCHANGED;
		sobj = &t -> obj;
	}
 
	if (dest == NULL) {
		assert (rdb == t -> rdb);
		res = xref_serialize_obj (rdb, sobj);
	} else
		res = obj_dup (sobj, dest);
	obj_destroy (&obj);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

enum s_status t_tag_dup (struct t_tag *stag, struct t_tag *dtag)
{
	enum s_status res;

	res = obj_string_dup (&stag -> name, &dtag -> name);
	if (res != S_CSC_OK)
		return res;

	dtag -> rid = stag -> rid;
	dtag -> ref = stag -> ref;

	return S_CSC_OK;
}

void t_tag_destroy (struct t_tag *tag)
{
	obj_string_destroy (&tag -> name);
}

int t_tag_is_valid (struct t_tag *tag)
{
	return tag -> ref.oid;
}

int t_tag_same (const struct t_tag *tag1, const struct t_tag *tag2)
{
	return 	tag1 -> rid == tag2 -> rid &&
		tag1 -> ref.gnum == tag2 -> ref.gnum &&
		tag1 -> ref.oid == tag2 -> ref.oid &&
		obj_string_equal_s (&tag1 -> name, &tag2 -> name);
}

enum s_status t_tag_get_checksum (struct t_tag *tag, struct hash *cks)
{
	struct hash_status *ctxt;
	enum s_status res;

	if (tag == NULL) {
		memset (cks, 0, sizeof (struct hash));
		return S_CSC_OK;
	}

	res = hash_init (&ctxt);
	if (res != S_CSC_OK)
		return res;

	res = hash_update (ctxt, &tag -> rid, sizeof (pdf_integer_t));
	if (res != S_CSC_OK)
		return res;

	res = hash_update (ctxt, &tag -> ref, sizeof (struct obj_reference));
	if (res != S_CSC_OK)
		return res;

	if (tag -> name.s != NULL) {
		res = hash_update (ctxt, tag -> name.s, tag -> name.size);
		if (res != S_CSC_OK)
			return res;
	}

	res = hash_final (ctxt, cks);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

void t_cache_unref (struct t_type *type, struct ref_owner *p, 
		    struct t_cache_closure *cc)
{
	t_unref (type, p, cc -> fd, cc -> fini);
}

static void cache_unref (void *ptr, struct ref_owner *p, void *closure)
{
	struct t_cache_closure *cc;

	cc = closure;

	t_cache_unref (ptr, p, closure);
}

enum s_status t_cache_init (struct xref_db *rdb, const char *xkey,
			    struct t_cache_closure *cache_closure)
{
	/* This assertion fails when using SUGAR_IMPLEMENTATIONS.
	 * These objects are so customized that are not cachable */
	assert (cache_closure -> fd != NULL);

	return xref_cache_add (rdb, xkey, cache_unref, cache_closure);
}

void t_cache_destroy (struct xref_db *rdb, const char *xkey)
{
	if (rdb != NULL)
		xref_cache_remove (rdb, xkey);
}
