/*
 * PDFlow -- xobj.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <string.h>

#include "types/xobj.h"
#include "types/colour.h"
#include "fmt/fields.h"

static enum s_status dyn_ifc (enum xref_op op, struct xref_field *fd, 
			      void *data, int *stop);

static struct cobj_enum_desc subtype_desc [] = {
	{ XO_T_NONE, "/Default" },
	{ XO_T_PS, "PS" },
	{ XO_T_FORM, "Form" },
	{ XO_T_IMAGE, "Image" },
	{ -1, NULL }
};

static struct cobj_enum_desc subtype2_desc [] = {
	{ XO_T2_NONE, "/Default" },
	{ XO_T2_PS, "PS" },
	{ -1, NULL }
};

static struct cobj_enum_desc group_subtype_desc [] = {
	{ XO_GS_TRANSPARENCY, "Transparency" },
	{ -1, NULL }
};

#define XREF_IMPLEMENTATION struct xo_image
#define XO_IMAGE
#include "types/xobj.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct xo_alt_image
#define XO_ALT_IMAGE
#include "types/xobj.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct xo_form_group
#define XO_FORM_GROUP
#include "types/xobj.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct xo_form_reference
#define XO_FORM_REFERENCE
#include "types/xobj.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct xo_form
#define XO_FORM
#include "types/xobj.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct xo_ps
#define XO_PS
#include "types/xobj.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct xo_object
#define XO_OBJECT
#include "types/xobj.tab.h"
#undef XREF_IMPLEMENTATION

T_IMPLEMENTATIONS (xo, xo_object, &xo_object_field);
T_IMPLEMENTATIONS (xo_form_group, xo_form_group, &xo_form_group_field);
T_IMPLEMENTATIONS (xo_form_reference, xo_form_reference, 
		   &xo_form_reference_field);
T_IMPLEMENTATIONS (xo_alt_image, xo_alt_image, &xo_alt_image_field);

struct xo_ifc {
	enum xo_subtype canonical_subtype;

	enum xo_subtype subtype;
	enum xo_subtype2 subtype2;

	struct xref_field *fd;

};

static struct xo_ifc ps_field = {
	XO_T_PS, XO_T_PS, XO_T2_NONE, &xo_ps_field
};

static struct xo_ifc ps_alt_field = {
	XO_T_PS, XO_T_FORM, XO_T2_PS, &xo_ps_field
};

static struct xo_ifc ps_form_field = {
	XO_T_FORM, XO_T_FORM, XO_T2_NONE, &xo_form_field
};

static struct xo_ifc ps_image_field = {
	XO_T_IMAGE, XO_T_IMAGE, XO_T2_NONE, &xo_image_field
};

static struct xo_ifc *ifcs [] = {
	&ps_field, &ps_alt_field, &ps_form_field, &ps_image_field,
	NULL
};

static enum s_status dyn_ifc (enum xref_op op, struct xref_field *fd, 
			      void *data, int *stop)
{
	struct xo_object *xo;
	int i;

	xo = data;

	if (xo -> ifc == NULL) {
		for (i = 0; ifcs [i] != NULL; i++)
			if (xo -> subtype.type == ifcs [i] -> subtype &&
			    xo -> subtype2.type == ifcs [i] -> subtype2)
				break;

		xo -> ifc = ifcs [i];
	}

	if (xo -> ifc == NULL) {
		switch (op) {
		case XREF_OP_PARSE:
			return obj_error (&xo -> t.obj, 
					  "Unsupported XObject type");
			break;

		case XREF_OP_SERIALIZE:
			return s_make_indirect 
				(S_CSC_INTERNAL_ERROR,
				 "Trying to serialize unknown XObject type");

		case XREF_OP_DESTROY:
			/* Can only happen if no initialization
			 * has been done, otherwise the interface 
			 * should have already been set
			 */
			*stop = 1;
			return S_CSC_OK;
		}
	}

	*fd = *xo -> ifc -> fd;

	return S_CSC_OK;
}

enum xo_subtype xo_get_subtype (struct xo_object *xo)
{
	return xo -> ifc -> canonical_subtype;
}

